# -*- coding: latin-1 -*-
############################################################################
# Gnocis
# Bjørn Bredesen, 2018-2019
# bjorn.bredesen@ii.uib.no
############################################################################

__name__ = "gnocis"
__version__ = "0.1.0"
__summary__ = (
    "Gnocis is a system for the analysis and the modelling of cis-regulatory DNA sequences."
)
__uri__ = "https://github.com/bjornbredesen/gnocis"
__author__ = "Bjørn Bredesen"
__email__ = "bjorn@bjornbredesen.no"
__license__ = "MIT"
__title__ = "gnocis"
__release__ = True

