/* Generated by Cython 0.29.24 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_24"
#define CYTHON_HEX_VERSION 0x001D18F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__gnocis__regions
#define __PYX_HAVE_API__gnocis__regions
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "gnocis/regions.pyx",
  "gnocis/regions.pxd",
  "stringsource",
  "gnocis/sequences.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6gnocis_9sequences_sequenceLabel;
struct __pyx_obj_6gnocis_9sequences_sequence;
struct __pyx_obj_6gnocis_9sequences_sequences;
struct __pyx_obj_6gnocis_9sequences_sequenceStream;
struct __pyx_obj_6gnocis_9sequences_sequenceGenerator;
struct __pyx_obj_6gnocis_9sequences_MarkovChain;
struct __pyx_obj_6gnocis_9sequences_IID;
struct __pyx_obj_6gnocis_7regions_region;
struct __pyx_obj_6gnocis_7regions_regions;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr;
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr;
struct __pyx_opt_args_6gnocis_9sequences_streamFASTA;
struct __pyx_opt_args_6gnocis_9sequences_stream2bit;
struct __pyx_opt_args_6gnocis_9sequences_getSequenceStreamFromPath;

/* "sequences.pxd":53
 * 
 * # Streams a FASTA file in blocks.
 * cpdef streamFASTA(path, wantBlockSize = *, spacePrune = *, dropChr = *, restrictToSequences = *)             # <<<<<<<<<<<<<<
 * 
 * # Streams a 2bit file in blocks.
 */
struct __pyx_opt_args_6gnocis_9sequences_streamFASTA {
  int __pyx_n;
  PyObject *wantBlockSize;
  PyObject *spacePrune;
  PyObject *dropChr;
  PyObject *restrictToSequences;
};

/* "sequences.pxd":56
 * 
 * # Streams a 2bit file in blocks.
 * cpdef stream2bit(path, wantBlockSize = *, spacePrune = *, dropChr = *, restrictToSequences = *)             # <<<<<<<<<<<<<<
 * 
 * # Gets a sequence stream based on a path, deciding on the format from the path.
 */
struct __pyx_opt_args_6gnocis_9sequences_stream2bit {
  int __pyx_n;
  PyObject *wantBlockSize;
  PyObject *spacePrune;
  PyObject *dropChr;
  PyObject *restrictToSequences;
};

/* "sequences.pxd":59
 * 
 * # Gets a sequence stream based on a path, deciding on the format from the path.
 * cpdef getSequenceStreamFromPath(path, wantBlockSize = *, spacePrune = *, dropChr = *, restrictToSequences = *)             # <<<<<<<<<<<<<<
 * 
 * ###########################################################################
 */
struct __pyx_opt_args_6gnocis_9sequences_getSequenceStreamFromPath {
  int __pyx_n;
  PyObject *wantBlockSize;
  PyObject *spacePrune;
  PyObject *dropChr;
  PyObject *restrictToSequences;
};

/* "sequences.pxd":17
 * 
 * # Represents a sequence label
 * cdef class sequenceLabel:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name
 */
struct __pyx_obj_6gnocis_9sequences_sequenceLabel {
  PyObject_HEAD
  PyObject *name;
  float value;
};


/* "sequences.pxd":27
 * 
 * # Represents and loads DNA sequences.
 * cdef class sequence:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name, seq, path, annotation
 */
struct __pyx_obj_6gnocis_9sequences_sequence {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gnocis_9sequences_sequence *__pyx_vtab;
  PyObject *name;
  PyObject *seq;
  PyObject *path;
  PyObject *annotation;
  struct __pyx_obj_6gnocis_7regions_region *sourceRegion;
  PyObject *cbytes;
  PyObject *cbytesIndexed;
  PyObject *labels;
};


/* "sequences.pxd":37
 * 
 * # Represents a set of DNA sequences.
 * cdef class sequences:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name
 */
struct __pyx_obj_6gnocis_9sequences_sequences {
  PyObject_HEAD
  PyObject *name;
  PyObject *sequences;
};


/* "sequences.pxd":49
 * 
 * # Represents sequence streams.
 * cdef class sequenceStream:             # <<<<<<<<<<<<<<
 * 	pass
 * 
 */
struct __pyx_obj_6gnocis_9sequences_sequenceStream {
  PyObject_HEAD
};


/* "sequences.pxd":65
 * 
 * # Represents sequence streams
 * cdef class sequenceGenerator:             # <<<<<<<<<<<<<<
 * 	pass
 * 
 */
struct __pyx_obj_6gnocis_9sequences_sequenceGenerator {
  PyObject_HEAD
};


/* "sequences.pxd":68
 * 	pass
 * 
 * cdef class MarkovChain(sequenceGenerator):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int degree, pseudoCounts, nGenerated
 */
struct __pyx_obj_6gnocis_9sequences_MarkovChain {
  struct __pyx_obj_6gnocis_9sequences_sequenceGenerator __pyx_base;
  int degree;
  int pseudoCounts;
  int nGenerated;
  bool addReverseComplements;
  bool prepared;
  PyObject *trainingSequences;
  PyObject *spectrum;
  PyObject *initialDistribution;
  PyObject *probspectrum;
  PyObject *comparableSpectrum;
};


/* "sequences.pxd":78
 * 
 * # IID sequence model
 * cdef class IID(sequenceGenerator):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int pseudoCounts, nGenerated
 */
struct __pyx_obj_6gnocis_9sequences_IID {
  struct __pyx_obj_6gnocis_9sequences_sequenceGenerator __pyx_base;
  int pseudoCounts;
  int nGenerated;
  bool addComplements;
  bool prepared;
  PyObject *trainingSequences;
  PyObject *spectrum;
  PyObject *ntDistribution;
};


/* "gnocis/regions.pxd":16
 * 
 * # Represents regions.
 * cdef class region:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str seq
 */
struct __pyx_obj_6gnocis_7regions_region {
  PyObject_HEAD
  PyObject *seq;
  PY_LONG_LONG start;
  PY_LONG_LONG end;
  bool strand;
  float score;
  PyObject *source;
  PyObject *feature;
  PyObject *group;
  PyObject *ext;
  PyObject *markers;
};


/* "gnocis/regions.pxd":27
 * 
 * # Represents region sets. Supports operands '+' for adding regions from two sets with no merging, '|' for merging two sets, and '^' for excluding the regions in the latter set from the first.
 * cdef class regions:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name
 */
struct __pyx_obj_6gnocis_7regions_regions {
  PyObject_HEAD
  PyObject *name;
  PyObject *regions;
};


/* "gnocis/regions.pyx":161
 * 		return self.exclusion(other)
 * 
 * 	def bstr(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:return: Returns a concatenated list of regions formatted as `seq:start..end (strand)`.
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self;
};


/* "gnocis/regions.pyx":166
 * 		:rtype: str
 * 		"""
 * 		return ', '.join( r.bstr() for r in self.regions )             # <<<<<<<<<<<<<<
 * 
 * 	# Returns a random division of the regions
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *__pyx_outer_scope;
  PyObject *__pyx_v_r;
};


/* "gnocis/regions.pyx":431
 * 
 * 	# Calculates the overlap precision to another set
 * 	def nucleotidePrecision(self, other):             # <<<<<<<<<<<<<<
 * 		""" Calculates the nucleotide precision to another set.
 * 
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision {
  PyObject_HEAD
  PyObject *__pyx_v_other;
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self;
};


/* "gnocis/regions.pyx":440
 * 		:rtype: float
 * 		"""
 * 		return sum(len(r) for r in self & other) / sum(len(r) for r in self)             # <<<<<<<<<<<<<<
 * 
 * 	# Calculates the overlap precision to another set
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};

struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/regions.pyx":443
 * 
 * 	# Calculates the overlap precision to another set
 * 	def dummy(self, genome, useSeq = None):             # <<<<<<<<<<<<<<
 * 		""" Generates a dummy region set, with the same region lengths but random positions.
 * 
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy {
  PyObject_HEAD
  PyObject *__pyx_v_seqLens;
};


/* "gnocis/regions.pyx":458
 * 		if useSeq is not None:
 * 			seqLens = { k: seqLens[k] for k in seqLens if k in useSeq }
 * 		genomeLen = sum(seqLens[k] for k in seqLens)             # <<<<<<<<<<<<<<
 * 		rs = self.regions
 * 		if useSeq is not None:
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *__pyx_outer_scope;
  PyObject *__pyx_v_k;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/regions.pyx":675
 * 
 * 	# Prints basic statistics about the set.
 * 	def printStatistics(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		Outputs basic statistics.
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self;
};


/* "gnocis/regions.pyx":681
 * 		print('Region set statistics - ' + self.name)
 * 		print(' - Regions: %d'%len(self.regions))
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))             # <<<<<<<<<<<<<<
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *__pyx_outer_scope;
  PyObject *__pyx_v_r;
};


/* "gnocis/regions.pyx":683
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))             # <<<<<<<<<<<<<<
 * 			print(' - Total length: %d nt'%( sum( len(r) for r in self.regions ) ))
 * 
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};

struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};

struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "gnocis/regions.pyx":684
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))
 * 			print(' - Total length: %d nt'%( sum( len(r) for r in self.regions ) ))             # <<<<<<<<<<<<<<
 * 
 * # Loads regions from a BED file.
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "gnocis/regions.pyx":726
 * 
 * # Loads regions from a gzipped BED file.
 * def loadBEDGZ(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ {
  PyObject_HEAD
  PyObject *__pyx_v_f;
};


/* "gnocis/regions.pyx":744
 * 	with gzip.open(path, 'rb') as f:
 * 		rgns = []
 * 		for l in (y.decode('utf-8') for y in f):             # <<<<<<<<<<<<<<
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *__pyx_outer_scope;
  PyObject *__pyx_v_y;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/regions.pyx":788
 * 
 * # Loads regions from a gzipped GFF file.
 * def loadGFFGZ(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ {
  PyObject_HEAD
  PyObject *__pyx_v_f;
};


/* "gnocis/regions.pyx":798
 * 	"""
 * 	with gzip.open(path, 'rb') as f:
 * 		raw_gff = [ y for y in (x.split('\t') for x in (y.decode('utf-8') for y in f) if not x.startswith('#')) if len(y)==9 ]             # <<<<<<<<<<<<<<
 * 		rgns = [ region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) for r in raw_gff ]
 * 	rs = regions('GFF GZ file: ' + path, rgns)
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *__pyx_outer_scope;
  PyObject *__pyx_v_genexpr;
  PyObject *__pyx_v_x;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};

struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *__pyx_outer_scope;
  PyObject *__pyx_v_y;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/regions.pyx":1072
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)), measureName = 'Nucleotide precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)             # <<<<<<<<<<<<<<
 * 
 * def nucleotideRegionF1Barplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 {
  PyObject_HEAD
  PyObject *__pyx_v_ps;
  PyObject *__pyx_v_rs;
};

struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};

struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/regions.pyx":1105
 * 	:type showValues: bool, optional
 * 	"""
 * 	def measure(ps, rs):             # <<<<<<<<<<<<<<
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)
 * 		r = len(rs.overlap(ps)) / len(rs)
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure {
  PyObject_HEAD
  PyObject *__pyx_v_ps;
  PyObject *__pyx_v_rs;
};


/* "gnocis/regions.pyx":1106
 * 	"""
 * 	def measure(ps, rs):
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)             # <<<<<<<<<<<<<<
 * 		r = len(rs.overlap(ps)) / len(rs)
 * 		return 2 * (p*r) / (p+r)
 */
struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};

struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};



/* "sequences.pxd":27
 * 
 * # Represents and loads DNA sequences.
 * cdef class sequence:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name, seq, path, annotation
 */

struct __pyx_vtabstruct_6gnocis_9sequences_sequence {
  PyObject *(*getBytes)(struct __pyx_obj_6gnocis_9sequences_sequence *);
  PyObject *(*getBytesIndexed)(struct __pyx_obj_6gnocis_9sequences_sequence *);
};
static struct __pyx_vtabstruct_6gnocis_9sequences_sequence *__pyx_vtabptr_6gnocis_9sequences_sequence;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyIntFromDouble.proto */
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE PyObject* __Pyx_PyInt_FromDouble(double value);
#else
#define __Pyx_PyInt_FromDouble(value) PyLong_FromDouble(value)
#endif

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_PY_LONG_LONG(PY_LONG_LONG value, Py_ssize_t width, char padding_char, char format_char);

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* py_dict_keys.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyObjectFormat.proto */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* s, PyObject* f);
#else
#define __Pyx_PyObject_Format(s, f) PyObject_Format(s, f)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* unicode_tailmatch.proto */
static int __Pyx_PyUnicode_Tailmatch(
    PyObject* s, PyObject* substr, Py_ssize_t start, Py_ssize_t end, int direction);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value);

/* CIntFromPy.proto */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libcpp' */

/* Module declarations from 'gnocis.sequences' */
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequenceLabel = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequence = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequences = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequenceStream = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequenceGenerator = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_MarkovChain = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_IID = 0;
static PyObject *(*__pyx_f_6gnocis_9sequences_streamFASTA)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6gnocis_9sequences_streamFASTA *__pyx_optional_args); /*proto*/
static PyObject *(*__pyx_f_6gnocis_9sequences_stream2bit)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6gnocis_9sequences_stream2bit *__pyx_optional_args); /*proto*/
static PyObject *(*__pyx_f_6gnocis_9sequences_getSequenceStreamFromPath)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6gnocis_9sequences_getSequenceStreamFromPath *__pyx_optional_args); /*proto*/

/* Module declarations from 'gnocis.regions' */
static PyTypeObject *__pyx_ptype_6gnocis_7regions_region = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions_regions = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct__bstr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_3_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_4_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_5_dummy = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_6_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_7_printStatistics = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_8_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_9_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_10_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_11_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_12_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_14_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_16_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_17_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_19_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_20_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_21_measure = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_22_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions___pyx_scope_struct_23_genexpr = 0;
static PyObject *__pyx_f_6gnocis_7regions___pyx_unpickle_region__set_state(struct __pyx_obj_6gnocis_7regions_region *, PyObject *); /*proto*/
static PyObject *__pyx_f_6gnocis_7regions___pyx_unpickle_regions__set_state(struct __pyx_obj_6gnocis_7regions_regions *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "gnocis.regions"
extern int __pyx_module_is_main_gnocis__regions;
int __pyx_module_is_main_gnocis__regions = 0;

/* Implementation of 'gnocis.regions' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_zip;
static const char __pyx_k_[] = ".";
static const char __pyx_k_1[] = "\t.\t1\n";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_2f[] = ".2f";
static const char __pyx_k_CI[] = "CI";
static const char __pyx_k__3[] = ":";
static const char __pyx_k__4[] = "..";
static const char __pyx_k__5[] = " (";
static const char __pyx_k__6[] = "+";
static const char __pyx_k__7[] = "-";
static const char __pyx_k__8[] = ")";
static const char __pyx_k__9[] = "; ";
static const char __pyx_k_ax[] = "ax";
static const char __pyx_k_bp[] = " bp)";
static const char __pyx_k_bw[] = "bw";
static const char __pyx_k_ci[] = "ci";
static const char __pyx_k_cv[] = "cv";
static const char __pyx_k_cx[] = "cx";
static const char __pyx_k_io[] = "io";
static const char __pyx_k_nt[] = " nt)";
static const char __pyx_k_ps[] = "ps";
static const char __pyx_k_rb[] = "rb";
static const char __pyx_k_rs[] = "rs";
static const char __pyx_k_End[] = "End";
static const char __pyx_k_Seq[] = "Seq.";
static const char __pyx_k__10[] = ")>";
static const char __pyx_k__11[] = " + ";
static const char __pyx_k__12[] = ", ";
static const char __pyx_k__13[] = "";
static const char __pyx_k__14[] = " | ";
static const char __pyx_k__15[] = " & ";
static const char __pyx_k__16[] = " ^ ";
static const char __pyx_k__17[] = " ";
static const char __pyx_k__19[] = "\t";
static const char __pyx_k__21[] = "\n";
static const char __pyx_k__22[] = "#";
static const char __pyx_k__27[] = "*";
static const char __pyx_k__28[] = " +/- ";
static const char __pyx_k__29[] = " %";
static const char __pyx_k__30[] = "_";
static const char __pyx_k_all[] = "all";
static const char __pyx_k_bar[] = "bar";
static const char __pyx_k_bio[] = "bio";
static const char __pyx_k_chr[] = "chr";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_ext[] = "ext";
static const char __pyx_k_fig[] = "fig";
static const char __pyx_k_flt[] = "flt";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_loc[] = "loc";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_plt[] = "plt";
static const char __pyx_k_png[] = "png";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_prs[] = "prs";
static const char __pyx_k_psi[] = "psi";
static const char __pyx_k_rsi[] = "rsi";
static const char __pyx_k_seq[] = "seq";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_vCI[] = "vCI";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_2f_2[] = "%.2f";
static const char __pyx_k_HTML[] = "HTML";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bstr[] = "bstr";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_gzip[] = "gzip";
static const char __pyx_k_html[] = "html";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mean[] = "mean";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_ps_2[] = "_ps";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_rgns[] = "rgns";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_text[] = "text";
static const char __pyx_k_yerr[] = "yerr";
static const char __pyx_k_Start[] = "Start";
static const char __pyx_k_Table[] = "Table: ";
static const char __pyx_k_align[] = "align";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_dummy[] = " (dummy)";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_group[] = "group";
static const char __pyx_k_label[] = "label";
static const char __pyx_k_lower[] = "lower";
static const char __pyx_k_merge[] = "merge";
static const char __pyx_k_mtick[] = "mtick";
static const char __pyx_k_parts[] = "parts";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ratio[] = "ratio";
static const char __pyx_k_score[] = "score";
static const char __pyx_k_slots[] = "__slots__";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_style[] = "style";
static const char __pyx_k_table[] = "table";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_vMean[] = "vMean";
static const char __pyx_k_width[] = "width";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_yaxis[] = "yaxis";
static const char __pyx_k_Length[] = "Length";
static const char __pyx_k_Region[] = "Region<";
static const char __pyx_k_Strand[] = "Strand";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_base64[] = "base64";
static const char __pyx_k_bottom[] = "bottom";
static const char __pyx_k_center[] = "center";
static const char __pyx_k_choice[] = "choice";
static const char __pyx_k_common[] = "common";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_genome[] = "genome";
static const char __pyx_k_ggplot[] = "ggplot";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_legend[] = "legend";
static const char __pyx_k_models[] = "models";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_nt_max[] = " nt - max.: ";
static const char __pyx_k_nt_min[] = " nt (min.: ";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_region[] = "region";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_strand[] = "strand";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_useSeq[] = "useSeq";
static const char __pyx_k_xticks[] = "xticks";
static const char __pyx_k_yticks[] = "yticks";
static const char __pyx_k_BytesIO[] = "BytesIO";
static const char __pyx_k_Score_f[] = "Score: %f";
static const char __pyx_k_as_dict[] = "_as_dict_";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_display[] = "display";
static const char __pyx_k_dropChr[] = "dropChr";
static const char __pyx_k_dummy_2[] = "dummy";
static const char __pyx_k_encoded[] = "encoded";
static const char __pyx_k_feature[] = "feature";
static const char __pyx_k_figsize[] = "figsize";
static const char __pyx_k_fltName[] = "fltName";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_ioputil[] = "ioputil";
static const char __pyx_k_loadBED[] = "loadBED";
static const char __pyx_k_loadGFF[] = "loadGFF";
static const char __pyx_k_markers[] = "markers";
static const char __pyx_k_measure[] = "measure";
static const char __pyx_k_nctable[] = "nctable";
static const char __pyx_k_outpath[] = "outpath";
static const char __pyx_k_overlap[] = "overlap";
static const char __pyx_k_raw_gff[] = "raw_gff";
static const char __pyx_k_regions[] = "regions";
static const char __pyx_k_repeats[] = "repeats";
static const char __pyx_k_savefig[] = "savefig";
static const char __pyx_k_scaling[] = "scaling";
static const char __pyx_k_shuffle[] = "shuffle";
static const char __pyx_k_split_A[] = " (split A)";
static const char __pyx_k_split_B[] = " (split B)";
static const char __pyx_k_statfun[] = "statfun";
static const char __pyx_k_uniform[] = "uniform";
static const char __pyx_k_BED_file[] = "BED file: ";
static const char __pyx_k_GFF_file[] = "GFF file: ";
static const char __pyx_k_Source_s[] = "Source: %s";
static const char __pyx_k_fancybox[] = "fancybox";
static const char __pyx_k_fontsize[] = "fontsize";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_getvalue[] = "getvalue";
static const char __pyx_k_positive[] = "positive";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_rotation[] = "rotation";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_subplots[] = "subplots";
static const char __pyx_k_Feature_s[] = "Feature: %s";
static const char __pyx_k_Regions_d[] = " - Regions: %d";
static const char __pyx_k_b64encode[] = "b64encode";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_exclusion[] = "exclusion";
static const char __pyx_k_isPercent[] = "isPercent";
static const char __pyx_k_legendLoc[] = "legendLoc";
static const char __pyx_k_loadBEDGZ[] = "loadBEDGZ";
static const char __pyx_k_loadGFFGZ[] = "loadGFFGZ";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_regions_2[] = " regions)";
static const char __pyx_k_repr_html[] = "_repr_html_";
static const char __pyx_k_sequences[] = "sequences";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_recentered[] = " (recentered - ";
static const char __pyx_k_regionSets[] = "regionSets";
static const char __pyx_k_returnHTML[] = "returnHTML";
static const char __pyx_k_set_xticks[] = "set_xticks";
static const char __pyx_k_set_ylabel[] = "set_ylabel";
static const char __pyx_k_showLegend[] = "showLegend";
static const char __pyx_k_showValues[] = "showValues";
static const char __pyx_k_startswith[] = "startswith";
static const char __pyx_k_upper_left[] = "upper left";
static const char __pyx_k_GFF_GZ_file[] = "GFF GZ file: ";
static const char __pyx_k_Mean_length[] = " - Mean length: ";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_Predictions[] = "Predictions";
static const char __pyx_k_Regions_set[] = "Regions set: ";
static const char __pyx_k_initialSort[] = "initialSort";
static const char __pyx_k_mainRegions[] = "mainRegions";
static const char __pyx_k_measureName[] = "measureName";
static const char __pyx_k_testRegions[] = "testRegions";
static const char __pyx_k_Region_set_s[] = "Region set<%s>";
static const char __pyx_k_bboxAnchorTo[] = "bboxAnchorTo";
static const char __pyx_k_fontsizeAxis[] = "fontsizeAxis";
static const char __pyx_k_intersection[] = "intersection";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_sourceRegion[] = "sourceRegion";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_tight_layout[] = "tight_layout";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_regionBarplot[] = "regionBarplot";
static const char __pyx_k_bbox_to_anchor[] = "bbox_to_anchor";
static const char __pyx_k_fontsizeLabels[] = "fontsizeLabels";
static const char __pyx_k_fontsizeLegend[] = "fontsizeLegend";
static const char __pyx_k_fontsizeValues[] = "fontsizeValues";
static const char __pyx_k_gnocis_regions[] = "gnocis.regions";
static const char __pyx_k_hasTestRegions[] = "hasTestRegions";
static const char __pyx_k_predictionSets[] = "predictionSets";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_sequenceLengths[] = "sequenceLengths";
static const char __pyx_k_set_xticklabels[] = "set_xticklabels";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_PercentFormatter[] = "PercentFormatter";
static const char __pyx_k_extractSequences[] = "extractSequences";
static const char __pyx_k_overlapping_with[] = " overlapping with ";
static const char __pyx_k_Overlap_precision[] = "Overlap precision";
static const char __pyx_k_Total_length_d_nt[] = " - Total length: %d nt";
static const char __pyx_k_matplotlib_pyplot[] = "matplotlib.pyplot";
static const char __pyx_k_matplotlib_ticker[] = "matplotlib.ticker";
static const char __pyx_k_predictionBarplot[] = "predictionBarplot";
static const char __pyx_k_verticalalignment[] = "verticalalignment";
static const char __pyx_k_CVModelPredictions[] = "CVModelPredictions";
static const char __pyx_k_Region_sequences_s[] = " - Region sequences: %s";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_gnocis_regions_pyx[] = "gnocis/regions.pyx";
static const char __pyx_k_loadCoordinateList[] = "loadCoordinateList";
static const char __pyx_k_sort_locals_lambda[] = "sort.<locals>.<lambda>";
static const char __pyx_k_Overlap_sensitivity[] = "Overlap sensitivity";
static const char __pyx_k_bootstrap_sample_of[] = " (bootstrap sample of ";
static const char __pyx_k_bstr_locals_genexpr[] = "bstr.<locals>.genexpr";
static const char __pyx_k_horizontalalignment[] = "horizontalalignment";
static const char __pyx_k_merge_locals_lambda[] = "merge.<locals>.<lambda>";
static const char __pyx_k_pyx_unpickle_region[] = "__pyx_unpickle_region";
static const char __pyx_k_set_major_formatter[] = "set_major_formatter";
static const char __pyx_k_Coordinate_list_file[] = "Coordinate list file: ";
static const char __pyx_k_IPython_core_display[] = "IPython.core.display";
static const char __pyx_k_Nucleotide_precision[] = "Nucleotide precision";
static const char __pyx_k_dummy_locals_genexpr[] = "dummy.<locals>.genexpr";
static const char __pyx_k_from_sequence_stream[] = " - from sequence stream: ";
static const char __pyx_k_not_overlapping_with[] = " not overlapping with ";
static const char __pyx_k_pyx_unpickle_regions[] = "__pyx_unpickle_regions";
static const char __pyx_k_BED_file_syntax_error[] = "BED file syntax error";
static const char __pyx_k_Region_set_statistics[] = "Region set statistics - ";
static const char __pyx_k_overlapPrecisionBarplot[] = "overlapPrecisionBarplot";
static const char __pyx_k_loadBEDGZ_locals_genexpr[] = "loadBEDGZ.<locals>.genexpr";
static const char __pyx_k_loadGFFGZ_locals_genexpr[] = "loadGFFGZ.<locals>.genexpr";
static const char __pyx_k_nucleotideRegionF1Barplot[] = "nucleotideRegionF1Barplot";
static const char __pyx_k_overlapSensitivityBarplot[] = "overlapSensitivityBarplot";
static const char __pyx_k_Nucleotide_region_F1_score[] = "Nucleotide/region F1 score";
static const char __pyx_k_nucleotidePrecisionBarplot[] = "nucleotidePrecisionBarplot";
static const char __pyx_k_Coordinate_list_syntax_error[] = "Coordinate list syntax error";
static const char __pyx_k_printStatistics_locals_genexpr[] = "printStatistics.<locals>.genexpr";
static const char __pyx_k_img_src_data_image_png_base64_s[] = "<img src='data:image/png;base64,%s'>";
static const char __pyx_k_loadGFFGZ_locals_genexpr_locals[] = "loadGFFGZ.<locals>.genexpr.<locals>.genexpr";
static const char __pyx_k_predictionBarplot_locals_lambda[] = "predictionBarplot.<locals>.<lambda>";
static const char __pyx_k_Incompatible_checksums_s_vs_0x11[] = "Incompatible checksums (%s vs 0x11eeeec = (end, ext, feature, group, markers, score, seq, source, start, strand))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x9d[] = "Incompatible checksums (%s vs 0x9d99455 = (name, regions))";
static const char __pyx_k_nucleotidePrecisionBarplot_local[] = "nucleotidePrecisionBarplot.<locals>.<lambda>.<locals>.genexpr";
static const char __pyx_k_nucleotidePrecision_locals_genex[] = "nucleotidePrecision.<locals>.genexpr";
static const char __pyx_k_nucleotideRegionF1Barplot_locals[] = "nucleotideRegionF1Barplot.<locals>.measure.<locals>.genexpr";
static const char __pyx_k_overlapPrecisionBarplot_locals_l[] = "overlapPrecisionBarplot.<locals>.<lambda>";
static const char __pyx_k_overlapSensitivityBarplot_locals[] = "overlapSensitivityBarplot.<locals>.<lambda>";
static const char __pyx_k_nucleotidePrecisionBarplot_local_2[] = "nucleotidePrecisionBarplot.<locals>.<lambda>";
static const char __pyx_k_nucleotideRegionF1Barplot_locals_2[] = "nucleotideRegionF1Barplot.<locals>.measure";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_kp_u_1;
static PyObject *__pyx_kp_u_2f;
static PyObject *__pyx_kp_u_2f_2;
static PyObject *__pyx_kp_u_BED_file;
static PyObject *__pyx_kp_u_BED_file_syntax_error;
static PyObject *__pyx_n_s_BytesIO;
static PyObject *__pyx_n_s_CI;
static PyObject *__pyx_n_s_CVModelPredictions;
static PyObject *__pyx_kp_u_Coordinate_list_file;
static PyObject *__pyx_kp_u_Coordinate_list_syntax_error;
static PyObject *__pyx_n_u_End;
static PyObject *__pyx_kp_u_Feature_s;
static PyObject *__pyx_kp_u_GFF_GZ_file;
static PyObject *__pyx_kp_u_GFF_file;
static PyObject *__pyx_n_s_HTML;
static PyObject *__pyx_n_s_IPython_core_display;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x11;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x9d;
static PyObject *__pyx_n_u_Length;
static PyObject *__pyx_kp_u_Mean_length;
static PyObject *__pyx_kp_u_None;
static PyObject *__pyx_kp_u_Nucleotide_precision;
static PyObject *__pyx_kp_u_Nucleotide_region_F1_score;
static PyObject *__pyx_kp_u_Overlap_precision;
static PyObject *__pyx_kp_u_Overlap_sensitivity;
static PyObject *__pyx_n_s_PercentFormatter;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_u_Predictions;
static PyObject *__pyx_kp_u_Region;
static PyObject *__pyx_kp_u_Region_sequences_s;
static PyObject *__pyx_kp_u_Region_set_s;
static PyObject *__pyx_kp_u_Region_set_statistics;
static PyObject *__pyx_kp_u_Regions_d;
static PyObject *__pyx_kp_u_Regions_set;
static PyObject *__pyx_kp_u_Score_f;
static PyObject *__pyx_kp_u_Seq;
static PyObject *__pyx_kp_u_Source_s;
static PyObject *__pyx_n_u_Start;
static PyObject *__pyx_n_u_Strand;
static PyObject *__pyx_kp_u_Table;
static PyObject *__pyx_kp_u_Total_length_d_nt;
static PyObject *__pyx_kp_u__10;
static PyObject *__pyx_kp_u__11;
static PyObject *__pyx_kp_u__12;
static PyObject *__pyx_kp_u__13;
static PyObject *__pyx_kp_u__14;
static PyObject *__pyx_kp_u__15;
static PyObject *__pyx_kp_u__16;
static PyObject *__pyx_kp_u__17;
static PyObject *__pyx_kp_u__19;
static PyObject *__pyx_kp_u__21;
static PyObject *__pyx_kp_u__22;
static PyObject *__pyx_n_s__27;
static PyObject *__pyx_kp_u__28;
static PyObject *__pyx_kp_u__29;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_n_s__30;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_kp_u__5;
static PyObject *__pyx_kp_u__6;
static PyObject *__pyx_kp_u__7;
static PyObject *__pyx_kp_u__8;
static PyObject *__pyx_kp_u__9;
static PyObject *__pyx_n_s_align;
static PyObject *__pyx_n_u_all;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_as_dict;
static PyObject *__pyx_n_s_ax;
static PyObject *__pyx_n_s_b64encode;
static PyObject *__pyx_n_s_bar;
static PyObject *__pyx_n_s_base64;
static PyObject *__pyx_n_s_bboxAnchorTo;
static PyObject *__pyx_n_s_bbox_to_anchor;
static PyObject *__pyx_n_s_bio;
static PyObject *__pyx_kp_u_bootstrap_sample_of;
static PyObject *__pyx_n_u_bottom;
static PyObject *__pyx_kp_u_bp;
static PyObject *__pyx_n_s_bstr;
static PyObject *__pyx_n_s_bstr_locals_genexpr;
static PyObject *__pyx_n_s_bw;
static PyObject *__pyx_n_u_center;
static PyObject *__pyx_n_s_choice;
static PyObject *__pyx_n_u_chr;
static PyObject *__pyx_n_s_ci;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_common;
static PyObject *__pyx_n_s_context;
static PyObject *__pyx_n_s_cv;
static PyObject *__pyx_n_s_cx;
static PyObject *__pyx_n_u_d;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_display;
static PyObject *__pyx_n_s_dropChr;
static PyObject *__pyx_kp_u_dummy;
static PyObject *__pyx_n_s_dummy_2;
static PyObject *__pyx_n_s_dummy_locals_genexpr;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_encoded;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_u_end;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_exclusion;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_u_ext;
static PyObject *__pyx_n_s_extractSequences;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_u_f;
static PyObject *__pyx_n_s_fancybox;
static PyObject *__pyx_n_s_feature;
static PyObject *__pyx_n_u_feature;
static PyObject *__pyx_n_s_fig;
static PyObject *__pyx_n_s_figsize;
static PyObject *__pyx_n_s_flt;
static PyObject *__pyx_n_s_fltName;
static PyObject *__pyx_n_s_fontsize;
static PyObject *__pyx_n_s_fontsizeAxis;
static PyObject *__pyx_n_s_fontsizeLabels;
static PyObject *__pyx_n_s_fontsizeLegend;
static PyObject *__pyx_n_s_fontsizeValues;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_kp_u_from_sequence_stream;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_genome;
static PyObject *__pyx_n_u_genome;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_getvalue;
static PyObject *__pyx_n_u_ggplot;
static PyObject *__pyx_n_s_gnocis_regions;
static PyObject *__pyx_kp_s_gnocis_regions_pyx;
static PyObject *__pyx_n_s_group;
static PyObject *__pyx_n_u_group;
static PyObject *__pyx_n_s_gzip;
static PyObject *__pyx_n_s_hasTestRegions;
static PyObject *__pyx_n_s_horizontalalignment;
static PyObject *__pyx_n_s_html;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_kp_u_img_src_data_image_png_base64_s;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_initialSort;
static PyObject *__pyx_n_s_intersection;
static PyObject *__pyx_n_s_io;
static PyObject *__pyx_n_s_ioputil;
static PyObject *__pyx_n_s_isPercent;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_u_l;
static PyObject *__pyx_n_s_label;
static PyObject *__pyx_n_s_legend;
static PyObject *__pyx_n_s_legendLoc;
static PyObject *__pyx_n_s_loadBED;
static PyObject *__pyx_n_s_loadBEDGZ;
static PyObject *__pyx_n_s_loadBEDGZ_locals_genexpr;
static PyObject *__pyx_n_s_loadCoordinateList;
static PyObject *__pyx_n_s_loadGFF;
static PyObject *__pyx_n_s_loadGFFGZ;
static PyObject *__pyx_n_s_loadGFFGZ_locals_genexpr;
static PyObject *__pyx_n_s_loadGFFGZ_locals_genexpr_locals;
static PyObject *__pyx_n_s_loc;
static PyObject *__pyx_n_s_lower;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mainRegions;
static PyObject *__pyx_n_u_markers;
static PyObject *__pyx_n_s_matplotlib_pyplot;
static PyObject *__pyx_n_s_matplotlib_ticker;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_mean;
static PyObject *__pyx_n_s_measure;
static PyObject *__pyx_n_s_measureName;
static PyObject *__pyx_n_s_merge;
static PyObject *__pyx_n_s_merge_locals_lambda;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_models;
static PyObject *__pyx_n_s_mtick;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_nctable;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_u_not_overlapping_with;
static PyObject *__pyx_kp_u_nt;
static PyObject *__pyx_kp_u_nt_max;
static PyObject *__pyx_kp_u_nt_min;
static PyObject *__pyx_n_s_nucleotidePrecisionBarplot;
static PyObject *__pyx_n_s_nucleotidePrecisionBarplot_local;
static PyObject *__pyx_n_s_nucleotidePrecisionBarplot_local_2;
static PyObject *__pyx_n_s_nucleotidePrecision_locals_genex;
static PyObject *__pyx_n_s_nucleotideRegionF1Barplot;
static PyObject *__pyx_n_s_nucleotideRegionF1Barplot_locals;
static PyObject *__pyx_n_s_nucleotideRegionF1Barplot_locals_2;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_outpath;
static PyObject *__pyx_n_s_overlap;
static PyObject *__pyx_n_s_overlapPrecisionBarplot;
static PyObject *__pyx_n_s_overlapPrecisionBarplot_locals_l;
static PyObject *__pyx_n_s_overlapSensitivityBarplot;
static PyObject *__pyx_n_s_overlapSensitivityBarplot_locals;
static PyObject *__pyx_kp_u_overlapping_with;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_parts;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_plt;
static PyObject *__pyx_n_u_png;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_positive;
static PyObject *__pyx_n_s_predictionBarplot;
static PyObject *__pyx_n_s_predictionBarplot_locals_lambda;
static PyObject *__pyx_n_s_predictionSets;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_printStatistics_locals_genexpr;
static PyObject *__pyx_n_s_prs;
static PyObject *__pyx_n_s_ps;
static PyObject *__pyx_n_s_ps_2;
static PyObject *__pyx_n_s_psi;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_region;
static PyObject *__pyx_n_s_pyx_unpickle_regions;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_n_u_r;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_ratio;
static PyObject *__pyx_n_s_raw_gff;
static PyObject *__pyx_n_u_rb;
static PyObject *__pyx_kp_u_recentered;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_region;
static PyObject *__pyx_n_s_regionBarplot;
static PyObject *__pyx_n_s_regionSets;
static PyObject *__pyx_n_s_regions;
static PyObject *__pyx_kp_u_regions_2;
static PyObject *__pyx_n_s_repeats;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_repr_html;
static PyObject *__pyx_n_s_returnHTML;
static PyObject *__pyx_n_s_rgns;
static PyObject *__pyx_n_s_rotation;
static PyObject *__pyx_n_s_rs;
static PyObject *__pyx_n_s_rsi;
static PyObject *__pyx_n_s_savefig;
static PyObject *__pyx_n_s_scaling;
static PyObject *__pyx_n_s_score;
static PyObject *__pyx_n_u_score;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_seq;
static PyObject *__pyx_n_u_seq;
static PyObject *__pyx_n_s_sequenceLengths;
static PyObject *__pyx_n_s_sequences;
static PyObject *__pyx_n_s_set_major_formatter;
static PyObject *__pyx_n_s_set_xticklabels;
static PyObject *__pyx_n_s_set_xticks;
static PyObject *__pyx_n_s_set_ylabel;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_showLegend;
static PyObject *__pyx_n_s_showValues;
static PyObject *__pyx_n_s_shuffle;
static PyObject *__pyx_n_s_slots;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_sort_locals_lambda;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_source;
static PyObject *__pyx_n_u_source;
static PyObject *__pyx_n_s_sourceRegion;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_kp_u_split_A;
static PyObject *__pyx_kp_u_split_B;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_u_start;
static PyObject *__pyx_n_s_startswith;
static PyObject *__pyx_n_s_statfun;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_strand;
static PyObject *__pyx_n_u_strand;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_style;
static PyObject *__pyx_n_s_subplots;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_table;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_testRegions;
static PyObject *__pyx_n_s_text;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tight_layout;
static PyObject *__pyx_n_s_uniform;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_kp_u_upper_left;
static PyObject *__pyx_n_s_useSeq;
static PyObject *__pyx_kp_u_utf_8;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_vCI;
static PyObject *__pyx_n_s_vMean;
static PyObject *__pyx_n_s_verticalalignment;
static PyObject *__pyx_n_u_w;
static PyObject *__pyx_n_s_width;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_xticks;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_yaxis;
static PyObject *__pyx_n_s_yerr;
static PyObject *__pyx_n_s_yticks;
static PyObject *__pyx_n_s_zip;
static int __pyx_pf_6gnocis_7regions_6region___init__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_seq, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_strand, PyObject *__pyx_v_score, PyObject *__pyx_v_source, PyObject *__pyx_v_feature, PyObject *__pyx_v_group, PyObject *__pyx_v_dropChr); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_2singleton(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_4center(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_6gnocis_7regions_6region_6__len__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_8__str__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_10__repr__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_12bstr(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_3seq___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_3seq_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_3seq_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_5start___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_5start_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_3end___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_3end_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_6strand___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_6strand_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_5score___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_5score_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_6source___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_6source_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_6source_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_7feature___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_7feature_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_7feature_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_5group___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_5group_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_5group_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_3ext___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_3ext_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_3ext_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_7markers___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_7markers_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gnocis_7regions_6region_7markers_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_14__reduce_cython__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6region_16__setstate_cython__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6gnocis_7regions_7regions___init__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_rgns, PyObject *__pyx_v_initialSort); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_2__iter__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_4__getitem__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static Py_ssize_t __pyx_pf_6gnocis_7regions_7regions_6__len__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_8__str__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_10_as_dict_(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_12table(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_14_repr_html_(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_16__repr__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_18__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_20__or__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_22__and__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_24__xor__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_4bstr_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_26bstr(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_28randomSplit(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_ratio); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_30sample(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_n); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_32filter(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_fltName, PyObject *__pyx_v_flt); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_34flatten(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_36merge(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_38intersection(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_40exclusion(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_42overlap(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_44nonOverlap(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_46overlapSensitivity(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_48overlapPrecision(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_19nucleotidePrecision_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_19nucleotidePrecision_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_50nucleotidePrecision(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_5dummy_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_52dummy(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_genome, PyObject *__pyx_v_useSeq); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_54expected(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_genome, PyObject *__pyx_v_statfun, PyObject *__pyx_v_useSeq, PyObject *__pyx_v_repeats); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_56recenter(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PY_LONG_LONG __pyx_v_size); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_58deltaResize(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_sizeDelta); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_60rename(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_newname); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda5(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_62sort(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_64extractSequences(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_seqs); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_66extractSequencesFrom2bit(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_68extractSequencesFromFASTA(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_70extract(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_src); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_72saveGFF(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_74saveBED(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_76saveCoordinateList(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_6genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_9genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_12genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_78printStatistics(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_4name___get__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_7regions_4name_2__set__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gnocis_7regions_7regions_4name_4__del__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_7regions___get__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static int __pyx_pf_6gnocis_7regions_7regions_7regions_2__set__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gnocis_7regions_7regions_7regions_4__del__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_80__reduce_cython__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_7regions_82__setstate_cython__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_loadBED(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_9loadBEDGZ_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_2loadBEDGZ(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_4loadGFF(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_dropChr); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_9loadGFFGZ_9genexpr24_7genexpr_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_9loadGFFGZ_9genexpr24_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_6loadGFFGZ(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_dropChr); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_8loadCoordinateList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_dropChr); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_10regionBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mainRegions, PyObject *__pyx_v_measure, PyObject *__pyx_v_measureName, PyObject *__pyx_v_testRegions, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeValues, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues, PyObject *__pyx_v_isPercent); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda14(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rs, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_12predictionBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda15(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ps, PyObject *__pyx_v_rs); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_14overlapSensitivityBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_regionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda16(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ps, PyObject *__pyx_v_rs); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_16overlapPrecisionBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_regionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda17(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ps, PyObject *__pyx_v_rs); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_18nucleotidePrecisionBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_regionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_measure(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ps, PyObject *__pyx_v_rs); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_20nucleotideRegionF1Barplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_regionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_22__pyx_unpickle_region(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6gnocis_7regions_24__pyx_unpickle_regions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6gnocis_7regions_region(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions_regions(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct__bstr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_5_dummy(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_6_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_7_printStatistics(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_8_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_9_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_10_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_11_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_12_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_14_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_16_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_17_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_19_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_20_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_21_measure(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_22_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_23_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_keys = {0, &__pyx_n_s_keys, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_strip = {0, &__pyx_n_s_strip, 0, 0, 0};
static PyObject *__pyx_float_0_;
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_float_0_5;
static PyObject *__pyx_float_neg_0_15;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_18;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_18804460;
static PyObject *__pyx_int_165254229;
static PyObject *__pyx_slice__2;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__59;
/* Late includes */

/* "gnocis/regions.pyx":53
 * 	__slots__ = ('seq', 'start', 'end', 'strand', 'score', 'source', 'feature', 'markers', 'group', 'ext')
 * 
 * 	def __init__(self, seq, start, end, strand = True, score = 0.0, source = '.', feature = '.', group = '.', dropChr = True):             # <<<<<<<<<<<<<<
 * 		if dropChr:
 * 			if seq.lower().startswith('chr'):
 */

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_seq = 0;
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_end = 0;
  PyObject *__pyx_v_strand = 0;
  PyObject *__pyx_v_score = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_feature = 0;
  PyObject *__pyx_v_group = 0;
  PyObject *__pyx_v_dropChr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_seq,&__pyx_n_s_start,&__pyx_n_s_end,&__pyx_n_s_strand,&__pyx_n_s_score,&__pyx_n_s_source,&__pyx_n_s_feature,&__pyx_n_s_group,&__pyx_n_s_dropChr,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[3] = ((PyObject *)Py_True);
    values[4] = ((PyObject *)__pyx_float_0_0);
    values[5] = ((PyObject *)__pyx_kp_u_);
    values[6] = ((PyObject *)__pyx_kp_u_);
    values[7] = ((PyObject *)__pyx_kp_u_);
    values[8] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 9, 1); __PYX_ERR(0, 53, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 9, 2); __PYX_ERR(0, 53, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strand);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_score);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feature);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_group);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dropChr);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 53, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_seq = values[0];
    __pyx_v_start = values[1];
    __pyx_v_end = values[2];
    __pyx_v_strand = values[3];
    __pyx_v_score = values[4];
    __pyx_v_source = values[5];
    __pyx_v_feature = values[6];
    __pyx_v_group = values[7];
    __pyx_v_dropChr = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 53, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.region.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_6region___init__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), __pyx_v_seq, __pyx_v_start, __pyx_v_end, __pyx_v_strand, __pyx_v_score, __pyx_v_source, __pyx_v_feature, __pyx_v_group, __pyx_v_dropChr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region___init__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_seq, PyObject *__pyx_v_start, PyObject *__pyx_v_end, PyObject *__pyx_v_strand, PyObject *__pyx_v_score, PyObject *__pyx_v_source, PyObject *__pyx_v_feature, PyObject *__pyx_v_group, PyObject *__pyx_v_dropChr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PY_LONG_LONG __pyx_t_6;
  PY_LONG_LONG __pyx_t_7;
  bool __pyx_t_8;
  float __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_seq);

  /* "gnocis/regions.pyx":54
 * 
 * 	def __init__(self, seq, start, end, strand = True, score = 0.0, source = '.', feature = '.', group = '.', dropChr = True):
 * 		if dropChr:             # <<<<<<<<<<<<<<
 * 			if seq.lower().startswith('chr'):
 * 				seq = seq[3:]
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_dropChr); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 54, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "gnocis/regions.pyx":55
 * 	def __init__(self, seq, start, end, strand = True, score = 0.0, source = '.', feature = '.', group = '.', dropChr = True):
 * 		if dropChr:
 * 			if seq.lower().startswith('chr'):             # <<<<<<<<<<<<<<
 * 				seq = seq[3:]
 * 		self.seq, self.start, self.end, self.strand, self.score = seq, start, end, strand, score
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_seq, __pyx_n_s_lower); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_startswith); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_n_u_chr) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_u_chr);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "gnocis/regions.pyx":56
 * 		if dropChr:
 * 			if seq.lower().startswith('chr'):
 * 				seq = seq[3:]             # <<<<<<<<<<<<<<
 * 		self.seq, self.start, self.end, self.strand, self.score = seq, start, end, strand, score
 * 		if start > end: self.start, self.end = end, start
 */
      __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_seq, 3, 0, NULL, NULL, &__pyx_slice__2, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_seq, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "gnocis/regions.pyx":55
 * 	def __init__(self, seq, start, end, strand = True, score = 0.0, source = '.', feature = '.', group = '.', dropChr = True):
 * 		if dropChr:
 * 			if seq.lower().startswith('chr'):             # <<<<<<<<<<<<<<
 * 				seq = seq[3:]
 * 		self.seq, self.start, self.end, self.strand, self.score = seq, start, end, strand, score
 */
    }

    /* "gnocis/regions.pyx":54
 * 
 * 	def __init__(self, seq, start, end, strand = True, score = 0.0, source = '.', feature = '.', group = '.', dropChr = True):
 * 		if dropChr:             # <<<<<<<<<<<<<<
 * 			if seq.lower().startswith('chr'):
 * 				seq = seq[3:]
 */
  }

  /* "gnocis/regions.pyx":57
 * 			if seq.lower().startswith('chr'):
 * 				seq = seq[3:]
 * 		self.seq, self.start, self.end, self.strand, self.score = seq, start, end, strand, score             # <<<<<<<<<<<<<<
 * 		if start > end: self.start, self.end = end, start
 * 		self.source, self.feature, self.group = source, feature, group
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_seq))||((__pyx_v_seq) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_seq)->tp_name), 0))) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_seq;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_start); if (unlikely((__pyx_t_6 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_end); if (unlikely((__pyx_t_7 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_strand); if (unlikely((__pyx_t_8 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_t_9 = __pyx_PyFloat_AsFloat(__pyx_v_score); if (unlikely((__pyx_t_9 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->seq);
  __Pyx_DECREF(__pyx_v_self->seq);
  __pyx_v_self->seq = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_v_self->start = __pyx_t_6;
  __pyx_v_self->end = __pyx_t_7;
  __pyx_v_self->strand = __pyx_t_8;
  __pyx_v_self->score = __pyx_t_9;

  /* "gnocis/regions.pyx":58
 * 				seq = seq[3:]
 * 		self.seq, self.start, self.end, self.strand, self.score = seq, start, end, strand, score
 * 		if start > end: self.start, self.end = end, start             # <<<<<<<<<<<<<<
 * 		self.source, self.feature, self.group = source, feature, group
 * 		self.ext = None
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_start, __pyx_v_end, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {
    __pyx_t_7 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_end); if (unlikely((__pyx_t_7 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_start); if (unlikely((__pyx_t_6 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L1_error)
    __pyx_v_self->start = __pyx_t_7;
    __pyx_v_self->end = __pyx_t_6;
  }

  /* "gnocis/regions.pyx":59
 * 		self.seq, self.start, self.end, self.strand, self.score = seq, start, end, strand, score
 * 		if start > end: self.start, self.end = end, start
 * 		self.source, self.feature, self.group = source, feature, group             # <<<<<<<<<<<<<<
 * 		self.ext = None
 * 
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_source))||((__pyx_v_source) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_source)->tp_name), 0))) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_source;
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_feature))||((__pyx_v_feature) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_feature)->tp_name), 0))) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_t_4 = __pyx_v_feature;
  __Pyx_INCREF(__pyx_t_4);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_group))||((__pyx_v_group) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_group)->tp_name), 0))) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_group;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->source);
  __Pyx_DECREF(__pyx_v_self->source);
  __pyx_v_self->source = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->feature);
  __Pyx_DECREF(__pyx_v_self->feature);
  __pyx_v_self->feature = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->group);
  __Pyx_DECREF(__pyx_v_self->group);
  __pyx_v_self->group = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "gnocis/regions.pyx":60
 * 		if start > end: self.start, self.end = end, start
 * 		self.source, self.feature, self.group = source, feature, group
 * 		self.ext = None             # <<<<<<<<<<<<<<
 * 
 * 	def singleton(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ext);
  __Pyx_DECREF(__pyx_v_self->ext);
  __pyx_v_self->ext = Py_None;

  /* "gnocis/regions.pyx":53
 * 	__slots__ = ('seq', 'start', 'end', 'strand', 'score', 'source', 'feature', 'markers', 'group', 'ext')
 * 
 * 	def __init__(self, seq, start, end, strand = True, score = 0.0, source = '.', feature = '.', group = '.', dropChr = True):             # <<<<<<<<<<<<<<
 * 		if dropChr:
 * 			if seq.lower().startswith('chr'):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.regions.region.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_seq);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":62
 * 		self.ext = None
 * 
 * 	def singleton(self):             # <<<<<<<<<<<<<<
 * 		return regions(str(self), [ self ])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_3singleton(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_3singleton(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("singleton (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_2singleton(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_2singleton(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("singleton", 0);

  /* "gnocis/regions.pyx":63
 * 
 * 	def singleton(self):
 * 		return regions(str(self), [ self ])             # <<<<<<<<<<<<<<
 * 
 * 	def center(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self));
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":62
 * 		self.ext = None
 * 
 * 	def singleton(self):             # <<<<<<<<<<<<<<
 * 		return regions(str(self), [ self ])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.region.singleton", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":65
 * 		return regions(str(self), [ self ])
 * 
 * 	def center(self):             # <<<<<<<<<<<<<<
 * 		return int((self.start + self.end) / 2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_5center(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_5center(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("center (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_4center(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_4center(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("center", 0);

  /* "gnocis/regions.pyx":66
 * 
 * 	def center(self):
 * 		return int((self.start + self.end) / 2)             # <<<<<<<<<<<<<<
 * 
 * 	def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromDouble((((double)(__pyx_v_self->start + __pyx_v_self->end)) / 2.0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":65
 * 		return regions(str(self), [ self ])
 * 
 * 	def center(self):             # <<<<<<<<<<<<<<
 * 		return int((self.start + self.end) / 2)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.center", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":68
 * 		return int((self.start + self.end) / 2)
 * 
 * 	def __len__(self):             # <<<<<<<<<<<<<<
 * 		return self.end+1-self.start
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6gnocis_7regions_6region_7__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6gnocis_7regions_6region_7__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_6__len__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6gnocis_7regions_6region_6__len__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "gnocis/regions.pyx":69
 * 
 * 	def __len__(self):
 * 		return self.end+1-self.start             # <<<<<<<<<<<<<<
 * 
 * 	def __str__(self):
 */
  __pyx_r = ((__pyx_v_self->end + 1) - __pyx_v_self->start);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":68
 * 		return int((self.start + self.end) / 2)
 * 
 * 	def __len__(self):             # <<<<<<<<<<<<<<
 * 		return self.end+1-self.start
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":71
 * 		return self.end+1-self.start
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		rname = '%s:%d..%d (%s)'%(self.seq, self.start, self.end, '+' if self.strand else '-')
 * 		ext = [ 'Score: %f'%self.score ]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_9__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_9__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_8__str__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_8__str__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_v_rname = NULL;
  PyObject *__pyx_v_ext = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "gnocis/regions.pyx":72
 * 
 * 	def __str__(self):
 * 		rname = '%s:%d..%d (%s)'%(self.seq, self.start, self.end, '+' if self.strand else '-')             # <<<<<<<<<<<<<<
 * 		ext = [ 'Score: %f'%self.score ]
 * 		if self.source != '.':
 */
  __pyx_t_1 = PyTuple_New(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyUnicode_Unicode(__pyx_v_self->seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__3);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__3);
  __pyx_t_4 = __Pyx_PyUnicode_From_PY_LONG_LONG(__pyx_v_self->start, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__4);
  __pyx_t_4 = __Pyx_PyUnicode_From_PY_LONG_LONG(__pyx_v_self->end, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__5);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__5);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u__5);
  if ((__pyx_v_self->strand != 0)) {
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_4 = __pyx_kp_u__6;
  } else {
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_4 = __pyx_kp_u__7;
  }
  __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__8);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__8);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_kp_u__8);
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 8, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rname = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":73
 * 	def __str__(self):
 * 		rname = '%s:%d..%d (%s)'%(self.seq, self.start, self.end, '+' if self.strand else '-')
 * 		ext = [ 'Score: %f'%self.score ]             # <<<<<<<<<<<<<<
 * 		if self.source != '.':
 * 			ext.append( 'Source: %s'%self.source )
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyUnicode_Format(__pyx_kp_u_Score_f, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_ext = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":74
 * 		rname = '%s:%d..%d (%s)'%(self.seq, self.start, self.end, '+' if self.strand else '-')
 * 		ext = [ 'Score: %f'%self.score ]
 * 		if self.source != '.':             # <<<<<<<<<<<<<<
 * 			ext.append( 'Source: %s'%self.source )
 * 		if self.feature != '.':
 */
  __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_self->source, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 74, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "gnocis/regions.pyx":75
 * 		ext = [ 'Score: %f'%self.score ]
 * 		if self.source != '.':
 * 			ext.append( 'Source: %s'%self.source )             # <<<<<<<<<<<<<<
 * 		if self.feature != '.':
 * 			ext.append( 'Feature: %s'%self.feature )
 */
    __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_Source_s, __pyx_v_self->source); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_ext, __pyx_t_5); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "gnocis/regions.pyx":74
 * 		rname = '%s:%d..%d (%s)'%(self.seq, self.start, self.end, '+' if self.strand else '-')
 * 		ext = [ 'Score: %f'%self.score ]
 * 		if self.source != '.':             # <<<<<<<<<<<<<<
 * 			ext.append( 'Source: %s'%self.source )
 * 		if self.feature != '.':
 */
  }

  /* "gnocis/regions.pyx":76
 * 		if self.source != '.':
 * 			ext.append( 'Source: %s'%self.source )
 * 		if self.feature != '.':             # <<<<<<<<<<<<<<
 * 			ext.append( 'Feature: %s'%self.feature )
 * 		return 'Region<%s (%s)>'%(rname, '; '.join(ext))
 */
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_self->feature, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 76, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_7 != 0);
  if (__pyx_t_6) {

    /* "gnocis/regions.pyx":77
 * 			ext.append( 'Source: %s'%self.source )
 * 		if self.feature != '.':
 * 			ext.append( 'Feature: %s'%self.feature )             # <<<<<<<<<<<<<<
 * 		return 'Region<%s (%s)>'%(rname, '; '.join(ext))
 * 
 */
    __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_Feature_s, __pyx_v_self->feature); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_ext, __pyx_t_5); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "gnocis/regions.pyx":76
 * 		if self.source != '.':
 * 			ext.append( 'Source: %s'%self.source )
 * 		if self.feature != '.':             # <<<<<<<<<<<<<<
 * 			ext.append( 'Feature: %s'%self.feature )
 * 		return 'Region<%s (%s)>'%(rname, '; '.join(ext))
 */
  }

  /* "gnocis/regions.pyx":78
 * 		if self.feature != '.':
 * 			ext.append( 'Feature: %s'%self.feature )
 * 		return 'Region<%s (%s)>'%(rname, '; '.join(ext))             # <<<<<<<<<<<<<<
 * 
 * 	def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Region);
  __pyx_t_2 += 7;
  __Pyx_GIVEREF(__pyx_kp_u_Region);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Region);
  __Pyx_INCREF(__pyx_v_rname);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_rname) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_rname) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_v_rname);
  __Pyx_GIVEREF(__pyx_v_rname);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_rname);
  __Pyx_INCREF(__pyx_kp_u__5);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__5);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__5);
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__9, __pyx_v_ext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_INCREF(__pyx_kp_u__10);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__10);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_kp_u__10);
  __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_5, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":71
 * 		return self.end+1-self.start
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		rname = '%s:%d..%d (%s)'%(self.seq, self.start, self.end, '+' if self.strand else '-')
 * 		ext = [ 'Score: %f'%self.score ]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.regions.region.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rname);
  __Pyx_XDECREF(__pyx_v_ext);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":80
 * 		return 'Region<%s (%s)>'%(rname, '; '.join(ext))
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		return self.__str__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_11__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_11__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_10__repr__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_10__repr__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "gnocis/regions.pyx":81
 * 
 * 	def __repr__(self):
 * 		return self.__str__()             # <<<<<<<<<<<<<<
 * 
 * 	def bstr(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":80
 * 		return 'Region<%s (%s)>'%(rname, '; '.join(ext))
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		return self.__str__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.region.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":83
 * 		return self.__str__()
 * 
 * 	def bstr(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:return: Returns the region formatted as a string: `seq:start..end (strand)`.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_13bstr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6gnocis_7regions_6region_12bstr[] = "\n\t\t:return: Returns the region formatted as a string: `seq:start..end (strand)`.\n\t\t:rtype: str\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_6region_13bstr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bstr (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_12bstr(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_12bstr(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bstr", 0);

  /* "gnocis/regions.pyx":88
 * 		:rtype: str
 * 		"""
 * 		return '%s:%d..%d (%s)'%(self.seq, self.start, self.end, '+' if self.strand else '-')             # <<<<<<<<<<<<<<
 * 
 * # Represents region sets. Supports operands '+' for adding regions from two sets with no merging, '|' for merging two sets, and '^' for excluding the regions in the latter set from the first.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyUnicode_Unicode(__pyx_v_self->seq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__3);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__3);
  __pyx_t_4 = __Pyx_PyUnicode_From_PY_LONG_LONG(__pyx_v_self->start, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__4);
  __pyx_t_4 = __Pyx_PyUnicode_From_PY_LONG_LONG(__pyx_v_self->end, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__5);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__5);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u__5);
  if ((__pyx_v_self->strand != 0)) {
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_4 = __pyx_kp_u__6;
  } else {
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_4 = __pyx_kp_u__7;
  }
  __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__8);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__8);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_kp_u__8);
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 8, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":83
 * 		return self.__str__()
 * 
 * 	def bstr(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:return: Returns the region formatted as a string: `seq:start..end (strand)`.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.regions.region.bstr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":18
 * cdef class region:
 * 
 * 	cdef public str seq             # <<<<<<<<<<<<<<
 * 	cdef public long long start, end
 * 	cdef public bool strand
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_3seq_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_3seq_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_3seq___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_3seq___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->seq);
  __pyx_r = __pyx_v_self->seq;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_3seq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_3seq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_3seq_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_3seq_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->seq);
  __Pyx_DECREF(__pyx_v_self->seq);
  __pyx_v_self->seq = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.seq.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_3seq_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_3seq_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_3seq_4__del__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_3seq_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->seq);
  __Pyx_DECREF(__pyx_v_self->seq);
  __pyx_v_self->seq = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":19
 * 
 * 	cdef public str seq
 * 	cdef public long long start, end             # <<<<<<<<<<<<<<
 * 	cdef public bool strand
 * 	cdef public float score
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_5start_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_5start_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_5start___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_5start___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->start); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.start.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_5start_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_5start_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_5start_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PY_LONG_LONG __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_value); if (unlikely((__pyx_t_1 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(1, 19, __pyx_L1_error)
  __pyx_v_self->start = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.region.start.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_3end_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_3end_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_3end___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_3end___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->end); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.end.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_3end_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_3end_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_3end_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PY_LONG_LONG __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_value); if (unlikely((__pyx_t_1 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(1, 19, __pyx_L1_error)
  __pyx_v_self->end = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.region.end.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":20
 * 	cdef public str seq
 * 	cdef public long long start, end
 * 	cdef public bool strand             # <<<<<<<<<<<<<<
 * 	cdef public float score
 * 	cdef public str source, feature, group
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_6strand_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_6strand_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_6strand___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_6strand___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->strand); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.strand.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_6strand_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_6strand_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_6strand_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_6strand_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_v_self->strand = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.region.strand.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":21
 * 	cdef public long long start, end
 * 	cdef public bool strand
 * 	cdef public float score             # <<<<<<<<<<<<<<
 * 	cdef public str source, feature, group
 * 	cdef public ext
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_5score_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_5score_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_5score___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_5score___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.score.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_5score_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_5score_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_5score_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  float __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsFloat(__pyx_v_value); if (unlikely((__pyx_t_1 == (float)-1) && PyErr_Occurred())) __PYX_ERR(1, 21, __pyx_L1_error)
  __pyx_v_self->score = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.region.score.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":22
 * 	cdef public bool strand
 * 	cdef public float score
 * 	cdef public str source, feature, group             # <<<<<<<<<<<<<<
 * 	cdef public ext
 * 	cdef public set markers
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_6source_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_6source_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_6source___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_6source___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->source);
  __pyx_r = __pyx_v_self->source;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_6source_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_6source_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_6source_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_6source_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 22, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->source);
  __Pyx_DECREF(__pyx_v_self->source);
  __pyx_v_self->source = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.source.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_6source_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_6source_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_6source_4__del__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_6source_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->source);
  __Pyx_DECREF(__pyx_v_self->source);
  __pyx_v_self->source = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_7feature_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_7feature_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_7feature___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_7feature___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->feature);
  __pyx_r = __pyx_v_self->feature;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_7feature_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_7feature_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_7feature_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_7feature_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 22, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->feature);
  __Pyx_DECREF(__pyx_v_self->feature);
  __pyx_v_self->feature = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.feature.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_7feature_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_7feature_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_7feature_4__del__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_7feature_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->feature);
  __Pyx_DECREF(__pyx_v_self->feature);
  __pyx_v_self->feature = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_5group_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_5group_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_5group___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_5group___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->group);
  __pyx_r = __pyx_v_self->group;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_5group_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_5group_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_5group_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_5group_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 22, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->group);
  __Pyx_DECREF(__pyx_v_self->group);
  __pyx_v_self->group = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.group.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_5group_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_5group_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_5group_4__del__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_5group_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->group);
  __Pyx_DECREF(__pyx_v_self->group);
  __pyx_v_self->group = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":23
 * 	cdef public float score
 * 	cdef public str source, feature, group
 * 	cdef public ext             # <<<<<<<<<<<<<<
 * 	cdef public set markers
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_3ext_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_3ext_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_3ext___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_3ext___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ext);
  __pyx_r = __pyx_v_self->ext;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_3ext_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_3ext_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_3ext_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_3ext_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->ext);
  __Pyx_DECREF(__pyx_v_self->ext);
  __pyx_v_self->ext = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_3ext_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_3ext_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_3ext_4__del__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_3ext_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ext);
  __Pyx_DECREF(__pyx_v_self->ext);
  __pyx_v_self->ext = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":24
 * 	cdef public str source, feature, group
 * 	cdef public ext
 * 	cdef public set markers             # <<<<<<<<<<<<<<
 * 
 * # Represents region sets. Supports operands '+' for adding regions from two sets with no merging, '|' for merging two sets, and '^' for excluding the regions in the latter set from the first.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_7markers_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_7markers_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_7markers___get__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_7markers___get__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->markers);
  __pyx_r = __pyx_v_self->markers;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_7markers_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_7markers_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_7markers_2__set__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_7markers_2__set__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PySet_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 24, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->markers);
  __Pyx_DECREF(__pyx_v_self->markers);
  __pyx_v_self->markers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.markers.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_6region_7markers_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gnocis_7regions_6region_7markers_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_7markers_4__del__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_6region_7markers_4__del__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->markers);
  __Pyx_DECREF(__pyx_v_self->markers);
  __pyx_v_self->markers = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_14__reduce_cython__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_14__reduce_cython__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.end, self.ext, self.feature, self.group, self.markers, self.score, self.seq, self.source, self.start, self.strand)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->end); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->score); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->start); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->strand); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(10); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->ext);
  __Pyx_GIVEREF(__pyx_v_self->ext);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self->ext);
  __Pyx_INCREF(__pyx_v_self->feature);
  __Pyx_GIVEREF(__pyx_v_self->feature);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_self->feature);
  __Pyx_INCREF(__pyx_v_self->group);
  __Pyx_GIVEREF(__pyx_v_self->group);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_self->group);
  __Pyx_INCREF(__pyx_v_self->markers);
  __Pyx_GIVEREF(__pyx_v_self->markers);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->markers);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->seq);
  __Pyx_GIVEREF(__pyx_v_self->seq);
  PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_v_self->seq);
  __Pyx_INCREF(__pyx_v_self->source);
  __Pyx_GIVEREF(__pyx_v_self->source);
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_v_self->source);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 8, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 9, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.end, self.ext, self.feature, self.group, self.markers, self.score, self.seq, self.source, self.start, self.strand)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.end, self.ext, self.feature, self.group, self.markers, self.score, self.seq, self.source, self.start, self.strand)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.ext is not None or self.feature is not None or self.group is not None or self.markers is not None or self.seq is not None or self.source is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.end, self.ext, self.feature, self.group, self.markers, self.score, self.seq, self.source, self.start, self.strand)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.ext is not None or self.feature is not None or self.group is not None or self.markers is not None or self.seq is not None or self.source is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->ext != Py_None);
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_self->feature != ((PyObject*)Py_None));
    __pyx_t_6 = (__pyx_t_8 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->group != ((PyObject*)Py_None));
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_self->markers != ((PyObject*)Py_None));
    __pyx_t_6 = (__pyx_t_8 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_7 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->seq != ((PyObject*)Py_None));
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_self->source != ((PyObject*)Py_None));
    __pyx_t_6 = (__pyx_t_8 != 0);
    __pyx_t_7 = __pyx_t_6;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.ext is not None or self.feature is not None or self.group is not None or self.markers is not None or self.seq is not None or self.source is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, None), state
 *     else:
 */
  __pyx_t_7 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":13
 *         use_setstate = self.ext is not None or self.feature is not None or self.group is not None or self.markers is not None or self.seq is not None or self.source is not None
 *     if use_setstate:
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_region); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_18804460);
    __Pyx_GIVEREF(__pyx_int_18804460);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_18804460);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.ext is not None or self.feature is not None or self.group is not None or self.markers is not None or self.seq is not None or self.source is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, None), state
 *     else:
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_region__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_region); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_18804460);
    __Pyx_GIVEREF(__pyx_int_18804460);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_18804460);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.regions.region.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_region__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_6region_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_6region_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_6region_16__setstate_cython__(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_6region_16__setstate_cython__(struct __pyx_obj_6gnocis_7regions_region *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_region__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6gnocis_7regions___pyx_unpickle_region__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_region, (type(self), 0x11eeeec, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_region__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.region.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":108
 * 	#__slots__ = ('name', 'regions')
 * 
 * 	def __init__(self, name, rgns, initialSort = True):             # <<<<<<<<<<<<<<
 * 		self.name, self.regions = name, rgns
 * 		if initialSort and len(rgns) != 0:
 */

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_7regions_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gnocis_7regions_7regions_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_rgns = 0;
  PyObject *__pyx_v_initialSort = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_rgns,&__pyx_n_s_initialSort,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rgns)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_initialSort);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 108, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_rgns = values[1];
    __pyx_v_initialSort = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 108, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions___init__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), __pyx_v_name, __pyx_v_rgns, __pyx_v_initialSort);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_7regions___init__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_rgns, PyObject *__pyx_v_initialSort) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gnocis/regions.pyx":109
 * 
 * 	def __init__(self, name, rgns, initialSort = True):
 * 		self.name, self.regions = name, rgns             # <<<<<<<<<<<<<<
 * 		if initialSort and len(rgns) != 0:
 * 			self.sort()
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_name))||((__pyx_v_name) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_name)->tp_name), 0))) __PYX_ERR(0, 109, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_name;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_v_rgns))||((__pyx_v_rgns) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_rgns)->tp_name), 0))) __PYX_ERR(0, 109, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_rgns;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->regions);
  __Pyx_DECREF(__pyx_v_self->regions);
  __pyx_v_self->regions = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":110
 * 	def __init__(self, name, rgns, initialSort = True):
 * 		self.name, self.regions = name, rgns
 * 		if initialSort and len(rgns) != 0:             # <<<<<<<<<<<<<<
 * 			self.sort()
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_initialSort); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 110, __pyx_L1_error)
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = PyObject_Length(__pyx_v_rgns); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 110, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_5 != 0) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "gnocis/regions.pyx":111
 * 		self.name, self.regions = name, rgns
 * 		if initialSort and len(rgns) != 0:
 * 			self.sort()             # <<<<<<<<<<<<<<
 * 
 * 	def __iter__(self):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/regions.pyx":110
 * 	def __init__(self, name, rgns, initialSort = True):
 * 		self.name, self.regions = name, rgns
 * 		if initialSort and len(rgns) != 0:             # <<<<<<<<<<<<<<
 * 			self.sort()
 * 
 */
  }

  /* "gnocis/regions.pyx":108
 * 	#__slots__ = ('name', 'regions')
 * 
 * 	def __init__(self, name, rgns, initialSort = True):             # <<<<<<<<<<<<<<
 * 		self.name, self.regions = name, rgns
 * 		if initialSort and len(rgns) != 0:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.regions.regions.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":113
 * 			self.sort()
 * 
 * 	def __iter__(self):             # <<<<<<<<<<<<<<
 * 		return self.regions.__iter__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_3__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_2__iter__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_2__iter__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "gnocis/regions.pyx":114
 * 
 * 	def __iter__(self):
 * 		return self.regions.__iter__()             # <<<<<<<<<<<<<<
 * 
 * 	def __getitem__(self, i):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->regions, __pyx_n_s_iter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":113
 * 			self.sort()
 * 
 * 	def __iter__(self):             # <<<<<<<<<<<<<<
 * 		return self.regions.__iter__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.regions.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":116
 * 		return self.regions.__iter__()
 * 
 * 	def __getitem__(self, i):             # <<<<<<<<<<<<<<
 * 		return self.regions[i]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_5__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_4__getitem__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_4__getitem__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "gnocis/regions.pyx":117
 * 
 * 	def __getitem__(self, i):
 * 		return self.regions[i]             # <<<<<<<<<<<<<<
 * 
 * 	def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 117, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->regions, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":116
 * 		return self.regions.__iter__()
 * 
 * 	def __getitem__(self, i):             # <<<<<<<<<<<<<<
 * 		return self.regions[i]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.regions.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":119
 * 		return self.regions[i]
 * 
 * 	def __len__(self):             # <<<<<<<<<<<<<<
 * 		return len(self.regions)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6gnocis_7regions_7regions_7__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6gnocis_7regions_7regions_7__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_6__len__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6gnocis_7regions_7regions_6__len__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "gnocis/regions.pyx":120
 * 
 * 	def __len__(self):
 * 		return len(self.regions)             # <<<<<<<<<<<<<<
 * 
 * 	def __str__(self):
 */
  __pyx_t_1 = __pyx_v_self->regions;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 120, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":119
 * 		return self.regions[i]
 * 
 * 	def __len__(self):             # <<<<<<<<<<<<<<
 * 		return len(self.regions)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.regions.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":122
 * 		return len(self.regions)
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Region set<%s>'%(self.name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_9__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_9__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_8__str__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_8__str__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "gnocis/regions.pyx":123
 * 
 * 	def __str__(self):
 * 		return 'Region set<%s>'%(self.name)             # <<<<<<<<<<<<<<
 * 
 * 	def _as_dict_(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyUnicode_Format(__pyx_kp_u_Region_set_s, __pyx_v_self->name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":122
 * 		return len(self.regions)
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Region set<%s>'%(self.name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.regions.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":125
 * 		return 'Region set<%s>'%(self.name)
 * 
 * 	def _as_dict_(self):             # <<<<<<<<<<<<<<
 * 		return {
 * 			'Seq.': [ r.seq for r in self ],
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_11_as_dict_(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_11_as_dict_(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_as_dict_ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_10_as_dict_(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_10_as_dict_(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_7genexpr__pyx_v_r = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_r = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_r = NULL;
  PyObject *__pyx_8genexpr3__pyx_v_r = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_as_dict_", 0);

  /* "gnocis/regions.pyx":126
 * 
 * 	def _as_dict_(self):
 * 		return {             # <<<<<<<<<<<<<<
 * 			'Seq.': [ r.seq for r in self ],
 * 			'Start': [ r.start for r in self ],
 */
  __Pyx_XDECREF(__pyx_r);

  /* "gnocis/regions.pyx":127
 * 	def _as_dict_(self):
 * 		return {
 * 			'Seq.': [ r.seq for r in self ],             # <<<<<<<<<<<<<<
 * 			'Start': [ r.start for r in self ],
 * 			'End': [ r.end for r in self ],
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
      __pyx_t_3 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 127, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 127, __pyx_L5_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 127, __pyx_L5_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 127, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_r, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_7genexpr__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 127, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_r); __pyx_7genexpr__pyx_v_r = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_r); __pyx_7genexpr__pyx_v_r = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_1, __pyx_kp_u_Seq, __pyx_t_2) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  { /* enter inner scope */

    /* "gnocis/regions.pyx":128
 * 		return {
 * 			'Seq.': [ r.seq for r in self ],
 * 			'Start': [ r.start for r in self ],             # <<<<<<<<<<<<<<
 * 			'End': [ r.end for r in self ],
 * 			'Strand': [ '+' if r.strand else '-' for r in self ],
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
      __pyx_t_3 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 128, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 128, __pyx_L11_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 128, __pyx_L11_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 128, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 128, __pyx_L11_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 128, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 128, __pyx_L11_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_r, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr1__pyx_v_r, __pyx_n_s_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 128, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 128, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_r); __pyx_8genexpr1__pyx_v_r = 0;
    goto __pyx_L14_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_r); __pyx_8genexpr1__pyx_v_r = 0;
    goto __pyx_L1_error;
    __pyx_L14_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_Start, __pyx_t_2) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  { /* enter inner scope */

    /* "gnocis/regions.pyx":129
 * 			'Seq.': [ r.seq for r in self ],
 * 			'Start': [ r.start for r in self ],
 * 			'End': [ r.end for r in self ],             # <<<<<<<<<<<<<<
 * 			'Strand': [ '+' if r.strand else '-' for r in self ],
 * 			'Length': [ len(r) for r in self ],
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L17_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
      __pyx_t_3 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 129, __pyx_L17_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 129, __pyx_L17_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 129, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 129, __pyx_L17_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 129, __pyx_L17_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 129, __pyx_L17_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_r, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr2__pyx_v_r, __pyx_n_s_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 129, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 129, __pyx_L17_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_r); __pyx_8genexpr2__pyx_v_r = 0;
    goto __pyx_L20_exit_scope;
    __pyx_L17_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_r); __pyx_8genexpr2__pyx_v_r = 0;
    goto __pyx_L1_error;
    __pyx_L20_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_End, __pyx_t_2) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  { /* enter inner scope */

    /* "gnocis/regions.pyx":130
 * 			'Start': [ r.start for r in self ],
 * 			'End': [ r.end for r in self ],
 * 			'Strand': [ '+' if r.strand else '-' for r in self ],             # <<<<<<<<<<<<<<
 * 			'Length': [ len(r) for r in self ],
 * 		}
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
      __pyx_t_3 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 130, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 130, __pyx_L23_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 130, __pyx_L23_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 130, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 130, __pyx_L23_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 130, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 130, __pyx_L23_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_r, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr3__pyx_v_r, __pyx_n_s_strand); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 130, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 130, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_8) {
        __Pyx_INCREF(__pyx_kp_u__6);
        __pyx_t_6 = __pyx_kp_u__6;
      } else {
        __Pyx_INCREF(__pyx_kp_u__7);
        __pyx_t_6 = __pyx_kp_u__7;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 130, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_r); __pyx_8genexpr3__pyx_v_r = 0;
    goto __pyx_L26_exit_scope;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_r); __pyx_8genexpr3__pyx_v_r = 0;
    goto __pyx_L1_error;
    __pyx_L26_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_Strand, __pyx_t_2) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  { /* enter inner scope */

    /* "gnocis/regions.pyx":131
 * 			'End': [ r.end for r in self ],
 * 			'Strand': [ '+' if r.strand else '-' for r in self ],
 * 			'Length': [ len(r) for r in self ],             # <<<<<<<<<<<<<<
 * 		}
 * 
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L29_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
      __pyx_t_3 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 131, __pyx_L29_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 131, __pyx_L29_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 131, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 131, __pyx_L29_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 131, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 131, __pyx_L29_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_r, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_9 = PyObject_Length(__pyx_8genexpr4__pyx_v_r); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 131, __pyx_L29_error)
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 131, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 131, __pyx_L29_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_r); __pyx_8genexpr4__pyx_v_r = 0;
    goto __pyx_L32_exit_scope;
    __pyx_L29_error:;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_r); __pyx_8genexpr4__pyx_v_r = 0;
    goto __pyx_L1_error;
    __pyx_L32_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_Length, __pyx_t_2) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":125
 * 		return 'Region set<%s>'%(self.name)
 * 
 * 	def _as_dict_(self):             # <<<<<<<<<<<<<<
 * 		return {
 * 			'Seq.': [ r.seq for r in self ],
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.regions.regions._as_dict_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_r);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_r);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_r);
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_r);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":134
 * 		}
 * 
 * 	def table(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Table: ' + self.__str__(),
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_13table(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_13table(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("table (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_12table(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_12table(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("table", 0);

  /* "gnocis/regions.pyx":135
 * 
 * 	def table(self):
 * 		return nctable(             # <<<<<<<<<<<<<<
 * 			'Table: ' + self.__str__(),
 * 			self._as_dict_(),
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_nctable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gnocis/regions.pyx":136
 * 	def table(self):
 * 		return nctable(
 * 			'Table: ' + self.__str__(),             # <<<<<<<<<<<<<<
 * 			self._as_dict_(),
 * 			align = { 'Seq.': 'l' }
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_kp_u_Table, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":137
 * 		return nctable(
 * 			'Table: ' + self.__str__(),
 * 			self._as_dict_(),             # <<<<<<<<<<<<<<
 * 			align = { 'Seq.': 'l' }
 * 		)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_as_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gnocis/regions.pyx":135
 * 
 * 	def table(self):
 * 		return nctable(             # <<<<<<<<<<<<<<
 * 			'Table: ' + self.__str__(),
 * 			self._as_dict_(),
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":138
 * 			'Table: ' + self.__str__(),
 * 			self._as_dict_(),
 * 			align = { 'Seq.': 'l' }             # <<<<<<<<<<<<<<
 * 		)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_kp_u_Seq, __pyx_n_u_l) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_align, __pyx_t_3) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/regions.pyx":135
 * 
 * 	def table(self):
 * 		return nctable(             # <<<<<<<<<<<<<<
 * 			'Table: ' + self.__str__(),
 * 			self._as_dict_(),
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":134
 * 		}
 * 
 * 	def table(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Table: ' + self.__str__(),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.regions.regions.table", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":141
 * 		)
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		return self.table()._repr_html_()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_15_repr_html_(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_15_repr_html_(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_repr_html_ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_14_repr_html_(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_14_repr_html_(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_repr_html_", 0);

  /* "gnocis/regions.pyx":142
 * 
 * 	def _repr_html_(self):
 * 		return self.table()._repr_html_()             # <<<<<<<<<<<<<<
 * 
 * 	def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_table); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_repr_html); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":141
 * 		)
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		return self.table()._repr_html_()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.regions.regions._repr_html_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":144
 * 		return self.table()._repr_html_()
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		return self.table().__repr__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_17__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_17__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_16__repr__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_16__repr__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "gnocis/regions.pyx":145
 * 
 * 	def __repr__(self):
 * 		return self.table().__repr__()             # <<<<<<<<<<<<<<
 * 
 * 	def __add__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_table); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_repr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":144
 * 		return self.table()._repr_html_()
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		return self.table().__repr__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.regions.regions.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":147
 * 		return self.table().__repr__()
 * 
 * 	def __add__(self, other):             # <<<<<<<<<<<<<<
 * 		rs = regions('%s + %s'%(self.name, other.name), self.regions + other.regions)
 * 		rs.sort()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_19__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_19__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_18__add__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_18__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_rs = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__add__", 0);

  /* "gnocis/regions.pyx":148
 * 
 * 	def __add__(self, other):
 * 		rs = regions('%s + %s'%(self.name, other.name), self.regions + other.regions)             # <<<<<<<<<<<<<<
 * 		rs.sort()
 * 		return rs
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__11);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__11);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__11);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_5), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_regions); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyNumber_Add(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_rs = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "gnocis/regions.pyx":149
 * 	def __add__(self, other):
 * 		rs = regions('%s + %s'%(self.name, other.name), self.regions + other.regions)
 * 		rs.sort()             # <<<<<<<<<<<<<<
 * 		return rs
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rs), __pyx_n_s_sort); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "gnocis/regions.pyx":150
 * 		rs = regions('%s + %s'%(self.name, other.name), self.regions + other.regions)
 * 		rs.sort()
 * 		return rs             # <<<<<<<<<<<<<<
 * 
 * 	def __or__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_rs));
  __pyx_r = ((PyObject *)__pyx_v_rs);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":147
 * 		return self.table().__repr__()
 * 
 * 	def __add__(self, other):             # <<<<<<<<<<<<<<
 * 		rs = regions('%s + %s'%(self.name, other.name), self.regions + other.regions)
 * 		rs.sort()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.regions.regions.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_rs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":152
 * 		return rs
 * 
 * 	def __or__(self, other):             # <<<<<<<<<<<<<<
 * 		return self.merge(other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_21__or__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_21__or__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__or__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_20__or__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_20__or__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__or__", 0);

  /* "gnocis/regions.pyx":153
 * 
 * 	def __or__(self, other):
 * 		return self.merge(other)             # <<<<<<<<<<<<<<
 * 
 * 	def __and__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_merge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":152
 * 		return rs
 * 
 * 	def __or__(self, other):             # <<<<<<<<<<<<<<
 * 		return self.merge(other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.regions.__or__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":155
 * 		return self.merge(other)
 * 
 * 	def __and__(self, other):             # <<<<<<<<<<<<<<
 * 		return self.intersection(other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_23__and__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_23__and__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__and__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_22__and__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_22__and__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__and__", 0);

  /* "gnocis/regions.pyx":156
 * 
 * 	def __and__(self, other):
 * 		return self.intersection(other)             # <<<<<<<<<<<<<<
 * 
 * 	def __xor__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_intersection); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":155
 * 		return self.merge(other)
 * 
 * 	def __and__(self, other):             # <<<<<<<<<<<<<<
 * 		return self.intersection(other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.regions.__and__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":158
 * 		return self.intersection(other)
 * 
 * 	def __xor__(self, other):             # <<<<<<<<<<<<<<
 * 		return self.exclusion(other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_25__xor__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_25__xor__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__xor__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_24__xor__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_24__xor__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__xor__", 0);

  /* "gnocis/regions.pyx":159
 * 
 * 	def __xor__(self, other):
 * 		return self.exclusion(other)             # <<<<<<<<<<<<<<
 * 
 * 	def bstr(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exclusion); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":158
 * 		return self.intersection(other)
 * 
 * 	def __xor__(self, other):             # <<<<<<<<<<<<<<
 * 		return self.exclusion(other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.regions.__xor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":161
 * 		return self.exclusion(other)
 * 
 * 	def bstr(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:return: Returns a concatenated list of regions formatted as `seq:start..end (strand)`.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_27bstr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_26bstr[] = "\n\t\t:return: Returns a concatenated list of regions formatted as `seq:start..end (strand)`.\n\t\t:rtype: str\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_27bstr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bstr (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_26bstr(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_4bstr_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":166
 * 		:rtype: str
 * 		"""
 * 		return ', '.join( r.bstr() for r in self.regions )             # <<<<<<<<<<<<<<
 * 
 * 	# Returns a random division of the regions
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_4bstr_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_1_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 166, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_4bstr_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_bstr_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.bstr.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_4bstr_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 166, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 166, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 166, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_r, __pyx_n_s_bstr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":161
 * 		return self.exclusion(other)
 * 
 * 	def bstr(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:return: Returns a concatenated list of regions formatted as `seq:start..end (strand)`.
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_26bstr(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *__pyx_cur_scope;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_4bstr_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bstr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct__bstr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct__bstr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 161, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "gnocis/regions.pyx":166
 * 		:rtype: str
 * 		"""
 * 		return ', '.join( r.bstr() for r in self.regions )             # <<<<<<<<<<<<<<
 * 
 * 	# Returns a random division of the regions
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_7regions_4bstr_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__12, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":161
 * 		return self.exclusion(other)
 * 
 * 	def bstr(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:return: Returns a concatenated list of regions formatted as `seq:start..end (strand)`.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.regions.regions.bstr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_4bstr_2generator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":169
 * 
 * 	# Returns a random division of the regions
 * 	def randomSplit(self, ratio = 0.5):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:param ratio: Ratio for the split, between 0 and 1. For the return touple, `(A, B)`, a ratio of 0 means all the regions are in `B`, and opposite for a ratio of 1.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_29randomSplit(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_28randomSplit[] = "\n\t\t:param ratio: Ratio for the split, between 0 and 1. For the return touple, `(A, B)`, a ratio of 0 means all the regions are in `B`, and opposite for a ratio of 1.\n\t\t:type ratio: float\n\t\t\n\t\t:return: Returns a random split of the regions into two independent sets, with the given ratio.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_29randomSplit(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ratio = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("randomSplit (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ratio,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_float_0_5);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ratio);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "randomSplit") < 0)) __PYX_ERR(0, 169, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ratio = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("randomSplit", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 169, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.randomSplit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_28randomSplit(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), __pyx_v_ratio);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_28randomSplit(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_ratio) {
  PyObject *__pyx_v_rgns = NULL;
  PyObject *__pyx_v_isplit = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("randomSplit", 0);

  /* "gnocis/regions.pyx":177
 * 		:rtype: regions
 * 		"""
 * 		rgns = self.regions[:]             # <<<<<<<<<<<<<<
 * 		isplit = int(len(rgns)*ratio)
 * 		random.shuffle(rgns)
 */
  if (unlikely(__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 177, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GetSlice(__pyx_v_self->regions, 0, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rgns = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":178
 * 		"""
 * 		rgns = self.regions[:]
 * 		isplit = int(len(rgns)*ratio)             # <<<<<<<<<<<<<<
 * 		random.shuffle(rgns)
 * 		return regions(self.name + ' (split A)', rgns[:isplit]),\
 */
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_rgns); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 178, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_v_ratio); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_isplit = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":179
 * 		rgns = self.regions[:]
 * 		isplit = int(len(rgns)*ratio)
 * 		random.shuffle(rgns)             # <<<<<<<<<<<<<<
 * 		return regions(self.name + ' (split A)', rgns[:isplit]),\
 * 			   regions(self.name + ' (split B)', rgns[isplit:])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_random); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_shuffle); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_rgns) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_rgns);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":180
 * 		isplit = int(len(rgns)*ratio)
 * 		random.shuffle(rgns)
 * 		return regions(self.name + ' (split A)', rgns[:isplit]),\             # <<<<<<<<<<<<<<
 * 			   regions(self.name + ' (split B)', rgns[isplit:])
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_self->name, __pyx_kp_u_split_A); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_isplit);
  __pyx_t_4 = __pyx_v_isplit;
  __pyx_t_5 = (__pyx_t_4 == Py_None);
  if (__pyx_t_5) {
    __pyx_t_2 = PY_SSIZE_T_MAX;
  } else {
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_t_4); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 180, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_6;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyList_GetSlice(__pyx_v_rgns, 0, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/regions.pyx":181
 * 		random.shuffle(rgns)
 * 		return regions(self.name + ' (split A)', rgns[:isplit]),\
 * 			   regions(self.name + ' (split B)', rgns[isplit:])             # <<<<<<<<<<<<<<
 * 
 * 	# Returns a random sample of the regopns
 */
  __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_self->name, __pyx_kp_u_split_B); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_isplit);
  __pyx_t_1 = __pyx_v_isplit;
  __pyx_t_5 = (__pyx_t_1 == Py_None);
  if (__pyx_t_5) {
    __pyx_t_2 = 0;
  } else {
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 181, __pyx_L1_error)
    __pyx_t_2 = __pyx_t_6;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyList_GetSlice(__pyx_v_rgns, __pyx_t_2, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "gnocis/regions.pyx":180
 * 		isplit = int(len(rgns)*ratio)
 * 		random.shuffle(rgns)
 * 		return regions(self.name + ' (split A)', rgns[:isplit]),\             # <<<<<<<<<<<<<<
 * 			   regions(self.name + ' (split B)', rgns[isplit:])
 * 
 */
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":169
 * 
 * 	# Returns a random division of the regions
 * 	def randomSplit(self, ratio = 0.5):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:param ratio: Ratio for the split, between 0 and 1. For the return touple, `(A, B)`, a ratio of 0 means all the regions are in `B`, and opposite for a ratio of 1.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.regions.regions.randomSplit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rgns);
  __Pyx_XDECREF(__pyx_v_isplit);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":184
 * 
 * 	# Returns a random sample of the regopns
 * 	def sample(self, n = 0):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:param n: Number of regions to pick. If 0, the same number will be selected as are in the full set.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_31sample(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_30sample[] = "\n\t\t:param n: Number of regions to pick. If 0, the same number will be selected as are in the full set.\n\t\t:type n: int\n\t\t\n\t\t:return: Returns a random sample of the regions of size n. Regions are selected with replacement, and the same region can occur multiple times.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_31sample(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_n = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sample (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sample") < 0)) __PYX_ERR(0, 184, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_n = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sample", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.sample", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_30sample(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_30sample(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_n) {
  PyObject *__pyx_v_smp = NULL;
  CYTHON_UNUSED PyObject *__pyx_8genexpr6__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sample", 0);
  __Pyx_INCREF(__pyx_v_n);

  /* "gnocis/regions.pyx":192
 * 		:rtype: regions
 * 		"""
 * 		if n <= 0: n = len(self)             # <<<<<<<<<<<<<<
 * 		smp = [ random.choice(self.regions) for i in range(n) ]
 * 		return regions(self.name + ' (bootstrap sample of ' + str(n) + ' regions)', smp)
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_n, __pyx_int_0, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {
    __pyx_t_3 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 192, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_n, __pyx_t_1);
    __pyx_t_1 = 0;
  }

  /* "gnocis/regions.pyx":193
 * 		"""
 * 		if n <= 0: n = len(self)
 * 		smp = [ random.choice(self.regions) for i in range(n) ]             # <<<<<<<<<<<<<<
 * 		return regions(self.name + ' (bootstrap sample of ' + str(n) + ' regions)', smp)
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_n); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_3 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 193, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 193, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 193, __pyx_L6_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 193, __pyx_L6_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_6(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 193, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_i, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_random); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 193, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_choice); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 193, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_self->regions) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_self->regions);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 193, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_i); __pyx_8genexpr6__pyx_v_i = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_i); __pyx_8genexpr6__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_smp = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":194
 * 		if n <= 0: n = len(self)
 * 		smp = [ random.choice(self.regions) for i in range(n) ]
 * 		return regions(self.name + ' (bootstrap sample of ' + str(n) + ' regions)', smp)             # <<<<<<<<<<<<<<
 * 
 * 	# Returns a filtered subset of regions.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_self->name, __pyx_kp_u_bootstrap_sample_of); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_t_4, __pyx_kp_u_regions_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_smp);
  __Pyx_GIVEREF(__pyx_v_smp);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_smp);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":184
 * 
 * 	# Returns a random sample of the regopns
 * 	def sample(self, n = 0):             # <<<<<<<<<<<<<<
 * 		"""
 * 		:param n: Number of regions to pick. If 0, the same number will be selected as are in the full set.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("gnocis.regions.regions.sample", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_smp);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":197
 * 
 * 	# Returns a filtered subset of regions.
 * 	def filter(self, fltName, flt):             # <<<<<<<<<<<<<<
 * 		""" Returns a region set filtered with the input lambda function.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_33filter(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_32filter[] = " Returns a region set filtered with the input lambda function.\n\t\t\n\t\t:param fltName: Name of the filter. Appended to the region set name.\n\t\t:param flt: Lambda function to be applied to every region, returning `True` if the region is to be included, or otherwise `False`.\n\t\t:type fltName: str\n\t\t:type flt: lambda\n\t\t\n\t\t:return: Region set filtered with the lambda function.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_33filter(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fltName = 0;
  PyObject *__pyx_v_flt = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("filter (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fltName,&__pyx_n_s_flt,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fltName)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("filter", 1, 2, 2, 1); __PYX_ERR(0, 197, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "filter") < 0)) __PYX_ERR(0, 197, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_fltName = values[0];
    __pyx_v_flt = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("filter", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 197, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.filter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_32filter(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), __pyx_v_fltName, __pyx_v_flt);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_32filter(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_fltName, PyObject *__pyx_v_flt) {
  PyObject *__pyx_8genexpr7__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("filter", 0);

  /* "gnocis/regions.pyx":208
 * 		:rtype: regions
 * 		"""
 * 		return regions(self.name + ' (' + fltName + ')', [ r for r in self.regions if flt(r) ])             # <<<<<<<<<<<<<<
 * 
 * 	# Returns the merged region set for this and another set.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_self->name, __pyx_kp_u__5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_v_fltName); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_kp_u__8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_v_self->regions == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 208, __pyx_L5_error)
    }
    __pyx_t_3 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    for (;;) {
      if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 208, __pyx_L5_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 208, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_r, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_v_flt);
      __pyx_t_6 = __pyx_v_flt; __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_8genexpr7__pyx_v_r) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_8genexpr7__pyx_v_r);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 208, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 208, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_8) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_8genexpr7__pyx_v_r))) __PYX_ERR(0, 208, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_r); __pyx_8genexpr7__pyx_v_r = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_r); __pyx_8genexpr7__pyx_v_r = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":197
 * 
 * 	# Returns a filtered subset of regions.
 * 	def filter(self, fltName, flt):             # <<<<<<<<<<<<<<
 * 		""" Returns a region set filtered with the input lambda function.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.regions.regions.filter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":211
 * 
 * 	# Returns the merged region set for this and another set.
 * 	def flatten(self):             # <<<<<<<<<<<<<<
 * 		""" Returns a flattened set of regions, with internally overlapping regions merged.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_35flatten(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_34flatten[] = " Returns a flattened set of regions, with internally overlapping regions merged.\n\t\t\n\t\t:return: Flattened region set.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_35flatten(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("flatten (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_34flatten(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_34flatten(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("flatten", 0);

  /* "gnocis/regions.pyx":217
 * 		:rtype: regions
 * 		"""
 * 		return self.merge(regions('', []))             # <<<<<<<<<<<<<<
 * 
 * 	# Returns the merged region set for this and another set.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_merge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_u__13);
  __Pyx_GIVEREF(__pyx_kp_u__13);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u__13);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":211
 * 
 * 	# Returns the merged region set for this and another set.
 * 	def flatten(self):             # <<<<<<<<<<<<<<
 * 		""" Returns a flattened set of regions, with internally overlapping regions merged.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.regions.regions.flatten", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":220
 * 
 * 	# Returns the merged region set for this and another set.
 * 	def merge(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the merged set of regions in this set and another.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_37merge(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_36merge[] = " Gets the merged set of regions in this set and another.\n\t\t\n\t\t:param other: Other region set.\n\t\t:type other: regions\n\t\t\n\t\t:return: Merged region set.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_37merge(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("merge (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gnocis_7regions_regions, 1, "other", 0))) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_36merge(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":245
 * 		sortedKeys = sorted(rdict.keys())
 * 		for cseq in sortedKeys:
 * 			rdict[cseq] = sorted(rdict[cseq], key = lambda x: x.start)             # <<<<<<<<<<<<<<
 * 		lastRgn = region(rlist[0].seq, rlist[0].start, rlist[0].end)
 * 		newlist = [ lastRgn ]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_5merge_lambda1(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_7regions_5merge_lambda1 = {"lambda1", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_5merge_lambda1, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_7regions_7regions_5merge_lambda1(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda1 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda1(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda1", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.regions.merge.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":220
 * 
 * 	# Returns the merged region set for this and another set.
 * 	def merge(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the merged set of regions in this set and another.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_36merge(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other) {
  PyObject *__pyx_v_rlist = 0;
  PyObject *__pyx_v_newlist = 0;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rgn = 0;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_lastRgn = 0;
  PyObject *__pyx_v_rdict = 0;
  PyObject *__pyx_v_sortedKeys = 0;
  PyObject *__pyx_v_cseq = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PY_LONG_LONG __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("merge", 0);

  /* "gnocis/regions.pyx":234
 * 		cdef dict rdict
 * 		cdef list sortedKeys
 * 		rlist = self.regions + other.regions             # <<<<<<<<<<<<<<
 * 		if len(rlist) == 0:
 * 			return regions('%s | %s'%(self.name, other.name), [], initialSort = False)
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_self->regions, __pyx_v_other->regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rlist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":235
 * 		cdef list sortedKeys
 * 		rlist = self.regions + other.regions
 * 		if len(rlist) == 0:             # <<<<<<<<<<<<<<
 * 			return regions('%s | %s'%(self.name, other.name), [], initialSort = False)
 * 		rdict = {}
 */
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_rlist); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 235, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 == 0) != 0);
  if (__pyx_t_3) {

    /* "gnocis/regions.pyx":236
 * 		rlist = self.regions + other.regions
 * 		if len(rlist) == 0:
 * 			return regions('%s | %s'%(self.name, other.name), [], initialSort = False)             # <<<<<<<<<<<<<<
 * 		rdict = {}
 * 		for rgn in rlist:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_4 = 127;
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__14);
    __pyx_t_2 += 3;
    __Pyx_GIVEREF(__pyx_kp_u__14);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__14);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_other->name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_initialSort, Py_False) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "gnocis/regions.pyx":235
 * 		cdef list sortedKeys
 * 		rlist = self.regions + other.regions
 * 		if len(rlist) == 0:             # <<<<<<<<<<<<<<
 * 			return regions('%s | %s'%(self.name, other.name), [], initialSort = False)
 * 		rdict = {}
 */
  }

  /* "gnocis/regions.pyx":237
 * 		if len(rlist) == 0:
 * 			return regions('%s | %s'%(self.name, other.name), [], initialSort = False)
 * 		rdict = {}             # <<<<<<<<<<<<<<
 * 		for rgn in rlist:
 * 			if not rgn.seq in rdict:
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_rdict = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":238
 * 			return regions('%s | %s'%(self.name, other.name), [], initialSort = False)
 * 		rdict = {}
 * 		for rgn in rlist:             # <<<<<<<<<<<<<<
 * 			if not rgn.seq in rdict:
 * 				rdict[rgn.seq] = [rgn]
 */
  __pyx_t_5 = __pyx_v_rlist; __Pyx_INCREF(__pyx_t_5); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_5)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_2); __Pyx_INCREF(__pyx_t_1); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 238, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 238, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_rgn, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/regions.pyx":239
 * 		rdict = {}
 * 		for rgn in rlist:
 * 			if not rgn.seq in rdict:             # <<<<<<<<<<<<<<
 * 				rdict[rgn.seq] = [rgn]
 * 			else:
 */
    __pyx_t_3 = (__Pyx_PyDict_ContainsTF(__pyx_v_rgn->seq, __pyx_v_rdict, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 239, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_3 != 0);
    if (__pyx_t_7) {

      /* "gnocis/regions.pyx":240
 * 		for rgn in rlist:
 * 			if not rgn.seq in rdict:
 * 				rdict[rgn.seq] = [rgn]             # <<<<<<<<<<<<<<
 * 			else:
 * 				rdict[rgn.seq].append(rgn)
 */
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(((PyObject *)__pyx_v_rgn));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_rgn));
      PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_rgn));
      if (unlikely(PyDict_SetItem(__pyx_v_rdict, __pyx_v_rgn->seq, __pyx_t_1) < 0)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "gnocis/regions.pyx":239
 * 		rdict = {}
 * 		for rgn in rlist:
 * 			if not rgn.seq in rdict:             # <<<<<<<<<<<<<<
 * 				rdict[rgn.seq] = [rgn]
 * 			else:
 */
      goto __pyx_L6;
    }

    /* "gnocis/regions.pyx":242
 * 				rdict[rgn.seq] = [rgn]
 * 			else:
 * 				rdict[rgn.seq].append(rgn)             # <<<<<<<<<<<<<<
 * 		sortedKeys = sorted(rdict.keys())
 * 		for cseq in sortedKeys:
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_rdict, __pyx_v_rgn->seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_Append(__pyx_t_1, ((PyObject *)__pyx_v_rgn)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 242, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L6:;

    /* "gnocis/regions.pyx":238
 * 			return regions('%s | %s'%(self.name, other.name), [], initialSort = False)
 * 		rdict = {}
 * 		for rgn in rlist:             # <<<<<<<<<<<<<<
 * 			if not rgn.seq in rdict:
 * 				rdict[rgn.seq] = [rgn]
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":243
 * 			else:
 * 				rdict[rgn.seq].append(rgn)
 * 		sortedKeys = sorted(rdict.keys())             # <<<<<<<<<<<<<<
 * 		for cseq in sortedKeys:
 * 			rdict[cseq] = sorted(rdict[cseq], key = lambda x: x.start)
 */
  __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_rdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_8 = PyList_Sort(__pyx_t_5); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 243, __pyx_L1_error)
  __pyx_v_sortedKeys = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":244
 * 				rdict[rgn.seq].append(rgn)
 * 		sortedKeys = sorted(rdict.keys())
 * 		for cseq in sortedKeys:             # <<<<<<<<<<<<<<
 * 			rdict[cseq] = sorted(rdict[cseq], key = lambda x: x.start)
 * 		lastRgn = region(rlist[0].seq, rlist[0].start, rlist[0].end)
 */
  if (unlikely(__pyx_v_sortedKeys == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 244, __pyx_L1_error)
  }
  __pyx_t_5 = __pyx_v_sortedKeys; __Pyx_INCREF(__pyx_t_5); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_5)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_2); __Pyx_INCREF(__pyx_t_6); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 244, __pyx_L1_error)
    #else
    __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_cseq, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "gnocis/regions.pyx":245
 * 		sortedKeys = sorted(rdict.keys())
 * 		for cseq in sortedKeys:
 * 			rdict[cseq] = sorted(rdict[cseq], key = lambda x: x.start)             # <<<<<<<<<<<<<<
 * 		lastRgn = region(rlist[0].seq, rlist[0].start, rlist[0].end)
 * 		newlist = [ lastRgn ]
 */
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_rdict, __pyx_v_cseq); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_7regions_7regions_5merge_lambda1, 0, __pyx_n_s_merge_locals_lambda, NULL, __pyx_n_s_gnocis_regions, __pyx_d, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_key, __pyx_t_9) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(PyDict_SetItem(__pyx_v_rdict, __pyx_v_cseq, __pyx_t_9) < 0)) __PYX_ERR(0, 245, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "gnocis/regions.pyx":244
 * 				rdict[rgn.seq].append(rgn)
 * 		sortedKeys = sorted(rdict.keys())
 * 		for cseq in sortedKeys:             # <<<<<<<<<<<<<<
 * 			rdict[cseq] = sorted(rdict[cseq], key = lambda x: x.start)
 * 		lastRgn = region(rlist[0].seq, rlist[0].start, rlist[0].end)
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":246
 * 		for cseq in sortedKeys:
 * 			rdict[cseq] = sorted(rdict[cseq], key = lambda x: x.start)
 * 		lastRgn = region(rlist[0].seq, rlist[0].start, rlist[0].end)             # <<<<<<<<<<<<<<
 * 		newlist = [ lastRgn ]
 * 		for cseq in sortedKeys:
 */
  __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_rlist, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_seq); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_rlist, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_rlist, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_1);
  __pyx_t_9 = 0;
  __pyx_t_6 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_lastRgn = ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":247
 * 			rdict[cseq] = sorted(rdict[cseq], key = lambda x: x.start)
 * 		lastRgn = region(rlist[0].seq, rlist[0].start, rlist[0].end)
 * 		newlist = [ lastRgn ]             # <<<<<<<<<<<<<<
 * 		for cseq in sortedKeys:
 * 			for rgn in rdict[cseq]:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_lastRgn));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_lastRgn));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_lastRgn));
  __pyx_v_newlist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":248
 * 		lastRgn = region(rlist[0].seq, rlist[0].start, rlist[0].end)
 * 		newlist = [ lastRgn ]
 * 		for cseq in sortedKeys:             # <<<<<<<<<<<<<<
 * 			for rgn in rdict[cseq]:
 * 				if lastRgn.end+1 < rgn.start or lastRgn.seq != rgn.seq:
 */
  if (unlikely(__pyx_v_sortedKeys == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 248, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_sortedKeys; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 248, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_cseq, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "gnocis/regions.pyx":249
 * 		newlist = [ lastRgn ]
 * 		for cseq in sortedKeys:
 * 			for rgn in rdict[cseq]:             # <<<<<<<<<<<<<<
 * 				if lastRgn.end+1 < rgn.start or lastRgn.seq != rgn.seq:
 * 					lastRgn = region(rgn.seq, rgn.start, rgn.end)
 */
    __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_rdict, __pyx_v_cseq); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_6 = __pyx_t_5; __Pyx_INCREF(__pyx_t_6); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 249, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 249, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 249, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 249, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 249, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 249, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_11(__pyx_t_6);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 249, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 249, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_rgn, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "gnocis/regions.pyx":250
 * 		for cseq in sortedKeys:
 * 			for rgn in rdict[cseq]:
 * 				if lastRgn.end+1 < rgn.start or lastRgn.seq != rgn.seq:             # <<<<<<<<<<<<<<
 * 					lastRgn = region(rgn.seq, rgn.start, rgn.end)
 * 					newlist.append( lastRgn )
 */
      __pyx_t_3 = (((__pyx_v_lastRgn->end + 1) < __pyx_v_rgn->start) != 0);
      if (!__pyx_t_3) {
      } else {
        __pyx_t_7 = __pyx_t_3;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_lastRgn->seq, __pyx_v_rgn->seq, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 250, __pyx_L1_error)
      __pyx_t_12 = (__pyx_t_3 != 0);
      __pyx_t_7 = __pyx_t_12;
      __pyx_L14_bool_binop_done:;
      if (__pyx_t_7) {

        /* "gnocis/regions.pyx":251
 * 			for rgn in rdict[cseq]:
 * 				if lastRgn.end+1 < rgn.start or lastRgn.seq != rgn.seq:
 * 					lastRgn = region(rgn.seq, rgn.start, rgn.end)             # <<<<<<<<<<<<<<
 * 					newlist.append( lastRgn )
 * 				elif lastRgn.end < rgn.end:
 */
        __pyx_t_5 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_rgn->start); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_9 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_rgn->end); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_INCREF(__pyx_v_rgn->seq);
        __Pyx_GIVEREF(__pyx_v_rgn->seq);
        PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_rgn->seq);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_t_9);
        __pyx_t_5 = 0;
        __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_13, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF_SET(__pyx_v_lastRgn, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_9));
        __pyx_t_9 = 0;

        /* "gnocis/regions.pyx":252
 * 				if lastRgn.end+1 < rgn.start or lastRgn.seq != rgn.seq:
 * 					lastRgn = region(rgn.seq, rgn.start, rgn.end)
 * 					newlist.append( lastRgn )             # <<<<<<<<<<<<<<
 * 				elif lastRgn.end < rgn.end:
 * 					lastRgn.end = rgn.end
 */
        __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_newlist, ((PyObject *)__pyx_v_lastRgn)); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 252, __pyx_L1_error)

        /* "gnocis/regions.pyx":250
 * 		for cseq in sortedKeys:
 * 			for rgn in rdict[cseq]:
 * 				if lastRgn.end+1 < rgn.start or lastRgn.seq != rgn.seq:             # <<<<<<<<<<<<<<
 * 					lastRgn = region(rgn.seq, rgn.start, rgn.end)
 * 					newlist.append( lastRgn )
 */
        goto __pyx_L13;
      }

      /* "gnocis/regions.pyx":253
 * 					lastRgn = region(rgn.seq, rgn.start, rgn.end)
 * 					newlist.append( lastRgn )
 * 				elif lastRgn.end < rgn.end:             # <<<<<<<<<<<<<<
 * 					lastRgn.end = rgn.end
 * 		return regions('%s | %s'%(self.name, other.name), newlist, initialSort = False)
 */
      __pyx_t_7 = ((__pyx_v_lastRgn->end < __pyx_v_rgn->end) != 0);
      if (__pyx_t_7) {

        /* "gnocis/regions.pyx":254
 * 					newlist.append( lastRgn )
 * 				elif lastRgn.end < rgn.end:
 * 					lastRgn.end = rgn.end             # <<<<<<<<<<<<<<
 * 		return regions('%s | %s'%(self.name, other.name), newlist, initialSort = False)
 * 
 */
        __pyx_t_14 = __pyx_v_rgn->end;
        __pyx_v_lastRgn->end = __pyx_t_14;

        /* "gnocis/regions.pyx":253
 * 					lastRgn = region(rgn.seq, rgn.start, rgn.end)
 * 					newlist.append( lastRgn )
 * 				elif lastRgn.end < rgn.end:             # <<<<<<<<<<<<<<
 * 					lastRgn.end = rgn.end
 * 		return regions('%s | %s'%(self.name, other.name), newlist, initialSort = False)
 */
      }
      __pyx_L13:;

      /* "gnocis/regions.pyx":249
 * 		newlist = [ lastRgn ]
 * 		for cseq in sortedKeys:
 * 			for rgn in rdict[cseq]:             # <<<<<<<<<<<<<<
 * 				if lastRgn.end+1 < rgn.start or lastRgn.seq != rgn.seq:
 * 					lastRgn = region(rgn.seq, rgn.start, rgn.end)
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "gnocis/regions.pyx":248
 * 		lastRgn = region(rlist[0].seq, rlist[0].start, rlist[0].end)
 * 		newlist = [ lastRgn ]
 * 		for cseq in sortedKeys:             # <<<<<<<<<<<<<<
 * 			for rgn in rdict[cseq]:
 * 				if lastRgn.end+1 < rgn.start or lastRgn.seq != rgn.seq:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":255
 * 				elif lastRgn.end < rgn.end:
 * 					lastRgn.end = rgn.end
 * 		return regions('%s | %s'%(self.name, other.name), newlist, initialSort = False)             # <<<<<<<<<<<<<<
 * 
 * 	# Returns the intersection of this set and another set.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_4 = 127;
  __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_4;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __Pyx_INCREF(__pyx_kp_u__14);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__14);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__14);
  __pyx_t_6 = __Pyx_PyUnicode_Unicode(__pyx_v_other->name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_4;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
  __Pyx_INCREF(__pyx_v_newlist);
  __Pyx_GIVEREF(__pyx_v_newlist);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_newlist);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_initialSort, Py_False) < 0) __PYX_ERR(0, 255, __pyx_L1_error)
  __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_9;
  __pyx_t_9 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":220
 * 
 * 	# Returns the merged region set for this and another set.
 * 	def merge(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the merged set of regions in this set and another.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("gnocis.regions.regions.merge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rlist);
  __Pyx_XDECREF(__pyx_v_newlist);
  __Pyx_XDECREF((PyObject *)__pyx_v_rgn);
  __Pyx_XDECREF((PyObject *)__pyx_v_lastRgn);
  __Pyx_XDECREF(__pyx_v_rdict);
  __Pyx_XDECREF(__pyx_v_sortedKeys);
  __Pyx_XDECREF(__pyx_v_cseq);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":258
 * 
 * 	# Returns the intersection of this set and another set.
 * 	def intersection(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the intersection of regions in this set and another.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_39intersection(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_38intersection[] = " Gets the intersection of regions in this set and another.\n\t\t\n\t\t:param other: Other region set.\n\t\t:type other: regions\n\t\t\n\t\t:return: Intersection region set.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_39intersection(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("intersection (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gnocis_7regions_regions, 1, "other", 0))) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_38intersection(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_38intersection(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other) {
  PyObject *__pyx_v_intersection = 0;
  PyObject *__pyx_v_rBs = 0;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rA = 0;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rB = 0;
  int __pyx_v_iB;
  int __pyx_v_firstRelevantIndexB;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PY_LONG_LONG __pyx_t_10;
  PY_LONG_LONG __pyx_t_11;
  PY_LONG_LONG __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  Py_UCS4 __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("intersection", 0);

  /* "gnocis/regions.pyx":271
 * 		cdef region rA, rB
 * 		cdef int iB, firstRelevantIndexB
 * 		intersection = []             # <<<<<<<<<<<<<<
 * 		rBs = other.regions
 * 		for rA in self.regions:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_intersection = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":272
 * 		cdef int iB, firstRelevantIndexB
 * 		intersection = []
 * 		rBs = other.regions             # <<<<<<<<<<<<<<
 * 		for rA in self.regions:
 * 			firstRelevantIndexB = 0
 */
  __pyx_t_1 = __pyx_v_other->regions;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_rBs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":273
 * 		intersection = []
 * 		rBs = other.regions
 * 		for rA in self.regions:             # <<<<<<<<<<<<<<
 * 			firstRelevantIndexB = 0
 * 			for iB,rB in enumerate(rBs):
 */
  if (unlikely(__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 273, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 273, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_rA, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":274
 * 		rBs = other.regions
 * 		for rA in self.regions:
 * 			firstRelevantIndexB = 0             # <<<<<<<<<<<<<<
 * 			for iB,rB in enumerate(rBs):
 * 				if rB.seq > rA.seq or (rB.seq == rA.seq and rA.end < rB.start):
 */
    __pyx_v_firstRelevantIndexB = 0;

    /* "gnocis/regions.pyx":275
 * 		for rA in self.regions:
 * 			firstRelevantIndexB = 0
 * 			for iB,rB in enumerate(rBs):             # <<<<<<<<<<<<<<
 * 				if rB.seq > rA.seq or (rB.seq == rA.seq and rA.end < rB.start):
 * 					break
 */
    __pyx_t_4 = 0;
    __pyx_t_3 = __pyx_v_rBs; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
    for (;;) {
      if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 275, __pyx_L1_error)
      #else
      __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_rB, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_6));
      __pyx_t_6 = 0;
      __pyx_v_iB = __pyx_t_4;
      __pyx_t_4 = (__pyx_t_4 + 1);

      /* "gnocis/regions.pyx":276
 * 			firstRelevantIndexB = 0
 * 			for iB,rB in enumerate(rBs):
 * 				if rB.seq > rA.seq or (rB.seq == rA.seq and rA.end < rB.start):             # <<<<<<<<<<<<<<
 * 					break
 * 				if rB.seq < rA.seq or (rB.seq == rA.seq and rB.end < rA.start):
 */
      __pyx_t_6 = PyObject_RichCompare(__pyx_v_rB->seq, __pyx_v_rA->seq, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 276, __pyx_L1_error)
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 276, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (!__pyx_t_8) {
      } else {
        __pyx_t_7 = __pyx_t_8;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_rB->seq, __pyx_v_rA->seq, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 276, __pyx_L1_error)
      __pyx_t_9 = (__pyx_t_8 != 0);
      if (__pyx_t_9) {
      } else {
        __pyx_t_7 = __pyx_t_9;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_9 = ((__pyx_v_rA->end < __pyx_v_rB->start) != 0);
      __pyx_t_7 = __pyx_t_9;
      __pyx_L8_bool_binop_done:;
      if (__pyx_t_7) {

        /* "gnocis/regions.pyx":277
 * 			for iB,rB in enumerate(rBs):
 * 				if rB.seq > rA.seq or (rB.seq == rA.seq and rA.end < rB.start):
 * 					break             # <<<<<<<<<<<<<<
 * 				if rB.seq < rA.seq or (rB.seq == rA.seq and rB.end < rA.start):
 * 					firstRelevantIndexB = iB
 */
        goto __pyx_L6_break;

        /* "gnocis/regions.pyx":276
 * 			firstRelevantIndexB = 0
 * 			for iB,rB in enumerate(rBs):
 * 				if rB.seq > rA.seq or (rB.seq == rA.seq and rA.end < rB.start):             # <<<<<<<<<<<<<<
 * 					break
 * 				if rB.seq < rA.seq or (rB.seq == rA.seq and rB.end < rA.start):
 */
      }

      /* "gnocis/regions.pyx":278
 * 				if rB.seq > rA.seq or (rB.seq == rA.seq and rA.end < rB.start):
 * 					break
 * 				if rB.seq < rA.seq or (rB.seq == rA.seq and rB.end < rA.start):             # <<<<<<<<<<<<<<
 * 					firstRelevantIndexB = iB
 * 					continue
 */
      __pyx_t_6 = PyObject_RichCompare(__pyx_v_rB->seq, __pyx_v_rA->seq, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 278, __pyx_L1_error)
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 278, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (!__pyx_t_9) {
      } else {
        __pyx_t_7 = __pyx_t_9;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_v_rB->seq, __pyx_v_rA->seq, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 278, __pyx_L1_error)
      __pyx_t_8 = (__pyx_t_9 != 0);
      if (__pyx_t_8) {
      } else {
        __pyx_t_7 = __pyx_t_8;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_8 = ((__pyx_v_rB->end < __pyx_v_rA->start) != 0);
      __pyx_t_7 = __pyx_t_8;
      __pyx_L12_bool_binop_done:;
      if (__pyx_t_7) {

        /* "gnocis/regions.pyx":279
 * 					break
 * 				if rB.seq < rA.seq or (rB.seq == rA.seq and rB.end < rA.start):
 * 					firstRelevantIndexB = iB             # <<<<<<<<<<<<<<
 * 					continue
 * 				intersection.append( region(rA.seq, max(rA.start, rB.start), min(rA.end, rB.end)) )
 */
        __pyx_v_firstRelevantIndexB = __pyx_v_iB;

        /* "gnocis/regions.pyx":280
 * 				if rB.seq < rA.seq or (rB.seq == rA.seq and rB.end < rA.start):
 * 					firstRelevantIndexB = iB
 * 					continue             # <<<<<<<<<<<<<<
 * 				intersection.append( region(rA.seq, max(rA.start, rB.start), min(rA.end, rB.end)) )
 * 			if firstRelevantIndexB > 0:
 */
        goto __pyx_L5_continue;

        /* "gnocis/regions.pyx":278
 * 				if rB.seq > rA.seq or (rB.seq == rA.seq and rA.end < rB.start):
 * 					break
 * 				if rB.seq < rA.seq or (rB.seq == rA.seq and rB.end < rA.start):             # <<<<<<<<<<<<<<
 * 					firstRelevantIndexB = iB
 * 					continue
 */
      }

      /* "gnocis/regions.pyx":281
 * 					firstRelevantIndexB = iB
 * 					continue
 * 				intersection.append( region(rA.seq, max(rA.start, rB.start), min(rA.end, rB.end)) )             # <<<<<<<<<<<<<<
 * 			if firstRelevantIndexB > 0:
 * 				rBs = rBs[firstRelevantIndexB:]
 */
      __pyx_t_10 = __pyx_v_rB->start;
      __pyx_t_11 = __pyx_v_rA->start;
      if (((__pyx_t_10 > __pyx_t_11) != 0)) {
        __pyx_t_12 = __pyx_t_10;
      } else {
        __pyx_t_12 = __pyx_t_11;
      }
      __pyx_t_6 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_12 = __pyx_v_rB->end;
      __pyx_t_10 = __pyx_v_rA->end;
      if (((__pyx_t_12 < __pyx_t_10) != 0)) {
        __pyx_t_11 = __pyx_t_12;
      } else {
        __pyx_t_11 = __pyx_t_10;
      }
      __pyx_t_13 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = PyTuple_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_v_rA->seq);
      __Pyx_GIVEREF(__pyx_v_rA->seq);
      PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_v_rA->seq);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_13);
      PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_t_13);
      __pyx_t_6 = 0;
      __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_14, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_intersection, __pyx_t_13); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "gnocis/regions.pyx":275
 * 		for rA in self.regions:
 * 			firstRelevantIndexB = 0
 * 			for iB,rB in enumerate(rBs):             # <<<<<<<<<<<<<<
 * 				if rB.seq > rA.seq or (rB.seq == rA.seq and rA.end < rB.start):
 * 					break
 */
      __pyx_L5_continue:;
    }
    __pyx_L6_break:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":282
 * 					continue
 * 				intersection.append( region(rA.seq, max(rA.start, rB.start), min(rA.end, rB.end)) )
 * 			if firstRelevantIndexB > 0:             # <<<<<<<<<<<<<<
 * 				rBs = rBs[firstRelevantIndexB:]
 * 		return regions('%s & %s'%(self.name, other.name), intersection)
 */
    __pyx_t_7 = ((__pyx_v_firstRelevantIndexB > 0) != 0);
    if (__pyx_t_7) {

      /* "gnocis/regions.pyx":283
 * 				intersection.append( region(rA.seq, max(rA.start, rB.start), min(rA.end, rB.end)) )
 * 			if firstRelevantIndexB > 0:
 * 				rBs = rBs[firstRelevantIndexB:]             # <<<<<<<<<<<<<<
 * 		return regions('%s & %s'%(self.name, other.name), intersection)
 * 
 */
      if (unlikely(__pyx_v_rBs == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 283, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyList_GetSlice(__pyx_v_rBs, __pyx_v_firstRelevantIndexB, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_rBs, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "gnocis/regions.pyx":282
 * 					continue
 * 				intersection.append( region(rA.seq, max(rA.start, rB.start), min(rA.end, rB.end)) )
 * 			if firstRelevantIndexB > 0:             # <<<<<<<<<<<<<<
 * 				rBs = rBs[firstRelevantIndexB:]
 * 		return regions('%s & %s'%(self.name, other.name), intersection)
 */
    }

    /* "gnocis/regions.pyx":273
 * 		intersection = []
 * 		rBs = other.regions
 * 		for rA in self.regions:             # <<<<<<<<<<<<<<
 * 			firstRelevantIndexB = 0
 * 			for iB,rB in enumerate(rBs):
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":284
 * 			if firstRelevantIndexB > 0:
 * 				rBs = rBs[firstRelevantIndexB:]
 * 		return regions('%s & %s'%(self.name, other.name), intersection)             # <<<<<<<<<<<<<<
 * 
 * 	# Returns the excluded region set for this and another set.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_16 = 127;
  __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_16;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u__15);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__15);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__15);
  __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_other->name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_16;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_intersection);
  __Pyx_GIVEREF(__pyx_v_intersection);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_intersection);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":258
 * 
 * 	# Returns the intersection of this set and another set.
 * 	def intersection(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the intersection of regions in this set and another.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("gnocis.regions.regions.intersection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_intersection);
  __Pyx_XDECREF(__pyx_v_rBs);
  __Pyx_XDECREF((PyObject *)__pyx_v_rA);
  __Pyx_XDECREF((PyObject *)__pyx_v_rB);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":287
 * 
 * 	# Returns the excluded region set for this and another set.
 * 	def exclusion(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the regions of this set with regions of another set excluded.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_41exclusion(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_40exclusion[] = " Gets the regions of this set with regions of another set excluded.\n\t\t\n\t\t:param other: Other region set.\n\t\t:type other: regions\n\t\t\n\t\t:return: Excluded region set.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_41exclusion(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exclusion (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gnocis_7regions_regions, 1, "other", 0))) __PYX_ERR(0, 287, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_40exclusion(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_40exclusion(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other) {
  PyObject *__pyx_v_output = NULL;
  PyObject *__pyx_v_exSet = NULL;
  PyObject *__pyx_v_i = NULL;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_oi = NULL;
  PyObject *__pyx_v_firstRelevantIndex = NULL;
  PyObject *__pyx_v_ei = NULL;
  PyObject *__pyx_v_e = NULL;
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_rs = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  PY_LONG_LONG __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_UCS4 __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("exclusion", 0);

  /* "gnocis/regions.pyx":296
 * 		:rtype: regions
 * 		"""
 * 		output = []             # <<<<<<<<<<<<<<
 * 		exSet = other.regions
 * 		for i in self.regions:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_output = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":297
 * 		"""
 * 		output = []
 * 		exSet = other.regions             # <<<<<<<<<<<<<<
 * 		for i in self.regions:
 * 			oi = region(i.seq, i.start, i.end, i.strand, i.score, i.source, i.feature, group = i.group)
 */
  __pyx_t_1 = __pyx_v_other->regions;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_exSet = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":298
 * 		output = []
 * 		exSet = other.regions
 * 		for i in self.regions:             # <<<<<<<<<<<<<<
 * 			oi = region(i.seq, i.start, i.end, i.strand, i.score, i.source, i.feature, group = i.group)
 * 			firstRelevantIndex = 0
 */
  if (unlikely(__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 298, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 298, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":299
 * 		exSet = other.regions
 * 		for i in self.regions:
 * 			oi = region(i.seq, i.start, i.end, i.strand, i.score, i.source, i.feature, group = i.group)             # <<<<<<<<<<<<<<
 * 			firstRelevantIndex = 0
 * 			for ei,e in enumerate(exSet):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_end); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_strand); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_score); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_source); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_feature); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyTuple_New(7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_9);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_group); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_group, __pyx_t_8) < 0) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_oi, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "gnocis/regions.pyx":300
 * 		for i in self.regions:
 * 			oi = region(i.seq, i.start, i.end, i.strand, i.score, i.source, i.feature, group = i.group)
 * 			firstRelevantIndex = 0             # <<<<<<<<<<<<<<
 * 			for ei,e in enumerate(exSet):
 * 				if e.seq < oi.seq or (e.seq == oi.seq and e.end < oi.start):
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_XDECREF_SET(__pyx_v_firstRelevantIndex, __pyx_int_0);

    /* "gnocis/regions.pyx":301
 * 			oi = region(i.seq, i.start, i.end, i.strand, i.score, i.source, i.feature, group = i.group)
 * 			firstRelevantIndex = 0
 * 			for ei,e in enumerate(exSet):             # <<<<<<<<<<<<<<
 * 				if e.seq < oi.seq or (e.seq == oi.seq and e.end < oi.start):
 * 					firstRelevantIndex = ei+1
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_8 = __pyx_int_0;
    __pyx_t_9 = __pyx_v_exSet; __Pyx_INCREF(__pyx_t_9); __pyx_t_11 = 0;
    for (;;) {
      if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_9)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_10 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_11); __Pyx_INCREF(__pyx_t_10); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 301, __pyx_L1_error)
      #else
      __pyx_t_10 = PySequence_ITEM(__pyx_t_9, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_ei, __pyx_t_8);
      __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_t_8, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_8);
      __pyx_t_8 = __pyx_t_10;
      __pyx_t_10 = 0;

      /* "gnocis/regions.pyx":302
 * 			firstRelevantIndex = 0
 * 			for ei,e in enumerate(exSet):
 * 				if e.seq < oi.seq or (e.seq == oi.seq and e.end < oi.start):             # <<<<<<<<<<<<<<
 * 					firstRelevantIndex = ei+1
 * 					continue
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_seq); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_10, __pyx_v_oi->seq, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (!__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_seq); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_7, __pyx_v_oi->seq, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_end); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = PyObject_RichCompare(__pyx_t_7, __pyx_t_10, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 302, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_12 = __pyx_t_13;
      __pyx_L8_bool_binop_done:;
      if (__pyx_t_12) {

        /* "gnocis/regions.pyx":303
 * 			for ei,e in enumerate(exSet):
 * 				if e.seq < oi.seq or (e.seq == oi.seq and e.end < oi.start):
 * 					firstRelevantIndex = ei+1             # <<<<<<<<<<<<<<
 * 					continue
 * 				if e.seq > oi.seq or (e.seq == oi.seq and e.start > oi.end):
 */
        __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_v_ei, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 303, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF_SET(__pyx_v_firstRelevantIndex, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "gnocis/regions.pyx":304
 * 				if e.seq < oi.seq or (e.seq == oi.seq and e.end < oi.start):
 * 					firstRelevantIndex = ei+1
 * 					continue             # <<<<<<<<<<<<<<
 * 				if e.seq > oi.seq or (e.seq == oi.seq and e.start > oi.end):
 * 					break
 */
        goto __pyx_L5_continue;

        /* "gnocis/regions.pyx":302
 * 			firstRelevantIndex = 0
 * 			for ei,e in enumerate(exSet):
 * 				if e.seq < oi.seq or (e.seq == oi.seq and e.end < oi.start):             # <<<<<<<<<<<<<<
 * 					firstRelevantIndex = ei+1
 * 					continue
 */
      }

      /* "gnocis/regions.pyx":305
 * 					firstRelevantIndex = ei+1
 * 					continue
 * 				if e.seq > oi.seq or (e.seq == oi.seq and e.start > oi.end):             # <<<<<<<<<<<<<<
 * 					break
 * 				if oi.start >= e.start and oi.end <=e.end:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_seq); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyObject_RichCompare(__pyx_t_6, __pyx_v_oi->seq, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (!__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_seq); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_10, __pyx_v_oi->seq, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_10, __pyx_t_6, Py_GT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_12 = __pyx_t_13;
      __pyx_L12_bool_binop_done:;
      if (__pyx_t_12) {

        /* "gnocis/regions.pyx":306
 * 					continue
 * 				if e.seq > oi.seq or (e.seq == oi.seq and e.start > oi.end):
 * 					break             # <<<<<<<<<<<<<<
 * 				if oi.start >= e.start and oi.end <=e.end:
 * 					oi = None
 */
        goto __pyx_L6_break;

        /* "gnocis/regions.pyx":305
 * 					firstRelevantIndex = ei+1
 * 					continue
 * 				if e.seq > oi.seq or (e.seq == oi.seq and e.start > oi.end):             # <<<<<<<<<<<<<<
 * 					break
 * 				if oi.start >= e.start and oi.end <=e.end:
 */
      }

      /* "gnocis/regions.pyx":307
 * 				if e.seq > oi.seq or (e.seq == oi.seq and e.start > oi.end):
 * 					break
 * 				if oi.start >= e.start and oi.end <=e.end:             # <<<<<<<<<<<<<<
 * 					oi = None
 * 					break
 */
      __pyx_t_7 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->start); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyObject_RichCompare(__pyx_t_7, __pyx_t_6, Py_GE); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_10 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->end); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_10, __pyx_t_6, Py_LE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_12 = __pyx_t_13;
      __pyx_L16_bool_binop_done:;
      if (__pyx_t_12) {

        /* "gnocis/regions.pyx":308
 * 					break
 * 				if oi.start >= e.start and oi.end <=e.end:
 * 					oi = None             # <<<<<<<<<<<<<<
 * 					break
 * 				if e.start > oi.start and e.end < oi.end:
 */
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_oi, ((struct __pyx_obj_6gnocis_7regions_region *)Py_None));

        /* "gnocis/regions.pyx":309
 * 				if oi.start >= e.start and oi.end <=e.end:
 * 					oi = None
 * 					break             # <<<<<<<<<<<<<<
 * 				if e.start > oi.start and e.end < oi.end:
 * 					output.append( region(oi.seq, oi.start, e.start-1, oi.strand, oi.score, oi.source, oi.feature, group = oi.group) )
 */
        goto __pyx_L6_break;

        /* "gnocis/regions.pyx":307
 * 				if e.seq > oi.seq or (e.seq == oi.seq and e.start > oi.end):
 * 					break
 * 				if oi.start >= e.start and oi.end <=e.end:             # <<<<<<<<<<<<<<
 * 					oi = None
 * 					break
 */
      }

      /* "gnocis/regions.pyx":310
 * 					oi = None
 * 					break
 * 				if e.start > oi.start and e.end < oi.end:             # <<<<<<<<<<<<<<
 * 					output.append( region(oi.seq, oi.start, e.start-1, oi.strand, oi.score, oi.source, oi.feature, group = oi.group) )
 * 					oi = region(oi.seq, e.end+1, oi.end, oi.strand, oi.score, oi.source, oi.feature, group = oi.group)
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_start); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyObject_RichCompare(__pyx_t_7, __pyx_t_6, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L19_bool_binop_done;
      }
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_end); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_6 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_10, __pyx_t_6, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_12 = __pyx_t_13;
      __pyx_L19_bool_binop_done:;
      if (__pyx_t_12) {

        /* "gnocis/regions.pyx":311
 * 					break
 * 				if e.start > oi.start and e.end < oi.end:
 * 					output.append( region(oi.seq, oi.start, e.start-1, oi.strand, oi.score, oi.source, oi.feature, group = oi.group) )             # <<<<<<<<<<<<<<
 * 					oi = region(oi.seq, e.end+1, oi.end, oi.strand, oi.score, oi.source, oi.feature, group = oi.group)
 * 				elif e.start < oi.start:
 */
        __pyx_t_7 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->start); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_start); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_10 = __Pyx_PyInt_SubtractObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_oi->strand); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_oi->score); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = PyTuple_New(7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_v_oi->seq);
        __Pyx_GIVEREF(__pyx_v_oi->seq);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_oi->seq);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_5);
        __Pyx_INCREF(__pyx_v_oi->source);
        __Pyx_GIVEREF(__pyx_v_oi->source);
        PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_oi->source);
        __Pyx_INCREF(__pyx_v_oi->feature);
        __Pyx_GIVEREF(__pyx_v_oi->feature);
        PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_oi->feature);
        __pyx_t_7 = 0;
        __pyx_t_10 = 0;
        __pyx_t_6 = 0;
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_group, __pyx_v_oi->group) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
        __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_output, __pyx_t_6); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 311, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "gnocis/regions.pyx":312
 * 				if e.start > oi.start and e.end < oi.end:
 * 					output.append( region(oi.seq, oi.start, e.start-1, oi.strand, oi.score, oi.source, oi.feature, group = oi.group) )
 * 					oi = region(oi.seq, e.end+1, oi.end, oi.strand, oi.score, oi.source, oi.feature, group = oi.group)             # <<<<<<<<<<<<<<
 * 				elif e.start < oi.start:
 * 					if e.end >= oi.start:
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->end); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_oi->strand); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = PyFloat_FromDouble(__pyx_v_oi->score); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_7 = PyTuple_New(7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_oi->seq);
        __Pyx_GIVEREF(__pyx_v_oi->seq);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_oi->seq);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_10);
        __Pyx_INCREF(__pyx_v_oi->source);
        __Pyx_GIVEREF(__pyx_v_oi->source);
        PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_v_oi->source);
        __Pyx_INCREF(__pyx_v_oi->feature);
        __Pyx_GIVEREF(__pyx_v_oi->feature);
        PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_v_oi->feature);
        __pyx_t_5 = 0;
        __pyx_t_6 = 0;
        __pyx_t_4 = 0;
        __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_group, __pyx_v_oi->group) < 0) __PYX_ERR(0, 312, __pyx_L1_error)
        __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_7, __pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 312, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF_SET(__pyx_v_oi, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "gnocis/regions.pyx":310
 * 					oi = None
 * 					break
 * 				if e.start > oi.start and e.end < oi.end:             # <<<<<<<<<<<<<<
 * 					output.append( region(oi.seq, oi.start, e.start-1, oi.strand, oi.score, oi.source, oi.feature, group = oi.group) )
 * 					oi = region(oi.seq, e.end+1, oi.end, oi.strand, oi.score, oi.source, oi.feature, group = oi.group)
 */
        goto __pyx_L18;
      }

      /* "gnocis/regions.pyx":313
 * 					output.append( region(oi.seq, oi.start, e.start-1, oi.strand, oi.score, oi.source, oi.feature, group = oi.group) )
 * 					oi = region(oi.seq, e.end+1, oi.end, oi.strand, oi.score, oi.source, oi.feature, group = oi.group)
 * 				elif e.start < oi.start:             # <<<<<<<<<<<<<<
 * 					if e.end >= oi.start:
 * 						oi.start = e.end+1
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_10 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_4, __pyx_t_10, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 313, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 313, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_12) {

        /* "gnocis/regions.pyx":314
 * 					oi = region(oi.seq, e.end+1, oi.end, oi.strand, oi.score, oi.source, oi.feature, group = oi.group)
 * 				elif e.start < oi.start:
 * 					if e.end >= oi.start:             # <<<<<<<<<<<<<<
 * 						oi.start = e.end+1
 * 				else:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_end); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 314, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 314, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_4 = PyObject_RichCompare(__pyx_t_7, __pyx_t_10, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 314, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 314, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_12) {

          /* "gnocis/regions.pyx":315
 * 				elif e.start < oi.start:
 * 					if e.end >= oi.start:
 * 						oi.start = e.end+1             # <<<<<<<<<<<<<<
 * 				else:
 * 					if e.start <= oi.end:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 315, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 315, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_15 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_10); if (unlikely((__pyx_t_15 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 315, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_v_oi->start = __pyx_t_15;

          /* "gnocis/regions.pyx":314
 * 					oi = region(oi.seq, e.end+1, oi.end, oi.strand, oi.score, oi.source, oi.feature, group = oi.group)
 * 				elif e.start < oi.start:
 * 					if e.end >= oi.start:             # <<<<<<<<<<<<<<
 * 						oi.start = e.end+1
 * 				else:
 */
        }

        /* "gnocis/regions.pyx":313
 * 					output.append( region(oi.seq, oi.start, e.start-1, oi.strand, oi.score, oi.source, oi.feature, group = oi.group) )
 * 					oi = region(oi.seq, e.end+1, oi.end, oi.strand, oi.score, oi.source, oi.feature, group = oi.group)
 * 				elif e.start < oi.start:             # <<<<<<<<<<<<<<
 * 					if e.end >= oi.start:
 * 						oi.start = e.end+1
 */
        goto __pyx_L18;
      }

      /* "gnocis/regions.pyx":317
 * 						oi.start = e.end+1
 * 				else:
 * 					if e.start <= oi.end:             # <<<<<<<<<<<<<<
 * 						oi.end = e.start-1
 * 			exSet = exSet[firstRelevantIndex:]
 */
      /*else*/ {
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_4 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_oi->end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = PyObject_RichCompare(__pyx_t_10, __pyx_t_4, Py_LE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_12) {

          /* "gnocis/regions.pyx":318
 * 				else:
 * 					if e.start <= oi.end:
 * 						oi.end = e.start-1             # <<<<<<<<<<<<<<
 * 			exSet = exSet[firstRelevantIndex:]
 * 			if oi:
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_start); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 318, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_7, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 318, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_15 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_4); if (unlikely((__pyx_t_15 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_v_oi->end = __pyx_t_15;

          /* "gnocis/regions.pyx":317
 * 						oi.start = e.end+1
 * 				else:
 * 					if e.start <= oi.end:             # <<<<<<<<<<<<<<
 * 						oi.end = e.start-1
 * 			exSet = exSet[firstRelevantIndex:]
 */
        }
      }
      __pyx_L18:;

      /* "gnocis/regions.pyx":301
 * 			oi = region(i.seq, i.start, i.end, i.strand, i.score, i.source, i.feature, group = i.group)
 * 			firstRelevantIndex = 0
 * 			for ei,e in enumerate(exSet):             # <<<<<<<<<<<<<<
 * 				if e.seq < oi.seq or (e.seq == oi.seq and e.end < oi.start):
 * 					firstRelevantIndex = ei+1
 */
      __pyx_L5_continue:;
    }
    __pyx_L6_break:;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "gnocis/regions.pyx":319
 * 					if e.start <= oi.end:
 * 						oi.end = e.start-1
 * 			exSet = exSet[firstRelevantIndex:]             # <<<<<<<<<<<<<<
 * 			if oi:
 * 				output.append( oi )
 */
    if (unlikely(__pyx_v_exSet == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 319, __pyx_L1_error)
    }
    __Pyx_INCREF(__pyx_v_firstRelevantIndex);
    __pyx_t_8 = __pyx_v_firstRelevantIndex;
    __pyx_t_12 = (__pyx_t_8 == Py_None);
    if (__pyx_t_12) {
      __pyx_t_11 = 0;
    } else {
      __pyx_t_16 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_16 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 319, __pyx_L1_error)
      __pyx_t_11 = __pyx_t_16;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyList_GetSlice(__pyx_v_exSet, __pyx_t_11, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF_SET(__pyx_v_exSet, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "gnocis/regions.pyx":320
 * 						oi.end = e.start-1
 * 			exSet = exSet[firstRelevantIndex:]
 * 			if oi:             # <<<<<<<<<<<<<<
 * 				output.append( oi )
 * 		rs = regions('%s ^ %s'%(self.name, other.name), output)
 */
    __pyx_t_12 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_oi)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 320, __pyx_L1_error)
    if (__pyx_t_12) {

      /* "gnocis/regions.pyx":321
 * 			exSet = exSet[firstRelevantIndex:]
 * 			if oi:
 * 				output.append( oi )             # <<<<<<<<<<<<<<
 * 		rs = regions('%s ^ %s'%(self.name, other.name), output)
 * 		return rs
 */
      __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_output, ((PyObject *)__pyx_v_oi)); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 321, __pyx_L1_error)

      /* "gnocis/regions.pyx":320
 * 						oi.end = e.start-1
 * 			exSet = exSet[firstRelevantIndex:]
 * 			if oi:             # <<<<<<<<<<<<<<
 * 				output.append( oi )
 * 		rs = regions('%s ^ %s'%(self.name, other.name), output)
 */
    }

    /* "gnocis/regions.pyx":298
 * 		output = []
 * 		exSet = other.regions
 * 		for i in self.regions:             # <<<<<<<<<<<<<<
 * 			oi = region(i.seq, i.start, i.end, i.strand, i.score, i.source, i.feature, group = i.group)
 * 			firstRelevantIndex = 0
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":322
 * 			if oi:
 * 				output.append( oi )
 * 		rs = regions('%s ^ %s'%(self.name, other.name), output)             # <<<<<<<<<<<<<<
 * 		return rs
 * 
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_17 = 127;
  __pyx_t_8 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_17;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __Pyx_INCREF(__pyx_kp_u__16);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__16);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__16);
  __pyx_t_8 = __Pyx_PyUnicode_Unicode(__pyx_v_other->name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_17;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_8);
  __Pyx_INCREF(__pyx_v_output);
  __Pyx_GIVEREF(__pyx_v_output);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_output);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_1, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rs = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "gnocis/regions.pyx":323
 * 				output.append( oi )
 * 		rs = regions('%s ^ %s'%(self.name, other.name), output)
 * 		return rs             # <<<<<<<<<<<<<<
 * 
 * 	# Returns the subset of regions from this set that overlap with another set. Note that this assumes that the regions are sorted in both lists, with the class sort() method.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_rs));
  __pyx_r = ((PyObject *)__pyx_v_rs);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":287
 * 
 * 	# Returns the excluded region set for this and another set.
 * 	def exclusion(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the regions of this set with regions of another set excluded.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("gnocis.regions.regions.exclusion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_output);
  __Pyx_XDECREF(__pyx_v_exSet);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF((PyObject *)__pyx_v_oi);
  __Pyx_XDECREF(__pyx_v_firstRelevantIndex);
  __Pyx_XDECREF(__pyx_v_ei);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF((PyObject *)__pyx_v_rs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":326
 * 
 * 	# Returns the subset of regions from this set that overlap with another set. Note that this assumes that the regions are sorted in both lists, with the class sort() method.
 * 	def overlap(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the regions in this set that overlap with another.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_43overlap(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_42overlap[] = " Gets the regions in this set that overlap with another.\n\t\t\n\t\t:param other: Other region set.\n\t\t:type other: regions\n\t\t\n\t\t:return: Set of overlapping regions.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_43overlap(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gnocis_7regions_regions, 1, "other", 0))) __PYX_ERR(0, 326, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_42overlap(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_42overlap(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other) {
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rA = 0;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rB = 0;
  PyObject *__pyx_v_rAs = 0;
  PyObject *__pyx_v_rBs = 0;
  PyObject *__pyx_v_rlist = 0;
  int __pyx_v_iB;
  int __pyx_v_firstRelevantIndexB;
  int __pyx_v_lenB;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("overlap", 0);

  /* "gnocis/regions.pyx":340
 * 		cdef int iB, firstRelevantIndexB
 * 		cdef int lenB
 * 		rlist = []             # <<<<<<<<<<<<<<
 * 		rAs = self.regions
 * 		rBs = other.regions
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rlist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":341
 * 		cdef int lenB
 * 		rlist = []
 * 		rAs = self.regions             # <<<<<<<<<<<<<<
 * 		rBs = other.regions
 * 		lenB = len(rBs)
 */
  __pyx_t_1 = __pyx_v_self->regions;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_rAs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":342
 * 		rlist = []
 * 		rAs = self.regions
 * 		rBs = other.regions             # <<<<<<<<<<<<<<
 * 		lenB = len(rBs)
 * 		firstRelevantIndexB = 0
 */
  __pyx_t_1 = __pyx_v_other->regions;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_rBs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":343
 * 		rAs = self.regions
 * 		rBs = other.regions
 * 		lenB = len(rBs)             # <<<<<<<<<<<<<<
 * 		firstRelevantIndexB = 0
 * 		for rA in rAs:
 */
  if (unlikely(__pyx_v_rBs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 343, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_rBs); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 343, __pyx_L1_error)
  __pyx_v_lenB = __pyx_t_2;

  /* "gnocis/regions.pyx":344
 * 		rBs = other.regions
 * 		lenB = len(rBs)
 * 		firstRelevantIndexB = 0             # <<<<<<<<<<<<<<
 * 		for rA in rAs:
 * 			iB = firstRelevantIndexB
 */
  __pyx_v_firstRelevantIndexB = 0;

  /* "gnocis/regions.pyx":345
 * 		lenB = len(rBs)
 * 		firstRelevantIndexB = 0
 * 		for rA in rAs:             # <<<<<<<<<<<<<<
 * 			iB = firstRelevantIndexB
 * 			while True:
 */
  if (unlikely(__pyx_v_rAs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 345, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_rAs; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 345, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_rA, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":346
 * 		firstRelevantIndexB = 0
 * 		for rA in rAs:
 * 			iB = firstRelevantIndexB             # <<<<<<<<<<<<<<
 * 			while True:
 * 				if iB >= lenB:
 */
    __pyx_v_iB = __pyx_v_firstRelevantIndexB;

    /* "gnocis/regions.pyx":347
 * 		for rA in rAs:
 * 			iB = firstRelevantIndexB
 * 			while True:             # <<<<<<<<<<<<<<
 * 				if iB >= lenB:
 * 					break
 */
    while (1) {

      /* "gnocis/regions.pyx":348
 * 			iB = firstRelevantIndexB
 * 			while True:
 * 				if iB >= lenB:             # <<<<<<<<<<<<<<
 * 					break
 * 				rB = rBs[iB]
 */
      __pyx_t_4 = ((__pyx_v_iB >= __pyx_v_lenB) != 0);
      if (__pyx_t_4) {

        /* "gnocis/regions.pyx":349
 * 			while True:
 * 				if iB >= lenB:
 * 					break             # <<<<<<<<<<<<<<
 * 				rB = rBs[iB]
 * 				if rB.seq > rA.seq:
 */
        goto __pyx_L6_break;

        /* "gnocis/regions.pyx":348
 * 			iB = firstRelevantIndexB
 * 			while True:
 * 				if iB >= lenB:             # <<<<<<<<<<<<<<
 * 					break
 * 				rB = rBs[iB]
 */
      }

      /* "gnocis/regions.pyx":350
 * 				if iB >= lenB:
 * 					break
 * 				rB = rBs[iB]             # <<<<<<<<<<<<<<
 * 				if rB.seq > rA.seq:
 * 					break
 */
      if (unlikely(__pyx_v_rBs == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 350, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_rBs, __pyx_v_iB, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 350, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_rB, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "gnocis/regions.pyx":351
 * 					break
 * 				rB = rBs[iB]
 * 				if rB.seq > rA.seq:             # <<<<<<<<<<<<<<
 * 					break
 * 				if rB.seq < rA.seq:
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_rB->seq, __pyx_v_rA->seq, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 351, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 351, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_4) {

        /* "gnocis/regions.pyx":352
 * 				rB = rBs[iB]
 * 				if rB.seq > rA.seq:
 * 					break             # <<<<<<<<<<<<<<
 * 				if rB.seq < rA.seq:
 * 					iB += 1
 */
        goto __pyx_L6_break;

        /* "gnocis/regions.pyx":351
 * 					break
 * 				rB = rBs[iB]
 * 				if rB.seq > rA.seq:             # <<<<<<<<<<<<<<
 * 					break
 * 				if rB.seq < rA.seq:
 */
      }

      /* "gnocis/regions.pyx":353
 * 				if rB.seq > rA.seq:
 * 					break
 * 				if rB.seq < rA.seq:             # <<<<<<<<<<<<<<
 * 					iB += 1
 * 					firstRelevantIndexB = iB
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_rB->seq, __pyx_v_rA->seq, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_4) {

        /* "gnocis/regions.pyx":354
 * 					break
 * 				if rB.seq < rA.seq:
 * 					iB += 1             # <<<<<<<<<<<<<<
 * 					firstRelevantIndexB = iB
 * 					continue
 */
        __pyx_v_iB = (__pyx_v_iB + 1);

        /* "gnocis/regions.pyx":355
 * 				if rB.seq < rA.seq:
 * 					iB += 1
 * 					firstRelevantIndexB = iB             # <<<<<<<<<<<<<<
 * 					continue
 * 				if rB.end < rA.start:
 */
        __pyx_v_firstRelevantIndexB = __pyx_v_iB;

        /* "gnocis/regions.pyx":356
 * 					iB += 1
 * 					firstRelevantIndexB = iB
 * 					continue             # <<<<<<<<<<<<<<
 * 				if rB.end < rA.start:
 * 					iB += 1
 */
        goto __pyx_L5_continue;

        /* "gnocis/regions.pyx":353
 * 				if rB.seq > rA.seq:
 * 					break
 * 				if rB.seq < rA.seq:             # <<<<<<<<<<<<<<
 * 					iB += 1
 * 					firstRelevantIndexB = iB
 */
      }

      /* "gnocis/regions.pyx":357
 * 					firstRelevantIndexB = iB
 * 					continue
 * 				if rB.end < rA.start:             # <<<<<<<<<<<<<<
 * 					iB += 1
 * 					firstRelevantIndexB = iB
 */
      __pyx_t_4 = ((__pyx_v_rB->end < __pyx_v_rA->start) != 0);
      if (__pyx_t_4) {

        /* "gnocis/regions.pyx":358
 * 					continue
 * 				if rB.end < rA.start:
 * 					iB += 1             # <<<<<<<<<<<<<<
 * 					firstRelevantIndexB = iB
 * 					continue
 */
        __pyx_v_iB = (__pyx_v_iB + 1);

        /* "gnocis/regions.pyx":359
 * 				if rB.end < rA.start:
 * 					iB += 1
 * 					firstRelevantIndexB = iB             # <<<<<<<<<<<<<<
 * 					continue
 * 				firstRelevantIndexB = iB
 */
        __pyx_v_firstRelevantIndexB = __pyx_v_iB;

        /* "gnocis/regions.pyx":360
 * 					iB += 1
 * 					firstRelevantIndexB = iB
 * 					continue             # <<<<<<<<<<<<<<
 * 				firstRelevantIndexB = iB
 * 				if rA.end < rB.start:
 */
        goto __pyx_L5_continue;

        /* "gnocis/regions.pyx":357
 * 					firstRelevantIndexB = iB
 * 					continue
 * 				if rB.end < rA.start:             # <<<<<<<<<<<<<<
 * 					iB += 1
 * 					firstRelevantIndexB = iB
 */
      }

      /* "gnocis/regions.pyx":361
 * 					firstRelevantIndexB = iB
 * 					continue
 * 				firstRelevantIndexB = iB             # <<<<<<<<<<<<<<
 * 				if rA.end < rB.start:
 * 					break
 */
      __pyx_v_firstRelevantIndexB = __pyx_v_iB;

      /* "gnocis/regions.pyx":362
 * 					continue
 * 				firstRelevantIndexB = iB
 * 				if rA.end < rB.start:             # <<<<<<<<<<<<<<
 * 					break
 * 				rlist.append(rA)
 */
      __pyx_t_4 = ((__pyx_v_rA->end < __pyx_v_rB->start) != 0);
      if (__pyx_t_4) {

        /* "gnocis/regions.pyx":363
 * 				firstRelevantIndexB = iB
 * 				if rA.end < rB.start:
 * 					break             # <<<<<<<<<<<<<<
 * 				rlist.append(rA)
 * 				break
 */
        goto __pyx_L6_break;

        /* "gnocis/regions.pyx":362
 * 					continue
 * 				firstRelevantIndexB = iB
 * 				if rA.end < rB.start:             # <<<<<<<<<<<<<<
 * 					break
 * 				rlist.append(rA)
 */
      }

      /* "gnocis/regions.pyx":364
 * 				if rA.end < rB.start:
 * 					break
 * 				rlist.append(rA)             # <<<<<<<<<<<<<<
 * 				break
 * 		return regions('%s overlapping with %s'%(self.name, other.name), rlist, initialSort = False)
 */
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_rlist, ((PyObject *)__pyx_v_rA)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 364, __pyx_L1_error)

      /* "gnocis/regions.pyx":365
 * 					break
 * 				rlist.append(rA)
 * 				break             # <<<<<<<<<<<<<<
 * 		return regions('%s overlapping with %s'%(self.name, other.name), rlist, initialSort = False)
 * 
 */
      goto __pyx_L6_break;
      __pyx_L5_continue:;
    }
    __pyx_L6_break:;

    /* "gnocis/regions.pyx":345
 * 		lenB = len(rBs)
 * 		firstRelevantIndexB = 0
 * 		for rA in rAs:             # <<<<<<<<<<<<<<
 * 			iB = firstRelevantIndexB
 * 			while True:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":366
 * 				rlist.append(rA)
 * 				break
 * 		return regions('%s overlapping with %s'%(self.name, other.name), rlist, initialSort = False)             # <<<<<<<<<<<<<<
 * 
 * 	# Returns the subset of regions from this set that do not overlap with another set. Note that this assumes that the regions are sorted in both lists, with the class sort() method.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_6 = 127;
  __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_overlapping_with);
  __pyx_t_2 += 18;
  __Pyx_GIVEREF(__pyx_kp_u_overlapping_with);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_overlapping_with);
  __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_other->name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_6;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_rlist);
  __Pyx_GIVEREF(__pyx_v_rlist);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_rlist);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_initialSort, Py_False) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":326
 * 
 * 	# Returns the subset of regions from this set that overlap with another set. Note that this assumes that the regions are sorted in both lists, with the class sort() method.
 * 	def overlap(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the regions in this set that overlap with another.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.regions.regions.overlap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_rA);
  __Pyx_XDECREF((PyObject *)__pyx_v_rB);
  __Pyx_XDECREF(__pyx_v_rAs);
  __Pyx_XDECREF(__pyx_v_rBs);
  __Pyx_XDECREF(__pyx_v_rlist);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":369
 * 
 * 	# Returns the subset of regions from this set that do not overlap with another set. Note that this assumes that the regions are sorted in both lists, with the class sort() method.
 * 	def nonOverlap(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the regions in this set that do not overlap with another.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_45nonOverlap(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_44nonOverlap[] = " Gets the regions in this set that do not overlap with another.\n\t\t\n\t\t:param other: Other region set.\n\t\t:type other: regions\n\t\t\n\t\t:return: Set of non-overlapping regions.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_45nonOverlap(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nonOverlap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gnocis_7regions_regions, 1, "other", 0))) __PYX_ERR(0, 369, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_44nonOverlap(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_44nonOverlap(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_other) {
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rA = 0;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rB = 0;
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_rs = 0;
  PyObject *__pyx_v_rBs = 0;
  bool __pyx_v_over;
  int __pyx_v_iB;
  int __pyx_v_firstRelevantIndexB;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nonOverlap", 0);

  /* "gnocis/regions.pyx":383
 * 		cdef bool over
 * 		cdef int iB, firstRelevantIndexB
 * 		rs = regions('%s not overlapping with %s'%(self.name, other.name), [])             # <<<<<<<<<<<<<<
 * 		rBs = other.regions
 * 		for rA in self.regions:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __pyx_t_4 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_not_overlapping_with);
  __pyx_t_2 += 22;
  __Pyx_GIVEREF(__pyx_kp_u_not_overlapping_with);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_not_overlapping_with);
  __pyx_t_4 = __Pyx_PyUnicode_Unicode(__pyx_v_other->name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_rs = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":384
 * 		cdef int iB, firstRelevantIndexB
 * 		rs = regions('%s not overlapping with %s'%(self.name, other.name), [])
 * 		rBs = other.regions             # <<<<<<<<<<<<<<
 * 		for rA in self.regions:
 * 			firstRelevantIndexB = 0
 */
  __pyx_t_1 = __pyx_v_other->regions;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_rBs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":385
 * 		rs = regions('%s not overlapping with %s'%(self.name, other.name), [])
 * 		rBs = other.regions
 * 		for rA in self.regions:             # <<<<<<<<<<<<<<
 * 			firstRelevantIndexB = 0
 * 			over = False
 */
  if (unlikely(__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 385, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 385, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_rA, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "gnocis/regions.pyx":386
 * 		rBs = other.regions
 * 		for rA in self.regions:
 * 			firstRelevantIndexB = 0             # <<<<<<<<<<<<<<
 * 			over = False
 * 			for iB,rB in enumerate(rBs):
 */
    __pyx_v_firstRelevantIndexB = 0;

    /* "gnocis/regions.pyx":387
 * 		for rA in self.regions:
 * 			firstRelevantIndexB = 0
 * 			over = False             # <<<<<<<<<<<<<<
 * 			for iB,rB in enumerate(rBs):
 * 				if rB.seq > rA.seq:
 */
    __pyx_v_over = 0;

    /* "gnocis/regions.pyx":388
 * 			firstRelevantIndexB = 0
 * 			over = False
 * 			for iB,rB in enumerate(rBs):             # <<<<<<<<<<<<<<
 * 				if rB.seq > rA.seq:
 * 					break
 */
    __pyx_t_6 = 0;
    __pyx_t_5 = __pyx_v_rBs; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
    for (;;) {
      if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 388, __pyx_L1_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_rB, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_4));
      __pyx_t_4 = 0;
      __pyx_v_iB = __pyx_t_6;
      __pyx_t_6 = (__pyx_t_6 + 1);

      /* "gnocis/regions.pyx":389
 * 			over = False
 * 			for iB,rB in enumerate(rBs):
 * 				if rB.seq > rA.seq:             # <<<<<<<<<<<<<<
 * 					break
 * 				if rB.seq < rA.seq:
 */
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_rB->seq, __pyx_v_rA->seq, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_8) {

        /* "gnocis/regions.pyx":390
 * 			for iB,rB in enumerate(rBs):
 * 				if rB.seq > rA.seq:
 * 					break             # <<<<<<<<<<<<<<
 * 				if rB.seq < rA.seq:
 * 					continue
 */
        goto __pyx_L6_break;

        /* "gnocis/regions.pyx":389
 * 			over = False
 * 			for iB,rB in enumerate(rBs):
 * 				if rB.seq > rA.seq:             # <<<<<<<<<<<<<<
 * 					break
 * 				if rB.seq < rA.seq:
 */
      }

      /* "gnocis/regions.pyx":391
 * 				if rB.seq > rA.seq:
 * 					break
 * 				if rB.seq < rA.seq:             # <<<<<<<<<<<<<<
 * 					continue
 * 				if rB.end < rA.start:
 */
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_rB->seq, __pyx_v_rA->seq, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 391, __pyx_L1_error)
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_8) {

        /* "gnocis/regions.pyx":392
 * 					break
 * 				if rB.seq < rA.seq:
 * 					continue             # <<<<<<<<<<<<<<
 * 				if rB.end < rA.start:
 * 					continue
 */
        goto __pyx_L5_continue;

        /* "gnocis/regions.pyx":391
 * 				if rB.seq > rA.seq:
 * 					break
 * 				if rB.seq < rA.seq:             # <<<<<<<<<<<<<<
 * 					continue
 * 				if rB.end < rA.start:
 */
      }

      /* "gnocis/regions.pyx":393
 * 				if rB.seq < rA.seq:
 * 					continue
 * 				if rB.end < rA.start:             # <<<<<<<<<<<<<<
 * 					continue
 * 				firstRelevantIndexB = iB
 */
      __pyx_t_8 = ((__pyx_v_rB->end < __pyx_v_rA->start) != 0);
      if (__pyx_t_8) {

        /* "gnocis/regions.pyx":394
 * 					continue
 * 				if rB.end < rA.start:
 * 					continue             # <<<<<<<<<<<<<<
 * 				firstRelevantIndexB = iB
 * 				if rA.end < rB.start:
 */
        goto __pyx_L5_continue;

        /* "gnocis/regions.pyx":393
 * 				if rB.seq < rA.seq:
 * 					continue
 * 				if rB.end < rA.start:             # <<<<<<<<<<<<<<
 * 					continue
 * 				firstRelevantIndexB = iB
 */
      }

      /* "gnocis/regions.pyx":395
 * 				if rB.end < rA.start:
 * 					continue
 * 				firstRelevantIndexB = iB             # <<<<<<<<<<<<<<
 * 				if rA.end < rB.start:
 * 					break
 */
      __pyx_v_firstRelevantIndexB = __pyx_v_iB;

      /* "gnocis/regions.pyx":396
 * 					continue
 * 				firstRelevantIndexB = iB
 * 				if rA.end < rB.start:             # <<<<<<<<<<<<<<
 * 					break
 * 				over = True
 */
      __pyx_t_8 = ((__pyx_v_rA->end < __pyx_v_rB->start) != 0);
      if (__pyx_t_8) {

        /* "gnocis/regions.pyx":397
 * 				firstRelevantIndexB = iB
 * 				if rA.end < rB.start:
 * 					break             # <<<<<<<<<<<<<<
 * 				over = True
 * 				break
 */
        goto __pyx_L6_break;

        /* "gnocis/regions.pyx":396
 * 					continue
 * 				firstRelevantIndexB = iB
 * 				if rA.end < rB.start:             # <<<<<<<<<<<<<<
 * 					break
 * 				over = True
 */
      }

      /* "gnocis/regions.pyx":398
 * 				if rA.end < rB.start:
 * 					break
 * 				over = True             # <<<<<<<<<<<<<<
 * 				break
 * 			if not over:
 */
      __pyx_v_over = 1;

      /* "gnocis/regions.pyx":399
 * 					break
 * 				over = True
 * 				break             # <<<<<<<<<<<<<<
 * 			if not over:
 * 				rs.regions.append(rA)
 */
      goto __pyx_L6_break;

      /* "gnocis/regions.pyx":388
 * 			firstRelevantIndexB = 0
 * 			over = False
 * 			for iB,rB in enumerate(rBs):             # <<<<<<<<<<<<<<
 * 				if rB.seq > rA.seq:
 * 					break
 */
      __pyx_L5_continue:;
    }
    __pyx_L6_break:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "gnocis/regions.pyx":400
 * 				over = True
 * 				break
 * 			if not over:             # <<<<<<<<<<<<<<
 * 				rs.regions.append(rA)
 * 			if firstRelevantIndexB > 0:
 */
    __pyx_t_8 = ((!(__pyx_v_over != 0)) != 0);
    if (__pyx_t_8) {

      /* "gnocis/regions.pyx":401
 * 				break
 * 			if not over:
 * 				rs.regions.append(rA)             # <<<<<<<<<<<<<<
 * 			if firstRelevantIndexB > 0:
 * 				rBs = rBs[firstRelevantIndexB:]
 */
      if (unlikely(__pyx_v_rs->regions == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 401, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_rs->regions, ((PyObject *)__pyx_v_rA)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 401, __pyx_L1_error)

      /* "gnocis/regions.pyx":400
 * 				over = True
 * 				break
 * 			if not over:             # <<<<<<<<<<<<<<
 * 				rs.regions.append(rA)
 * 			if firstRelevantIndexB > 0:
 */
    }

    /* "gnocis/regions.pyx":402
 * 			if not over:
 * 				rs.regions.append(rA)
 * 			if firstRelevantIndexB > 0:             # <<<<<<<<<<<<<<
 * 				rBs = rBs[firstRelevantIndexB:]
 * 		return rs
 */
    __pyx_t_8 = ((__pyx_v_firstRelevantIndexB > 0) != 0);
    if (__pyx_t_8) {

      /* "gnocis/regions.pyx":403
 * 				rs.regions.append(rA)
 * 			if firstRelevantIndexB > 0:
 * 				rBs = rBs[firstRelevantIndexB:]             # <<<<<<<<<<<<<<
 * 		return rs
 * 
 */
      if (unlikely(__pyx_v_rBs == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 403, __pyx_L1_error)
      }
      __pyx_t_5 = __Pyx_PyList_GetSlice(__pyx_v_rBs, __pyx_v_firstRelevantIndexB, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_rBs, ((PyObject*)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "gnocis/regions.pyx":402
 * 			if not over:
 * 				rs.regions.append(rA)
 * 			if firstRelevantIndexB > 0:             # <<<<<<<<<<<<<<
 * 				rBs = rBs[firstRelevantIndexB:]
 * 		return rs
 */
    }

    /* "gnocis/regions.pyx":385
 * 		rs = regions('%s not overlapping with %s'%(self.name, other.name), [])
 * 		rBs = other.regions
 * 		for rA in self.regions:             # <<<<<<<<<<<<<<
 * 			firstRelevantIndexB = 0
 * 			over = False
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":404
 * 			if firstRelevantIndexB > 0:
 * 				rBs = rBs[firstRelevantIndexB:]
 * 		return rs             # <<<<<<<<<<<<<<
 * 
 * 	# Calculates the overlap sensitivity to another set
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_rs));
  __pyx_r = ((PyObject *)__pyx_v_rs);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":369
 * 
 * 	# Returns the subset of regions from this set that do not overlap with another set. Note that this assumes that the regions are sorted in both lists, with the class sort() method.
 * 	def nonOverlap(self, regions other):             # <<<<<<<<<<<<<<
 * 		""" Gets the regions in this set that do not overlap with another.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.regions.regions.nonOverlap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_rA);
  __Pyx_XDECREF((PyObject *)__pyx_v_rB);
  __Pyx_XDECREF((PyObject *)__pyx_v_rs);
  __Pyx_XDECREF(__pyx_v_rBs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":407
 * 
 * 	# Calculates the overlap sensitivity to another set
 * 	def overlapSensitivity(self, other):             # <<<<<<<<<<<<<<
 * 		""" Calculates the overlap sensitivity to another set.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_47overlapSensitivity(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_46overlapSensitivity[] = " Calculates the overlap sensitivity to another set.\n\t\t\n\t\t:param other: Other region set.\n\t\t:type other: regions\n\t\t\n\t\t:return: Overlap sensitivity.\n\t\t:rtype: float\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_47overlapSensitivity(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlapSensitivity (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_46overlapSensitivity(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_46overlapSensitivity(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("overlapSensitivity", 0);

  /* "gnocis/regions.pyx":416
 * 		:rtype: float
 * 		"""
 * 		return len(other.overlap(self)) / len(other)             # <<<<<<<<<<<<<<
 * 
 * 	# Calculates the overlap precision to another set
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_overlap); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_self)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_self));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 416, __pyx_L1_error)
  if (unlikely(__pyx_t_5 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 416, __pyx_L1_error)
  }
  __pyx_t_1 = PyFloat_FromDouble((((double)__pyx_t_4) / ((double)__pyx_t_5))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":407
 * 
 * 	# Calculates the overlap sensitivity to another set
 * 	def overlapSensitivity(self, other):             # <<<<<<<<<<<<<<
 * 		""" Calculates the overlap sensitivity to another set.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.regions.overlapSensitivity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":419
 * 
 * 	# Calculates the overlap precision to another set
 * 	def overlapPrecision(self, other):             # <<<<<<<<<<<<<<
 * 		""" Calculates the overlap precision to another set.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_49overlapPrecision(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_48overlapPrecision[] = " Calculates the overlap precision to another set.\n\t\t\n\t\t:param other: Other region set.\n\t\t:type other: regions\n\t\t\n\t\t:return: Overlap precision.\n\t\t:rtype: float\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_49overlapPrecision(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlapPrecision (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_48overlapPrecision(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_48overlapPrecision(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("overlapPrecision", 0);

  /* "gnocis/regions.pyx":428
 * 		:rtype: float
 * 		"""
 * 		return len(self.overlap(other)) / len(self)             # <<<<<<<<<<<<<<
 * 
 * 	# Calculates the overlap precision to another set
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_overlap); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 428, __pyx_L1_error)
  if (unlikely(__pyx_t_5 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 428, __pyx_L1_error)
  }
  __pyx_t_1 = PyFloat_FromDouble((((double)__pyx_t_4) / ((double)__pyx_t_5))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":419
 * 
 * 	# Calculates the overlap precision to another set
 * 	def overlapPrecision(self, other):             # <<<<<<<<<<<<<<
 * 		""" Calculates the overlap precision to another set.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.regions.overlapPrecision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":431
 * 
 * 	# Calculates the overlap precision to another set
 * 	def nucleotidePrecision(self, other):             # <<<<<<<<<<<<<<
 * 		""" Calculates the nucleotide precision to another set.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_51nucleotidePrecision(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_50nucleotidePrecision[] = " Calculates the nucleotide precision to another set.\n\t\t\n\t\t:param other: Other region set.\n\t\t:type other: regions\n\t\t\n\t\t:return: Nucleotide precision.\n\t\t:rtype: float\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_51nucleotidePrecision(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nucleotidePrecision (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_50nucleotidePrecision(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":440
 * 		:rtype: float
 * 		"""
 * 		return sum(len(r) for r in self & other) / sum(len(r) for r in self)             # <<<<<<<<<<<<<<
 * 
 * 	# Calculates the overlap precision to another set
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_19nucleotidePrecision_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_3_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 440, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_nucleotidePrecision_locals_genex, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.nucleotidePrecision.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 440, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 440, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_other)) { __Pyx_RaiseClosureNameError("other"); __PYX_ERR(0, 440, __pyx_L1_error) }
  __pyx_t_1 = PyNumber_And(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self), __pyx_cur_scope->__pyx_outer_scope->__pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 440, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_5 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 440, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 440, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_19nucleotidePrecision_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_4_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 440, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_5generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_nucleotidePrecision_locals_genex, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.nucleotidePrecision.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 440, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 440, __pyx_L1_error) }
  if (likely(PyList_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self))) {
    __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 440, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 440, __pyx_L1_error)
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 440, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":431
 * 
 * 	# Calculates the overlap precision to another set
 * 	def nucleotidePrecision(self, other):             # <<<<<<<<<<<<<<
 * 		""" Calculates the nucleotide precision to another set.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_50nucleotidePrecision(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *__pyx_cur_scope;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_2generator1 = 0;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_5generator2 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nucleotidePrecision", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 431, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_other = __pyx_v_other;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_other);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_other);

  /* "gnocis/regions.pyx":440
 * 		:rtype: float
 * 		"""
 * 		return sum(len(r) for r in self & other) / sum(len(r) for r in self)             # <<<<<<<<<<<<<<
 * 
 * 	# Calculates the overlap precision to another set
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_7regions_19nucleotidePrecision_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_7regions_19nucleotidePrecision_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":431
 * 
 * 	# Calculates the overlap precision to another set
 * 	def nucleotidePrecision(self, other):             # <<<<<<<<<<<<<<
 * 		""" Calculates the nucleotide precision to another set.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.regions.nucleotidePrecision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_2generator1);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_19nucleotidePrecision_5generator2);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":443
 * 
 * 	# Calculates the overlap precision to another set
 * 	def dummy(self, genome, useSeq = None):             # <<<<<<<<<<<<<<
 * 		""" Generates a dummy region set, with the same region lengths but random positions.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_53dummy(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_52dummy[] = " Generates a dummy region set, with the same region lengths but random positions.\n\t\t\n\t\t:param genome: Genome.\n\t\t:param useSeq: Chromosomes to focus on.\n\t\t:type genome: genome\n\t\t:type useSeq: list\n\t\t\n\t\t:return: Dummy region set.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_53dummy(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_genome = 0;
  PyObject *__pyx_v_useSeq = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dummy (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_genome,&__pyx_n_s_useSeq,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_genome)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_useSeq);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dummy") < 0)) __PYX_ERR(0, 443, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_genome = values[0];
    __pyx_v_useSeq = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dummy", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 443, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.dummy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_52dummy(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), __pyx_v_genome, __pyx_v_useSeq);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_5dummy_2generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":458
 * 		if useSeq is not None:
 * 			seqLens = { k: seqLens[k] for k in seqLens if k in useSeq }
 * 		genomeLen = sum(seqLens[k] for k in seqLens)             # <<<<<<<<<<<<<<
 * 		rs = self.regions
 * 		if useSeq is not None:
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_5dummy_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_6_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_6_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 458, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_5dummy_2generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_dummy_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.dummy.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_5dummy_2generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 458, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens)) { __Pyx_RaiseClosureNameError("seqLens"); __PYX_ERR(0, 458, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 458, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 458, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 458, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 458, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens)) { __Pyx_RaiseClosureNameError("seqLens"); __PYX_ERR(0, 458, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens, __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 458, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":443
 * 
 * 	# Calculates the overlap precision to another set
 * 	def dummy(self, genome, useSeq = None):             # <<<<<<<<<<<<<<
 * 		""" Generates a dummy region set, with the same region lengths but random positions.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_52dummy(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_genome, PyObject *__pyx_v_useSeq) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *__pyx_cur_scope;
  PyObject *__pyx_v_dr = NULL;
  PyObject *__pyx_v_genomeLen = NULL;
  PyObject *__pyx_v_rs = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_pos = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_k = NULL;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_5dummy_2generator3 = 0;
  PyObject *__pyx_9genexpr12__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dummy", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_5_dummy(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_5_dummy, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 443, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/regions.pyx":454
 * 		:rtype: regions
 * 		"""
 * 		dr = []             # <<<<<<<<<<<<<<
 * 		seqLens = genome.sequenceLengths()
 * 		if useSeq is not None:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dr = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":455
 * 		"""
 * 		dr = []
 * 		seqLens = genome.sequenceLengths()             # <<<<<<<<<<<<<<
 * 		if useSeq is not None:
 * 			seqLens = { k: seqLens[k] for k in seqLens if k in useSeq }
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_genome, __pyx_n_s_sequenceLengths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_seqLens = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":456
 * 		dr = []
 * 		seqLens = genome.sequenceLengths()
 * 		if useSeq is not None:             # <<<<<<<<<<<<<<
 * 			seqLens = { k: seqLens[k] for k in seqLens if k in useSeq }
 * 		genomeLen = sum(seqLens[k] for k in seqLens)
 */
  __pyx_t_4 = (__pyx_v_useSeq != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "gnocis/regions.pyx":457
 * 		seqLens = genome.sequenceLengths()
 * 		if useSeq is not None:
 * 			seqLens = { k: seqLens[k] for k in seqLens if k in useSeq }             # <<<<<<<<<<<<<<
 * 		genomeLen = sum(seqLens[k] for k in seqLens)
 * 		rs = self.regions
 */
    { /* enter inner scope */
      __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_seqLens)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_seqLens)) {
        __pyx_t_2 = __pyx_cur_scope->__pyx_v_seqLens; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
      } else {
        __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_seqLens); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 457, __pyx_L6_error)
      }
      for (;;) {
        if (likely(!__pyx_t_7)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 457, __pyx_L6_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 457, __pyx_L6_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_7(__pyx_t_2);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 457, __pyx_L6_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_k, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_9genexpr10__pyx_v_k, __pyx_v_useSeq, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 457, __pyx_L6_error)
        __pyx_t_4 = (__pyx_t_5 != 0);
        if (__pyx_t_4) {
          __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_seqLens, __pyx_9genexpr10__pyx_v_k); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 457, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_9genexpr10__pyx_v_k, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 457, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_k); __pyx_9genexpr10__pyx_v_k = 0;
      goto __pyx_L10_exit_scope;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_k); __pyx_9genexpr10__pyx_v_k = 0;
      goto __pyx_L1_error;
      __pyx_L10_exit_scope:;
    } /* exit inner scope */
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_seqLens);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_seqLens, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/regions.pyx":456
 * 		dr = []
 * 		seqLens = genome.sequenceLengths()
 * 		if useSeq is not None:             # <<<<<<<<<<<<<<
 * 			seqLens = { k: seqLens[k] for k in seqLens if k in useSeq }
 * 		genomeLen = sum(seqLens[k] for k in seqLens)
 */
  }

  /* "gnocis/regions.pyx":458
 * 		if useSeq is not None:
 * 			seqLens = { k: seqLens[k] for k in seqLens if k in useSeq }
 * 		genomeLen = sum(seqLens[k] for k in seqLens)             # <<<<<<<<<<<<<<
 * 		rs = self.regions
 * 		if useSeq is not None:
 */
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_7regions_5dummy_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_genomeLen = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":459
 * 			seqLens = { k: seqLens[k] for k in seqLens if k in useSeq }
 * 		genomeLen = sum(seqLens[k] for k in seqLens)
 * 		rs = self.regions             # <<<<<<<<<<<<<<
 * 		if useSeq is not None:
 * 			rs = [ r for r in rs if r.seq in useSeq ]
 */
  __pyx_t_2 = __pyx_v_self->regions;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_rs = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":460
 * 		genomeLen = sum(seqLens[k] for k in seqLens)
 * 		rs = self.regions
 * 		if useSeq is not None:             # <<<<<<<<<<<<<<
 * 			rs = [ r for r in rs if r.seq in useSeq ]
 * 		for r in rs:
 */
  __pyx_t_4 = (__pyx_v_useSeq != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "gnocis/regions.pyx":461
 * 		rs = self.regions
 * 		if useSeq is not None:
 * 			rs = [ r for r in rs if r.seq in useSeq ]             # <<<<<<<<<<<<<<
 * 		for r in rs:
 * 			pos = int(random.random() * (genomeLen - len(r)))
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L14_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__pyx_v_rs == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 461, __pyx_L14_error)
      }
      __pyx_t_1 = __pyx_v_rs; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
      for (;;) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 461, __pyx_L14_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 461, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_r, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr12__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 461, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_useSeq, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 461, __pyx_L14_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_4 = (__pyx_t_5 != 0);
        if (__pyx_t_4) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_9genexpr12__pyx_v_r))) __PYX_ERR(0, 461, __pyx_L14_error)
        }
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_r); __pyx_9genexpr12__pyx_v_r = 0;
      goto __pyx_L18_exit_scope;
      __pyx_L14_error:;
      __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_r); __pyx_9genexpr12__pyx_v_r = 0;
      goto __pyx_L1_error;
      __pyx_L18_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_rs, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "gnocis/regions.pyx":460
 * 		genomeLen = sum(seqLens[k] for k in seqLens)
 * 		rs = self.regions
 * 		if useSeq is not None:             # <<<<<<<<<<<<<<
 * 			rs = [ r for r in rs if r.seq in useSeq ]
 * 		for r in rs:
 */
  }

  /* "gnocis/regions.pyx":462
 * 		if useSeq is not None:
 * 			rs = [ r for r in rs if r.seq in useSeq ]
 * 		for r in rs:             # <<<<<<<<<<<<<<
 * 			pos = int(random.random() * (genomeLen - len(r)))
 * 			for k in seqLens:
 */
  if (unlikely(__pyx_v_rs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 462, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_v_rs; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
  for (;;) {
    if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 462, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/regions.pyx":463
 * 			rs = [ r for r in rs if r.seq in useSeq ]
 * 		for r in rs:
 * 			pos = int(random.random() * (genomeLen - len(r)))             # <<<<<<<<<<<<<<
 * 			for k in seqLens:
 * 				if pos < seqLens[k]:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_random); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_random); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = PyObject_Length(__pyx_v_r); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 463, __pyx_L1_error)
    __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_genomeLen, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyNumber_Multiply(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_pos, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":464
 * 		for r in rs:
 * 			pos = int(random.random() * (genomeLen - len(r)))
 * 			for k in seqLens:             # <<<<<<<<<<<<<<
 * 				if pos < seqLens[k]:
 * 					dr.append( region(seq = k, start = pos, end = pos + len(r) - 1) )
 */
    if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_seqLens)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_seqLens)) {
      __pyx_t_3 = __pyx_cur_scope->__pyx_v_seqLens; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_seqLens); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 464, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_8); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 464, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 464, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_8); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 464, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 464, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_3);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 464, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "gnocis/regions.pyx":465
 * 			pos = int(random.random() * (genomeLen - len(r)))
 * 			for k in seqLens:
 * 				if pos < seqLens[k]:             # <<<<<<<<<<<<<<
 * 					dr.append( region(seq = k, start = pos, end = pos + len(r) - 1) )
 * 					break
 */
      __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_seqLens, __pyx_v_k); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_pos, __pyx_t_8, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_4) {

        /* "gnocis/regions.pyx":466
 * 			for k in seqLens:
 * 				if pos < seqLens[k]:
 * 					dr.append( region(seq = k, start = pos, end = pos + len(r) - 1) )             # <<<<<<<<<<<<<<
 * 					break
 * 				pos -= seqLens[k]
 */
        __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_seq, __pyx_v_k) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_start, __pyx_v_pos) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
        __pyx_t_10 = PyObject_Length(__pyx_v_r); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 466, __pyx_L1_error)
        __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 466, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = PyNumber_Add(__pyx_v_pos, __pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 466, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_t_11, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 466, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_end, __pyx_t_8) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 466, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_dr, __pyx_t_8); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 466, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "gnocis/regions.pyx":467
 * 				if pos < seqLens[k]:
 * 					dr.append( region(seq = k, start = pos, end = pos + len(r) - 1) )
 * 					break             # <<<<<<<<<<<<<<
 * 				pos -= seqLens[k]
 * 		return regions(self.name + ' (dummy)', dr)
 */
        goto __pyx_L22_break;

        /* "gnocis/regions.pyx":465
 * 			pos = int(random.random() * (genomeLen - len(r)))
 * 			for k in seqLens:
 * 				if pos < seqLens[k]:             # <<<<<<<<<<<<<<
 * 					dr.append( region(seq = k, start = pos, end = pos + len(r) - 1) )
 * 					break
 */
      }

      /* "gnocis/regions.pyx":468
 * 					dr.append( region(seq = k, start = pos, end = pos + len(r) - 1) )
 * 					break
 * 				pos -= seqLens[k]             # <<<<<<<<<<<<<<
 * 		return regions(self.name + ' (dummy)', dr)
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_seqLens, __pyx_v_k); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = PyNumber_InPlaceSubtract(__pyx_v_pos, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF_SET(__pyx_v_pos, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "gnocis/regions.pyx":464
 * 		for r in rs:
 * 			pos = int(random.random() * (genomeLen - len(r)))
 * 			for k in seqLens:             # <<<<<<<<<<<<<<
 * 				if pos < seqLens[k]:
 * 					dr.append( region(seq = k, start = pos, end = pos + len(r) - 1) )
 */
    }
    __pyx_L22_break:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":462
 * 		if useSeq is not None:
 * 			rs = [ r for r in rs if r.seq in useSeq ]
 * 		for r in rs:             # <<<<<<<<<<<<<<
 * 			pos = int(random.random() * (genomeLen - len(r)))
 * 			for k in seqLens:
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":469
 * 					break
 * 				pos -= seqLens[k]
 * 		return regions(self.name + ' (dummy)', dr)             # <<<<<<<<<<<<<<
 * 
 * 	def expected(self, genome, statfun, useSeq = None, repeats = 100):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_self->name, __pyx_kp_u_dummy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_dr);
  __Pyx_GIVEREF(__pyx_v_dr);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_dr);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":443
 * 
 * 	# Calculates the overlap precision to another set
 * 	def dummy(self, genome, useSeq = None):             # <<<<<<<<<<<<<<
 * 		""" Generates a dummy region set, with the same region lengths but random positions.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("gnocis.regions.regions.dummy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dr);
  __Pyx_XDECREF(__pyx_v_genomeLen);
  __Pyx_XDECREF(__pyx_v_rs);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_k);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_5dummy_2generator3);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_r);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":471
 * 		return regions(self.name + ' (dummy)', dr)
 * 
 * 	def expected(self, genome, statfun, useSeq = None, repeats = 100):             # <<<<<<<<<<<<<<
 * 		""" Calculates an expected statistic by random dummy set creation and averaging.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_55expected(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_54expected[] = " Calculates an expected statistic by random dummy set creation and averaging.\n\t\t\n\t\t:param genome: Genome.\n\t\t:param statfun: Function to apply to region sets that returns desired statistic.\n\t\t:param useSeq: Chromosomes to focus on.\n\t\t:param repeats: Number of repeats for calculating statistic.\n\t\t:type genome: genome\n\t\t:type statfun: Function\n\t\t:type useSeq: list\n\t\t:type repeats: int\n\t\t\n\t\t:return: Average statistic.\n\t\t:rtype: float\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_55expected(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_genome = 0;
  PyObject *__pyx_v_statfun = 0;
  PyObject *__pyx_v_useSeq = 0;
  PyObject *__pyx_v_repeats = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("expected (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_genome,&__pyx_n_s_statfun,&__pyx_n_s_useSeq,&__pyx_n_s_repeats,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)__pyx_int_100);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_genome)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_statfun)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("expected", 0, 2, 4, 1); __PYX_ERR(0, 471, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_useSeq);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_repeats);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "expected") < 0)) __PYX_ERR(0, 471, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_genome = values[0];
    __pyx_v_statfun = values[1];
    __pyx_v_useSeq = values[2];
    __pyx_v_repeats = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("expected", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 471, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.expected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_54expected(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), __pyx_v_genome, __pyx_v_statfun, __pyx_v_useSeq, __pyx_v_repeats);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_54expected(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_genome, PyObject *__pyx_v_statfun, PyObject *__pyx_v_useSeq, PyObject *__pyx_v_repeats) {
  PyObject *__pyx_v_stats = NULL;
  CYTHON_UNUSED PyObject *__pyx_9genexpr13__pyx_v__ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("expected", 0);

  /* "gnocis/regions.pyx":486
 * 		:rtype: float
 * 		"""
 * 		stats = [             # <<<<<<<<<<<<<<
 * 			statfun(self.dummy(genome = genome, useSeq = useSeq))
 * 			for _ in range(repeats)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "gnocis/regions.pyx":488
 * 		stats = [
 * 			statfun(self.dummy(genome = genome, useSeq = useSeq))
 * 			for _ in range(repeats)             # <<<<<<<<<<<<<<
 * 		]
 * 		return sum(stats) / len(stats)
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_repeats); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 488, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 488, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 488, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 488, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v__, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "gnocis/regions.pyx":487
 * 		"""
 * 		stats = [
 * 			statfun(self.dummy(genome = genome, useSeq = useSeq))             # <<<<<<<<<<<<<<
 * 			for _ in range(repeats)
 * 		]
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dummy_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 487, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 487, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_genome, __pyx_v_genome) < 0) __PYX_ERR(0, 487, __pyx_L5_error)
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_useSeq, __pyx_v_useSeq) < 0) __PYX_ERR(0, 487, __pyx_L5_error)
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 487, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_v_statfun);
      __pyx_t_7 = __pyx_v_statfun; __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 486, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "gnocis/regions.pyx":488
 * 		stats = [
 * 			statfun(self.dummy(genome = genome, useSeq = useSeq))
 * 			for _ in range(repeats)             # <<<<<<<<<<<<<<
 * 		]
 * 		return sum(stats) / len(stats)
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v__); __pyx_9genexpr13__pyx_v__ = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v__); __pyx_9genexpr13__pyx_v__ = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_v_stats = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":490
 * 			for _ in range(repeats)
 * 		]
 * 		return sum(stats) / len(stats)             # <<<<<<<<<<<<<<
 * 
 * 	# Returns a set of recentered regions, with regions randomly placed within larger regions.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_v_stats); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_stats); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 490, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":471
 * 		return regions(self.name + ' (dummy)', dr)
 * 
 * 	def expected(self, genome, statfun, useSeq = None, repeats = 100):             # <<<<<<<<<<<<<<
 * 		""" Calculates an expected statistic by random dummy set creation and averaging.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("gnocis.regions.regions.expected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_stats);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":493
 * 
 * 	# Returns a set of recentered regions, with regions randomly placed within larger regions.
 * 	def recenter(self, long long size):             # <<<<<<<<<<<<<<
 * 		""" Gets a set of randomly recentered regions. If the target size is smaller than a given region, a region of the desired size is randomly placed within the region. If the desired size is larger, a region of the desired size is placed with equal center to the center of the region.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_57recenter(PyObject *__pyx_v_self, PyObject *__pyx_arg_size); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_56recenter[] = " Gets a set of randomly recentered regions. If the target size is smaller than a given region, a region of the desired size is randomly placed within the region. If the desired size is larger, a region of the desired size is placed with equal center to the center of the region.\n\t\t\n\t\t:param size: Desired region size.\n\t\t:type size: int\n\t\t\n\t\t:return: Set of randomly recentered regions.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_57recenter(PyObject *__pyx_v_self, PyObject *__pyx_arg_size) {
  PY_LONG_LONG __pyx_v_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recenter (wrapper)", 0);
  assert(__pyx_arg_size); {
    __pyx_v_size = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_arg_size); if (unlikely((__pyx_v_size == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 493, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.recenter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_56recenter(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PY_LONG_LONG)__pyx_v_size));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_56recenter(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PY_LONG_LONG __pyx_v_size) {
  PyObject *__pyx_v_rlist = 0;
  PY_LONG_LONG __pyx_v_rstart;
  PY_LONG_LONG __pyx_v_rend;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rgn = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PY_LONG_LONG __pyx_t_11;
  int __pyx_t_12;
  Py_UCS4 __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recenter", 0);

  /* "gnocis/regions.pyx":502
 * 		:rtype: regions
 * 		"""
 * 		cdef list rlist = []             # <<<<<<<<<<<<<<
 * 		cdef long long rstart, rend
 * 		cdef region rgn
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rlist = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":505
 * 		cdef long long rstart, rend
 * 		cdef region rgn
 * 		for rgn in self.regions:             # <<<<<<<<<<<<<<
 * 			if len(rgn) < size:
 * 				#rstart = <long long>( ( rgn.start + rgn.end ) / 2 ) - <long long>(size/2)
 */
  if (unlikely(__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 505, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 505, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 505, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_rgn, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":506
 * 		cdef region rgn
 * 		for rgn in self.regions:
 * 			if len(rgn) < size:             # <<<<<<<<<<<<<<
 * 				#rstart = <long long>( ( rgn.start + rgn.end ) / 2 ) - <long long>(size/2)
 * 				rstart = <long long>( ( rgn.start / 2 ) + ( rgn.end / 2 ) ) - <long long>(size/2)
 */
    __pyx_t_4 = PyObject_Length(((PyObject *)__pyx_v_rgn)); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 506, __pyx_L1_error)
    __pyx_t_5 = ((__pyx_t_4 < __pyx_v_size) != 0);
    if (__pyx_t_5) {

      /* "gnocis/regions.pyx":508
 * 			if len(rgn) < size:
 * 				#rstart = <long long>( ( rgn.start + rgn.end ) / 2 ) - <long long>(size/2)
 * 				rstart = <long long>( ( rgn.start / 2 ) + ( rgn.end / 2 ) ) - <long long>(size/2)             # <<<<<<<<<<<<<<
 * 			elif len(rgn) > size:
 * 				rstart = rgn.start + <long long>( random.uniform( 0, len(rgn)-size ) )
 */
      __pyx_v_rstart = (((PY_LONG_LONG)((((double)__pyx_v_rgn->start) / 2.0) + (((double)__pyx_v_rgn->end) / 2.0))) - ((PY_LONG_LONG)(((double)__pyx_v_size) / 2.0)));

      /* "gnocis/regions.pyx":506
 * 		cdef region rgn
 * 		for rgn in self.regions:
 * 			if len(rgn) < size:             # <<<<<<<<<<<<<<
 * 				#rstart = <long long>( ( rgn.start + rgn.end ) / 2 ) - <long long>(size/2)
 * 				rstart = <long long>( ( rgn.start / 2 ) + ( rgn.end / 2 ) ) - <long long>(size/2)
 */
      goto __pyx_L5;
    }

    /* "gnocis/regions.pyx":509
 * 				#rstart = <long long>( ( rgn.start + rgn.end ) / 2 ) - <long long>(size/2)
 * 				rstart = <long long>( ( rgn.start / 2 ) + ( rgn.end / 2 ) ) - <long long>(size/2)
 * 			elif len(rgn) > size:             # <<<<<<<<<<<<<<
 * 				rstart = rgn.start + <long long>( random.uniform( 0, len(rgn)-size ) )
 * 				#rstart += <long long>( random.uniform( 0, len(rgn)-size ) )
 */
    __pyx_t_4 = PyObject_Length(((PyObject *)__pyx_v_rgn)); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 509, __pyx_L1_error)
    __pyx_t_5 = ((__pyx_t_4 > __pyx_v_size) != 0);
    if (__pyx_t_5) {

      /* "gnocis/regions.pyx":510
 * 				rstart = <long long>( ( rgn.start / 2 ) + ( rgn.end / 2 ) ) - <long long>(size/2)
 * 			elif len(rgn) > size:
 * 				rstart = rgn.start + <long long>( random.uniform( 0, len(rgn)-size ) )             # <<<<<<<<<<<<<<
 * 				#rstart += <long long>( random.uniform( 0, len(rgn)-size ) )
 * 			else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_random); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_uniform); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_4 = PyObject_Length(((PyObject *)__pyx_v_rgn)); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 510, __pyx_L1_error)
      __pyx_t_6 = __Pyx_PyInt_From_PY_LONG_LONG((__pyx_t_4 - __pyx_v_size)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_int_0, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_int_0, __pyx_t_6};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 510, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_int_0);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_11 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_3); if (unlikely((__pyx_t_11 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_rstart = (__pyx_v_rgn->start + ((PY_LONG_LONG)__pyx_t_11));

      /* "gnocis/regions.pyx":509
 * 				#rstart = <long long>( ( rgn.start + rgn.end ) / 2 ) - <long long>(size/2)
 * 				rstart = <long long>( ( rgn.start / 2 ) + ( rgn.end / 2 ) ) - <long long>(size/2)
 * 			elif len(rgn) > size:             # <<<<<<<<<<<<<<
 * 				rstart = rgn.start + <long long>( random.uniform( 0, len(rgn)-size ) )
 * 				#rstart += <long long>( random.uniform( 0, len(rgn)-size ) )
 */
      goto __pyx_L5;
    }

    /* "gnocis/regions.pyx":513
 * 				#rstart += <long long>( random.uniform( 0, len(rgn)-size ) )
 * 			else:
 * 				rstart = rgn.start             # <<<<<<<<<<<<<<
 * 			rend = rstart + size - 1
 * 			rlist.append(region(rgn.seq, rstart, rend, strand = rgn.strand, score = rgn.score, source = rgn.source, feature = rgn.feature, group = rgn.group))
 */
    /*else*/ {
      __pyx_t_11 = __pyx_v_rgn->start;
      __pyx_v_rstart = __pyx_t_11;
    }
    __pyx_L5:;

    /* "gnocis/regions.pyx":514
 * 			else:
 * 				rstart = rgn.start
 * 			rend = rstart + size - 1             # <<<<<<<<<<<<<<
 * 			rlist.append(region(rgn.seq, rstart, rend, strand = rgn.strand, score = rgn.score, source = rgn.source, feature = rgn.feature, group = rgn.group))
 * 		return regions('%s (recentered - %d bp)'%(self.name, size), rlist)
 */
    __pyx_v_rend = ((__pyx_v_rstart + __pyx_v_size) - 1);

    /* "gnocis/regions.pyx":515
 * 				rstart = rgn.start
 * 			rend = rstart + size - 1
 * 			rlist.append(region(rgn.seq, rstart, rend, strand = rgn.strand, score = rgn.score, source = rgn.source, feature = rgn.feature, group = rgn.group))             # <<<<<<<<<<<<<<
 * 		return regions('%s (recentered - %d bp)'%(self.name, size), rlist)
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_rstart); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_rend); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_rgn->seq);
    __Pyx_GIVEREF(__pyx_v_rgn->seq);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_rgn->seq);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_7);
    __pyx_t_3 = 0;
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_rgn->strand); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_strand, __pyx_t_3) < 0) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyFloat_FromDouble(__pyx_v_rgn->score); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_score, __pyx_t_3) < 0) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_source, __pyx_v_rgn->source) < 0) __PYX_ERR(0, 515, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_feature, __pyx_v_rgn->feature) < 0) __PYX_ERR(0, 515, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_group, __pyx_v_rgn->group) < 0) __PYX_ERR(0, 515, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_10, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_rlist, __pyx_t_3); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":505
 * 		cdef long long rstart, rend
 * 		cdef region rgn
 * 		for rgn in self.regions:             # <<<<<<<<<<<<<<
 * 			if len(rgn) < size:
 * 				#rstart = <long long>( ( rgn.start + rgn.end ) / 2 ) - <long long>(size/2)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":516
 * 			rend = rstart + size - 1
 * 			rlist.append(region(rgn.seq, rstart, rend, strand = rgn.strand, score = rgn.score, source = rgn.source, feature = rgn.feature, group = rgn.group))
 * 		return regions('%s (recentered - %d bp)'%(self.name, size), rlist)             # <<<<<<<<<<<<<<
 * 
 * 	# Returns a set of delta-resized regions.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_13 = 127;
  __pyx_t_3 = __Pyx_PyUnicode_Unicode(__pyx_v_self->name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_13;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_recentered);
  __pyx_t_2 += 15;
  __Pyx_GIVEREF(__pyx_kp_u_recentered);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u_recentered);
  __pyx_t_3 = __Pyx_PyUnicode_From_PY_LONG_LONG(__pyx_v_size, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u_bp);
  __pyx_t_2 += 4;
  __Pyx_GIVEREF(__pyx_kp_u_bp);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_bp);
  __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_2, __pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_rlist);
  __Pyx_GIVEREF(__pyx_v_rlist);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_rlist);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":493
 * 
 * 	# Returns a set of recentered regions, with regions randomly placed within larger regions.
 * 	def recenter(self, long long size):             # <<<<<<<<<<<<<<
 * 		""" Gets a set of randomly recentered regions. If the target size is smaller than a given region, a region of the desired size is randomly placed within the region. If the desired size is larger, a region of the desired size is placed with equal center to the center of the region.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("gnocis.regions.regions.recenter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rlist);
  __Pyx_XDECREF((PyObject *)__pyx_v_rgn);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":519
 * 
 * 	# Returns a set of delta-resized regions.
 * 	def deltaResize(self, sizeDelta):             # <<<<<<<<<<<<<<
 * 		""" Gets a set of resized regions. The delta size is subtracted from the start and added to the end.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_59deltaResize(PyObject *__pyx_v_self, PyObject *__pyx_v_sizeDelta); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_58deltaResize[] = " Gets a set of resized regions. The delta size is subtracted from the start and added to the end.\n\t\t\n\t\t:param sizeDelta: Desired region size change.\n\t\t:type size: int\n\t\t\n\t\t:return: Set of resized regions.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_59deltaResize(PyObject *__pyx_v_self, PyObject *__pyx_v_sizeDelta) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("deltaResize (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_58deltaResize(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_sizeDelta));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_58deltaResize(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_sizeDelta) {
  PyObject *__pyx_9genexpr14__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("deltaResize", 0);

  /* "gnocis/regions.pyx":528
 * 		:rtype: regions
 * 		"""
 * 		return regions(self.name + ' (%s %d bp)'%('+' if sizeDelta >= 0 else '-', sizeDelta), [             # <<<<<<<<<<<<<<
 * 			region(r.seq, r.start - sizeDelta, r.end + sizeDelta, r.strand, r.score, r.source, r.feature, r.group)
 * 			for r in self.regions
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u__5);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__5);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_sizeDelta, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_4 = __pyx_kp_u__6;
  } else {
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_4 = __pyx_kp_u__7;
  }
  __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__17);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__17);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__17);
  __pyx_t_5 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_v_sizeDelta), __pyx_n_u_d); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_bp);
  __pyx_t_2 += 4;
  __Pyx_GIVEREF(__pyx_kp_u_bp);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_bp);
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_self->name, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  { /* enter inner scope */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "gnocis/regions.pyx":530
 * 		return regions(self.name + ' (%s %d bp)'%('+' if sizeDelta >= 0 else '-', sizeDelta), [
 * 			region(r.seq, r.start - sizeDelta, r.end + sizeDelta, r.strand, r.score, r.source, r.feature, r.group)
 * 			for r in self.regions             # <<<<<<<<<<<<<<
 * 			if ( (r.end + sizeDelta) - (r.start - sizeDelta) ) > 0
 * 		])
 */
    if (unlikely(__pyx_v_self->regions == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 530, __pyx_L5_error)
    }
    __pyx_t_4 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_4); __pyx_t_2 = 0;
    for (;;) {
      if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_4)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_2); __Pyx_INCREF(__pyx_t_7); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 530, __pyx_L5_error)
      #else
      __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 530, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_r, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "gnocis/regions.pyx":531
 * 			region(r.seq, r.start - sizeDelta, r.end + sizeDelta, r.strand, r.score, r.source, r.feature, r.group)
 * 			for r in self.regions
 * 			if ( (r.end + sizeDelta) - (r.start - sizeDelta) ) > 0             # <<<<<<<<<<<<<<
 * 		])
 * 
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_end); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 531, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PyNumber_Add(__pyx_t_7, __pyx_v_sizeDelta); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 531, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_start); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 531, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = PyNumber_Subtract(__pyx_t_7, __pyx_v_sizeDelta); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 531, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyNumber_Subtract(__pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 531, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_7, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 531, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 531, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_6) {

        /* "gnocis/regions.pyx":529
 * 		"""
 * 		return regions(self.name + ' (%s %d bp)'%('+' if sizeDelta >= 0 else '-', sizeDelta), [
 * 			region(r.seq, r.start - sizeDelta, r.end + sizeDelta, r.strand, r.score, r.source, r.feature, r.group)             # <<<<<<<<<<<<<<
 * 			for r in self.regions
 * 			if ( (r.end + sizeDelta) - (r.start - sizeDelta) ) > 0
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_start); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyNumber_Subtract(__pyx_t_7, __pyx_v_sizeDelta); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_end); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = PyNumber_Add(__pyx_t_7, __pyx_v_sizeDelta); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_strand); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_score); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_source); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_feature); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_r, __pyx_n_s_group); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = PyTuple_New(8); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_14);
        __pyx_t_9 = 0;
        __pyx_t_8 = 0;
        __pyx_t_10 = 0;
        __pyx_t_7 = 0;
        __pyx_t_11 = 0;
        __pyx_t_12 = 0;
        __pyx_t_13 = 0;
        __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_15, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_14))) __PYX_ERR(0, 528, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

        /* "gnocis/regions.pyx":531
 * 			region(r.seq, r.start - sizeDelta, r.end + sizeDelta, r.strand, r.score, r.source, r.feature, r.group)
 * 			for r in self.regions
 * 			if ( (r.end + sizeDelta) - (r.start - sizeDelta) ) > 0             # <<<<<<<<<<<<<<
 * 		])
 * 
 */
      }

      /* "gnocis/regions.pyx":530
 * 		return regions(self.name + ' (%s %d bp)'%('+' if sizeDelta >= 0 else '-', sizeDelta), [
 * 			region(r.seq, r.start - sizeDelta, r.end + sizeDelta, r.strand, r.score, r.source, r.feature, r.group)
 * 			for r in self.regions             # <<<<<<<<<<<<<<
 * 			if ( (r.end + sizeDelta) - (r.start - sizeDelta) ) > 0
 * 		])
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_r); __pyx_9genexpr14__pyx_v_r = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_r); __pyx_9genexpr14__pyx_v_r = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */

  /* "gnocis/regions.pyx":528
 * 		:rtype: regions
 * 		"""
 * 		return regions(self.name + ' (%s %d bp)'%('+' if sizeDelta >= 0 else '-', sizeDelta), [             # <<<<<<<<<<<<<<
 * 			region(r.seq, r.start - sizeDelta, r.end + sizeDelta, r.strand, r.score, r.source, r.feature, r.group)
 * 			for r in self.regions
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":519
 * 
 * 	# Returns a set of delta-resized regions.
 * 	def deltaResize(self, sizeDelta):             # <<<<<<<<<<<<<<
 * 		""" Gets a set of resized regions. The delta size is subtracted from the start and added to the end.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("gnocis.regions.regions.deltaResize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":535
 * 
 * 	# Returns a renamed region set.
 * 	def rename(self, newname):             # <<<<<<<<<<<<<<
 * 		""":param newname: New name.
 * 		:type newname: str
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_61rename(PyObject *__pyx_v_self, PyObject *__pyx_v_newname); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_60rename[] = ":param newname: New name.\n\t\t:type newname: str\n\t\t\n\t\t:return: Renamed region set.\n\t\t:rtype: regions\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_61rename(PyObject *__pyx_v_self, PyObject *__pyx_v_newname) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rename (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_60rename(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_newname));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_60rename(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_newname) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("rename", 0);

  /* "gnocis/regions.pyx":542
 * 		:rtype: regions
 * 		"""
 * 		return regions(newname, self.regions)             # <<<<<<<<<<<<<<
 * 
 * 	# Sorts regions, first by start coordinate, then by sequence.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_newname);
  __Pyx_GIVEREF(__pyx_v_newname);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_newname);
  __Pyx_INCREF(__pyx_v_self->regions);
  __Pyx_GIVEREF(__pyx_v_self->regions);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->regions);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":535
 * 
 * 	# Returns a renamed region set.
 * 	def rename(self, newname):             # <<<<<<<<<<<<<<
 * 		""":param newname: New name.
 * 		:type newname: str
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.regions.regions.rename", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":545
 * 
 * 	# Sorts regions, first by start coordinate, then by sequence.
 * 	def sort(self):             # <<<<<<<<<<<<<<
 * 		""" Sorts the set, first by start coordinate, then by sequence."""
 * 		cdef dict rdict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_63sort(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_62sort[] = " Sorts the set, first by start coordinate, then by sequence.";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_63sort(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sort (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_62sort(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":560
 * 		self.regions = []
 * 		for cseq in sortedKeys:
 * 			self.regions += sorted(rdict[cseq], key = lambda x: x.start)             # <<<<<<<<<<<<<<
 * 
 * 	# Extracts region sequences from a FASTA file.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_4sort_lambda5(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_7regions_4sort_lambda5 = {"lambda5", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_4sort_lambda5, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_7regions_7regions_4sort_lambda5(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda5 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda5(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda5(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda5", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.regions.sort.lambda5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":545
 * 
 * 	# Sorts regions, first by start coordinate, then by sequence.
 * 	def sort(self):             # <<<<<<<<<<<<<<
 * 		""" Sorts the set, first by start coordinate, then by sequence."""
 * 		cdef dict rdict
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_62sort(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_v_rdict = 0;
  PyObject *__pyx_v_sortedKeys = 0;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_rgn = 0;
  PyObject *__pyx_v_cseq = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sort", 0);

  /* "gnocis/regions.pyx":551
 * 		cdef region rgn, x
 * 		cdef str cseq
 * 		rdict = {}             # <<<<<<<<<<<<<<
 * 		for rgn in self.regions:
 * 			if not rgn.seq in rdict:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rdict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":552
 * 		cdef str cseq
 * 		rdict = {}
 * 		for rgn in self.regions:             # <<<<<<<<<<<<<<
 * 			if not rgn.seq in rdict:
 * 				rdict[rgn.seq] = [rgn]
 */
  if (unlikely(__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 552, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 552, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_rgn, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":553
 * 		rdict = {}
 * 		for rgn in self.regions:
 * 			if not rgn.seq in rdict:             # <<<<<<<<<<<<<<
 * 				rdict[rgn.seq] = [rgn]
 * 			else:
 */
    __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_v_rgn->seq, __pyx_v_rdict, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 553, __pyx_L1_error)
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "gnocis/regions.pyx":554
 * 		for rgn in self.regions:
 * 			if not rgn.seq in rdict:
 * 				rdict[rgn.seq] = [rgn]             # <<<<<<<<<<<<<<
 * 			else:
 * 				rdict[rgn.seq].append(rgn)
 */
      __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 554, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(((PyObject *)__pyx_v_rgn));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_rgn));
      PyList_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_rgn));
      if (unlikely(PyDict_SetItem(__pyx_v_rdict, __pyx_v_rgn->seq, __pyx_t_3) < 0)) __PYX_ERR(0, 554, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "gnocis/regions.pyx":553
 * 		rdict = {}
 * 		for rgn in self.regions:
 * 			if not rgn.seq in rdict:             # <<<<<<<<<<<<<<
 * 				rdict[rgn.seq] = [rgn]
 * 			else:
 */
      goto __pyx_L5;
    }

    /* "gnocis/regions.pyx":556
 * 				rdict[rgn.seq] = [rgn]
 * 			else:
 * 				rdict[rgn.seq].append(rgn)             # <<<<<<<<<<<<<<
 * 		sortedKeys = sorted(rdict.keys())
 * 		self.regions = []
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_rdict, __pyx_v_rgn->seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 556, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyObject_Append(__pyx_t_3, ((PyObject *)__pyx_v_rgn)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 556, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L5:;

    /* "gnocis/regions.pyx":552
 * 		cdef str cseq
 * 		rdict = {}
 * 		for rgn in self.regions:             # <<<<<<<<<<<<<<
 * 			if not rgn.seq in rdict:
 * 				rdict[rgn.seq] = [rgn]
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":557
 * 			else:
 * 				rdict[rgn.seq].append(rgn)
 * 		sortedKeys = sorted(rdict.keys())             # <<<<<<<<<<<<<<
 * 		self.regions = []
 * 		for cseq in sortedKeys:
 */
  __pyx_t_3 = __Pyx_PyDict_Keys(__pyx_v_rdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_6 = PyList_Sort(__pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 557, __pyx_L1_error)
  __pyx_v_sortedKeys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":558
 * 				rdict[rgn.seq].append(rgn)
 * 		sortedKeys = sorted(rdict.keys())
 * 		self.regions = []             # <<<<<<<<<<<<<<
 * 		for cseq in sortedKeys:
 * 			self.regions += sorted(rdict[cseq], key = lambda x: x.start)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->regions);
  __Pyx_DECREF(__pyx_v_self->regions);
  __pyx_v_self->regions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":559
 * 		sortedKeys = sorted(rdict.keys())
 * 		self.regions = []
 * 		for cseq in sortedKeys:             # <<<<<<<<<<<<<<
 * 			self.regions += sorted(rdict[cseq], key = lambda x: x.start)
 * 
 */
  if (unlikely(__pyx_v_sortedKeys == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 559, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_sortedKeys; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_7); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 559, __pyx_L1_error)
    #else
    __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    #endif
    if (!(likely(PyUnicode_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_cseq, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "gnocis/regions.pyx":560
 * 		self.regions = []
 * 		for cseq in sortedKeys:
 * 			self.regions += sorted(rdict[cseq], key = lambda x: x.start)             # <<<<<<<<<<<<<<
 * 
 * 	# Extracts region sequences from a FASTA file.
 */
    __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_rdict, __pyx_v_cseq); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_7regions_7regions_4sort_lambda5, 0, __pyx_n_s_sort_locals_lambda, NULL, __pyx_n_s_gnocis_regions, __pyx_d, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_key, __pyx_t_8) < 0) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_self->regions, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_7);
    __Pyx_GOTREF(__pyx_v_self->regions);
    __Pyx_DECREF(__pyx_v_self->regions);
    __pyx_v_self->regions = ((PyObject*)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "gnocis/regions.pyx":559
 * 		sortedKeys = sorted(rdict.keys())
 * 		self.regions = []
 * 		for cseq in sortedKeys:             # <<<<<<<<<<<<<<
 * 			self.regions += sorted(rdict[cseq], key = lambda x: x.start)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":545
 * 
 * 	# Sorts regions, first by start coordinate, then by sequence.
 * 	def sort(self):             # <<<<<<<<<<<<<<
 * 		""" Sorts the set, first by start coordinate, then by sequence."""
 * 		cdef dict rdict
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("gnocis.regions.regions.sort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rdict);
  __Pyx_XDECREF(__pyx_v_sortedKeys);
  __Pyx_XDECREF((PyObject *)__pyx_v_rgn);
  __Pyx_XDECREF(__pyx_v_cseq);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":563
 * 
 * 	# Extracts region sequences from a FASTA file.
 * 	def extractSequences(self, seqs):             # <<<<<<<<<<<<<<
 * 		""" Extracts region sequences from a sequence set or stream.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_65extractSequences(PyObject *__pyx_v_self, PyObject *__pyx_v_seqs); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_64extractSequences[] = " Extracts region sequences from a sequence set or stream.\n\t\t\n\t\t:param seqs: Sequences/sequence stream to extract region sequences from.\n\t\t:type seqs: sequences/sequenceStream\n\t\t\n\t\t:return: Extracted region sequences.\n\t\t:rtype: sequences\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_65extractSequences(PyObject *__pyx_v_self, PyObject *__pyx_v_seqs) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extractSequences (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_64extractSequences(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_seqs));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_64extractSequences(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_seqs) {
  PyObject *__pyx_v_cSeqRegions = NULL;
  struct __pyx_obj_6gnocis_9sequences_sequences *__pyx_v_ret = NULL;
  PyObject *__pyx_v_cseq = NULL;
  PyObject *__pyx_v_aSeq = NULL;
  PyObject *__pyx_v_bSeq = NULL;
  PyObject *__pyx_v_blk = NULL;
  PyObject *__pyx_v_firstValidRegion = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_9genexpr15__pyx_v_r = NULL;
  PyObject *__pyx_9genexpr16__pyx_v_r = NULL;
  PyObject *__pyx_9genexpr17__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  Py_ssize_t __pyx_t_15;
  long __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  Py_ssize_t __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extractSequences", 0);

  /* "gnocis/regions.pyx":572
 * 		:rtype: sequences
 * 		"""
 * 		cSeqRegions = []             # <<<<<<<<<<<<<<
 * 		ret = sequences('Regions set: ' + self.name + ' - from sequence stream: ' + str(seqs), [])
 * 		if isinstance(seqs, sequences) or isinstance(seqs, list):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cSeqRegions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":573
 * 		"""
 * 		cSeqRegions = []
 * 		ret = sequences('Regions set: ' + self.name + ' - from sequence stream: ' + str(seqs), [])             # <<<<<<<<<<<<<<
 * 		if isinstance(seqs, sequences) or isinstance(seqs, list):
 * 			for cseq in seqs:
 */
  __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_kp_u_Regions_set, __pyx_v_self->name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_kp_u_from_sequence_stream); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_seqs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_sequences), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ret = ((struct __pyx_obj_6gnocis_9sequences_sequences *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":574
 * 		cSeqRegions = []
 * 		ret = sequences('Regions set: ' + self.name + ' - from sequence stream: ' + str(seqs), [])
 * 		if isinstance(seqs, sequences) or isinstance(seqs, list):             # <<<<<<<<<<<<<<
 * 			for cseq in seqs:
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_seqs, __pyx_ptype_6gnocis_9sequences_sequences); 
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = PyList_Check(__pyx_v_seqs); 
  __pyx_t_5 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "gnocis/regions.pyx":575
 * 		ret = sequences('Regions set: ' + self.name + ' - from sequence stream: ' + str(seqs), [])
 * 		if isinstance(seqs, sequences) or isinstance(seqs, list):
 * 			for cseq in seqs:             # <<<<<<<<<<<<<<
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]
 * 		elif isinstance(seqs, sequenceStream):
 */
    if (likely(PyList_CheckExact(__pyx_v_seqs)) || PyTuple_CheckExact(__pyx_v_seqs)) {
      __pyx_t_1 = __pyx_v_seqs; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_seqs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 575, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 575, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 575, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_8(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 575, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_cseq, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "gnocis/regions.pyx":576
 * 		if isinstance(seqs, sequences) or isinstance(seqs, list):
 * 			for cseq in seqs:
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]             # <<<<<<<<<<<<<<
 * 		elif isinstance(seqs, sequenceStream):
 * 			for cseq in seqs:
 */
      { /* enter inner scope */
        __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 576, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(__pyx_v_self->regions == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 576, __pyx_L10_error)
        }
        __pyx_t_3 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
        for (;;) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_10); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 576, __pyx_L10_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 576, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
          __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_r, __pyx_t_10);
          __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr15__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 576, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cseq, __pyx_n_s_name); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 576, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = PyObject_RichCompare(__pyx_t_10, __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 576, __pyx_L10_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 576, __pyx_L10_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (__pyx_t_4) {
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_cseq, __pyx_n_s_seq); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr15__pyx_v_r, __pyx_n_s_start); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr15__pyx_v_r, __pyx_n_s_end); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_13 = __Pyx_PyInt_AddObjC(__pyx_t_10, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_t_12, 0, 0, &__pyx_t_11, &__pyx_t_13, NULL, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            PyTuple_SET_ITEM(__pyx_t_13, 0, Py_None);
            __Pyx_GIVEREF(__pyx_t_10);
            PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_10);
            __pyx_t_10 = 0;
            __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_sourceRegion, __pyx_9genexpr15__pyx_v_r) < 0) __PYX_ERR(0, 576, __pyx_L10_error)
            __pyx_t_11 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_sequence), __pyx_t_13, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_11))) __PYX_ERR(0, 576, __pyx_L10_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_r); __pyx_9genexpr15__pyx_v_r = 0;
        goto __pyx_L14_exit_scope;
        __pyx_L10_error:;
        __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_r); __pyx_9genexpr15__pyx_v_r = 0;
        goto __pyx_L1_error;
        __pyx_L14_exit_scope:;
      } /* exit inner scope */
      __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_ret->sequences, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_v_ret->sequences);
      __Pyx_DECREF(__pyx_v_ret->sequences);
      __pyx_v_ret->sequences = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "gnocis/regions.pyx":575
 * 		ret = sequences('Regions set: ' + self.name + ' - from sequence stream: ' + str(seqs), [])
 * 		if isinstance(seqs, sequences) or isinstance(seqs, list):
 * 			for cseq in seqs:             # <<<<<<<<<<<<<<
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]
 * 		elif isinstance(seqs, sequenceStream):
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/regions.pyx":574
 * 		cSeqRegions = []
 * 		ret = sequences('Regions set: ' + self.name + ' - from sequence stream: ' + str(seqs), [])
 * 		if isinstance(seqs, sequences) or isinstance(seqs, list):             # <<<<<<<<<<<<<<
 * 			for cseq in seqs:
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]
 */
    goto __pyx_L3;
  }

  /* "gnocis/regions.pyx":577
 * 			for cseq in seqs:
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]
 * 		elif isinstance(seqs, sequenceStream):             # <<<<<<<<<<<<<<
 * 			for cseq in seqs:
 * 				cSeqRegions = [ { 'start':r.start, 'end':r.end, 'seq':sequence(None, '', sourceRegion = r) } for r in self.regions if r.seq == cseq.name ]
 */
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_seqs, __pyx_ptype_6gnocis_9sequences_sequenceStream); 
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "gnocis/regions.pyx":578
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]
 * 		elif isinstance(seqs, sequenceStream):
 * 			for cseq in seqs:             # <<<<<<<<<<<<<<
 * 				cSeqRegions = [ { 'start':r.start, 'end':r.end, 'seq':sequence(None, '', sourceRegion = r) } for r in self.regions if r.seq == cseq.name ]
 * 				if len(cSeqRegions) == 0: continue
 */
    if (likely(PyList_CheckExact(__pyx_v_seqs)) || PyTuple_CheckExact(__pyx_v_seqs)) {
      __pyx_t_1 = __pyx_v_seqs; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_seqs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 578, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 578, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 578, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_8(__pyx_t_1);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 578, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_cseq, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "gnocis/regions.pyx":579
 * 		elif isinstance(seqs, sequenceStream):
 * 			for cseq in seqs:
 * 				cSeqRegions = [ { 'start':r.start, 'end':r.end, 'seq':sequence(None, '', sourceRegion = r) } for r in self.regions if r.seq == cseq.name ]             # <<<<<<<<<<<<<<
 * 				if len(cSeqRegions) == 0: continue
 * 				ret.sequences += [ r['seq'] for r in cSeqRegions ]
 */
      { /* enter inner scope */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__pyx_v_self->regions == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
          __PYX_ERR(0, 579, __pyx_L19_error)
        }
        __pyx_t_2 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_2); __pyx_t_9 = 0;
        for (;;) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_11); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 579, __pyx_L19_error)
          #else
          __pyx_t_11 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          #endif
          __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_r, __pyx_t_11);
          __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr16__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_cseq, __pyx_n_s_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 579, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_13 = PyObject_RichCompare(__pyx_t_11, __pyx_t_10, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 579, __pyx_L19_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 579, __pyx_L19_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (__pyx_t_5) {
            __pyx_t_13 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr16__pyx_v_r, __pyx_n_s_start); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (PyDict_SetItem(__pyx_t_13, __pyx_n_u_start, __pyx_t_10) < 0) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr16__pyx_v_r, __pyx_n_s_end); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (PyDict_SetItem(__pyx_t_13, __pyx_n_u_end, __pyx_t_10) < 0) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_sourceRegion, __pyx_9genexpr16__pyx_v_r) < 0) __PYX_ERR(0, 579, __pyx_L19_error)
            __pyx_t_11 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_sequence), __pyx_tuple__18, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (PyDict_SetItem(__pyx_t_13, __pyx_n_u_seq, __pyx_t_11) < 0) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_13))) __PYX_ERR(0, 579, __pyx_L19_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_r); __pyx_9genexpr16__pyx_v_r = 0;
        goto __pyx_L23_exit_scope;
        __pyx_L19_error:;
        __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_r); __pyx_9genexpr16__pyx_v_r = 0;
        goto __pyx_L1_error;
        __pyx_L23_exit_scope:;
      } /* exit inner scope */
      __Pyx_DECREF_SET(__pyx_v_cSeqRegions, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "gnocis/regions.pyx":580
 * 			for cseq in seqs:
 * 				cSeqRegions = [ { 'start':r.start, 'end':r.end, 'seq':sequence(None, '', sourceRegion = r) } for r in self.regions if r.seq == cseq.name ]
 * 				if len(cSeqRegions) == 0: continue             # <<<<<<<<<<<<<<
 * 				ret.sequences += [ r['seq'] for r in cSeqRegions ]
 * 				aSeq, bSeq = 0, 0
 */
      __pyx_t_9 = PyList_GET_SIZE(__pyx_v_cSeqRegions); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 580, __pyx_L1_error)
      __pyx_t_5 = ((__pyx_t_9 == 0) != 0);
      if (__pyx_t_5) {
        goto __pyx_L15_continue;
      }

      /* "gnocis/regions.pyx":581
 * 				cSeqRegions = [ { 'start':r.start, 'end':r.end, 'seq':sequence(None, '', sourceRegion = r) } for r in self.regions if r.seq == cseq.name ]
 * 				if len(cSeqRegions) == 0: continue
 * 				ret.sequences += [ r['seq'] for r in cSeqRegions ]             # <<<<<<<<<<<<<<
 * 				aSeq, bSeq = 0, 0
 * 				for blk in cseq:
 */
      { /* enter inner scope */
        __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __pyx_v_cSeqRegions; __Pyx_INCREF(__pyx_t_2); __pyx_t_9 = 0;
        for (;;) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_13 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_13); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 581, __pyx_L27_error)
          #else
          __pyx_t_13 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 581, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_13);
          #endif
          __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_r, __pyx_t_13);
          __pyx_t_13 = 0;
          __pyx_t_13 = __Pyx_PyObject_Dict_GetItem(__pyx_9genexpr17__pyx_v_r, __pyx_n_u_seq); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 581, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_13))) __PYX_ERR(0, 581, __pyx_L27_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_r); __pyx_9genexpr17__pyx_v_r = 0;
        goto __pyx_L30_exit_scope;
        __pyx_L27_error:;
        __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_r); __pyx_9genexpr17__pyx_v_r = 0;
        goto __pyx_L1_error;
        __pyx_L30_exit_scope:;
      } /* exit inner scope */
      __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_ret->sequences, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_v_ret->sequences);
      __Pyx_DECREF(__pyx_v_ret->sequences);
      __pyx_v_ret->sequences = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "gnocis/regions.pyx":582
 * 				if len(cSeqRegions) == 0: continue
 * 				ret.sequences += [ r['seq'] for r in cSeqRegions ]
 * 				aSeq, bSeq = 0, 0             # <<<<<<<<<<<<<<
 * 				for blk in cseq:
 * 					bSeq = aSeq + len(blk)
 */
      __pyx_t_2 = __pyx_int_0;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = __pyx_int_0;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_aSeq, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_bSeq, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "gnocis/regions.pyx":583
 * 				ret.sequences += [ r['seq'] for r in cSeqRegions ]
 * 				aSeq, bSeq = 0, 0
 * 				for blk in cseq:             # <<<<<<<<<<<<<<
 * 					bSeq = aSeq + len(blk)
 * 					# Go through regions. If any overlap with the window, extract.
 */
      if (likely(PyList_CheckExact(__pyx_v_cseq)) || PyTuple_CheckExact(__pyx_v_cseq)) {
        __pyx_t_3 = __pyx_v_cseq; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
        __pyx_t_14 = NULL;
      } else {
        __pyx_t_9 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_cseq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_14 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 583, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_14)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 583, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 583, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_14(__pyx_t_3);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 583, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_blk, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "gnocis/regions.pyx":584
 * 				aSeq, bSeq = 0, 0
 * 				for blk in cseq:
 * 					bSeq = aSeq + len(blk)             # <<<<<<<<<<<<<<
 * 					# Go through regions. If any overlap with the window, extract.
 * 					# If any precede the window, dump them.
 */
        __pyx_t_15 = PyObject_Length(__pyx_v_blk); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 584, __pyx_L1_error)
        __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_13 = PyNumber_Add(__pyx_v_aSeq, __pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF_SET(__pyx_v_bSeq, __pyx_t_13);
        __pyx_t_13 = 0;

        /* "gnocis/regions.pyx":587
 * 					# Go through regions. If any overlap with the window, extract.
 * 					# If any precede the window, dump them.
 * 					firstValidRegion = 0             # <<<<<<<<<<<<<<
 * 					for i,r in enumerate(cSeqRegions):
 * 						if r['end'] + 1 < aSeq:
 */
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_XDECREF_SET(__pyx_v_firstValidRegion, __pyx_int_0);

        /* "gnocis/regions.pyx":588
 * 					# If any precede the window, dump them.
 * 					firstValidRegion = 0
 * 					for i,r in enumerate(cSeqRegions):             # <<<<<<<<<<<<<<
 * 						if r['end'] + 1 < aSeq:
 * 							firstValidRegion = i + 1
 */
        __Pyx_INCREF(__pyx_int_0);
        __pyx_t_13 = __pyx_int_0;
        __pyx_t_2 = __pyx_v_cSeqRegions; __Pyx_INCREF(__pyx_t_2); __pyx_t_15 = 0;
        for (;;) {
          if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_15); __Pyx_INCREF(__pyx_t_11); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 588, __pyx_L1_error)
          #else
          __pyx_t_11 = PySequence_ITEM(__pyx_t_2, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 588, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          #endif
          __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_11);
          __pyx_t_11 = 0;
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_13);
          __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_t_13, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 588, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_13);
          __pyx_t_13 = __pyx_t_11;
          __pyx_t_11 = 0;

          /* "gnocis/regions.pyx":589
 * 					firstValidRegion = 0
 * 					for i,r in enumerate(cSeqRegions):
 * 						if r['end'] + 1 < aSeq:             # <<<<<<<<<<<<<<
 * 							firstValidRegion = i + 1
 * 							continue
 */
          __pyx_t_11 = __Pyx_PyObject_Dict_GetItem(__pyx_v_r, __pyx_n_u_end); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 589, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_t_11, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 589, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = PyObject_RichCompare(__pyx_t_10, __pyx_v_aSeq, Py_LT); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 589, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 589, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (__pyx_t_5) {

            /* "gnocis/regions.pyx":590
 * 					for i,r in enumerate(cSeqRegions):
 * 						if r['end'] + 1 < aSeq:
 * 							firstValidRegion = i + 1             # <<<<<<<<<<<<<<
 * 							continue
 * 						if r['start'] > bSeq:
 */
            __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 590, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF_SET(__pyx_v_firstValidRegion, __pyx_t_11);
            __pyx_t_11 = 0;

            /* "gnocis/regions.pyx":591
 * 						if r['end'] + 1 < aSeq:
 * 							firstValidRegion = i + 1
 * 							continue             # <<<<<<<<<<<<<<
 * 						if r['start'] > bSeq:
 * 							break
 */
            goto __pyx_L33_continue;

            /* "gnocis/regions.pyx":589
 * 					firstValidRegion = 0
 * 					for i,r in enumerate(cSeqRegions):
 * 						if r['end'] + 1 < aSeq:             # <<<<<<<<<<<<<<
 * 							firstValidRegion = i + 1
 * 							continue
 */
          }

          /* "gnocis/regions.pyx":592
 * 							firstValidRegion = i + 1
 * 							continue
 * 						if r['start'] > bSeq:             # <<<<<<<<<<<<<<
 * 							break
 * 						r['seq'].seq += blk[ max(r['start']-aSeq, 0) : min(r['end']+1-aSeq, bSeq-aSeq) ]
 */
          __pyx_t_11 = __Pyx_PyObject_Dict_GetItem(__pyx_v_r, __pyx_n_u_start); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 592, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = PyObject_RichCompare(__pyx_t_11, __pyx_v_bSeq, Py_GT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 592, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 592, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_5) {

            /* "gnocis/regions.pyx":593
 * 							continue
 * 						if r['start'] > bSeq:
 * 							break             # <<<<<<<<<<<<<<
 * 						r['seq'].seq += blk[ max(r['start']-aSeq, 0) : min(r['end']+1-aSeq, bSeq-aSeq) ]
 * 					if firstValidRegion > 0:
 */
            goto __pyx_L34_break;

            /* "gnocis/regions.pyx":592
 * 							firstValidRegion = i + 1
 * 							continue
 * 						if r['start'] > bSeq:             # <<<<<<<<<<<<<<
 * 							break
 * 						r['seq'].seq += blk[ max(r['start']-aSeq, 0) : min(r['end']+1-aSeq, bSeq-aSeq) ]
 */
          }

          /* "gnocis/regions.pyx":594
 * 						if r['start'] > bSeq:
 * 							break
 * 						r['seq'].seq += blk[ max(r['start']-aSeq, 0) : min(r['end']+1-aSeq, bSeq-aSeq) ]             # <<<<<<<<<<<<<<
 * 					if firstValidRegion > 0:
 * 						cSeqRegions = cSeqRegions[firstValidRegion:]
 */
          __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_v_r, __pyx_n_u_seq); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_seq); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_16 = 0;
          __pyx_t_12 = __Pyx_PyObject_Dict_GetItem(__pyx_v_r, __pyx_n_u_start); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_17 = PyNumber_Subtract(__pyx_t_12, __pyx_v_aSeq); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_18 = __Pyx_PyInt_From_long(__pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_19 = PyObject_RichCompare(__pyx_t_18, __pyx_t_17, Py_GT); __Pyx_XGOTREF(__pyx_t_19); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_19); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (__pyx_t_5) {
            __pyx_t_19 = __Pyx_PyInt_From_long(__pyx_t_16); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 594, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_19);
            __pyx_t_12 = __pyx_t_19;
            __pyx_t_19 = 0;
          } else {
            __Pyx_INCREF(__pyx_t_17);
            __pyx_t_12 = __pyx_t_17;
          }
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_17 = PyNumber_Subtract(__pyx_v_bSeq, __pyx_v_aSeq); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_19 = __Pyx_PyObject_Dict_GetItem(__pyx_v_r, __pyx_n_u_end); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_18 = __Pyx_PyInt_AddObjC(__pyx_t_19, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          __pyx_t_19 = PyNumber_Subtract(__pyx_t_18, __pyx_v_aSeq); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_19);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          __pyx_t_20 = PyObject_RichCompare(__pyx_t_17, __pyx_t_19, Py_LT); __Pyx_XGOTREF(__pyx_t_20); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 594, __pyx_L1_error)
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_20); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          if (__pyx_t_5) {
            __Pyx_INCREF(__pyx_t_17);
            __pyx_t_18 = __pyx_t_17;
          } else {
            __Pyx_INCREF(__pyx_t_19);
            __pyx_t_18 = __pyx_t_19;
          }
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_17 = __Pyx_PyObject_GetSlice(__pyx_v_blk, 0, 0, &__pyx_t_12, &__pyx_t_18, NULL, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          __pyx_t_18 = PyNumber_InPlaceAdd(__pyx_t_11, __pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_t_10, __pyx_n_s_seq, __pyx_t_18) < 0) __PYX_ERR(0, 594, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "gnocis/regions.pyx":588
 * 					# If any precede the window, dump them.
 * 					firstValidRegion = 0
 * 					for i,r in enumerate(cSeqRegions):             # <<<<<<<<<<<<<<
 * 						if r['end'] + 1 < aSeq:
 * 							firstValidRegion = i + 1
 */
          __pyx_L33_continue:;
        }
        __pyx_L34_break:;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

        /* "gnocis/regions.pyx":595
 * 							break
 * 						r['seq'].seq += blk[ max(r['start']-aSeq, 0) : min(r['end']+1-aSeq, bSeq-aSeq) ]
 * 					if firstValidRegion > 0:             # <<<<<<<<<<<<<<
 * 						cSeqRegions = cSeqRegions[firstValidRegion:]
 * 					aSeq = bSeq
 */
        __pyx_t_13 = PyObject_RichCompare(__pyx_v_firstValidRegion, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 595, __pyx_L1_error)
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 595, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_5) {

          /* "gnocis/regions.pyx":596
 * 						r['seq'].seq += blk[ max(r['start']-aSeq, 0) : min(r['end']+1-aSeq, bSeq-aSeq) ]
 * 					if firstValidRegion > 0:
 * 						cSeqRegions = cSeqRegions[firstValidRegion:]             # <<<<<<<<<<<<<<
 * 					aSeq = bSeq
 * 		return ret
 */
          __Pyx_INCREF(__pyx_v_firstValidRegion);
          __pyx_t_13 = __pyx_v_firstValidRegion;
          __pyx_t_5 = (__pyx_t_13 == Py_None);
          if (__pyx_t_5) {
            __pyx_t_15 = 0;
          } else {
            __pyx_t_21 = __Pyx_PyIndex_AsSsize_t(__pyx_t_13); if (unlikely((__pyx_t_21 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 596, __pyx_L1_error)
            __pyx_t_15 = __pyx_t_21;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_13 = __Pyx_PyList_GetSlice(__pyx_v_cSeqRegions, __pyx_t_15, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 596, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF_SET(__pyx_v_cSeqRegions, ((PyObject*)__pyx_t_13));
          __pyx_t_13 = 0;

          /* "gnocis/regions.pyx":595
 * 							break
 * 						r['seq'].seq += blk[ max(r['start']-aSeq, 0) : min(r['end']+1-aSeq, bSeq-aSeq) ]
 * 					if firstValidRegion > 0:             # <<<<<<<<<<<<<<
 * 						cSeqRegions = cSeqRegions[firstValidRegion:]
 * 					aSeq = bSeq
 */
        }

        /* "gnocis/regions.pyx":597
 * 					if firstValidRegion > 0:
 * 						cSeqRegions = cSeqRegions[firstValidRegion:]
 * 					aSeq = bSeq             # <<<<<<<<<<<<<<
 * 		return ret
 * 
 */
        __Pyx_INCREF(__pyx_v_bSeq);
        __Pyx_DECREF_SET(__pyx_v_aSeq, __pyx_v_bSeq);

        /* "gnocis/regions.pyx":583
 * 				ret.sequences += [ r['seq'] for r in cSeqRegions ]
 * 				aSeq, bSeq = 0, 0
 * 				for blk in cseq:             # <<<<<<<<<<<<<<
 * 					bSeq = aSeq + len(blk)
 * 					# Go through regions. If any overlap with the window, extract.
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "gnocis/regions.pyx":578
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]
 * 		elif isinstance(seqs, sequenceStream):
 * 			for cseq in seqs:             # <<<<<<<<<<<<<<
 * 				cSeqRegions = [ { 'start':r.start, 'end':r.end, 'seq':sequence(None, '', sourceRegion = r) } for r in self.regions if r.seq == cseq.name ]
 * 				if len(cSeqRegions) == 0: continue
 */
      __pyx_L15_continue:;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/regions.pyx":577
 * 			for cseq in seqs:
 * 				ret.sequences += [ sequence(None, cseq.seq[r.start:r.end+1], sourceRegion = r) for r in self.regions if r.seq == cseq.name ]
 * 		elif isinstance(seqs, sequenceStream):             # <<<<<<<<<<<<<<
 * 			for cseq in seqs:
 * 				cSeqRegions = [ { 'start':r.start, 'end':r.end, 'seq':sequence(None, '', sourceRegion = r) } for r in self.regions if r.seq == cseq.name ]
 */
  }
  __pyx_L3:;

  /* "gnocis/regions.pyx":598
 * 						cSeqRegions = cSeqRegions[firstValidRegion:]
 * 					aSeq = bSeq
 * 		return ret             # <<<<<<<<<<<<<<
 * 
 * 	# Extracts region sequences from a 2bit file.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_ret));
  __pyx_r = ((PyObject *)__pyx_v_ret);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":563
 * 
 * 	# Extracts region sequences from a FASTA file.
 * 	def extractSequences(self, seqs):             # <<<<<<<<<<<<<<
 * 		""" Extracts region sequences from a sequence set or stream.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("gnocis.regions.regions.extractSequences", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cSeqRegions);
  __Pyx_XDECREF((PyObject *)__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_cseq);
  __Pyx_XDECREF(__pyx_v_aSeq);
  __Pyx_XDECREF(__pyx_v_bSeq);
  __Pyx_XDECREF(__pyx_v_blk);
  __Pyx_XDECREF(__pyx_v_firstValidRegion);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_r);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_r);
  __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":601
 * 
 * 	# Extracts region sequences from a 2bit file.
 * 	def extractSequencesFrom2bit(self, path):             # <<<<<<<<<<<<<<
 * 		""" Extracts region sequences from a 2bit file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_67extractSequencesFrom2bit(PyObject *__pyx_v_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_66extractSequencesFrom2bit[] = " Extracts region sequences from a 2bit file.\n\t\t\n\t\t:param path: Path to 2bit file.\n\t\t:type path: str\n\t\t\n\t\t:return: Extracted region sequences.\n\t\t:rtype: sequences\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_67extractSequencesFrom2bit(PyObject *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extractSequencesFrom2bit (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_66extractSequencesFrom2bit(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_66extractSequencesFrom2bit(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extractSequencesFrom2bit", 0);

  /* "gnocis/regions.pyx":610
 * 		:rtype: sequences
 * 		"""
 * 		return self.extractSequences( stream2bit(path) )             # <<<<<<<<<<<<<<
 * 
 * 	# Extracts region sequences from a FASTA file.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extractSequences); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6gnocis_9sequences_stream2bit(__pyx_v_path, 0, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":601
 * 
 * 	# Extracts region sequences from a 2bit file.
 * 	def extractSequencesFrom2bit(self, path):             # <<<<<<<<<<<<<<
 * 		""" Extracts region sequences from a 2bit file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.regions.regions.extractSequencesFrom2bit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":613
 * 
 * 	# Extracts region sequences from a FASTA file.
 * 	def extractSequencesFromFASTA(self, path):             # <<<<<<<<<<<<<<
 * 		""" Extracts region sequences from a FASTA file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_69extractSequencesFromFASTA(PyObject *__pyx_v_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_68extractSequencesFromFASTA[] = " Extracts region sequences from a FASTA file.\n\t\t\n\t\t:param path: Path to FASTA file.\n\t\t:type path: str\n\t\t\n\t\t:return: Extracted region sequences.\n\t\t:rtype: sequences\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_69extractSequencesFromFASTA(PyObject *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extractSequencesFromFASTA (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_68extractSequencesFromFASTA(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_68extractSequencesFromFASTA(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extractSequencesFromFASTA", 0);

  /* "gnocis/regions.pyx":622
 * 		:rtype: sequences
 * 		"""
 * 		return self.extractSequences( streamFASTA(path) )             # <<<<<<<<<<<<<<
 * 
 * 	# Extracts region sequences from a sequence list, sequence stream, or path.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extractSequences); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6gnocis_9sequences_streamFASTA(__pyx_v_path, 0, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":613
 * 
 * 	# Extracts region sequences from a FASTA file.
 * 	def extractSequencesFromFASTA(self, path):             # <<<<<<<<<<<<<<
 * 		""" Extracts region sequences from a FASTA file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.regions.regions.extractSequencesFromFASTA", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":625
 * 
 * 	# Extracts region sequences from a sequence list, sequence stream, or path.
 * 	def extract(self, src):             # <<<<<<<<<<<<<<
 * 		""" Extracts region sequences from a sequence set, sequence stream, or a file by path.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_71extract(PyObject *__pyx_v_self, PyObject *__pyx_v_src); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_70extract[] = " Extracts region sequences from a sequence set, sequence stream, or a file by path.\n\t\t\n\t\t:param src: Sequence set, sequence stream, or sequence file path.\n\t\t:type src: sequences/sequenceStream/str\n\t\t\n\t\t:return: Extracted region sequences.\n\t\t:rtype: sequences\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_71extract(PyObject *__pyx_v_self, PyObject *__pyx_v_src) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extract (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_70extract(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_src));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_70extract(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_src) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extract", 0);
  __Pyx_INCREF(__pyx_v_src);

  /* "gnocis/regions.pyx":634
 * 		:rtype: sequences
 * 		"""
 * 		if isinstance(src, str):             # <<<<<<<<<<<<<<
 * 			src = getSequenceStreamFromPath(src)
 * 		if type(src).__name__ == 'genome':
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_src); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "gnocis/regions.pyx":635
 * 		"""
 * 		if isinstance(src, str):
 * 			src = getSequenceStreamFromPath(src)             # <<<<<<<<<<<<<<
 * 		if type(src).__name__ == 'genome':
 * 			src = src.sequences
 */
    __pyx_t_3 = __pyx_f_6gnocis_9sequences_getSequenceStreamFromPath(__pyx_v_src, 0, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_src, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":634
 * 		:rtype: sequences
 * 		"""
 * 		if isinstance(src, str):             # <<<<<<<<<<<<<<
 * 			src = getSequenceStreamFromPath(src)
 * 		if type(src).__name__ == 'genome':
 */
  }

  /* "gnocis/regions.pyx":636
 * 		if isinstance(src, str):
 * 			src = getSequenceStreamFromPath(src)
 * 		if type(src).__name__ == 'genome':             # <<<<<<<<<<<<<<
 * 			src = src.sequences
 * 		if isinstance(src, sequences) or isinstance(src, list) or isinstance(src, sequenceStream):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_src)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_genome, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "gnocis/regions.pyx":637
 * 			src = getSequenceStreamFromPath(src)
 * 		if type(src).__name__ == 'genome':
 * 			src = src.sequences             # <<<<<<<<<<<<<<
 * 		if isinstance(src, sequences) or isinstance(src, list) or isinstance(src, sequenceStream):
 * 			return self.extractSequences(src)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_sequences); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_src, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "gnocis/regions.pyx":636
 * 		if isinstance(src, str):
 * 			src = getSequenceStreamFromPath(src)
 * 		if type(src).__name__ == 'genome':             # <<<<<<<<<<<<<<
 * 			src = src.sequences
 * 		if isinstance(src, sequences) or isinstance(src, list) or isinstance(src, sequenceStream):
 */
  }

  /* "gnocis/regions.pyx":638
 * 		if type(src).__name__ == 'genome':
 * 			src = src.sequences
 * 		if isinstance(src, sequences) or isinstance(src, list) or isinstance(src, sequenceStream):             # <<<<<<<<<<<<<<
 * 			return self.extractSequences(src)
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_src, __pyx_ptype_6gnocis_9sequences_sequences); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = PyList_Check(__pyx_v_src); 
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_src, __pyx_ptype_6gnocis_9sequences_sequenceStream); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "gnocis/regions.pyx":639
 * 			src = src.sequences
 * 		if isinstance(src, sequences) or isinstance(src, list) or isinstance(src, sequenceStream):
 * 			return self.extractSequences(src)             # <<<<<<<<<<<<<<
 * 
 * 	# Saves regions to GFF file.
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extractSequences); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_src) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_src);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "gnocis/regions.pyx":638
 * 		if type(src).__name__ == 'genome':
 * 			src = src.sequences
 * 		if isinstance(src, sequences) or isinstance(src, list) or isinstance(src, sequenceStream):             # <<<<<<<<<<<<<<
 * 			return self.extractSequences(src)
 * 
 */
  }

  /* "gnocis/regions.pyx":625
 * 
 * 	# Extracts region sequences from a sequence list, sequence stream, or path.
 * 	def extract(self, src):             # <<<<<<<<<<<<<<
 * 		""" Extracts region sequences from a sequence set, sequence stream, or a file by path.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.regions.regions.extract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":642
 * 
 * 	# Saves regions to GFF file.
 * 	def saveGFF(self, path):             # <<<<<<<<<<<<<<
 * 		""" Saves the regions to a General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_73saveGFF(PyObject *__pyx_v_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_72saveGFF[] = " Saves the regions to a General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.\n\t\t\n\t\t:param path: Output file path.\n\t\t:type path: str\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_73saveGFF(PyObject *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("saveGFF (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_72saveGFF(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_72saveGFF(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_v_fOut = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("saveGFF", 0);

  /* "gnocis/regions.pyx":648
 * 		:type path: str
 * 		"""
 * 		with open(path, 'w') as fOut:             # <<<<<<<<<<<<<<
 * 			for r in self.regions:
 * 				fOut.write('%s\t%s\t%s\t%d\t%d\t%f\t%s\t.\t1\n'%(r.seq, r.source, r.feature, r.start, r.end + 1, r.score, '+' if r.strand else '-'))
 */
  /*with:*/ {
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_path);
    __Pyx_GIVEREF(__pyx_v_path);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_path);
    __Pyx_INCREF(__pyx_n_u_w);
    __Pyx_GIVEREF(__pyx_n_u_w);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_w);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 648, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {
          __pyx_v_fOut = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":649
 * 		"""
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:             # <<<<<<<<<<<<<<
 * 				fOut.write('%s\t%s\t%s\t%d\t%d\t%f\t%s\t.\t1\n'%(r.seq, r.source, r.feature, r.start, r.end + 1, r.score, '+' if r.strand else '-'))
 * 
 */
          if (unlikely(__pyx_v_self->regions == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
            __PYX_ERR(0, 649, __pyx_L7_error)
          }
          __pyx_t_4 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
          for (;;) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 649, __pyx_L7_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 649, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
            __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":650
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:
 * 				fOut.write('%s\t%s\t%s\t%d\t%d\t%f\t%s\t.\t1\n'%(r.seq, r.source, r.feature, r.start, r.end + 1, r.score, '+' if r.strand else '-'))             # <<<<<<<<<<<<<<
 * 
 * 	# Saves regions to BED file.
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_fOut, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = PyTuple_New(14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_10 = 0;
            __pyx_t_11 = 127;
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_12), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_13);
            __pyx_t_13 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u__19);
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_source); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_12 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_13), __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u__19);
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_feature); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_12), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_13);
            __pyx_t_13 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_kp_u__19);
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_start); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_12 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_13), __pyx_n_u_d); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_kp_u__19);
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_end); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyInt_AddObjC(__pyx_t_12, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_12 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_13), __pyx_n_u_d); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_5, 8, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 9, __pyx_kp_u__19);
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_score); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_n_u_f); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_5, 10, __pyx_t_13);
            __pyx_t_13 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 11, __pyx_kp_u__19);
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_strand); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (__pyx_t_14) {
              __Pyx_INCREF(__pyx_kp_u__6);
              __pyx_t_13 = __pyx_kp_u__6;
            } else {
              __Pyx_INCREF(__pyx_kp_u__7);
              __pyx_t_13 = __pyx_kp_u__7;
            }
            __pyx_t_12 = __Pyx_PyUnicode_Unicode(__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_5, 12, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_INCREF(__pyx_kp_u_1);
            __pyx_t_10 += 5;
            __Pyx_GIVEREF(__pyx_kp_u_1);
            PyTuple_SET_ITEM(__pyx_t_5, 13, __pyx_kp_u_1);
            __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_5, 14, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_12);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":649
 * 		"""
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:             # <<<<<<<<<<<<<<
 * 				fOut.write('%s\t%s\t%s\t%d\t%d\t%f\t%s\t.\t1\n'%(r.seq, r.source, r.feature, r.start, r.end + 1, r.score, '+' if r.strand else '-'))
 * 
 */
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":648
 * 		:type path: str
 * 		"""
 * 		with open(path, 'w') as fOut:             # <<<<<<<<<<<<<<
 * 			for r in self.regions:
 * 				fOut.write('%s\t%s\t%s\t%d\t%d\t%f\t%s\t.\t1\n'%(r.seq, r.source, r.feature, r.start, r.end + 1, r.score, '+' if r.strand else '-'))
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.regions.saveGFF", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 648, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_12 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 648, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_12, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 648, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (__pyx_t_14 < 0) __PYX_ERR(0, 648, __pyx_L9_except_error)
          __pyx_t_16 = ((!(__pyx_t_14 != 0)) != 0);
          if (__pyx_t_16) {
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_XGIVEREF(__pyx_t_1);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_2, __pyx_t_1);
            __pyx_t_4 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
            __PYX_ERR(0, 648, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 648, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L18;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L18:;
  }

  /* "gnocis/regions.pyx":642
 * 
 * 	# Saves regions to GFF file.
 * 	def saveGFF(self, path):             # <<<<<<<<<<<<<<
 * 		""" Saves the regions to a General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("gnocis.regions.regions.saveGFF", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fOut);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":653
 * 
 * 	# Saves regions to BED file.
 * 	def saveBED(self, path):             # <<<<<<<<<<<<<<
 * 		""" Saves the regions to a BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_75saveBED(PyObject *__pyx_v_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_74saveBED[] = " Saves the regions to a BED (https://www.ensembl.org/info/website/upload/bed.html) file.\n\t\t\n\t\t:param path: Output file path.\n\t\t:type path: str\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_75saveBED(PyObject *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("saveBED (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_74saveBED(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_74saveBED(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_v_fOut = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("saveBED", 0);

  /* "gnocis/regions.pyx":659
 * 		:type path: str
 * 		"""
 * 		with open(path, 'w') as fOut:             # <<<<<<<<<<<<<<
 * 			for r in self.regions:
 * 				fOut.write('%s\t%d\t%d\t%s\t%f\t%s\n'%(r.seq, r.start, r.end, r.feature, r.score, '+' if r.strand else '-'))
 */
  /*with:*/ {
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_path);
    __Pyx_GIVEREF(__pyx_v_path);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_path);
    __Pyx_INCREF(__pyx_n_u_w);
    __Pyx_GIVEREF(__pyx_n_u_w);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_w);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 659, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {
          __pyx_v_fOut = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":660
 * 		"""
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:             # <<<<<<<<<<<<<<
 * 				fOut.write('%s\t%d\t%d\t%s\t%f\t%s\n'%(r.seq, r.start, r.end, r.feature, r.score, '+' if r.strand else '-'))
 * 
 */
          if (unlikely(__pyx_v_self->regions == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
            __PYX_ERR(0, 660, __pyx_L7_error)
          }
          __pyx_t_4 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
          for (;;) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 660, __pyx_L7_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 660, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
            __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":661
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:
 * 				fOut.write('%s\t%d\t%d\t%s\t%f\t%s\n'%(r.seq, r.start, r.end, r.feature, r.score, '+' if r.strand else '-'))             # <<<<<<<<<<<<<<
 * 
 * 	# Saves regions to coordinate list file.
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_fOut, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = PyTuple_New(12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_10 = 0;
            __pyx_t_11 = 127;
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_12), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_13);
            __pyx_t_13 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u__19);
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_start); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_12 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_13), __pyx_n_u_d); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u__19);
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_end); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_12), __pyx_n_u_d); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_13);
            __pyx_t_13 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_kp_u__19);
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_feature); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_12 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_13), __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_kp_u__19);
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_score); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_n_u_f); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_5, 8, __pyx_t_13);
            __pyx_t_13 = 0;
            __Pyx_INCREF(__pyx_kp_u__19);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__19);
            PyTuple_SET_ITEM(__pyx_t_5, 9, __pyx_kp_u__19);
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_strand); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (__pyx_t_14) {
              __Pyx_INCREF(__pyx_kp_u__6);
              __pyx_t_13 = __pyx_kp_u__6;
            } else {
              __Pyx_INCREF(__pyx_kp_u__7);
              __pyx_t_13 = __pyx_kp_u__7;
            }
            __pyx_t_12 = __Pyx_PyUnicode_Unicode(__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_5, 10, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_INCREF(__pyx_kp_u__21);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__21);
            PyTuple_SET_ITEM(__pyx_t_5, 11, __pyx_kp_u__21);
            __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_5, 12, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_12);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 661, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":660
 * 		"""
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:             # <<<<<<<<<<<<<<
 * 				fOut.write('%s\t%d\t%d\t%s\t%f\t%s\n'%(r.seq, r.start, r.end, r.feature, r.score, '+' if r.strand else '-'))
 * 
 */
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":659
 * 		:type path: str
 * 		"""
 * 		with open(path, 'w') as fOut:             # <<<<<<<<<<<<<<
 * 			for r in self.regions:
 * 				fOut.write('%s\t%d\t%d\t%s\t%f\t%s\n'%(r.seq, r.start, r.end, r.feature, r.score, '+' if r.strand else '-'))
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.regions.saveBED", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 659, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_12 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 659, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_12, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 659, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (__pyx_t_14 < 0) __PYX_ERR(0, 659, __pyx_L9_except_error)
          __pyx_t_16 = ((!(__pyx_t_14 != 0)) != 0);
          if (__pyx_t_16) {
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_XGIVEREF(__pyx_t_1);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_2, __pyx_t_1);
            __pyx_t_4 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
            __PYX_ERR(0, 659, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 659, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L18;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L18:;
  }

  /* "gnocis/regions.pyx":653
 * 
 * 	# Saves regions to BED file.
 * 	def saveBED(self, path):             # <<<<<<<<<<<<<<
 * 		""" Saves the regions to a BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("gnocis.regions.regions.saveBED", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fOut);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":664
 * 
 * 	# Saves regions to coordinate list file.
 * 	def saveCoordinateList(self, path):             # <<<<<<<<<<<<<<
 * 		""" Saves the regions to a coordinate list file (a coordinate per line, with the format `seq:start..end`).
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_77saveCoordinateList(PyObject *__pyx_v_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_76saveCoordinateList[] = " Saves the regions to a coordinate list file (a coordinate per line, with the format `seq:start..end`).\n\t\t\n\t\t:param path: Output file path.\n\t\t:type path: str\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_77saveCoordinateList(PyObject *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("saveCoordinateList (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_76saveCoordinateList(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_76saveCoordinateList(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_v_fOut = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("saveCoordinateList", 0);

  /* "gnocis/regions.pyx":670
 * 		:type path: str
 * 		"""
 * 		with open(path, 'w') as fOut:             # <<<<<<<<<<<<<<
 * 			for r in self.regions:
 * 				fOut.write('%s:%d..%d\n'%(r.seq, r.start, r.end))
 */
  /*with:*/ {
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_path);
    __Pyx_GIVEREF(__pyx_v_path);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_path);
    __Pyx_INCREF(__pyx_n_u_w);
    __Pyx_GIVEREF(__pyx_n_u_w);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_w);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 670, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {
          __pyx_v_fOut = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":671
 * 		"""
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:             # <<<<<<<<<<<<<<
 * 				fOut.write('%s:%d..%d\n'%(r.seq, r.start, r.end))
 * 
 */
          if (unlikely(__pyx_v_self->regions == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
            __PYX_ERR(0, 671, __pyx_L7_error)
          }
          __pyx_t_4 = __pyx_v_self->regions; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
          for (;;) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 671, __pyx_L7_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
            __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":672
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:
 * 				fOut.write('%s:%d..%d\n'%(r.seq, r.start, r.end))             # <<<<<<<<<<<<<<
 * 
 * 	# Prints basic statistics about the set.
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_fOut, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = PyTuple_New(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_10 = 0;
            __pyx_t_11 = 127;
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_12), __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_13);
            __pyx_t_13 = 0;
            __Pyx_INCREF(__pyx_kp_u__3);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__3);
            PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u__3);
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_start); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_12 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_13), __pyx_n_u_d); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_12);
            PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_12);
            __pyx_t_12 = 0;
            __Pyx_INCREF(__pyx_kp_u__4);
            __pyx_t_10 += 2;
            __Pyx_GIVEREF(__pyx_kp_u__4);
            PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u__4);
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_r, __pyx_n_s_end); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_12), __pyx_n_u_d); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
            __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_13);
            __pyx_t_13 = 0;
            __Pyx_INCREF(__pyx_kp_u__21);
            __pyx_t_10 += 1;
            __Pyx_GIVEREF(__pyx_kp_u__21);
            PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_kp_u__21);
            __pyx_t_13 = __Pyx_PyUnicode_Join(__pyx_t_5, 6, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_13);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 672, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":671
 * 		"""
 * 		with open(path, 'w') as fOut:
 * 			for r in self.regions:             # <<<<<<<<<<<<<<
 * 				fOut.write('%s:%d..%d\n'%(r.seq, r.start, r.end))
 * 
 */
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":670
 * 		:type path: str
 * 		"""
 * 		with open(path, 'w') as fOut:             # <<<<<<<<<<<<<<
 * 			for r in self.regions:
 * 				fOut.write('%s:%d..%d\n'%(r.seq, r.start, r.end))
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.regions.saveCoordinateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 670, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_13 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 670, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_13, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 670, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (__pyx_t_15 < 0) __PYX_ERR(0, 670, __pyx_L9_except_error)
          __pyx_t_16 = ((!(__pyx_t_15 != 0)) != 0);
          if (__pyx_t_16) {
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_XGIVEREF(__pyx_t_1);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_2, __pyx_t_1);
            __pyx_t_4 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
            __PYX_ERR(0, 670, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 670, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L18;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L18:;
  }

  /* "gnocis/regions.pyx":664
 * 
 * 	# Saves regions to coordinate list file.
 * 	def saveCoordinateList(self, path):             # <<<<<<<<<<<<<<
 * 		""" Saves the regions to a coordinate list file (a coordinate per line, with the format `seq:start..end`).
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("gnocis.regions.regions.saveCoordinateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fOut);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":675
 * 
 * 	# Prints basic statistics about the set.
 * 	def printStatistics(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		Outputs basic statistics.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_79printStatistics(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6gnocis_7regions_7regions_78printStatistics[] = "\n\t\tOutputs basic statistics.\n\t\t";
static PyObject *__pyx_pw_6gnocis_7regions_7regions_79printStatistics(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("printStatistics (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_78printStatistics(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":681
 * 		print('Region set statistics - ' + self.name)
 * 		print(' - Regions: %d'%len(self.regions))
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))             # <<<<<<<<<<<<<<
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_8_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_8_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 681, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_15printStatistics_2generator4, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_printStatistics_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.printStatistics.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 681, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 681, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 681, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 681, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_r, __pyx_n_s_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_5generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":683
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))             # <<<<<<<<<<<<<<
 * 			print(' - Total length: %d nt'%( sum( len(r) for r in self.regions ) ))
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_9_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_9_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 683, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_15printStatistics_5generator5, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_printStatistics_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.printStatistics.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_5generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 683, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 683, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 683, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 683, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_8generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_6genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_10_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_10_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 683, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_15printStatistics_8generator6, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_printStatistics_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.printStatistics.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_8generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 683, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 683, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 683, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 683, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_11generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_9genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_11_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_11_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 683, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_15printStatistics_11generator7, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_printStatistics_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.printStatistics.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_11generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 683, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 683, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 683, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 683, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_14generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":684
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))
 * 			print(' - Total length: %d nt'%( sum( len(r) for r in self.regions ) ))             # <<<<<<<<<<<<<<
 * 
 * # Loads regions from a BED file.
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_15printStatistics_12genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_12_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_12_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 684, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_7regions_15printStatistics_14generator8, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_printStatistics_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.regions.printStatistics.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_14generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 684, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 684, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 684, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->regions; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 684, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 684, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 684, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":675
 * 
 * 	# Prints basic statistics about the set.
 * 	def printStatistics(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		Outputs basic statistics.
 */

static PyObject *__pyx_pf_6gnocis_7regions_7regions_78printStatistics(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *__pyx_cur_scope;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_2generator4 = 0;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_5generator5 = 0;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_8generator6 = 0;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_11generator7 = 0;
  PyObject *__pyx_gb_6gnocis_7regions_7regions_15printStatistics_14generator8 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("printStatistics", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_7_printStatistics(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_7_printStatistics, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 675, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "gnocis/regions.pyx":679
 * 		Outputs basic statistics.
 * 		"""
 * 		print('Region set statistics - ' + self.name)             # <<<<<<<<<<<<<<
 * 		print(' - Regions: %d'%len(self.regions))
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))
 */
  __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_kp_u_Region_set_statistics, __pyx_cur_scope->__pyx_v_self->name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":680
 * 		"""
 * 		print('Region set statistics - ' + self.name)
 * 		print(' - Regions: %d'%len(self.regions))             # <<<<<<<<<<<<<<
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))
 * 		if len(self.regions) > 0:
 */
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->regions;
  __Pyx_INCREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 680, __pyx_L1_error)
  }
  __pyx_t_3 = PyList_GET_SIZE(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyUnicode_Format(__pyx_kp_u_Regions_d, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":681
 * 		print('Region set statistics - ' + self.name)
 * 		print(' - Regions: %d'%len(self.regions))
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))             # <<<<<<<<<<<<<<
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))
 */
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_7regions_15printStatistics_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_5 = PyList_Sort(__pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 681, __pyx_L1_error)
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__12, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_Region_sequences_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":682
 * 		print(' - Regions: %d'%len(self.regions))
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))
 * 		if len(self.regions) > 0:             # <<<<<<<<<<<<<<
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))
 * 			print(' - Total length: %d nt'%( sum( len(r) for r in self.regions ) ))
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->regions;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 682, __pyx_L1_error)
  }
  __pyx_t_3 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((__pyx_t_3 > 0) != 0);
  if (__pyx_t_6) {

    /* "gnocis/regions.pyx":683
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))             # <<<<<<<<<<<<<<
 * 			print(' - Total length: %d nt'%( sum( len(r) for r in self.regions ) ))
 * 
 */
    __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_7 = 127;
    __Pyx_INCREF(__pyx_kp_u_Mean_length);
    __pyx_t_3 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_Mean_length);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Mean_length);
    __pyx_t_2 = __pyx_pf_6gnocis_7regions_7regions_15printStatistics_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __pyx_cur_scope->__pyx_v_self->regions;
    __Pyx_INCREF(__pyx_t_2);
    if (unlikely(__pyx_t_2 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 683, __pyx_L1_error)
    }
    __pyx_t_8 = PyList_GET_SIZE(__pyx_t_2); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Format(__pyx_t_9, __pyx_kp_u_2f); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_nt_min);
    __pyx_t_3 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_nt_min);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_nt_min);
    __pyx_t_2 = __pyx_pf_6gnocis_7regions_7regions_15printStatistics_6genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_9), __pyx_n_u_d); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_nt_max);
    __pyx_t_3 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_nt_max);
    PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_nt_max);
    __pyx_t_2 = __pyx_pf_6gnocis_7regions_7regions_15printStatistics_9genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_9), __pyx_n_u_d); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_7;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_nt);
    __pyx_t_3 += 4;
    __Pyx_GIVEREF(__pyx_kp_u_nt);
    PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u_nt);
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/regions.pyx":684
 * 		if len(self.regions) > 0:
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))
 * 			print(' - Total length: %d nt'%( sum( len(r) for r in self.regions ) ))             # <<<<<<<<<<<<<<
 * 
 * # Loads regions from a BED file.
 */
    __pyx_t_1 = __pyx_pf_6gnocis_7regions_7regions_15printStatistics_12genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Total_length_d_nt, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 684, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/regions.pyx":682
 * 		print(' - Regions: %d'%len(self.regions))
 * 		print(' - Region sequences: %s'%', '.join(sorted(set( r.seq for r in self.regions ))))
 * 		if len(self.regions) > 0:             # <<<<<<<<<<<<<<
 * 			print(' - Mean length: %.2f nt (min.: %d nt - max.: %d nt)'%( sum( len(r) for r in self.regions ) / len(self.regions), min( len(r) for r in self.regions), max( len(r) for r in self.regions) ))
 * 			print(' - Total length: %d nt'%( sum( len(r) for r in self.regions ) ))
 */
  }

  /* "gnocis/regions.pyx":675
 * 
 * 	# Prints basic statistics about the set.
 * 	def printStatistics(self):             # <<<<<<<<<<<<<<
 * 		"""
 * 		Outputs basic statistics.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("gnocis.regions.regions.printStatistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_15printStatistics_2generator4);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_15printStatistics_5generator5);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_15printStatistics_8generator6);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_15printStatistics_11generator7);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_7regions_15printStatistics_14generator8);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":29
 * cdef class regions:
 * 
 * 	cdef public str name             # <<<<<<<<<<<<<<
 * 	cdef public list regions
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_4name___get__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_4name___get__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_7regions_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_7regions_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_4name_2__set__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_7regions_4name_2__set__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 29, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.regions.name.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_7regions_4name_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gnocis_7regions_7regions_4name_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_4name_4__del__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_7regions_4name_4__del__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pxd":30
 * 
 * 	cdef public str name
 * 	cdef public list regions             # <<<<<<<<<<<<<<
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_7regions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_7regions_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_7regions___get__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_7regions___get__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->regions);
  __pyx_r = __pyx_v_self->regions;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_7regions_7regions_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gnocis_7regions_7regions_7regions_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_7regions_2__set__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_7regions_7regions_2__set__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 30, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->regions);
  __Pyx_DECREF(__pyx_v_self->regions);
  __pyx_v_self->regions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.regions.regions.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gnocis_7regions_7regions_7regions_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gnocis_7regions_7regions_7regions_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_7regions_4__del__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gnocis_7regions_7regions_7regions_4__del__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->regions);
  __Pyx_DECREF(__pyx_v_self->regions);
  __pyx_v_self->regions = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_81__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_81__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_80__reduce_cython__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_80__reduce_cython__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name, self.regions)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __Pyx_INCREF(__pyx_v_self->regions);
  __Pyx_GIVEREF(__pyx_v_self->regions);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->regions);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name, self.regions)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name, self.regions)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None or self.regions is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name, self.regions)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None or self.regions is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->name != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->regions != ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_5 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None or self.regions is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None or self.regions is not None
 *     if use_setstate:
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_regions); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_165254229);
    __Pyx_GIVEREF(__pyx_int_165254229);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_165254229);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None or self.regions is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, None), state
 *     else:
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_regions__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_regions); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_165254229);
    __Pyx_GIVEREF(__pyx_int_165254229);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_165254229);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.regions.regions.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_regions__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7regions_83__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6gnocis_7regions_7regions_83__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_7regions_82__setstate_cython__(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_7regions_82__setstate_cython__(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_regions__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6gnocis_7regions___pyx_unpickle_regions__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_regions, (type(self), 0x9d99455, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_regions__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.regions.regions.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":687
 * 
 * # Loads regions from a BED file.
 * def loadBED(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_1loadBED(PyObject *__pyx_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_6gnocis_7regions_loadBED[] = " Loads regions from a BED (https://www.ensembl.org/info/website/upload/bed.html) file.\n\t\n\t:param path: Path to the input file.\n\t:type path: str\n\t\n\t:return: Loaded regions.\n\t:rtype: regions\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_1loadBED = {"loadBED", (PyCFunction)__pyx_pw_6gnocis_7regions_1loadBED, METH_O, __pyx_doc_6gnocis_7regions_loadBED};
static PyObject *__pyx_pw_6gnocis_7regions_1loadBED(PyObject *__pyx_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("loadBED (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_loadBED(__pyx_self, ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_loadBED(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_v_rgns = 0;
  PyObject *__pyx_v_l = 0;
  PyObject *__pyx_v_seq = 0;
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_end = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_feature = 0;
  PyObject *__pyx_v_group = 0;
  bool __pyx_v_strand;
  float __pyx_v_score;
  PyObject *__pyx_v_parts = 0;
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_rs = 0;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  double __pyx_t_14;
  bool __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("loadBED", 0);

  /* "gnocis/regions.pyx":702
 * 	cdef list parts
 * 	cdef regions rs
 * 	with open(path, 'r') as f:             # <<<<<<<<<<<<<<
 * 		rgns = []
 * 		for l in f:
 */
  /*with:*/ {
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_path);
    __Pyx_GIVEREF(__pyx_v_path);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_path);
    __Pyx_INCREF(__pyx_n_u_r);
    __Pyx_GIVEREF(__pyx_n_u_r);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_r);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 702, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {
          __pyx_v_f = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":703
 * 	cdef regions rs
 * 	with open(path, 'r') as f:
 * 		rgns = []             # <<<<<<<<<<<<<<
 * 		for l in f:
 * 			if l.startswith('#'): continue
 */
          __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 703, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_v_rgns = ((PyObject*)__pyx_t_4);
          __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":704
 * 	with open(path, 'r') as f:
 * 		rgns = []
 * 		for l in f:             # <<<<<<<<<<<<<<
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 */
          if (likely(PyList_CheckExact(__pyx_v_f)) || PyTuple_CheckExact(__pyx_v_f)) {
            __pyx_t_4 = __pyx_v_f; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
            __pyx_t_10 = NULL;
          } else {
            __pyx_t_9 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 704, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_10 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 704, __pyx_L7_error)
          }
          for (;;) {
            if (likely(!__pyx_t_10)) {
              if (likely(PyList_CheckExact(__pyx_t_4))) {
                if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 704, __pyx_L7_error)
                #else
                __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 704, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_2);
                #endif
              } else {
                if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 704, __pyx_L7_error)
                #else
                __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 704, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_2);
                #endif
              }
            } else {
              __pyx_t_2 = __pyx_t_10(__pyx_t_4);
              if (unlikely(!__pyx_t_2)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 704, __pyx_L7_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_2);
            }
            if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 704, __pyx_L7_error)
            __Pyx_XDECREF_SET(__pyx_v_l, ((PyObject*)__pyx_t_2));
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":705
 * 		rgns = []
 * 		for l in f:
 * 			if l.startswith('#'): continue             # <<<<<<<<<<<<<<
 * 			parts = l.strip().split('\t')
 * 			if len(parts) < 3: raise Exception('BED file syntax error')
 */
            if (unlikely(__pyx_v_l == Py_None)) {
              PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "startswith");
              __PYX_ERR(0, 705, __pyx_L7_error)
            }
            __pyx_t_11 = __Pyx_PyUnicode_Tailmatch(__pyx_v_l, __pyx_kp_u__22, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 705, __pyx_L7_error)
            if ((__pyx_t_11 != 0)) {
              goto __pyx_L13_continue;
            }

            /* "gnocis/regions.pyx":706
 * 		for l in f:
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')             # <<<<<<<<<<<<<<
 * 			if len(parts) < 3: raise Exception('BED file syntax error')
 * 			seq, start, end = parts[0:3]
 */
            __pyx_t_1 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_strip, __pyx_v_l); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_split); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 706, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_kp_u__19) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u__19);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 706, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 706, __pyx_L7_error)
            __Pyx_XDECREF_SET(__pyx_v_parts, ((PyObject*)__pyx_t_2));
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":707
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 * 			if len(parts) < 3: raise Exception('BED file syntax error')             # <<<<<<<<<<<<<<
 * 			seq, start, end = parts[0:3]
 * 			strand = True
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 707, __pyx_L7_error)
            }
            __pyx_t_12 = PyList_GET_SIZE(__pyx_v_parts); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 707, __pyx_L7_error)
            __pyx_t_11 = ((__pyx_t_12 < 3) != 0);
            if (unlikely(__pyx_t_11)) {
              __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_Raise(__pyx_t_2, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __PYX_ERR(0, 707, __pyx_L7_error)
            }

            /* "gnocis/regions.pyx":708
 * 			parts = l.strip().split('\t')
 * 			if len(parts) < 3: raise Exception('BED file syntax error')
 * 			seq, start, end = parts[0:3]             # <<<<<<<<<<<<<<
 * 			strand = True
 * 			score = 0.0
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 708, __pyx_L7_error)
            }
            __pyx_t_2 = __Pyx_PyList_GetSlice(__pyx_v_parts, 0, 3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 708, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (1) {
              PyObject* sequence = __pyx_t_2;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 3)) {
                if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 708, __pyx_L7_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
              __pyx_t_13 = PyList_GET_ITEM(sequence, 2); 
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_13);
              #else
              __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 708, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 708, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_13 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 708, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_13);
              #endif
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
            if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 708, __pyx_L7_error)
            if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 708, __pyx_L7_error)
            if (!(likely(PyUnicode_CheckExact(__pyx_t_13))||((__pyx_t_13) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_13)->tp_name), 0))) __PYX_ERR(0, 708, __pyx_L7_error)
            __Pyx_XDECREF_SET(__pyx_v_seq, ((PyObject*)__pyx_t_5));
            __pyx_t_5 = 0;
            __Pyx_XDECREF_SET(__pyx_v_start, ((PyObject*)__pyx_t_1));
            __pyx_t_1 = 0;
            __Pyx_XDECREF_SET(__pyx_v_end, ((PyObject*)__pyx_t_13));
            __pyx_t_13 = 0;

            /* "gnocis/regions.pyx":709
 * 			if len(parts) < 3: raise Exception('BED file syntax error')
 * 			seq, start, end = parts[0:3]
 * 			strand = True             # <<<<<<<<<<<<<<
 * 			score = 0.0
 * 			source = '.'
 */
            __pyx_v_strand = 1;

            /* "gnocis/regions.pyx":710
 * 			seq, start, end = parts[0:3]
 * 			strand = True
 * 			score = 0.0             # <<<<<<<<<<<<<<
 * 			source = '.'
 * 			feature = '.'
 */
            __pyx_v_score = 0.0;

            /* "gnocis/regions.pyx":711
 * 			strand = True
 * 			score = 0.0
 * 			source = '.'             # <<<<<<<<<<<<<<
 * 			feature = '.'
 * 			group = '.'
 */
            __Pyx_INCREF(__pyx_kp_u_);
            __Pyx_XDECREF_SET(__pyx_v_source, __pyx_kp_u_);

            /* "gnocis/regions.pyx":712
 * 			score = 0.0
 * 			source = '.'
 * 			feature = '.'             # <<<<<<<<<<<<<<
 * 			group = '.'
 * 			if len(parts) > 3:
 */
            __Pyx_INCREF(__pyx_kp_u_);
            __Pyx_XDECREF_SET(__pyx_v_feature, __pyx_kp_u_);

            /* "gnocis/regions.pyx":713
 * 			source = '.'
 * 			feature = '.'
 * 			group = '.'             # <<<<<<<<<<<<<<
 * 			if len(parts) > 3:
 * 				feature = parts[3]
 */
            __Pyx_INCREF(__pyx_kp_u_);
            __Pyx_XDECREF_SET(__pyx_v_group, __pyx_kp_u_);

            /* "gnocis/regions.pyx":714
 * 			feature = '.'
 * 			group = '.'
 * 			if len(parts) > 3:             # <<<<<<<<<<<<<<
 * 				feature = parts[3]
 * 			if len(parts) > 4:
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 714, __pyx_L7_error)
            }
            __pyx_t_12 = PyList_GET_SIZE(__pyx_v_parts); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 714, __pyx_L7_error)
            __pyx_t_11 = ((__pyx_t_12 > 3) != 0);
            if (__pyx_t_11) {

              /* "gnocis/regions.pyx":715
 * 			group = '.'
 * 			if len(parts) > 3:
 * 				feature = parts[3]             # <<<<<<<<<<<<<<
 * 			if len(parts) > 4:
 * 				score = float(parts[4])
 */
              if (unlikely(__pyx_v_parts == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 715, __pyx_L7_error)
              }
              __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_parts, 3, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 715, __pyx_L7_error)
              __Pyx_DECREF_SET(__pyx_v_feature, ((PyObject*)__pyx_t_2));
              __pyx_t_2 = 0;

              /* "gnocis/regions.pyx":714
 * 			feature = '.'
 * 			group = '.'
 * 			if len(parts) > 3:             # <<<<<<<<<<<<<<
 * 				feature = parts[3]
 * 			if len(parts) > 4:
 */
            }

            /* "gnocis/regions.pyx":716
 * 			if len(parts) > 3:
 * 				feature = parts[3]
 * 			if len(parts) > 4:             # <<<<<<<<<<<<<<
 * 				score = float(parts[4])
 * 			if len(parts) > 5:
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 716, __pyx_L7_error)
            }
            __pyx_t_12 = PyList_GET_SIZE(__pyx_v_parts); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 716, __pyx_L7_error)
            __pyx_t_11 = ((__pyx_t_12 > 4) != 0);
            if (__pyx_t_11) {

              /* "gnocis/regions.pyx":717
 * 				feature = parts[3]
 * 			if len(parts) > 4:
 * 				score = float(parts[4])             # <<<<<<<<<<<<<<
 * 			if len(parts) > 5:
 * 				strand = parts[5] == '+'
 */
              if (unlikely(__pyx_v_parts == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 717, __pyx_L7_error)
              }
              __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_parts, 4, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_14 = __Pyx_PyObject_AsDouble(__pyx_t_2); if (unlikely(__pyx_t_14 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_v_score = __pyx_t_14;

              /* "gnocis/regions.pyx":716
 * 			if len(parts) > 3:
 * 				feature = parts[3]
 * 			if len(parts) > 4:             # <<<<<<<<<<<<<<
 * 				score = float(parts[4])
 * 			if len(parts) > 5:
 */
            }

            /* "gnocis/regions.pyx":718
 * 			if len(parts) > 4:
 * 				score = float(parts[4])
 * 			if len(parts) > 5:             # <<<<<<<<<<<<<<
 * 				strand = parts[5] == '+'
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 718, __pyx_L7_error)
            }
            __pyx_t_12 = PyList_GET_SIZE(__pyx_v_parts); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 718, __pyx_L7_error)
            __pyx_t_11 = ((__pyx_t_12 > 5) != 0);
            if (__pyx_t_11) {

              /* "gnocis/regions.pyx":719
 * 				score = float(parts[4])
 * 			if len(parts) > 5:
 * 				strand = parts[5] == '+'             # <<<<<<<<<<<<<<
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 * 	rs = regions('BED file: ' + path, rgns)
 */
              if (unlikely(__pyx_v_parts == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 719, __pyx_L7_error)
              }
              __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_parts, 5, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_13 = PyObject_RichCompare(__pyx_t_2, __pyx_kp_u__6, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 719, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely((__pyx_t_15 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 719, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __pyx_v_strand = __pyx_t_15;

              /* "gnocis/regions.pyx":718
 * 			if len(parts) > 4:
 * 				score = float(parts[4])
 * 			if len(parts) > 5:             # <<<<<<<<<<<<<<
 * 				strand = parts[5] == '+'
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 */
            }

            /* "gnocis/regions.pyx":720
 * 			if len(parts) > 5:
 * 				strand = parts[5] == '+'
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )             # <<<<<<<<<<<<<<
 * 	rs = regions('BED file: ' + path, rgns)
 * 	#rs.sort()
 */
            __pyx_t_13 = __Pyx_PyNumber_Float(__pyx_v_start); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = __Pyx_PyNumber_Float(__pyx_v_end); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = __Pyx_PyBool_FromLong(__pyx_v_strand); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_5 = PyFloat_FromDouble(__pyx_v_score); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_16 = PyTuple_New(7); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_INCREF(__pyx_v_seq);
            __Pyx_GIVEREF(__pyx_v_seq);
            PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_seq);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_16, 2, __pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_16, 3, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_16, 4, __pyx_t_5);
            __Pyx_INCREF(__pyx_v_source);
            __Pyx_GIVEREF(__pyx_v_source);
            PyTuple_SET_ITEM(__pyx_t_16, 5, __pyx_v_source);
            __Pyx_INCREF(__pyx_v_feature);
            __Pyx_GIVEREF(__pyx_v_feature);
            PyTuple_SET_ITEM(__pyx_t_16, 6, __pyx_v_feature);
            __pyx_t_2 = 0;
            __pyx_t_1 = 0;
            __pyx_t_13 = 0;
            __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_group, __pyx_v_group) < 0) __PYX_ERR(0, 720, __pyx_L7_error)
            __pyx_t_13 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_16, __pyx_t_5); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_rgns, __pyx_t_13); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 720, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

            /* "gnocis/regions.pyx":704
 * 	with open(path, 'r') as f:
 * 		rgns = []
 * 		for l in f:             # <<<<<<<<<<<<<<
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 */
            __pyx_L13_continue:;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":702
 * 	cdef list parts
 * 	cdef regions rs
 * 	with open(path, 'r') as f:             # <<<<<<<<<<<<<<
 * 		rgns = []
 * 		for l in f:
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.loadBED", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_13, &__pyx_t_5) < 0) __PYX_ERR(0, 702, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_16 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_13, __pyx_t_5); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 702, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_16, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 702, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (__pyx_t_11 < 0) __PYX_ERR(0, 702, __pyx_L9_except_error)
          __pyx_t_19 = ((!(__pyx_t_11 != 0)) != 0);
          if (__pyx_t_19) {
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_5);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_13, __pyx_t_5);
            __pyx_t_4 = 0; __pyx_t_13 = 0; __pyx_t_5 = 0; 
            __PYX_ERR(0, 702, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 702, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L23;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L23:;
  }

  /* "gnocis/regions.pyx":721
 * 				strand = parts[5] == '+'
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 * 	rs = regions('BED file: ' + path, rgns)             # <<<<<<<<<<<<<<
 * 	#rs.sort()
 * 	return rs
 */
  __pyx_t_5 = PyNumber_Add(__pyx_kp_u_BED_file, __pyx_v_path); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(!__pyx_v_rgns)) { __Pyx_RaiseUnboundLocalError("rgns"); __PYX_ERR(0, 721, __pyx_L1_error) }
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_rgns);
  __Pyx_GIVEREF(__pyx_v_rgns);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_v_rgns);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_v_rs = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":723
 * 	rs = regions('BED file: ' + path, rgns)
 * 	#rs.sort()
 * 	return rs             # <<<<<<<<<<<<<<
 * 
 * # Loads regions from a gzipped BED file.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_rs));
  __pyx_r = ((PyObject *)__pyx_v_rs);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":687
 * 
 * # Loads regions from a BED file.
 * def loadBED(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("gnocis.regions.loadBED", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rgns);
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_seq);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_end);
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XDECREF(__pyx_v_feature);
  __Pyx_XDECREF(__pyx_v_group);
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XDECREF((PyObject *)__pyx_v_rs);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":726
 * 
 * # Loads regions from a gzipped BED file.
 * def loadBEDGZ(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_3loadBEDGZ(PyObject *__pyx_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_6gnocis_7regions_2loadBEDGZ[] = " Loads regions from a gzipped BED (https://www.ensembl.org/info/website/upload/bed.html) file.\n\t\n\t:param path: Path to the input file.\n\t:type path: str\n\t\n\t:return: Loaded regions.\n\t:rtype: regions\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_3loadBEDGZ = {"loadBEDGZ", (PyCFunction)__pyx_pw_6gnocis_7regions_3loadBEDGZ, METH_O, __pyx_doc_6gnocis_7regions_2loadBEDGZ};
static PyObject *__pyx_pw_6gnocis_7regions_3loadBEDGZ(PyObject *__pyx_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("loadBEDGZ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_7regions_2loadBEDGZ(__pyx_self, ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_9loadBEDGZ_2generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":744
 * 	with gzip.open(path, 'rb') as f:
 * 		rgns = []
 * 		for l in (y.decode('utf-8') for y in f):             # <<<<<<<<<<<<<<
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 */

static PyObject *__pyx_pf_6gnocis_7regions_9loadBEDGZ_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_14_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_14_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 744, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_9loadBEDGZ_2generator9, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_loadBEDGZ_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.loadBEDGZ.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_9loadBEDGZ_2generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 744, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f)) { __Pyx_RaiseClosureNameError("f"); __PYX_ERR(0, 744, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_f; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 744, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 744, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 744, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 744, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_y);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_y, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_y, __pyx_n_s_decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_utf_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 744, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":726
 * 
 * # Loads regions from a gzipped BED file.
 * def loadBEDGZ(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_2loadBEDGZ(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *__pyx_cur_scope;
  PyObject *__pyx_v_rgns = 0;
  PyObject *__pyx_v_l = 0;
  PyObject *__pyx_v_seq = 0;
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_end = 0;
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_feature = 0;
  PyObject *__pyx_v_group = 0;
  bool __pyx_v_strand;
  float __pyx_v_score;
  PyObject *__pyx_v_parts = 0;
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_rs = 0;
  PyObject *__pyx_gb_6gnocis_7regions_9loadBEDGZ_2generator9 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  double __pyx_t_15;
  bool __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("loadBEDGZ", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 726, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/regions.pyx":742
 * 	cdef regions rs
 * 	#with open(path, 'r') as f:
 * 	with gzip.open(path, 'rb') as f:             # <<<<<<<<<<<<<<
 * 		rgns = []
 * 		for l in (y.decode('utf-8') for y in f):
 */
  /*with:*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_gzip); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_open); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_path, __pyx_n_u_rb};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_path, __pyx_n_u_rb};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_path);
      __Pyx_GIVEREF(__pyx_v_path);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_path);
      __Pyx_INCREF(__pyx_n_u_rb);
      __Pyx_GIVEREF(__pyx_n_u_rb);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_n_u_rb);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 742, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {
          __Pyx_GIVEREF(__pyx_t_5);
          __pyx_cur_scope->__pyx_v_f = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":743
 * 	#with open(path, 'r') as f:
 * 	with gzip.open(path, 'rb') as f:
 * 		rgns = []             # <<<<<<<<<<<<<<
 * 		for l in (y.decode('utf-8') for y in f):
 * 			if l.startswith('#'): continue
 */
          __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 743, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_v_rgns = ((PyObject*)__pyx_t_5);
          __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":744
 * 	with gzip.open(path, 'rb') as f:
 * 		rgns = []
 * 		for l in (y.decode('utf-8') for y in f):             # <<<<<<<<<<<<<<
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 */
          __pyx_t_5 = __pyx_pf_6gnocis_7regions_9loadBEDGZ_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 744, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
            __pyx_t_1 = __pyx_t_5; __Pyx_INCREF(__pyx_t_1); __pyx_t_10 = 0;
            __pyx_t_11 = NULL;
          } else {
            __pyx_t_10 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 744, __pyx_L7_error)
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          for (;;) {
            if (likely(!__pyx_t_11)) {
              if (likely(PyList_CheckExact(__pyx_t_1))) {
                if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_1)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 744, __pyx_L7_error)
                #else
                __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 744, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_5);
                #endif
              } else {
                if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 744, __pyx_L7_error)
                #else
                __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 744, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_5);
                #endif
              }
            } else {
              __pyx_t_5 = __pyx_t_11(__pyx_t_1);
              if (unlikely(!__pyx_t_5)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 744, __pyx_L7_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_5);
            }
            if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 744, __pyx_L7_error)
            __Pyx_XDECREF_SET(__pyx_v_l, ((PyObject*)__pyx_t_5));
            __pyx_t_5 = 0;

            /* "gnocis/regions.pyx":745
 * 		rgns = []
 * 		for l in (y.decode('utf-8') for y in f):
 * 			if l.startswith('#'): continue             # <<<<<<<<<<<<<<
 * 			parts = l.strip().split('\t')
 * 			if len(parts) < 3: raise Exception('BED file syntax error')
 */
            if (unlikely(__pyx_v_l == Py_None)) {
              PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "startswith");
              __PYX_ERR(0, 745, __pyx_L7_error)
            }
            __pyx_t_12 = __Pyx_PyUnicode_Tailmatch(__pyx_v_l, __pyx_kp_u__22, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 745, __pyx_L7_error)
            if ((__pyx_t_12 != 0)) {
              goto __pyx_L13_continue;
            }

            /* "gnocis/regions.pyx":746
 * 		for l in (y.decode('utf-8') for y in f):
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')             # <<<<<<<<<<<<<<
 * 			if len(parts) < 3: raise Exception('BED file syntax error')
 * 			seq, start, end = parts[0:3]
 */
            __pyx_t_3 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_strip, __pyx_v_l); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 746, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u__19) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u__19);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 746, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (!(likely(PyList_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 746, __pyx_L7_error)
            __Pyx_XDECREF_SET(__pyx_v_parts, ((PyObject*)__pyx_t_5));
            __pyx_t_5 = 0;

            /* "gnocis/regions.pyx":747
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 * 			if len(parts) < 3: raise Exception('BED file syntax error')             # <<<<<<<<<<<<<<
 * 			seq, start, end = parts[0:3]
 * 			strand = True
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 747, __pyx_L7_error)
            }
            __pyx_t_13 = PyList_GET_SIZE(__pyx_v_parts); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 747, __pyx_L7_error)
            __pyx_t_12 = ((__pyx_t_13 < 3) != 0);
            if (unlikely(__pyx_t_12)) {
              __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 747, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_Raise(__pyx_t_5, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __PYX_ERR(0, 747, __pyx_L7_error)
            }

            /* "gnocis/regions.pyx":748
 * 			parts = l.strip().split('\t')
 * 			if len(parts) < 3: raise Exception('BED file syntax error')
 * 			seq, start, end = parts[0:3]             # <<<<<<<<<<<<<<
 * 			strand = True
 * 			score = 0.0
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 748, __pyx_L7_error)
            }
            __pyx_t_5 = __Pyx_PyList_GetSlice(__pyx_v_parts, 0, 3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 748, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (1) {
              PyObject* sequence = __pyx_t_5;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 3)) {
                if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 748, __pyx_L7_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
              __pyx_t_14 = PyList_GET_ITEM(sequence, 2); 
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_14);
              #else
              __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 748, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_14 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 748, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_14);
              #endif
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
            if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 748, __pyx_L7_error)
            if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 748, __pyx_L7_error)
            if (!(likely(PyUnicode_CheckExact(__pyx_t_14))||((__pyx_t_14) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_14)->tp_name), 0))) __PYX_ERR(0, 748, __pyx_L7_error)
            __Pyx_XDECREF_SET(__pyx_v_seq, ((PyObject*)__pyx_t_2));
            __pyx_t_2 = 0;
            __Pyx_XDECREF_SET(__pyx_v_start, ((PyObject*)__pyx_t_3));
            __pyx_t_3 = 0;
            __Pyx_XDECREF_SET(__pyx_v_end, ((PyObject*)__pyx_t_14));
            __pyx_t_14 = 0;

            /* "gnocis/regions.pyx":749
 * 			if len(parts) < 3: raise Exception('BED file syntax error')
 * 			seq, start, end = parts[0:3]
 * 			strand = True             # <<<<<<<<<<<<<<
 * 			score = 0.0
 * 			source = '.'
 */
            __pyx_v_strand = 1;

            /* "gnocis/regions.pyx":750
 * 			seq, start, end = parts[0:3]
 * 			strand = True
 * 			score = 0.0             # <<<<<<<<<<<<<<
 * 			source = '.'
 * 			feature = '.'
 */
            __pyx_v_score = 0.0;

            /* "gnocis/regions.pyx":751
 * 			strand = True
 * 			score = 0.0
 * 			source = '.'             # <<<<<<<<<<<<<<
 * 			feature = '.'
 * 			group = '.'
 */
            __Pyx_INCREF(__pyx_kp_u_);
            __Pyx_XDECREF_SET(__pyx_v_source, __pyx_kp_u_);

            /* "gnocis/regions.pyx":752
 * 			score = 0.0
 * 			source = '.'
 * 			feature = '.'             # <<<<<<<<<<<<<<
 * 			group = '.'
 * 			if len(parts) > 3:
 */
            __Pyx_INCREF(__pyx_kp_u_);
            __Pyx_XDECREF_SET(__pyx_v_feature, __pyx_kp_u_);

            /* "gnocis/regions.pyx":753
 * 			source = '.'
 * 			feature = '.'
 * 			group = '.'             # <<<<<<<<<<<<<<
 * 			if len(parts) > 3:
 * 				feature = parts[3]
 */
            __Pyx_INCREF(__pyx_kp_u_);
            __Pyx_XDECREF_SET(__pyx_v_group, __pyx_kp_u_);

            /* "gnocis/regions.pyx":754
 * 			feature = '.'
 * 			group = '.'
 * 			if len(parts) > 3:             # <<<<<<<<<<<<<<
 * 				feature = parts[3]
 * 			if len(parts) > 4:
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 754, __pyx_L7_error)
            }
            __pyx_t_13 = PyList_GET_SIZE(__pyx_v_parts); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 754, __pyx_L7_error)
            __pyx_t_12 = ((__pyx_t_13 > 3) != 0);
            if (__pyx_t_12) {

              /* "gnocis/regions.pyx":755
 * 			group = '.'
 * 			if len(parts) > 3:
 * 				feature = parts[3]             # <<<<<<<<<<<<<<
 * 			if len(parts) > 4:
 * 				score = float(parts[4])
 */
              if (unlikely(__pyx_v_parts == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 755, __pyx_L7_error)
              }
              __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_parts, 3, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 755, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 755, __pyx_L7_error)
              __Pyx_DECREF_SET(__pyx_v_feature, ((PyObject*)__pyx_t_5));
              __pyx_t_5 = 0;

              /* "gnocis/regions.pyx":754
 * 			feature = '.'
 * 			group = '.'
 * 			if len(parts) > 3:             # <<<<<<<<<<<<<<
 * 				feature = parts[3]
 * 			if len(parts) > 4:
 */
            }

            /* "gnocis/regions.pyx":756
 * 			if len(parts) > 3:
 * 				feature = parts[3]
 * 			if len(parts) > 4:             # <<<<<<<<<<<<<<
 * 				score = float(parts[4])
 * 			if len(parts) > 5:
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 756, __pyx_L7_error)
            }
            __pyx_t_13 = PyList_GET_SIZE(__pyx_v_parts); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 756, __pyx_L7_error)
            __pyx_t_12 = ((__pyx_t_13 > 4) != 0);
            if (__pyx_t_12) {

              /* "gnocis/regions.pyx":757
 * 				feature = parts[3]
 * 			if len(parts) > 4:
 * 				score = float(parts[4])             # <<<<<<<<<<<<<<
 * 			if len(parts) > 5:
 * 				strand = parts[5] == '+'
 */
              if (unlikely(__pyx_v_parts == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 757, __pyx_L7_error)
              }
              __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_parts, 4, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 757, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_15 = __Pyx_PyObject_AsDouble(__pyx_t_5); if (unlikely(__pyx_t_15 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 757, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_v_score = __pyx_t_15;

              /* "gnocis/regions.pyx":756
 * 			if len(parts) > 3:
 * 				feature = parts[3]
 * 			if len(parts) > 4:             # <<<<<<<<<<<<<<
 * 				score = float(parts[4])
 * 			if len(parts) > 5:
 */
            }

            /* "gnocis/regions.pyx":758
 * 			if len(parts) > 4:
 * 				score = float(parts[4])
 * 			if len(parts) > 5:             # <<<<<<<<<<<<<<
 * 				strand = parts[5] == '+'
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 */
            if (unlikely(__pyx_v_parts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 758, __pyx_L7_error)
            }
            __pyx_t_13 = PyList_GET_SIZE(__pyx_v_parts); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 758, __pyx_L7_error)
            __pyx_t_12 = ((__pyx_t_13 > 5) != 0);
            if (__pyx_t_12) {

              /* "gnocis/regions.pyx":759
 * 				score = float(parts[4])
 * 			if len(parts) > 5:
 * 				strand = parts[5] == '+'             # <<<<<<<<<<<<<<
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 * 	rs = regions('BED file: ' + path, rgns)
 */
              if (unlikely(__pyx_v_parts == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 759, __pyx_L7_error)
              }
              __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_parts, 5, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 759, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_14 = PyObject_RichCompare(__pyx_t_5, __pyx_kp_u__6, Py_EQ); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 759, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely((__pyx_t_16 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 759, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_v_strand = __pyx_t_16;

              /* "gnocis/regions.pyx":758
 * 			if len(parts) > 4:
 * 				score = float(parts[4])
 * 			if len(parts) > 5:             # <<<<<<<<<<<<<<
 * 				strand = parts[5] == '+'
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 */
            }

            /* "gnocis/regions.pyx":760
 * 			if len(parts) > 5:
 * 				strand = parts[5] == '+'
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )             # <<<<<<<<<<<<<<
 * 	rs = regions('BED file: ' + path, rgns)
 * 	return rs
 */
            __pyx_t_14 = __Pyx_PyNumber_Float(__pyx_v_start); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_14); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_14 = __Pyx_PyNumber_Float(__pyx_v_end); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_14); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_14 = __Pyx_PyBool_FromLong(__pyx_v_strand); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_2 = PyFloat_FromDouble(__pyx_v_score); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_17 = PyTuple_New(7); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_INCREF(__pyx_v_seq);
            __Pyx_GIVEREF(__pyx_v_seq);
            PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_v_seq);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_17, 2, __pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_14);
            PyTuple_SET_ITEM(__pyx_t_17, 3, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_17, 4, __pyx_t_2);
            __Pyx_INCREF(__pyx_v_source);
            __Pyx_GIVEREF(__pyx_v_source);
            PyTuple_SET_ITEM(__pyx_t_17, 5, __pyx_v_source);
            __Pyx_INCREF(__pyx_v_feature);
            __Pyx_GIVEREF(__pyx_v_feature);
            PyTuple_SET_ITEM(__pyx_t_17, 6, __pyx_v_feature);
            __pyx_t_5 = 0;
            __pyx_t_3 = 0;
            __pyx_t_14 = 0;
            __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_group, __pyx_v_group) < 0) __PYX_ERR(0, 760, __pyx_L7_error)
            __pyx_t_14 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_17, __pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_rgns, __pyx_t_14); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 760, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

            /* "gnocis/regions.pyx":744
 * 	with gzip.open(path, 'rb') as f:
 * 		rgns = []
 * 		for l in (y.decode('utf-8') for y in f):             # <<<<<<<<<<<<<<
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 */
            __pyx_L13_continue:;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "gnocis/regions.pyx":742
 * 	cdef regions rs
 * 	#with open(path, 'r') as f:
 * 	with gzip.open(path, 'rb') as f:             # <<<<<<<<<<<<<<
 * 		rgns = []
 * 		for l in (y.decode('utf-8') for y in f):
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.loadBEDGZ", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_14, &__pyx_t_2) < 0) __PYX_ERR(0, 742, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_17 = PyTuple_Pack(3, __pyx_t_1, __pyx_t_14, __pyx_t_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 742, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_17, NULL);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 742, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_19);
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (__pyx_t_12 < 0) __PYX_ERR(0, 742, __pyx_L9_except_error)
          __pyx_t_20 = ((!(__pyx_t_12 != 0)) != 0);
          if (__pyx_t_20) {
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_2);
            __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_14, __pyx_t_2);
            __pyx_t_1 = 0; __pyx_t_14 = 0; __pyx_t_2 = 0; 
            __PYX_ERR(0, 742, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_6) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 742, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L23;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L1_error;
    __pyx_L23:;
  }

  /* "gnocis/regions.pyx":761
 * 				strand = parts[5] == '+'
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 * 	rs = regions('BED file: ' + path, rgns)             # <<<<<<<<<<<<<<
 * 	return rs
 * 
 */
  __pyx_t_2 = PyNumber_Add(__pyx_kp_u_BED_file, __pyx_v_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_rgns)) { __Pyx_RaiseUnboundLocalError("rgns"); __PYX_ERR(0, 761, __pyx_L1_error) }
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_rgns);
  __Pyx_GIVEREF(__pyx_v_rgns);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_rgns);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_v_rs = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":762
 * 			rgns.append( region( seq, int(float(start)), int(float(end)), strand, score, source, feature, group = group ) )
 * 	rs = regions('BED file: ' + path, rgns)
 * 	return rs             # <<<<<<<<<<<<<<
 * 
 * # Loads regions from a GFF file.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_rs));
  __pyx_r = ((PyObject *)__pyx_v_rs);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":726
 * 
 * # Loads regions from a gzipped BED file.
 * def loadBEDGZ(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("gnocis.regions.loadBEDGZ", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rgns);
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_seq);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_end);
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XDECREF(__pyx_v_feature);
  __Pyx_XDECREF(__pyx_v_group);
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XDECREF((PyObject *)__pyx_v_rs);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_9loadBEDGZ_2generator9);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":765
 * 
 * # Loads regions from a GFF file.
 * def loadGFF(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_5loadGFF(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_4loadGFF[] = " Loads regions from a General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.\n\t\n\t:param path: Path to the input file.\n\t:type path: str\n\t\n\t:return: Loaded regions.\n\t:rtype: regions\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_5loadGFF = {"loadGFF", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_5loadGFF, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_4loadGFF};
static PyObject *__pyx_pw_6gnocis_7regions_5loadGFF(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_dropChr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("loadGFF (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,&__pyx_n_s_dropChr,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dropChr);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "loadGFF") < 0)) __PYX_ERR(0, 765, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_path = values[0];
    __pyx_v_dropChr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("loadGFF", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 765, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.loadGFF", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_4loadGFF(__pyx_self, __pyx_v_path, __pyx_v_dropChr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_4loadGFF(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_dropChr) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_rgns = 0;
  PyObject *__pyx_v_r = 0;
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_rs = 0;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("loadGFF", 0);

  /* "gnocis/regions.pyx":777
 * 	cdef list rgns, r
 * 	cdef regions rs
 * 	rgns = []             # <<<<<<<<<<<<<<
 * 	with open(path, 'r') as f:
 * 		for x in f:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rgns = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":778
 * 	cdef regions rs
 * 	rgns = []
 * 	with open(path, 'r') as f:             # <<<<<<<<<<<<<<
 * 		for x in f:
 * 			if x.startswith('#'): continue
 */
  /*with:*/ {
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_path);
    __Pyx_GIVEREF(__pyx_v_path);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_path);
    __Pyx_INCREF(__pyx_n_u_r);
    __Pyx_GIVEREF(__pyx_n_u_r);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_r);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 778, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {
          __pyx_v_f = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":779
 * 	rgns = []
 * 	with open(path, 'r') as f:
 * 		for x in f:             # <<<<<<<<<<<<<<
 * 			if x.startswith('#'): continue
 * 			r = x.strip().split('\t')
 */
          if (likely(PyList_CheckExact(__pyx_v_f)) || PyTuple_CheckExact(__pyx_v_f)) {
            __pyx_t_4 = __pyx_v_f; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
            __pyx_t_10 = NULL;
          } else {
            __pyx_t_9 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 779, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_10 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 779, __pyx_L7_error)
          }
          for (;;) {
            if (likely(!__pyx_t_10)) {
              if (likely(PyList_CheckExact(__pyx_t_4))) {
                if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 779, __pyx_L7_error)
                #else
                __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 779, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_2);
                #endif
              } else {
                if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 779, __pyx_L7_error)
                #else
                __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 779, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_2);
                #endif
              }
            } else {
              __pyx_t_2 = __pyx_t_10(__pyx_t_4);
              if (unlikely(!__pyx_t_2)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 779, __pyx_L7_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_2);
            }
            if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 779, __pyx_L7_error)
            __Pyx_XDECREF_SET(__pyx_v_x, ((PyObject*)__pyx_t_2));
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":780
 * 	with open(path, 'r') as f:
 * 		for x in f:
 * 			if x.startswith('#'): continue             # <<<<<<<<<<<<<<
 * 			r = x.strip().split('\t')
 * 			if len(r) != 9: continue
 */
            if (unlikely(__pyx_v_x == Py_None)) {
              PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "startswith");
              __PYX_ERR(0, 780, __pyx_L7_error)
            }
            __pyx_t_11 = __Pyx_PyUnicode_Tailmatch(__pyx_v_x, __pyx_kp_u__22, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 780, __pyx_L7_error)
            if ((__pyx_t_11 != 0)) {
              goto __pyx_L13_continue;
            }

            /* "gnocis/regions.pyx":781
 * 		for x in f:
 * 			if x.startswith('#'): continue
 * 			r = x.strip().split('\t')             # <<<<<<<<<<<<<<
 * 			if len(r) != 9: continue
 * 			rgns.append( region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) )
 */
            __pyx_t_1 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_strip, __pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_split); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 781, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_kp_u__19) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u__19);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 781, __pyx_L7_error)
            __Pyx_XDECREF_SET(__pyx_v_r, ((PyObject*)__pyx_t_2));
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":782
 * 			if x.startswith('#'): continue
 * 			r = x.strip().split('\t')
 * 			if len(r) != 9: continue             # <<<<<<<<<<<<<<
 * 			rgns.append( region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) )
 * 	rs = regions('GFF file: ' + path, rgns)
 */
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
              __PYX_ERR(0, 782, __pyx_L7_error)
            }
            __pyx_t_12 = PyList_GET_SIZE(__pyx_v_r); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 782, __pyx_L7_error)
            __pyx_t_11 = ((__pyx_t_12 != 9) != 0);
            if (__pyx_t_11) {
              goto __pyx_L13_continue;
            }

            /* "gnocis/regions.pyx":783
 * 			r = x.strip().split('\t')
 * 			if len(r) != 9: continue
 * 			rgns.append( region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) )             # <<<<<<<<<<<<<<
 * 	rs = regions('GFF file: ' + path, rgns)
 * 	return rs
 */
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 783, __pyx_L7_error)
            }
            __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 783, __pyx_L7_error)
            }
            __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_r, 3, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 783, __pyx_L7_error)
            }
            __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_r, 4, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_13 = __Pyx_PyNumber_Float(__pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 783, __pyx_L7_error)
            }
            __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_r, 6, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_14 = PyObject_RichCompare(__pyx_t_1, __pyx_kp_u__7, Py_NE); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 783, __pyx_L7_error)
            }
            __pyx_t_15 = __Pyx_GetItemInt_List(__pyx_v_r, 5, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_11 = (__Pyx_PyUnicode_Equals(__pyx_t_15, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (__pyx_t_11) {
              if (unlikely(__pyx_v_r == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 783, __pyx_L7_error)
              }
              __pyx_t_15 = __Pyx_GetItemInt_List(__pyx_v_r, 5, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 783, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_16 = __Pyx_PyNumber_Float(__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 783, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_1 = __pyx_t_16;
              __pyx_t_16 = 0;
            } else {
              __Pyx_INCREF(__pyx_float_0_0);
              __pyx_t_1 = __pyx_float_0_0;
            }
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 783, __pyx_L7_error)
            }
            __pyx_t_16 = __Pyx_GetItemInt_List(__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 783, __pyx_L7_error)
            }
            __pyx_t_15 = __Pyx_GetItemInt_List(__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_17 = PyTuple_New(7); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_13);
            PyTuple_SET_ITEM(__pyx_t_17, 2, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyTuple_SET_ITEM(__pyx_t_17, 3, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_17, 4, __pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_16);
            PyTuple_SET_ITEM(__pyx_t_17, 5, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_15);
            PyTuple_SET_ITEM(__pyx_t_17, 6, __pyx_t_15);
            __pyx_t_2 = 0;
            __pyx_t_5 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_1 = 0;
            __pyx_t_16 = 0;
            __pyx_t_15 = 0;
            __pyx_t_15 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);
            if (unlikely(__pyx_v_r == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(0, 783, __pyx_L7_error)
            }
            __pyx_t_16 = __Pyx_GetItemInt_List(__pyx_v_r, 8, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_group, __pyx_t_16) < 0) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_dropChr, __pyx_v_dropChr) < 0) __PYX_ERR(0, 783, __pyx_L7_error)
            __pyx_t_16 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_17, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_rgns, __pyx_t_16); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 783, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "gnocis/regions.pyx":779
 * 	rgns = []
 * 	with open(path, 'r') as f:
 * 		for x in f:             # <<<<<<<<<<<<<<
 * 			if x.startswith('#'): continue
 * 			r = x.strip().split('\t')
 */
            __pyx_L13_continue:;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":778
 * 	cdef regions rs
 * 	rgns = []
 * 	with open(path, 'r') as f:             # <<<<<<<<<<<<<<
 * 		for x in f:
 * 			if x.startswith('#'): continue
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.loadGFF", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_16, &__pyx_t_15) < 0) __PYX_ERR(0, 778, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_17 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_16, __pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 778, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_17, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 778, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_19);
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (__pyx_t_11 < 0) __PYX_ERR(0, 778, __pyx_L9_except_error)
          __pyx_t_20 = ((!(__pyx_t_11 != 0)) != 0);
          if (__pyx_t_20) {
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_16, __pyx_t_15);
            __pyx_t_4 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; 
            __PYX_ERR(0, 778, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 778, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L20;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L20:;
  }

  /* "gnocis/regions.pyx":784
 * 			if len(r) != 9: continue
 * 			rgns.append( region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) )
 * 	rs = regions('GFF file: ' + path, rgns)             # <<<<<<<<<<<<<<
 * 	return rs
 * 
 */
  __pyx_t_15 = PyNumber_Add(__pyx_kp_u_GFF_file, __pyx_v_path); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_15);
  __Pyx_INCREF(__pyx_v_rgns);
  __Pyx_GIVEREF(__pyx_v_rgns);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_v_rgns);
  __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_16, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_v_rs = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "gnocis/regions.pyx":785
 * 			rgns.append( region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) )
 * 	rs = regions('GFF file: ' + path, rgns)
 * 	return rs             # <<<<<<<<<<<<<<
 * 
 * # Loads regions from a gzipped GFF file.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_rs));
  __pyx_r = ((PyObject *)__pyx_v_rs);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":765
 * 
 * # Loads regions from a GFF file.
 * def loadGFF(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("gnocis.regions.loadGFF", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_rgns);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF((PyObject *)__pyx_v_rs);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":788
 * 
 * # Loads regions from a gzipped GFF file.
 * def loadGFFGZ(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_7loadGFFGZ(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_6loadGFFGZ[] = " Loads regions from a gzipped General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.\n\t\n\t:param path: Path to the input file.\n\t:type path: str\n\t\n\t:return: Loaded regions.\n\t:rtype: regions\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_7loadGFFGZ = {"loadGFFGZ", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_7loadGFFGZ, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_6loadGFFGZ};
static PyObject *__pyx_pw_6gnocis_7regions_7loadGFFGZ(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_dropChr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("loadGFFGZ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,&__pyx_n_s_dropChr,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dropChr);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "loadGFFGZ") < 0)) __PYX_ERR(0, 788, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_path = values[0];
    __pyx_v_dropChr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("loadGFFGZ", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 788, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.loadGFFGZ", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_6loadGFFGZ(__pyx_self, __pyx_v_path, __pyx_v_dropChr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_9loadGFFGZ_9genexpr24_2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */
static PyObject *__pyx_gb_6gnocis_7regions_9loadGFFGZ_9genexpr24_7genexpr_2generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":798
 * 	"""
 * 	with gzip.open(path, 'rb') as f:
 * 		raw_gff = [ y for y in (x.split('\t') for x in (y.decode('utf-8') for y in f) if not x.startswith('#')) if len(y)==9 ]             # <<<<<<<<<<<<<<
 * 		rgns = [ region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) for r in raw_gff ]
 * 	rs = regions('GFF GZ file: ' + path, rgns)
 */

static PyObject *__pyx_pf_6gnocis_7regions_9loadGFFGZ_9genexpr24_7genexpr_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_17_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_17_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 798, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_9loadGFFGZ_9genexpr24_7genexpr_2generator11, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_loadGFFGZ_locals_genexpr_locals, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.loadGFFGZ.genexpr24.genexpr.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_9loadGFFGZ_9genexpr24_7genexpr_2generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 798, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_f)) { __Pyx_RaiseClosureNameError("f"); __PYX_ERR(0, 798, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_f)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_f)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_f; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_f); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 798, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_y);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_y, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_y, __pyx_n_s_decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_utf_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 798, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_9loadGFFGZ_9genexpr24_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *__pyx_cur_scope;
  PyObject *__pyx_gb_6gnocis_7regions_9loadGFFGZ_9genexpr24_7genexpr_2generator11 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_16_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_16_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 798, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_9loadGFFGZ_9genexpr24_2generator10, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_loadGFFGZ_locals_genexpr, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.loadGFFGZ.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_9loadGFFGZ_9genexpr24_7genexpr_2generator11);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_9loadGFFGZ_9genexpr24_2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 798, __pyx_L1_error)
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_9loadGFFGZ_9genexpr24_7genexpr_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 798, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_x);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_x, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_x, __pyx_n_s_startswith); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u__22) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u__22);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = ((!__pyx_t_7) != 0);
    if (__pyx_t_8) {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_x, __pyx_n_s_split); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u__19) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u__19);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      __Pyx_XGIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 798, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":788
 * 
 * # Loads regions from a gzipped GFF file.
 * def loadGFFGZ(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_6loadGFFGZ(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_dropChr) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *__pyx_cur_scope;
  PyObject *__pyx_v_raw_gff = NULL;
  PyObject *__pyx_v_rgns = NULL;
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_rs = NULL;
  PyObject *__pyx_9genexpr24__pyx_v_y = NULL;
  PyObject *__pyx_9genexpr24__pyx_v_0 = NULL;
  PyObject *__pyx_9genexpr27__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("loadGFFGZ", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 788, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/regions.pyx":797
 * 	:rtype: regions
 * 	"""
 * 	with gzip.open(path, 'rb') as f:             # <<<<<<<<<<<<<<
 * 		raw_gff = [ y for y in (x.split('\t') for x in (y.decode('utf-8') for y in f) if not x.startswith('#')) if len(y)==9 ]
 * 		rgns = [ region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) for r in raw_gff ]
 */
  /*with:*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_gzip); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_open); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_path, __pyx_n_u_rb};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_path, __pyx_n_u_rb};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_path);
      __Pyx_GIVEREF(__pyx_v_path);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_path);
      __Pyx_INCREF(__pyx_n_u_rb);
      __Pyx_GIVEREF(__pyx_n_u_rb);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_n_u_rb);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 797, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 797, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {
          __Pyx_GIVEREF(__pyx_t_5);
          __pyx_cur_scope->__pyx_v_f = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":798
 * 	"""
 * 	with gzip.open(path, 'rb') as f:
 * 		raw_gff = [ y for y in (x.split('\t') for x in (y.decode('utf-8') for y in f) if not x.startswith('#')) if len(y)==9 ]             # <<<<<<<<<<<<<<
 * 		rgns = [ region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) for r in raw_gff ]
 * 	rs = regions('GFF GZ file: ' + path, rgns)
 */
          { /* enter inner scope */
            __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = __pyx_pf_6gnocis_7regions_9loadGFFGZ_9genexpr24_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
              __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
              __pyx_t_11 = NULL;
            } else {
              __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L15_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 798, __pyx_L15_error)
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            for (;;) {
              if (likely(!__pyx_t_11)) {
                if (likely(PyList_CheckExact(__pyx_t_3))) {
                  if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L15_error)
                  #else
                  __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L15_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  #endif
                } else {
                  if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L15_error)
                  #else
                  __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L15_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  #endif
                }
              } else {
                __pyx_t_1 = __pyx_t_11(__pyx_t_3);
                if (unlikely(!__pyx_t_1)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 798, __pyx_L15_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_1);
              }
              __Pyx_XDECREF_SET(__pyx_9genexpr24__pyx_v_y, __pyx_t_1);
              __pyx_t_1 = 0;
              __pyx_t_12 = PyObject_Length(__pyx_9genexpr24__pyx_v_y); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 798, __pyx_L15_error)
              __pyx_t_13 = ((__pyx_t_12 == 9) != 0);
              if (__pyx_t_13) {
                if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_9genexpr24__pyx_v_y))) __PYX_ERR(0, 798, __pyx_L15_error)
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_y); __pyx_9genexpr24__pyx_v_y = 0;
            goto __pyx_L19_exit_scope;
            __pyx_L15_error:;
            __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_y); __pyx_9genexpr24__pyx_v_y = 0;
            goto __pyx_L7_error;
            __pyx_L19_exit_scope:;
          } /* exit inner scope */
          __pyx_v_raw_gff = ((PyObject*)__pyx_t_5);
          __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":799
 * 	with gzip.open(path, 'rb') as f:
 * 		raw_gff = [ y for y in (x.split('\t') for x in (y.decode('utf-8') for y in f) if not x.startswith('#')) if len(y)==9 ]
 * 		rgns = [ region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) for r in raw_gff ]             # <<<<<<<<<<<<<<
 * 	rs = regions('GFF GZ file: ' + path, rgns)
 * 	rs.sort()
 */
          { /* enter inner scope */
            __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 799, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_3 = __pyx_v_raw_gff; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
            for (;;) {
              if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 799, __pyx_L22_error)
              #else
              __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
              __Pyx_XDECREF_SET(__pyx_9genexpr27__pyx_v_r, __pyx_t_1);
              __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_2 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_14 = __Pyx_PyNumber_Float(__pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_15 = __Pyx_PyNumber_Float(__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyNumber_Int(__pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_15 = __Pyx_PyInt_SubtractObjC(__pyx_t_14, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_16 = PyObject_RichCompare(__pyx_t_14, __pyx_kp_u__7, Py_NE); __Pyx_XGOTREF(__pyx_t_16); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_17 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_13 = (__Pyx_PyUnicode_Equals(__pyx_t_17, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (__pyx_t_13) {
                __pyx_t_17 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 799, __pyx_L22_error)
                __Pyx_GOTREF(__pyx_t_17);
                __pyx_t_18 = __Pyx_PyNumber_Float(__pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 799, __pyx_L22_error)
                __Pyx_GOTREF(__pyx_t_18);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_14 = __pyx_t_18;
                __pyx_t_18 = 0;
              } else {
                __Pyx_INCREF(__pyx_float_0_0);
                __pyx_t_14 = __pyx_float_0_0;
              }
              __pyx_t_18 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_17 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_19 = PyTuple_New(7); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_19);
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_2);
              PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_2);
              __Pyx_GIVEREF(__pyx_t_15);
              PyTuple_SET_ITEM(__pyx_t_19, 2, __pyx_t_15);
              __Pyx_GIVEREF(__pyx_t_16);
              PyTuple_SET_ITEM(__pyx_t_19, 3, __pyx_t_16);
              __Pyx_GIVEREF(__pyx_t_14);
              PyTuple_SET_ITEM(__pyx_t_19, 4, __pyx_t_14);
              __Pyx_GIVEREF(__pyx_t_18);
              PyTuple_SET_ITEM(__pyx_t_19, 5, __pyx_t_18);
              __Pyx_GIVEREF(__pyx_t_17);
              PyTuple_SET_ITEM(__pyx_t_19, 6, __pyx_t_17);
              __pyx_t_1 = 0;
              __pyx_t_2 = 0;
              __pyx_t_15 = 0;
              __pyx_t_16 = 0;
              __pyx_t_14 = 0;
              __pyx_t_18 = 0;
              __pyx_t_17 = 0;
              __pyx_t_17 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_18 = __Pyx_GetItemInt(__pyx_9genexpr27__pyx_v_r, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_18);
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_group, __pyx_t_18) < 0) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_dropChr, __pyx_v_dropChr) < 0) __PYX_ERR(0, 799, __pyx_L22_error)
              __pyx_t_18 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_19, __pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_18);
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_18))) __PYX_ERR(0, 799, __pyx_L22_error)
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_r); __pyx_9genexpr27__pyx_v_r = 0;
            goto __pyx_L25_exit_scope;
            __pyx_L22_error:;
            __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_r); __pyx_9genexpr27__pyx_v_r = 0;
            goto __pyx_L7_error;
            __pyx_L25_exit_scope:;
          } /* exit inner scope */
          __pyx_v_rgns = ((PyObject*)__pyx_t_5);
          __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":797
 * 	:rtype: regions
 * 	"""
 * 	with gzip.open(path, 'rb') as f:             # <<<<<<<<<<<<<<
 * 		raw_gff = [ y for y in (x.split('\t') for x in (y.decode('utf-8') for y in f) if not x.startswith('#')) if len(y)==9 ]
 * 		rgns = [ region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) for r in raw_gff ]
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.loadGFFGZ", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_3, &__pyx_t_18) < 0) __PYX_ERR(0, 797, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_17 = PyTuple_Pack(3, __pyx_t_5, __pyx_t_3, __pyx_t_18); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 797, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_17, NULL);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 797, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_20);
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_20);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          if (__pyx_t_13 < 0) __PYX_ERR(0, 797, __pyx_L9_except_error)
          __pyx_t_21 = ((!(__pyx_t_13 != 0)) != 0);
          if (__pyx_t_21) {
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_3);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_3, __pyx_t_18);
            __pyx_t_5 = 0; __pyx_t_3 = 0; __pyx_t_18 = 0; 
            __PYX_ERR(0, 797, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_6) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 797, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L29;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L1_error;
    __pyx_L29:;
  }

  /* "gnocis/regions.pyx":800
 * 		raw_gff = [ y for y in (x.split('\t') for x in (y.decode('utf-8') for y in f) if not x.startswith('#')) if len(y)==9 ]
 * 		rgns = [ region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) for r in raw_gff ]
 * 	rs = regions('GFF GZ file: ' + path, rgns)             # <<<<<<<<<<<<<<
 * 	rs.sort()
 * 	return rs
 */
  __pyx_t_18 = PyNumber_Add(__pyx_kp_u_GFF_GZ_file, __pyx_v_path); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (unlikely(!__pyx_v_rgns)) { __Pyx_RaiseUnboundLocalError("rgns"); __PYX_ERR(0, 800, __pyx_L1_error) }
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_18);
  __Pyx_INCREF(__pyx_v_rgns);
  __Pyx_GIVEREF(__pyx_v_rgns);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rgns);
  __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_3, NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_rs = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_18);
  __pyx_t_18 = 0;

  /* "gnocis/regions.pyx":801
 * 		rgns = [ region( r[0], int(float(r[3])), int(float(r[4]))-1, r[6] != '-', float(r[5]) if r[5] != '.' else 0.0, r[1], r[2], group = r[8], dropChr = dropChr ) for r in raw_gff ]
 * 	rs = regions('GFF GZ file: ' + path, rgns)
 * 	rs.sort()             # <<<<<<<<<<<<<<
 * 	return rs
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rs), __pyx_n_s_sort); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_18 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "gnocis/regions.pyx":802
 * 	rs = regions('GFF GZ file: ' + path, rgns)
 * 	rs.sort()
 * 	return rs             # <<<<<<<<<<<<<<
 * 
 * # Loads regions from a coordinate list file.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_rs));
  __pyx_r = ((PyObject *)__pyx_v_rs);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":788
 * 
 * # Loads regions from a gzipped GFF file.
 * def loadGFFGZ(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("gnocis.regions.loadGFFGZ", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_raw_gff);
  __Pyx_XDECREF(__pyx_v_rgns);
  __Pyx_XDECREF((PyObject *)__pyx_v_rs);
  __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_y);
  __Pyx_XDECREF(__pyx_9genexpr24__pyx_v_0);
  __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_r);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":805
 * 
 * # Loads regions from a coordinate list file.
 * def loadCoordinateList(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a coordinate list file (one region per line, with the format `seq:start..end`).
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_9loadCoordinateList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_8loadCoordinateList[] = " Loads regions from a coordinate list file (one region per line, with the format `seq:start..end`).\n\t\n\t:param path: Path to the input file.\n\t:type path: str\n\t\n\t:return: Loaded regions.\n\t:rtype: regions\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_9loadCoordinateList = {"loadCoordinateList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_9loadCoordinateList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_8loadCoordinateList};
static PyObject *__pyx_pw_6gnocis_7regions_9loadCoordinateList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_dropChr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("loadCoordinateList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,&__pyx_n_s_dropChr,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dropChr);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "loadCoordinateList") < 0)) __PYX_ERR(0, 805, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_path = values[0];
    __pyx_v_dropChr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("loadCoordinateList", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 805, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.loadCoordinateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_8loadCoordinateList(__pyx_self, __pyx_v_path, __pyx_v_dropChr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_8loadCoordinateList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_dropChr) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_rgns = NULL;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_pos = NULL;
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_rs = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("loadCoordinateList", 0);

  /* "gnocis/regions.pyx":814
 * 	:rtype: regions
 * 	"""
 * 	with open(path, 'r') as f:             # <<<<<<<<<<<<<<
 * 		rgns = []
 * 		for l in f:
 */
  /*with:*/ {
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_path);
    __Pyx_GIVEREF(__pyx_v_path);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_path);
    __Pyx_INCREF(__pyx_n_u_r);
    __Pyx_GIVEREF(__pyx_n_u_r);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_r);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_2, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 814, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {
          __pyx_v_f = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":815
 * 	"""
 * 	with open(path, 'r') as f:
 * 		rgns = []             # <<<<<<<<<<<<<<
 * 		for l in f:
 * 			e = l.strip().split(':')
 */
          __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 815, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_v_rgns = ((PyObject*)__pyx_t_4);
          __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":816
 * 	with open(path, 'r') as f:
 * 		rgns = []
 * 		for l in f:             # <<<<<<<<<<<<<<
 * 			e = l.strip().split(':')
 * 			if len(e) != 2:
 */
          if (likely(PyList_CheckExact(__pyx_v_f)) || PyTuple_CheckExact(__pyx_v_f)) {
            __pyx_t_4 = __pyx_v_f; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
            __pyx_t_10 = NULL;
          } else {
            __pyx_t_9 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 816, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_10 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 816, __pyx_L7_error)
          }
          for (;;) {
            if (likely(!__pyx_t_10)) {
              if (likely(PyList_CheckExact(__pyx_t_4))) {
                if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 816, __pyx_L7_error)
                #else
                __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 816, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_2);
                #endif
              } else {
                if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 816, __pyx_L7_error)
                #else
                __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 816, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_2);
                #endif
              }
            } else {
              __pyx_t_2 = __pyx_t_10(__pyx_t_4);
              if (unlikely(!__pyx_t_2)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 816, __pyx_L7_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_2);
            }
            __Pyx_XDECREF_SET(__pyx_v_l, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":817
 * 		rgns = []
 * 		for l in f:
 * 			e = l.strip().split(':')             # <<<<<<<<<<<<<<
 * 			if len(e) != 2:
 * 				raise Exception('Coordinate list syntax error')
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_l, __pyx_n_s_strip); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 817, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_11 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_11)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_11);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_split); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 817, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_kp_u__3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u__3);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 817, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":818
 * 		for l in f:
 * 			e = l.strip().split(':')
 * 			if len(e) != 2:             # <<<<<<<<<<<<<<
 * 				raise Exception('Coordinate list syntax error')
 * 			pos = e[1].split('..')
 */
            __pyx_t_12 = PyObject_Length(__pyx_v_e); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 818, __pyx_L7_error)
            __pyx_t_13 = ((__pyx_t_12 != 2) != 0);
            if (unlikely(__pyx_t_13)) {

              /* "gnocis/regions.pyx":819
 * 			e = l.strip().split(':')
 * 			if len(e) != 2:
 * 				raise Exception('Coordinate list syntax error')             # <<<<<<<<<<<<<<
 * 			pos = e[1].split('..')
 * 			if len(pos) != 2:
 */
              __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 819, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_Raise(__pyx_t_2, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __PYX_ERR(0, 819, __pyx_L7_error)

              /* "gnocis/regions.pyx":818
 * 		for l in f:
 * 			e = l.strip().split(':')
 * 			if len(e) != 2:             # <<<<<<<<<<<<<<
 * 				raise Exception('Coordinate list syntax error')
 * 			pos = e[1].split('..')
 */
            }

            /* "gnocis/regions.pyx":820
 * 			if len(e) != 2:
 * 				raise Exception('Coordinate list syntax error')
 * 			pos = e[1].split('..')             # <<<<<<<<<<<<<<
 * 			if len(pos) != 2:
 * 				raise Exception('Coordinate list syntax error')
 */
            __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_e, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 820, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_split); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_kp_u__4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u__4);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF_SET(__pyx_v_pos, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":821
 * 				raise Exception('Coordinate list syntax error')
 * 			pos = e[1].split('..')
 * 			if len(pos) != 2:             # <<<<<<<<<<<<<<
 * 				raise Exception('Coordinate list syntax error')
 * 			rgns.append( region(e[0], int(float(pos[0])), int(float(pos[1])), dropChr = dropChr) )
 */
            __pyx_t_12 = PyObject_Length(__pyx_v_pos); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 821, __pyx_L7_error)
            __pyx_t_13 = ((__pyx_t_12 != 2) != 0);
            if (unlikely(__pyx_t_13)) {

              /* "gnocis/regions.pyx":822
 * 			pos = e[1].split('..')
 * 			if len(pos) != 2:
 * 				raise Exception('Coordinate list syntax error')             # <<<<<<<<<<<<<<
 * 			rgns.append( region(e[0], int(float(pos[0])), int(float(pos[1])), dropChr = dropChr) )
 * 	rs = regions('Coordinate list file: ' + path, rgns)
 */
              __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_Raise(__pyx_t_2, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __PYX_ERR(0, 822, __pyx_L7_error)

              /* "gnocis/regions.pyx":821
 * 				raise Exception('Coordinate list syntax error')
 * 			pos = e[1].split('..')
 * 			if len(pos) != 2:             # <<<<<<<<<<<<<<
 * 				raise Exception('Coordinate list syntax error')
 * 			rgns.append( region(e[0], int(float(pos[0])), int(float(pos[1])), dropChr = dropChr) )
 */
            }

            /* "gnocis/regions.pyx":823
 * 			if len(pos) != 2:
 * 				raise Exception('Coordinate list syntax error')
 * 			rgns.append( region(e[0], int(float(pos[0])), int(float(pos[1])), dropChr = dropChr) )             # <<<<<<<<<<<<<<
 * 	rs = regions('Coordinate list file: ' + path, rgns)
 * 	rs.sort()
 */
            __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_e, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_pos, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_5 = __Pyx_PyNumber_Float(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_pos, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_11 = __Pyx_PyNumber_Float(__pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_t_5);
            __pyx_t_2 = 0;
            __pyx_t_1 = 0;
            __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dropChr, __pyx_v_dropChr) < 0) __PYX_ERR(0, 823, __pyx_L7_error)
            __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_t_11, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_rgns, __pyx_t_1); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 823, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "gnocis/regions.pyx":816
 * 	with open(path, 'r') as f:
 * 		rgns = []
 * 		for l in f:             # <<<<<<<<<<<<<<
 * 			e = l.strip().split(':')
 * 			if len(e) != 2:
 */
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "gnocis/regions.pyx":814
 * 	:rtype: regions
 * 	"""
 * 	with open(path, 'r') as f:             # <<<<<<<<<<<<<<
 * 		rgns = []
 * 		for l in f:
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.loadCoordinateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_5) < 0) __PYX_ERR(0, 814, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_11 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 814, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 814, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (__pyx_t_13 < 0) __PYX_ERR(0, 814, __pyx_L9_except_error)
          __pyx_t_16 = ((!(__pyx_t_13 != 0)) != 0);
          if (__pyx_t_16) {
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_XGIVEREF(__pyx_t_5);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_1, __pyx_t_5);
            __pyx_t_4 = 0; __pyx_t_1 = 0; __pyx_t_5 = 0; 
            __PYX_ERR(0, 814, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_3) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 814, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L20;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L1_error;
    __pyx_L20:;
  }

  /* "gnocis/regions.pyx":824
 * 				raise Exception('Coordinate list syntax error')
 * 			rgns.append( region(e[0], int(float(pos[0])), int(float(pos[1])), dropChr = dropChr) )
 * 	rs = regions('Coordinate list file: ' + path, rgns)             # <<<<<<<<<<<<<<
 * 	rs.sort()
 * 	return rs
 */
  __pyx_t_5 = PyNumber_Add(__pyx_kp_u_Coordinate_list_file, __pyx_v_path); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(!__pyx_v_rgns)) { __Pyx_RaiseUnboundLocalError("rgns"); __PYX_ERR(0, 824, __pyx_L1_error) }
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_rgns);
  __Pyx_GIVEREF(__pyx_v_rgns);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_rgns);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rs = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":825
 * 			rgns.append( region(e[0], int(float(pos[0])), int(float(pos[1])), dropChr = dropChr) )
 * 	rs = regions('Coordinate list file: ' + path, rgns)
 * 	rs.sort()             # <<<<<<<<<<<<<<
 * 	return rs
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rs), __pyx_n_s_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/regions.pyx":826
 * 	rs = regions('Coordinate list file: ' + path, rgns)
 * 	rs.sort()
 * 	return rs             # <<<<<<<<<<<<<<
 * 
 * # Generates a barplot for a region-based measure
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_rs));
  __pyx_r = ((PyObject *)__pyx_v_rs);
  goto __pyx_L0;

  /* "gnocis/regions.pyx":805
 * 
 * # Loads regions from a coordinate list file.
 * def loadCoordinateList(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a coordinate list file (one region per line, with the format `seq:start..end`).
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("gnocis.regions.loadCoordinateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_rgns);
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XDECREF((PyObject *)__pyx_v_rs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":829
 * 
 * # Generates a barplot for a region-based measure
 * def regionBarplot(mainRegions, measure, measureName, testRegions = None, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeValues = 12, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = False, isPercent = False):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of a measure (lambda) for predicted regions.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_11regionBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_10regionBarplot[] = " Generates a barplot of a measure (lambda) for predicted regions.\n\t\n\t:param mainRegions: List of main region sets.\n\t:param measure: Function to compute measure.\n\t:param measureName: Name of measure.\n\t:param testRegions: List of test region sets, to check against (or None if not relevant).\n\t:param figsize: Figure size.\n\t:param outpath: Output path.\n\t:param returnHTML: If True, an HTML node will be returned.\n\t:param fontsizeLabels: Size of label font.\n\t:param fontsizeLegend: Size of legend font.\n\t:param fontsizeAxis: Size of axis font.\n\t:param style: Plot style to use.\n\t:param showLegend: Flag for whether or not to render legend.\n\t:param bboxAnchorTo: Legend anchor point.\n\t:param legendLoc: Legend location.\n\t:param showValues: If True, values will be plotted.\n\t:param isPercent: If True, values will be plotted as percentages.\n\t\n\t:type mainRegions: list\n\t:type measure: function\n\t:type measureName: str\n\t:type testRegions: list, optional\n\t:type figsize: tuple, optional\n\t:type outpath: str, optional\n\t:type returnHTML: bool, optional\n\t:type fontsizeLabels: float, optional\n\t:type fontsizeLegend: float, optional\n\t:type fontsizeAxis: float, optional\n\t:type style: str, optional\n\t:type showLegend: bool, optional\n\t:type bboxAnchorTo: tuple, optional\n\t:type legendLoc: str, optional\n\t:type showValues: bool, optional\n\t:type isPercent: bool, optional\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_11regionBarplot = {"regionBarplot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_11regionBarplot, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_10regionBarplot};
static PyObject *__pyx_pw_6gnocis_7regions_11regionBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mainRegions = 0;
  PyObject *__pyx_v_measure = 0;
  PyObject *__pyx_v_measureName = 0;
  PyObject *__pyx_v_testRegions = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_returnHTML = 0;
  PyObject *__pyx_v_fontsizeLabels = 0;
  PyObject *__pyx_v_fontsizeValues = 0;
  PyObject *__pyx_v_fontsizeLegend = 0;
  PyObject *__pyx_v_fontsizeAxis = 0;
  PyObject *__pyx_v_style = 0;
  PyObject *__pyx_v_showLegend = 0;
  PyObject *__pyx_v_bboxAnchorTo = 0;
  PyObject *__pyx_v_legendLoc = 0;
  PyObject *__pyx_v_showValues = 0;
  PyObject *__pyx_v_isPercent = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("regionBarplot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mainRegions,&__pyx_n_s_measure,&__pyx_n_s_measureName,&__pyx_n_s_testRegions,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_returnHTML,&__pyx_n_s_fontsizeLabels,&__pyx_n_s_fontsizeValues,&__pyx_n_s_fontsizeLegend,&__pyx_n_s_fontsizeAxis,&__pyx_n_s_style,&__pyx_n_s_showLegend,&__pyx_n_s_bboxAnchorTo,&__pyx_n_s_legendLoc,&__pyx_n_s_showValues,&__pyx_n_s_isPercent,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)__pyx_tuple__25);
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)Py_False);
    values[7] = ((PyObject *)__pyx_int_18);
    values[8] = ((PyObject *)__pyx_int_12);
    values[9] = ((PyObject *)__pyx_int_12);
    values[10] = ((PyObject *)__pyx_int_10);
    values[11] = ((PyObject *)__pyx_n_u_ggplot);
    values[12] = ((PyObject *)Py_True);
    values[13] = ((PyObject *)__pyx_tuple__26);
    values[14] = ((PyObject *)__pyx_kp_u_upper_left);
    values[15] = ((PyObject *)Py_False);
    values[16] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mainRegions)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_measure)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("regionBarplot", 0, 3, 17, 1); __PYX_ERR(0, 829, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_measureName)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("regionBarplot", 0, 3, 17, 2); __PYX_ERR(0, 829, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_testRegions);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_returnHTML);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLabels);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeValues);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLegend);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeAxis);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showLegend);
          if (value) { values[12] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bboxAnchorTo);
          if (value) { values[13] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_legendLoc);
          if (value) { values[14] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showValues);
          if (value) { values[15] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_isPercent);
          if (value) { values[16] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "regionBarplot") < 0)) __PYX_ERR(0, 829, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mainRegions = values[0];
    __pyx_v_measure = values[1];
    __pyx_v_measureName = values[2];
    __pyx_v_testRegions = values[3];
    __pyx_v_figsize = values[4];
    __pyx_v_outpath = values[5];
    __pyx_v_returnHTML = values[6];
    __pyx_v_fontsizeLabels = values[7];
    __pyx_v_fontsizeValues = values[8];
    __pyx_v_fontsizeLegend = values[9];
    __pyx_v_fontsizeAxis = values[10];
    __pyx_v_style = values[11];
    __pyx_v_showLegend = values[12];
    __pyx_v_bboxAnchorTo = values[13];
    __pyx_v_legendLoc = values[14];
    __pyx_v_showValues = values[15];
    __pyx_v_isPercent = values[16];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("regionBarplot", 0, 3, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 829, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.regionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_10regionBarplot(__pyx_self, __pyx_v_mainRegions, __pyx_v_measure, __pyx_v_measureName, __pyx_v_testRegions, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_returnHTML, __pyx_v_fontsizeLabels, __pyx_v_fontsizeValues, __pyx_v_fontsizeLegend, __pyx_v_fontsizeAxis, __pyx_v_style, __pyx_v_showLegend, __pyx_v_bboxAnchorTo, __pyx_v_legendLoc, __pyx_v_showValues, __pyx_v_isPercent);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_10regionBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mainRegions, PyObject *__pyx_v_measure, PyObject *__pyx_v_measureName, PyObject *__pyx_v_testRegions, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeValues, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues, PyObject *__pyx_v_isPercent) {
  PyObject *__pyx_v_plt = NULL;
  PyObject *__pyx_v_base64 = NULL;
  PyObject *__pyx_v_BytesIO = NULL;
  PyObject *__pyx_v_display = NULL;
  PyObject *__pyx_v_HTML = NULL;
  PyObject *__pyx_v_mtick = NULL;
  double __pyx_v_width;
  double __pyx_v_bw;
  PyObject *__pyx_v_fig = NULL;
  PyObject *__pyx_v_ax = NULL;
  double __pyx_v_scaling;
  PyObject *__pyx_v_hasTestRegions = NULL;
  PyObject *__pyx_v_psi = NULL;
  PyObject *__pyx_v_ps = NULL;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_prs = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_vMean = NULL;
  PyObject *__pyx_v_vCI = NULL;
  PyObject *__pyx_v_cx = NULL;
  PyObject *__pyx_v_m = NULL;
  PyObject *__pyx_v_ci = NULL;
  PyObject *__pyx_v_cv = NULL;
  PyObject *__pyx_v_bio = NULL;
  PyObject *__pyx_v_encoded = NULL;
  PyObject *__pyx_v_html = NULL;
  Py_ssize_t __pyx_9genexpr28__pyx_v_rsi;
  PyObject *__pyx_9genexpr29__pyx_v_rs = NULL;
  PyObject *__pyx_9genexpr30__pyx_v__ps = NULL;
  PyObject *__pyx_9genexpr31__pyx_v_cv = NULL;
  PyObject *__pyx_9genexpr32__pyx_v_cv = NULL;
  PyObject *__pyx_9genexpr33__pyx_v_rs = NULL;
  Py_ssize_t __pyx_9genexpr34__pyx_v_i;
  PyObject *__pyx_9genexpr35__pyx_v_rs = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  long __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  double __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *(*__pyx_t_21)(PyObject *);
  PyObject *(*__pyx_t_22)(PyObject *);
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  int __pyx_t_27;
  PyObject *__pyx_t_28 = NULL;
  Py_UCS4 __pyx_t_29;
  PyObject *__pyx_t_30 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("regionBarplot", 0);
  __Pyx_INCREF(__pyx_v_testRegions);

  /* "gnocis/regions.pyx":866
 * 	:type isPercent: bool, optional
 * 	"""
 * 	import matplotlib.pyplot as plt             # <<<<<<<<<<<<<<
 * 	import base64
 * 	from io import BytesIO
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__27);
  __Pyx_GIVEREF(__pyx_n_s__27);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__27);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_matplotlib_pyplot, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_plt = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":867
 * 	"""
 * 	import matplotlib.pyplot as plt
 * 	import base64             # <<<<<<<<<<<<<<
 * 	from io import BytesIO
 * 	from IPython.core.display import display, HTML
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_base64, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_base64 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":868
 * 	import matplotlib.pyplot as plt
 * 	import base64
 * 	from io import BytesIO             # <<<<<<<<<<<<<<
 * 	from IPython.core.display import display, HTML
 * 	import matplotlib.ticker as mtick
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_BytesIO);
  __Pyx_GIVEREF(__pyx_n_s_BytesIO);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_BytesIO);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_io, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_BytesIO); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_BytesIO = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":869
 * 	import base64
 * 	from io import BytesIO
 * 	from IPython.core.display import display, HTML             # <<<<<<<<<<<<<<
 * 	import matplotlib.ticker as mtick
 * 	with plt.style.context(style):
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_display);
  __Pyx_GIVEREF(__pyx_n_s_display);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_display);
  __Pyx_INCREF(__pyx_n_s_HTML);
  __Pyx_GIVEREF(__pyx_n_s_HTML);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_HTML);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_IPython_core_display, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_display); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_display = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_HTML); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_HTML = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":870
 * 	from io import BytesIO
 * 	from IPython.core.display import display, HTML
 * 	import matplotlib.ticker as mtick             # <<<<<<<<<<<<<<
 * 	with plt.style.context(style):
 * 		width = 0.5
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__27);
  __Pyx_GIVEREF(__pyx_n_s__27);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__27);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_matplotlib_ticker, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_mtick = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":871
 * 	from IPython.core.display import display, HTML
 * 	import matplotlib.ticker as mtick
 * 	with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 		width = 0.5
 * 		bw = width / max(len(mainRegions)-1, 1)
 */
  /*with:*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_style); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_style) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_style);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 871, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {

          /* "gnocis/regions.pyx":872
 * 	import matplotlib.ticker as mtick
 * 	with plt.style.context(style):
 * 		width = 0.5             # <<<<<<<<<<<<<<
 * 		bw = width / max(len(mainRegions)-1, 1)
 * 		fig, ax = plt.subplots(figsize = figsize)
 */
          __pyx_v_width = 0.5;

          /* "gnocis/regions.pyx":873
 * 	with plt.style.context(style):
 * 		width = 0.5
 * 		bw = width / max(len(mainRegions)-1, 1)             # <<<<<<<<<<<<<<
 * 		fig, ax = plt.subplots(figsize = figsize)
 * 		scaling = 100. if isPercent else 1.
 */
          __pyx_t_9 = 1;
          __pyx_t_10 = PyObject_Length(__pyx_v_mainRegions); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 873, __pyx_L7_error)
          __pyx_t_11 = (__pyx_t_10 - 1);
          if (((__pyx_t_9 > __pyx_t_11) != 0)) {
            __pyx_t_10 = __pyx_t_9;
          } else {
            __pyx_t_10 = __pyx_t_11;
          }
          __pyx_t_11 = __pyx_t_10;
          if (unlikely(__pyx_t_11 == 0)) {
            PyErr_SetString(PyExc_ZeroDivisionError, "float division");
            __PYX_ERR(0, 873, __pyx_L7_error)
          }
          __pyx_v_bw = (__pyx_v_width / ((double)__pyx_t_11));

          /* "gnocis/regions.pyx":874
 * 		width = 0.5
 * 		bw = width / max(len(mainRegions)-1, 1)
 * 		fig, ax = plt.subplots(figsize = figsize)             # <<<<<<<<<<<<<<
 * 		scaling = 100. if isPercent else 1.
 * 		hasTestRegions = testRegions is not None
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_subplots); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 874, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 874, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 874, __pyx_L7_error)
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 874, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
            PyObject* sequence = __pyx_t_2;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 874, __pyx_L7_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            #else
            __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 874, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 874, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 874, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
            index = 0; __pyx_t_3 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L13_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_3);
            index = 1; __pyx_t_1 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L13_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_1);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(0, 874, __pyx_L7_error)
            __pyx_t_12 = NULL;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            goto __pyx_L14_unpacking_done;
            __pyx_L13_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_12 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 874, __pyx_L7_error)
            __pyx_L14_unpacking_done:;
          }
          __pyx_v_fig = __pyx_t_3;
          __pyx_t_3 = 0;
          __pyx_v_ax = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "gnocis/regions.pyx":875
 * 		bw = width / max(len(mainRegions)-1, 1)
 * 		fig, ax = plt.subplots(figsize = figsize)
 * 		scaling = 100. if isPercent else 1.             # <<<<<<<<<<<<<<
 * 		hasTestRegions = testRegions is not None
 * 		if testRegions is None:
 */
          __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_isPercent); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 875, __pyx_L7_error)
          if (__pyx_t_14) {
            __pyx_t_13 = 100.;
          } else {
            __pyx_t_13 = 1.;
          }
          __pyx_v_scaling = __pyx_t_13;

          /* "gnocis/regions.pyx":876
 * 		fig, ax = plt.subplots(figsize = figsize)
 * 		scaling = 100. if isPercent else 1.
 * 		hasTestRegions = testRegions is not None             # <<<<<<<<<<<<<<
 * 		if testRegions is None:
 * 			testRegions = [ 0 ]
 */
          __pyx_t_14 = (__pyx_v_testRegions != Py_None);
          __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 876, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_v_hasTestRegions = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "gnocis/regions.pyx":877
 * 		scaling = 100. if isPercent else 1.
 * 		hasTestRegions = testRegions is not None
 * 		if testRegions is None:             # <<<<<<<<<<<<<<
 * 			testRegions = [ 0 ]
 * 		for psi, ps in enumerate(mainRegions):
 */
          __pyx_t_14 = (__pyx_v_testRegions == Py_None);
          __pyx_t_15 = (__pyx_t_14 != 0);
          if (__pyx_t_15) {

            /* "gnocis/regions.pyx":878
 * 		hasTestRegions = testRegions is not None
 * 		if testRegions is None:
 * 			testRegions = [ 0 ]             # <<<<<<<<<<<<<<
 * 		for psi, ps in enumerate(mainRegions):
 * 			x = [
 */
            __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 878, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_INCREF(__pyx_int_0);
            __Pyx_GIVEREF(__pyx_int_0);
            PyList_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
            __Pyx_DECREF_SET(__pyx_v_testRegions, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":877
 * 		scaling = 100. if isPercent else 1.
 * 		hasTestRegions = testRegions is not None
 * 		if testRegions is None:             # <<<<<<<<<<<<<<
 * 			testRegions = [ 0 ]
 * 		for psi, ps in enumerate(mainRegions):
 */
          }

          /* "gnocis/regions.pyx":879
 * 		if testRegions is None:
 * 			testRegions = [ 0 ]
 * 		for psi, ps in enumerate(mainRegions):             # <<<<<<<<<<<<<<
 * 			x = [
 * 				float(rsi) - (width/2) + (bw*psi)
 */
          __Pyx_INCREF(__pyx_int_0);
          __pyx_t_2 = __pyx_int_0;
          if (likely(PyList_CheckExact(__pyx_v_mainRegions)) || PyTuple_CheckExact(__pyx_v_mainRegions)) {
            __pyx_t_1 = __pyx_v_mainRegions; __Pyx_INCREF(__pyx_t_1); __pyx_t_11 = 0;
            __pyx_t_16 = NULL;
          } else {
            __pyx_t_11 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_mainRegions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_16 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 879, __pyx_L7_error)
          }
          for (;;) {
            if (likely(!__pyx_t_16)) {
              if (likely(PyList_CheckExact(__pyx_t_1))) {
                if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_1)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 879, __pyx_L7_error)
                #else
                __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 879, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_3);
                #endif
              } else {
                if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 879, __pyx_L7_error)
                #else
                __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 879, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_3);
                #endif
              }
            } else {
              __pyx_t_3 = __pyx_t_16(__pyx_t_1);
              if (unlikely(!__pyx_t_3)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 879, __pyx_L7_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_3);
            }
            __Pyx_XDECREF_SET(__pyx_v_ps, __pyx_t_3);
            __pyx_t_3 = 0;
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_XDECREF_SET(__pyx_v_psi, __pyx_t_2);
            __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 879, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2);
            __pyx_t_2 = __pyx_t_3;
            __pyx_t_3 = 0;

            /* "gnocis/regions.pyx":880
 * 			testRegions = [ 0 ]
 * 		for psi, ps in enumerate(mainRegions):
 * 			x = [             # <<<<<<<<<<<<<<
 * 				float(rsi) - (width/2) + (bw*psi)
 * 				for rsi in range(len(testRegions))
 */
            { /* enter inner scope */
              __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_3);

              /* "gnocis/regions.pyx":882
 * 			x = [
 * 				float(rsi) - (width/2) + (bw*psi)
 * 				for rsi in range(len(testRegions))             # <<<<<<<<<<<<<<
 * 			]
 * 			if isinstance(ps, CVModelPredictions):
 */
              __pyx_t_10 = PyObject_Length(__pyx_v_testRegions); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 882, __pyx_L7_error)
              __pyx_t_17 = __pyx_t_10;
              for (__pyx_t_18 = 0; __pyx_t_18 < __pyx_t_17; __pyx_t_18+=1) {
                __pyx_9genexpr28__pyx_v_rsi = __pyx_t_18;

                /* "gnocis/regions.pyx":881
 * 		for psi, ps in enumerate(mainRegions):
 * 			x = [
 * 				float(rsi) - (width/2) + (bw*psi)             # <<<<<<<<<<<<<<
 * 				for rsi in range(len(testRegions))
 * 			]
 */
                __pyx_t_5 = PyFloat_FromDouble((((double)__pyx_9genexpr28__pyx_v_rsi) - (__pyx_v_width / 2.0))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 881, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_19 = PyFloat_FromDouble(__pyx_v_bw); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 881, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_19);
                __pyx_t_20 = PyNumber_Multiply(__pyx_t_19, __pyx_v_psi); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 881, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __pyx_t_19 = PyNumber_Add(__pyx_t_5, __pyx_t_20); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 881, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_19))) __PYX_ERR(0, 880, __pyx_L7_error)
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              }
            } /* exit inner scope */
            __Pyx_XDECREF_SET(__pyx_v_x, ((PyObject*)__pyx_t_3));
            __pyx_t_3 = 0;

            /* "gnocis/regions.pyx":884
 * 				for rsi in range(len(testRegions))
 * 			]
 * 			if isinstance(ps, CVModelPredictions):             # <<<<<<<<<<<<<<
 * 				# TODO Add multiclass support
 * 				prs = ps.regions(label = positive)
 */
            __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_CVModelPredictions); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_15 = PyObject_IsInstance(__pyx_v_ps, __pyx_t_3); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 884, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_14 = (__pyx_t_15 != 0);
            if (__pyx_t_14) {

              /* "gnocis/regions.pyx":886
 * 			if isinstance(ps, CVModelPredictions):
 * 				# TODO Add multiclass support
 * 				prs = ps.regions(label = positive)             # <<<<<<<<<<<<<<
 * 				v = [
 * 					[
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ps, __pyx_n_s_regions); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 886, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_19 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 886, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_19);
              __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_n_s_positive); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 886, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_20);
              if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_label, __pyx_t_20) < 0) __PYX_ERR(0, 886, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_19); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 886, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_XDECREF_SET(__pyx_v_prs, __pyx_t_20);
              __pyx_t_20 = 0;

              /* "gnocis/regions.pyx":887
 * 				# TODO Add multiclass support
 * 				prs = ps.regions(label = positive)
 * 				v = [             # <<<<<<<<<<<<<<
 * 					[
 * 						scaling * measure(_ps, rs) #100. * len(rs.overlap(_ps)) / len(rs)
 */
              { /* enter inner scope */
                __pyx_t_20 = PyList_New(0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 887, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_20);

                /* "gnocis/regions.pyx":892
 * 						for _ps in prs
 * 					]
 * 					for rs in testRegions             # <<<<<<<<<<<<<<
 * 				]
 * 				vMean = [
 */
                if (likely(PyList_CheckExact(__pyx_v_testRegions)) || PyTuple_CheckExact(__pyx_v_testRegions)) {
                  __pyx_t_19 = __pyx_v_testRegions; __Pyx_INCREF(__pyx_t_19); __pyx_t_10 = 0;
                  __pyx_t_21 = NULL;
                } else {
                  __pyx_t_10 = -1; __pyx_t_19 = PyObject_GetIter(__pyx_v_testRegions); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 892, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_19);
                  __pyx_t_21 = Py_TYPE(__pyx_t_19)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 892, __pyx_L23_error)
                }
                for (;;) {
                  if (likely(!__pyx_t_21)) {
                    if (likely(PyList_CheckExact(__pyx_t_19))) {
                      if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_19)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_3 = PyList_GET_ITEM(__pyx_t_19, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 892, __pyx_L23_error)
                      #else
                      __pyx_t_3 = PySequence_ITEM(__pyx_t_19, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L23_error)
                      __Pyx_GOTREF(__pyx_t_3);
                      #endif
                    } else {
                      if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_19)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_19, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 892, __pyx_L23_error)
                      #else
                      __pyx_t_3 = PySequence_ITEM(__pyx_t_19, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L23_error)
                      __Pyx_GOTREF(__pyx_t_3);
                      #endif
                    }
                  } else {
                    __pyx_t_3 = __pyx_t_21(__pyx_t_19);
                    if (unlikely(!__pyx_t_3)) {
                      PyObject* exc_type = PyErr_Occurred();
                      if (exc_type) {
                        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                        else __PYX_ERR(0, 892, __pyx_L23_error)
                      }
                      break;
                    }
                    __Pyx_GOTREF(__pyx_t_3);
                  }
                  __Pyx_XDECREF_SET(__pyx_9genexpr29__pyx_v_rs, __pyx_t_3);
                  __pyx_t_3 = 0;
                  { /* enter inner scope */

                    /* "gnocis/regions.pyx":888
 * 				prs = ps.regions(label = positive)
 * 				v = [
 * 					[             # <<<<<<<<<<<<<<
 * 						scaling * measure(_ps, rs) #100. * len(rs.overlap(_ps)) / len(rs)
 * 						for _ps in prs
 */
                    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 888, __pyx_L28_error)
                    __Pyx_GOTREF(__pyx_t_3);

                    /* "gnocis/regions.pyx":890
 * 					[
 * 						scaling * measure(_ps, rs) #100. * len(rs.overlap(_ps)) / len(rs)
 * 						for _ps in prs             # <<<<<<<<<<<<<<
 * 					]
 * 					for rs in testRegions
 */
                    if (likely(PyList_CheckExact(__pyx_v_prs)) || PyTuple_CheckExact(__pyx_v_prs)) {
                      __pyx_t_5 = __pyx_v_prs; __Pyx_INCREF(__pyx_t_5); __pyx_t_17 = 0;
                      __pyx_t_22 = NULL;
                    } else {
                      __pyx_t_17 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_prs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 890, __pyx_L28_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      __pyx_t_22 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 890, __pyx_L28_error)
                    }
                    for (;;) {
                      if (likely(!__pyx_t_22)) {
                        if (likely(PyList_CheckExact(__pyx_t_5))) {
                          if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_5)) break;
                          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                          __pyx_t_23 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_17); __Pyx_INCREF(__pyx_t_23); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 890, __pyx_L28_error)
                          #else
                          __pyx_t_23 = PySequence_ITEM(__pyx_t_5, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 890, __pyx_L28_error)
                          __Pyx_GOTREF(__pyx_t_23);
                          #endif
                        } else {
                          if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                          __pyx_t_23 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_17); __Pyx_INCREF(__pyx_t_23); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 890, __pyx_L28_error)
                          #else
                          __pyx_t_23 = PySequence_ITEM(__pyx_t_5, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 890, __pyx_L28_error)
                          __Pyx_GOTREF(__pyx_t_23);
                          #endif
                        }
                      } else {
                        __pyx_t_23 = __pyx_t_22(__pyx_t_5);
                        if (unlikely(!__pyx_t_23)) {
                          PyObject* exc_type = PyErr_Occurred();
                          if (exc_type) {
                            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                            else __PYX_ERR(0, 890, __pyx_L28_error)
                          }
                          break;
                        }
                        __Pyx_GOTREF(__pyx_t_23);
                      }
                      __Pyx_XDECREF_SET(__pyx_9genexpr30__pyx_v__ps, __pyx_t_23);
                      __pyx_t_23 = 0;

                      /* "gnocis/regions.pyx":889
 * 				v = [
 * 					[
 * 						scaling * measure(_ps, rs) #100. * len(rs.overlap(_ps)) / len(rs)             # <<<<<<<<<<<<<<
 * 						for _ps in prs
 * 					]
 */
                      __pyx_t_23 = PyFloat_FromDouble(__pyx_v_scaling); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 889, __pyx_L28_error)
                      __Pyx_GOTREF(__pyx_t_23);
                      __Pyx_INCREF(__pyx_v_measure);
                      __pyx_t_25 = __pyx_v_measure; __pyx_t_26 = NULL;
                      __pyx_t_27 = 0;
                      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_25))) {
                        __pyx_t_26 = PyMethod_GET_SELF(__pyx_t_25);
                        if (likely(__pyx_t_26)) {
                          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_25);
                          __Pyx_INCREF(__pyx_t_26);
                          __Pyx_INCREF(function);
                          __Pyx_DECREF_SET(__pyx_t_25, function);
                          __pyx_t_27 = 1;
                        }
                      }
                      #if CYTHON_FAST_PYCALL
                      if (PyFunction_Check(__pyx_t_25)) {
                        PyObject *__pyx_temp[3] = {__pyx_t_26, __pyx_9genexpr30__pyx_v__ps, __pyx_9genexpr29__pyx_v_rs};
                        __pyx_t_24 = __Pyx_PyFunction_FastCall(__pyx_t_25, __pyx_temp+1-__pyx_t_27, 2+__pyx_t_27); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 889, __pyx_L28_error)
                        __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
                        __Pyx_GOTREF(__pyx_t_24);
                      } else
                      #endif
                      #if CYTHON_FAST_PYCCALL
                      if (__Pyx_PyFastCFunction_Check(__pyx_t_25)) {
                        PyObject *__pyx_temp[3] = {__pyx_t_26, __pyx_9genexpr30__pyx_v__ps, __pyx_9genexpr29__pyx_v_rs};
                        __pyx_t_24 = __Pyx_PyCFunction_FastCall(__pyx_t_25, __pyx_temp+1-__pyx_t_27, 2+__pyx_t_27); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 889, __pyx_L28_error)
                        __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
                        __Pyx_GOTREF(__pyx_t_24);
                      } else
                      #endif
                      {
                        __pyx_t_28 = PyTuple_New(2+__pyx_t_27); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 889, __pyx_L28_error)
                        __Pyx_GOTREF(__pyx_t_28);
                        if (__pyx_t_26) {
                          __Pyx_GIVEREF(__pyx_t_26); PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_26); __pyx_t_26 = NULL;
                        }
                        __Pyx_INCREF(__pyx_9genexpr30__pyx_v__ps);
                        __Pyx_GIVEREF(__pyx_9genexpr30__pyx_v__ps);
                        PyTuple_SET_ITEM(__pyx_t_28, 0+__pyx_t_27, __pyx_9genexpr30__pyx_v__ps);
                        __Pyx_INCREF(__pyx_9genexpr29__pyx_v_rs);
                        __Pyx_GIVEREF(__pyx_9genexpr29__pyx_v_rs);
                        PyTuple_SET_ITEM(__pyx_t_28, 1+__pyx_t_27, __pyx_9genexpr29__pyx_v_rs);
                        __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_t_28, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 889, __pyx_L28_error)
                        __Pyx_GOTREF(__pyx_t_24);
                        __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
                      }
                      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                      __pyx_t_25 = PyNumber_Multiply(__pyx_t_23, __pyx_t_24); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 889, __pyx_L28_error)
                      __Pyx_GOTREF(__pyx_t_25);
                      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_25))) __PYX_ERR(0, 888, __pyx_L28_error)
                      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

                      /* "gnocis/regions.pyx":890
 * 					[
 * 						scaling * measure(_ps, rs) #100. * len(rs.overlap(_ps)) / len(rs)
 * 						for _ps in prs             # <<<<<<<<<<<<<<
 * 					]
 * 					for rs in testRegions
 */
                    }
                    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                    __Pyx_XDECREF(__pyx_9genexpr30__pyx_v__ps); __pyx_9genexpr30__pyx_v__ps = 0;
                    goto __pyx_L31_exit_scope;
                    __pyx_L28_error:;
                    __Pyx_XDECREF(__pyx_9genexpr30__pyx_v__ps); __pyx_9genexpr30__pyx_v__ps = 0;
                    goto __pyx_L23_error;
                    __pyx_L31_exit_scope:;
                  } /* exit inner scope */
                  if (unlikely(__Pyx_ListComp_Append(__pyx_t_20, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 887, __pyx_L23_error)
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                  /* "gnocis/regions.pyx":892
 * 						for _ps in prs
 * 					]
 * 					for rs in testRegions             # <<<<<<<<<<<<<<
 * 				]
 * 				vMean = [
 */
                }
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_XDECREF(__pyx_9genexpr29__pyx_v_rs); __pyx_9genexpr29__pyx_v_rs = 0;
                goto __pyx_L32_exit_scope;
                __pyx_L23_error:;
                __Pyx_XDECREF(__pyx_9genexpr29__pyx_v_rs); __pyx_9genexpr29__pyx_v_rs = 0;
                goto __pyx_L7_error;
                __pyx_L32_exit_scope:;
              } /* exit inner scope */
              __Pyx_XDECREF_SET(__pyx_v_v, ((PyObject*)__pyx_t_20));
              __pyx_t_20 = 0;

              /* "gnocis/regions.pyx":894
 * 					for rs in testRegions
 * 				]
 * 				vMean = [             # <<<<<<<<<<<<<<
 * 					mean(cv)
 * 					for cv in v
 */
              { /* enter inner scope */
                __pyx_t_20 = PyList_New(0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 894, __pyx_L35_error)
                __Pyx_GOTREF(__pyx_t_20);

                /* "gnocis/regions.pyx":896
 * 				vMean = [
 * 					mean(cv)
 * 					for cv in v             # <<<<<<<<<<<<<<
 * 				]
 * 				vCI = [
 */
                __pyx_t_19 = __pyx_v_v; __Pyx_INCREF(__pyx_t_19); __pyx_t_10 = 0;
                for (;;) {
                  if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_19)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_3 = PyList_GET_ITEM(__pyx_t_19, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 896, __pyx_L35_error)
                  #else
                  __pyx_t_3 = PySequence_ITEM(__pyx_t_19, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L35_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  #endif
                  __Pyx_XDECREF_SET(__pyx_9genexpr31__pyx_v_cv, __pyx_t_3);
                  __pyx_t_3 = 0;

                  /* "gnocis/regions.pyx":895
 * 				]
 * 				vMean = [
 * 					mean(cv)             # <<<<<<<<<<<<<<
 * 					for cv in v
 * 				]
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_mean); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 895, __pyx_L35_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __pyx_t_25 = NULL;
                  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                    __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_5);
                    if (likely(__pyx_t_25)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                      __Pyx_INCREF(__pyx_t_25);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_5, function);
                    }
                  }
                  __pyx_t_3 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_25, __pyx_9genexpr31__pyx_v_cv) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_9genexpr31__pyx_v_cv);
                  __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
                  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 895, __pyx_L35_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  if (unlikely(__Pyx_ListComp_Append(__pyx_t_20, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 894, __pyx_L35_error)
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                  /* "gnocis/regions.pyx":896
 * 				vMean = [
 * 					mean(cv)
 * 					for cv in v             # <<<<<<<<<<<<<<
 * 				]
 * 				vCI = [
 */
                }
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_XDECREF(__pyx_9genexpr31__pyx_v_cv); __pyx_9genexpr31__pyx_v_cv = 0;
                goto __pyx_L38_exit_scope;
                __pyx_L35_error:;
                __Pyx_XDECREF(__pyx_9genexpr31__pyx_v_cv); __pyx_9genexpr31__pyx_v_cv = 0;
                goto __pyx_L7_error;
                __pyx_L38_exit_scope:;
              } /* exit inner scope */
              __Pyx_XDECREF_SET(__pyx_v_vMean, ((PyObject*)__pyx_t_20));
              __pyx_t_20 = 0;

              /* "gnocis/regions.pyx":898
 * 					for cv in v
 * 				]
 * 				vCI = [             # <<<<<<<<<<<<<<
 * 					CI(cv)
 * 					for cv in v
 */
              { /* enter inner scope */
                __pyx_t_20 = PyList_New(0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 898, __pyx_L41_error)
                __Pyx_GOTREF(__pyx_t_20);

                /* "gnocis/regions.pyx":900
 * 				vCI = [
 * 					CI(cv)
 * 					for cv in v             # <<<<<<<<<<<<<<
 * 				]
 * 				ax.bar(x, vMean, bw, label = ps.name, yerr = vCI)
 */
                __pyx_t_19 = __pyx_v_v; __Pyx_INCREF(__pyx_t_19); __pyx_t_10 = 0;
                for (;;) {
                  if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_19)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_3 = PyList_GET_ITEM(__pyx_t_19, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 900, __pyx_L41_error)
                  #else
                  __pyx_t_3 = PySequence_ITEM(__pyx_t_19, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L41_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  #endif
                  __Pyx_XDECREF_SET(__pyx_9genexpr32__pyx_v_cv, __pyx_t_3);
                  __pyx_t_3 = 0;

                  /* "gnocis/regions.pyx":899
 * 				]
 * 				vCI = [
 * 					CI(cv)             # <<<<<<<<<<<<<<
 * 					for cv in v
 * 				]
 */
                  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CI); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 899, __pyx_L41_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __pyx_t_25 = NULL;
                  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                    __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_5);
                    if (likely(__pyx_t_25)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                      __Pyx_INCREF(__pyx_t_25);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_5, function);
                    }
                  }
                  __pyx_t_3 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_25, __pyx_9genexpr32__pyx_v_cv) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_9genexpr32__pyx_v_cv);
                  __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
                  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 899, __pyx_L41_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  if (unlikely(__Pyx_ListComp_Append(__pyx_t_20, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 898, __pyx_L41_error)
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                  /* "gnocis/regions.pyx":900
 * 				vCI = [
 * 					CI(cv)
 * 					for cv in v             # <<<<<<<<<<<<<<
 * 				]
 * 				ax.bar(x, vMean, bw, label = ps.name, yerr = vCI)
 */
                }
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_XDECREF(__pyx_9genexpr32__pyx_v_cv); __pyx_9genexpr32__pyx_v_cv = 0;
                goto __pyx_L44_exit_scope;
                __pyx_L41_error:;
                __Pyx_XDECREF(__pyx_9genexpr32__pyx_v_cv); __pyx_9genexpr32__pyx_v_cv = 0;
                goto __pyx_L7_error;
                __pyx_L44_exit_scope:;
              } /* exit inner scope */
              __Pyx_XDECREF_SET(__pyx_v_vCI, ((PyObject*)__pyx_t_20));
              __pyx_t_20 = 0;

              /* "gnocis/regions.pyx":902
 * 					for cv in v
 * 				]
 * 				ax.bar(x, vMean, bw, label = ps.name, yerr = vCI)             # <<<<<<<<<<<<<<
 * 				if showValues:
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):
 */
              __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_ax, __pyx_n_s_bar); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 902, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_20);
              __pyx_t_19 = PyFloat_FromDouble(__pyx_v_bw); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 902, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_19);
              __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 902, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_INCREF(__pyx_v_x);
              __Pyx_GIVEREF(__pyx_v_x);
              PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_x);
              __Pyx_INCREF(__pyx_v_vMean);
              __Pyx_GIVEREF(__pyx_v_vMean);
              PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_vMean);
              __Pyx_GIVEREF(__pyx_t_19);
              PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_19);
              __pyx_t_19 = 0;
              __pyx_t_19 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 902, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_19);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_ps, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 902, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_label, __pyx_t_5) < 0) __PYX_ERR(0, 902, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_yerr, __pyx_v_vCI) < 0) __PYX_ERR(0, 902, __pyx_L7_error)
              __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_t_3, __pyx_t_19); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 902, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "gnocis/regions.pyx":903
 * 				]
 * 				ax.bar(x, vMean, bw, label = ps.name, yerr = vCI)
 * 				if showValues:             # <<<<<<<<<<<<<<
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):
 * 						plt.text(cx, m+ci,
 */
              __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_showValues); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 903, __pyx_L7_error)
              if (__pyx_t_14) {

                /* "gnocis/regions.pyx":904
 * 				ax.bar(x, vMean, bw, label = ps.name, yerr = vCI)
 * 				if showValues:
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):             # <<<<<<<<<<<<<<
 * 						plt.text(cx, m+ci,
 * 							'%.2f +/- %.2f'%(m, ci) + (' %' if isPercent else ''),
 */
                __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 904, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_INCREF(__pyx_v_vMean);
                __Pyx_GIVEREF(__pyx_v_vMean);
                PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_vMean);
                __Pyx_INCREF(__pyx_v_vCI);
                __Pyx_GIVEREF(__pyx_v_vCI);
                PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_vCI);
                __pyx_t_19 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 904, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 904, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_INCREF(__pyx_v_x);
                __Pyx_GIVEREF(__pyx_v_x);
                PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_x);
                __Pyx_GIVEREF(__pyx_t_19);
                PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_19);
                __pyx_t_19 = 0;
                __pyx_t_19 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 904, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                if (likely(PyList_CheckExact(__pyx_t_19)) || PyTuple_CheckExact(__pyx_t_19)) {
                  __pyx_t_5 = __pyx_t_19; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
                  __pyx_t_21 = NULL;
                } else {
                  __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_19); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 904, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __pyx_t_21 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 904, __pyx_L7_error)
                }
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                for (;;) {
                  if (likely(!__pyx_t_21)) {
                    if (likely(PyList_CheckExact(__pyx_t_5))) {
                      if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_19 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_19); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 904, __pyx_L7_error)
                      #else
                      __pyx_t_19 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 904, __pyx_L7_error)
                      __Pyx_GOTREF(__pyx_t_19);
                      #endif
                    } else {
                      if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_19 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_19); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 904, __pyx_L7_error)
                      #else
                      __pyx_t_19 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 904, __pyx_L7_error)
                      __Pyx_GOTREF(__pyx_t_19);
                      #endif
                    }
                  } else {
                    __pyx_t_19 = __pyx_t_21(__pyx_t_5);
                    if (unlikely(!__pyx_t_19)) {
                      PyObject* exc_type = PyErr_Occurred();
                      if (exc_type) {
                        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                        else __PYX_ERR(0, 904, __pyx_L7_error)
                      }
                      break;
                    }
                    __Pyx_GOTREF(__pyx_t_19);
                  }
                  if ((likely(PyTuple_CheckExact(__pyx_t_19))) || (PyList_CheckExact(__pyx_t_19))) {
                    PyObject* sequence = __pyx_t_19;
                    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                    if (unlikely(size != 2)) {
                      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                      __PYX_ERR(0, 904, __pyx_L7_error)
                    }
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    if (likely(PyTuple_CheckExact(sequence))) {
                      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
                      __pyx_t_20 = PyTuple_GET_ITEM(sequence, 1); 
                    } else {
                      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
                      __pyx_t_20 = PyList_GET_ITEM(sequence, 1); 
                    }
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_20);
                    #else
                    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 904, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __pyx_t_20 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 904, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_20);
                    #endif
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  } else {
                    Py_ssize_t index = -1;
                    __pyx_t_25 = PyObject_GetIter(__pyx_t_19); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 904, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_25);
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                    __pyx_t_12 = Py_TYPE(__pyx_t_25)->tp_iternext;
                    index = 0; __pyx_t_3 = __pyx_t_12(__pyx_t_25); if (unlikely(!__pyx_t_3)) goto __pyx_L48_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_3);
                    index = 1; __pyx_t_20 = __pyx_t_12(__pyx_t_25); if (unlikely(!__pyx_t_20)) goto __pyx_L48_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_20);
                    if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_25), 2) < 0) __PYX_ERR(0, 904, __pyx_L7_error)
                    __pyx_t_12 = NULL;
                    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                    goto __pyx_L49_unpacking_done;
                    __pyx_L48_unpacking_failed:;
                    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                    __pyx_t_12 = NULL;
                    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                    __PYX_ERR(0, 904, __pyx_L7_error)
                    __pyx_L49_unpacking_done:;
                  }
                  __Pyx_XDECREF_SET(__pyx_v_cx, __pyx_t_3);
                  __pyx_t_3 = 0;
                  if ((likely(PyTuple_CheckExact(__pyx_t_20))) || (PyList_CheckExact(__pyx_t_20))) {
                    PyObject* sequence = __pyx_t_20;
                    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                    if (unlikely(size != 2)) {
                      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                      __PYX_ERR(0, 904, __pyx_L7_error)
                    }
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    if (likely(PyTuple_CheckExact(sequence))) {
                      __pyx_t_25 = PyTuple_GET_ITEM(sequence, 0); 
                      __pyx_t_24 = PyTuple_GET_ITEM(sequence, 1); 
                    } else {
                      __pyx_t_25 = PyList_GET_ITEM(sequence, 0); 
                      __pyx_t_24 = PyList_GET_ITEM(sequence, 1); 
                    }
                    __Pyx_INCREF(__pyx_t_25);
                    __Pyx_INCREF(__pyx_t_24);
                    #else
                    __pyx_t_25 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 904, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_25);
                    __pyx_t_24 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 904, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_24);
                    #endif
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  } else {
                    Py_ssize_t index = -1;
                    __pyx_t_23 = PyObject_GetIter(__pyx_t_20); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 904, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_23);
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                    __pyx_t_12 = Py_TYPE(__pyx_t_23)->tp_iternext;
                    index = 0; __pyx_t_25 = __pyx_t_12(__pyx_t_23); if (unlikely(!__pyx_t_25)) goto __pyx_L50_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_25);
                    index = 1; __pyx_t_24 = __pyx_t_12(__pyx_t_23); if (unlikely(!__pyx_t_24)) goto __pyx_L50_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_24);
                    if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_23), 2) < 0) __PYX_ERR(0, 904, __pyx_L7_error)
                    __pyx_t_12 = NULL;
                    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                    goto __pyx_L51_unpacking_done;
                    __pyx_L50_unpacking_failed:;
                    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                    __pyx_t_12 = NULL;
                    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                    __PYX_ERR(0, 904, __pyx_L7_error)
                    __pyx_L51_unpacking_done:;
                  }
                  __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_25);
                  __pyx_t_25 = 0;
                  __Pyx_XDECREF_SET(__pyx_v_ci, __pyx_t_24);
                  __pyx_t_24 = 0;

                  /* "gnocis/regions.pyx":905
 * 				if showValues:
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):
 * 						plt.text(cx, m+ci,             # <<<<<<<<<<<<<<
 * 							'%.2f +/- %.2f'%(m, ci) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 */
                  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_text); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 905, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_19);
                  __pyx_t_20 = PyNumber_Add(__pyx_v_m, __pyx_v_ci); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 905, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_20);

                  /* "gnocis/regions.pyx":906
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):
 * 						plt.text(cx, m+ci,
 * 							'%.2f +/- %.2f'%(m, ci) + (' %' if isPercent else ''),             # <<<<<<<<<<<<<<
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 * 							fontsize = fontsizeValues)
 */
                  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 906, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __pyx_t_17 = 0;
                  __pyx_t_29 = 127;
                  __pyx_t_24 = __Pyx_PyObject_Format(__pyx_v_m, __pyx_kp_u_2f); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 906, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_24);
                  __pyx_t_29 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_24) > __pyx_t_29) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_24) : __pyx_t_29;
                  __pyx_t_17 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_24);
                  __Pyx_GIVEREF(__pyx_t_24);
                  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_24);
                  __pyx_t_24 = 0;
                  __Pyx_INCREF(__pyx_kp_u__28);
                  __pyx_t_17 += 5;
                  __Pyx_GIVEREF(__pyx_kp_u__28);
                  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u__28);
                  __pyx_t_24 = __Pyx_PyObject_Format(__pyx_v_ci, __pyx_kp_u_2f); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 906, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_24);
                  __pyx_t_29 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_24) > __pyx_t_29) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_24) : __pyx_t_29;
                  __pyx_t_17 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_24);
                  __Pyx_GIVEREF(__pyx_t_24);
                  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_24);
                  __pyx_t_24 = 0;
                  __pyx_t_24 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_17, __pyx_t_29); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 906, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_24);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_isPercent); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 906, __pyx_L7_error)
                  if (__pyx_t_14) {
                    __Pyx_INCREF(__pyx_kp_u__29);
                    __pyx_t_3 = __pyx_kp_u__29;
                  } else {
                    __Pyx_INCREF(__pyx_kp_u__13);
                    __pyx_t_3 = __pyx_kp_u__13;
                  }
                  __pyx_t_25 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_24, __pyx_t_3); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 906, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_25);
                  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                  /* "gnocis/regions.pyx":905
 * 				if showValues:
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):
 * 						plt.text(cx, m+ci,             # <<<<<<<<<<<<<<
 * 							'%.2f +/- %.2f'%(m, ci) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 */
                  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 905, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_INCREF(__pyx_v_cx);
                  __Pyx_GIVEREF(__pyx_v_cx);
                  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_cx);
                  __Pyx_GIVEREF(__pyx_t_20);
                  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_20);
                  __Pyx_GIVEREF(__pyx_t_25);
                  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_25);
                  __pyx_t_20 = 0;
                  __pyx_t_25 = 0;

                  /* "gnocis/regions.pyx":907
 * 						plt.text(cx, m+ci,
 * 							'%.2f +/- %.2f'%(m, ci) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',             # <<<<<<<<<<<<<<
 * 							fontsize = fontsizeValues)
 * 			else:
 */
                  __pyx_t_25 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 907, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_25);
                  if (PyDict_SetItem(__pyx_t_25, __pyx_n_s_horizontalalignment, __pyx_n_u_center) < 0) __PYX_ERR(0, 907, __pyx_L7_error)
                  if (PyDict_SetItem(__pyx_t_25, __pyx_n_s_verticalalignment, __pyx_n_u_bottom) < 0) __PYX_ERR(0, 907, __pyx_L7_error)

                  /* "gnocis/regions.pyx":908
 * 							'%.2f +/- %.2f'%(m, ci) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 * 							fontsize = fontsizeValues)             # <<<<<<<<<<<<<<
 * 			else:
 * 				v = [
 */
                  if (PyDict_SetItem(__pyx_t_25, __pyx_n_s_fontsize, __pyx_v_fontsizeValues) < 0) __PYX_ERR(0, 907, __pyx_L7_error)

                  /* "gnocis/regions.pyx":905
 * 				if showValues:
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):
 * 						plt.text(cx, m+ci,             # <<<<<<<<<<<<<<
 * 							'%.2f +/- %.2f'%(m, ci) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 */
                  __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_3, __pyx_t_25); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 905, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_20);
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

                  /* "gnocis/regions.pyx":904
 * 				ax.bar(x, vMean, bw, label = ps.name, yerr = vCI)
 * 				if showValues:
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):             # <<<<<<<<<<<<<<
 * 						plt.text(cx, m+ci,
 * 							'%.2f +/- %.2f'%(m, ci) + (' %' if isPercent else ''),
 */
                }
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

                /* "gnocis/regions.pyx":903
 * 				]
 * 				ax.bar(x, vMean, bw, label = ps.name, yerr = vCI)
 * 				if showValues:             # <<<<<<<<<<<<<<
 * 					for cx, (m, ci) in zip(x, zip(vMean, vCI)):
 * 						plt.text(cx, m+ci,
 */
              }

              /* "gnocis/regions.pyx":884
 * 				for rsi in range(len(testRegions))
 * 			]
 * 			if isinstance(ps, CVModelPredictions):             # <<<<<<<<<<<<<<
 * 				# TODO Add multiclass support
 * 				prs = ps.regions(label = positive)
 */
              goto __pyx_L20;
            }

            /* "gnocis/regions.pyx":910
 * 							fontsize = fontsizeValues)
 * 			else:
 * 				v = [             # <<<<<<<<<<<<<<
 * 					scaling * measure(ps, rs) #100. * len(rs.overlap(ps)) / len(rs)
 * 					for rs in testRegions
 */
            /*else*/ {
              { /* enter inner scope */
                __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 910, __pyx_L54_error)
                __Pyx_GOTREF(__pyx_t_5);

                /* "gnocis/regions.pyx":912
 * 				v = [
 * 					scaling * measure(ps, rs) #100. * len(rs.overlap(ps)) / len(rs)
 * 					for rs in testRegions             # <<<<<<<<<<<<<<
 * 				]
 * 				ax.bar(x, v, bw, label = ps.name)
 */
                if (likely(PyList_CheckExact(__pyx_v_testRegions)) || PyTuple_CheckExact(__pyx_v_testRegions)) {
                  __pyx_t_20 = __pyx_v_testRegions; __Pyx_INCREF(__pyx_t_20); __pyx_t_10 = 0;
                  __pyx_t_21 = NULL;
                } else {
                  __pyx_t_10 = -1; __pyx_t_20 = PyObject_GetIter(__pyx_v_testRegions); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 912, __pyx_L54_error)
                  __Pyx_GOTREF(__pyx_t_20);
                  __pyx_t_21 = Py_TYPE(__pyx_t_20)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 912, __pyx_L54_error)
                }
                for (;;) {
                  if (likely(!__pyx_t_21)) {
                    if (likely(PyList_CheckExact(__pyx_t_20))) {
                      if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_20)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_25 = PyList_GET_ITEM(__pyx_t_20, __pyx_t_10); __Pyx_INCREF(__pyx_t_25); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 912, __pyx_L54_error)
                      #else
                      __pyx_t_25 = PySequence_ITEM(__pyx_t_20, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 912, __pyx_L54_error)
                      __Pyx_GOTREF(__pyx_t_25);
                      #endif
                    } else {
                      if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_20)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_25 = PyTuple_GET_ITEM(__pyx_t_20, __pyx_t_10); __Pyx_INCREF(__pyx_t_25); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 912, __pyx_L54_error)
                      #else
                      __pyx_t_25 = PySequence_ITEM(__pyx_t_20, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 912, __pyx_L54_error)
                      __Pyx_GOTREF(__pyx_t_25);
                      #endif
                    }
                  } else {
                    __pyx_t_25 = __pyx_t_21(__pyx_t_20);
                    if (unlikely(!__pyx_t_25)) {
                      PyObject* exc_type = PyErr_Occurred();
                      if (exc_type) {
                        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                        else __PYX_ERR(0, 912, __pyx_L54_error)
                      }
                      break;
                    }
                    __Pyx_GOTREF(__pyx_t_25);
                  }
                  __Pyx_XDECREF_SET(__pyx_9genexpr33__pyx_v_rs, __pyx_t_25);
                  __pyx_t_25 = 0;

                  /* "gnocis/regions.pyx":911
 * 			else:
 * 				v = [
 * 					scaling * measure(ps, rs) #100. * len(rs.overlap(ps)) / len(rs)             # <<<<<<<<<<<<<<
 * 					for rs in testRegions
 * 				]
 */
                  __pyx_t_25 = PyFloat_FromDouble(__pyx_v_scaling); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 911, __pyx_L54_error)
                  __Pyx_GOTREF(__pyx_t_25);
                  __Pyx_INCREF(__pyx_v_measure);
                  __pyx_t_19 = __pyx_v_measure; __pyx_t_24 = NULL;
                  __pyx_t_27 = 0;
                  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
                    __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_19);
                    if (likely(__pyx_t_24)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
                      __Pyx_INCREF(__pyx_t_24);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_19, function);
                      __pyx_t_27 = 1;
                    }
                  }
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_19)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_24, __pyx_v_ps, __pyx_9genexpr33__pyx_v_rs};
                    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_27, 2+__pyx_t_27); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 911, __pyx_L54_error)
                    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                    __Pyx_GOTREF(__pyx_t_3);
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_19)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_24, __pyx_v_ps, __pyx_9genexpr33__pyx_v_rs};
                    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_27, 2+__pyx_t_27); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 911, __pyx_L54_error)
                    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                    __Pyx_GOTREF(__pyx_t_3);
                  } else
                  #endif
                  {
                    __pyx_t_23 = PyTuple_New(2+__pyx_t_27); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 911, __pyx_L54_error)
                    __Pyx_GOTREF(__pyx_t_23);
                    if (__pyx_t_24) {
                      __Pyx_GIVEREF(__pyx_t_24); PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_24); __pyx_t_24 = NULL;
                    }
                    __Pyx_INCREF(__pyx_v_ps);
                    __Pyx_GIVEREF(__pyx_v_ps);
                    PyTuple_SET_ITEM(__pyx_t_23, 0+__pyx_t_27, __pyx_v_ps);
                    __Pyx_INCREF(__pyx_9genexpr33__pyx_v_rs);
                    __Pyx_GIVEREF(__pyx_9genexpr33__pyx_v_rs);
                    PyTuple_SET_ITEM(__pyx_t_23, 1+__pyx_t_27, __pyx_9genexpr33__pyx_v_rs);
                    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_23, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 911, __pyx_L54_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  __pyx_t_19 = PyNumber_Multiply(__pyx_t_25, __pyx_t_3); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 911, __pyx_L54_error)
                  __Pyx_GOTREF(__pyx_t_19);
                  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_19))) __PYX_ERR(0, 910, __pyx_L54_error)
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                  /* "gnocis/regions.pyx":912
 * 				v = [
 * 					scaling * measure(ps, rs) #100. * len(rs.overlap(ps)) / len(rs)
 * 					for rs in testRegions             # <<<<<<<<<<<<<<
 * 				]
 * 				ax.bar(x, v, bw, label = ps.name)
 */
                }
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_XDECREF(__pyx_9genexpr33__pyx_v_rs); __pyx_9genexpr33__pyx_v_rs = 0;
                goto __pyx_L57_exit_scope;
                __pyx_L54_error:;
                __Pyx_XDECREF(__pyx_9genexpr33__pyx_v_rs); __pyx_9genexpr33__pyx_v_rs = 0;
                goto __pyx_L7_error;
                __pyx_L57_exit_scope:;
              } /* exit inner scope */
              __Pyx_XDECREF_SET(__pyx_v_v, ((PyObject*)__pyx_t_5));
              __pyx_t_5 = 0;

              /* "gnocis/regions.pyx":914
 * 					for rs in testRegions
 * 				]
 * 				ax.bar(x, v, bw, label = ps.name)             # <<<<<<<<<<<<<<
 * 				if showValues:
 * 					for cx, cv in zip(x, v):
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_ax, __pyx_n_s_bar); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 914, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_20 = PyFloat_FromDouble(__pyx_v_bw); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 914, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_20);
              __pyx_t_19 = PyTuple_New(3); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 914, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_19);
              __Pyx_INCREF(__pyx_v_x);
              __Pyx_GIVEREF(__pyx_v_x);
              PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_v_x);
              __Pyx_INCREF(__pyx_v_v);
              __Pyx_GIVEREF(__pyx_v_v);
              PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_v_v);
              __Pyx_GIVEREF(__pyx_t_20);
              PyTuple_SET_ITEM(__pyx_t_19, 2, __pyx_t_20);
              __pyx_t_20 = 0;
              __pyx_t_20 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 914, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_20);
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ps, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 914, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_3);
              if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_label, __pyx_t_3) < 0) __PYX_ERR(0, 914, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_19, __pyx_t_20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 914, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "gnocis/regions.pyx":915
 * 				]
 * 				ax.bar(x, v, bw, label = ps.name)
 * 				if showValues:             # <<<<<<<<<<<<<<
 * 					for cx, cv in zip(x, v):
 * 						plt.text(cx, cv,
 */
              __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_showValues); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 915, __pyx_L7_error)
              if (__pyx_t_14) {

                /* "gnocis/regions.pyx":916
 * 				ax.bar(x, v, bw, label = ps.name)
 * 				if showValues:
 * 					for cx, cv in zip(x, v):             # <<<<<<<<<<<<<<
 * 						plt.text(cx, cv,
 * 							'%.2f'%(cv) + (' %' if isPercent else ''),
 */
                __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 916, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_INCREF(__pyx_v_x);
                __Pyx_GIVEREF(__pyx_v_x);
                PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_x);
                __Pyx_INCREF(__pyx_v_v);
                __Pyx_GIVEREF(__pyx_v_v);
                PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_v);
                __pyx_t_20 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 916, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                if (likely(PyList_CheckExact(__pyx_t_20)) || PyTuple_CheckExact(__pyx_t_20)) {
                  __pyx_t_3 = __pyx_t_20; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
                  __pyx_t_21 = NULL;
                } else {
                  __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 916, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __pyx_t_21 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 916, __pyx_L7_error)
                }
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                for (;;) {
                  if (likely(!__pyx_t_21)) {
                    if (likely(PyList_CheckExact(__pyx_t_3))) {
                      if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_20 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_20); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 916, __pyx_L7_error)
                      #else
                      __pyx_t_20 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 916, __pyx_L7_error)
                      __Pyx_GOTREF(__pyx_t_20);
                      #endif
                    } else {
                      if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_20 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_20); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 916, __pyx_L7_error)
                      #else
                      __pyx_t_20 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 916, __pyx_L7_error)
                      __Pyx_GOTREF(__pyx_t_20);
                      #endif
                    }
                  } else {
                    __pyx_t_20 = __pyx_t_21(__pyx_t_3);
                    if (unlikely(!__pyx_t_20)) {
                      PyObject* exc_type = PyErr_Occurred();
                      if (exc_type) {
                        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                        else __PYX_ERR(0, 916, __pyx_L7_error)
                      }
                      break;
                    }
                    __Pyx_GOTREF(__pyx_t_20);
                  }
                  if ((likely(PyTuple_CheckExact(__pyx_t_20))) || (PyList_CheckExact(__pyx_t_20))) {
                    PyObject* sequence = __pyx_t_20;
                    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                    if (unlikely(size != 2)) {
                      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                      __PYX_ERR(0, 916, __pyx_L7_error)
                    }
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    if (likely(PyTuple_CheckExact(sequence))) {
                      __pyx_t_19 = PyTuple_GET_ITEM(sequence, 0); 
                      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
                    } else {
                      __pyx_t_19 = PyList_GET_ITEM(sequence, 0); 
                      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
                    }
                    __Pyx_INCREF(__pyx_t_19);
                    __Pyx_INCREF(__pyx_t_5);
                    #else
                    __pyx_t_19 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 916, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 916, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    #endif
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  } else {
                    Py_ssize_t index = -1;
                    __pyx_t_25 = PyObject_GetIter(__pyx_t_20); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 916, __pyx_L7_error)
                    __Pyx_GOTREF(__pyx_t_25);
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                    __pyx_t_12 = Py_TYPE(__pyx_t_25)->tp_iternext;
                    index = 0; __pyx_t_19 = __pyx_t_12(__pyx_t_25); if (unlikely(!__pyx_t_19)) goto __pyx_L61_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_19);
                    index = 1; __pyx_t_5 = __pyx_t_12(__pyx_t_25); if (unlikely(!__pyx_t_5)) goto __pyx_L61_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_5);
                    if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_25), 2) < 0) __PYX_ERR(0, 916, __pyx_L7_error)
                    __pyx_t_12 = NULL;
                    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                    goto __pyx_L62_unpacking_done;
                    __pyx_L61_unpacking_failed:;
                    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                    __pyx_t_12 = NULL;
                    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                    __PYX_ERR(0, 916, __pyx_L7_error)
                    __pyx_L62_unpacking_done:;
                  }
                  __Pyx_XDECREF_SET(__pyx_v_cx, __pyx_t_19);
                  __pyx_t_19 = 0;
                  __Pyx_XDECREF_SET(__pyx_v_cv, __pyx_t_5);
                  __pyx_t_5 = 0;

                  /* "gnocis/regions.pyx":917
 * 				if showValues:
 * 					for cx, cv in zip(x, v):
 * 						plt.text(cx, cv,             # <<<<<<<<<<<<<<
 * 							'%.2f'%(cv) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 */
                  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_text); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 917, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_20);

                  /* "gnocis/regions.pyx":918
 * 					for cx, cv in zip(x, v):
 * 						plt.text(cx, cv,
 * 							'%.2f'%(cv) + (' %' if isPercent else ''),             # <<<<<<<<<<<<<<
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 * 							fontsize = fontsizeValues)
 */
                  __pyx_t_5 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_2f_2, __pyx_v_cv); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 918, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_isPercent); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 918, __pyx_L7_error)
                  if (__pyx_t_14) {
                    __Pyx_INCREF(__pyx_kp_u__29);
                    __pyx_t_19 = __pyx_kp_u__29;
                  } else {
                    __Pyx_INCREF(__pyx_kp_u__13);
                    __pyx_t_19 = __pyx_kp_u__13;
                  }
                  __pyx_t_25 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_5, __pyx_t_19); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 918, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_25);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                  /* "gnocis/regions.pyx":917
 * 				if showValues:
 * 					for cx, cv in zip(x, v):
 * 						plt.text(cx, cv,             # <<<<<<<<<<<<<<
 * 							'%.2f'%(cv) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 */
                  __pyx_t_19 = PyTuple_New(3); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 917, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_19);
                  __Pyx_INCREF(__pyx_v_cx);
                  __Pyx_GIVEREF(__pyx_v_cx);
                  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_v_cx);
                  __Pyx_INCREF(__pyx_v_cv);
                  __Pyx_GIVEREF(__pyx_v_cv);
                  PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_v_cv);
                  __Pyx_GIVEREF(__pyx_t_25);
                  PyTuple_SET_ITEM(__pyx_t_19, 2, __pyx_t_25);
                  __pyx_t_25 = 0;

                  /* "gnocis/regions.pyx":919
 * 						plt.text(cx, cv,
 * 							'%.2f'%(cv) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',             # <<<<<<<<<<<<<<
 * 							fontsize = fontsizeValues)
 * 		ax.set_ylabel(measureName, fontsize = fontsizeLabels)
 */
                  __pyx_t_25 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 919, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_25);
                  if (PyDict_SetItem(__pyx_t_25, __pyx_n_s_horizontalalignment, __pyx_n_u_center) < 0) __PYX_ERR(0, 919, __pyx_L7_error)
                  if (PyDict_SetItem(__pyx_t_25, __pyx_n_s_verticalalignment, __pyx_n_u_bottom) < 0) __PYX_ERR(0, 919, __pyx_L7_error)

                  /* "gnocis/regions.pyx":920
 * 							'%.2f'%(cv) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 * 							fontsize = fontsizeValues)             # <<<<<<<<<<<<<<
 * 		ax.set_ylabel(measureName, fontsize = fontsizeLabels)
 * 		ax.set_xticks([ float(i) for i in range(len(testRegions)) ])
 */
                  if (PyDict_SetItem(__pyx_t_25, __pyx_n_s_fontsize, __pyx_v_fontsizeValues) < 0) __PYX_ERR(0, 919, __pyx_L7_error)

                  /* "gnocis/regions.pyx":917
 * 				if showValues:
 * 					for cx, cv in zip(x, v):
 * 						plt.text(cx, cv,             # <<<<<<<<<<<<<<
 * 							'%.2f'%(cv) + (' %' if isPercent else ''),
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 */
                  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_t_19, __pyx_t_25); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L7_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

                  /* "gnocis/regions.pyx":916
 * 				ax.bar(x, v, bw, label = ps.name)
 * 				if showValues:
 * 					for cx, cv in zip(x, v):             # <<<<<<<<<<<<<<
 * 						plt.text(cx, cv,
 * 							'%.2f'%(cv) + (' %' if isPercent else ''),
 */
                }
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

                /* "gnocis/regions.pyx":915
 * 				]
 * 				ax.bar(x, v, bw, label = ps.name)
 * 				if showValues:             # <<<<<<<<<<<<<<
 * 					for cx, cv in zip(x, v):
 * 						plt.text(cx, cv,
 */
              }
            }
            __pyx_L20:;

            /* "gnocis/regions.pyx":879
 * 		if testRegions is None:
 * 			testRegions = [ 0 ]
 * 		for psi, ps in enumerate(mainRegions):             # <<<<<<<<<<<<<<
 * 			x = [
 * 				float(rsi) - (width/2) + (bw*psi)
 */
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "gnocis/regions.pyx":921
 * 							horizontalalignment = 'center', verticalalignment = 'bottom',
 * 							fontsize = fontsizeValues)
 * 		ax.set_ylabel(measureName, fontsize = fontsizeLabels)             # <<<<<<<<<<<<<<
 * 		ax.set_xticks([ float(i) for i in range(len(testRegions)) ])
 * 		ax.set_xticklabels([ rs.name if hasTestRegions else '' for rs in testRegions ])
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ax, __pyx_n_s_set_ylabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 921, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_measureName);
          __Pyx_GIVEREF(__pyx_v_measureName);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_measureName);
          __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 921, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_fontsize, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 921, __pyx_L7_error)
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 921, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":922
 * 							fontsize = fontsizeValues)
 * 		ax.set_ylabel(measureName, fontsize = fontsizeLabels)
 * 		ax.set_xticks([ float(i) for i in range(len(testRegions)) ])             # <<<<<<<<<<<<<<
 * 		ax.set_xticklabels([ rs.name if hasTestRegions else '' for rs in testRegions ])
 * 		plt.xticks(fontsize = fontsizeLabels, rotation = 0)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ax, __pyx_n_s_set_xticks); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 922, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          { /* enter inner scope */
            __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_11 = PyObject_Length(__pyx_v_testRegions); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 922, __pyx_L7_error)
            __pyx_t_10 = __pyx_t_11;
            for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_10; __pyx_t_17+=1) {
              __pyx_9genexpr34__pyx_v_i = __pyx_t_17;
              __pyx_t_2 = PyFloat_FromDouble(((double)__pyx_9genexpr34__pyx_v_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 922, __pyx_L7_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
          } /* exit inner scope */
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 922, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":923
 * 		ax.set_ylabel(measureName, fontsize = fontsizeLabels)
 * 		ax.set_xticks([ float(i) for i in range(len(testRegions)) ])
 * 		ax.set_xticklabels([ rs.name if hasTestRegions else '' for rs in testRegions ])             # <<<<<<<<<<<<<<
 * 		plt.xticks(fontsize = fontsizeLabels, rotation = 0)
 * 		plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ax, __pyx_n_s_set_xticklabels); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 923, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          { /* enter inner scope */
            __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 923, __pyx_L67_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (likely(PyList_CheckExact(__pyx_v_testRegions)) || PyTuple_CheckExact(__pyx_v_testRegions)) {
              __pyx_t_2 = __pyx_v_testRegions; __Pyx_INCREF(__pyx_t_2); __pyx_t_11 = 0;
              __pyx_t_16 = NULL;
            } else {
              __pyx_t_11 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_testRegions); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L67_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_16 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 923, __pyx_L67_error)
            }
            for (;;) {
              if (likely(!__pyx_t_16)) {
                if (likely(PyList_CheckExact(__pyx_t_2))) {
                  if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_25 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_11); __Pyx_INCREF(__pyx_t_25); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 923, __pyx_L67_error)
                  #else
                  __pyx_t_25 = PySequence_ITEM(__pyx_t_2, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 923, __pyx_L67_error)
                  __Pyx_GOTREF(__pyx_t_25);
                  #endif
                } else {
                  if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_25 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_11); __Pyx_INCREF(__pyx_t_25); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 923, __pyx_L67_error)
                  #else
                  __pyx_t_25 = PySequence_ITEM(__pyx_t_2, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 923, __pyx_L67_error)
                  __Pyx_GOTREF(__pyx_t_25);
                  #endif
                }
              } else {
                __pyx_t_25 = __pyx_t_16(__pyx_t_2);
                if (unlikely(!__pyx_t_25)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 923, __pyx_L67_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_25);
              }
              __Pyx_XDECREF_SET(__pyx_9genexpr35__pyx_v_rs, __pyx_t_25);
              __pyx_t_25 = 0;
              __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_hasTestRegions); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 923, __pyx_L67_error)
              if (__pyx_t_14) {
                __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr35__pyx_v_rs, __pyx_n_s_name); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 923, __pyx_L67_error)
                __Pyx_GOTREF(__pyx_t_19);
                __pyx_t_25 = __pyx_t_19;
                __pyx_t_19 = 0;
              } else {
                __Pyx_INCREF(__pyx_kp_u__13);
                __pyx_t_25 = __pyx_kp_u__13;
              }
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_25))) __PYX_ERR(0, 923, __pyx_L67_error)
              __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_9genexpr35__pyx_v_rs); __pyx_9genexpr35__pyx_v_rs = 0;
            goto __pyx_L70_exit_scope;
            __pyx_L67_error:;
            __Pyx_XDECREF(__pyx_9genexpr35__pyx_v_rs); __pyx_9genexpr35__pyx_v_rs = 0;
            goto __pyx_L7_error;
            __pyx_L70_exit_scope:;
          } /* exit inner scope */
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 923, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":924
 * 		ax.set_xticks([ float(i) for i in range(len(testRegions)) ])
 * 		ax.set_xticklabels([ rs.name if hasTestRegions else '' for rs in testRegions ])
 * 		plt.xticks(fontsize = fontsizeLabels, rotation = 0)             # <<<<<<<<<<<<<<
 * 		plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 		if isPercent:
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_xticks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 924, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 924, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_fontsize, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 924, __pyx_L7_error)
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_rotation, __pyx_int_0) < 0) __PYX_ERR(0, 924, __pyx_L7_error)
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "gnocis/regions.pyx":925
 * 		ax.set_xticklabels([ rs.name if hasTestRegions else '' for rs in testRegions ])
 * 		plt.xticks(fontsize = fontsizeLabels, rotation = 0)
 * 		plt.yticks(fontsize = fontsizeAxis, rotation = 0)             # <<<<<<<<<<<<<<
 * 		if isPercent:
 * 			ax.yaxis.set_major_formatter(mtick.PercentFormatter())
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_yticks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 925, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_fontsize, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 925, __pyx_L7_error)
          if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_rotation, __pyx_int_0) < 0) __PYX_ERR(0, 925, __pyx_L7_error)
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 925, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "gnocis/regions.pyx":926
 * 		plt.xticks(fontsize = fontsizeLabels, rotation = 0)
 * 		plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 		if isPercent:             # <<<<<<<<<<<<<<
 * 			ax.yaxis.set_major_formatter(mtick.PercentFormatter())
 * 		if showLegend:
 */
          __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_isPercent); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 926, __pyx_L7_error)
          if (__pyx_t_14) {

            /* "gnocis/regions.pyx":927
 * 		plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 		if isPercent:
 * 			ax.yaxis.set_major_formatter(mtick.PercentFormatter())             # <<<<<<<<<<<<<<
 * 		if showLegend:
 * 			ax.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ax, __pyx_n_s_yaxis); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 927, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_set_major_formatter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 927, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_mtick, __pyx_n_s_PercentFormatter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 927, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_25 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_25)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_25);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
              }
            }
            __pyx_t_3 = (__pyx_t_25) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_25) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
            __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 927, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 927, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "gnocis/regions.pyx":926
 * 		plt.xticks(fontsize = fontsizeLabels, rotation = 0)
 * 		plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 		if isPercent:             # <<<<<<<<<<<<<<
 * 			ax.yaxis.set_major_formatter(mtick.PercentFormatter())
 * 		if showLegend:
 */
          }

          /* "gnocis/regions.pyx":928
 * 		if isPercent:
 * 			ax.yaxis.set_major_formatter(mtick.PercentFormatter())
 * 		if showLegend:             # <<<<<<<<<<<<<<
 * 			ax.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 		fig.tight_layout()
 */
          __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_showLegend); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 928, __pyx_L7_error)
          if (__pyx_t_14) {

            /* "gnocis/regions.pyx":929
 * 			ax.yaxis.set_major_formatter(mtick.PercentFormatter())
 * 		if showLegend:
 * 			ax.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)             # <<<<<<<<<<<<<<
 * 		fig.tight_layout()
 * 		if outpath is None:
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_ax, __pyx_n_s_legend); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 929, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_bbox_to_anchor, __pyx_v_bboxAnchorTo) < 0) __PYX_ERR(0, 929, __pyx_L7_error)
            if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_loc, __pyx_v_legendLoc) < 0) __PYX_ERR(0, 929, __pyx_L7_error)
            if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_fontsize, __pyx_v_fontsizeLegend) < 0) __PYX_ERR(0, 929, __pyx_L7_error)
            if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_fancybox, Py_True) < 0) __PYX_ERR(0, 929, __pyx_L7_error)
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 929, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "gnocis/regions.pyx":928
 * 		if isPercent:
 * 			ax.yaxis.set_major_formatter(mtick.PercentFormatter())
 * 		if showLegend:             # <<<<<<<<<<<<<<
 * 			ax.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 		fig.tight_layout()
 */
          }

          /* "gnocis/regions.pyx":930
 * 		if showLegend:
 * 			ax.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 		fig.tight_layout()             # <<<<<<<<<<<<<<
 * 		if outpath is None:
 * 			bio = BytesIO()
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_tight_layout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 930, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 930, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "gnocis/regions.pyx":931
 * 			ax.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 		fig.tight_layout()
 * 		if outpath is None:             # <<<<<<<<<<<<<<
 * 			bio = BytesIO()
 * 			fig.savefig(bio, format='png')
 */
          __pyx_t_14 = (__pyx_v_outpath == Py_None);
          __pyx_t_15 = (__pyx_t_14 != 0);
          if (__pyx_t_15) {

            /* "gnocis/regions.pyx":932
 * 		fig.tight_layout()
 * 		if outpath is None:
 * 			bio = BytesIO()             # <<<<<<<<<<<<<<
 * 			fig.savefig(bio, format='png')
 * 			plt.close('all')
 */
            __Pyx_INCREF(__pyx_v_BytesIO);
            __pyx_t_1 = __pyx_v_BytesIO; __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 932, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_v_bio = __pyx_t_3;
            __pyx_t_3 = 0;

            /* "gnocis/regions.pyx":933
 * 		if outpath is None:
 * 			bio = BytesIO()
 * 			fig.savefig(bio, format='png')             # <<<<<<<<<<<<<<
 * 			plt.close('all')
 * 			encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_savefig); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_INCREF(__pyx_v_bio);
            __Pyx_GIVEREF(__pyx_v_bio);
            PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_bio);
            __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 933, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_format, __pyx_n_u_png) < 0) __PYX_ERR(0, 933, __pyx_L7_error)
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":934
 * 			bio = BytesIO()
 * 			fig.savefig(bio, format='png')
 * 			plt.close('all')             # <<<<<<<<<<<<<<
 * 			encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 			html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_close); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 934, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_n_u_all) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_u_all);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 934, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":935
 * 			fig.savefig(bio, format='png')
 * 			plt.close('all')
 * 			encoded = base64.b64encode(bio.getvalue()).decode('utf-8')             # <<<<<<<<<<<<<<
 * 			html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 			if returnHTML:
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_base64, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 935, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_v_bio, __pyx_n_s_getvalue); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 935, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_25);
            __pyx_t_19 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_25))) {
              __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_25);
              if (likely(__pyx_t_19)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_25);
                __Pyx_INCREF(__pyx_t_19);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_25, function);
              }
            }
            __pyx_t_3 = (__pyx_t_19) ? __Pyx_PyObject_CallOneArg(__pyx_t_25, __pyx_t_19) : __Pyx_PyObject_CallNoArg(__pyx_t_25);
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            __pyx_t_25 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_25)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_25);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_5 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_25, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
            __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 935, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 935, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u_utf_8);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 935, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_v_encoded = __pyx_t_2;
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":936
 * 			plt.close('all')
 * 			encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 			html = '<img src=\'data:image/png;base64,%s\'>'%encoded             # <<<<<<<<<<<<<<
 * 			if returnHTML:
 * 				return html
 */
            __pyx_t_2 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_img_src_data_image_png_base64_s, __pyx_v_encoded); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 936, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_v_html = ((PyObject*)__pyx_t_2);
            __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":937
 * 			encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 			html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 			if returnHTML:             # <<<<<<<<<<<<<<
 * 				return html
 * 			display(HTML(html))
 */
            __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_returnHTML); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 937, __pyx_L7_error)
            if (__pyx_t_15) {

              /* "gnocis/regions.pyx":938
 * 			html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 			if returnHTML:
 * 				return html             # <<<<<<<<<<<<<<
 * 			display(HTML(html))
 * 		else:
 */
              __Pyx_XDECREF(__pyx_r);
              __Pyx_INCREF(__pyx_v_html);
              __pyx_r = __pyx_v_html;
              goto __pyx_L11_try_return;

              /* "gnocis/regions.pyx":937
 * 			encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 			html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 			if returnHTML:             # <<<<<<<<<<<<<<
 * 				return html
 * 			display(HTML(html))
 */
            }

            /* "gnocis/regions.pyx":939
 * 			if returnHTML:
 * 				return html
 * 			display(HTML(html))             # <<<<<<<<<<<<<<
 * 		else:
 * 			fig.savefig(outpath)
 */
            __Pyx_INCREF(__pyx_v_HTML);
            __pyx_t_5 = __pyx_v_HTML; __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_v_html) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_html);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_INCREF(__pyx_v_display);
            __pyx_t_5 = __pyx_v_display; __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":931
 * 			ax.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 		fig.tight_layout()
 * 		if outpath is None:             # <<<<<<<<<<<<<<
 * 			bio = BytesIO()
 * 			fig.savefig(bio, format='png')
 */
            goto __pyx_L73;
          }

          /* "gnocis/regions.pyx":941
 * 			display(HTML(html))
 * 		else:
 * 			fig.savefig(outpath)             # <<<<<<<<<<<<<<
 * 			plt.close('all')
 * 
 */
          /*else*/ {
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_savefig); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 941, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_v_outpath) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_outpath);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 941, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "gnocis/regions.pyx":942
 * 		else:
 * 			fig.savefig(outpath)
 * 			plt.close('all')             # <<<<<<<<<<<<<<
 * 
 * def predictionBarplot(predictionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_close); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 942, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_n_u_all) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_u_all);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 942, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __pyx_L73:;

          /* "gnocis/regions.pyx":871
 * 	from IPython.core.display import display, HTML
 * 	import matplotlib.ticker as mtick
 * 	with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 		width = 0.5
 * 		bw = width / max(len(mainRegions)-1, 1)
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
        __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
        __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
        __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
        __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("gnocis.regions.regionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(0, 871, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_30 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 871, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_30);
          __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_30);
          __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
          if (__pyx_t_15 < 0) __PYX_ERR(0, 871, __pyx_L9_except_error)
          __pyx_t_14 = ((!(__pyx_t_15 != 0)) != 0);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_XGIVEREF(__pyx_t_1);
            __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_5, __pyx_t_1);
            __pyx_t_2 = 0; __pyx_t_5 = 0; __pyx_t_1 = 0; 
            __PYX_ERR(0, 871, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L11_try_return:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L4_return;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_4) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 871, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L4_return: {
        __pyx_t_8 = __pyx_r;
        __pyx_r = 0;
        if (__pyx_t_4) {
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__20, NULL);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 871, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __pyx_r = __pyx_t_8;
        __pyx_t_8 = 0;
        goto __pyx_L0;
      }
      __pyx_L6:;
    }
    goto __pyx_L78;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L1_error;
    __pyx_L78:;
  }

  /* "gnocis/regions.pyx":829
 * 
 * # Generates a barplot for a region-based measure
 * def regionBarplot(mainRegions, measure, measureName, testRegions = None, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeValues = 12, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = False, isPercent = False):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of a measure (lambda) for predicted regions.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_AddTraceback("gnocis.regions.regionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_plt);
  __Pyx_XDECREF(__pyx_v_base64);
  __Pyx_XDECREF(__pyx_v_BytesIO);
  __Pyx_XDECREF(__pyx_v_display);
  __Pyx_XDECREF(__pyx_v_HTML);
  __Pyx_XDECREF(__pyx_v_mtick);
  __Pyx_XDECREF(__pyx_v_fig);
  __Pyx_XDECREF(__pyx_v_ax);
  __Pyx_XDECREF(__pyx_v_hasTestRegions);
  __Pyx_XDECREF(__pyx_v_psi);
  __Pyx_XDECREF(__pyx_v_ps);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_prs);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_vMean);
  __Pyx_XDECREF(__pyx_v_vCI);
  __Pyx_XDECREF(__pyx_v_cx);
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XDECREF(__pyx_v_ci);
  __Pyx_XDECREF(__pyx_v_cv);
  __Pyx_XDECREF(__pyx_v_bio);
  __Pyx_XDECREF(__pyx_v_encoded);
  __Pyx_XDECREF(__pyx_v_html);
  __Pyx_XDECREF(__pyx_9genexpr29__pyx_v_rs);
  __Pyx_XDECREF(__pyx_9genexpr30__pyx_v__ps);
  __Pyx_XDECREF(__pyx_9genexpr31__pyx_v_cv);
  __Pyx_XDECREF(__pyx_9genexpr32__pyx_v_cv);
  __Pyx_XDECREF(__pyx_9genexpr33__pyx_v_rs);
  __Pyx_XDECREF(__pyx_9genexpr35__pyx_v_rs);
  __Pyx_XDECREF(__pyx_v_testRegions);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":944
 * 			plt.close('all')
 * 
 * def predictionBarplot(predictionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_13predictionBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_12predictionBarplot[] = " Generates a prediction overlap precision barplot.\n\t\n\t:param predictionSets: List of prediction region sets.\n\t:param figsize: Figure size.\n\t:param outpath: Output path.\n\t:param returnHTML: If True, an HTML node will be returned.\n\t:param fontsizeLabels: Size of label font.\n\t:param fontsizeLegend: Size of legend font.\n\t:param fontsizeAxis: Size of axis font.\n\t:param style: Plot style to use.\n\t:param showLegend: Flag for whether or not to render legend.\n\t:param bboxAnchorTo: Legend anchor point.\n\t:param legendLoc: Legend location.\n\t:param showValues: If True, values will be plotted.\n\t\n\t:type predictionSets: list\n\t:type figsize: tuple, optional\n\t:type outpath: str, optional\n\t:type returnHTML: bool, optional\n\t:type fontsizeLabels: float, optional\n\t:type fontsizeLegend: float, optional\n\t:type fontsizeAxis: float, optional\n\t:type style: str, optional\n\t:type showLegend: bool, optional\n\t:type bboxAnchorTo: tuple, optional\n\t:type legendLoc: str, optional\n\t:type showValues: bool, optional\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_13predictionBarplot = {"predictionBarplot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_13predictionBarplot, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_12predictionBarplot};
static PyObject *__pyx_pw_6gnocis_7regions_13predictionBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_predictionSets = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_returnHTML = 0;
  PyObject *__pyx_v_fontsizeLabels = 0;
  PyObject *__pyx_v_fontsizeLegend = 0;
  PyObject *__pyx_v_fontsizeAxis = 0;
  PyObject *__pyx_v_style = 0;
  PyObject *__pyx_v_showLegend = 0;
  PyObject *__pyx_v_bboxAnchorTo = 0;
  PyObject *__pyx_v_legendLoc = 0;
  PyObject *__pyx_v_showValues = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predictionBarplot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_predictionSets,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_returnHTML,&__pyx_n_s_fontsizeLabels,&__pyx_n_s_fontsizeLegend,&__pyx_n_s_fontsizeAxis,&__pyx_n_s_style,&__pyx_n_s_showLegend,&__pyx_n_s_bboxAnchorTo,&__pyx_n_s_legendLoc,&__pyx_n_s_showValues,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    values[1] = ((PyObject *)__pyx_tuple__25);
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)Py_False);
    values[4] = ((PyObject *)__pyx_int_18);
    values[5] = ((PyObject *)__pyx_int_12);
    values[6] = ((PyObject *)__pyx_int_10);
    values[7] = ((PyObject *)__pyx_n_u_ggplot);
    values[8] = ((PyObject *)Py_True);
    values[9] = ((PyObject *)__pyx_tuple__26);
    values[10] = ((PyObject *)__pyx_kp_u_upper_left);
    values[11] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_predictionSets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_returnHTML);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLabels);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLegend);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeAxis);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showLegend);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bboxAnchorTo);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_legendLoc);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showValues);
          if (value) { values[11] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "predictionBarplot") < 0)) __PYX_ERR(0, 944, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_predictionSets = values[0];
    __pyx_v_figsize = values[1];
    __pyx_v_outpath = values[2];
    __pyx_v_returnHTML = values[3];
    __pyx_v_fontsizeLabels = values[4];
    __pyx_v_fontsizeLegend = values[5];
    __pyx_v_fontsizeAxis = values[6];
    __pyx_v_style = values[7];
    __pyx_v_showLegend = values[8];
    __pyx_v_bboxAnchorTo = values[9];
    __pyx_v_legendLoc = values[10];
    __pyx_v_showValues = values[11];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("predictionBarplot", 0, 1, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 944, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.predictionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_12predictionBarplot(__pyx_self, __pyx_v_predictionSets, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_returnHTML, __pyx_v_fontsizeLabels, __pyx_v_fontsizeLegend, __pyx_v_fontsizeAxis, __pyx_v_style, __pyx_v_showLegend, __pyx_v_bboxAnchorTo, __pyx_v_legendLoc, __pyx_v_showValues);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":973
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, measure = (lambda rs, _: len(rs)), measureName = 'Predictions', testRegions = None, figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)             # <<<<<<<<<<<<<<
 * 
 * def overlapSensitivityBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_17predictionBarplot_lambda14(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_17predictionBarplot_lambda14 = {"lambda14", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_17predictionBarplot_lambda14, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_7regions_17predictionBarplot_lambda14(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_rs = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda14 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rs,&__pyx_n_s__30,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__30)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lambda14", 1, 2, 2, 1); __PYX_ERR(0, 973, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lambda14") < 0)) __PYX_ERR(0, 973, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_rs = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda14", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 973, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.predictionBarplot.lambda14", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda14(__pyx_self, __pyx_v_rs, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda14(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_rs, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda14", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_Length(__pyx_v_rs); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 973, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.regions.predictionBarplot.lambda14", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":944
 * 			plt.close('all')
 * 
 * def predictionBarplot(predictionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_12predictionBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("predictionBarplot", 0);

  /* "gnocis/regions.pyx":973
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, measure = (lambda rs, _: len(rs)), measureName = 'Predictions', testRegions = None, figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)             # <<<<<<<<<<<<<<
 * 
 * def overlapSensitivityBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_regionBarplot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mainRegions, __pyx_v_predictionSets) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_7regions_17predictionBarplot_lambda14, 0, __pyx_n_s_predictionBarplot_locals_lambda, NULL, __pyx_n_s_gnocis_regions, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measure, __pyx_t_3) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measureName, __pyx_n_u_Predictions) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_testRegions, Py_None) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_outpath, __pyx_v_outpath) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_returnHTML, __pyx_v_returnHTML) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLabels, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLegend, __pyx_v_fontsizeLegend) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeAxis, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_style, __pyx_v_style) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showLegend, __pyx_v_showLegend) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bboxAnchorTo, __pyx_v_bboxAnchorTo) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_legendLoc, __pyx_v_legendLoc) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showValues, __pyx_v_showValues) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_isPercent, Py_False) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/regions.pyx":944
 * 			plt.close('all')
 * 
 * def predictionBarplot(predictionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.predictionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":975
 * 	regionBarplot(mainRegions = predictionSets, measure = (lambda rs, _: len(rs)), measureName = 'Predictions', testRegions = None, figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)
 * 
 * def overlapSensitivityBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap sensitivity barplot.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_15overlapSensitivityBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_14overlapSensitivityBarplot[] = " Generates a prediction overlap sensitivity barplot.\n\t\n\t:param predictionSets: List of prediction region sets.\n\t:param regionSets: List of validation region sets.\n\t:param figsize: Figure size.\n\t:param outpath: Output path.\n\t:param returnHTML: If True, an HTML node will be returned.\n\t:param fontsizeLabels: Size of label font.\n\t:param fontsizeLegend: Size of legend font.\n\t:param fontsizeAxis: Size of axis font.\n\t:param style: Plot style to use.\n\t:param showLegend: Flag for whether or not to render legend.\n\t:param bboxAnchorTo: Legend anchor point.\n\t:param legendLoc: Legend location.\n\t:param showValues: If True, values will be plotted.\n\t\n\t:type predictionSets: list\n\t:type regionSets: list\n\t:type figsize: tuple, optional\n\t:type outpath: str, optional\n\t:type returnHTML: bool, optional\n\t:type fontsizeLabels: float, optional\n\t:type fontsizeLegend: float, optional\n\t:type fontsizeAxis: float, optional\n\t:type style: str, optional\n\t:type showLegend: bool, optional\n\t:type bboxAnchorTo: tuple, optional\n\t:type legendLoc: str, optional\n\t:type showValues: bool, optional\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_15overlapSensitivityBarplot = {"overlapSensitivityBarplot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_15overlapSensitivityBarplot, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_14overlapSensitivityBarplot};
static PyObject *__pyx_pw_6gnocis_7regions_15overlapSensitivityBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_predictionSets = 0;
  PyObject *__pyx_v_regionSets = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_returnHTML = 0;
  PyObject *__pyx_v_fontsizeLabels = 0;
  PyObject *__pyx_v_fontsizeLegend = 0;
  PyObject *__pyx_v_fontsizeAxis = 0;
  PyObject *__pyx_v_style = 0;
  PyObject *__pyx_v_showLegend = 0;
  PyObject *__pyx_v_bboxAnchorTo = 0;
  PyObject *__pyx_v_legendLoc = 0;
  PyObject *__pyx_v_showValues = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlapSensitivityBarplot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_predictionSets,&__pyx_n_s_regionSets,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_returnHTML,&__pyx_n_s_fontsizeLabels,&__pyx_n_s_fontsizeLegend,&__pyx_n_s_fontsizeAxis,&__pyx_n_s_style,&__pyx_n_s_showLegend,&__pyx_n_s_bboxAnchorTo,&__pyx_n_s_legendLoc,&__pyx_n_s_showValues,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)__pyx_tuple__25);
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_False);
    values[5] = ((PyObject *)__pyx_int_18);
    values[6] = ((PyObject *)__pyx_int_12);
    values[7] = ((PyObject *)__pyx_int_10);
    values[8] = ((PyObject *)__pyx_n_u_ggplot);
    values[9] = ((PyObject *)Py_True);
    values[10] = ((PyObject *)__pyx_tuple__26);
    values[11] = ((PyObject *)__pyx_kp_u_upper_left);
    values[12] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_predictionSets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_regionSets)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("overlapSensitivityBarplot", 0, 2, 13, 1); __PYX_ERR(0, 975, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_returnHTML);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLabels);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLegend);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeAxis);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showLegend);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bboxAnchorTo);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_legendLoc);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showValues);
          if (value) { values[12] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "overlapSensitivityBarplot") < 0)) __PYX_ERR(0, 975, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_predictionSets = values[0];
    __pyx_v_regionSets = values[1];
    __pyx_v_figsize = values[2];
    __pyx_v_outpath = values[3];
    __pyx_v_returnHTML = values[4];
    __pyx_v_fontsizeLabels = values[5];
    __pyx_v_fontsizeLegend = values[6];
    __pyx_v_fontsizeAxis = values[7];
    __pyx_v_style = values[8];
    __pyx_v_showLegend = values[9];
    __pyx_v_bboxAnchorTo = values[10];
    __pyx_v_legendLoc = values[11];
    __pyx_v_showValues = values[12];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("overlapSensitivityBarplot", 0, 2, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 975, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.overlapSensitivityBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_14overlapSensitivityBarplot(__pyx_self, __pyx_v_predictionSets, __pyx_v_regionSets, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_returnHTML, __pyx_v_fontsizeLabels, __pyx_v_fontsizeLegend, __pyx_v_fontsizeAxis, __pyx_v_style, __pyx_v_showLegend, __pyx_v_bboxAnchorTo, __pyx_v_legendLoc, __pyx_v_showValues);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1006
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(rs.overlap(ps)) / len(rs)), measureName = 'Overlap sensitivity', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)             # <<<<<<<<<<<<<<
 * 
 * def overlapPrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_25overlapSensitivityBarplot_lambda15(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_25overlapSensitivityBarplot_lambda15 = {"lambda15", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_25overlapSensitivityBarplot_lambda15, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_7regions_25overlapSensitivityBarplot_lambda15(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ps = 0;
  PyObject *__pyx_v_rs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda15 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ps,&__pyx_n_s_rs,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ps)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lambda15", 1, 2, 2, 1); __PYX_ERR(0, 1006, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lambda15") < 0)) __PYX_ERR(0, 1006, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_ps = values[0];
    __pyx_v_rs = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda15", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1006, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.overlapSensitivityBarplot.lambda15", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda15(__pyx_self, __pyx_v_ps, __pyx_v_rs);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda15(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ps, PyObject *__pyx_v_rs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda15", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_rs, __pyx_n_s_overlap); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_ps) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_ps);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = PyObject_Length(__pyx_v_rs); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (unlikely(__pyx_t_5 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 1006, __pyx_L1_error)
  }
  __pyx_t_1 = PyFloat_FromDouble((((double)__pyx_t_4) / ((double)__pyx_t_5))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.overlapSensitivityBarplot.lambda15", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":975
 * 	regionBarplot(mainRegions = predictionSets, measure = (lambda rs, _: len(rs)), measureName = 'Predictions', testRegions = None, figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)
 * 
 * def overlapSensitivityBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap sensitivity barplot.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_14overlapSensitivityBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_regionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("overlapSensitivityBarplot", 0);

  /* "gnocis/regions.pyx":1006
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(rs.overlap(ps)) / len(rs)), measureName = 'Overlap sensitivity', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)             # <<<<<<<<<<<<<<
 * 
 * def overlapPrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_regionBarplot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mainRegions, __pyx_v_predictionSets) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_testRegions, __pyx_v_regionSets) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_7regions_25overlapSensitivityBarplot_lambda15, 0, __pyx_n_s_overlapSensitivityBarplot_locals, NULL, __pyx_n_s_gnocis_regions, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measure, __pyx_t_3) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measureName, __pyx_kp_u_Overlap_sensitivity) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_outpath, __pyx_v_outpath) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_returnHTML, __pyx_v_returnHTML) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLabels, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLegend, __pyx_v_fontsizeLegend) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeAxis, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_style, __pyx_v_style) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showLegend, __pyx_v_showLegend) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bboxAnchorTo, __pyx_v_bboxAnchorTo) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_legendLoc, __pyx_v_legendLoc) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showValues, __pyx_v_showValues) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_isPercent, Py_True) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/regions.pyx":975
 * 	regionBarplot(mainRegions = predictionSets, measure = (lambda rs, _: len(rs)), measureName = 'Predictions', testRegions = None, figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)
 * 
 * def overlapSensitivityBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap sensitivity barplot.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.overlapSensitivityBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1008
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(rs.overlap(ps)) / len(rs)), measureName = 'Overlap sensitivity', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def overlapPrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_17overlapPrecisionBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_16overlapPrecisionBarplot[] = " Generates a prediction overlap precision barplot.\n\t\n\t:param predictionSets: List of prediction region sets.\n\t:param regionSets: List of validation region sets.\n\t:param figsize: Figure size.\n\t:param outpath: Output path.\n\t:param returnHTML: If True, an HTML node will be returned.\n\t:param fontsizeLabels: Size of label font.\n\t:param fontsizeLegend: Size of legend font.\n\t:param fontsizeAxis: Size of axis font.\n\t:param style: Plot style to use.\n\t:param showLegend: Flag for whether or not to render legend.\n\t:param bboxAnchorTo: Legend anchor point.\n\t:param legendLoc: Legend location.\n\t:param showValues: If True, values will be plotted.\n\t\n\t:type predictionSets: list\n\t:type regionSets: list\n\t:type figsize: tuple, optional\n\t:type outpath: str, optional\n\t:type returnHTML: bool, optional\n\t:type fontsizeLabels: float, optional\n\t:type fontsizeLegend: float, optional\n\t:type fontsizeAxis: float, optional\n\t:type style: str, optional\n\t:type showLegend: bool, optional\n\t:type bboxAnchorTo: tuple, optional\n\t:type legendLoc: str, optional\n\t:type showValues: bool, optional\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_17overlapPrecisionBarplot = {"overlapPrecisionBarplot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_17overlapPrecisionBarplot, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_16overlapPrecisionBarplot};
static PyObject *__pyx_pw_6gnocis_7regions_17overlapPrecisionBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_predictionSets = 0;
  PyObject *__pyx_v_regionSets = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_returnHTML = 0;
  PyObject *__pyx_v_fontsizeLabels = 0;
  PyObject *__pyx_v_fontsizeLegend = 0;
  PyObject *__pyx_v_fontsizeAxis = 0;
  PyObject *__pyx_v_style = 0;
  PyObject *__pyx_v_showLegend = 0;
  PyObject *__pyx_v_bboxAnchorTo = 0;
  PyObject *__pyx_v_legendLoc = 0;
  PyObject *__pyx_v_showValues = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlapPrecisionBarplot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_predictionSets,&__pyx_n_s_regionSets,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_returnHTML,&__pyx_n_s_fontsizeLabels,&__pyx_n_s_fontsizeLegend,&__pyx_n_s_fontsizeAxis,&__pyx_n_s_style,&__pyx_n_s_showLegend,&__pyx_n_s_bboxAnchorTo,&__pyx_n_s_legendLoc,&__pyx_n_s_showValues,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)__pyx_tuple__25);
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_False);
    values[5] = ((PyObject *)__pyx_int_18);
    values[6] = ((PyObject *)__pyx_int_12);
    values[7] = ((PyObject *)__pyx_int_10);
    values[8] = ((PyObject *)__pyx_n_u_ggplot);
    values[9] = ((PyObject *)Py_True);
    values[10] = ((PyObject *)__pyx_tuple__26);
    values[11] = ((PyObject *)__pyx_kp_u_upper_left);
    values[12] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_predictionSets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_regionSets)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("overlapPrecisionBarplot", 0, 2, 13, 1); __PYX_ERR(0, 1008, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_returnHTML);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLabels);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLegend);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeAxis);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showLegend);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bboxAnchorTo);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_legendLoc);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showValues);
          if (value) { values[12] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "overlapPrecisionBarplot") < 0)) __PYX_ERR(0, 1008, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_predictionSets = values[0];
    __pyx_v_regionSets = values[1];
    __pyx_v_figsize = values[2];
    __pyx_v_outpath = values[3];
    __pyx_v_returnHTML = values[4];
    __pyx_v_fontsizeLabels = values[5];
    __pyx_v_fontsizeLegend = values[6];
    __pyx_v_fontsizeAxis = values[7];
    __pyx_v_style = values[8];
    __pyx_v_showLegend = values[9];
    __pyx_v_bboxAnchorTo = values[10];
    __pyx_v_legendLoc = values[11];
    __pyx_v_showValues = values[12];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("overlapPrecisionBarplot", 0, 2, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1008, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.overlapPrecisionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_16overlapPrecisionBarplot(__pyx_self, __pyx_v_predictionSets, __pyx_v_regionSets, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_returnHTML, __pyx_v_fontsizeLabels, __pyx_v_fontsizeLegend, __pyx_v_fontsizeAxis, __pyx_v_style, __pyx_v_showLegend, __pyx_v_bboxAnchorTo, __pyx_v_legendLoc, __pyx_v_showValues);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1039
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(ps.overlap(rs)) / len(ps)), measureName = 'Overlap precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)             # <<<<<<<<<<<<<<
 * 
 * def nucleotidePrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_23overlapPrecisionBarplot_lambda16(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_23overlapPrecisionBarplot_lambda16 = {"lambda16", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_23overlapPrecisionBarplot_lambda16, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_7regions_23overlapPrecisionBarplot_lambda16(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ps = 0;
  PyObject *__pyx_v_rs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda16 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ps,&__pyx_n_s_rs,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ps)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lambda16", 1, 2, 2, 1); __PYX_ERR(0, 1039, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lambda16") < 0)) __PYX_ERR(0, 1039, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_ps = values[0];
    __pyx_v_rs = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda16", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1039, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.overlapPrecisionBarplot.lambda16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda16(__pyx_self, __pyx_v_ps, __pyx_v_rs);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda16(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ps, PyObject *__pyx_v_rs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda16", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ps, __pyx_n_s_overlap); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_rs) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_rs);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = PyObject_Length(__pyx_v_ps); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (unlikely(__pyx_t_5 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 1039, __pyx_L1_error)
  }
  __pyx_t_1 = PyFloat_FromDouble((((double)__pyx_t_4) / ((double)__pyx_t_5))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.overlapPrecisionBarplot.lambda16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1008
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(rs.overlap(ps)) / len(rs)), measureName = 'Overlap sensitivity', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def overlapPrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_16overlapPrecisionBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_regionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("overlapPrecisionBarplot", 0);

  /* "gnocis/regions.pyx":1039
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(ps.overlap(rs)) / len(ps)), measureName = 'Overlap precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)             # <<<<<<<<<<<<<<
 * 
 * def nucleotidePrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_regionBarplot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mainRegions, __pyx_v_predictionSets) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_testRegions, __pyx_v_regionSets) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_7regions_23overlapPrecisionBarplot_lambda16, 0, __pyx_n_s_overlapPrecisionBarplot_locals_l, NULL, __pyx_n_s_gnocis_regions, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measure, __pyx_t_3) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measureName, __pyx_kp_u_Overlap_precision) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_outpath, __pyx_v_outpath) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_returnHTML, __pyx_v_returnHTML) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLabels, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLegend, __pyx_v_fontsizeLegend) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeAxis, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_style, __pyx_v_style) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showLegend, __pyx_v_showLegend) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bboxAnchorTo, __pyx_v_bboxAnchorTo) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_legendLoc, __pyx_v_legendLoc) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showValues, __pyx_v_showValues) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_isPercent, Py_True) < 0) __PYX_ERR(0, 1039, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/regions.pyx":1008
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(rs.overlap(ps)) / len(rs)), measureName = 'Overlap sensitivity', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def overlapPrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.overlapPrecisionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1041
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(ps.overlap(rs)) / len(ps)), measureName = 'Overlap precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotidePrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction nucleotide precision barplot.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_19nucleotidePrecisionBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_18nucleotidePrecisionBarplot[] = " Generates a prediction nucleotide precision barplot.\n\t\n\t:param predictionSets: List of prediction region sets.\n\t:param regionSets: List of validation region sets.\n\t:param figsize: Figure size.\n\t:param outpath: Output path.\n\t:param returnHTML: If True, an HTML node will be returned.\n\t:param fontsizeLabels: Size of label font.\n\t:param fontsizeLegend: Size of legend font.\n\t:param fontsizeAxis: Size of axis font.\n\t:param style: Plot style to use.\n\t:param showLegend: Flag for whether or not to render legend.\n\t:param bboxAnchorTo: Legend anchor point.\n\t:param legendLoc: Legend location.\n\t:param showValues: If True, values will be plotted.\n\t\n\t:type predictionSets: list\n\t:type regionSets: list\n\t:type figsize: tuple, optional\n\t:type outpath: str, optional\n\t:type returnHTML: bool, optional\n\t:type fontsizeLabels: float, optional\n\t:type fontsizeLegend: float, optional\n\t:type fontsizeAxis: float, optional\n\t:type style: str, optional\n\t:type showLegend: bool, optional\n\t:type bboxAnchorTo: tuple, optional\n\t:type legendLoc: str, optional\n\t:type showValues: bool, optional\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_19nucleotidePrecisionBarplot = {"nucleotidePrecisionBarplot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_19nucleotidePrecisionBarplot, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_18nucleotidePrecisionBarplot};
static PyObject *__pyx_pw_6gnocis_7regions_19nucleotidePrecisionBarplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_predictionSets = 0;
  PyObject *__pyx_v_regionSets = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_returnHTML = 0;
  PyObject *__pyx_v_fontsizeLabels = 0;
  PyObject *__pyx_v_fontsizeLegend = 0;
  PyObject *__pyx_v_fontsizeAxis = 0;
  PyObject *__pyx_v_style = 0;
  PyObject *__pyx_v_showLegend = 0;
  PyObject *__pyx_v_bboxAnchorTo = 0;
  PyObject *__pyx_v_legendLoc = 0;
  PyObject *__pyx_v_showValues = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nucleotidePrecisionBarplot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_predictionSets,&__pyx_n_s_regionSets,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_returnHTML,&__pyx_n_s_fontsizeLabels,&__pyx_n_s_fontsizeLegend,&__pyx_n_s_fontsizeAxis,&__pyx_n_s_style,&__pyx_n_s_showLegend,&__pyx_n_s_bboxAnchorTo,&__pyx_n_s_legendLoc,&__pyx_n_s_showValues,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)__pyx_tuple__25);
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_False);
    values[5] = ((PyObject *)__pyx_int_18);
    values[6] = ((PyObject *)__pyx_int_12);
    values[7] = ((PyObject *)__pyx_int_10);
    values[8] = ((PyObject *)__pyx_n_u_ggplot);
    values[9] = ((PyObject *)Py_True);
    values[10] = ((PyObject *)__pyx_tuple__26);
    values[11] = ((PyObject *)__pyx_kp_u_upper_left);
    values[12] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_predictionSets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_regionSets)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("nucleotidePrecisionBarplot", 0, 2, 13, 1); __PYX_ERR(0, 1041, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_returnHTML);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLabels);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLegend);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeAxis);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showLegend);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bboxAnchorTo);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_legendLoc);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showValues);
          if (value) { values[12] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "nucleotidePrecisionBarplot") < 0)) __PYX_ERR(0, 1041, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_predictionSets = values[0];
    __pyx_v_regionSets = values[1];
    __pyx_v_figsize = values[2];
    __pyx_v_outpath = values[3];
    __pyx_v_returnHTML = values[4];
    __pyx_v_fontsizeLabels = values[5];
    __pyx_v_fontsizeLegend = values[6];
    __pyx_v_fontsizeAxis = values[7];
    __pyx_v_style = values[8];
    __pyx_v_showLegend = values[9];
    __pyx_v_bboxAnchorTo = values[10];
    __pyx_v_legendLoc = values[11];
    __pyx_v_showValues = values[12];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("nucleotidePrecisionBarplot", 0, 2, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1041, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.nucleotidePrecisionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_18nucleotidePrecisionBarplot(__pyx_self, __pyx_v_predictionSets, __pyx_v_regionSets, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_returnHTML, __pyx_v_fontsizeLabels, __pyx_v_fontsizeLegend, __pyx_v_fontsizeAxis, __pyx_v_style, __pyx_v_showLegend, __pyx_v_bboxAnchorTo, __pyx_v_legendLoc, __pyx_v_showValues);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1072
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)), measureName = 'Nucleotide precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)             # <<<<<<<<<<<<<<
 * 
 * def nucleotideRegionF1Barplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_26nucleotidePrecisionBarplot_lambda17(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_26nucleotidePrecisionBarplot_lambda17 = {"lambda17", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_26nucleotidePrecisionBarplot_lambda17, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_7regions_26nucleotidePrecisionBarplot_lambda17(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ps = 0;
  PyObject *__pyx_v_rs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda17 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ps,&__pyx_n_s_rs,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ps)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lambda17", 1, 2, 2, 1); __PYX_ERR(0, 1072, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lambda17") < 0)) __PYX_ERR(0, 1072, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_ps = values[0];
    __pyx_v_rs = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda17", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1072, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.nucleotidePrecisionBarplot.lambda17", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda17(__pyx_self, __pyx_v_ps, __pyx_v_rs);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_2generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

static PyObject *__pyx_pf_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_19_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_19_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1072, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_2generator12, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_nucleotidePrecisionBarplot_local, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.nucleotidePrecisionBarplot.lambda17.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_2generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps)) { __Pyx_RaiseClosureNameError("ps"); __PYX_ERR(0, 1072, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_rs)) { __Pyx_RaiseClosureNameError("rs"); __PYX_ERR(0, 1072, __pyx_L1_error) }
  __pyx_t_1 = PyNumber_And(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_rs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1072, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1072, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1072, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_5 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1072, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1072, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_5generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

static PyObject *__pyx_pf_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_20_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_20_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1072, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_5generator13, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_nucleotidePrecisionBarplot_local, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.nucleotidePrecisionBarplot.lambda17.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_5generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps)) { __Pyx_RaiseClosureNameError("ps"); __PYX_ERR(0, 1072, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1072, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1072, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1072, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1072, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1072, __pyx_L1_error)
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1072, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda17(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ps, PyObject *__pyx_v_rs) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *__pyx_cur_scope;
  PyObject *__pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_2generator12 = 0;
  PyObject *__pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_5generator13 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda17", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1072, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_ps = __pyx_v_ps;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ps);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ps);
  __pyx_cur_scope->__pyx_v_rs = __pyx_v_rs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_rs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_rs);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.nucleotidePrecisionBarplot.lambda17", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_2generator12);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_26nucleotidePrecisionBarplot_8lambda17_5generator13);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1041
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(ps.overlap(rs)) / len(ps)), measureName = 'Overlap precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotidePrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction nucleotide precision barplot.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_18nucleotidePrecisionBarplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_regionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nucleotidePrecisionBarplot", 0);

  /* "gnocis/regions.pyx":1072
 * 	:type showValues: bool, optional
 * 	"""
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)), measureName = 'Nucleotide precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)             # <<<<<<<<<<<<<<
 * 
 * def nucleotideRegionF1Barplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_regionBarplot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mainRegions, __pyx_v_predictionSets) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_testRegions, __pyx_v_regionSets) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_7regions_26nucleotidePrecisionBarplot_lambda17, 0, __pyx_n_s_nucleotidePrecisionBarplot_local_2, NULL, __pyx_n_s_gnocis_regions, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measure, __pyx_t_3) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measureName, __pyx_kp_u_Nucleotide_precision) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_outpath, __pyx_v_outpath) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_returnHTML, __pyx_v_returnHTML) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLabels, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLegend, __pyx_v_fontsizeLegend) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeAxis, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_style, __pyx_v_style) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showLegend, __pyx_v_showLegend) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bboxAnchorTo, __pyx_v_bboxAnchorTo) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_legendLoc, __pyx_v_legendLoc) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showValues, __pyx_v_showValues) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_isPercent, Py_True) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/regions.pyx":1041
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(ps.overlap(rs)) / len(ps)), measureName = 'Overlap precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotidePrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction nucleotide precision barplot.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.nucleotidePrecisionBarplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1074
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)), measureName = 'Nucleotide precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotideRegionF1Barplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of F1-measures, where the recall is based on region overlap sensitivity and precision is based on nucleotide precision.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_21nucleotideRegionF1Barplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_7regions_20nucleotideRegionF1Barplot[] = " Generates a barplot of F1-measures, where the recall is based on region overlap sensitivity and precision is based on nucleotide precision.\n\t\n\t:param predictionSets: List of prediction region sets.\n\t:param regionSets: List of validation region sets.\n\t:param figsize: Figure size.\n\t:param outpath: Output path.\n\t:param returnHTML: If True, an HTML node will be returned.\n\t:param fontsizeLabels: Size of label font.\n\t:param fontsizeLegend: Size of legend font.\n\t:param fontsizeAxis: Size of axis font.\n\t:param style: Plot style to use.\n\t:param showLegend: Flag for whether or not to render legend.\n\t:param bboxAnchorTo: Legend anchor point.\n\t:param legendLoc: Legend location.\n\t:param showValues: If True, values will be plotted.\n\t\n\t:type predictionSets: list\n\t:type regionSets: list\n\t:type figsize: tuple, optional\n\t:type outpath: str, optional\n\t:type returnHTML: bool, optional\n\t:type fontsizeLabels: float, optional\n\t:type fontsizeLegend: float, optional\n\t:type fontsizeAxis: float, optional\n\t:type style: str, optional\n\t:type showLegend: bool, optional\n\t:type bboxAnchorTo: tuple, optional\n\t:type legendLoc: str, optional\n\t:type showValues: bool, optional\n\t";
static PyMethodDef __pyx_mdef_6gnocis_7regions_21nucleotideRegionF1Barplot = {"nucleotideRegionF1Barplot", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_21nucleotideRegionF1Barplot, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_20nucleotideRegionF1Barplot};
static PyObject *__pyx_pw_6gnocis_7regions_21nucleotideRegionF1Barplot(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_predictionSets = 0;
  PyObject *__pyx_v_regionSets = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_returnHTML = 0;
  PyObject *__pyx_v_fontsizeLabels = 0;
  PyObject *__pyx_v_fontsizeLegend = 0;
  PyObject *__pyx_v_fontsizeAxis = 0;
  PyObject *__pyx_v_style = 0;
  PyObject *__pyx_v_showLegend = 0;
  PyObject *__pyx_v_bboxAnchorTo = 0;
  PyObject *__pyx_v_legendLoc = 0;
  PyObject *__pyx_v_showValues = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nucleotideRegionF1Barplot (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_predictionSets,&__pyx_n_s_regionSets,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_returnHTML,&__pyx_n_s_fontsizeLabels,&__pyx_n_s_fontsizeLegend,&__pyx_n_s_fontsizeAxis,&__pyx_n_s_style,&__pyx_n_s_showLegend,&__pyx_n_s_bboxAnchorTo,&__pyx_n_s_legendLoc,&__pyx_n_s_showValues,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)__pyx_tuple__25);
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_False);
    values[5] = ((PyObject *)__pyx_int_18);
    values[6] = ((PyObject *)__pyx_int_12);
    values[7] = ((PyObject *)__pyx_int_10);
    values[8] = ((PyObject *)__pyx_n_u_ggplot);
    values[9] = ((PyObject *)Py_True);
    values[10] = ((PyObject *)__pyx_tuple__26);
    values[11] = ((PyObject *)__pyx_kp_u_upper_left);
    values[12] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_predictionSets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_regionSets)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("nucleotideRegionF1Barplot", 0, 2, 13, 1); __PYX_ERR(0, 1074, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_returnHTML);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLabels);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLegend);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeAxis);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showLegend);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bboxAnchorTo);
          if (value) { values[10] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_legendLoc);
          if (value) { values[11] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_showValues);
          if (value) { values[12] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "nucleotideRegionF1Barplot") < 0)) __PYX_ERR(0, 1074, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_predictionSets = values[0];
    __pyx_v_regionSets = values[1];
    __pyx_v_figsize = values[2];
    __pyx_v_outpath = values[3];
    __pyx_v_returnHTML = values[4];
    __pyx_v_fontsizeLabels = values[5];
    __pyx_v_fontsizeLegend = values[6];
    __pyx_v_fontsizeAxis = values[7];
    __pyx_v_style = values[8];
    __pyx_v_showLegend = values[9];
    __pyx_v_bboxAnchorTo = values[10];
    __pyx_v_legendLoc = values[11];
    __pyx_v_showValues = values[12];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("nucleotideRegionF1Barplot", 0, 2, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1074, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.nucleotideRegionF1Barplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_20nucleotideRegionF1Barplot(__pyx_self, __pyx_v_predictionSets, __pyx_v_regionSets, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_returnHTML, __pyx_v_fontsizeLabels, __pyx_v_fontsizeLegend, __pyx_v_fontsizeAxis, __pyx_v_style, __pyx_v_showLegend, __pyx_v_bboxAnchorTo, __pyx_v_legendLoc, __pyx_v_showValues);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1105
 * 	:type showValues: bool, optional
 * 	"""
 * 	def measure(ps, rs):             # <<<<<<<<<<<<<<
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)
 * 		r = len(rs.overlap(ps)) / len(rs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_25nucleotideRegionF1Barplot_1measure(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_25nucleotideRegionF1Barplot_1measure = {"measure", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_25nucleotideRegionF1Barplot_1measure, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_7regions_25nucleotideRegionF1Barplot_1measure(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ps = 0;
  PyObject *__pyx_v_rs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("measure (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ps,&__pyx_n_s_rs,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ps)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("measure", 1, 2, 2, 1); __PYX_ERR(0, 1105, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "measure") < 0)) __PYX_ERR(0, 1105, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_ps = values[0];
    __pyx_v_rs = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("measure", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1105, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.nucleotideRegionF1Barplot.measure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_measure(__pyx_self, __pyx_v_ps, __pyx_v_rs);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_2generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/regions.pyx":1106
 * 	"""
 * 	def measure(ps, rs):
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)             # <<<<<<<<<<<<<<
 * 		r = len(rs.overlap(ps)) / len(rs)
 * 		return 2 * (p*r) / (p+r)
 */

static PyObject *__pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_22_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_22_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1106, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_2generator14, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_nucleotideRegionF1Barplot_locals, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.nucleotideRegionF1Barplot.measure.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_2generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1106, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps)) { __Pyx_RaiseClosureNameError("ps"); __PYX_ERR(0, 1106, __pyx_L1_error) }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_rs)) { __Pyx_RaiseClosureNameError("rs"); __PYX_ERR(0, 1106, __pyx_L1_error) }
  __pyx_t_1 = PyNumber_And(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_rs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1106, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1106, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_5 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1106, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1106, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_5generator15(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

static PyObject *__pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_23_genexpr(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_23_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1106, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_5generator15, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_nucleotideRegionF1Barplot_locals, __pyx_n_s_gnocis_regions); if (unlikely(!gen)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.regions.nucleotideRegionF1Barplot.measure.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_5generator15(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1106, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps)) { __Pyx_RaiseClosureNameError("ps"); __PYX_ERR(0, 1106, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1106, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1106, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1106, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = PyObject_Length(__pyx_cur_scope->__pyx_v_r); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1106, __pyx_L1_error)
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1106, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1105
 * 	:type showValues: bool, optional
 * 	"""
 * 	def measure(ps, rs):             # <<<<<<<<<<<<<<
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)
 * 		r = len(rs.overlap(ps)) / len(rs)
 */

static PyObject *__pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_measure(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ps, PyObject *__pyx_v_rs) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *__pyx_cur_scope;
  PyObject *__pyx_v_p = NULL;
  double __pyx_v_r;
  PyObject *__pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_2generator14 = 0;
  PyObject *__pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_5generator15 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("measure", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *)__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_21_measure(__pyx_ptype_6gnocis_7regions___pyx_scope_struct_21_measure, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1105, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_ps = __pyx_v_ps;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ps);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ps);
  __pyx_cur_scope->__pyx_v_rs = __pyx_v_rs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_rs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_rs);

  /* "gnocis/regions.pyx":1106
 * 	"""
 * 	def measure(ps, rs):
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)             # <<<<<<<<<<<<<<
 * 		r = len(rs.overlap(ps)) / len(rs)
 * 		return 2 * (p*r) / (p+r)
 */
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_pf_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_p = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":1107
 * 	def measure(ps, rs):
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)
 * 		r = len(rs.overlap(ps)) / len(rs)             # <<<<<<<<<<<<<<
 * 		return 2 * (p*r) / (p+r)
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = measure, measureName = 'Nucleotide/region F1 score', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_rs, __pyx_n_s_overlap); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_cur_scope->__pyx_v_ps) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_cur_scope->__pyx_v_ps);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_rs;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_5 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 1107, __pyx_L1_error)
  }
  __pyx_v_r = (((double)__pyx_t_4) / ((double)__pyx_t_5));

  /* "gnocis/regions.pyx":1108
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)
 * 		r = len(rs.overlap(ps)) / len(rs)
 * 		return 2 * (p*r) / (p+r)             # <<<<<<<<<<<<<<
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = measure, measureName = 'Nucleotide/region F1 score', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_r); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_p, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_r); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyNumber_Add(__pyx_v_p, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gnocis/regions.pyx":1105
 * 	:type showValues: bool, optional
 * 	"""
 * 	def measure(ps, rs):             # <<<<<<<<<<<<<<
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)
 * 		r = len(rs.overlap(ps)) / len(rs)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.nucleotideRegionF1Barplot.measure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_2generator14);
  __Pyx_XDECREF(__pyx_gb_6gnocis_7regions_25nucleotideRegionF1Barplot_7measure_5generator15);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/regions.pyx":1074
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)), measureName = 'Nucleotide precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotideRegionF1Barplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of F1-measures, where the recall is based on region overlap sensitivity and precision is based on nucleotide precision.
 * 
 */

static PyObject *__pyx_pf_6gnocis_7regions_20nucleotideRegionF1Barplot(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_predictionSets, PyObject *__pyx_v_regionSets, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_style, PyObject *__pyx_v_showLegend, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc, PyObject *__pyx_v_showValues) {
  PyObject *__pyx_v_measure = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("nucleotideRegionF1Barplot", 0);

  /* "gnocis/regions.pyx":1105
 * 	:type showValues: bool, optional
 * 	"""
 * 	def measure(ps, rs):             # <<<<<<<<<<<<<<
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)
 * 		r = len(rs.overlap(ps)) / len(rs)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_7regions_25nucleotideRegionF1Barplot_1measure, 0, __pyx_n_s_nucleotideRegionF1Barplot_locals_2, NULL, __pyx_n_s_gnocis_regions, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_measure = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":1109
 * 		r = len(rs.overlap(ps)) / len(rs)
 * 		return 2 * (p*r) / (p+r)
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = measure, measureName = 'Nucleotide/region F1 score', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_regionBarplot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mainRegions, __pyx_v_predictionSets) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_testRegions, __pyx_v_regionSets) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measure, __pyx_v_measure) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_measureName, __pyx_kp_u_Nucleotide_region_F1_score) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_outpath, __pyx_v_outpath) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_returnHTML, __pyx_v_returnHTML) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLabels, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeLegend, __pyx_v_fontsizeLegend) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fontsizeAxis, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_style, __pyx_v_style) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showLegend, __pyx_v_showLegend) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bboxAnchorTo, __pyx_v_bboxAnchorTo) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_legendLoc, __pyx_v_legendLoc) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_showValues, __pyx_v_showValues) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_isPercent, Py_False) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/regions.pyx":1074
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)), measureName = 'Nucleotide precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotideRegionF1Barplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of F1-measures, where the recall is based on region overlap sensitivity and precision is based on nucleotide precision.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.regions.nucleotideRegionF1Barplot", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_measure);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_region(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_23__pyx_unpickle_region(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_23__pyx_unpickle_region = {"__pyx_unpickle_region", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_23__pyx_unpickle_region, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_7regions_23__pyx_unpickle_region(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_region (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_region", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_region", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_region") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_region", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.__pyx_unpickle_region", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_22__pyx_unpickle_region(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_22__pyx_unpickle_region(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_region", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x11eeeec:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x11eeeec = (end, ext, feature, group, markers, score, seq, source, start, strand))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x11eeeec) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x11eeeec:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x11eeeec = (end, ext, feature, group, markers, score, seq, source, start, strand))" % __pyx_checksum)
 *     __pyx_result = region.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x11eeeec:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x11eeeec = (end, ext, feature, group, markers, score, seq, source, start, strand))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = region.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x11, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x11eeeec:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x11eeeec = (end, ext, feature, group, markers, score, seq, source, start, strand))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x11eeeec = (end, ext, feature, group, markers, score, seq, source, start, strand))" % __pyx_checksum)
 *     __pyx_result = region.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_region__set_state(<region> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6gnocis_7regions_region), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x11eeeec = (end, ext, feature, group, markers, score, seq, source, start, strand))" % __pyx_checksum)
 *     __pyx_result = region.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_region__set_state(<region> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = region.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_region__set_state(<region> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_region__set_state(region __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6gnocis_7regions___pyx_unpickle_region__set_state(((struct __pyx_obj_6gnocis_7regions_region *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x11eeeec = (end, ext, feature, group, markers, score, seq, source, start, strand))" % __pyx_checksum)
 *     __pyx_result = region.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_region__set_state(<region> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_region__set_state(<region> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_region__set_state(region __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.ext = __pyx_state[1]; __pyx_result.feature = __pyx_state[2]; __pyx_result.group = __pyx_state[3]; __pyx_result.markers = __pyx_state[4]; __pyx_result.score = __pyx_state[5]; __pyx_result.seq = __pyx_state[6]; __pyx_result.source = __pyx_state[7]; __pyx_result.start = __pyx_state[8]; __pyx_result.strand = __pyx_state[9]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_region(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.regions.__pyx_unpickle_region", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_region__set_state(<region> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_region__set_state(region __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.ext = __pyx_state[1]; __pyx_result.feature = __pyx_state[2]; __pyx_result.group = __pyx_state[3]; __pyx_result.markers = __pyx_state[4]; __pyx_result.score = __pyx_state[5]; __pyx_result.seq = __pyx_state[6]; __pyx_result.source = __pyx_state[7]; __pyx_result.start = __pyx_state[8]; __pyx_result.strand = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6gnocis_7regions___pyx_unpickle_region__set_state(struct __pyx_obj_6gnocis_7regions_region *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PY_LONG_LONG __pyx_t_2;
  float __pyx_t_3;
  bool __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_region__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_region__set_state(region __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.ext = __pyx_state[1]; __pyx_result.feature = __pyx_state[2]; __pyx_result.group = __pyx_state[3]; __pyx_result.markers = __pyx_state[4]; __pyx_result.score = __pyx_state[5]; __pyx_result.seq = __pyx_state[6]; __pyx_result.source = __pyx_state[7]; __pyx_result.start = __pyx_state[8]; __pyx_result.strand = __pyx_state[9]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_1); if (unlikely((__pyx_t_2 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->end = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->ext);
  __Pyx_DECREF(__pyx_v___pyx_result->ext);
  __pyx_v___pyx_result->ext = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->feature);
  __Pyx_DECREF(__pyx_v___pyx_result->feature);
  __pyx_v___pyx_result->feature = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->group);
  __Pyx_DECREF(__pyx_v___pyx_result->group);
  __pyx_v___pyx_result->group = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PySet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->markers);
  __Pyx_DECREF(__pyx_v___pyx_result->markers);
  __pyx_v___pyx_result->markers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_3 == (float)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->score = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->seq);
  __Pyx_DECREF(__pyx_v___pyx_result->seq);
  __pyx_v___pyx_result->seq = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->source);
  __Pyx_DECREF(__pyx_v___pyx_result->source);
  __pyx_v___pyx_result->source = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_t_1); if (unlikely((__pyx_t_2 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->start = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->strand = __pyx_t_4;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_region__set_state(region __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.ext = __pyx_state[1]; __pyx_result.feature = __pyx_state[2]; __pyx_result.group = __pyx_state[3]; __pyx_result.markers = __pyx_state[4]; __pyx_result.score = __pyx_state[5]; __pyx_result.seq = __pyx_state[6]; __pyx_result.source = __pyx_state[7]; __pyx_result.start = __pyx_state[8]; __pyx_result.strand = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_6 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = ((__pyx_t_6 > 10) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  __pyx_t_5 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "(tree fragment)":14
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.ext = __pyx_state[1]; __pyx_result.feature = __pyx_state[2]; __pyx_result.group = __pyx_state[3]; __pyx_result.markers = __pyx_state[4]; __pyx_result.score = __pyx_state[5]; __pyx_result.seq = __pyx_state[6]; __pyx_result.source = __pyx_state[7]; __pyx_result.start = __pyx_state[8]; __pyx_result.strand = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[10])             # <<<<<<<<<<<<<<
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_11, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_region__set_state(region __pyx_result, tuple __pyx_state):
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.ext = __pyx_state[1]; __pyx_result.feature = __pyx_state[2]; __pyx_result.group = __pyx_state[3]; __pyx_result.markers = __pyx_state[4]; __pyx_result.score = __pyx_state[5]; __pyx_result.seq = __pyx_state[6]; __pyx_result.source = __pyx_state[7]; __pyx_result.start = __pyx_state[8]; __pyx_result.strand = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[10])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_region__set_state(<region> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_region__set_state(region __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.ext = __pyx_state[1]; __pyx_result.feature = __pyx_state[2]; __pyx_result.group = __pyx_state[3]; __pyx_result.markers = __pyx_state[4]; __pyx_result.score = __pyx_state[5]; __pyx_result.seq = __pyx_state[6]; __pyx_result.source = __pyx_state[7]; __pyx_result.start = __pyx_state[8]; __pyx_result.strand = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("gnocis.regions.__pyx_unpickle_region__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_regions(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_7regions_25__pyx_unpickle_regions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_7regions_25__pyx_unpickle_regions = {"__pyx_unpickle_regions", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_25__pyx_unpickle_regions, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_7regions_25__pyx_unpickle_regions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_regions (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_regions", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_regions", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_regions") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_regions", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.regions.__pyx_unpickle_regions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_7regions_24__pyx_unpickle_regions(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_7regions_24__pyx_unpickle_regions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_regions", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9d99455:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9d99455 = (name, regions))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x9d99455) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9d99455:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9d99455 = (name, regions))" % __pyx_checksum)
 *     __pyx_result = regions.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x9d99455:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9d99455 = (name, regions))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = regions.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x9d, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9d99455:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9d99455 = (name, regions))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9d99455 = (name, regions))" % __pyx_checksum)
 *     __pyx_result = regions.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_regions__set_state(<regions> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9d99455 = (name, regions))" % __pyx_checksum)
 *     __pyx_result = regions.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_regions__set_state(<regions> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = regions.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_regions__set_state(<regions> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_regions__set_state(regions __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6gnocis_7regions___pyx_unpickle_regions__set_state(((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9d99455 = (name, regions))" % __pyx_checksum)
 *     __pyx_result = regions.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_regions__set_state(<regions> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_regions__set_state(<regions> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_regions__set_state(regions __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.regions = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_regions(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.regions.__pyx_unpickle_regions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_regions__set_state(<regions> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_regions__set_state(regions __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.regions = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6gnocis_7regions___pyx_unpickle_regions__set_state(struct __pyx_obj_6gnocis_7regions_regions *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_regions__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_regions__set_state(regions __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.regions = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->regions);
  __Pyx_DECREF(__pyx_v___pyx_result->regions);
  __pyx_v___pyx_result->regions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_regions__set_state(regions __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.regions = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 2) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.regions = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_regions__set_state(regions __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.regions = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_regions__set_state(<regions> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_regions__set_state(regions __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]; __pyx_result.regions = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("gnocis.regions.__pyx_unpickle_regions__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_6gnocis_7regions_region(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gnocis_7regions_region *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gnocis_7regions_region *)o);
  p->seq = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->source = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->feature = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->group = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->ext = Py_None; Py_INCREF(Py_None);
  p->markers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions_region(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions_region *p = (struct __pyx_obj_6gnocis_7regions_region *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->seq);
  Py_CLEAR(p->source);
  Py_CLEAR(p->feature);
  Py_CLEAR(p->group);
  Py_CLEAR(p->ext);
  Py_CLEAR(p->markers);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6gnocis_7regions_region(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions_region *p = (struct __pyx_obj_6gnocis_7regions_region *)o;
  if (p->ext) {
    e = (*v)(p->ext, a); if (e) return e;
  }
  if (p->markers) {
    e = (*v)(p->markers, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions_region(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions_region *p = (struct __pyx_obj_6gnocis_7regions_region *)o;
  tmp = ((PyObject*)p->ext);
  p->ext = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->markers);
  p->markers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_seq(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_3seq_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_seq(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_3seq_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gnocis_7regions_6region_3seq_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_start(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_5start_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_start(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_5start_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_end(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_3end_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_end(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_3end_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_strand(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_6strand_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_strand(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_6strand_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_score(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_5score_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_score(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_5score_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_source(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_6source_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_source(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_6source_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gnocis_7regions_6region_6source_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_feature(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_7feature_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_feature(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_7feature_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gnocis_7regions_6region_7feature_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_group(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_5group_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_group(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_5group_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gnocis_7regions_6region_5group_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_ext(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_3ext_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_ext(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_3ext_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gnocis_7regions_6region_3ext_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_6region_markers(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_6region_7markers_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_6region_markers(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_6region_7markers_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gnocis_7regions_6region_7markers_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_6gnocis_7regions_region[] = {
  {"singleton", (PyCFunction)__pyx_pw_6gnocis_7regions_6region_3singleton, METH_NOARGS, 0},
  {"center", (PyCFunction)__pyx_pw_6gnocis_7regions_6region_5center, METH_NOARGS, 0},
  {"bstr", (PyCFunction)__pyx_pw_6gnocis_7regions_6region_13bstr, METH_NOARGS, __pyx_doc_6gnocis_7regions_6region_12bstr},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6gnocis_7regions_6region_15__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6gnocis_7regions_6region_17__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gnocis_7regions_region[] = {
  {(char *)"seq", __pyx_getprop_6gnocis_7regions_6region_seq, __pyx_setprop_6gnocis_7regions_6region_seq, (char *)0, 0},
  {(char *)"start", __pyx_getprop_6gnocis_7regions_6region_start, __pyx_setprop_6gnocis_7regions_6region_start, (char *)0, 0},
  {(char *)"end", __pyx_getprop_6gnocis_7regions_6region_end, __pyx_setprop_6gnocis_7regions_6region_end, (char *)0, 0},
  {(char *)"strand", __pyx_getprop_6gnocis_7regions_6region_strand, __pyx_setprop_6gnocis_7regions_6region_strand, (char *)0, 0},
  {(char *)"score", __pyx_getprop_6gnocis_7regions_6region_score, __pyx_setprop_6gnocis_7regions_6region_score, (char *)0, 0},
  {(char *)"source", __pyx_getprop_6gnocis_7regions_6region_source, __pyx_setprop_6gnocis_7regions_6region_source, (char *)0, 0},
  {(char *)"feature", __pyx_getprop_6gnocis_7regions_6region_feature, __pyx_setprop_6gnocis_7regions_6region_feature, (char *)0, 0},
  {(char *)"group", __pyx_getprop_6gnocis_7regions_6region_group, __pyx_setprop_6gnocis_7regions_6region_group, (char *)0, 0},
  {(char *)"ext", __pyx_getprop_6gnocis_7regions_6region_ext, __pyx_setprop_6gnocis_7regions_6region_ext, (char *)0, 0},
  {(char *)"markers", __pyx_getprop_6gnocis_7regions_6region_markers, __pyx_setprop_6gnocis_7regions_6region_markers, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_region = {
  __pyx_pw_6gnocis_7regions_6region_7__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_region = {
  __pyx_pw_6gnocis_7regions_6region_7__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6gnocis_7regions_region = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.region", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions_region), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions_region, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6gnocis_7regions_6region_11__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_region, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_region, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6gnocis_7regions_6region_9__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n\tThe `region` class represents a region within a sequence. A region has a sequence name, a start and end, and a strandedness, with optional additional annotation.\n\t\n\t:param seq: Sequence name.\n\t:param start: Start coordinate.\n\t:param end: End coordinate (inclusive).\n\t:param strand: Strandedness. `True` for the + (forward) strand (default), and `False` for the - (reverse) strand.\n\t:param score: Score.\n\t:param source: Soure.\n\t:param feature: Feature.\n\t:param group: Group.\n\t:param dropChr: `True` if a leading `chr` in `seq` is to be dropped (default).\n\t\n\t:type seq: str\n\t:type start: int\n\t:type end: int\n\t:type strand: bool, optional\n\t:type score: float, optional\n\t:type source: str, optional\n\t:type feature: str, optional\n\t:type group: str, optional\n\t:type dropChr: bool, optional\n\t\n\tThe length of a region `r` can be calculated with `len(r)`.\n\t", /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions_region, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions_region, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gnocis_7regions_region, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gnocis_7regions_region, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gnocis_7regions_6region_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions_region, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_6gnocis_7regions_regions(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gnocis_7regions_regions *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gnocis_7regions_regions *)o);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->regions = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions_regions(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions_regions *p = (struct __pyx_obj_6gnocis_7regions_regions *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->regions);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6gnocis_7regions_regions(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions_regions *p = (struct __pyx_obj_6gnocis_7regions_regions *)o;
  if (p->regions) {
    e = (*v)(p->regions, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions_regions(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions_regions *p = (struct __pyx_obj_6gnocis_7regions_regions *)o;
  tmp = ((PyObject*)p->regions);
  p->regions = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6gnocis_7regions_regions(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_6gnocis_7regions_7regions_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_7regions_4name_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_7regions_name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_7regions_4name_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gnocis_7regions_7regions_4name_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6gnocis_7regions_7regions_regions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gnocis_7regions_7regions_7regions_1__get__(o);
}

static int __pyx_setprop_6gnocis_7regions_7regions_regions(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gnocis_7regions_7regions_7regions_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gnocis_7regions_7regions_7regions_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_6gnocis_7regions_regions[] = {
  {"_as_dict_", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_11_as_dict_, METH_NOARGS, 0},
  {"table", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_13table, METH_NOARGS, 0},
  {"_repr_html_", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_15_repr_html_, METH_NOARGS, 0},
  {"bstr", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_27bstr, METH_NOARGS, __pyx_doc_6gnocis_7regions_7regions_26bstr},
  {"randomSplit", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_7regions_29randomSplit, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_7regions_28randomSplit},
  {"sample", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_7regions_31sample, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_7regions_30sample},
  {"filter", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_7regions_33filter, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_7regions_32filter},
  {"flatten", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_35flatten, METH_NOARGS, __pyx_doc_6gnocis_7regions_7regions_34flatten},
  {"merge", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_37merge, METH_O, __pyx_doc_6gnocis_7regions_7regions_36merge},
  {"intersection", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_39intersection, METH_O, __pyx_doc_6gnocis_7regions_7regions_38intersection},
  {"exclusion", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_41exclusion, METH_O, __pyx_doc_6gnocis_7regions_7regions_40exclusion},
  {"overlap", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_43overlap, METH_O, __pyx_doc_6gnocis_7regions_7regions_42overlap},
  {"nonOverlap", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_45nonOverlap, METH_O, __pyx_doc_6gnocis_7regions_7regions_44nonOverlap},
  {"overlapSensitivity", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_47overlapSensitivity, METH_O, __pyx_doc_6gnocis_7regions_7regions_46overlapSensitivity},
  {"overlapPrecision", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_49overlapPrecision, METH_O, __pyx_doc_6gnocis_7regions_7regions_48overlapPrecision},
  {"nucleotidePrecision", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_51nucleotidePrecision, METH_O, __pyx_doc_6gnocis_7regions_7regions_50nucleotidePrecision},
  {"dummy", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_7regions_53dummy, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_7regions_52dummy},
  {"expected", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_7regions_7regions_55expected, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_7regions_7regions_54expected},
  {"recenter", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_57recenter, METH_O, __pyx_doc_6gnocis_7regions_7regions_56recenter},
  {"deltaResize", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_59deltaResize, METH_O, __pyx_doc_6gnocis_7regions_7regions_58deltaResize},
  {"rename", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_61rename, METH_O, __pyx_doc_6gnocis_7regions_7regions_60rename},
  {"sort", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_63sort, METH_NOARGS, __pyx_doc_6gnocis_7regions_7regions_62sort},
  {"extractSequences", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_65extractSequences, METH_O, __pyx_doc_6gnocis_7regions_7regions_64extractSequences},
  {"extractSequencesFrom2bit", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_67extractSequencesFrom2bit, METH_O, __pyx_doc_6gnocis_7regions_7regions_66extractSequencesFrom2bit},
  {"extractSequencesFromFASTA", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_69extractSequencesFromFASTA, METH_O, __pyx_doc_6gnocis_7regions_7regions_68extractSequencesFromFASTA},
  {"extract", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_71extract, METH_O, __pyx_doc_6gnocis_7regions_7regions_70extract},
  {"saveGFF", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_73saveGFF, METH_O, __pyx_doc_6gnocis_7regions_7regions_72saveGFF},
  {"saveBED", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_75saveBED, METH_O, __pyx_doc_6gnocis_7regions_7regions_74saveBED},
  {"saveCoordinateList", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_77saveCoordinateList, METH_O, __pyx_doc_6gnocis_7regions_7regions_76saveCoordinateList},
  {"printStatistics", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_79printStatistics, METH_NOARGS, __pyx_doc_6gnocis_7regions_7regions_78printStatistics},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_81__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6gnocis_7regions_7regions_83__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gnocis_7regions_regions[] = {
  {(char *)"name", __pyx_getprop_6gnocis_7regions_7regions_name, __pyx_setprop_6gnocis_7regions_7regions_name, (char *)0, 0},
  {(char *)"regions", __pyx_getprop_6gnocis_7regions_7regions_regions, __pyx_setprop_6gnocis_7regions_7regions_regions, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number_regions = {
  __pyx_pw_6gnocis_7regions_7regions_19__add__, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  __pyx_pw_6gnocis_7regions_7regions_23__and__, /*nb_and*/
  __pyx_pw_6gnocis_7regions_7regions_25__xor__, /*nb_xor*/
  __pyx_pw_6gnocis_7regions_7regions_21__or__, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_regions = {
  __pyx_pw_6gnocis_7regions_7regions_7__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6gnocis_7regions_regions, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_regions = {
  __pyx_pw_6gnocis_7regions_7regions_7__len__, /*mp_length*/
  __pyx_pw_6gnocis_7regions_7regions_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6gnocis_7regions_regions = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.regions", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions_regions), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions_regions, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6gnocis_7regions_7regions_17__repr__, /*tp_repr*/
  &__pyx_tp_as_number_regions, /*tp_as_number*/
  &__pyx_tp_as_sequence_regions, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_regions, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6gnocis_7regions_7regions_9__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n\tThe `regions` class represents a set of regions.\n\t\n\t:param name: Name for the sequence set.\n\t:param rgns: List of regions to include.\n\t:param initialSort: Whether or not to sort the regions (default is `True`).\n\t\n\t:type name: str\n\t:type rgns: list\n\t:type initialSort: bool\n\t\n\tFor a region set `RS`, `len(RS)` gives the number of regions, `RS[i]` gives the `i`th region in the set, and `[ r for r in RS ]` lists regions in the set. For sets `A` and `B`, `A+B` gives a sorted region set of regions from both sets, `A|B` gives the merged set, `A&B` the conjunction, and `A^B` the regions of `A` with `B` excluded.\n\t", /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions_regions, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions_regions, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_6gnocis_7regions_7regions_3__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gnocis_7regions_regions, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gnocis_7regions_regions, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gnocis_7regions_7regions_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions_regions, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct__bstr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct__bstr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct__bstr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct__bstr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct__bstr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct__bstr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct__bstr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct__bstr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct__bstr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct__bstr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct__bstr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct__bstr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_6gnocis_7regions_regions *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct__bstr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct__bstr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct__bstr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct__bstr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct__bstr, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct__bstr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct__bstr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_1_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_1_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_other);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *)o;
  if (p->__pyx_v_other) {
    e = (*v)(p->__pyx_v_other, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision *)o;
  tmp = ((PyObject*)p->__pyx_v_other);
  p->__pyx_v_other = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_6gnocis_7regions_regions *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_2_nucleotidePrecision", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_3_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_3_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_3_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_3_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_3_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_4_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_4_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_4_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_4_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_4_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_4_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_4_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_4_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_4_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_4_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_4_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_4_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_4_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_4_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_5_dummy[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_5_dummy = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_5_dummy(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_5_dummy > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_5_dummy[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_5_dummy];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_5_dummy(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_seqLens);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_5_dummy < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_5_dummy[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_5_dummy++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_5_dummy(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *)o;
  if (p->__pyx_v_seqLens) {
    e = (*v)(p->__pyx_v_seqLens, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_5_dummy(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy *)o;
  tmp = ((PyObject*)p->__pyx_v_seqLens);
  p->__pyx_v_seqLens = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_5_dummy = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_5_dummy", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_5_dummy), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_5_dummy, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_5_dummy, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_5_dummy, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_5_dummy, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_6_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_6_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_6_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_6_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_6_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_6_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_6_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_6_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_6_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_6_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_6_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_6_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_6_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_6_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_6_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_6_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_6_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_7_printStatistics[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_7_printStatistics = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_7_printStatistics(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_7_printStatistics > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_7_printStatistics[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_7_printStatistics];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_7_printStatistics(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_7_printStatistics < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_7_printStatistics[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_7_printStatistics++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_7_printStatistics(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_7_printStatistics(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_6gnocis_7regions_regions *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_7_printStatistics = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_7_printStatistics", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_7_printStatistics), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_7_printStatistics, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_7_printStatistics, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_7_printStatistics, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_7_printStatistics, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_8_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_8_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_8_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_8_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_8_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_8_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_8_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_8_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_8_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_8_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_8_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_8_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_8_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_8_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_8_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_8_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_8_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_9_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_9_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_9_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_9_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_9_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_9_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_9_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_9_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_9_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_9_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_9_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_9_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_9_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_9_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_9_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_9_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_9_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_10_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_10_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_10_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_10_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_10_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_10_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_10_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_10_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_10_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_10_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_10_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_10_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_10_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_10_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_10_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_10_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_10_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_11_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_11_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_11_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_11_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_11_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_11_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_11_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_11_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_11_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_11_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_11_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_11_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_11_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_11_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_11_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_11_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_11_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_12_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_12_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_12_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_12_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_12_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_12_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_12_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_12_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_12_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_12_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_12_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_12_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_12_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_12_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_12_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_12_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_12_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_f);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *)o;
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ *)o;
  tmp = ((PyObject*)p->__pyx_v_f);
  p->__pyx_v_f = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_13_loadBEDGZ", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_14_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_14_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_14_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_14_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_14_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_14_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_14_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_y);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_14_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_14_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_14_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_14_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_y) {
    e = (*v)(p->__pyx_v_y, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_14_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_14_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_14_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_14_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_14_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_14_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_f);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *)o;
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ *)o;
  tmp = ((PyObject*)p->__pyx_v_f);
  p->__pyx_v_f = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_15_loadGFFGZ", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_16_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_16_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_16_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_16_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_16_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_16_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_16_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_genexpr);
  Py_CLEAR(p->__pyx_v_x);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_16_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_16_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_16_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_16_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_genexpr) {
    e = (*v)(p->__pyx_v_genexpr, a); if (e) return e;
  }
  if (p->__pyx_v_x) {
    e = (*v)(p->__pyx_v_x, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_16_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_16_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_16_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_16_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_16_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_16_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_17_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_17_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_17_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_17_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_17_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_17_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_17_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_y);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_17_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_17_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_17_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_17_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_y) {
    e = (*v)(p->__pyx_v_y, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_17_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_17_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_17_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_17_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_17_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_17_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ps);
  Py_CLEAR(p->__pyx_v_rs);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *)o;
  if (p->__pyx_v_ps) {
    e = (*v)(p->__pyx_v_ps, a); if (e) return e;
  }
  if (p->__pyx_v_rs) {
    e = (*v)(p->__pyx_v_rs, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 *)o;
  tmp = ((PyObject*)p->__pyx_v_ps);
  p->__pyx_v_ps = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_rs);
  p->__pyx_v_rs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_18___pyx_lambda_funcdef_lambda17", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_19_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_19_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_19_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_19_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_19_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_19_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_19_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_19_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_19_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_19_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_19_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_19_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_19_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_19_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_19_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_19_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_19_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_20_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_20_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_20_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_20_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_20_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_20_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_20_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_20_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_20_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_20_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_20_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_20_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_20_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_20_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_20_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_20_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_20_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_21_measure[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_21_measure = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_21_measure(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_21_measure > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_21_measure[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_21_measure];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_21_measure(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ps);
  Py_CLEAR(p->__pyx_v_rs);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_21_measure < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_21_measure[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_21_measure++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_21_measure(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *)o;
  if (p->__pyx_v_ps) {
    e = (*v)(p->__pyx_v_ps, a); if (e) return e;
  }
  if (p->__pyx_v_rs) {
    e = (*v)(p->__pyx_v_rs, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_21_measure(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure *)o;
  tmp = ((PyObject*)p->__pyx_v_ps);
  p->__pyx_v_ps = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_rs);
  p->__pyx_v_rs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_21_measure = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_21_measure", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_21_measure), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_21_measure, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_21_measure, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_7regions___pyx_scope_struct_21_measure, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_21_measure, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_22_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_22_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_22_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_22_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_22_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_22_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_22_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_22_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_22_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_22_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_22_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_22_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_22_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_22_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_22_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_22_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_22_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *__pyx_freelist_6gnocis_7regions___pyx_scope_struct_23_genexpr[8];
static int __pyx_freecount_6gnocis_7regions___pyx_scope_struct_23_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_7regions___pyx_scope_struct_23_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_23_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_7regions___pyx_scope_struct_23_genexpr[--__pyx_freecount_6gnocis_7regions___pyx_scope_struct_23_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_23_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_7regions___pyx_scope_struct_23_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr)))) {
    __pyx_freelist_6gnocis_7regions___pyx_scope_struct_23_genexpr[__pyx_freecount_6gnocis_7regions___pyx_scope_struct_23_genexpr++] = ((struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_23_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *p = (struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_7regions___pyx_scope_struct_23_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.regions.__pyx_scope_struct_23_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_7regions___pyx_scope_struct_23_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_7regions___pyx_scope_struct_23_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_7regions___pyx_scope_struct_23_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_7regions___pyx_scope_struct_23_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_regions(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_regions},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "regions",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_kp_u_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 1, 0, 0},
  {&__pyx_kp_u_2f, __pyx_k_2f, sizeof(__pyx_k_2f), 0, 1, 0, 0},
  {&__pyx_kp_u_2f_2, __pyx_k_2f_2, sizeof(__pyx_k_2f_2), 0, 1, 0, 0},
  {&__pyx_kp_u_BED_file, __pyx_k_BED_file, sizeof(__pyx_k_BED_file), 0, 1, 0, 0},
  {&__pyx_kp_u_BED_file_syntax_error, __pyx_k_BED_file_syntax_error, sizeof(__pyx_k_BED_file_syntax_error), 0, 1, 0, 0},
  {&__pyx_n_s_BytesIO, __pyx_k_BytesIO, sizeof(__pyx_k_BytesIO), 0, 0, 1, 1},
  {&__pyx_n_s_CI, __pyx_k_CI, sizeof(__pyx_k_CI), 0, 0, 1, 1},
  {&__pyx_n_s_CVModelPredictions, __pyx_k_CVModelPredictions, sizeof(__pyx_k_CVModelPredictions), 0, 0, 1, 1},
  {&__pyx_kp_u_Coordinate_list_file, __pyx_k_Coordinate_list_file, sizeof(__pyx_k_Coordinate_list_file), 0, 1, 0, 0},
  {&__pyx_kp_u_Coordinate_list_syntax_error, __pyx_k_Coordinate_list_syntax_error, sizeof(__pyx_k_Coordinate_list_syntax_error), 0, 1, 0, 0},
  {&__pyx_n_u_End, __pyx_k_End, sizeof(__pyx_k_End), 0, 1, 0, 1},
  {&__pyx_kp_u_Feature_s, __pyx_k_Feature_s, sizeof(__pyx_k_Feature_s), 0, 1, 0, 0},
  {&__pyx_kp_u_GFF_GZ_file, __pyx_k_GFF_GZ_file, sizeof(__pyx_k_GFF_GZ_file), 0, 1, 0, 0},
  {&__pyx_kp_u_GFF_file, __pyx_k_GFF_file, sizeof(__pyx_k_GFF_file), 0, 1, 0, 0},
  {&__pyx_n_s_HTML, __pyx_k_HTML, sizeof(__pyx_k_HTML), 0, 0, 1, 1},
  {&__pyx_n_s_IPython_core_display, __pyx_k_IPython_core_display, sizeof(__pyx_k_IPython_core_display), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x11, __pyx_k_Incompatible_checksums_s_vs_0x11, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x11), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x9d, __pyx_k_Incompatible_checksums_s_vs_0x9d, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x9d), 0, 0, 1, 0},
  {&__pyx_n_u_Length, __pyx_k_Length, sizeof(__pyx_k_Length), 0, 1, 0, 1},
  {&__pyx_kp_u_Mean_length, __pyx_k_Mean_length, sizeof(__pyx_k_Mean_length), 0, 1, 0, 0},
  {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
  {&__pyx_kp_u_Nucleotide_precision, __pyx_k_Nucleotide_precision, sizeof(__pyx_k_Nucleotide_precision), 0, 1, 0, 0},
  {&__pyx_kp_u_Nucleotide_region_F1_score, __pyx_k_Nucleotide_region_F1_score, sizeof(__pyx_k_Nucleotide_region_F1_score), 0, 1, 0, 0},
  {&__pyx_kp_u_Overlap_precision, __pyx_k_Overlap_precision, sizeof(__pyx_k_Overlap_precision), 0, 1, 0, 0},
  {&__pyx_kp_u_Overlap_sensitivity, __pyx_k_Overlap_sensitivity, sizeof(__pyx_k_Overlap_sensitivity), 0, 1, 0, 0},
  {&__pyx_n_s_PercentFormatter, __pyx_k_PercentFormatter, sizeof(__pyx_k_PercentFormatter), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_u_Predictions, __pyx_k_Predictions, sizeof(__pyx_k_Predictions), 0, 1, 0, 1},
  {&__pyx_kp_u_Region, __pyx_k_Region, sizeof(__pyx_k_Region), 0, 1, 0, 0},
  {&__pyx_kp_u_Region_sequences_s, __pyx_k_Region_sequences_s, sizeof(__pyx_k_Region_sequences_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Region_set_s, __pyx_k_Region_set_s, sizeof(__pyx_k_Region_set_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Region_set_statistics, __pyx_k_Region_set_statistics, sizeof(__pyx_k_Region_set_statistics), 0, 1, 0, 0},
  {&__pyx_kp_u_Regions_d, __pyx_k_Regions_d, sizeof(__pyx_k_Regions_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Regions_set, __pyx_k_Regions_set, sizeof(__pyx_k_Regions_set), 0, 1, 0, 0},
  {&__pyx_kp_u_Score_f, __pyx_k_Score_f, sizeof(__pyx_k_Score_f), 0, 1, 0, 0},
  {&__pyx_kp_u_Seq, __pyx_k_Seq, sizeof(__pyx_k_Seq), 0, 1, 0, 0},
  {&__pyx_kp_u_Source_s, __pyx_k_Source_s, sizeof(__pyx_k_Source_s), 0, 1, 0, 0},
  {&__pyx_n_u_Start, __pyx_k_Start, sizeof(__pyx_k_Start), 0, 1, 0, 1},
  {&__pyx_n_u_Strand, __pyx_k_Strand, sizeof(__pyx_k_Strand), 0, 1, 0, 1},
  {&__pyx_kp_u_Table, __pyx_k_Table, sizeof(__pyx_k_Table), 0, 1, 0, 0},
  {&__pyx_kp_u_Total_length_d_nt, __pyx_k_Total_length_d_nt, sizeof(__pyx_k_Total_length_d_nt), 0, 1, 0, 0},
  {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
  {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
  {&__pyx_kp_u__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 1, 0, 0},
  {&__pyx_kp_u__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 1, 0, 0},
  {&__pyx_kp_u__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 1, 0, 0},
  {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
  {&__pyx_kp_u__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 1, 0, 0},
  {&__pyx_kp_u__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 1, 0, 0},
  {&__pyx_kp_u__19, __pyx_k__19, sizeof(__pyx_k__19), 0, 1, 0, 0},
  {&__pyx_kp_u__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 1, 0, 0},
  {&__pyx_kp_u__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 1, 0, 0},
  {&__pyx_n_s__27, __pyx_k__27, sizeof(__pyx_k__27), 0, 0, 1, 1},
  {&__pyx_kp_u__28, __pyx_k__28, sizeof(__pyx_k__28), 0, 1, 0, 0},
  {&__pyx_kp_u__29, __pyx_k__29, sizeof(__pyx_k__29), 0, 1, 0, 0},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_n_s__30, __pyx_k__30, sizeof(__pyx_k__30), 0, 0, 1, 1},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
  {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
  {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
  {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
  {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
  {&__pyx_n_s_align, __pyx_k_align, sizeof(__pyx_k_align), 0, 0, 1, 1},
  {&__pyx_n_u_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 1, 0, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_as_dict, __pyx_k_as_dict, sizeof(__pyx_k_as_dict), 0, 0, 1, 1},
  {&__pyx_n_s_ax, __pyx_k_ax, sizeof(__pyx_k_ax), 0, 0, 1, 1},
  {&__pyx_n_s_b64encode, __pyx_k_b64encode, sizeof(__pyx_k_b64encode), 0, 0, 1, 1},
  {&__pyx_n_s_bar, __pyx_k_bar, sizeof(__pyx_k_bar), 0, 0, 1, 1},
  {&__pyx_n_s_base64, __pyx_k_base64, sizeof(__pyx_k_base64), 0, 0, 1, 1},
  {&__pyx_n_s_bboxAnchorTo, __pyx_k_bboxAnchorTo, sizeof(__pyx_k_bboxAnchorTo), 0, 0, 1, 1},
  {&__pyx_n_s_bbox_to_anchor, __pyx_k_bbox_to_anchor, sizeof(__pyx_k_bbox_to_anchor), 0, 0, 1, 1},
  {&__pyx_n_s_bio, __pyx_k_bio, sizeof(__pyx_k_bio), 0, 0, 1, 1},
  {&__pyx_kp_u_bootstrap_sample_of, __pyx_k_bootstrap_sample_of, sizeof(__pyx_k_bootstrap_sample_of), 0, 1, 0, 0},
  {&__pyx_n_u_bottom, __pyx_k_bottom, sizeof(__pyx_k_bottom), 0, 1, 0, 1},
  {&__pyx_kp_u_bp, __pyx_k_bp, sizeof(__pyx_k_bp), 0, 1, 0, 0},
  {&__pyx_n_s_bstr, __pyx_k_bstr, sizeof(__pyx_k_bstr), 0, 0, 1, 1},
  {&__pyx_n_s_bstr_locals_genexpr, __pyx_k_bstr_locals_genexpr, sizeof(__pyx_k_bstr_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_bw, __pyx_k_bw, sizeof(__pyx_k_bw), 0, 0, 1, 1},
  {&__pyx_n_u_center, __pyx_k_center, sizeof(__pyx_k_center), 0, 1, 0, 1},
  {&__pyx_n_s_choice, __pyx_k_choice, sizeof(__pyx_k_choice), 0, 0, 1, 1},
  {&__pyx_n_u_chr, __pyx_k_chr, sizeof(__pyx_k_chr), 0, 1, 0, 1},
  {&__pyx_n_s_ci, __pyx_k_ci, sizeof(__pyx_k_ci), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_common, __pyx_k_common, sizeof(__pyx_k_common), 0, 0, 1, 1},
  {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
  {&__pyx_n_s_cv, __pyx_k_cv, sizeof(__pyx_k_cv), 0, 0, 1, 1},
  {&__pyx_n_s_cx, __pyx_k_cx, sizeof(__pyx_k_cx), 0, 0, 1, 1},
  {&__pyx_n_u_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 1, 0, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_display, __pyx_k_display, sizeof(__pyx_k_display), 0, 0, 1, 1},
  {&__pyx_n_s_dropChr, __pyx_k_dropChr, sizeof(__pyx_k_dropChr), 0, 0, 1, 1},
  {&__pyx_kp_u_dummy, __pyx_k_dummy, sizeof(__pyx_k_dummy), 0, 1, 0, 0},
  {&__pyx_n_s_dummy_2, __pyx_k_dummy_2, sizeof(__pyx_k_dummy_2), 0, 0, 1, 1},
  {&__pyx_n_s_dummy_locals_genexpr, __pyx_k_dummy_locals_genexpr, sizeof(__pyx_k_dummy_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_encoded, __pyx_k_encoded, sizeof(__pyx_k_encoded), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_u_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 1, 0, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_exclusion, __pyx_k_exclusion, sizeof(__pyx_k_exclusion), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_u_ext, __pyx_k_ext, sizeof(__pyx_k_ext), 0, 1, 0, 1},
  {&__pyx_n_s_extractSequences, __pyx_k_extractSequences, sizeof(__pyx_k_extractSequences), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_u_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 1, 0, 1},
  {&__pyx_n_s_fancybox, __pyx_k_fancybox, sizeof(__pyx_k_fancybox), 0, 0, 1, 1},
  {&__pyx_n_s_feature, __pyx_k_feature, sizeof(__pyx_k_feature), 0, 0, 1, 1},
  {&__pyx_n_u_feature, __pyx_k_feature, sizeof(__pyx_k_feature), 0, 1, 0, 1},
  {&__pyx_n_s_fig, __pyx_k_fig, sizeof(__pyx_k_fig), 0, 0, 1, 1},
  {&__pyx_n_s_figsize, __pyx_k_figsize, sizeof(__pyx_k_figsize), 0, 0, 1, 1},
  {&__pyx_n_s_flt, __pyx_k_flt, sizeof(__pyx_k_flt), 0, 0, 1, 1},
  {&__pyx_n_s_fltName, __pyx_k_fltName, sizeof(__pyx_k_fltName), 0, 0, 1, 1},
  {&__pyx_n_s_fontsize, __pyx_k_fontsize, sizeof(__pyx_k_fontsize), 0, 0, 1, 1},
  {&__pyx_n_s_fontsizeAxis, __pyx_k_fontsizeAxis, sizeof(__pyx_k_fontsizeAxis), 0, 0, 1, 1},
  {&__pyx_n_s_fontsizeLabels, __pyx_k_fontsizeLabels, sizeof(__pyx_k_fontsizeLabels), 0, 0, 1, 1},
  {&__pyx_n_s_fontsizeLegend, __pyx_k_fontsizeLegend, sizeof(__pyx_k_fontsizeLegend), 0, 0, 1, 1},
  {&__pyx_n_s_fontsizeValues, __pyx_k_fontsizeValues, sizeof(__pyx_k_fontsizeValues), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_kp_u_from_sequence_stream, __pyx_k_from_sequence_stream, sizeof(__pyx_k_from_sequence_stream), 0, 1, 0, 0},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_genome, __pyx_k_genome, sizeof(__pyx_k_genome), 0, 0, 1, 1},
  {&__pyx_n_u_genome, __pyx_k_genome, sizeof(__pyx_k_genome), 0, 1, 0, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_getvalue, __pyx_k_getvalue, sizeof(__pyx_k_getvalue), 0, 0, 1, 1},
  {&__pyx_n_u_ggplot, __pyx_k_ggplot, sizeof(__pyx_k_ggplot), 0, 1, 0, 1},
  {&__pyx_n_s_gnocis_regions, __pyx_k_gnocis_regions, sizeof(__pyx_k_gnocis_regions), 0, 0, 1, 1},
  {&__pyx_kp_s_gnocis_regions_pyx, __pyx_k_gnocis_regions_pyx, sizeof(__pyx_k_gnocis_regions_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_group, __pyx_k_group, sizeof(__pyx_k_group), 0, 0, 1, 1},
  {&__pyx_n_u_group, __pyx_k_group, sizeof(__pyx_k_group), 0, 1, 0, 1},
  {&__pyx_n_s_gzip, __pyx_k_gzip, sizeof(__pyx_k_gzip), 0, 0, 1, 1},
  {&__pyx_n_s_hasTestRegions, __pyx_k_hasTestRegions, sizeof(__pyx_k_hasTestRegions), 0, 0, 1, 1},
  {&__pyx_n_s_horizontalalignment, __pyx_k_horizontalalignment, sizeof(__pyx_k_horizontalalignment), 0, 0, 1, 1},
  {&__pyx_n_s_html, __pyx_k_html, sizeof(__pyx_k_html), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_kp_u_img_src_data_image_png_base64_s, __pyx_k_img_src_data_image_png_base64_s, sizeof(__pyx_k_img_src_data_image_png_base64_s), 0, 1, 0, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_initialSort, __pyx_k_initialSort, sizeof(__pyx_k_initialSort), 0, 0, 1, 1},
  {&__pyx_n_s_intersection, __pyx_k_intersection, sizeof(__pyx_k_intersection), 0, 0, 1, 1},
  {&__pyx_n_s_io, __pyx_k_io, sizeof(__pyx_k_io), 0, 0, 1, 1},
  {&__pyx_n_s_ioputil, __pyx_k_ioputil, sizeof(__pyx_k_ioputil), 0, 0, 1, 1},
  {&__pyx_n_s_isPercent, __pyx_k_isPercent, sizeof(__pyx_k_isPercent), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_u_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 1, 0, 1},
  {&__pyx_n_s_label, __pyx_k_label, sizeof(__pyx_k_label), 0, 0, 1, 1},
  {&__pyx_n_s_legend, __pyx_k_legend, sizeof(__pyx_k_legend), 0, 0, 1, 1},
  {&__pyx_n_s_legendLoc, __pyx_k_legendLoc, sizeof(__pyx_k_legendLoc), 0, 0, 1, 1},
  {&__pyx_n_s_loadBED, __pyx_k_loadBED, sizeof(__pyx_k_loadBED), 0, 0, 1, 1},
  {&__pyx_n_s_loadBEDGZ, __pyx_k_loadBEDGZ, sizeof(__pyx_k_loadBEDGZ), 0, 0, 1, 1},
  {&__pyx_n_s_loadBEDGZ_locals_genexpr, __pyx_k_loadBEDGZ_locals_genexpr, sizeof(__pyx_k_loadBEDGZ_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_loadCoordinateList, __pyx_k_loadCoordinateList, sizeof(__pyx_k_loadCoordinateList), 0, 0, 1, 1},
  {&__pyx_n_s_loadGFF, __pyx_k_loadGFF, sizeof(__pyx_k_loadGFF), 0, 0, 1, 1},
  {&__pyx_n_s_loadGFFGZ, __pyx_k_loadGFFGZ, sizeof(__pyx_k_loadGFFGZ), 0, 0, 1, 1},
  {&__pyx_n_s_loadGFFGZ_locals_genexpr, __pyx_k_loadGFFGZ_locals_genexpr, sizeof(__pyx_k_loadGFFGZ_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_loadGFFGZ_locals_genexpr_locals, __pyx_k_loadGFFGZ_locals_genexpr_locals, sizeof(__pyx_k_loadGFFGZ_locals_genexpr_locals), 0, 0, 1, 1},
  {&__pyx_n_s_loc, __pyx_k_loc, sizeof(__pyx_k_loc), 0, 0, 1, 1},
  {&__pyx_n_s_lower, __pyx_k_lower, sizeof(__pyx_k_lower), 0, 0, 1, 1},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mainRegions, __pyx_k_mainRegions, sizeof(__pyx_k_mainRegions), 0, 0, 1, 1},
  {&__pyx_n_u_markers, __pyx_k_markers, sizeof(__pyx_k_markers), 0, 1, 0, 1},
  {&__pyx_n_s_matplotlib_pyplot, __pyx_k_matplotlib_pyplot, sizeof(__pyx_k_matplotlib_pyplot), 0, 0, 1, 1},
  {&__pyx_n_s_matplotlib_ticker, __pyx_k_matplotlib_ticker, sizeof(__pyx_k_matplotlib_ticker), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_mean, __pyx_k_mean, sizeof(__pyx_k_mean), 0, 0, 1, 1},
  {&__pyx_n_s_measure, __pyx_k_measure, sizeof(__pyx_k_measure), 0, 0, 1, 1},
  {&__pyx_n_s_measureName, __pyx_k_measureName, sizeof(__pyx_k_measureName), 0, 0, 1, 1},
  {&__pyx_n_s_merge, __pyx_k_merge, sizeof(__pyx_k_merge), 0, 0, 1, 1},
  {&__pyx_n_s_merge_locals_lambda, __pyx_k_merge_locals_lambda, sizeof(__pyx_k_merge_locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_models, __pyx_k_models, sizeof(__pyx_k_models), 0, 0, 1, 1},
  {&__pyx_n_s_mtick, __pyx_k_mtick, sizeof(__pyx_k_mtick), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_nctable, __pyx_k_nctable, sizeof(__pyx_k_nctable), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_u_not_overlapping_with, __pyx_k_not_overlapping_with, sizeof(__pyx_k_not_overlapping_with), 0, 1, 0, 0},
  {&__pyx_kp_u_nt, __pyx_k_nt, sizeof(__pyx_k_nt), 0, 1, 0, 0},
  {&__pyx_kp_u_nt_max, __pyx_k_nt_max, sizeof(__pyx_k_nt_max), 0, 1, 0, 0},
  {&__pyx_kp_u_nt_min, __pyx_k_nt_min, sizeof(__pyx_k_nt_min), 0, 1, 0, 0},
  {&__pyx_n_s_nucleotidePrecisionBarplot, __pyx_k_nucleotidePrecisionBarplot, sizeof(__pyx_k_nucleotidePrecisionBarplot), 0, 0, 1, 1},
  {&__pyx_n_s_nucleotidePrecisionBarplot_local, __pyx_k_nucleotidePrecisionBarplot_local, sizeof(__pyx_k_nucleotidePrecisionBarplot_local), 0, 0, 1, 1},
  {&__pyx_n_s_nucleotidePrecisionBarplot_local_2, __pyx_k_nucleotidePrecisionBarplot_local_2, sizeof(__pyx_k_nucleotidePrecisionBarplot_local_2), 0, 0, 1, 1},
  {&__pyx_n_s_nucleotidePrecision_locals_genex, __pyx_k_nucleotidePrecision_locals_genex, sizeof(__pyx_k_nucleotidePrecision_locals_genex), 0, 0, 1, 1},
  {&__pyx_n_s_nucleotideRegionF1Barplot, __pyx_k_nucleotideRegionF1Barplot, sizeof(__pyx_k_nucleotideRegionF1Barplot), 0, 0, 1, 1},
  {&__pyx_n_s_nucleotideRegionF1Barplot_locals, __pyx_k_nucleotideRegionF1Barplot_locals, sizeof(__pyx_k_nucleotideRegionF1Barplot_locals), 0, 0, 1, 1},
  {&__pyx_n_s_nucleotideRegionF1Barplot_locals_2, __pyx_k_nucleotideRegionF1Barplot_locals_2, sizeof(__pyx_k_nucleotideRegionF1Barplot_locals_2), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_outpath, __pyx_k_outpath, sizeof(__pyx_k_outpath), 0, 0, 1, 1},
  {&__pyx_n_s_overlap, __pyx_k_overlap, sizeof(__pyx_k_overlap), 0, 0, 1, 1},
  {&__pyx_n_s_overlapPrecisionBarplot, __pyx_k_overlapPrecisionBarplot, sizeof(__pyx_k_overlapPrecisionBarplot), 0, 0, 1, 1},
  {&__pyx_n_s_overlapPrecisionBarplot_locals_l, __pyx_k_overlapPrecisionBarplot_locals_l, sizeof(__pyx_k_overlapPrecisionBarplot_locals_l), 0, 0, 1, 1},
  {&__pyx_n_s_overlapSensitivityBarplot, __pyx_k_overlapSensitivityBarplot, sizeof(__pyx_k_overlapSensitivityBarplot), 0, 0, 1, 1},
  {&__pyx_n_s_overlapSensitivityBarplot_locals, __pyx_k_overlapSensitivityBarplot_locals, sizeof(__pyx_k_overlapSensitivityBarplot_locals), 0, 0, 1, 1},
  {&__pyx_kp_u_overlapping_with, __pyx_k_overlapping_with, sizeof(__pyx_k_overlapping_with), 0, 1, 0, 0},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_parts, __pyx_k_parts, sizeof(__pyx_k_parts), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_plt, __pyx_k_plt, sizeof(__pyx_k_plt), 0, 0, 1, 1},
  {&__pyx_n_u_png, __pyx_k_png, sizeof(__pyx_k_png), 0, 1, 0, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_positive, __pyx_k_positive, sizeof(__pyx_k_positive), 0, 0, 1, 1},
  {&__pyx_n_s_predictionBarplot, __pyx_k_predictionBarplot, sizeof(__pyx_k_predictionBarplot), 0, 0, 1, 1},
  {&__pyx_n_s_predictionBarplot_locals_lambda, __pyx_k_predictionBarplot_locals_lambda, sizeof(__pyx_k_predictionBarplot_locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_predictionSets, __pyx_k_predictionSets, sizeof(__pyx_k_predictionSets), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_printStatistics_locals_genexpr, __pyx_k_printStatistics_locals_genexpr, sizeof(__pyx_k_printStatistics_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_prs, __pyx_k_prs, sizeof(__pyx_k_prs), 0, 0, 1, 1},
  {&__pyx_n_s_ps, __pyx_k_ps, sizeof(__pyx_k_ps), 0, 0, 1, 1},
  {&__pyx_n_s_ps_2, __pyx_k_ps_2, sizeof(__pyx_k_ps_2), 0, 0, 1, 1},
  {&__pyx_n_s_psi, __pyx_k_psi, sizeof(__pyx_k_psi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_region, __pyx_k_pyx_unpickle_region, sizeof(__pyx_k_pyx_unpickle_region), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_regions, __pyx_k_pyx_unpickle_regions, sizeof(__pyx_k_pyx_unpickle_regions), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_n_u_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 1, 0, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_ratio, __pyx_k_ratio, sizeof(__pyx_k_ratio), 0, 0, 1, 1},
  {&__pyx_n_s_raw_gff, __pyx_k_raw_gff, sizeof(__pyx_k_raw_gff), 0, 0, 1, 1},
  {&__pyx_n_u_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 1, 0, 1},
  {&__pyx_kp_u_recentered, __pyx_k_recentered, sizeof(__pyx_k_recentered), 0, 1, 0, 0},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_region, __pyx_k_region, sizeof(__pyx_k_region), 0, 0, 1, 1},
  {&__pyx_n_s_regionBarplot, __pyx_k_regionBarplot, sizeof(__pyx_k_regionBarplot), 0, 0, 1, 1},
  {&__pyx_n_s_regionSets, __pyx_k_regionSets, sizeof(__pyx_k_regionSets), 0, 0, 1, 1},
  {&__pyx_n_s_regions, __pyx_k_regions, sizeof(__pyx_k_regions), 0, 0, 1, 1},
  {&__pyx_kp_u_regions_2, __pyx_k_regions_2, sizeof(__pyx_k_regions_2), 0, 1, 0, 0},
  {&__pyx_n_s_repeats, __pyx_k_repeats, sizeof(__pyx_k_repeats), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_repr_html, __pyx_k_repr_html, sizeof(__pyx_k_repr_html), 0, 0, 1, 1},
  {&__pyx_n_s_returnHTML, __pyx_k_returnHTML, sizeof(__pyx_k_returnHTML), 0, 0, 1, 1},
  {&__pyx_n_s_rgns, __pyx_k_rgns, sizeof(__pyx_k_rgns), 0, 0, 1, 1},
  {&__pyx_n_s_rotation, __pyx_k_rotation, sizeof(__pyx_k_rotation), 0, 0, 1, 1},
  {&__pyx_n_s_rs, __pyx_k_rs, sizeof(__pyx_k_rs), 0, 0, 1, 1},
  {&__pyx_n_s_rsi, __pyx_k_rsi, sizeof(__pyx_k_rsi), 0, 0, 1, 1},
  {&__pyx_n_s_savefig, __pyx_k_savefig, sizeof(__pyx_k_savefig), 0, 0, 1, 1},
  {&__pyx_n_s_scaling, __pyx_k_scaling, sizeof(__pyx_k_scaling), 0, 0, 1, 1},
  {&__pyx_n_s_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 0, 1, 1},
  {&__pyx_n_u_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 1, 0, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_seq, __pyx_k_seq, sizeof(__pyx_k_seq), 0, 0, 1, 1},
  {&__pyx_n_u_seq, __pyx_k_seq, sizeof(__pyx_k_seq), 0, 1, 0, 1},
  {&__pyx_n_s_sequenceLengths, __pyx_k_sequenceLengths, sizeof(__pyx_k_sequenceLengths), 0, 0, 1, 1},
  {&__pyx_n_s_sequences, __pyx_k_sequences, sizeof(__pyx_k_sequences), 0, 0, 1, 1},
  {&__pyx_n_s_set_major_formatter, __pyx_k_set_major_formatter, sizeof(__pyx_k_set_major_formatter), 0, 0, 1, 1},
  {&__pyx_n_s_set_xticklabels, __pyx_k_set_xticklabels, sizeof(__pyx_k_set_xticklabels), 0, 0, 1, 1},
  {&__pyx_n_s_set_xticks, __pyx_k_set_xticks, sizeof(__pyx_k_set_xticks), 0, 0, 1, 1},
  {&__pyx_n_s_set_ylabel, __pyx_k_set_ylabel, sizeof(__pyx_k_set_ylabel), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_showLegend, __pyx_k_showLegend, sizeof(__pyx_k_showLegend), 0, 0, 1, 1},
  {&__pyx_n_s_showValues, __pyx_k_showValues, sizeof(__pyx_k_showValues), 0, 0, 1, 1},
  {&__pyx_n_s_shuffle, __pyx_k_shuffle, sizeof(__pyx_k_shuffle), 0, 0, 1, 1},
  {&__pyx_n_s_slots, __pyx_k_slots, sizeof(__pyx_k_slots), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_sort_locals_lambda, __pyx_k_sort_locals_lambda, sizeof(__pyx_k_sort_locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
  {&__pyx_n_u_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 1, 0, 1},
  {&__pyx_n_s_sourceRegion, __pyx_k_sourceRegion, sizeof(__pyx_k_sourceRegion), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_kp_u_split_A, __pyx_k_split_A, sizeof(__pyx_k_split_A), 0, 1, 0, 0},
  {&__pyx_kp_u_split_B, __pyx_k_split_B, sizeof(__pyx_k_split_B), 0, 1, 0, 0},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_u_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 1, 0, 1},
  {&__pyx_n_s_startswith, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {&__pyx_n_s_statfun, __pyx_k_statfun, sizeof(__pyx_k_statfun), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_strand, __pyx_k_strand, sizeof(__pyx_k_strand), 0, 0, 1, 1},
  {&__pyx_n_u_strand, __pyx_k_strand, sizeof(__pyx_k_strand), 0, 1, 0, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_style, __pyx_k_style, sizeof(__pyx_k_style), 0, 0, 1, 1},
  {&__pyx_n_s_subplots, __pyx_k_subplots, sizeof(__pyx_k_subplots), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_table, __pyx_k_table, sizeof(__pyx_k_table), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_testRegions, __pyx_k_testRegions, sizeof(__pyx_k_testRegions), 0, 0, 1, 1},
  {&__pyx_n_s_text, __pyx_k_text, sizeof(__pyx_k_text), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tight_layout, __pyx_k_tight_layout, sizeof(__pyx_k_tight_layout), 0, 0, 1, 1},
  {&__pyx_n_s_uniform, __pyx_k_uniform, sizeof(__pyx_k_uniform), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_kp_u_upper_left, __pyx_k_upper_left, sizeof(__pyx_k_upper_left), 0, 1, 0, 0},
  {&__pyx_n_s_useSeq, __pyx_k_useSeq, sizeof(__pyx_k_useSeq), 0, 0, 1, 1},
  {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_vCI, __pyx_k_vCI, sizeof(__pyx_k_vCI), 0, 0, 1, 1},
  {&__pyx_n_s_vMean, __pyx_k_vMean, sizeof(__pyx_k_vMean), 0, 0, 1, 1},
  {&__pyx_n_s_verticalalignment, __pyx_k_verticalalignment, sizeof(__pyx_k_verticalalignment), 0, 0, 1, 1},
  {&__pyx_n_u_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 1, 0, 1},
  {&__pyx_n_s_width, __pyx_k_width, sizeof(__pyx_k_width), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_xticks, __pyx_k_xticks, sizeof(__pyx_k_xticks), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_yaxis, __pyx_k_yaxis, sizeof(__pyx_k_yaxis), 0, 0, 1, 1},
  {&__pyx_n_s_yerr, __pyx_k_yerr, sizeof(__pyx_k_yerr), 0, 0, 1, 1},
  {&__pyx_n_s_yticks, __pyx_k_yticks, sizeof(__pyx_k_yticks), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 193, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 275, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 440, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 648, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 679, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 683, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 683, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 904, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "gnocis/regions.pyx":56
 * 		if dropChr:
 * 			if seq.lower().startswith('chr'):
 * 				seq = seq[3:]             # <<<<<<<<<<<<<<
 * 		self.seq, self.start, self.end, self.strand, self.score = seq, start, end, strand, score
 * 		if start > end: self.start, self.end = end, start
 */
  __pyx_slice__2 = PySlice_New(__pyx_int_3, Py_None, Py_None); if (unlikely(!__pyx_slice__2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__2);
  __Pyx_GIVEREF(__pyx_slice__2);

  /* "gnocis/regions.pyx":579
 * 		elif isinstance(seqs, sequenceStream):
 * 			for cseq in seqs:
 * 				cSeqRegions = [ { 'start':r.start, 'end':r.end, 'seq':sequence(None, '', sourceRegion = r) } for r in self.regions if r.seq == cseq.name ]             # <<<<<<<<<<<<<<
 * 				if len(cSeqRegions) == 0: continue
 * 				ret.sequences += [ r['seq'] for r in cSeqRegions ]
 */
  __pyx_tuple__18 = PyTuple_Pack(2, Py_None, __pyx_kp_u__13); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "gnocis/regions.pyx":648
 * 		:type path: str
 * 		"""
 * 		with open(path, 'w') as fOut:             # <<<<<<<<<<<<<<
 * 			for r in self.regions:
 * 				fOut.write('%s\t%s\t%s\t%d\t%d\t%f\t%s\t.\t1\n'%(r.seq, r.source, r.feature, r.start, r.end + 1, r.score, '+' if r.strand else '-'))
 */
  __pyx_tuple__20 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "gnocis/regions.pyx":707
 * 			if l.startswith('#'): continue
 * 			parts = l.strip().split('\t')
 * 			if len(parts) < 3: raise Exception('BED file syntax error')             # <<<<<<<<<<<<<<
 * 			seq, start, end = parts[0:3]
 * 			strand = True
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_u_BED_file_syntax_error); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "gnocis/regions.pyx":819
 * 			e = l.strip().split(':')
 * 			if len(e) != 2:
 * 				raise Exception('Coordinate list syntax error')             # <<<<<<<<<<<<<<
 * 			pos = e[1].split('..')
 * 			if len(pos) != 2:
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_u_Coordinate_list_syntax_error); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "gnocis/regions.pyx":829
 * 
 * # Generates a barplot for a region-based measure
 * def regionBarplot(mainRegions, measure, measureName, testRegions = None, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeValues = 12, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = False, isPercent = False):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of a measure (lambda) for predicted regions.
 * 
 */
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_int_8, __pyx_int_8); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_float_0_, __pyx_float_neg_0_15); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "gnocis/regions.pyx":1105
 * 	:type showValues: bool, optional
 * 	"""
 * 	def measure(ps, rs):             # <<<<<<<<<<<<<<
 * 		p = sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)
 * 		r = len(rs.overlap(ps)) / len(rs)
 */
  __pyx_tuple__31 = PyTuple_Pack(7, __pyx_n_s_ps, __pyx_n_s_rs, __pyx_n_s_p, __pyx_n_s_r, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(2, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_measure, 1105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 1105, __pyx_L1_error)

  /* "gnocis/regions.pyx":51
 * 	"""
 * 
 * 	__slots__ = ('seq', 'start', 'end', 'strand', 'score', 'source', 'feature', 'markers', 'group', 'ext')             # <<<<<<<<<<<<<<
 * 
 * 	def __init__(self, seq, start, end, strand = True, score = 0.0, source = '.', feature = '.', group = '.', dropChr = True):
 */
  __pyx_tuple__33 = PyTuple_Pack(10, __pyx_n_u_seq, __pyx_n_u_start, __pyx_n_u_end, __pyx_n_u_strand, __pyx_n_u_score, __pyx_n_u_source, __pyx_n_u_feature, __pyx_n_u_markers, __pyx_n_u_group, __pyx_n_u_ext); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "gnocis/regions.pyx":687
 * 
 * # Loads regions from a BED file.
 * def loadBED(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(14, __pyx_n_s_path, __pyx_n_s_rgns, __pyx_n_s_l, __pyx_n_s_seq, __pyx_n_s_start, __pyx_n_s_end, __pyx_n_s_source, __pyx_n_s_feature, __pyx_n_s_group, __pyx_n_s_strand, __pyx_n_s_score, __pyx_n_s_parts, __pyx_n_s_rs, __pyx_n_s_f); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_loadBED, 687, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 687, __pyx_L1_error)

  /* "gnocis/regions.pyx":726
 * 
 * # Loads regions from a gzipped BED file.
 * def loadBEDGZ(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */
  __pyx_tuple__36 = PyTuple_Pack(17, __pyx_n_s_path, __pyx_n_s_rgns, __pyx_n_s_l, __pyx_n_s_y, __pyx_n_s_seq, __pyx_n_s_start, __pyx_n_s_end, __pyx_n_s_source, __pyx_n_s_feature, __pyx_n_s_group, __pyx_n_s_strand, __pyx_n_s_score, __pyx_n_s_parts, __pyx_n_s_rs, __pyx_n_s_f, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(1, 0, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_loadBEDGZ, 726, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 726, __pyx_L1_error)

  /* "gnocis/regions.pyx":765
 * 
 * # Loads regions from a GFF file.
 * def loadGFF(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */
  __pyx_tuple__38 = PyTuple_Pack(7, __pyx_n_s_path, __pyx_n_s_dropChr, __pyx_n_s_x, __pyx_n_s_rgns, __pyx_n_s_r, __pyx_n_s_rs, __pyx_n_s_f); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(2, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_loadGFF, 765, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 765, __pyx_L1_error)

  /* "gnocis/regions.pyx":788
 * 
 * # Loads regions from a gzipped GFF file.
 * def loadGFFGZ(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */
  __pyx_tuple__40 = PyTuple_Pack(10, __pyx_n_s_path, __pyx_n_s_dropChr, __pyx_n_s_f, __pyx_n_s_raw_gff, __pyx_n_s_rgns, __pyx_n_s_rs, __pyx_n_s_y, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_r); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_loadGFFGZ, 788, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 788, __pyx_L1_error)

  /* "gnocis/regions.pyx":805
 * 
 * # Loads regions from a coordinate list file.
 * def loadCoordinateList(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a coordinate list file (one region per line, with the format `seq:start..end`).
 * 
 */
  __pyx_tuple__42 = PyTuple_Pack(8, __pyx_n_s_path, __pyx_n_s_dropChr, __pyx_n_s_f, __pyx_n_s_rgns, __pyx_n_s_l, __pyx_n_s_e, __pyx_n_s_pos, __pyx_n_s_rs); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(2, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_loadCoordinateList, 805, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(0, 805, __pyx_L1_error)

  /* "gnocis/regions.pyx":829
 * 
 * # Generates a barplot for a region-based measure
 * def regionBarplot(mainRegions, measure, measureName, testRegions = None, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeValues = 12, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = False, isPercent = False):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of a measure (lambda) for predicted regions.
 * 
 */
  __pyx_tuple__44 = PyTuple_Pack(51, __pyx_n_s_mainRegions, __pyx_n_s_measure, __pyx_n_s_measureName, __pyx_n_s_testRegions, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_returnHTML, __pyx_n_s_fontsizeLabels, __pyx_n_s_fontsizeValues, __pyx_n_s_fontsizeLegend, __pyx_n_s_fontsizeAxis, __pyx_n_s_style, __pyx_n_s_showLegend, __pyx_n_s_bboxAnchorTo, __pyx_n_s_legendLoc, __pyx_n_s_showValues, __pyx_n_s_isPercent, __pyx_n_s_plt, __pyx_n_s_base64, __pyx_n_s_BytesIO, __pyx_n_s_display, __pyx_n_s_HTML, __pyx_n_s_mtick, __pyx_n_s_width, __pyx_n_s_bw, __pyx_n_s_fig, __pyx_n_s_ax, __pyx_n_s_scaling, __pyx_n_s_hasTestRegions, __pyx_n_s_psi, __pyx_n_s_ps, __pyx_n_s_x, __pyx_n_s_prs, __pyx_n_s_v, __pyx_n_s_vMean, __pyx_n_s_vCI, __pyx_n_s_cx, __pyx_n_s_m, __pyx_n_s_ci, __pyx_n_s_cv, __pyx_n_s_bio, __pyx_n_s_encoded, __pyx_n_s_html, __pyx_n_s_rsi, __pyx_n_s_rs, __pyx_n_s_ps_2, __pyx_n_s_cv, __pyx_n_s_cv, __pyx_n_s_rs, __pyx_n_s_i, __pyx_n_s_rs); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(17, 0, 51, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_regionBarplot, 829, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(0, 829, __pyx_L1_error)

  /* "gnocis/regions.pyx":944
 * 			plt.close('all')
 * 
 * def predictionBarplot(predictionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */
  __pyx_tuple__46 = PyTuple_Pack(12, __pyx_n_s_predictionSets, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_returnHTML, __pyx_n_s_fontsizeLabels, __pyx_n_s_fontsizeLegend, __pyx_n_s_fontsizeAxis, __pyx_n_s_style, __pyx_n_s_showLegend, __pyx_n_s_bboxAnchorTo, __pyx_n_s_legendLoc, __pyx_n_s_showValues); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(12, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_predictionBarplot, 944, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 944, __pyx_L1_error)

  /* "gnocis/regions.pyx":975
 * 	regionBarplot(mainRegions = predictionSets, measure = (lambda rs, _: len(rs)), measureName = 'Predictions', testRegions = None, figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)
 * 
 * def overlapSensitivityBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap sensitivity barplot.
 * 
 */
  __pyx_tuple__48 = PyTuple_Pack(13, __pyx_n_s_predictionSets, __pyx_n_s_regionSets, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_returnHTML, __pyx_n_s_fontsizeLabels, __pyx_n_s_fontsizeLegend, __pyx_n_s_fontsizeAxis, __pyx_n_s_style, __pyx_n_s_showLegend, __pyx_n_s_bboxAnchorTo, __pyx_n_s_legendLoc, __pyx_n_s_showValues); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(13, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_overlapSensitivityBarplot, 975, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 975, __pyx_L1_error)

  /* "gnocis/regions.pyx":1008
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(rs.overlap(ps)) / len(rs)), measureName = 'Overlap sensitivity', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def overlapPrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */
  __pyx_tuple__50 = PyTuple_Pack(13, __pyx_n_s_predictionSets, __pyx_n_s_regionSets, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_returnHTML, __pyx_n_s_fontsizeLabels, __pyx_n_s_fontsizeLegend, __pyx_n_s_fontsizeAxis, __pyx_n_s_style, __pyx_n_s_showLegend, __pyx_n_s_bboxAnchorTo, __pyx_n_s_legendLoc, __pyx_n_s_showValues); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(13, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_overlapPrecisionBarplot, 1008, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 1008, __pyx_L1_error)

  /* "gnocis/regions.pyx":1041
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(ps.overlap(rs)) / len(ps)), measureName = 'Overlap precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotidePrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction nucleotide precision barplot.
 * 
 */
  __pyx_tuple__52 = PyTuple_Pack(13, __pyx_n_s_predictionSets, __pyx_n_s_regionSets, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_returnHTML, __pyx_n_s_fontsizeLabels, __pyx_n_s_fontsizeLegend, __pyx_n_s_fontsizeAxis, __pyx_n_s_style, __pyx_n_s_showLegend, __pyx_n_s_bboxAnchorTo, __pyx_n_s_legendLoc, __pyx_n_s_showValues); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(13, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_nucleotidePrecisionBarplot, 1041, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 1041, __pyx_L1_error)

  /* "gnocis/regions.pyx":1074
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)), measureName = 'Nucleotide precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotideRegionF1Barplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of F1-measures, where the recall is based on region overlap sensitivity and precision is based on nucleotide precision.
 * 
 */
  __pyx_tuple__54 = PyTuple_Pack(15, __pyx_n_s_predictionSets, __pyx_n_s_regionSets, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_returnHTML, __pyx_n_s_fontsizeLabels, __pyx_n_s_fontsizeLegend, __pyx_n_s_fontsizeAxis, __pyx_n_s_style, __pyx_n_s_showLegend, __pyx_n_s_bboxAnchorTo, __pyx_n_s_legendLoc, __pyx_n_s_showValues, __pyx_n_s_measure, __pyx_n_s_measure); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(13, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_regions_pyx, __pyx_n_s_nucleotideRegionF1Barplot, 1074, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 1074, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_region(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__56 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_region, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__58 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_regions, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_keys.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyUnicode_Type_strip.type = (PyObject*)&PyUnicode_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_ = PyFloat_FromDouble(0.); if (unlikely(!__pyx_float_0_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_5 = PyFloat_FromDouble(0.5); if (unlikely(!__pyx_float_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_0_15 = PyFloat_FromDouble(-0.15); if (unlikely(!__pyx_float_neg_0_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_18 = PyInt_FromLong(18); if (unlikely(!__pyx_int_18)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_18804460 = PyInt_FromLong(18804460L); if (unlikely(!__pyx_int_18804460)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_165254229 = PyInt_FromLong(165254229L); if (unlikely(!__pyx_int_165254229)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6gnocis_7regions_region) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions_region.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions_region.tp_dictoffset && __pyx_type_6gnocis_7regions_region.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions_region.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_region, (PyObject *)&__pyx_type_6gnocis_7regions_region) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6gnocis_7regions_region) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_ptype_6gnocis_7regions_region = &__pyx_type_6gnocis_7regions_region;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions_regions) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions_regions.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions_regions.tp_dictoffset && __pyx_type_6gnocis_7regions_regions.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions_regions.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_regions, (PyObject *)&__pyx_type_6gnocis_7regions_regions) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6gnocis_7regions_regions) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_ptype_6gnocis_7regions_regions = &__pyx_type_6gnocis_7regions_regions;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct__bstr) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct__bstr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct__bstr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct__bstr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct__bstr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct__bstr = &__pyx_type_6gnocis_7regions___pyx_scope_struct__bstr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_1_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision) < 0) __PYX_ERR(0, 431, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision = &__pyx_type_6gnocis_7regions___pyx_scope_struct_2_nucleotidePrecision;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_3_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_3_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_3_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_3_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_3_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_3_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_4_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_4_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_4_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_4_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_4_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_4_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_5_dummy) < 0) __PYX_ERR(0, 443, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_5_dummy.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_5_dummy.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_5_dummy.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_5_dummy.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_5_dummy = &__pyx_type_6gnocis_7regions___pyx_scope_struct_5_dummy;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_6_genexpr) < 0) __PYX_ERR(0, 458, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_6_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_6_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_6_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_6_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_6_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_6_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_7_printStatistics) < 0) __PYX_ERR(0, 675, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_7_printStatistics.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_7_printStatistics.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_7_printStatistics.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_7_printStatistics.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_7_printStatistics = &__pyx_type_6gnocis_7regions___pyx_scope_struct_7_printStatistics;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_8_genexpr) < 0) __PYX_ERR(0, 681, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_8_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_8_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_8_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_8_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_8_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_8_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_9_genexpr) < 0) __PYX_ERR(0, 683, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_9_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_9_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_9_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_9_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_9_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_9_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_10_genexpr) < 0) __PYX_ERR(0, 683, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_10_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_10_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_10_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_10_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_10_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_10_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_11_genexpr) < 0) __PYX_ERR(0, 683, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_11_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_11_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_11_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_11_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_11_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_11_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_12_genexpr) < 0) __PYX_ERR(0, 684, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_12_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_12_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_12_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_12_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_12_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_12_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ) < 0) __PYX_ERR(0, 726, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ = &__pyx_type_6gnocis_7regions___pyx_scope_struct_13_loadBEDGZ;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_14_genexpr) < 0) __PYX_ERR(0, 744, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_14_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_14_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_14_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_14_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_14_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_14_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ) < 0) __PYX_ERR(0, 788, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ = &__pyx_type_6gnocis_7regions___pyx_scope_struct_15_loadGFFGZ;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_16_genexpr) < 0) __PYX_ERR(0, 798, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_16_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_16_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_16_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_16_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_16_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_16_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_17_genexpr) < 0) __PYX_ERR(0, 798, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_17_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_17_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_17_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_17_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_17_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_17_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17 = &__pyx_type_6gnocis_7regions___pyx_scope_struct_18___pyx_lambda_funcdef_lambda17;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_19_genexpr) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_19_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_19_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_19_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_19_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_19_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_19_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_20_genexpr) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_20_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_20_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_20_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_20_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_20_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_20_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_21_measure) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_21_measure.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_21_measure.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_21_measure.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_21_measure.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_21_measure = &__pyx_type_6gnocis_7regions___pyx_scope_struct_21_measure;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_22_genexpr) < 0) __PYX_ERR(0, 1106, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_22_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_22_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_22_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_22_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_22_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_22_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_7regions___pyx_scope_struct_23_genexpr) < 0) __PYX_ERR(0, 1106, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_7regions___pyx_scope_struct_23_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_7regions___pyx_scope_struct_23_genexpr.tp_dictoffset && __pyx_type_6gnocis_7regions___pyx_scope_struct_23_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_7regions___pyx_scope_struct_23_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_7regions___pyx_scope_struct_23_genexpr = &__pyx_type_6gnocis_7regions___pyx_scope_struct_23_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("gnocis.sequences"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6gnocis_9sequences_sequenceLabel = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequenceLabel", sizeof(struct __pyx_obj_6gnocis_9sequences_sequenceLabel), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequenceLabel) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_sequence = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequence", sizeof(struct __pyx_obj_6gnocis_9sequences_sequence), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequence) __PYX_ERR(3, 27, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_9sequences_sequence = (struct __pyx_vtabstruct_6gnocis_9sequences_sequence*)__Pyx_GetVtable(__pyx_ptype_6gnocis_9sequences_sequence->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_9sequences_sequence)) __PYX_ERR(3, 27, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_sequences = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequences", sizeof(struct __pyx_obj_6gnocis_9sequences_sequences), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequences) __PYX_ERR(3, 37, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_sequenceStream = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequenceStream", sizeof(struct __pyx_obj_6gnocis_9sequences_sequenceStream), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequenceStream) __PYX_ERR(3, 49, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_sequenceGenerator = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequenceGenerator", sizeof(struct __pyx_obj_6gnocis_9sequences_sequenceGenerator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequenceGenerator) __PYX_ERR(3, 65, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_MarkovChain = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "MarkovChain", sizeof(struct __pyx_obj_6gnocis_9sequences_MarkovChain), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_MarkovChain) __PYX_ERR(3, 68, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_IID = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "IID", sizeof(struct __pyx_obj_6gnocis_9sequences_IID), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_IID) __PYX_ERR(3, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("gnocis.sequences"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "streamFASTA", (void (**)(void))&__pyx_f_6gnocis_9sequences_streamFASTA, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6gnocis_9sequences_streamFASTA *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "stream2bit", (void (**)(void))&__pyx_f_6gnocis_9sequences_stream2bit, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6gnocis_9sequences_stream2bit *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "getSequenceStreamFromPath", (void (**)(void))&__pyx_f_6gnocis_9sequences_getSequenceStreamFromPath, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6gnocis_9sequences_getSequenceStreamFromPath *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initregions(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initregions(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_regions(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_regions(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_regions(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'regions' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_regions(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("regions", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_gnocis__regions) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "gnocis.regions")) {
      if (unlikely(PyDict_SetItemString(modules, "gnocis.regions", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "gnocis/regions.pyx":10
 * 
 * from __future__ import division
 * import random             # <<<<<<<<<<<<<<
 * import gzip
 * from libcpp cimport bool
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_random, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_random, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":11
 * from __future__ import division
 * import random
 * import gzip             # <<<<<<<<<<<<<<
 * from libcpp cimport bool
 * from .sequences cimport *
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_gzip, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_gzip, __pyx_t_1) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":14
 * from libcpp cimport bool
 * from .sequences cimport *
 * from .ioputil import nctable             # <<<<<<<<<<<<<<
 * from .models import CVModelPredictions
 * from .sequences import positive
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_nctable);
  __Pyx_GIVEREF(__pyx_n_s_nctable);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_nctable);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_ioputil, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_nctable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nctable, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":15
 * from .sequences cimport *
 * from .ioputil import nctable
 * from .models import CVModelPredictions             # <<<<<<<<<<<<<<
 * from .sequences import positive
 * from .common import mean, CI
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_CVModelPredictions);
  __Pyx_GIVEREF(__pyx_n_s_CVModelPredictions);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_CVModelPredictions);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_models, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CVModelPredictions); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CVModelPredictions, __pyx_t_2) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":16
 * from .ioputil import nctable
 * from .models import CVModelPredictions
 * from .sequences import positive             # <<<<<<<<<<<<<<
 * from .common import mean, CI
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_positive);
  __Pyx_GIVEREF(__pyx_n_s_positive);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_positive);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_sequences, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_positive); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_positive, __pyx_t_1) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/regions.pyx":17
 * from .models import CVModelPredictions
 * from .sequences import positive
 * from .common import mean, CI             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_mean);
  __Pyx_GIVEREF(__pyx_n_s_mean);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_mean);
  __Pyx_INCREF(__pyx_n_s_CI);
  __Pyx_GIVEREF(__pyx_n_s_CI);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_CI);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_common, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_mean); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mean, __pyx_t_2) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CI); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CI, __pyx_t_2) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":51
 * 	"""
 * 
 * 	__slots__ = ('seq', 'start', 'end', 'strand', 'score', 'source', 'feature', 'markers', 'group', 'ext')             # <<<<<<<<<<<<<<
 * 
 * 	def __init__(self, seq, start, end, strand = True, score = 0.0, source = '.', feature = '.', group = '.', dropChr = True):
 */
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gnocis_7regions_region->tp_dict, __pyx_n_s_slots, __pyx_tuple__33) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6gnocis_7regions_region);

  /* "gnocis/regions.pyx":687
 * 
 * # Loads regions from a BED file.
 * def loadBED(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_1loadBED, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_loadBED, __pyx_t_1) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":726
 * 
 * # Loads regions from a gzipped BED file.
 * def loadBEDGZ(path):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped BED (https://www.ensembl.org/info/website/upload/bed.html) file.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_3loadBEDGZ, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_loadBEDGZ, __pyx_t_1) < 0) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":765
 * 
 * # Loads regions from a GFF file.
 * def loadGFF(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_5loadGFF, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_loadGFF, __pyx_t_1) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":788
 * 
 * # Loads regions from a gzipped GFF file.
 * def loadGFFGZ(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a gzipped General Feature Format (https://www.ensembl.org/info/website/upload/gff.html) file.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_7loadGFFGZ, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_loadGFFGZ, __pyx_t_1) < 0) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":805
 * 
 * # Loads regions from a coordinate list file.
 * def loadCoordinateList(path, dropChr = True):             # <<<<<<<<<<<<<<
 * 	""" Loads regions from a coordinate list file (one region per line, with the format `seq:start..end`).
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_9loadCoordinateList, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_loadCoordinateList, __pyx_t_1) < 0) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":829
 * 
 * # Generates a barplot for a region-based measure
 * def regionBarplot(mainRegions, measure, measureName, testRegions = None, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeValues = 12, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = False, isPercent = False):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of a measure (lambda) for predicted regions.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_11regionBarplot, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_regionBarplot, __pyx_t_1) < 0) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":944
 * 			plt.close('all')
 * 
 * def predictionBarplot(predictionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_13predictionBarplot, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_predictionBarplot, __pyx_t_1) < 0) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":975
 * 	regionBarplot(mainRegions = predictionSets, measure = (lambda rs, _: len(rs)), measureName = 'Predictions', testRegions = None, figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = False)
 * 
 * def overlapSensitivityBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap sensitivity barplot.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_15overlapSensitivityBarplot, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_overlapSensitivityBarplot, __pyx_t_1) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":1008
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(rs.overlap(ps)) / len(rs)), measureName = 'Overlap sensitivity', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def overlapPrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction overlap precision barplot.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_17overlapPrecisionBarplot, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_overlapPrecisionBarplot, __pyx_t_1) < 0) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":1041
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: len(ps.overlap(rs)) / len(ps)), measureName = 'Overlap precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotidePrecisionBarplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a prediction nucleotide precision barplot.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_19nucleotidePrecisionBarplot, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nucleotidePrecisionBarplot, __pyx_t_1) < 0) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":1074
 * 	regionBarplot(mainRegions = predictionSets, testRegions = regionSets, measure = (lambda ps, rs: sum(len(r) for r in ps & rs) / sum(len(r) for r in ps)), measureName = 'Nucleotide precision', figsize = figsize, outpath = outpath, returnHTML = returnHTML, fontsizeLabels = fontsizeLabels, fontsizeLegend = fontsizeLegend, fontsizeAxis = fontsizeAxis, style = style, showLegend = showLegend, bboxAnchorTo = bboxAnchorTo, legendLoc = legendLoc, showValues = showValues, isPercent = True)
 * 
 * def nucleotideRegionF1Barplot(predictionSets, regionSets, figsize = (8, 8), outpath = None, returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, style = 'ggplot', showLegend = True, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left', showValues = True):             # <<<<<<<<<<<<<<
 * 	""" Generates a barplot of F1-measures, where the recall is based on region overlap sensitivity and precision is based on nucleotide precision.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_21nucleotideRegionF1Barplot, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nucleotideRegionF1Barplot, __pyx_t_1) < 0) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_region(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_23__pyx_unpickle_region, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_region, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_region__set_state(<region> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_region__set_state(region __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.end = __pyx_state[0]; __pyx_result.ext = __pyx_state[1]; __pyx_result.feature = __pyx_state[2]; __pyx_result.group = __pyx_state[3]; __pyx_result.markers = __pyx_state[4]; __pyx_result.score = __pyx_state[5]; __pyx_result.seq = __pyx_state[6]; __pyx_result.source = __pyx_state[7]; __pyx_result.start = __pyx_state[8]; __pyx_result.strand = __pyx_state[9]
 *     if len(__pyx_state) > 10 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_7regions_25__pyx_unpickle_regions, NULL, __pyx_n_s_gnocis_regions); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_regions, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/regions.pyx":1
 * # distutils: language=c++             # <<<<<<<<<<<<<<
 * # -*- coding: latin-1 -*-
 * ############################################################################
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init gnocis.regions", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init gnocis.regions");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyIntFromDouble */
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE PyObject* __Pyx_PyInt_FromDouble(double value) {
    if (value >= (double)LONG_MIN && value <= (double)LONG_MAX) {
        return PyInt_FromLong((long)value);
    }
    return PyLong_FromDouble(value);
}
#endif

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_PY_LONG_LONG(PY_LONG_LONG value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(PY_LONG_LONG)*3+2];
    char *dpos, *end = digits + sizeof(PY_LONG_LONG)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    PY_LONG_LONG remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (PY_LONG_LONG) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (PY_LONG_LONG) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (PY_LONG_LONG) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject** CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyList_New(0);
    dest = PyList_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyListObject*)src)->ob_item + start,
        ((PyListObject*)dest)->ob_item,
        length);
    return dest;
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyTuple_New(0);
    dest = PyTuple_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyTupleObject*)src)->ob_item + start,
        ((PyTupleObject*)dest)->ob_item,
        length);
    return dest;
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_keys */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_keys, d);
    else
        return PyDict_Keys(d);
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyObjectFormat */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* obj, PyObject* format_spec) {
    int ret;
    _PyUnicodeWriter writer;
    if (likely(PyFloat_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyFloat_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else if (likely(PyLong_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyLong_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else {
        return PyObject_Format(obj, format_spec);
    }
    if (unlikely(ret == -1)) {
        _PyUnicodeWriter_Dealloc(&writer);
        return NULL;
    }
    return _PyUnicodeWriter_Finish(&writer);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* unicode_tailmatch */
static int __Pyx_PyUnicode_TailmatchTuple(PyObject* s, PyObject* substrings,
                                          Py_ssize_t start, Py_ssize_t end, int direction) {
    Py_ssize_t i, count = PyTuple_GET_SIZE(substrings);
    for (i = 0; i < count; i++) {
        Py_ssize_t result;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        result = PyUnicode_Tailmatch(s, PyTuple_GET_ITEM(substrings, i),
                                     start, end, direction);
#else
        PyObject* sub = PySequence_ITEM(substrings, i);
        if (unlikely(!sub)) return -1;
        result = PyUnicode_Tailmatch(s, sub, start, end, direction);
        Py_DECREF(sub);
#endif
        if (result) {
            return (int) result;
        }
    }
    return 0;
}
static int __Pyx_PyUnicode_Tailmatch(PyObject* s, PyObject* substr,
                                     Py_ssize_t start, Py_ssize_t end, int direction) {
    if (unlikely(PyTuple_Check(substr))) {
        return __Pyx_PyUnicode_TailmatchTuple(s, substr, start, end, direction);
    }
    return (int) PyUnicode_Tailmatch(s, substr, start, end, direction);
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (PY_LONG_LONG) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, digit, digits[0])
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (PY_LONG_LONG) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case -1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            PY_LONG_LONG val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (PY_LONG_LONG) -1;
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
