/* Generated by Cython 0.29.24 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_24"
#define CYTHON_HEX_VERSION 0x001D18F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__gnocis__models
#define __PYX_HAVE_API__gnocis__models
/* Early includes */
#include <string.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "gnocis/models.pyx",
  "gnocis/regions.pxd",
  "gnocis/sequences.pxd",
  "gnocis/features.pxd",
  "gnocis/validation.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6gnocis_7regions_region;
struct __pyx_obj_6gnocis_7regions_regions;
struct __pyx_obj_6gnocis_9sequences_sequenceLabel;
struct __pyx_obj_6gnocis_9sequences_sequence;
struct __pyx_obj_6gnocis_9sequences_sequences;
struct __pyx_obj_6gnocis_9sequences_sequenceStream;
struct __pyx_obj_6gnocis_9sequences_sequenceGenerator;
struct __pyx_obj_6gnocis_9sequences_MarkovChain;
struct __pyx_obj_6gnocis_9sequences_IID;
struct __pyx_obj_6gnocis_8features_feature;
struct __pyx_obj_6gnocis_8features_features;
struct __pyx_obj_6gnocis_8features_scaledFeature;
struct __pyx_obj_6gnocis_8features_featureScaler;
struct __pyx_obj_6gnocis_8features_featureMotifOccurrenceFrequency;
struct __pyx_obj_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency;
struct __pyx_obj_6gnocis_8features_kSpectrumFeature;
struct __pyx_obj_6gnocis_8features_kSpectrum;
struct __pyx_obj_6gnocis_8features_kSpectrumMMFeature;
struct __pyx_obj_6gnocis_8features_kSpectrumMM;
struct __pyx_obj_6gnocis_8features_kSpectrumMMDFeature;
struct __pyx_obj_6gnocis_8features_kSpectrumMMD;
struct __pyx_obj_6gnocis_8features_kSpectrumFeatureGPS;
struct __pyx_obj_6gnocis_8features_kSpectrumGPS;
struct __pyx_obj_6gnocis_10validation_point2D;
struct __pyx_obj_6gnocis_10validation_validationPair;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC;
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr;
struct __pyx_opt_args_6gnocis_9sequences_streamFASTA;
struct __pyx_opt_args_6gnocis_9sequences_stream2bit;
struct __pyx_opt_args_6gnocis_9sequences_getSequenceStreamFromPath;

/* "sequences.pxd":53
 * 
 * # Streams a FASTA file in blocks.
 * cpdef streamFASTA(path, wantBlockSize = *, spacePrune = *, dropChr = *, restrictToSequences = *)             # <<<<<<<<<<<<<<
 * 
 * # Streams a 2bit file in blocks.
 */
struct __pyx_opt_args_6gnocis_9sequences_streamFASTA {
  int __pyx_n;
  PyObject *wantBlockSize;
  PyObject *spacePrune;
  PyObject *dropChr;
  PyObject *restrictToSequences;
};

/* "sequences.pxd":56
 * 
 * # Streams a 2bit file in blocks.
 * cpdef stream2bit(path, wantBlockSize = *, spacePrune = *, dropChr = *, restrictToSequences = *)             # <<<<<<<<<<<<<<
 * 
 * # Gets a sequence stream based on a path, deciding on the format from the path.
 */
struct __pyx_opt_args_6gnocis_9sequences_stream2bit {
  int __pyx_n;
  PyObject *wantBlockSize;
  PyObject *spacePrune;
  PyObject *dropChr;
  PyObject *restrictToSequences;
};

/* "sequences.pxd":59
 * 
 * # Gets a sequence stream based on a path, deciding on the format from the path.
 * cpdef getSequenceStreamFromPath(path, wantBlockSize = *, spacePrune = *, dropChr = *, restrictToSequences = *)             # <<<<<<<<<<<<<<
 * 
 * ###########################################################################
 */
struct __pyx_opt_args_6gnocis_9sequences_getSequenceStreamFromPath {
  int __pyx_n;
  PyObject *wantBlockSize;
  PyObject *spacePrune;
  PyObject *dropChr;
  PyObject *restrictToSequences;
};
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;
struct __pyx_defaults1;
typedef struct __pyx_defaults1 __pyx_defaults1;
struct __pyx_defaults2;
typedef struct __pyx_defaults2 __pyx_defaults2;
struct __pyx_defaults3;
typedef struct __pyx_defaults3 __pyx_defaults3;
struct __pyx_defaults4;
typedef struct __pyx_defaults4 __pyx_defaults4;
struct __pyx_defaults5;
typedef struct __pyx_defaults5 __pyx_defaults5;
struct __pyx_defaults6;
typedef struct __pyx_defaults6 __pyx_defaults6;
struct __pyx_defaults7;
typedef struct __pyx_defaults7 __pyx_defaults7;
struct __pyx_defaults8;
typedef struct __pyx_defaults8 __pyx_defaults8;
struct __pyx_defaults9;
typedef struct __pyx_defaults9 __pyx_defaults9;
struct __pyx_defaults10;
typedef struct __pyx_defaults10 __pyx_defaults10;
struct __pyx_defaults11;
typedef struct __pyx_defaults11 __pyx_defaults11;
struct __pyx_defaults12;
typedef struct __pyx_defaults12 __pyx_defaults12;
struct __pyx_defaults {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults1 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults2 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults3 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults4 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults5 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults6 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults7 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults8 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults9 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults10 {
  PyObject *__pyx_arg_trainedModels;
};
struct __pyx_defaults11 {
  PyObject *__pyx_arg_labelPositive;
  PyObject *__pyx_arg_labelNegative;
};
struct __pyx_defaults12 {
  PyObject *__pyx_arg_labelPositive;
};

/* "regions.pxd":16
 * 
 * # Represents regions.
 * cdef class region:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str seq
 */
struct __pyx_obj_6gnocis_7regions_region {
  PyObject_HEAD
  PyObject *seq;
  PY_LONG_LONG start;
  PY_LONG_LONG end;
  bool strand;
  float score;
  PyObject *source;
  PyObject *feature;
  PyObject *group;
  PyObject *ext;
  PyObject *markers;
};


/* "regions.pxd":27
 * 
 * # Represents region sets. Supports operands '+' for adding regions from two sets with no merging, '|' for merging two sets, and '^' for excluding the regions in the latter set from the first.
 * cdef class regions:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name
 */
struct __pyx_obj_6gnocis_7regions_regions {
  PyObject_HEAD
  PyObject *name;
  PyObject *regions;
};


/* "sequences.pxd":17
 * 
 * # Represents a sequence label
 * cdef class sequenceLabel:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name
 */
struct __pyx_obj_6gnocis_9sequences_sequenceLabel {
  PyObject_HEAD
  PyObject *name;
  float value;
};


/* "sequences.pxd":27
 * 
 * # Represents and loads DNA sequences.
 * cdef class sequence:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name, seq, path, annotation
 */
struct __pyx_obj_6gnocis_9sequences_sequence {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gnocis_9sequences_sequence *__pyx_vtab;
  PyObject *name;
  PyObject *seq;
  PyObject *path;
  PyObject *annotation;
  struct __pyx_obj_6gnocis_7regions_region *sourceRegion;
  PyObject *cbytes;
  PyObject *cbytesIndexed;
  PyObject *labels;
};


/* "sequences.pxd":37
 * 
 * # Represents a set of DNA sequences.
 * cdef class sequences:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name
 */
struct __pyx_obj_6gnocis_9sequences_sequences {
  PyObject_HEAD
  PyObject *name;
  PyObject *sequences;
};


/* "sequences.pxd":49
 * 
 * # Represents sequence streams.
 * cdef class sequenceStream:             # <<<<<<<<<<<<<<
 * 	pass
 * 
 */
struct __pyx_obj_6gnocis_9sequences_sequenceStream {
  PyObject_HEAD
};


/* "sequences.pxd":65
 * 
 * # Represents sequence streams
 * cdef class sequenceGenerator:             # <<<<<<<<<<<<<<
 * 	pass
 * 
 */
struct __pyx_obj_6gnocis_9sequences_sequenceGenerator {
  PyObject_HEAD
};


/* "sequences.pxd":68
 * 	pass
 * 
 * cdef class MarkovChain(sequenceGenerator):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int degree, pseudoCounts, nGenerated
 */
struct __pyx_obj_6gnocis_9sequences_MarkovChain {
  struct __pyx_obj_6gnocis_9sequences_sequenceGenerator __pyx_base;
  int degree;
  int pseudoCounts;
  int nGenerated;
  bool addReverseComplements;
  bool prepared;
  PyObject *trainingSequences;
  PyObject *spectrum;
  PyObject *initialDistribution;
  PyObject *probspectrum;
  PyObject *comparableSpectrum;
};


/* "sequences.pxd":78
 * 
 * # IID sequence model
 * cdef class IID(sequenceGenerator):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int pseudoCounts, nGenerated
 */
struct __pyx_obj_6gnocis_9sequences_IID {
  struct __pyx_obj_6gnocis_9sequences_sequenceGenerator __pyx_base;
  int pseudoCounts;
  int nGenerated;
  bool addComplements;
  bool prepared;
  PyObject *trainingSequences;
  PyObject *spectrum;
  PyObject *ntDistribution;
};


/* "features.pxd":17
 * # Sequence features
 * 
 * cdef class feature:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public sequence cachedSequence
 */
struct __pyx_obj_6gnocis_8features_feature {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gnocis_8features_feature *__pyx_vtab;
  struct __pyx_obj_6gnocis_9sequences_sequence *cachedSequence;
  double cachedValue;
};


/* "features.pxd":24
 * 
 * # Represents feature sets.
 * cdef class features:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name
 */
struct __pyx_obj_6gnocis_8features_features {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gnocis_8features_features *__pyx_vtab;
  PyObject *name;
  PyObject *features;
  PyObject *values;
  int nvalues;
};


/* "features.pxd":33
 * 	cpdef list getAll(self, sequence seq)
 * 
 * cdef class scaledFeature(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public float vScale, vSub
 */
struct __pyx_obj_6gnocis_8features_scaledFeature {
  struct __pyx_obj_6gnocis_8features_feature __pyx_base;
  float vScale;
  float vSub;
  struct __pyx_obj_6gnocis_8features_feature *feature;
};


/* "features.pxd":40
 * 	cpdef double get(self, sequence seq)
 * 
 * cdef class featureScaler(features):             # <<<<<<<<<<<<<<
 * 
 * 	cpdef list getAll(self, sequence seq)
 */
struct __pyx_obj_6gnocis_8features_featureScaler {
  struct __pyx_obj_6gnocis_8features_features __pyx_base;
};


/* "features.pxd":49
 * 
 * # Sequence model feature for motif occurrence frequency.
 * cdef class featureMotifOccurrenceFrequency(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public object m
 */
struct __pyx_obj_6gnocis_8features_featureMotifOccurrenceFrequency {
  struct __pyx_obj_6gnocis_8features_feature __pyx_base;
  PyObject *m;
};


/* "features.pxd":56
 * 
 * # Sequence model feature for PREdictor-style motif pair occurrence frequency.
 * cdef class featurePREdictorMotifPairOccurrenceFrequency(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public object mA, mB
 */
struct __pyx_obj_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency {
  struct __pyx_obj_6gnocis_8features_feature __pyx_base;
  PyObject *mA;
  PyObject *mB;
  int distCut;
};


/* "features.pxd":68
 * 
 * # k-mer for k-spectrum feature set
 * cdef class kSpectrumFeature(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public kSpectrum parent
 */
struct __pyx_obj_6gnocis_8features_kSpectrumFeature {
  struct __pyx_obj_6gnocis_8features_feature __pyx_base;
  struct __pyx_obj_6gnocis_8features_kSpectrum *parent;
  PyObject *kmer;
  int index;
};


/* "features.pxd":77
 * 
 * # Extracts k-mer spectra from sequences
 * cdef class kSpectrum:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int nspectrum, nFeatures, bitmask
 */
struct __pyx_obj_6gnocis_8features_kSpectrum {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gnocis_8features_kSpectrum *__pyx_vtab;
  int nspectrum;
  int nFeatures;
  int bitmask;
  PyObject *kmerByIndex;
  PyObject *features;
  struct __pyx_obj_6gnocis_9sequences_sequence *cachedSequence;
  PyObject *cachedSpectrum;
};


/* "features.pxd":91
 * 
 * # k-mer for k-spectrum feature set
 * cdef class kSpectrumMMFeature(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public kSpectrumMM parent
 */
struct __pyx_obj_6gnocis_8features_kSpectrumMMFeature {
  struct __pyx_obj_6gnocis_8features_feature __pyx_base;
  struct __pyx_obj_6gnocis_8features_kSpectrumMM *parent;
  PyObject *kmer;
  int index;
};


/* "features.pxd":100
 * 
 * # Extracts k-mer spectra from sequences
 * cdef class kSpectrumMM:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int nspectrum, nFeatures, bitmask
 */
struct __pyx_obj_6gnocis_8features_kSpectrumMM {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMM *__pyx_vtab;
  int nspectrum;
  int nFeatures;
  int bitmask;
  PyObject *kmerByIndex;
  PyObject *features;
  struct __pyx_obj_6gnocis_9sequences_sequence *cachedSequence;
  PyObject *cachedSpectrum;
};


/* "features.pxd":116
 * 
 * # k-mer for k-spectrum feature set
 * cdef class kSpectrumMMDFeature(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public kSpectrumMMD parent
 */
struct __pyx_obj_6gnocis_8features_kSpectrumMMDFeature {
  struct __pyx_obj_6gnocis_8features_feature __pyx_base;
  struct __pyx_obj_6gnocis_8features_kSpectrumMMD *parent;
  PyObject *kmer;
  int index;
};


/* "features.pxd":125
 * 
 * # Extracts k-mer spectra from sequences
 * cdef class kSpectrumMMD:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int nspectrum, nFeatures, bitmask
 */
struct __pyx_obj_6gnocis_8features_kSpectrumMMD {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMD *__pyx_vtab;
  int nspectrum;
  int nFeatures;
  int bitmask;
  PyObject *kmerByIndex;
  PyObject *features;
  struct __pyx_obj_6gnocis_9sequences_sequence *cachedSequence;
  PyObject *cachedSpectrum;
};


/* "features.pxd":140
 * 
 * # k-mer for k-spectrum feature set
 * cdef class kSpectrumFeatureGPS(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public kSpectrumGPS parent
 */
struct __pyx_obj_6gnocis_8features_kSpectrumFeatureGPS {
  struct __pyx_obj_6gnocis_8features_feature __pyx_base;
  struct __pyx_obj_6gnocis_8features_kSpectrumGPS *parent;
  PyObject *kmer;
  int index;
  bool section;
};


/* "features.pxd":150
 * 
 * # Extracts k-mer spectra from sequences
 * cdef class kSpectrumGPS:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int nspectrum, nkmers, nFeatures, bitmask
 */
struct __pyx_obj_6gnocis_8features_kSpectrumGPS {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gnocis_8features_kSpectrumGPS *__pyx_vtab;
  int nspectrum;
  int nkmers;
  int nFeatures;
  int bitmask;
  PyObject *kmerByIndex;
  PyObject *features;
  struct __pyx_obj_6gnocis_9sequences_sequence *cachedSequence;
  PyObject *cachedSpectrum;
};


/* "validation.pxd":17
 * 
 * # Represents a two-dimensional point.
 * cdef class point2D:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public float x, y, rank
 */
struct __pyx_obj_6gnocis_10validation_point2D {
  PyObject_HEAD
  float x;
  float y;
  float rank;
};


/* "validation.pxd":22
 * 
 * # Represents a pair of a validation score and label (with the option of a name).
 * cdef class validationPair:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public float score
 */
struct __pyx_obj_6gnocis_10validation_validationPair {
  PyObject_HEAD
  float score;
  bool label;
  PyObject *name;
};


/* "gnocis/models.pyx":40
 * 
 * # Used for applying models in separate processes
 * def multiprocessModelDeployer(args):             # <<<<<<<<<<<<<<
 * 	cdef sequences seqs
 * 	cdef sequence seq, win
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_9sequences_sequence *__pyx_7genexpr__pyx_v_seq;
  int __pyx_v_windowSize;
  int __pyx_v_windowStep;
};


/* "gnocis/models.pyx":51
 * 		max(
 * 			(
 * 				model.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(windowSize, windowStep)
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *__pyx_outer_scope;
  PyObject *__pyx_v_win;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/models.pyx":87
 * 			break
 * 
 * def getPoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	while True:
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs {
  PyObject_HEAD
  PyObject *__pyx_v_pool;
};


/* "gnocis/models.pyx":90
 * 	pIDs = set()
 * 	while True:
 * 		pIDs = pIDs | set( v for v in pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) if v != None )             # <<<<<<<<<<<<<<
 * 		if len(pIDs) == len(pool._pool):
 * 			return pIDs
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *__pyx_outer_scope;
  Py_ssize_t __pyx_8genexpr4__pyx_v__;
  PyObject *__pyx_v_v;
};


/* "gnocis/models.pyx":94
 * 			return pIDs
 * 
 * def getPoolModelIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	mIDs = set()
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs {
  PyObject_HEAD
  PyObject *__pyx_v_r;
};


/* "gnocis/models.pyx":99
 * 	while True:
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )             # <<<<<<<<<<<<<<
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )
 * 		if len(pIDs) == len(pool._pool):
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *__pyx_outer_scope;
  PyObject *__pyx_v_mid;
  PyObject *__pyx_v_pid;
};


/* "gnocis/models.pyx":100
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )             # <<<<<<<<<<<<<<
 * 		if len(pIDs) == len(pool._pool):
 * 			return mIDs
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *__pyx_outer_scope;
  PyObject *__pyx_v_mid;
  PyObject *__pyx_v_pid;
};


/* "gnocis/models.pyx":113
 * 	return (processID, model != None)
 * 
 * def setPoolModel(pool, mdl):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	if mdl == None:
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel {
  PyObject_HEAD
  PyObject *__pyx_v_r;
};


/* "gnocis/models.pyx":122
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			# We append the process-IDs that we confirmed are done
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if not hasModel )             # <<<<<<<<<<<<<<
 * 			if len(pIDs) == len(pool._pool):
 * 				break
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *__pyx_outer_scope;
  PyObject *__pyx_v_hasModel;
  PyObject *__pyx_v_pid;
};


/* "gnocis/models.pyx":133
 * 			pool.map( setPoolProcessModel, [ mdl for _ in range(len(pool._pool)) ] )
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if hasModel )             # <<<<<<<<<<<<<<
 * 			if len(pIDs) == len(pool._pool):
 * 				break
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *__pyx_outer_scope;
  PyObject *__pyx_v_hasModel;
  PyObject *__pyx_v_pid;
};


/* "gnocis/models.pyx":318
 * 
 * 	# Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 	def calibrateGenomewidePrecision(self, positives, genome, factor = 1.0, precision = 0.8, bgModelOrder = 4, bgModel = None):             # <<<<<<<<<<<<<<
 * 		""" Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision {
  PyObject_HEAD
  PyObject *__pyx_v_positives;
  PyObject *__pyx_v_seqLens;
};


/* "gnocis/models.pyx":342
 * 			genome = getSequenceStreamFromPath(genome)
 * 		seqLens = genome.sequenceLengths()
 * 		genomeLen = sum(seqLens[chrom] for chrom in seqLens.keys())             # <<<<<<<<<<<<<<
 * 		# Train background model
 * 		if bgModel == None:
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *__pyx_outer_scope;
  PyObject *__pyx_v_chrom;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "gnocis/models.pyx":349
 * 				else MarkovChain(trainingSequences = genome, degree = bgModelOrder)
 * 		# Generate calibration sequences
 * 		meanLen = int(sum(len(s) for s in positives) / len(positives))             # <<<<<<<<<<<<<<
 * 		negatives = bgModel.generateStream(n = int(factor * genomeLen / meanLen), length = meanLen)
 * 		# Calibrate threshold
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *__pyx_outer_scope;
  PyObject *__pyx_v_s;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/models.pyx":356
 * 
 * 	# Applies model to a sequence with a sliding window, and returns the maximum score.
 * 	def scoreSequence(self, seq):             # <<<<<<<<<<<<<<
 * 		""" Scores a single sequence. The score is determined by applying the model with a sliding window, and taking the maximum score.
 * 
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence {
  PyObject_HEAD
  PyObject *__pyx_v_self;
  PyObject *__pyx_v_seq;
};


/* "gnocis/models.pyx":369
 * 		score = max(
 * 			(
 * 				self.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(self.windowSize, self.windowStep)
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *__pyx_outer_scope;
  PyObject *__pyx_v_win;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/models.pyx":595
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamCurves(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves {
  PyObject_HEAD
  PyObject *__pyx_v_winPosSet;
};


/* "gnocis/models.pyx":630
 * 
 * 		posScores = {}
 * 		for seq in set(winPos.seq for winPos in winPosSet):             # <<<<<<<<<<<<<<
 * 			scores = [
 * 				(int(winPos.start/self.windowStep), wScore)
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *__pyx_outer_scope;
  PyObject *__pyx_v_winPos;
};


/* "gnocis/models.pyx":752
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "gnocis/models.pyx":786
 * 	"""
 * 
 * 	def __init__(self, name, features, trainingSet, windowSize, windowStep, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_f;
  PyObject *__pyx_v_trainingNegatives;
  PyObject *__pyx_v_trainingPositives;
};


/* "gnocis/models.pyx":798
 * 		self.sortedweights = []
 * 		for f in features:
 * 			vPos = sum( f.get(s) for s in trainingPositives )             # <<<<<<<<<<<<<<
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )
 * 			if vPos == 0.0 or vNeg == 0.0:
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_s;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/models.pyx":799
 * 		for f in features:
 * 			vPos = sum( f.get(s) for s in trainingPositives )
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )             # <<<<<<<<<<<<<<
 * 			if vPos == 0.0 or vNeg == 0.0:
 * 				vPos += 1.0
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_s;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "gnocis/models.pyx":812
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "gnocis/models.pyx":919
 * 	"""
 * 
 * 	def __init__(self, models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = None, storeTrained = True):             # <<<<<<<<<<<<<<
 * 		self.models = []
 * 		self.PRC = {}
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_tsByLabel;
};


/* "gnocis/models.pyx":942
 * 			for lbl, seqs in zip(tslabels, trainingSet.withLabel(tslabels))
 * 		}
 * 		ntrain = int( min(len(tsByLabel[lbl]) for lbl in tsByLabel) * ratioTrainTest )             # <<<<<<<<<<<<<<
 * 		self.cvtrain = []
 * 		self.cvval = []
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_lbl;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "gnocis/models.pyx":1044
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC {
  PyObject_HEAD
  PyObject *__pyx_v_curves;
};


/* "gnocis/models.pyx":1064
 * 				for mdl in self.models:
 * 					curves = self.PRC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))             # <<<<<<<<<<<<<<
 * 					curvebyx = [
 * 						{
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *__pyx_outer_scope;
  PyObject *__pyx_v_c;
  PyObject *__pyx_v_pt;
};


/* "gnocis/models.pyx":1133
 * 			raise err
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC {
  PyObject_HEAD
  PyObject *__pyx_v_curves;
};


/* "gnocis/models.pyx":1151
 * 				for mdl in self.models:
 * 					curves = self.ROC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))             # <<<<<<<<<<<<<<
 * 					curvebyx = [
 * 						{
 */
struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *__pyx_outer_scope;
  PyObject *__pyx_v_c;
  PyObject *__pyx_v_pt;
};



/* "sequences.pxd":27
 * 
 * # Represents and loads DNA sequences.
 * cdef class sequence:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name, seq, path, annotation
 */

struct __pyx_vtabstruct_6gnocis_9sequences_sequence {
  PyObject *(*getBytes)(struct __pyx_obj_6gnocis_9sequences_sequence *);
  PyObject *(*getBytesIndexed)(struct __pyx_obj_6gnocis_9sequences_sequence *);
};
static struct __pyx_vtabstruct_6gnocis_9sequences_sequence *__pyx_vtabptr_6gnocis_9sequences_sequence;


/* "features.pxd":17
 * # Sequence features
 * 
 * cdef class feature:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public sequence cachedSequence
 */

struct __pyx_vtabstruct_6gnocis_8features_feature {
  double (*get)(struct __pyx_obj_6gnocis_8features_feature *, struct __pyx_obj_6gnocis_9sequences_sequence *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gnocis_8features_feature *__pyx_vtabptr_6gnocis_8features_feature;


/* "features.pxd":24
 * 
 * # Represents feature sets.
 * cdef class features:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public str name
 */

struct __pyx_vtabstruct_6gnocis_8features_features {
  PyObject *(*getAll)(struct __pyx_obj_6gnocis_8features_features *, struct __pyx_obj_6gnocis_9sequences_sequence *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gnocis_8features_features *__pyx_vtabptr_6gnocis_8features_features;


/* "features.pxd":33
 * 	cpdef list getAll(self, sequence seq)
 * 
 * cdef class scaledFeature(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public float vScale, vSub
 */

struct __pyx_vtabstruct_6gnocis_8features_scaledFeature {
  struct __pyx_vtabstruct_6gnocis_8features_feature __pyx_base;
};
static struct __pyx_vtabstruct_6gnocis_8features_scaledFeature *__pyx_vtabptr_6gnocis_8features_scaledFeature;


/* "features.pxd":40
 * 	cpdef double get(self, sequence seq)
 * 
 * cdef class featureScaler(features):             # <<<<<<<<<<<<<<
 * 
 * 	cpdef list getAll(self, sequence seq)
 */

struct __pyx_vtabstruct_6gnocis_8features_featureScaler {
  struct __pyx_vtabstruct_6gnocis_8features_features __pyx_base;
};
static struct __pyx_vtabstruct_6gnocis_8features_featureScaler *__pyx_vtabptr_6gnocis_8features_featureScaler;


/* "features.pxd":49
 * 
 * # Sequence model feature for motif occurrence frequency.
 * cdef class featureMotifOccurrenceFrequency(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public object m
 */

struct __pyx_vtabstruct_6gnocis_8features_featureMotifOccurrenceFrequency {
  struct __pyx_vtabstruct_6gnocis_8features_feature __pyx_base;
};
static struct __pyx_vtabstruct_6gnocis_8features_featureMotifOccurrenceFrequency *__pyx_vtabptr_6gnocis_8features_featureMotifOccurrenceFrequency;


/* "features.pxd":56
 * 
 * # Sequence model feature for PREdictor-style motif pair occurrence frequency.
 * cdef class featurePREdictorMotifPairOccurrenceFrequency(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public object mA, mB
 */

struct __pyx_vtabstruct_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency {
  struct __pyx_vtabstruct_6gnocis_8features_feature __pyx_base;
};
static struct __pyx_vtabstruct_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency *__pyx_vtabptr_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency;


/* "features.pxd":68
 * 
 * # k-mer for k-spectrum feature set
 * cdef class kSpectrumFeature(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public kSpectrum parent
 */

struct __pyx_vtabstruct_6gnocis_8features_kSpectrumFeature {
  struct __pyx_vtabstruct_6gnocis_8features_feature __pyx_base;
};
static struct __pyx_vtabstruct_6gnocis_8features_kSpectrumFeature *__pyx_vtabptr_6gnocis_8features_kSpectrumFeature;


/* "features.pxd":77
 * 
 * # Extracts k-mer spectra from sequences
 * cdef class kSpectrum:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int nspectrum, nFeatures, bitmask
 */

struct __pyx_vtabstruct_6gnocis_8features_kSpectrum {
  double (*extract)(struct __pyx_obj_6gnocis_8features_kSpectrum *, struct __pyx_obj_6gnocis_9sequences_sequence *, int);
};
static struct __pyx_vtabstruct_6gnocis_8features_kSpectrum *__pyx_vtabptr_6gnocis_8features_kSpectrum;


/* "features.pxd":91
 * 
 * # k-mer for k-spectrum feature set
 * cdef class kSpectrumMMFeature(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public kSpectrumMM parent
 */

struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMFeature {
  struct __pyx_vtabstruct_6gnocis_8features_feature __pyx_base;
};
static struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMFeature *__pyx_vtabptr_6gnocis_8features_kSpectrumMMFeature;


/* "features.pxd":100
 * 
 * # Extracts k-mer spectra from sequences
 * cdef class kSpectrumMM:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int nspectrum, nFeatures, bitmask
 */

struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMM {
  double (*extract)(struct __pyx_obj_6gnocis_8features_kSpectrumMM *, struct __pyx_obj_6gnocis_9sequences_sequence *, int);
};
static struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMM *__pyx_vtabptr_6gnocis_8features_kSpectrumMM;


/* "features.pxd":116
 * 
 * # k-mer for k-spectrum feature set
 * cdef class kSpectrumMMDFeature(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public kSpectrumMMD parent
 */

struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMDFeature {
  struct __pyx_vtabstruct_6gnocis_8features_feature __pyx_base;
};
static struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMDFeature *__pyx_vtabptr_6gnocis_8features_kSpectrumMMDFeature;


/* "features.pxd":125
 * 
 * # Extracts k-mer spectra from sequences
 * cdef class kSpectrumMMD:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int nspectrum, nFeatures, bitmask
 */

struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMD {
  double (*extract)(struct __pyx_obj_6gnocis_8features_kSpectrumMMD *, struct __pyx_obj_6gnocis_9sequences_sequence *, int);
};
static struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMD *__pyx_vtabptr_6gnocis_8features_kSpectrumMMD;


/* "features.pxd":140
 * 
 * # k-mer for k-spectrum feature set
 * cdef class kSpectrumFeatureGPS(feature):             # <<<<<<<<<<<<<<
 * 
 * 	cdef public kSpectrumGPS parent
 */

struct __pyx_vtabstruct_6gnocis_8features_kSpectrumFeatureGPS {
  struct __pyx_vtabstruct_6gnocis_8features_feature __pyx_base;
};
static struct __pyx_vtabstruct_6gnocis_8features_kSpectrumFeatureGPS *__pyx_vtabptr_6gnocis_8features_kSpectrumFeatureGPS;


/* "features.pxd":150
 * 
 * # Extracts k-mer spectra from sequences
 * cdef class kSpectrumGPS:             # <<<<<<<<<<<<<<
 * 
 * 	cdef public int nspectrum, nkmers, nFeatures, bitmask
 */

struct __pyx_vtabstruct_6gnocis_8features_kSpectrumGPS {
  double (*extract)(struct __pyx_obj_6gnocis_8features_kSpectrumGPS *, struct __pyx_obj_6gnocis_9sequences_sequence *, int);
};
static struct __pyx_vtabstruct_6gnocis_8features_kSpectrumGPS *__pyx_vtabptr_6gnocis_8features_kSpectrumGPS;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyObjectFormat.proto */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* s, PyObject* f);
#else
#define __Pyx_PyObject_Format(s, f) PyObject_Format(s, f)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_EqObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_EqObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (PyObject_RichCompare(op1, op2, Py_EQ))
    #endif

/* PyFloatBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyFloat_AddObjC(op1, op2, floatval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CyFunctionClassCell.proto */
static int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libcpp' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'gnocis.regions' */
static PyTypeObject *__pyx_ptype_6gnocis_7regions_region = 0;
static PyTypeObject *__pyx_ptype_6gnocis_7regions_regions = 0;

/* Module declarations from 'gnocis.sequences' */
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequenceLabel = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequence = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequences = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequenceStream = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_sequenceGenerator = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_MarkovChain = 0;
static PyTypeObject *__pyx_ptype_6gnocis_9sequences_IID = 0;
static PyObject *(*__pyx_f_6gnocis_9sequences_getSequenceStreamFromPath)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6gnocis_9sequences_getSequenceStreamFromPath *__pyx_optional_args); /*proto*/

/* Module declarations from 'gnocis.features' */
static PyTypeObject *__pyx_ptype_6gnocis_8features_feature = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_features = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_scaledFeature = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_featureScaler = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_featureMotifOccurrenceFrequency = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_kSpectrumFeature = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_kSpectrum = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_kSpectrumMMFeature = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_kSpectrumMM = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_kSpectrumMMDFeature = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_kSpectrumMMD = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_kSpectrumFeatureGPS = 0;
static PyTypeObject *__pyx_ptype_6gnocis_8features_kSpectrumGPS = 0;

/* Module declarations from 'gnocis.validation' */
static PyTypeObject *__pyx_ptype_6gnocis_10validation_point2D = 0;
static PyTypeObject *__pyx_ptype_6gnocis_10validation_validationPair = 0;

/* Module declarations from 'gnocis.models' */
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_3_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_5_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_6_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_7_setPoolModel = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_8_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_9_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_11_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_12_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_13_scoreSequence = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_14_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_16_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_17_getTrainer = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_18___init__ = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_19_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_20_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_21_getTrainer = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_22___init__ = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_23_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_24_plotPRC = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_25_genexpr = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_26_plotROC = 0;
static PyTypeObject *__pyx_ptype_6gnocis_6models___pyx_scope_struct_27_genexpr = 0;
#define __Pyx_MODULE_NAME "gnocis.models"
extern int __pyx_module_is_main_gnocis__models;
int __pyx_module_is_main_gnocis__models = 0;

/* Implementation of 'gnocis.models' */
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_print;
static const char __pyx_k_[] = " ";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_2f[] = ".2f";
static const char __pyx_k_CI[] = "CI";
static const char __pyx_k_CV[] = "CV";
static const char __pyx_k_FN[] = "FN";
static const char __pyx_k_FP[] = "FP";
static const char __pyx_k_TN[] = "TN";
static const char __pyx_k_TP[] = "TP";
static const char __pyx_k__2[] = "";
static const char __pyx_k__5[] = "*";
static const char __pyx_k__6[] = "--";
static const char __pyx_k__7[] = " %";
static const char __pyx_k_ci[] = "ci";
static const char __pyx_k_fv[] = "fv";
static const char __pyx_k_io[] = "io";
static const char __pyx_k_mp[] = "mp";
static const char __pyx_k_pt[] = "pt";
static const char __pyx_k_ry[] = "ry";
static const char __pyx_k_ts[] = "ts";
static const char __pyx_k_vp[] = "vp";
static const char __pyx_k_xs[] = "xs";
static const char __pyx_k_ACC[] = "ACC";
static const char __pyx_k_AUC[] = " - AUC = ";
static const char __pyx_k_INF[] = "-INF";
static const char __pyx_k_PRC[] = "PRC";
static const char __pyx_k_ROC[] = "ROC";
static const char __pyx_k__11[] = ">";
static const char __pyx_k__17[] = " +/- ";
static const char __pyx_k__29[] = "\n";
static const char __pyx_k__45[] = "_";
static const char __pyx_k_all[] = "all";
static const char __pyx_k_arg[] = "arg";
static const char __pyx_k_bio[] = "bio";
static const char __pyx_k_blk[] = "blk";
static const char __pyx_k_div[] = "<div>";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_fig[] = "fig";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_hdr[] = "hdr";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_lbl[] = "lbl";
static const char __pyx_k_loc[] = "loc";
static const char __pyx_k_log[] = "log";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_mdl[] = "mdl";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_nNT[] = "nNT";
static const char __pyx_k_plt[] = "plt";
static const char __pyx_k_png[] = "png";
static const char __pyx_k_rep[] = "rep";
static const char __pyx_k_ret[] = "ret";
static const char __pyx_k_seq[] = "seq";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_win[] = "win";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_AUCs[] = "AUCs";
static const char __pyx_k_HTML[] = "HTML";
static const char __pyx_k_Name[] = "Name";
static const char __pyx_k_Pool[] = "Pool";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_cACC[] = "cACC";
static const char __pyx_k_core[] = "core";
static const char __pyx_k_cseq[] = "cseq";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_done[] = "done";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_grey[] = "grey";
static const char __pyx_k_html[] = "html";
static const char __pyx_k_imdl[] = "imdl";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_mIDs[] = "mIDs";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_mean[] = "mean";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_pIDs[] = "pIDs";
static const char __pyx_k_plot[] = "plot";
static const char __pyx_k_pool[] = "_pool";
static const char __pyx_k_pred[] = "pred";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_seqs[] = "seqs";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_span[] = "span";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_task[] = "task";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_vNeg[] = "vNeg";
static const char __pyx_k_vPos[] = "vPos";
static const char __pyx_k_vneg[] = "vneg";
static const char __pyx_k_vpos[] = "vpos";
static const char __pyx_k_AUC_2[] = "; AUC = ";
static const char __pyx_k_Label[] = "Label";
static const char __pyx_k_Model[] = "Model";
static const char __pyx_k_Score[] = "Score";
static const char __pyx_k_alpha[] = "alpha";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_color[] = "color";
static const char __pyx_k_cores[] = "cores";
static const char __pyx_k_curve[] = "curve";
static const char __pyx_k_cvval[] = "cvval";
static const char __pyx_k_div_2[] = "</div>";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_fetch[] = "fetch";
static const char __pyx_k_label[] = "label";
static const char __pyx_k_model[] = "model";
static const char __pyx_k_mtask[] = "mtask";
static const char __pyx_k_newID[] = "newID";
static const char __pyx_k_npred[] = "npred";
static const char __pyx_k_nvneg[] = "nvneg";
static const char __pyx_k_nvpos[] = "nvpos";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_rtask[] = "rtask";
static const char __pyx_k_rvneg[] = "rvneg";
static const char __pyx_k_rvpos[] = "rvpos";
static const char __pyx_k_score[] = "score";
static const char __pyx_k_spawn[] = "spawn";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_stats[] = "stats";
static const char __pyx_k_steps[] = "steps";
static const char __pyx_k_style[] = "style";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_title[] = "title";
static const char __pyx_k_train[] = "train";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_Recall[] = "Recall";
static const char __pyx_k_Repeat[] = "Repeat";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_base64[] = "base64";
static const char __pyx_k_common[] = "common";
static const char __pyx_k_config[] = "config";
static const char __pyx_k_curves[] = "curves";
static const char __pyx_k_cvpred[] = "cvpred";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_degree[] = "degree";
static const char __pyx_k_factor[] = "factor";
static const char __pyx_k_figure[] = "figure";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_genome[] = "genome";
static const char __pyx_k_getAUC[] = "getAUC";
static const char __pyx_k_getAll[] = "getAll";
static const char __pyx_k_getPRC[] = "getPRC";
static const char __pyx_k_getROC[] = "getROC";
static const char __pyx_k_ggplot[] = "ggplot";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_labels[] = "labels";
static const char __pyx_k_legend[] = "legend";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_models[] = "models";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_motifs[] = "motifs";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_ntrain[] = "ntrain";
static const char __pyx_k_pool_2[] = "pool";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_rtable[] = "rtable";
static const char __pyx_k_scores[] = "scores";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_stream[] = "stream";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_vPairs[] = "vPairs";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_wScore[] = "wScore";
static const char __pyx_k_weight[] = "weight";
static const char __pyx_k_winPos[] = "winPos";
static const char __pyx_k_winSeq[] = "winSeq";
static const char __pyx_k_xlabel[] = "xlabel";
static const char __pyx_k_xticks[] = "xticks";
static const char __pyx_k_ylabel[] = "ylabel";
static const char __pyx_k_yticks[] = "yticks";
static const char __pyx_k_BytesIO[] = "BytesIO";
static const char __pyx_k_CICurve[] = "CICurve";
static const char __pyx_k_CMStats[] = "CMStats";
static const char __pyx_k_PRC_AUC[] = "PRC AUC";
static const char __pyx_k_ROC_AUC[] = "ROC AUC";
static const char __pyx_k_Repeats[] = "Repeats:";
static const char __pyx_k_asTable[] = "asTable";
static const char __pyx_k_bgModel[] = "bgModel";
static const char __pyx_k_cScores[] = "cScores";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_cvtrain[] = "cvtrain";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_display[] = "display";
static const char __pyx_k_div_div[] = "</div></div>";
static const char __pyx_k_encoded[] = "encoded";
static const char __pyx_k_figsize[] = "figsize";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_ioputil[] = "ioputil";
static const char __pyx_k_meanLen[] = "meanLen";
static const char __pyx_k_meanlen[] = "meanlen";
static const char __pyx_k_modelID[] = "modelID";
static const char __pyx_k_nctable[] = "nctable";
static const char __pyx_k_outpath[] = "outpath";
static const char __pyx_k_plotPRC[] = "plotPRC";
static const char __pyx_k_plotROC[] = "plotROC";
static const char __pyx_k_predict[] = "predict";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_procIDs[] = "procIDs";
static const char __pyx_k_regions[] = "regions";
static const char __pyx_k_repeats[] = "repeats";
static const char __pyx_k_sScores[] = "sScores";
static const char __pyx_k_savefig[] = "savefig";
static const char __pyx_k_seqLens[] = "seqLens";
static const char __pyx_k_shuffle[] = "shuffle";
static const char __pyx_k_weights[] = "weights";
static const char __pyx_k_windows[] = "windows";
static const char __pyx_k_addModel[] = "addModel";
static const char __pyx_k_curvebyx[] = "curvebyx";
static const char __pyx_k_fancybox[] = "fancybox";
static const char __pyx_k_features[] = "features";
static const char __pyx_k_fontsize[] = "fontsize";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_getvalue[] = "getvalue";
static const char __pyx_k_hasCores[] = "hasCores";
static const char __pyx_k_initargs[] = "initargs";
static const char __pyx_k_negative[] = "negative";
static const char __pyx_k_pairFreq[] = "pairFreq";
static const char __pyx_k_positive[] = "positive";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_rotation[] = "rotation";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_tslabels[] = "tslabels";
static const char __pyx_k_Precision[] = "Precision";
static const char __pyx_k_Repeats_d[] = " - Repeats: %d";
static const char __pyx_k_Sequences[] = "; Sequences: ";
static const char __pyx_k_b64encode[] = "b64encode";
static const char __pyx_k_blkScores[] = "blkScores";
static const char __pyx_k_calibrate[] = "calibrate";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_genomeLen[] = "genomeLen";
static const char __pyx_k_getPRCAUC[] = "getPRCAUC";
static const char __pyx_k_getROCAUC[] = "getROCAUC";
static const char __pyx_k_legendLoc[] = "legendLoc";
static const char __pyx_k_linestyle[] = "linestyle";
static const char __pyx_k_meanCurve[] = "meanCurve";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_nCoresUse[] = "nCoresUse";
static const char __pyx_k_negatives[] = "negatives";
static const char __pyx_k_posScores[] = "posScores";
static const char __pyx_k_positives[] = "positives";
static const char __pyx_k_precision[] = "precision";
static const char __pyx_k_processID[] = "processID";
static const char __pyx_k_repr_html[] = "_repr_html_";
static const char __pyx_k_sequences[] = "sequences";
static const char __pyx_k_setNCores[] = "setNCores";
static const char __pyx_k_threshold[] = "threshold";
static const char __pyx_k_tsByLabel[] = "tsByLabel";
static const char __pyx_k_winPosSet[] = "winPosSet";
static const char __pyx_k_winSeqSet[] = "winSeqSet";
static const char __pyx_k_withLabel[] = "withLabel";
static const char __pyx_k_Validation[] = "Validation";
static const char __pyx_k_aPrecision[] = "aPrecision";
static const char __pyx_k_aThreshold[] = "aThreshold";
static const char __pyx_k_bPrecision[] = "bPrecision";
static const char __pyx_k_bThreshold[] = "bThreshold";
static const char __pyx_k_calibrator[] = "calibrator";
static const char __pyx_k_getTrainer[] = "getTrainer";
static const char __pyx_k_maxModelID[] = "maxModelID";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_returnHTML[] = "returnHTML";
static const char __pyx_k_upper_left[] = "upper left";
static const char __pyx_k_validation[] = "validation";
static const char __pyx_k_wPrecision[] = "wPrecision";
static const char __pyx_k_wThreshold[] = "wThreshold";
static const char __pyx_k_windowSize[] = "windowSize";
static const char __pyx_k_windowStep[] = "windowStep";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_Mean_length[] = "Mean length";
static const char __pyx_k_Predictions[] = "Predictions";
static const char __pyx_k_curvebyxmax[] = "curvebyxmax";
static const char __pyx_k_curvebyxmin[] = "curvebyxmin";
static const char __pyx_k_getAUCTable[] = "getAUCTable";
static const char __pyx_k_get_context[] = "get_context";
static const char __pyx_k_initializer[] = "initializer";
static const char __pyx_k_nTreadFetch[] = "nTreadFetch";
static const char __pyx_k_noCropNames[] = "noCropNames";
static const char __pyx_k_predictCore[] = "predictCore";
static const char __pyx_k_predictions[] = "predictions";
static const char __pyx_k_ratioNegPos[] = "ratioNegPos";
static const char __pyx_k_scoreWindow[] = "scoreWindow";
static const char __pyx_k_trainingSet[] = "trainingSet";
static const char __pyx_k_upper_right[] = "upper right";
static const char __pyx_k_winScoreSet[] = "winScoreSet";
static const char __pyx_k_Training_set[] = "; Training set: ";
static const char __pyx_k_bboxAnchorTo[] = "bboxAnchorTo";
static const char __pyx_k_bgModelOrder[] = "bgModelOrder";
static const char __pyx_k_defaultValue[] = "defaultValue";
static const char __pyx_k_fill_between[] = "fill_between";
static const char __pyx_k_fontsizeAxis[] = "fontsizeAxis";
static const char __pyx_k_getStatTable[] = "getStatTable";
static const char __pyx_k_progressTask[] = "progressTask";
static const char __pyx_k_setPoolModel[] = "setPoolModel";
static const char __pyx_k_storeTrained[] = "storeTrained";
static const char __pyx_k_tight_layout[] = "tight_layout";
static const char __pyx_k_Configuration[] = "Configuration";
static const char __pyx_k_Predictions_2[] = "Predictions: ";
static const char __pyx_k_crossvalidate[] = "crossvalidate";
static const char __pyx_k_gnocis_models[] = "gnocis.models";
static const char __pyx_k_labelNegative[] = "labelNegative";
static const char __pyx_k_labelPositive[] = "labelPositive";
static const char __pyx_k_motifSpectrum[] = "motifSpectrum";
static const char __pyx_k_scoreSequence[] = "scoreSequence";
static const char __pyx_k_sequenceModel[] = "sequenceModel";
static const char __pyx_k_sortedweights[] = "sortedweights";
static const char __pyx_k_trainedModels[] = "trainedModels";
static const char __pyx_k_validationSet[] = "validationSet";
static const char __pyx_k_Negative_label[] = "; Negative label: ";
static const char __pyx_k_Positive_label[] = "; Positive label: ";
static const char __pyx_k_Training_set_2[] = "Training set";
static const char __pyx_k_Training_set_3[] = "Training set:";
static const char __pyx_k_Training_set_s[] = " - Training set: %s";
static const char __pyx_k_Validation_set[] = "Validation set";
static const char __pyx_k_bbox_to_anchor[] = "bbox_to_anchor";
static const char __pyx_k_fixedStepCurve[] = "fixedStepCurve";
static const char __pyx_k_fontsizeLabels[] = "fontsizeLabels";
static const char __pyx_k_fontsizeLegend[] = "fontsizeLegend";
static const char __pyx_k_generateStream[] = "generateStream";
static const char __pyx_k_negativeScores[] = "negativeScores";
static const char __pyx_k_positiveScores[] = "positiveScores";
static const char __pyx_k_ratioTrainTest[] = "ratioTrainTest";
static const char __pyx_k_crossvalidation[] = "crossvalidation";
static const char __pyx_k_getPoolModelIDs[] = "getPoolModelIDs";
static const char __pyx_k_multiprocessing[] = "multiprocessing";
static const char __pyx_k_sequenceLengths[] = "sequenceLengths";
static const char __pyx_k_setNThreadFetch[] = "setNThreadFetch";
static const char __pyx_k_wantedPrecision[] = "wantedPrecision";
static const char __pyx_k_Cross_validation[] = "Cross-validation";
static const char __pyx_k_Predicting_cores[] = "Predicting cores";
static const char __pyx_k_Train_test_ratio[] = "Train/test ratio:";
static const char __pyx_k_Validation_set_2[] = "Validation set:";
static const char __pyx_k_Validation_set_s[] = " - Validation set: %s";
static const char __pyx_k_getPoolProcessID[] = "getPoolProcessID";
static const char __pyx_k_maxThreadFetchNT[] = "maxThreadFetchNT";
static const char __pyx_k_predictedRegions[] = "predictedRegions";
static const char __pyx_k_predictionCurves[] = "predictionCurves";
static const char __pyx_k_setPoolProcessID[] = "setPoolProcessID";
static const char __pyx_k_classScoresSorted[] = "classScoresSorted";
static const char __pyx_k_getPoolProcessIDs[] = "getPoolProcessIDs";
static const char __pyx_k_getSequenceScores[] = "getSequenceScores";
static const char __pyx_k_gnocis_models_pyx[] = "gnocis/models.pyx";
static const char __pyx_k_matplotlib_pyplot[] = "matplotlib.pyplot";
static const char __pyx_k_motifPairSpectrum[] = "motifPairSpectrum";
static const char __pyx_k_trainingNegatives[] = "trainingNegatives";
static const char __pyx_k_trainingPositives[] = "trainingPositives";
static const char __pyx_k_trainingSequences[] = "trainingSequences";
static const char __pyx_k_CVModelPredictions[] = "CVModelPredictions<";
static const char __pyx_k_Calibrating_models[] = "Calibrating models";
static const char __pyx_k_Cross_validation_2[] = "Cross-validation\n";
static const char __pyx_k_Expected_at_random[] = "Expected at random";
static const char __pyx_k_True_Positive_Rate[] = "True Positive Rate";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_getConfusionMatrix[] = "getConfusionMatrix";
static const char __pyx_k_sequenceModelDummy[] = "sequenceModelDummy";
static const char __pyx_k_setCorePredictions[] = "setCorePredictions";
static const char __pyx_k_False_Positive_Rate[] = "False Positive Rate";
static const char __pyx_k_multiprocessingPool[] = "multiprocessingPool";
static const char __pyx_k_printTestStatistics[] = "printTestStatistics";
static const char __pyx_k_sequenceModel_score[] = "sequenceModel.score";
static const char __pyx_k_sequenceModel_train[] = "sequenceModel.train";
static const char __pyx_k_setPoolProcessModel[] = "setPoolProcessModel";
static const char __pyx_k_trainPREdictorModel[] = "trainPREdictorModel";
static const char __pyx_k_CVModelPredictions_2[] = "CVModelPredictions";
static const char __pyx_k_IPython_core_display[] = "IPython.core.display";
static const char __pyx_k_div_style_float_left[] = "<div style=\"float: left;\">";
static const char __pyx_k_sequenceModelLogOdds[] = "sequenceModelLogOdds";
static const char __pyx_k_sequenceModel___init[] = "sequenceModel.__init__";
static const char __pyx_k_sequenceModel_getPRC[] = "sequenceModel.getPRC";
static const char __pyx_k_sequenceModel_getROC[] = "sequenceModel.getROC";
static const char __pyx_k_Evaluation_statistics[] = "Evaluation statistics";
static const char __pyx_k_enableMultiprocessing[] = "enableMultiprocessing";
static const char __pyx_k_getConfigurationTable[] = "getConfigurationTable";
static const char __pyx_k_getPoolProcessModelID[] = "getPoolProcessModelID";
static const char __pyx_k_getPrecisionThreshold[] = "getPrecisionThreshold";
static const char __pyx_k_initializePoolProcess[] = "initializePoolProcess";
static const char __pyx_k_labelPredictionCurves[] = "labelPredictionCurves";
static const char __pyx_k_sequenceModel_plotPRC[] = "sequenceModel.plotPRC";
static const char __pyx_k_sequenceModel_predict[] = "sequenceModel.predict";
static const char __pyx_k_streamSequenceWindows[] = "streamSequenceWindows";
static const char __pyx_k_Dummy_model_Features_s[] = "Dummy model<Features: %s>";
static const char __pyx_k_Negatives_per_positive[] = "Negatives per positive:";
static const char __pyx_k_Predicting_genome_wide[] = "Predicting genome-wide";
static const char __pyx_k_Scored_sequences_Model[] = "Scored sequences - Model: ";
static const char __pyx_k_crossvalidation___init[] = "crossvalidation.__init__";
static const char __pyx_k_crossvalidation___repr[] = "crossvalidation.__repr__";
static const char __pyx_k_getPoolProcessHasModel[] = "getPoolProcessHasModel";
static const char __pyx_k_Log_odds_model_Features[] = "Log-odds model<Features: ";
static const char __pyx_k_crossvalidation_plotPRC[] = "crossvalidation.plotPRC";
static const char __pyx_k_crossvalidation_plotROC[] = "crossvalidation.plotROC";
static const char __pyx_k_crossvalidation_predict[] = "crossvalidation.predict";
static const char __pyx_k_getValidationStatistics[] = "getValidationStatistics";
static const char __pyx_k_sequenceModel_getPRCAUC[] = "sequenceModel.getPRCAUC";
static const char __pyx_k_sequenceModel_getROCAUC[] = "sequenceModel.getROCAUC";
static const char __pyx_k_crossvalidation_addModel[] = "crossvalidation.addModel";
static const char __pyx_k_initializePoolProcessIDs[] = "initializePoolProcessIDs";
static const char __pyx_k_sequenceModelDummy___str[] = "sequenceModelDummy.__str__";
static const char __pyx_k_sequenceModel___getstate[] = "sequenceModel.__getstate__";
static const char __pyx_k_sequenceModel___setstate[] = "sequenceModel.__setstate__";
static const char __pyx_k_CVModelPredictions___init[] = "CVModelPredictions.__init__";
static const char __pyx_k_CVModelPredictions___repr[] = "CVModelPredictions.__repr__";
static const char __pyx_k_createDummyPREdictorModel[] = "createDummyPREdictorModel";
static const char __pyx_k_crossvalidation_calibrate[] = "crossvalidation.calibrate";
static const char __pyx_k_div_div_style_float_right[] = "</div><div style=\"float: right;\">";
static const char __pyx_k_multiprocessModelDeployer[] = "multiprocessModelDeployer";
static const char __pyx_k_printValidationStatistics[] = "printValidationStatistics";
static const char __pyx_k_sequenceModelDummy___init[] = "sequenceModelDummy.__init__";
static const char __pyx_k_sequenceModelDummy___repr[] = "sequenceModelDummy.__repr__";
static const char __pyx_k_sequenceModel_predictCore[] = "sequenceModel.predictCore";
static const char __pyx_k_trainSinglePREdictorModel[] = "trainSinglePREdictorModel";
static const char __pyx_k_CVModelPredictions_regions[] = "CVModelPredictions.regions";
static const char __pyx_k_crossvalidation__repr_html[] = "crossvalidation._repr_html_";
static const char __pyx_k_sequenceModelLogOdds___str[] = "sequenceModelLogOdds.__str__";
static const char __pyx_k_crossvalidation_getAUCTable[] = "crossvalidation.getAUCTable";
static const char __pyx_k_getOptimalAccuracyThreshold[] = "getOptimalAccuracyThreshold";
static const char __pyx_k_predictSequenceStreamCurves[] = "predictSequenceStreamCurves";
static const char __pyx_k_sequenceModelLogOdds___init[] = "sequenceModelLogOdds.__init__";
static const char __pyx_k_sequenceModelLogOdds___repr[] = "sequenceModelLogOdds.__repr__";
static const char __pyx_k_sequenceModel_scoreSequence[] = "sequenceModel.scoreSequence";
static const char __pyx_k_setPoolModel_locals_genexpr[] = "setPoolModel.<locals>.genexpr";
static const char __pyx_k_calibrateGenomewidePrecision[] = "calibrateGenomewidePrecision";
static const char __pyx_k_div_b_Cross_validation_b_div[] = "<div><b>Cross-validation</b></div>";
static const char __pyx_k_getConfusionMatrixStatistics[] = "getConfusionMatrixStatistics";
static const char __pyx_k_predictSequenceStreamRegions[] = "predictSequenceStreamRegions";
static const char __pyx_k_CVModelPredictions__repr_html[] = "CVModelPredictions._repr_html_";
static const char __pyx_k_Generating_training_test_sets[] = "Generating training/test sets";
static const char __pyx_k_Training_sequences_per_repeat[] = "Training sequences per repeat";
static const char __pyx_k_sequenceModelDummy_getTrainer[] = "sequenceModelDummy.getTrainer";
static const char __pyx_k_CVModelPredictions_predictCore[] = "CVModelPredictions.predictCore";
static const char __pyx_k_This_model_takes_a_feature_set[] = "\n\tThis model takes a feature set, and scores input sequences by summing feature values, without weighting.\n\t\n\t:param name: Model name.\n\t:param features: The feature set.\n\t:param windowSize: Window size to use.\n\t:param windowStep: Window step size to use.\n\t\n\t:type name: str\n\t:type features: features\n\t:type windowSize: int\n\t:type windowStep: int\n\t";
static const char __pyx_k_Training_sequences_per_class_d[] = " - Training sequences per class: %d";
static const char __pyx_k_getPoolModelIDs_locals_genexpr[] = "getPoolModelIDs.<locals>.genexpr";
static const char __pyx_k_sequenceModelDummy_scoreWindow[] = "sequenceModelDummy.scoreWindow";
static const char __pyx_k_CVModelPredictions_getStatTable[] = "CVModelPredictions.getStatTable";
static const char __pyx_k_Constructs_a_log_odds_model_bas[] = "\n\tConstructs a log-odds model based on an input feature set and binary training set, and scores input sequences by summing log-odds-weighted feature values.\n\t\n\t:param name: Model name.\n\t:param features: The feature set.\n\t:param trainingSet: Training sequences.\n\t:param windowSize: Window size to use.\n\t:param windowStep: Window step size to use.\n\t:param labelPositive: Positive training class label.\n\t:param labelNegative: Negative training class label.\n\t\n\t:type name: str\n\t:type features: features\n\t:type trainingSet: sequences\n\t:type windowSize: int\n\t:type windowStep: int\n\t:type labelPositive: sequenceLabel\n\t:type labelNegative: sequenceLabel\n\t";
static const char __pyx_k_Helper_class_for_cross_validati[] = "\n\tHelper class for cross-validations. Accepts binary training and validation sets and constructs cross-validation sets for a desired number of repeats. If a separate validation set is not given, the training set is used. The cross-validation set for each repeat contains numbers of training and test sequences determined by a training-to-testing sequence ratio, as well as a negative-per-positive test sequence ratio. When constructing the validation set, identities are checked for against the training set, to avoid contamination (will not work if sequences are cloned). Holds models and validation statistics. Integrates with terminal and IPython for visualization of results. Stores training and testing data, so new models can be added.\n\t\n\t:param models: List of models to cross-validate.\n\t:param tpos: Positive training set.\n\t:param tneg: Negative training set.\n\t:param vpos: Positive validation set.\n\t:param vneg: Negative validation set.\n\t:param repeats: Number of experimental repeats. Default = 20.\n\t:param ratioTrainTest: Ratio of training to testing sequences. Default = 80%.\n\t:param ratioNegPos: Ratio of validation negatives to positives. Default = 100.\n\t\n\t:type models: list\n\t:type tpos: sequences\n\t:type tneg: sequences\n\t:type vpos: sequences, optional\n\t:type vneg: sequences, optional\n\t:type repeats: int, optional\n\t:type ratioTrainTest: float, optional\n\t:type ratioNegPos: float, optional\n\t";
static const char __pyx_k_Helper_class_that_represents_cr[] = "\n\tHelper class that represents cross-validated model predictions..\n\t\n\t:param model: Model.\n\t:param labelPredictionCurves: Labelled prediction curves.\n\t\n\t:type model: sequenceModel\n\t:type labelPredictionCurves: dict\n\t";
static const char __pyx_k_The_sequenceModel_class_is_an_a[] = "\n\tThe `sequenceModel` class is an abstract class for sequence models. A number of methods are implemented for machine learning and prediction with DNA sequences.\n\t\n\t:param name: Name of model.\n\t:param enableMultiprocessing: If True (default), multiprocessing is enabled.\n\t\n\t:type name: string\n\t:type enableMultiprocessing: bool\n\t";
static const char __pyx_k_Validation_sequences_per_repeat[] = "Validation sequences per repeat";
static const char __pyx_k_crossvalidation___init___locals[] = "crossvalidation.__init__.<locals>.genexpr";
static const char __pyx_k_img_src_data_image_png_base64_s[] = "<img src='data:image/png;base64,%s'>";
static const char __pyx_k_sequenceModelLogOdds_getTrainer[] = "sequenceModelLogOdds.getTrainer.<locals>.<lambda>";
static const char __pyx_k_sequenceModel_getSequenceScores[] = "sequenceModel.getSequenceScores";
static const char __pyx_k_CVModelPredictions_setCorePredic[] = "CVModelPredictions.setCorePredictions";
static const char __pyx_k_Warning_model_s_not_trained_Skip[] = "Warning: model %s not trained. Skipping.";
static const char __pyx_k_crossvalidation_getConfiguration[] = "crossvalidation.getConfigurationTable";
static const char __pyx_k_crossvalidation_plotPRC_locals_g[] = "crossvalidation.plotPRC.<locals>.genexpr";
static const char __pyx_k_crossvalidation_plotROC_locals_g[] = "crossvalidation.plotROC.<locals>.genexpr";
static const char __pyx_k_getPoolProcessIDs_locals_genexpr[] = "getPoolProcessIDs.<locals>.genexpr";
static const char __pyx_k_multiprocessModelDeployer_locals[] = "multiprocessModelDeployer.<locals>.genexpr";
static const char __pyx_k_sequenceModelDummy_getTrainer_lo[] = "sequenceModelDummy.getTrainer.<locals>.<lambda>";
static const char __pyx_k_sequenceModelLogOdds___init___lo[] = "sequenceModelLogOdds.__init__.<locals>.genexpr";
static const char __pyx_k_sequenceModelLogOdds_scoreWindow[] = "sequenceModelLogOdds.scoreWindow";
static const char __pyx_k_sequenceModel_calibrateGenomewid[] = "sequenceModel.calibrateGenomewidePrecision.<locals>.genexpr";
static const char __pyx_k_sequenceModel_getConfusionMatrix[] = "sequenceModel.getConfusionMatrix";
static const char __pyx_k_sequenceModel_getOptimalAccuracy[] = "sequenceModel.getOptimalAccuracyThreshold.<locals>.<lambda>";
static const char __pyx_k_sequenceModel_getPrecisionThresh[] = "sequenceModel.getPrecisionThreshold.<locals>.<lambda>";
static const char __pyx_k_sequenceModel_getValidationStati[] = "sequenceModel.getValidationStatistics";
static const char __pyx_k_sequenceModel_predictSequenceStr[] = "sequenceModel.predictSequenceStreamCurves.<locals>.genexpr";
static const char __pyx_k_sequenceModel_printTestStatistic[] = "sequenceModel.printTestStatistics";
static const char __pyx_k_sequenceModel_scoreSequence_loca[] = "sequenceModel.scoreSequence.<locals>.genexpr";
static const char __pyx_k_sequenceModelLogOdds_getTrainer_2[] = "sequenceModelLogOdds.getTrainer";
static const char __pyx_k_sequenceModel_calibrateGenomewid_2[] = "sequenceModel.calibrateGenomewidePrecision";
static const char __pyx_k_sequenceModel_getOptimalAccuracy_2[] = "sequenceModel.getOptimalAccuracyThreshold";
static const char __pyx_k_sequenceModel_getPrecisionThresh_2[] = "sequenceModel.getPrecisionThreshold";
static const char __pyx_k_sequenceModel_predictSequenceStr_2[] = "sequenceModel.predictSequenceStreamRegions";
static const char __pyx_k_sequenceModel_predictSequenceStr_3[] = "sequenceModel.predictSequenceStreamCurves";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_kp_u_2f;
static PyObject *__pyx_n_s_ACC;
static PyObject *__pyx_kp_u_AUC;
static PyObject *__pyx_kp_u_AUC_2;
static PyObject *__pyx_n_s_AUCs;
static PyObject *__pyx_n_s_BytesIO;
static PyObject *__pyx_n_s_CI;
static PyObject *__pyx_n_s_CICurve;
static PyObject *__pyx_n_s_CMStats;
static PyObject *__pyx_n_u_CV;
static PyObject *__pyx_kp_u_CVModelPredictions;
static PyObject *__pyx_n_s_CVModelPredictions_2;
static PyObject *__pyx_n_s_latin1_CVModelPredictions_2;
static PyObject *__pyx_n_s_CVModelPredictions___init;
static PyObject *__pyx_n_s_CVModelPredictions___repr;
static PyObject *__pyx_n_s_CVModelPredictions__repr_html;
static PyObject *__pyx_n_s_CVModelPredictions_getStatTable;
static PyObject *__pyx_n_s_CVModelPredictions_predictCore;
static PyObject *__pyx_n_s_CVModelPredictions_regions;
static PyObject *__pyx_n_s_CVModelPredictions_setCorePredic;
static PyObject *__pyx_kp_u_Calibrating_models;
static PyObject *__pyx_n_u_Configuration;
static PyObject *__pyx_kp_s_Constructs_a_log_odds_model_bas;
static PyObject *__pyx_kp_u_Cross_validation;
static PyObject *__pyx_kp_u_Cross_validation_2;
static PyObject *__pyx_kp_u_Dummy_model_Features_s;
static PyObject *__pyx_kp_u_Evaluation_statistics;
static PyObject *__pyx_kp_u_Expected_at_random;
static PyObject *__pyx_n_s_FN;
static PyObject *__pyx_n_s_FP;
static PyObject *__pyx_kp_u_False_Positive_Rate;
static PyObject *__pyx_kp_u_Generating_training_test_sets;
static PyObject *__pyx_n_s_HTML;
static PyObject *__pyx_kp_s_Helper_class_for_cross_validati;
static PyObject *__pyx_kp_s_Helper_class_that_represents_cr;
static PyObject *__pyx_kp_u_INF;
static PyObject *__pyx_n_s_IPython_core_display;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_u_Label;
static PyObject *__pyx_kp_u_Log_odds_model_Features;
static PyObject *__pyx_kp_u_Mean_length;
static PyObject *__pyx_n_u_Model;
static PyObject *__pyx_n_u_Name;
static PyObject *__pyx_kp_u_Negative_label;
static PyObject *__pyx_kp_u_Negatives_per_positive;
static PyObject *__pyx_n_s_PRC;
static PyObject *__pyx_kp_u_PRC_AUC;
static PyObject *__pyx_n_s_Pool;
static PyObject *__pyx_kp_u_Positive_label;
static PyObject *__pyx_n_u_Precision;
static PyObject *__pyx_kp_u_Predicting_cores;
static PyObject *__pyx_kp_u_Predicting_genome_wide;
static PyObject *__pyx_n_u_Predictions;
static PyObject *__pyx_kp_u_Predictions_2;
static PyObject *__pyx_n_s_ROC;
static PyObject *__pyx_kp_u_ROC_AUC;
static PyObject *__pyx_n_u_Recall;
static PyObject *__pyx_n_u_Repeat;
static PyObject *__pyx_kp_u_Repeats;
static PyObject *__pyx_kp_u_Repeats_d;
static PyObject *__pyx_n_u_Score;
static PyObject *__pyx_kp_u_Scored_sequences_Model;
static PyObject *__pyx_kp_u_Sequences;
static PyObject *__pyx_n_s_TN;
static PyObject *__pyx_n_s_TP;
static PyObject *__pyx_kp_s_The_sequenceModel_class_is_an_a;
static PyObject *__pyx_kp_s_This_model_takes_a_feature_set;
static PyObject *__pyx_kp_u_Train_test_ratio;
static PyObject *__pyx_kp_u_Training_sequences_per_class_d;
static PyObject *__pyx_kp_u_Training_sequences_per_repeat;
static PyObject *__pyx_kp_u_Training_set;
static PyObject *__pyx_kp_u_Training_set_2;
static PyObject *__pyx_kp_u_Training_set_3;
static PyObject *__pyx_kp_u_Training_set_s;
static PyObject *__pyx_kp_u_True_Positive_Rate;
static PyObject *__pyx_n_u_Validation;
static PyObject *__pyx_kp_u_Validation_sequences_per_repeat;
static PyObject *__pyx_kp_u_Validation_set;
static PyObject *__pyx_kp_u_Validation_set_2;
static PyObject *__pyx_kp_u_Validation_set_s;
static PyObject *__pyx_kp_u_Warning_model_s_not_trained_Skip;
static PyObject *__pyx_kp_u__11;
static PyObject *__pyx_kp_u__17;
static PyObject *__pyx_kp_u__2;
static PyObject *__pyx_kp_u__29;
static PyObject *__pyx_n_s__45;
static PyObject *__pyx_n_s__5;
static PyObject *__pyx_kp_u__6;
static PyObject *__pyx_kp_u__7;
static PyObject *__pyx_n_s_aPrecision;
static PyObject *__pyx_n_s_aThreshold;
static PyObject *__pyx_n_s_addModel;
static PyObject *__pyx_n_u_all;
static PyObject *__pyx_n_s_alpha;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_arg;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_asTable;
static PyObject *__pyx_n_s_b64encode;
static PyObject *__pyx_n_s_bPrecision;
static PyObject *__pyx_n_s_bThreshold;
static PyObject *__pyx_n_s_base64;
static PyObject *__pyx_n_s_bboxAnchorTo;
static PyObject *__pyx_n_s_bbox_to_anchor;
static PyObject *__pyx_n_s_bgModel;
static PyObject *__pyx_n_s_bgModelOrder;
static PyObject *__pyx_n_s_bio;
static PyObject *__pyx_n_s_blk;
static PyObject *__pyx_n_s_blkScores;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_cACC;
static PyObject *__pyx_n_s_cScores;
static PyObject *__pyx_n_s_calibrate;
static PyObject *__pyx_n_s_calibrateGenomewidePrecision;
static PyObject *__pyx_n_s_calibrator;
static PyObject *__pyx_n_s_ci;
static PyObject *__pyx_n_s_classScoresSorted;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_color;
static PyObject *__pyx_n_s_common;
static PyObject *__pyx_n_s_config;
static PyObject *__pyx_n_s_context;
static PyObject *__pyx_n_s_core;
static PyObject *__pyx_n_s_cores;
static PyObject *__pyx_n_s_createDummyPREdictorModel;
static PyObject *__pyx_n_s_crossvalidate;
static PyObject *__pyx_n_s_crossvalidation;
static PyObject *__pyx_n_s_latin1_crossvalidation;
static PyObject *__pyx_n_s_crossvalidation___init;
static PyObject *__pyx_n_s_crossvalidation___init___locals;
static PyObject *__pyx_n_s_crossvalidation___repr;
static PyObject *__pyx_n_s_crossvalidation__repr_html;
static PyObject *__pyx_n_s_crossvalidation_addModel;
static PyObject *__pyx_n_s_crossvalidation_calibrate;
static PyObject *__pyx_n_s_crossvalidation_getAUCTable;
static PyObject *__pyx_n_s_crossvalidation_getConfiguration;
static PyObject *__pyx_n_s_crossvalidation_plotPRC;
static PyObject *__pyx_n_s_crossvalidation_plotPRC_locals_g;
static PyObject *__pyx_n_s_crossvalidation_plotROC;
static PyObject *__pyx_n_s_crossvalidation_plotROC_locals_g;
static PyObject *__pyx_n_s_crossvalidation_predict;
static PyObject *__pyx_n_s_cseq;
static PyObject *__pyx_n_s_curve;
static PyObject *__pyx_n_s_curvebyx;
static PyObject *__pyx_n_s_curvebyxmax;
static PyObject *__pyx_n_s_curvebyxmin;
static PyObject *__pyx_n_s_curves;
static PyObject *__pyx_n_s_cvpred;
static PyObject *__pyx_n_s_cvtrain;
static PyObject *__pyx_n_s_cvval;
static PyObject *__pyx_n_u_d;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_s_defaultValue;
static PyObject *__pyx_n_s_degree;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_display;
static PyObject *__pyx_kp_u_div;
static PyObject *__pyx_kp_u_div_2;
static PyObject *__pyx_kp_u_div_b_Cross_validation_b_div;
static PyObject *__pyx_kp_u_div_div;
static PyObject *__pyx_kp_u_div_div_style_float_right;
static PyObject *__pyx_kp_u_div_style_float_left;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_done;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_enableMultiprocessing;
static PyObject *__pyx_n_s_encoded;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_factor;
static PyObject *__pyx_n_s_fancybox;
static PyObject *__pyx_n_s_features;
static PyObject *__pyx_n_s_fetch;
static PyObject *__pyx_n_s_fig;
static PyObject *__pyx_n_s_figsize;
static PyObject *__pyx_n_s_figure;
static PyObject *__pyx_n_s_fill_between;
static PyObject *__pyx_n_s_fixedStepCurve;
static PyObject *__pyx_n_s_fontsize;
static PyObject *__pyx_n_s_fontsizeAxis;
static PyObject *__pyx_n_s_fontsizeLabels;
static PyObject *__pyx_n_s_fontsizeLegend;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fv;
static PyObject *__pyx_n_s_generateStream;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_genome;
static PyObject *__pyx_n_s_genomeLen;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getAUC;
static PyObject *__pyx_n_s_getAUCTable;
static PyObject *__pyx_n_s_getAll;
static PyObject *__pyx_n_s_getConfigurationTable;
static PyObject *__pyx_n_s_getConfusionMatrix;
static PyObject *__pyx_n_s_getConfusionMatrixStatistics;
static PyObject *__pyx_n_s_getOptimalAccuracyThreshold;
static PyObject *__pyx_n_s_getPRC;
static PyObject *__pyx_n_s_getPRCAUC;
static PyObject *__pyx_n_s_getPoolModelIDs;
static PyObject *__pyx_n_s_getPoolModelIDs_locals_genexpr;
static PyObject *__pyx_n_s_getPoolProcessHasModel;
static PyObject *__pyx_n_s_getPoolProcessID;
static PyObject *__pyx_n_s_getPoolProcessIDs;
static PyObject *__pyx_n_s_getPoolProcessIDs_locals_genexpr;
static PyObject *__pyx_n_s_getPoolProcessModelID;
static PyObject *__pyx_n_s_getPrecisionThreshold;
static PyObject *__pyx_n_s_getROC;
static PyObject *__pyx_n_s_getROCAUC;
static PyObject *__pyx_n_s_getSequenceScores;
static PyObject *__pyx_n_s_getStatTable;
static PyObject *__pyx_n_s_getTrainer;
static PyObject *__pyx_n_s_getValidationStatistics;
static PyObject *__pyx_n_s_get_context;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_getvalue;
static PyObject *__pyx_n_u_ggplot;
static PyObject *__pyx_n_s_gnocis_models;
static PyObject *__pyx_kp_s_gnocis_models_pyx;
static PyObject *__pyx_n_u_grey;
static PyObject *__pyx_n_s_hasCores;
static PyObject *__pyx_n_s_hdr;
static PyObject *__pyx_n_s_html;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_imdl;
static PyObject *__pyx_kp_u_img_src_data_image_png_base64_s;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_initargs;
static PyObject *__pyx_n_s_initializePoolProcess;
static PyObject *__pyx_n_s_initializePoolProcessIDs;
static PyObject *__pyx_n_s_initializer;
static PyObject *__pyx_n_s_io;
static PyObject *__pyx_n_s_ioputil;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_label;
static PyObject *__pyx_n_s_labelNegative;
static PyObject *__pyx_n_s_labelPositive;
static PyObject *__pyx_n_s_labelPredictionCurves;
static PyObject *__pyx_n_s_labels;
static PyObject *__pyx_n_s_lbl;
static PyObject *__pyx_n_s_legend;
static PyObject *__pyx_n_s_legendLoc;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_s_linestyle;
static PyObject *__pyx_n_s_loc;
static PyObject *__pyx_n_s_log;
static PyObject *__pyx_n_s_mIDs;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_math;
static PyObject *__pyx_n_s_matplotlib_pyplot;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_maxModelID;
static PyObject *__pyx_n_s_maxThreadFetchNT;
static PyObject *__pyx_n_s_mdl;
static PyObject *__pyx_n_s_mean;
static PyObject *__pyx_n_s_meanCurve;
static PyObject *__pyx_n_s_meanLen;
static PyObject *__pyx_n_s_meanlen;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_model;
static PyObject *__pyx_n_u_model;
static PyObject *__pyx_n_s_modelID;
static PyObject *__pyx_n_u_modelID;
static PyObject *__pyx_n_s_models;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_motifPairSpectrum;
static PyObject *__pyx_n_s_motifSpectrum;
static PyObject *__pyx_n_s_motifs;
static PyObject *__pyx_n_s_mp;
static PyObject *__pyx_n_s_mtask;
static PyObject *__pyx_n_s_multiprocessModelDeployer;
static PyObject *__pyx_n_s_multiprocessModelDeployer_locals;
static PyObject *__pyx_n_s_multiprocessing;
static PyObject *__pyx_n_s_multiprocessingPool;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_nCoresUse;
static PyObject *__pyx_n_s_nNT;
static PyObject *__pyx_n_s_nTreadFetch;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_nctable;
static PyObject *__pyx_n_s_negative;
static PyObject *__pyx_n_s_negativeScores;
static PyObject *__pyx_n_s_negatives;
static PyObject *__pyx_n_u_negatives;
static PyObject *__pyx_n_s_newID;
static PyObject *__pyx_n_s_noCropNames;
static PyObject *__pyx_n_s_npred;
static PyObject *__pyx_n_s_ntrain;
static PyObject *__pyx_n_s_nvneg;
static PyObject *__pyx_n_s_nvpos;
static PyObject *__pyx_n_s_outpath;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_n_s_pIDs;
static PyObject *__pyx_n_s_pairFreq;
static PyObject *__pyx_n_s_plot;
static PyObject *__pyx_n_s_plotPRC;
static PyObject *__pyx_n_s_plotROC;
static PyObject *__pyx_n_s_plt;
static PyObject *__pyx_n_u_png;
static PyObject *__pyx_n_s_pool;
static PyObject *__pyx_n_s_pool_2;
static PyObject *__pyx_n_s_posScores;
static PyObject *__pyx_n_s_positive;
static PyObject *__pyx_n_s_positiveScores;
static PyObject *__pyx_n_s_positives;
static PyObject *__pyx_n_u_positives;
static PyObject *__pyx_n_s_precision;
static PyObject *__pyx_n_s_pred;
static PyObject *__pyx_n_s_predict;
static PyObject *__pyx_n_s_predictCore;
static PyObject *__pyx_n_s_predictSequenceStreamCurves;
static PyObject *__pyx_n_s_predictSequenceStreamRegions;
static PyObject *__pyx_n_s_predictedRegions;
static PyObject *__pyx_n_s_predictionCurves;
static PyObject *__pyx_n_s_predictions;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_printTestStatistics;
static PyObject *__pyx_n_s_printValidationStatistics;
static PyObject *__pyx_n_s_procIDs;
static PyObject *__pyx_n_s_processID;
static PyObject *__pyx_n_s_progressTask;
static PyObject *__pyx_n_s_pt;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_ratioNegPos;
static PyObject *__pyx_n_s_ratioTrainTest;
static PyObject *__pyx_n_s_regions;
static PyObject *__pyx_n_s_rep;
static PyObject *__pyx_n_s_repeats;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_repr_html;
static PyObject *__pyx_n_s_ret;
static PyObject *__pyx_n_s_returnHTML;
static PyObject *__pyx_n_s_rotation;
static PyObject *__pyx_n_s_rtable;
static PyObject *__pyx_n_s_rtask;
static PyObject *__pyx_n_s_rvneg;
static PyObject *__pyx_n_s_rvpos;
static PyObject *__pyx_n_s_ry;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_sScores;
static PyObject *__pyx_n_s_savefig;
static PyObject *__pyx_n_s_score;
static PyObject *__pyx_n_s_scoreSequence;
static PyObject *__pyx_n_s_scoreWindow;
static PyObject *__pyx_n_s_scores;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_seq;
static PyObject *__pyx_n_s_seqLens;
static PyObject *__pyx_n_s_seqs;
static PyObject *__pyx_n_s_sequenceLengths;
static PyObject *__pyx_n_s_sequenceModel;
static PyObject *__pyx_n_s_latin1_sequenceModel;
static PyObject *__pyx_n_s_sequenceModelDummy;
static PyObject *__pyx_n_s_latin1_sequenceModelDummy;
static PyObject *__pyx_n_s_sequenceModelDummy___init;
static PyObject *__pyx_n_s_sequenceModelDummy___repr;
static PyObject *__pyx_n_s_sequenceModelDummy___str;
static PyObject *__pyx_n_s_sequenceModelDummy_getTrainer;
static PyObject *__pyx_n_s_sequenceModelDummy_getTrainer_lo;
static PyObject *__pyx_n_s_sequenceModelDummy_scoreWindow;
static PyObject *__pyx_n_s_sequenceModelLogOdds;
static PyObject *__pyx_n_s_latin1_sequenceModelLogOdds;
static PyObject *__pyx_n_s_sequenceModelLogOdds___init;
static PyObject *__pyx_n_s_sequenceModelLogOdds___init___lo;
static PyObject *__pyx_n_s_sequenceModelLogOdds___repr;
static PyObject *__pyx_n_s_sequenceModelLogOdds___str;
static PyObject *__pyx_n_s_sequenceModelLogOdds_getTrainer;
static PyObject *__pyx_n_s_sequenceModelLogOdds_getTrainer_2;
static PyObject *__pyx_n_s_sequenceModelLogOdds_scoreWindow;
static PyObject *__pyx_n_s_sequenceModel___getstate;
static PyObject *__pyx_n_s_sequenceModel___init;
static PyObject *__pyx_n_s_sequenceModel___setstate;
static PyObject *__pyx_n_s_sequenceModel_calibrateGenomewid;
static PyObject *__pyx_n_s_sequenceModel_calibrateGenomewid_2;
static PyObject *__pyx_n_s_sequenceModel_getConfusionMatrix;
static PyObject *__pyx_n_s_sequenceModel_getOptimalAccuracy;
static PyObject *__pyx_n_s_sequenceModel_getOptimalAccuracy_2;
static PyObject *__pyx_n_s_sequenceModel_getPRC;
static PyObject *__pyx_n_s_sequenceModel_getPRCAUC;
static PyObject *__pyx_n_s_sequenceModel_getPrecisionThresh;
static PyObject *__pyx_n_s_sequenceModel_getPrecisionThresh_2;
static PyObject *__pyx_n_s_sequenceModel_getROC;
static PyObject *__pyx_n_s_sequenceModel_getROCAUC;
static PyObject *__pyx_n_s_sequenceModel_getSequenceScores;
static PyObject *__pyx_n_s_sequenceModel_getValidationStati;
static PyObject *__pyx_n_s_sequenceModel_plotPRC;
static PyObject *__pyx_n_s_sequenceModel_predict;
static PyObject *__pyx_n_s_sequenceModel_predictCore;
static PyObject *__pyx_n_s_sequenceModel_predictSequenceStr;
static PyObject *__pyx_n_s_sequenceModel_predictSequenceStr_2;
static PyObject *__pyx_n_s_sequenceModel_predictSequenceStr_3;
static PyObject *__pyx_n_s_sequenceModel_printTestStatistic;
static PyObject *__pyx_n_s_sequenceModel_score;
static PyObject *__pyx_n_s_sequenceModel_scoreSequence;
static PyObject *__pyx_n_s_sequenceModel_scoreSequence_loca;
static PyObject *__pyx_n_s_sequenceModel_train;
static PyObject *__pyx_n_s_sequences;
static PyObject *__pyx_n_s_setCorePredictions;
static PyObject *__pyx_n_s_setNCores;
static PyObject *__pyx_n_s_setNThreadFetch;
static PyObject *__pyx_n_s_setPoolModel;
static PyObject *__pyx_n_s_setPoolModel_locals_genexpr;
static PyObject *__pyx_n_s_setPoolProcessID;
static PyObject *__pyx_n_s_setPoolProcessModel;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_shuffle;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_sortedweights;
static PyObject *__pyx_n_s_span;
static PyObject *__pyx_n_u_spawn;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_stats;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_steps;
static PyObject *__pyx_n_s_storeTrained;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_stream;
static PyObject *__pyx_n_s_streamSequenceWindows;
static PyObject *__pyx_n_s_style;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_s_target;
static PyObject *__pyx_n_s_task;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_threshold;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tight_layout;
static PyObject *__pyx_n_u_title;
static PyObject *__pyx_n_s_train;
static PyObject *__pyx_n_s_trainPREdictorModel;
static PyObject *__pyx_n_s_trainSinglePREdictorModel;
static PyObject *__pyx_n_s_trainedModels;
static PyObject *__pyx_n_s_trainingNegatives;
static PyObject *__pyx_n_s_trainingPositives;
static PyObject *__pyx_n_s_trainingSequences;
static PyObject *__pyx_n_s_trainingSet;
static PyObject *__pyx_n_s_ts;
static PyObject *__pyx_n_s_tsByLabel;
static PyObject *__pyx_n_s_tslabels;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_kp_u_upper_left;
static PyObject *__pyx_kp_u_upper_right;
static PyObject *__pyx_kp_u_utf_8;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_vNeg;
static PyObject *__pyx_n_s_vPairs;
static PyObject *__pyx_n_s_vPos;
static PyObject *__pyx_n_s_validation;
static PyObject *__pyx_n_s_validationSet;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_vneg;
static PyObject *__pyx_n_s_vp;
static PyObject *__pyx_n_s_vpos;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_s_wPrecision;
static PyObject *__pyx_n_s_wScore;
static PyObject *__pyx_n_s_wThreshold;
static PyObject *__pyx_n_s_wantedPrecision;
static PyObject *__pyx_n_s_weight;
static PyObject *__pyx_n_s_weights;
static PyObject *__pyx_n_s_win;
static PyObject *__pyx_n_s_winPos;
static PyObject *__pyx_n_s_winPosSet;
static PyObject *__pyx_n_s_winScoreSet;
static PyObject *__pyx_n_s_winSeq;
static PyObject *__pyx_n_s_winSeqSet;
static PyObject *__pyx_n_s_windowSize;
static PyObject *__pyx_n_s_windowStep;
static PyObject *__pyx_n_s_windows;
static PyObject *__pyx_n_s_withLabel;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_xlabel;
static PyObject *__pyx_n_s_xs;
static PyObject *__pyx_n_s_xticks;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_ylabel;
static PyObject *__pyx_n_s_yticks;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_6gnocis_6models_setNThreadFetch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_25multiprocessModelDeployer_7genexpr_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_2multiprocessModelDeployer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_4initializePoolProcess(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_arg); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_6setPoolProcessID(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_newID); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_8getPoolProcessID(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_arg); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_10getPoolProcessModelID(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_arg); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_12initializePoolProcessIDs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pool); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_17getPoolProcessIDs_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_14getPoolProcessIDs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pool); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15getPoolModelIDs_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15getPoolModelIDs_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_16getPoolModelIDs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pool); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18setPoolProcessModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mdl); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_20getPoolProcessHasModel(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_arg); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_12setPoolModel_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_12setPoolModel_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_22setPoolModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pool, PyObject *__pyx_v_mdl); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_24setNCores(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_n); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_enableMultiprocessing); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_2__getstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_4__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_6getSequenceScores(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_34__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda6(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_8getOptimalAccuracyThreshold(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda7(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_10getPrecisionThreshold(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_positives, PyObject *__pyx_v_negatives, PyObject *__pyx_v_wantedPrecision); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_12calibrateGenomewidePrecision(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_positives, PyObject *__pyx_v_genome, PyObject *__pyx_v_factor, PyObject *__pyx_v_precision, PyObject *__pyx_v_bgModelOrder, PyObject *__pyx_v_bgModel); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_13scoreSequence_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_14scoreSequence(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seq); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_36__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_16getConfusionMatrix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_38__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_18getROC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_40__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_20getROCAUC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_42__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_22getPRC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_44__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_24getPRCAUC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_46__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_26plotPRC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_style); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_48__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_28getValidationStatistics(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_30predictSequenceStreamRegions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_stream); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_27predictSequenceStreamCurves_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_32predictSequenceStreamCurves(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_stream); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_34predictCore(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_genome); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_36predict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_stream, PyObject *__pyx_v_curves); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_38score(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_target, PyObject *__pyx_v_asTable); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_40train(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ts); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_50__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_42printTestStatistics(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_features, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy_4__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda12(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_ts); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy_6getTrainer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy_8scoreWindow(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seq); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_52__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8__init___3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_features, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_4__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda15(PyObject *__pyx_self, PyObject *__pyx_v_ts); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_6getTrainer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8scoreWindow(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seq); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_26trainSinglePREdictorModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_motifs, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_28createDummyPREdictorModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_motifs, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_30trainPREdictorModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_motifs, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_54__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_model, PyObject *__pyx_v_trainedModels, PyObject *__pyx_v_labelPredictionCurves); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_2regions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_label); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_4setCorePredictions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_label, PyObject *__pyx_v_cores); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_6predictCore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_genome); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_8getStatTable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_10__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_12_repr_html_(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_56__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_models, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_validationSet, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative, PyObject *__pyx_v_repeats, PyObject *__pyx_v_ratioTrainTest, PyObject *__pyx_v_ratioNegPos, PyObject *__pyx_v_storeTrained); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_2addModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_mdl); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_58__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_4calibrate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_calibrator, PyObject *__pyx_v_labelPositive); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_6predict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_genome); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_7plotPRC_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_8plotPRC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_style, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_7plotROC_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_10plotROC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_style, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_12getConfigurationTable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_14getAUCTable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_16__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_18_repr_html_(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gnocis_6models_32crossvalidate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_models, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_validationSet, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative, PyObject *__pyx_v_repeats, PyObject *__pyx_v_ratioTrainTest, PyObject *__pyx_v_ratioNegPos, PyObject *__pyx_v_storeTrained); /* proto */
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_5_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_6_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_7_setPoolModel(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_8_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_9_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_11_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_12_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_13_scoreSequence(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_14_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_16_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_17_getTrainer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_18___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_19_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_20_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_21_getTrainer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_22___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_23_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_24_plotPRC(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_25_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_26_plotROC(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_27_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float__3;
static PyObject *__pyx_float_0_;
static PyObject *__pyx_float_1_;
static PyObject *__pyx_float_6_;
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_float_0_8;
static PyObject *__pyx_float_1_0;
static PyObject *__pyx_float_100_;
static PyObject *__pyx_float_neg_0_15;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_14;
static PyObject *__pyx_int_18;
static PyObject *__pyx_int_20;
static PyObject *__pyx_int_219;
static PyObject *__pyx_int_500;
static PyObject *__pyx_int_1000;
static PyObject *__pyx_int_100000;
static PyObject *__pyx_k__12;
static PyObject *__pyx_k__13;
static PyObject *__pyx_k__14;
static PyObject *__pyx_k__15;
static PyObject *__pyx_k__31;
static PyObject *__pyx_k__32;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__18;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__94;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_tuple__100;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__118;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__144;
static PyObject *__pyx_tuple__146;
static PyObject *__pyx_tuple__148;
static PyObject *__pyx_tuple__150;
static PyObject *__pyx_tuple__152;
static PyObject *__pyx_tuple__154;
static PyObject *__pyx_tuple__156;
static PyObject *__pyx_tuple__158;
static PyObject *__pyx_tuple__159;
static PyObject *__pyx_tuple__161;
static PyObject *__pyx_tuple__162;
static PyObject *__pyx_tuple__164;
static PyObject *__pyx_tuple__166;
static PyObject *__pyx_tuple__168;
static PyObject *__pyx_tuple__170;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__59;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__72;
static PyObject *__pyx_codeobj__74;
static PyObject *__pyx_codeobj__77;
static PyObject *__pyx_codeobj__79;
static PyObject *__pyx_codeobj__81;
static PyObject *__pyx_codeobj__83;
static PyObject *__pyx_codeobj__85;
static PyObject *__pyx_codeobj__87;
static PyObject *__pyx_codeobj__89;
static PyObject *__pyx_codeobj__91;
static PyObject *__pyx_codeobj__93;
static PyObject *__pyx_codeobj__95;
static PyObject *__pyx_codeobj__97;
static PyObject *__pyx_codeobj__99;
static PyObject *__pyx_codeobj__102;
static PyObject *__pyx_codeobj__105;
static PyObject *__pyx_codeobj__107;
static PyObject *__pyx_codeobj__109;
static PyObject *__pyx_codeobj__111;
static PyObject *__pyx_codeobj__113;
static PyObject *__pyx_codeobj__115;
static PyObject *__pyx_codeobj__117;
static PyObject *__pyx_codeobj__119;
static PyObject *__pyx_codeobj__121;
static PyObject *__pyx_codeobj__123;
static PyObject *__pyx_codeobj__125;
static PyObject *__pyx_codeobj__127;
static PyObject *__pyx_codeobj__129;
static PyObject *__pyx_codeobj__131;
static PyObject *__pyx_codeobj__133;
static PyObject *__pyx_codeobj__135;
static PyObject *__pyx_codeobj__137;
static PyObject *__pyx_codeobj__139;
static PyObject *__pyx_codeobj__141;
static PyObject *__pyx_codeobj__143;
static PyObject *__pyx_codeobj__145;
static PyObject *__pyx_codeobj__147;
static PyObject *__pyx_codeobj__149;
static PyObject *__pyx_codeobj__151;
static PyObject *__pyx_codeobj__153;
static PyObject *__pyx_codeobj__155;
static PyObject *__pyx_codeobj__157;
static PyObject *__pyx_codeobj__160;
static PyObject *__pyx_codeobj__163;
static PyObject *__pyx_codeobj__165;
static PyObject *__pyx_codeobj__167;
static PyObject *__pyx_codeobj__169;
static PyObject *__pyx_codeobj__171;
/* Late includes */

/* "gnocis/models.pyx":33
 * maxThreadFetchNT = nTreadFetch * 1000
 * 
 * def setNThreadFetch(v):             # <<<<<<<<<<<<<<
 * 	global nTreadFetch
 * 	global maxThreadFetchNT
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_1setNThreadFetch(PyObject *__pyx_self, PyObject *__pyx_v_v); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_1setNThreadFetch = {"setNThreadFetch", (PyCFunction)__pyx_pw_6gnocis_6models_1setNThreadFetch, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_1setNThreadFetch(PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setNThreadFetch (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_setNThreadFetch(__pyx_self, ((PyObject *)__pyx_v_v));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_setNThreadFetch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setNThreadFetch", 0);

  /* "gnocis/models.pyx":36
 * 	global nTreadFetch
 * 	global maxThreadFetchNT
 * 	nTreadFetch = v             # <<<<<<<<<<<<<<
 * 	maxThreadFetchNT = nTreadFetch * 1000
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nTreadFetch, __pyx_v_v) < 0) __PYX_ERR(0, 36, __pyx_L1_error)

  /* "gnocis/models.pyx":37
 * 	global maxThreadFetchNT
 * 	nTreadFetch = v
 * 	maxThreadFetchNT = nTreadFetch * 1000             # <<<<<<<<<<<<<<
 * 
 * # Used for applying models in separate processes
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_nTreadFetch); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_int_1000); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_maxThreadFetchNT, __pyx_t_2) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":33
 * maxThreadFetchNT = nTreadFetch * 1000
 * 
 * def setNThreadFetch(v):             # <<<<<<<<<<<<<<
 * 	global nTreadFetch
 * 	global maxThreadFetchNT
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.setNThreadFetch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":40
 * 
 * # Used for applying models in separate processes
 * def multiprocessModelDeployer(args):             # <<<<<<<<<<<<<<
 * 	cdef sequences seqs
 * 	cdef sequence seq, win
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_3multiprocessModelDeployer(PyObject *__pyx_self, PyObject *__pyx_v_args); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_3multiprocessModelDeployer = {"multiprocessModelDeployer", (PyCFunction)__pyx_pw_6gnocis_6models_3multiprocessModelDeployer, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_3multiprocessModelDeployer(PyObject *__pyx_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("multiprocessModelDeployer (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_2multiprocessModelDeployer(__pyx_self, ((PyObject *)__pyx_v_args));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_25multiprocessModelDeployer_7genexpr_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":51
 * 		max(
 * 			(
 * 				model.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(windowSize, windowStep)
 */

static PyObject *__pyx_pf_6gnocis_6models_25multiprocessModelDeployer_7genexpr_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_1_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 51, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_25multiprocessModelDeployer_7genexpr_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_multiprocessModelDeployer_locals, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 51, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.multiprocessModelDeployer.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_25multiprocessModelDeployer_7genexpr_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 51, __pyx_L1_error)

  /* "gnocis/models.pyx":53
 * 				model.scoreWindow(win)
 * 				for win in
 * 				seq.windows(windowSize, windowStep)             # <<<<<<<<<<<<<<
 * 				if len(win) == windowSize
 * 			),
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_7genexpr__pyx_v_seq)) { __Pyx_RaiseClosureNameError("seq"); __PYX_ERR(0, 53, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_7genexpr__pyx_v_seq), __pyx_n_s_windows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_windowSize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_windowStep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 53, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 53, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 53, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 53, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_win);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_win, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":54
 * 				for win in
 * 				seq.windows(windowSize, windowStep)
 * 				if len(win) == windowSize             # <<<<<<<<<<<<<<
 * 			),
 * 			default = defaultValue
 */
    __pyx_t_10 = PyObject_Length(__pyx_cur_scope->__pyx_v_win); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 54, __pyx_L1_error)
    __pyx_t_11 = ((__pyx_t_10 == __pyx_cur_scope->__pyx_outer_scope->__pyx_v_windowSize) != 0);
    if (__pyx_t_11) {

      /* "gnocis/models.pyx":51
 * 		max(
 * 			(
 * 				model.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(windowSize, windowStep)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_model); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 51, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_scoreWindow); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_cur_scope->__pyx_v_win) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_cur_scope->__pyx_v_win);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      __Pyx_XGIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_8;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_9;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_9 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 51, __pyx_L1_error)

      /* "gnocis/models.pyx":54
 * 				for win in
 * 				seq.windows(windowSize, windowStep)
 * 				if len(win) == windowSize             # <<<<<<<<<<<<<<
 * 			),
 * 			default = defaultValue
 */
    }

    /* "gnocis/models.pyx":52
 * 			(
 * 				model.scoreWindow(win)
 * 				for win in             # <<<<<<<<<<<<<<
 * 				seq.windows(windowSize, windowStep)
 * 				if len(win) == windowSize
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "gnocis/models.pyx":51
 * 		max(
 * 			(
 * 				model.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(windowSize, windowStep)
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":40
 * 
 * # Used for applying models in separate processes
 * def multiprocessModelDeployer(args):             # <<<<<<<<<<<<<<
 * 	cdef sequences seqs
 * 	cdef sequence seq, win
 */

static PyObject *__pyx_pf_6gnocis_6models_2multiprocessModelDeployer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *__pyx_cur_scope;
  struct __pyx_obj_6gnocis_9sequences_sequences *__pyx_v_seqs = 0;
  PyObject *__pyx_v_defaultValue = 0;
  PyObject *__pyx_7genexpr__pyx_v_0 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("multiprocessModelDeployer", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer(__pyx_ptype_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 40, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/models.pyx":43
 * 	cdef sequences seqs
 * 	cdef sequence seq, win
 * 	cdef defaultValue = float('-INF')             # <<<<<<<<<<<<<<
 * 	global model
 * 	seqs = args
 */
  __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_kp_u_INF); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_defaultValue = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":45
 * 	cdef defaultValue = float('-INF')
 * 	global model
 * 	seqs = args             # <<<<<<<<<<<<<<
 * 	cdef int windowSize = model.windowSize
 * 	cdef int windowStep = model.windowStep
 */
  if (!(likely(((__pyx_v_args) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_args, __pyx_ptype_6gnocis_9sequences_sequences))))) __PYX_ERR(0, 45, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_args;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_seqs = ((struct __pyx_obj_6gnocis_9sequences_sequences *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":46
 * 	global model
 * 	seqs = args
 * 	cdef int windowSize = model.windowSize             # <<<<<<<<<<<<<<
 * 	cdef int windowStep = model.windowStep
 * 	return [
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_model); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_windowSize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_cur_scope->__pyx_v_windowSize = __pyx_t_3;

  /* "gnocis/models.pyx":47
 * 	seqs = args
 * 	cdef int windowSize = model.windowSize
 * 	cdef int windowStep = model.windowStep             # <<<<<<<<<<<<<<
 * 	return [
 * 		max(
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_model); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_windowStep); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_cur_scope->__pyx_v_windowStep = __pyx_t_3;

  /* "gnocis/models.pyx":48
 * 	cdef int windowSize = model.windowSize
 * 	cdef int windowStep = model.windowStep
 * 	return [             # <<<<<<<<<<<<<<
 * 		max(
 * 			(
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "gnocis/models.pyx":58
 * 			default = defaultValue
 * 		)
 * 		for seq in seqs             # <<<<<<<<<<<<<<
 * 	]
 * 
 */
    if (likely(PyList_CheckExact(((PyObject *)__pyx_v_seqs))) || PyTuple_CheckExact(((PyObject *)__pyx_v_seqs))) {
      __pyx_t_2 = ((PyObject *)__pyx_v_seqs); __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(((PyObject *)__pyx_v_seqs)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 58, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 58, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 58, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 58, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 58, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_5(__pyx_t_2);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 58, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_6gnocis_9sequences_sequence))))) __PYX_ERR(0, 58, __pyx_L1_error)
      __Pyx_XGOTREF(((PyObject *)__pyx_cur_scope->__pyx_7genexpr__pyx_v_seq));
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_7genexpr__pyx_v_seq, ((struct __pyx_obj_6gnocis_9sequences_sequence *)__pyx_t_6));
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_t_6 = 0;

      /* "gnocis/models.pyx":51
 * 		max(
 * 			(
 * 				model.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(windowSize, windowStep)
 */
      __pyx_t_6 = __pyx_pf_6gnocis_6models_25multiprocessModelDeployer_7genexpr_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 51, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "gnocis/models.pyx":49
 * 	cdef int windowStep = model.windowStep
 * 	return [
 * 		max(             # <<<<<<<<<<<<<<
 * 			(
 * 				model.scoreWindow(win)
 */
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "gnocis/models.pyx":56
 * 				if len(win) == windowSize
 * 			),
 * 			default = defaultValue             # <<<<<<<<<<<<<<
 * 		)
 * 		for seq in seqs
 */
      __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 56, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_default, __pyx_v_defaultValue) < 0) __PYX_ERR(0, 56, __pyx_L1_error)

      /* "gnocis/models.pyx":49
 * 	cdef int windowStep = model.windowStep
 * 	return [
 * 		max(             # <<<<<<<<<<<<<<
 * 			(
 * 				model.scoreWindow(win)
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 48, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "gnocis/models.pyx":58
 * 			default = defaultValue
 * 		)
 * 		for seq in seqs             # <<<<<<<<<<<<<<
 * 	]
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":40
 * 
 * # Used for applying models in separate processes
 * def multiprocessModelDeployer(args):             # <<<<<<<<<<<<<<
 * 	cdef sequences seqs
 * 	cdef sequence seq, win
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("gnocis.models.multiprocessModelDeployer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_seqs);
  __Pyx_XDECREF(__pyx_v_defaultValue);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_0);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":61
 * 	]
 * 
 * def initializePoolProcess(arg):             # <<<<<<<<<<<<<<
 * 	global processID, model
 * 	processID, model = None, None
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_5initializePoolProcess(PyObject *__pyx_self, PyObject *__pyx_v_arg); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_5initializePoolProcess = {"initializePoolProcess", (PyCFunction)__pyx_pw_6gnocis_6models_5initializePoolProcess, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_5initializePoolProcess(PyObject *__pyx_self, PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("initializePoolProcess (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_4initializePoolProcess(__pyx_self, ((PyObject *)__pyx_v_arg));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_4initializePoolProcess(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("initializePoolProcess", 0);

  /* "gnocis/models.pyx":63
 * def initializePoolProcess(arg):
 * 	global processID, model
 * 	processID, model = None, None             # <<<<<<<<<<<<<<
 * 
 * def setPoolProcessID(newID):
 */
  __pyx_t_1 = Py_None;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = Py_None;
  __Pyx_INCREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_processID, __pyx_t_1) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_model, __pyx_t_2) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":61
 * 	]
 * 
 * def initializePoolProcess(arg):             # <<<<<<<<<<<<<<
 * 	global processID, model
 * 	processID, model = None, None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.initializePoolProcess", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":65
 * 	processID, model = None, None
 * 
 * def setPoolProcessID(newID):             # <<<<<<<<<<<<<<
 * 	global processID
 * 	processID = newID
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_7setPoolProcessID(PyObject *__pyx_self, PyObject *__pyx_v_newID); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_7setPoolProcessID = {"setPoolProcessID", (PyCFunction)__pyx_pw_6gnocis_6models_7setPoolProcessID, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_7setPoolProcessID(PyObject *__pyx_self, PyObject *__pyx_v_newID) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setPoolProcessID (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_6setPoolProcessID(__pyx_self, ((PyObject *)__pyx_v_newID));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_6setPoolProcessID(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_newID) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setPoolProcessID", 0);

  /* "gnocis/models.pyx":67
 * def setPoolProcessID(newID):
 * 	global processID
 * 	processID = newID             # <<<<<<<<<<<<<<
 * 	return True
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_processID, __pyx_v_newID) < 0) __PYX_ERR(0, 67, __pyx_L1_error)

  /* "gnocis/models.pyx":68
 * 	global processID
 * 	processID = newID
 * 	return True             # <<<<<<<<<<<<<<
 * 
 * def getPoolProcessID(arg):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "gnocis/models.pyx":65
 * 	processID, model = None, None
 * 
 * def setPoolProcessID(newID):             # <<<<<<<<<<<<<<
 * 	global processID
 * 	processID = newID
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.setPoolProcessID", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":70
 * 	return True
 * 
 * def getPoolProcessID(arg):             # <<<<<<<<<<<<<<
 * 	global processID
 * 	return processID
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_9getPoolProcessID(PyObject *__pyx_self, PyObject *__pyx_v_arg); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_9getPoolProcessID = {"getPoolProcessID", (PyCFunction)__pyx_pw_6gnocis_6models_9getPoolProcessID, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_9getPoolProcessID(PyObject *__pyx_self, PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getPoolProcessID (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_8getPoolProcessID(__pyx_self, ((PyObject *)__pyx_v_arg));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_8getPoolProcessID(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getPoolProcessID", 0);

  /* "gnocis/models.pyx":72
 * def getPoolProcessID(arg):
 * 	global processID
 * 	return processID             # <<<<<<<<<<<<<<
 * 
 * def getPoolProcessModelID(arg):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_processID); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":70
 * 	return True
 * 
 * def getPoolProcessID(arg):             # <<<<<<<<<<<<<<
 * 	global processID
 * 	return processID
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.models.getPoolProcessID", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":74
 * 	return processID
 * 
 * def getPoolProcessModelID(arg):             # <<<<<<<<<<<<<<
 * 	global processID, model
 * 	if model == None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_11getPoolProcessModelID(PyObject *__pyx_self, PyObject *__pyx_v_arg); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_11getPoolProcessModelID = {"getPoolProcessModelID", (PyCFunction)__pyx_pw_6gnocis_6models_11getPoolProcessModelID, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_11getPoolProcessModelID(PyObject *__pyx_self, PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getPoolProcessModelID (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_10getPoolProcessModelID(__pyx_self, ((PyObject *)__pyx_v_arg));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_10getPoolProcessModelID(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getPoolProcessModelID", 0);

  /* "gnocis/models.pyx":76
 * def getPoolProcessModelID(arg):
 * 	global processID, model
 * 	if model == None:             # <<<<<<<<<<<<<<
 * 		return processID, None
 * 	return processID, '%d %s'%(model.modelID, str(model))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_model); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, Py_None, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "gnocis/models.pyx":77
 * 	global processID, model
 * 	if model == None:
 * 		return processID, None             # <<<<<<<<<<<<<<
 * 	return processID, '%d %s'%(model.modelID, str(model))
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_processID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 1, Py_None);
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gnocis/models.pyx":76
 * def getPoolProcessModelID(arg):
 * 	global processID, model
 * 	if model == None:             # <<<<<<<<<<<<<<
 * 		return processID, None
 * 	return processID, '%d %s'%(model.modelID, str(model))
 */
  }

  /* "gnocis/models.pyx":78
 * 	if model == None:
 * 		return processID, None
 * 	return processID, '%d %s'%(model.modelID, str(model))             # <<<<<<<<<<<<<<
 * 
 * def initializePoolProcessIDs(pool):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_processID); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_5 = 127;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_model); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_modelID); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_FormatAndDecref(__Pyx_PyNumber_IntOrLong(__pyx_t_7), __pyx_n_u_d); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_t_4 += 1;
  __Pyx_GIVEREF(__pyx_kp_u_);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_kp_u_);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_model); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_5;
  __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":74
 * 	return processID
 * 
 * def getPoolProcessModelID(arg):             # <<<<<<<<<<<<<<
 * 	global processID, model
 * 	if model == None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.models.getPoolProcessModelID", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":80
 * 	return processID, '%d %s'%(model.modelID, str(model))
 * 
 * def initializePoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	while True:
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13initializePoolProcessIDs(PyObject *__pyx_self, PyObject *__pyx_v_pool); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_13initializePoolProcessIDs = {"initializePoolProcessIDs", (PyCFunction)__pyx_pw_6gnocis_6models_13initializePoolProcessIDs, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_13initializePoolProcessIDs(PyObject *__pyx_self, PyObject *__pyx_v_pool) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("initializePoolProcessIDs (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_12initializePoolProcessIDs(__pyx_self, ((PyObject *)__pyx_v_pool));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_12initializePoolProcessIDs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pool) {
  PyObject *__pyx_v_procIDs = NULL;
  CYTHON_UNUSED Py_ssize_t __pyx_8genexpr2__pyx_v__;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("initializePoolProcessIDs", 0);

  /* "gnocis/models.pyx":81
 * 
 * def initializePoolProcessIDs(pool):
 * 	while True:             # <<<<<<<<<<<<<<
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )
 * 		procIDs = set( pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) )
 */
  while (1) {

    /* "gnocis/models.pyx":82
 * def initializePoolProcessIDs(pool):
 * 	while True:
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )             # <<<<<<<<<<<<<<
 * 		procIDs = set( pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) )
 * 		if len(procIDs) == len(pool._pool):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_map); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_setPoolProcessID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PySequence_List(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":83
 * 	while True:
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )
 * 		procIDs = set( pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) )             # <<<<<<<<<<<<<<
 * 		if len(procIDs) == len(pool._pool):
 * 			break
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_map); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_getPoolProcessID); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    { /* enter inner scope */
      __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = __pyx_t_5;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
        __pyx_8genexpr2__pyx_v__ = __pyx_t_10;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_empty_tuple))) __PYX_ERR(0, 83, __pyx_L1_error)
      }
    } /* exit inner scope */
    __pyx_t_3 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_8, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_8, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_t_4);
      __pyx_t_8 = 0;
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_procIDs, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":84
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )
 * 		procIDs = set( pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) )
 * 		if len(procIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 			break
 * 
 */
    __pyx_t_5 = PySet_GET_SIZE(__pyx_v_procIDs); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 84, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = ((__pyx_t_5 == __pyx_t_9) != 0);
    if (__pyx_t_11) {

      /* "gnocis/models.pyx":85
 * 		procIDs = set( pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) )
 * 		if len(procIDs) == len(pool._pool):
 * 			break             # <<<<<<<<<<<<<<
 * 
 * def getPoolProcessIDs(pool):
 */
      goto __pyx_L4_break;

      /* "gnocis/models.pyx":84
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )
 * 		procIDs = set( pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) )
 * 		if len(procIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 			break
 * 
 */
    }
  }
  __pyx_L4_break:;

  /* "gnocis/models.pyx":80
 * 	return processID, '%d %s'%(model.modelID, str(model))
 * 
 * def initializePoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	while True:
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("gnocis.models.initializePoolProcessIDs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_procIDs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":87
 * 			break
 * 
 * def getPoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	while True:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15getPoolProcessIDs(PyObject *__pyx_self, PyObject *__pyx_v_pool); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15getPoolProcessIDs = {"getPoolProcessIDs", (PyCFunction)__pyx_pw_6gnocis_6models_15getPoolProcessIDs, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_15getPoolProcessIDs(PyObject *__pyx_self, PyObject *__pyx_v_pool) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getPoolProcessIDs (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_14getPoolProcessIDs(__pyx_self, ((PyObject *)__pyx_v_pool));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_17getPoolProcessIDs_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":90
 * 	pIDs = set()
 * 	while True:
 * 		pIDs = pIDs | set( v for v in pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) if v != None )             # <<<<<<<<<<<<<<
 * 		if len(pIDs) == len(pool._pool):
 * 			return pIDs
 */

static PyObject *__pyx_pf_6gnocis_6models_17getPoolProcessIDs_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_3_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 90, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_17getPoolProcessIDs_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_getPoolProcessIDs_locals_genexpr, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.getPoolProcessIDs.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_17getPoolProcessIDs_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_pool)) { __Pyx_RaiseClosureNameError("pool"); __PYX_ERR(0, 90, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_pool, __pyx_n_s_map); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_getPoolProcessID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_pool)) { __Pyx_RaiseClosureNameError("pool"); __PYX_ERR(0, 90, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_cur_scope->__pyx_8genexpr4__pyx_v__ = __pyx_t_8;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_empty_tuple))) __PYX_ERR(0, 90, __pyx_L1_error)
    }
  } /* exit inner scope */
  __pyx_t_5 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_11 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 90, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_11)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 90, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 90, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_11(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 90, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_v, Py_None, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_12) {
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_v))) __PYX_ERR(0, 90, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":87
 * 			break
 * 
 * def getPoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	while True:
 */

static PyObject *__pyx_pf_6gnocis_6models_14getPoolProcessIDs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pool) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *__pyx_cur_scope;
  PyObject *__pyx_v_pIDs = NULL;
  PyObject *__pyx_gb_6gnocis_6models_17getPoolProcessIDs_2generator1 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getPoolProcessIDs", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs(__pyx_ptype_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 87, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_pool = __pyx_v_pool;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_pool);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_pool);

  /* "gnocis/models.pyx":88
 * 
 * def getPoolProcessIDs(pool):
 * 	pIDs = set()             # <<<<<<<<<<<<<<
 * 	while True:
 * 		pIDs = pIDs | set( v for v in pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) if v != None )
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pIDs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":89
 * def getPoolProcessIDs(pool):
 * 	pIDs = set()
 * 	while True:             # <<<<<<<<<<<<<<
 * 		pIDs = pIDs | set( v for v in pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) if v != None )
 * 		if len(pIDs) == len(pool._pool):
 */
  while (1) {

    /* "gnocis/models.pyx":90
 * 	pIDs = set()
 * 	while True:
 * 		pIDs = pIDs | set( v for v in pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) if v != None )             # <<<<<<<<<<<<<<
 * 		if len(pIDs) == len(pool._pool):
 * 			return pIDs
 */
    __pyx_t_1 = __pyx_pf_6gnocis_6models_17getPoolProcessIDs_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Or(__pyx_v_pIDs, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_pIDs, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":91
 * 	while True:
 * 		pIDs = pIDs | set( v for v in pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) if v != None )
 * 		if len(pIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 			return pIDs
 * 
 */
    __pyx_t_3 = PySet_GET_SIZE(__pyx_v_pIDs); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 91, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = ((__pyx_t_3 == __pyx_t_4) != 0);
    if (__pyx_t_5) {

      /* "gnocis/models.pyx":92
 * 		pIDs = pIDs | set( v for v in pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) if v != None )
 * 		if len(pIDs) == len(pool._pool):
 * 			return pIDs             # <<<<<<<<<<<<<<
 * 
 * def getPoolModelIDs(pool):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_pIDs);
      __pyx_r = __pyx_v_pIDs;
      goto __pyx_L0;

      /* "gnocis/models.pyx":91
 * 	while True:
 * 		pIDs = pIDs | set( v for v in pool.map( getPoolProcessID, [ () for _ in range(len(pool._pool)) ] ) if v != None )
 * 		if len(pIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 			return pIDs
 * 
 */
    }
  }

  /* "gnocis/models.pyx":87
 * 			break
 * 
 * def getPoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	while True:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.getPoolProcessIDs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pIDs);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_17getPoolProcessIDs_2generator1);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":94
 * 			return pIDs
 * 
 * def getPoolModelIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	mIDs = set()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_17getPoolModelIDs(PyObject *__pyx_self, PyObject *__pyx_v_pool); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_17getPoolModelIDs = {"getPoolModelIDs", (PyCFunction)__pyx_pw_6gnocis_6models_17getPoolModelIDs, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_17getPoolModelIDs(PyObject *__pyx_self, PyObject *__pyx_v_pool) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getPoolModelIDs (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_16getPoolModelIDs(__pyx_self, ((PyObject *)__pyx_v_pool));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_15getPoolModelIDs_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":99
 * 	while True:
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )             # <<<<<<<<<<<<<<
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )
 * 		if len(pIDs) == len(pool._pool):
 */

static PyObject *__pyx_pf_6gnocis_6models_15getPoolModelIDs_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_5_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_5_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 99, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_15getPoolModelIDs_2generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_getPoolModelIDs_locals_genexpr, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.getPoolModelIDs.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_15getPoolModelIDs_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 99, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) { __Pyx_RaiseClosureNameError("r"); __PYX_ERR(0, 99, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_r; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 99, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 99, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 99, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 99, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 99, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_pid);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_pid, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_mid);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_mid, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_4 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_pid, Py_None, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_9) {
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_pid))) __PYX_ERR(0, 99, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_15getPoolModelIDs_5generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":100
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )             # <<<<<<<<<<<<<<
 * 		if len(pIDs) == len(pool._pool):
 * 			return mIDs
 */

static PyObject *__pyx_pf_6gnocis_6models_15getPoolModelIDs_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_6_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_6_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 100, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_15getPoolModelIDs_5generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_getPoolModelIDs_locals_genexpr, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.getPoolModelIDs.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_15getPoolModelIDs_5generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 100, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) { __Pyx_RaiseClosureNameError("r"); __PYX_ERR(0, 100, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_r; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 100, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 100, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 100, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 100, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 100, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_pid);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_pid, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_mid);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_mid, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_4 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_pid, Py_None, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_9) {
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_mid))) __PYX_ERR(0, 100, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":94
 * 			return pIDs
 * 
 * def getPoolModelIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	mIDs = set()
 */

static PyObject *__pyx_pf_6gnocis_6models_16getPoolModelIDs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pool) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *__pyx_cur_scope;
  PyObject *__pyx_v_pIDs = NULL;
  PyObject *__pyx_v_mIDs = NULL;
  CYTHON_UNUSED Py_ssize_t __pyx_8genexpr5__pyx_v__;
  PyObject *__pyx_gb_6gnocis_6models_15getPoolModelIDs_2generator2 = 0;
  PyObject *__pyx_gb_6gnocis_6models_15getPoolModelIDs_5generator3 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getPoolModelIDs", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs(__pyx_ptype_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 94, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/models.pyx":95
 * 
 * def getPoolModelIDs(pool):
 * 	pIDs = set()             # <<<<<<<<<<<<<<
 * 	mIDs = set()
 * 	while True:
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pIDs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":96
 * def getPoolModelIDs(pool):
 * 	pIDs = set()
 * 	mIDs = set()             # <<<<<<<<<<<<<<
 * 	while True:
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_mIDs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":97
 * 	pIDs = set()
 * 	mIDs = set()
 * 	while True:             # <<<<<<<<<<<<<<
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )
 */
  while (1) {

    /* "gnocis/models.pyx":98
 * 	mIDs = set()
 * 	while True:
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )             # <<<<<<<<<<<<<<
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_map); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_getPoolProcessModelID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { /* enter inner scope */
      __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_8genexpr5__pyx_v__ = __pyx_t_8;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_empty_tuple))) __PYX_ERR(0, 98, __pyx_L1_error)
      }
    } /* exit inner scope */
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":99
 * 	while True:
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )             # <<<<<<<<<<<<<<
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )
 * 		if len(pIDs) == len(pool._pool):
 */
    __pyx_t_1 = __pyx_pf_6gnocis_6models_15getPoolModelIDs_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Or(__pyx_v_pIDs, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_pIDs, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":100
 * 		r = pool.map( getPoolProcessModelID, [ () for _ in range(len(pool._pool)) ] )
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )             # <<<<<<<<<<<<<<
 * 		if len(pIDs) == len(pool._pool):
 * 			return mIDs
 */
    __pyx_t_1 = __pyx_pf_6gnocis_6models_15getPoolModelIDs_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Or(__pyx_v_mIDs, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_mIDs, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":101
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )
 * 		if len(pIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 			return mIDs
 * 
 */
    __pyx_t_6 = PySet_GET_SIZE(__pyx_v_pIDs); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 101, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_11 = ((__pyx_t_6 == __pyx_t_7) != 0);
    if (__pyx_t_11) {

      /* "gnocis/models.pyx":102
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )
 * 		if len(pIDs) == len(pool._pool):
 * 			return mIDs             # <<<<<<<<<<<<<<
 * 
 * def setPoolProcessModel(mdl):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_mIDs);
      __pyx_r = __pyx_v_mIDs;
      goto __pyx_L0;

      /* "gnocis/models.pyx":101
 * 		pIDs = pIDs | set( pid for pid, mid in r if pid != None )
 * 		mIDs = mIDs | set( mid for pid, mid in r if pid != None )
 * 		if len(pIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 			return mIDs
 * 
 */
    }
  }

  /* "gnocis/models.pyx":94
 * 			return pIDs
 * 
 * def getPoolModelIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	mIDs = set()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("gnocis.models.getPoolModelIDs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pIDs);
  __Pyx_XDECREF(__pyx_v_mIDs);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_15getPoolModelIDs_2generator2);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_15getPoolModelIDs_5generator3);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":104
 * 			return mIDs
 * 
 * def setPoolProcessModel(mdl):             # <<<<<<<<<<<<<<
 * 	global model
 * 	model = mdl
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_19setPoolProcessModel(PyObject *__pyx_self, PyObject *__pyx_v_mdl); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_19setPoolProcessModel = {"setPoolProcessModel", (PyCFunction)__pyx_pw_6gnocis_6models_19setPoolProcessModel, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_19setPoolProcessModel(PyObject *__pyx_self, PyObject *__pyx_v_mdl) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setPoolProcessModel (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_18setPoolProcessModel(__pyx_self, ((PyObject *)__pyx_v_mdl));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18setPoolProcessModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mdl) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setPoolProcessModel", 0);

  /* "gnocis/models.pyx":106
 * def setPoolProcessModel(mdl):
 * 	global model
 * 	model = mdl             # <<<<<<<<<<<<<<
 * 	return True
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_model, __pyx_v_mdl) < 0) __PYX_ERR(0, 106, __pyx_L1_error)

  /* "gnocis/models.pyx":107
 * 	global model
 * 	model = mdl
 * 	return True             # <<<<<<<<<<<<<<
 * 
 * def getPoolProcessHasModel(arg):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "gnocis/models.pyx":104
 * 			return mIDs
 * 
 * def setPoolProcessModel(mdl):             # <<<<<<<<<<<<<<
 * 	global model
 * 	model = mdl
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.setPoolProcessModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":109
 * 	return True
 * 
 * def getPoolProcessHasModel(arg):             # <<<<<<<<<<<<<<
 * 	global model, processID
 * 	return (processID, model != None)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_21getPoolProcessHasModel(PyObject *__pyx_self, PyObject *__pyx_v_arg); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_21getPoolProcessHasModel = {"getPoolProcessHasModel", (PyCFunction)__pyx_pw_6gnocis_6models_21getPoolProcessHasModel, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_21getPoolProcessHasModel(PyObject *__pyx_self, PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getPoolProcessHasModel (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_20getPoolProcessHasModel(__pyx_self, ((PyObject *)__pyx_v_arg));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_20getPoolProcessHasModel(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getPoolProcessHasModel", 0);

  /* "gnocis/models.pyx":111
 * def getPoolProcessHasModel(arg):
 * 	global model, processID
 * 	return (processID, model != None)             # <<<<<<<<<<<<<<
 * 
 * def setPoolModel(pool, mdl):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_processID); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_model); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, Py_None, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":109
 * 	return True
 * 
 * def getPoolProcessHasModel(arg):             # <<<<<<<<<<<<<<
 * 	global model, processID
 * 	return (processID, model != None)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.models.getPoolProcessHasModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":113
 * 	return (processID, model != None)
 * 
 * def setPoolModel(pool, mdl):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	if mdl == None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_23setPoolModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_23setPoolModel = {"setPoolModel", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_23setPoolModel, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_23setPoolModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pool = 0;
  PyObject *__pyx_v_mdl = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setPoolModel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pool_2,&__pyx_n_s_mdl,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pool_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mdl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setPoolModel", 1, 2, 2, 1); __PYX_ERR(0, 113, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setPoolModel") < 0)) __PYX_ERR(0, 113, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_pool = values[0];
    __pyx_v_mdl = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setPoolModel", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 113, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.setPoolModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_22setPoolModel(__pyx_self, __pyx_v_pool, __pyx_v_mdl);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_12setPoolModel_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":122
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			# We append the process-IDs that we confirmed are done
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if not hasModel )             # <<<<<<<<<<<<<<
 * 			if len(pIDs) == len(pool._pool):
 * 				break
 */

static PyObject *__pyx_pf_6gnocis_6models_12setPoolModel_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_8_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_8_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 122, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_12setPoolModel_2generator4, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_setPoolModel_locals_genexpr, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.setPoolModel.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_12setPoolModel_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) { __Pyx_RaiseClosureNameError("r"); __PYX_ERR(0, 122, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_r; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 122, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 122, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 122, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 122, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 122, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_pid);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_pid, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_hasModel);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_hasModel, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_hasModel); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 122, __pyx_L1_error)
    __pyx_t_10 = ((!__pyx_t_9) != 0);
    if (__pyx_t_10) {
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_pid))) __PYX_ERR(0, 122, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_12setPoolModel_5generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":133
 * 			pool.map( setPoolProcessModel, [ mdl for _ in range(len(pool._pool)) ] )
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if hasModel )             # <<<<<<<<<<<<<<
 * 			if len(pIDs) == len(pool._pool):
 * 				break
 */

static PyObject *__pyx_pf_6gnocis_6models_12setPoolModel_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_9_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_9_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 133, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_12setPoolModel_5generator5, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_setPoolModel_locals_genexpr, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.setPoolModel.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_12setPoolModel_5generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) { __Pyx_RaiseClosureNameError("r"); __PYX_ERR(0, 133, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_r; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 133, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 133, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 133, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_pid);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_pid, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_hasModel);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_hasModel, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_hasModel); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
    if (__pyx_t_9) {
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_pid))) __PYX_ERR(0, 133, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":113
 * 	return (processID, model != None)
 * 
 * def setPoolModel(pool, mdl):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	if mdl == None:
 */

static PyObject *__pyx_pf_6gnocis_6models_22setPoolModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_pool, PyObject *__pyx_v_mdl) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *__pyx_cur_scope;
  PyObject *__pyx_v_pIDs = NULL;
  CYTHON_UNUSED Py_ssize_t __pyx_8genexpr8__pyx_v__;
  CYTHON_UNUSED Py_ssize_t __pyx_8genexpr9__pyx_v__;
  PyObject *__pyx_gb_6gnocis_6models_12setPoolModel_2generator4 = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_9genexpr11__pyx_v__;
  CYTHON_UNUSED Py_ssize_t __pyx_9genexpr12__pyx_v__;
  PyObject *__pyx_gb_6gnocis_6models_12setPoolModel_5generator5 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setPoolModel", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_7_setPoolModel(__pyx_ptype_6gnocis_6models___pyx_scope_struct_7_setPoolModel, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 113, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/models.pyx":114
 * 
 * def setPoolModel(pool, mdl):
 * 	pIDs = set()             # <<<<<<<<<<<<<<
 * 	if mdl == None:
 * 		# Purge models from all processes
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pIDs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":115
 * def setPoolModel(pool, mdl):
 * 	pIDs = set()
 * 	if mdl == None:             # <<<<<<<<<<<<<<
 * 		# Purge models from all processes
 * 		# The process-ID is used to ensure that all processes have updated data
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_mdl, Py_None, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "gnocis/models.pyx":118
 * 		# Purge models from all processes
 * 		# The process-ID is used to ensure that all processes have updated data
 * 		while True:             # <<<<<<<<<<<<<<
 * 			pool.map( setPoolProcessModel, [ None for _ in range(len(pool._pool)) ] )
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 */
    while (1) {

      /* "gnocis/models.pyx":119
 * 		# The process-ID is used to ensure that all processes have updated data
 * 		while True:
 * 			pool.map( setPoolProcessModel, [ None for _ in range(len(pool._pool)) ] )             # <<<<<<<<<<<<<<
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			# We append the process-IDs that we confirmed are done
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_map); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_setPoolProcessModel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      { /* enter inner scope */
        __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_8 = __pyx_t_7;
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
          __pyx_8genexpr8__pyx_v__ = __pyx_t_9;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)Py_None))) __PYX_ERR(0, 119, __pyx_L1_error)
        }
      } /* exit inner scope */
      __pyx_t_6 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_4, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_4, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_5);
        __pyx_t_4 = 0;
        __pyx_t_5 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "gnocis/models.pyx":120
 * 		while True:
 * 			pool.map( setPoolProcessModel, [ None for _ in range(len(pool._pool)) ] )
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )             # <<<<<<<<<<<<<<
 * 			# We append the process-IDs that we confirmed are done
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if not hasModel )
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_map); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_getPoolProcessHasModel); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      { /* enter inner scope */
        __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = __pyx_t_7;
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
          __pyx_8genexpr9__pyx_v__ = __pyx_t_9;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_empty_tuple))) __PYX_ERR(0, 120, __pyx_L1_error)
        }
      } /* exit inner scope */
      __pyx_t_4 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_11, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_11, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_11);
        PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_10, __pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_10, __pyx_t_5);
        __pyx_t_11 = 0;
        __pyx_t_5 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "gnocis/models.pyx":122
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			# We append the process-IDs that we confirmed are done
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if not hasModel )             # <<<<<<<<<<<<<<
 * 			if len(pIDs) == len(pool._pool):
 * 				break
 */
      __pyx_t_1 = __pyx_pf_6gnocis_6models_12setPoolModel_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Or(__pyx_v_pIDs, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_pIDs, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "gnocis/models.pyx":123
 * 			# We append the process-IDs that we confirmed are done
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if not hasModel )
 * 			if len(pIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 				break
 * 	if mdl != None:
 */
      __pyx_t_7 = PySet_GET_SIZE(__pyx_v_pIDs); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 123, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 123, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = ((__pyx_t_7 == __pyx_t_8) != 0);
      if (__pyx_t_2) {

        /* "gnocis/models.pyx":124
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if not hasModel )
 * 			if len(pIDs) == len(pool._pool):
 * 				break             # <<<<<<<<<<<<<<
 * 	if mdl != None:
 * 		# For non-None model, first purge the original model
 */
        goto __pyx_L5_break;

        /* "gnocis/models.pyx":123
 * 			# We append the process-IDs that we confirmed are done
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if not hasModel )
 * 			if len(pIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 				break
 * 	if mdl != None:
 */
      }
    }
    __pyx_L5_break:;

    /* "gnocis/models.pyx":115
 * def setPoolModel(pool, mdl):
 * 	pIDs = set()
 * 	if mdl == None:             # <<<<<<<<<<<<<<
 * 		# Purge models from all processes
 * 		# The process-ID is used to ensure that all processes have updated data
 */
  }

  /* "gnocis/models.pyx":125
 * 			if len(pIDs) == len(pool._pool):
 * 				break
 * 	if mdl != None:             # <<<<<<<<<<<<<<
 * 		# For non-None model, first purge the original model
 * 		setPoolModel(pool, None)
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_mdl, Py_None, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "gnocis/models.pyx":127
 * 	if mdl != None:
 * 		# For non-None model, first purge the original model
 * 		setPoolModel(pool, None)             # <<<<<<<<<<<<<<
 * 		# Then set the model for all processes
 * 		# The process-ID is used to ensure that all processes have updated data
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_setPoolModel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_pool, Py_None};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_pool, Py_None};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_pool);
      __Pyx_GIVEREF(__pyx_v_pool);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_10, __pyx_v_pool);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_10, Py_None);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":130
 * 		# Then set the model for all processes
 * 		# The process-ID is used to ensure that all processes have updated data
 * 		while True:             # <<<<<<<<<<<<<<
 * 			pool.map( setPoolProcessModel, [ mdl for _ in range(len(pool._pool)) ] )
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 */
    while (1) {

      /* "gnocis/models.pyx":131
 * 		# The process-ID is used to ensure that all processes have updated data
 * 		while True:
 * 			pool.map( setPoolProcessModel, [ mdl for _ in range(len(pool._pool)) ] )             # <<<<<<<<<<<<<<
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if hasModel )
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_map); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_setPoolProcessModel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      { /* enter inner scope */
        __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_8 = PyObject_Length(__pyx_t_11); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_7 = __pyx_t_8;
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_7; __pyx_t_9+=1) {
          __pyx_9genexpr11__pyx_v__ = __pyx_t_9;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_v_mdl))) __PYX_ERR(0, 131, __pyx_L1_error)
        }
      } /* exit inner scope */
      __pyx_t_11 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_5, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_5, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__pyx_t_11) {
          __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_11); __pyx_t_11 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_10, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_10, __pyx_t_6);
        __pyx_t_5 = 0;
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "gnocis/models.pyx":132
 * 		while True:
 * 			pool.map( setPoolProcessModel, [ mdl for _ in range(len(pool._pool)) ] )
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )             # <<<<<<<<<<<<<<
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if hasModel )
 * 			if len(pIDs) == len(pool._pool):
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_map); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_getPoolProcessHasModel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      { /* enter inner scope */
        __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_7 = __pyx_t_8;
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_7; __pyx_t_9+=1) {
          __pyx_9genexpr12__pyx_v__ = __pyx_t_9;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_empty_tuple))) __PYX_ERR(0, 132, __pyx_L1_error)
        }
      } /* exit inner scope */
      __pyx_t_5 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_6);
        __pyx_t_4 = 0;
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "gnocis/models.pyx":133
 * 			pool.map( setPoolProcessModel, [ mdl for _ in range(len(pool._pool)) ] )
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if hasModel )             # <<<<<<<<<<<<<<
 * 			if len(pIDs) == len(pool._pool):
 * 				break
 */
      __pyx_t_1 = __pyx_pf_6gnocis_6models_12setPoolModel_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Or(__pyx_v_pIDs, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_pIDs, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "gnocis/models.pyx":134
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if hasModel )
 * 			if len(pIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 				break
 * 
 */
      __pyx_t_8 = PySet_GET_SIZE(__pyx_v_pIDs); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 134, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pool, __pyx_n_s_pool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 134, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = ((__pyx_t_8 == __pyx_t_7) != 0);
      if (__pyx_t_2) {

        /* "gnocis/models.pyx":135
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if hasModel )
 * 			if len(pIDs) == len(pool._pool):
 * 				break             # <<<<<<<<<<<<<<
 * 
 * def setNCores(n):
 */
        goto __pyx_L13_break;

        /* "gnocis/models.pyx":134
 * 			r = pool.map( getPoolProcessHasModel, [ () for _ in range(len(pool._pool)) ] )
 * 			pIDs = pIDs | set( pid for pid, hasModel in r if hasModel )
 * 			if len(pIDs) == len(pool._pool):             # <<<<<<<<<<<<<<
 * 				break
 * 
 */
      }
    }
    __pyx_L13_break:;

    /* "gnocis/models.pyx":125
 * 			if len(pIDs) == len(pool._pool):
 * 				break
 * 	if mdl != None:             # <<<<<<<<<<<<<<
 * 		# For non-None model, first purge the original model
 * 		setPoolModel(pool, None)
 */
  }

  /* "gnocis/models.pyx":113
 * 	return (processID, model != None)
 * 
 * def setPoolModel(pool, mdl):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	if mdl == None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("gnocis.models.setPoolModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pIDs);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_12setPoolModel_2generator4);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_12setPoolModel_5generator5);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":137
 * 				break
 * 
 * def setNCores(n):             # <<<<<<<<<<<<<<
 * 	"""
 * 	Sets the number of cores to use for multiprocessing.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_25setNCores(PyObject *__pyx_self, PyObject *__pyx_v_n); /*proto*/
static char __pyx_doc_6gnocis_6models_24setNCores[] = "\n\tSets the number of cores to use for multiprocessing.\n\t\n\t:param n: The number of cores to use.\n\t\n\t:type n: int\n\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_25setNCores = {"setNCores", (PyCFunction)__pyx_pw_6gnocis_6models_25setNCores, METH_O, __pyx_doc_6gnocis_6models_24setNCores};
static PyObject *__pyx_pw_6gnocis_6models_25setNCores(PyObject *__pyx_self, PyObject *__pyx_v_n) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setNCores (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_24setNCores(__pyx_self, ((PyObject *)__pyx_v_n));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_24setNCores(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setNCores", 0);

  /* "gnocis/models.pyx":146
 * 	"""
 * 	global nCoresUse, multiprocessingPool
 * 	nCoresUse = n             # <<<<<<<<<<<<<<
 * 	if multiprocessingPool != None:
 * 		multiprocessingPool.close()
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nCoresUse, __pyx_v_n) < 0) __PYX_ERR(0, 146, __pyx_L1_error)

  /* "gnocis/models.pyx":147
 * 	global nCoresUse, multiprocessingPool
 * 	nCoresUse = n
 * 	if multiprocessingPool != None:             # <<<<<<<<<<<<<<
 * 		multiprocessingPool.close()
 * 		multiprocessingPool.join()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_multiprocessingPool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, Py_None, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "gnocis/models.pyx":148
 * 	nCoresUse = n
 * 	if multiprocessingPool != None:
 * 		multiprocessingPool.close()             # <<<<<<<<<<<<<<
 * 		multiprocessingPool.join()
 * 		del multiprocessingPool
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_multiprocessingPool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":149
 * 	if multiprocessingPool != None:
 * 		multiprocessingPool.close()
 * 		multiprocessingPool.join()             # <<<<<<<<<<<<<<
 * 		del multiprocessingPool
 * 		multiprocessingPool = None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_multiprocessingPool); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_join); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":150
 * 		multiprocessingPool.close()
 * 		multiprocessingPool.join()
 * 		del multiprocessingPool             # <<<<<<<<<<<<<<
 * 		multiprocessingPool = None
 * 	if n > 1:
 */
    if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_multiprocessingPool) < 0) __PYX_ERR(0, 150, __pyx_L1_error)

    /* "gnocis/models.pyx":151
 * 		multiprocessingPool.join()
 * 		del multiprocessingPool
 * 		multiprocessingPool = None             # <<<<<<<<<<<<<<
 * 	if n > 1:
 * 		multiprocessingPool = mp.get_context('spawn').Pool(nCoresUse, initializer = initializePoolProcess, initargs = ('',))
 */
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_multiprocessingPool, Py_None) < 0) __PYX_ERR(0, 151, __pyx_L1_error)

    /* "gnocis/models.pyx":147
 * 	global nCoresUse, multiprocessingPool
 * 	nCoresUse = n
 * 	if multiprocessingPool != None:             # <<<<<<<<<<<<<<
 * 		multiprocessingPool.close()
 * 		multiprocessingPool.join()
 */
  }

  /* "gnocis/models.pyx":152
 * 		del multiprocessingPool
 * 		multiprocessingPool = None
 * 	if n > 1:             # <<<<<<<<<<<<<<
 * 		multiprocessingPool = mp.get_context('spawn').Pool(nCoresUse, initializer = initializePoolProcess, initargs = ('',))
 * 		initializePoolProcessIDs(multiprocessingPool)
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_n, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 152, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "gnocis/models.pyx":153
 * 		multiprocessingPool = None
 * 	if n > 1:
 * 		multiprocessingPool = mp.get_context('spawn').Pool(nCoresUse, initializer = initializePoolProcess, initargs = ('',))             # <<<<<<<<<<<<<<
 * 		initializePoolProcessIDs(multiprocessingPool)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_mp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_get_context); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_n_u_spawn) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_u_spawn);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Pool); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nCoresUse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_initializePoolProcess); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_initializer, __pyx_t_5) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_initargs, __pyx_tuple__3) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_multiprocessingPool, __pyx_t_5) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "gnocis/models.pyx":154
 * 	if n > 1:
 * 		multiprocessingPool = mp.get_context('spawn').Pool(nCoresUse, initializer = initializePoolProcess, initargs = ('',))
 * 		initializePoolProcessIDs(multiprocessingPool)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_initializePoolProcessIDs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_multiprocessingPool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "gnocis/models.pyx":152
 * 		del multiprocessingPool
 * 		multiprocessingPool = None
 * 	if n > 1:             # <<<<<<<<<<<<<<
 * 		multiprocessingPool = mp.get_context('spawn').Pool(nCoresUse, initializer = initializePoolProcess, initargs = ('',))
 * 		initializePoolProcessIDs(multiprocessingPool)
 */
  }

  /* "gnocis/models.pyx":137
 * 				break
 * 
 * def setNCores(n):             # <<<<<<<<<<<<<<
 * 	"""
 * 	Sets the number of cores to use for multiprocessing.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.models.setNCores", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":174
 * 	"""
 * 
 * 	def __init__(self, name, enableMultiprocessing = True):             # <<<<<<<<<<<<<<
 * 		self.name = name
 * 		self.enableMultiprocessing = enableMultiprocessing
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_enableMultiprocessing = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_enableMultiprocessing,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 174, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_enableMultiprocessing);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 174, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_enableMultiprocessing = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 174, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel___init__(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_enableMultiprocessing);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_enableMultiprocessing) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gnocis/models.pyx":175
 * 
 * 	def __init__(self, name, enableMultiprocessing = True):
 * 		self.name = name             # <<<<<<<<<<<<<<
 * 		self.enableMultiprocessing = enableMultiprocessing
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 175, __pyx_L1_error)

  /* "gnocis/models.pyx":176
 * 	def __init__(self, name, enableMultiprocessing = True):
 * 		self.name = name
 * 		self.enableMultiprocessing = enableMultiprocessing             # <<<<<<<<<<<<<<
 * 
 * 	# For pickling (used with multiprocessing)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_enableMultiprocessing, __pyx_v_enableMultiprocessing) < 0) __PYX_ERR(0, 176, __pyx_L1_error)

  /* "gnocis/models.pyx":174
 * 	"""
 * 
 * 	def __init__(self, name, enableMultiprocessing = True):             # <<<<<<<<<<<<<<
 * 		self.name = name
 * 		self.enableMultiprocessing = enableMultiprocessing
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":179
 * 
 * 	# For pickling (used with multiprocessing)
 * 	def __getstate__(self):             # <<<<<<<<<<<<<<
 * 		return self.__dict__
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_3__getstate__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_3__getstate__ = {"__getstate__", (PyCFunction)__pyx_pw_6gnocis_6models_13sequenceModel_3__getstate__, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_3__getstate__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_2__getstate__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_2__getstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "gnocis/models.pyx":180
 * 	# For pickling (used with multiprocessing)
 * 	def __getstate__(self):
 * 		return self.__dict__             # <<<<<<<<<<<<<<
 * 
 * 	def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":179
 * 
 * 	# For pickling (used with multiprocessing)
 * 	def __getstate__(self):             # <<<<<<<<<<<<<<
 * 		return self.__dict__
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":182
 * 		return self.__dict__
 * 
 * 	def __setstate__(self, state):             # <<<<<<<<<<<<<<
 * 		self.__dict__ = state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_5__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_5__setstate__ = {"__setstate__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_5__setstate__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_5__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_state,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, 1); __PYX_ERR(0, 182, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setstate__") < 0)) __PYX_ERR(0, 182, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_state = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 182, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_4__setstate__(__pyx_self, __pyx_v_self, __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_4__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "gnocis/models.pyx":183
 * 
 * 	def __setstate__(self, state):
 * 		self.__dict__ = state             # <<<<<<<<<<<<<<
 * 
 * 	# Gets a list of scores for a sequence list or stream. A stream is recommended when the sequence list is large, in order to avoid running out of memory.
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_dict, __pyx_v_state) < 0) __PYX_ERR(0, 183, __pyx_L1_error)

  /* "gnocis/models.pyx":182
 * 		return self.__dict__
 * 
 * 	def __setstate__(self, state):             # <<<<<<<<<<<<<<
 * 		self.__dict__ = state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":186
 * 
 * 	# Gets a list of scores for a sequence list or stream. A stream is recommended when the sequence list is large, in order to avoid running out of memory.
 * 	def getSequenceScores(self, seqs):             # <<<<<<<<<<<<<<
 * 		""" Scores a set of sequences, returning the maximum window score for each. Multiprocessing is enabled by default, but can be disabled in the constructor.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_7getSequenceScores(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_6getSequenceScores[] = " Scores a set of sequences, returning the maximum window score for each. Multiprocessing is enabled by default, but can be disabled in the constructor.\n\t\t\n\t\t:param seqs: Sequences to score.\n\t\t\n\t\t:type seqs: sequences/sequenceStream\n\t\t\n\t\t:return: List of the maximum window score per sequence\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_7getSequenceScores = {"getSequenceScores", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_7getSequenceScores, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_6getSequenceScores};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_7getSequenceScores(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getSequenceScores (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getSequenceScores", 1, 2, 2, 1); __PYX_ERR(0, 186, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getSequenceScores") < 0)) __PYX_ERR(0, 186, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getSequenceScores", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 186, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getSequenceScores", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_6getSequenceScores(__pyx_self, __pyx_v_self, __pyx_v_seqs);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_6getSequenceScores(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs) {
  struct __pyx_obj_6gnocis_9sequences_sequences *__pyx_9genexpr14__pyx_v_blk = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_blkScores = NULL;
  float __pyx_9genexpr14__pyx_v_score;
  PyObject *__pyx_9genexpr15__pyx_v_blkScores = NULL;
  float __pyx_9genexpr15__pyx_v_score;
  struct __pyx_obj_6gnocis_9sequences_sequences *__pyx_9genexpr16__pyx_v_blk = NULL;
  struct __pyx_obj_6gnocis_9sequences_sequence *__pyx_9genexpr16__pyx_v_cseq = NULL;
  struct __pyx_obj_6gnocis_9sequences_sequence *__pyx_9genexpr17__pyx_v_cseq = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  Py_ssize_t __pyx_t_17;
  float __pyx_t_18;
  int __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getSequenceScores", 0);

  /* "gnocis/models.pyx":200
 * 		cdef sequences blk
 * 		cdef sequence cseq
 * 		if nCoresUse > 1 and self.enableMultiprocessing:             # <<<<<<<<<<<<<<
 * 			# Wipe the first model, to ensure we have a clean slate
 * 			if not 'modelID' in self.__dict__.keys():
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nCoresUse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_enableMultiprocessing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "gnocis/models.pyx":202
 * 		if nCoresUse > 1 and self.enableMultiprocessing:
 * 			# Wipe the first model, to ensure we have a clean slate
 * 			if not 'modelID' in self.__dict__.keys():             # <<<<<<<<<<<<<<
 * 				global maxModelID
 * 				self.modelID = maxModelID
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_keys); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_modelID, __pyx_t_3, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_1 != 0);
    if (__pyx_t_4) {

      /* "gnocis/models.pyx":204
 * 			if not 'modelID' in self.__dict__.keys():
 * 				global maxModelID
 * 				self.modelID = maxModelID             # <<<<<<<<<<<<<<
 * 				maxModelID += 1
 * 			setPoolModel(multiprocessingPool, self)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_maxModelID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_modelID, __pyx_t_3) < 0) __PYX_ERR(0, 204, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "gnocis/models.pyx":205
 * 				global maxModelID
 * 				self.modelID = maxModelID
 * 				maxModelID += 1             # <<<<<<<<<<<<<<
 * 			setPoolModel(multiprocessingPool, self)
 * 			# Process in parallel
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_maxModelID); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_maxModelID, __pyx_t_5) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "gnocis/models.pyx":202
 * 		if nCoresUse > 1 and self.enableMultiprocessing:
 * 			# Wipe the first model, to ensure we have a clean slate
 * 			if not 'modelID' in self.__dict__.keys():             # <<<<<<<<<<<<<<
 * 				global maxModelID
 * 				self.modelID = maxModelID
 */
    }

    /* "gnocis/models.pyx":206
 * 				self.modelID = maxModelID
 * 				maxModelID += 1
 * 			setPoolModel(multiprocessingPool, self)             # <<<<<<<<<<<<<<
 * 			# Process in parallel
 * 			if isinstance(seqs, sequenceStream):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_setPoolModel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_multiprocessingPool); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_2, __pyx_v_self};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_2, __pyx_v_self};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_2);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_self);
      __pyx_t_2 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "gnocis/models.pyx":208
 * 			setPoolModel(multiprocessingPool, self)
 * 			# Process in parallel
 * 			if isinstance(seqs, sequenceStream):             # <<<<<<<<<<<<<<
 * 				return [
 * 					score
 */
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_seqs, __pyx_ptype_6gnocis_9sequences_sequenceStream); 
    __pyx_t_1 = (__pyx_t_4 != 0);
    if (__pyx_t_1) {

      /* "gnocis/models.pyx":209
 * 			# Process in parallel
 * 			if isinstance(seqs, sequenceStream):
 * 				return [             # <<<<<<<<<<<<<<
 * 					score
 * 					for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 */
      __Pyx_XDECREF(__pyx_r);
      { /* enter inner scope */
        __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "gnocis/models.pyx":211
 * 				return [
 * 					score
 * 					for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)             # <<<<<<<<<<<<<<
 * 					for blkScores in multiprocessingPool.map(
 * 						multiprocessModelDeployer,
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_seqs, __pyx_n_s_fetch); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 211, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nTreadFetch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_maxThreadFetchNT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 211, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_t_2, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L10_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_t_2, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L10_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 211, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, __pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_6);
          __pyx_t_2 = 0;
          __pyx_t_6 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
          __pyx_t_8 = __pyx_t_3; __Pyx_INCREF(__pyx_t_8); __pyx_t_11 = 0;
          __pyx_t_12 = NULL;
        } else {
          __pyx_t_11 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 211, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_12 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 211, __pyx_L10_error)
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        for (;;) {
          if (likely(!__pyx_t_12)) {
            if (likely(PyList_CheckExact(__pyx_t_8))) {
              if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_8)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 211, __pyx_L10_error)
              #else
              __pyx_t_3 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            } else {
              if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 211, __pyx_L10_error)
              #else
              __pyx_t_3 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 211, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
            }
          } else {
            __pyx_t_3 = __pyx_t_12(__pyx_t_8);
            if (unlikely(!__pyx_t_3)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 211, __pyx_L10_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_3);
          }
          if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_9sequences_sequences))))) __PYX_ERR(0, 211, __pyx_L10_error)
          __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_blk, ((struct __pyx_obj_6gnocis_9sequences_sequences *)__pyx_t_3));
          __pyx_t_3 = 0;

          /* "gnocis/models.pyx":212
 * 					score
 * 					for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 * 					for blkScores in multiprocessingPool.map(             # <<<<<<<<<<<<<<
 * 						multiprocessModelDeployer,
 * 						blk.split(nCoresUse)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_multiprocessingPool); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 212, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_map); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 212, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "gnocis/models.pyx":213
 * 					for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 * 					for blkScores in multiprocessingPool.map(
 * 						multiprocessModelDeployer,             # <<<<<<<<<<<<<<
 * 						blk.split(nCoresUse)
 * 					)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_multiprocessModelDeployer); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 213, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_10);

          /* "gnocis/models.pyx":214
 * 					for blkScores in multiprocessingPool.map(
 * 						multiprocessModelDeployer,
 * 						blk.split(nCoresUse)             # <<<<<<<<<<<<<<
 * 					)
 * 					for score in blkScores
 */
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_9genexpr14__pyx_v_blk), __pyx_n_s_split); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 214, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_nCoresUse); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 214, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
            }
          }
          __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_13);
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_9 = NULL;
          __pyx_t_7 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_7 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_t_10, __pyx_t_2};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L10_error)
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
            PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_t_10, __pyx_t_2};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L10_error)
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 212, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_9) {
              __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_9); __pyx_t_9 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_10);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_7, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_2);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_7, __pyx_t_2);
            __pyx_t_10 = 0;
            __pyx_t_2 = 0;
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "gnocis/models.pyx":212
 * 					score
 * 					for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 * 					for blkScores in multiprocessingPool.map(             # <<<<<<<<<<<<<<
 * 						multiprocessModelDeployer,
 * 						blk.split(nCoresUse)
 */
          if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
            __pyx_t_6 = __pyx_t_3; __Pyx_INCREF(__pyx_t_6); __pyx_t_15 = 0;
            __pyx_t_16 = NULL;
          } else {
            __pyx_t_15 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 212, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_16 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 212, __pyx_L10_error)
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          for (;;) {
            if (likely(!__pyx_t_16)) {
              if (likely(PyList_CheckExact(__pyx_t_6))) {
                if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_6)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_15); __Pyx_INCREF(__pyx_t_3); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 212, __pyx_L10_error)
                #else
                __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L10_error)
                __Pyx_GOTREF(__pyx_t_3);
                #endif
              } else {
                if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_15); __Pyx_INCREF(__pyx_t_3); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 212, __pyx_L10_error)
                #else
                __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L10_error)
                __Pyx_GOTREF(__pyx_t_3);
                #endif
              }
            } else {
              __pyx_t_3 = __pyx_t_16(__pyx_t_6);
              if (unlikely(!__pyx_t_3)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 212, __pyx_L10_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_3);
            }
            if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 212, __pyx_L10_error)
            __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_blkScores, ((PyObject*)__pyx_t_3));
            __pyx_t_3 = 0;

            /* "gnocis/models.pyx":216
 * 						blk.split(nCoresUse)
 * 					)
 * 					for score in blkScores             # <<<<<<<<<<<<<<
 * 				]
 * 			elif isinstance(seqs, sequences) or isinstance(seqs, list):
 */
            if (unlikely(__pyx_9genexpr14__pyx_v_blkScores == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
              __PYX_ERR(0, 216, __pyx_L10_error)
            }
            __pyx_t_3 = __pyx_9genexpr14__pyx_v_blkScores; __Pyx_INCREF(__pyx_t_3); __pyx_t_17 = 0;
            for (;;) {
              if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_13 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_17); __Pyx_INCREF(__pyx_t_13); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 216, __pyx_L10_error)
              #else
              __pyx_t_13 = PySequence_ITEM(__pyx_t_3, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 216, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_13);
              #endif
              __pyx_t_18 = __pyx_PyFloat_AsFloat(__pyx_t_13); if (unlikely((__pyx_t_18 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 216, __pyx_L10_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __pyx_9genexpr14__pyx_v_score = __pyx_t_18;

              /* "gnocis/models.pyx":210
 * 			if isinstance(seqs, sequenceStream):
 * 				return [
 * 					score             # <<<<<<<<<<<<<<
 * 					for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 * 					for blkScores in multiprocessingPool.map(
 */
              __pyx_t_13 = PyFloat_FromDouble(__pyx_9genexpr14__pyx_v_score); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 210, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_13);
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_13))) __PYX_ERR(0, 209, __pyx_L10_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

              /* "gnocis/models.pyx":216
 * 						blk.split(nCoresUse)
 * 					)
 * 					for score in blkScores             # <<<<<<<<<<<<<<
 * 				]
 * 			elif isinstance(seqs, sequences) or isinstance(seqs, list):
 */
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "gnocis/models.pyx":212
 * 					score
 * 					for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 * 					for blkScores in multiprocessingPool.map(             # <<<<<<<<<<<<<<
 * 						multiprocessModelDeployer,
 * 						blk.split(nCoresUse)
 */
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "gnocis/models.pyx":211
 * 				return [
 * 					score
 * 					for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)             # <<<<<<<<<<<<<<
 * 					for blkScores in multiprocessingPool.map(
 * 						multiprocessModelDeployer,
 */
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF((PyObject *)__pyx_9genexpr14__pyx_v_blk); __pyx_9genexpr14__pyx_v_blk = 0;
        __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_blkScores); __pyx_9genexpr14__pyx_v_blkScores = 0;
        goto __pyx_L17_exit_scope;
        __pyx_L10_error:;
        __Pyx_XDECREF((PyObject *)__pyx_9genexpr14__pyx_v_blk); __pyx_9genexpr14__pyx_v_blk = 0;
        __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_blkScores); __pyx_9genexpr14__pyx_v_blkScores = 0;
        goto __pyx_L1_error;
        __pyx_L17_exit_scope:;
      } /* exit inner scope */
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "gnocis/models.pyx":208
 * 			setPoolModel(multiprocessingPool, self)
 * 			# Process in parallel
 * 			if isinstance(seqs, sequenceStream):             # <<<<<<<<<<<<<<
 * 				return [
 * 					score
 */
    }

    /* "gnocis/models.pyx":218
 * 					for score in blkScores
 * 				]
 * 			elif isinstance(seqs, sequences) or isinstance(seqs, list):             # <<<<<<<<<<<<<<
 * 				return [
 * 					score
 */
    __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_seqs, __pyx_ptype_6gnocis_9sequences_sequences); 
    __pyx_t_19 = (__pyx_t_4 != 0);
    if (!__pyx_t_19) {
    } else {
      __pyx_t_1 = __pyx_t_19;
      goto __pyx_L18_bool_binop_done;
    }
    __pyx_t_19 = PyList_Check(__pyx_v_seqs); 
    __pyx_t_4 = (__pyx_t_19 != 0);
    __pyx_t_1 = __pyx_t_4;
    __pyx_L18_bool_binop_done:;
    if (__pyx_t_1) {

      /* "gnocis/models.pyx":219
 * 				]
 * 			elif isinstance(seqs, sequences) or isinstance(seqs, list):
 * 				return [             # <<<<<<<<<<<<<<
 * 					score
 * 					for blkScores in multiprocessingPool.map(
 */
      __Pyx_XDECREF(__pyx_r);
      { /* enter inner scope */
        __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 219, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "gnocis/models.pyx":221
 * 				return [
 * 					score
 * 					for blkScores in multiprocessingPool.map(             # <<<<<<<<<<<<<<
 * 						multiprocessModelDeployer,
 * 						seqs.split(nCoresUse)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_multiprocessingPool); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 221, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_map); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 221, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "gnocis/models.pyx":222
 * 					score
 * 					for blkScores in multiprocessingPool.map(
 * 						multiprocessModelDeployer,             # <<<<<<<<<<<<<<
 * 						seqs.split(nCoresUse)
 * 					)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_multiprocessModelDeployer); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 222, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "gnocis/models.pyx":223
 * 					for blkScores in multiprocessingPool.map(
 * 						multiprocessModelDeployer,
 * 						seqs.split(nCoresUse)             # <<<<<<<<<<<<<<
 * 					)
 * 					for score in blkScores
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_seqs, __pyx_n_s_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 223, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_nCoresUse); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 223, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_13 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_10);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 223, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_6, __pyx_t_13};
          __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 221, __pyx_L22_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_6, __pyx_t_13};
          __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 221, __pyx_L22_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 221, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, __pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_13);
          __pyx_t_6 = 0;
          __pyx_t_13 = 0;
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 221, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "gnocis/models.pyx":221
 * 				return [
 * 					score
 * 					for blkScores in multiprocessingPool.map(             # <<<<<<<<<<<<<<
 * 						multiprocessModelDeployer,
 * 						seqs.split(nCoresUse)
 */
        if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
          __pyx_t_3 = __pyx_t_8; __Pyx_INCREF(__pyx_t_3); __pyx_t_11 = 0;
          __pyx_t_12 = NULL;
        } else {
          __pyx_t_11 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 221, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_12 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 221, __pyx_L22_error)
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        for (;;) {
          if (likely(!__pyx_t_12)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_8 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_11); __Pyx_INCREF(__pyx_t_8); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 221, __pyx_L22_error)
              #else
              __pyx_t_8 = PySequence_ITEM(__pyx_t_3, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 221, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_8);
              #endif
            } else {
              if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_11); __Pyx_INCREF(__pyx_t_8); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 221, __pyx_L22_error)
              #else
              __pyx_t_8 = PySequence_ITEM(__pyx_t_3, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 221, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_8);
              #endif
            }
          } else {
            __pyx_t_8 = __pyx_t_12(__pyx_t_3);
            if (unlikely(!__pyx_t_8)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 221, __pyx_L22_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_8);
          }
          if (!(likely(PyList_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_8)->tp_name), 0))) __PYX_ERR(0, 221, __pyx_L22_error)
          __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_blkScores, ((PyObject*)__pyx_t_8));
          __pyx_t_8 = 0;

          /* "gnocis/models.pyx":225
 * 						seqs.split(nCoresUse)
 * 					)
 * 					for score in blkScores             # <<<<<<<<<<<<<<
 * 				]
 * 		if isinstance(seqs, sequenceStream):
 */
          if (unlikely(__pyx_9genexpr15__pyx_v_blkScores == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
            __PYX_ERR(0, 225, __pyx_L22_error)
          }
          __pyx_t_8 = __pyx_9genexpr15__pyx_v_blkScores; __Pyx_INCREF(__pyx_t_8); __pyx_t_15 = 0;
          for (;;) {
            if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_15); __Pyx_INCREF(__pyx_t_10); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 225, __pyx_L22_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_8, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 225, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
            __pyx_t_18 = __pyx_PyFloat_AsFloat(__pyx_t_10); if (unlikely((__pyx_t_18 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 225, __pyx_L22_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_9genexpr15__pyx_v_score = __pyx_t_18;

            /* "gnocis/models.pyx":220
 * 			elif isinstance(seqs, sequences) or isinstance(seqs, list):
 * 				return [
 * 					score             # <<<<<<<<<<<<<<
 * 					for blkScores in multiprocessingPool.map(
 * 						multiprocessModelDeployer,
 */
            __pyx_t_10 = PyFloat_FromDouble(__pyx_9genexpr15__pyx_v_score); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 220, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 219, __pyx_L22_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "gnocis/models.pyx":225
 * 						seqs.split(nCoresUse)
 * 					)
 * 					for score in blkScores             # <<<<<<<<<<<<<<
 * 				]
 * 		if isinstance(seqs, sequenceStream):
 */
          }
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "gnocis/models.pyx":221
 * 				return [
 * 					score
 * 					for blkScores in multiprocessingPool.map(             # <<<<<<<<<<<<<<
 * 						multiprocessModelDeployer,
 * 						seqs.split(nCoresUse)
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_blkScores); __pyx_9genexpr15__pyx_v_blkScores = 0;
        goto __pyx_L27_exit_scope;
        __pyx_L22_error:;
        __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_blkScores); __pyx_9genexpr15__pyx_v_blkScores = 0;
        goto __pyx_L1_error;
        __pyx_L27_exit_scope:;
      } /* exit inner scope */
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "gnocis/models.pyx":218
 * 					for score in blkScores
 * 				]
 * 			elif isinstance(seqs, sequences) or isinstance(seqs, list):             # <<<<<<<<<<<<<<
 * 				return [
 * 					score
 */
    }

    /* "gnocis/models.pyx":200
 * 		cdef sequences blk
 * 		cdef sequence cseq
 * 		if nCoresUse > 1 and self.enableMultiprocessing:             # <<<<<<<<<<<<<<
 * 			# Wipe the first model, to ensure we have a clean slate
 * 			if not 'modelID' in self.__dict__.keys():
 */
  }

  /* "gnocis/models.pyx":227
 * 					for score in blkScores
 * 				]
 * 		if isinstance(seqs, sequenceStream):             # <<<<<<<<<<<<<<
 * 			return [
 * 				self.scoreSequence(cseq)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_seqs, __pyx_ptype_6gnocis_9sequences_sequenceStream); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "gnocis/models.pyx":228
 * 				]
 * 		if isinstance(seqs, sequenceStream):
 * 			return [             # <<<<<<<<<<<<<<
 * 				self.scoreSequence(cseq)
 * 				for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 */
    __Pyx_XDECREF(__pyx_r);
    { /* enter inner scope */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 228, __pyx_L31_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "gnocis/models.pyx":230
 * 			return [
 * 				self.scoreSequence(cseq)
 * 				for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)             # <<<<<<<<<<<<<<
 * 				for cseq in blk
 * 			]
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_seqs, __pyx_n_s_fetch); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 230, __pyx_L31_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_nTreadFetch); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 230, __pyx_L31_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_maxThreadFetchNT); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 230, __pyx_L31_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_13};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L31_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_13};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L31_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      {
        __pyx_t_2 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_7, __pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_7, __pyx_t_13);
        __pyx_t_10 = 0;
        __pyx_t_13 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_8 = __pyx_t_3; __Pyx_INCREF(__pyx_t_8); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_11 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 230, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_12 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 230, __pyx_L31_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 230, __pyx_L31_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 230, __pyx_L31_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_12(__pyx_t_8);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 230, __pyx_L31_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_9sequences_sequences))))) __PYX_ERR(0, 230, __pyx_L31_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_blk, ((struct __pyx_obj_6gnocis_9sequences_sequences *)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "gnocis/models.pyx":231
 * 				self.scoreSequence(cseq)
 * 				for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 * 				for cseq in blk             # <<<<<<<<<<<<<<
 * 			]
 * 		elif isinstance(seqs, sequences) or isinstance(seqs, list):
 */
        if (likely(PyList_CheckExact(((PyObject *)__pyx_9genexpr16__pyx_v_blk))) || PyTuple_CheckExact(((PyObject *)__pyx_9genexpr16__pyx_v_blk))) {
          __pyx_t_3 = ((PyObject *)__pyx_9genexpr16__pyx_v_blk); __Pyx_INCREF(__pyx_t_3); __pyx_t_15 = 0;
          __pyx_t_16 = NULL;
        } else {
          __pyx_t_15 = -1; __pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_9genexpr16__pyx_v_blk)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 231, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_16 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 231, __pyx_L31_error)
        }
        for (;;) {
          if (likely(!__pyx_t_16)) {
            if (likely(PyList_CheckExact(__pyx_t_3))) {
              if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_15); __Pyx_INCREF(__pyx_t_2); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 231, __pyx_L31_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 231, __pyx_L31_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_15); __Pyx_INCREF(__pyx_t_2); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 231, __pyx_L31_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 231, __pyx_L31_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_16(__pyx_t_3);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 231, __pyx_L31_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gnocis_9sequences_sequence))))) __PYX_ERR(0, 231, __pyx_L31_error)
          __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_cseq, ((struct __pyx_obj_6gnocis_9sequences_sequence *)__pyx_t_2));
          __pyx_t_2 = 0;

          /* "gnocis/models.pyx":229
 * 		if isinstance(seqs, sequenceStream):
 * 			return [
 * 				self.scoreSequence(cseq)             # <<<<<<<<<<<<<<
 * 				for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 * 				for cseq in blk
 */
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_scoreSequence); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 229, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_10 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
            }
          }
          __pyx_t_2 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_10, ((PyObject *)__pyx_9genexpr16__pyx_v_cseq)) : __Pyx_PyObject_CallOneArg(__pyx_t_13, ((PyObject *)__pyx_9genexpr16__pyx_v_cseq));
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L31_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 228, __pyx_L31_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "gnocis/models.pyx":231
 * 				self.scoreSequence(cseq)
 * 				for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)
 * 				for cseq in blk             # <<<<<<<<<<<<<<
 * 			]
 * 		elif isinstance(seqs, sequences) or isinstance(seqs, list):
 */
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "gnocis/models.pyx":230
 * 			return [
 * 				self.scoreSequence(cseq)
 * 				for blk in seqs.fetch(nTreadFetch, maxThreadFetchNT)             # <<<<<<<<<<<<<<
 * 				for cseq in blk
 * 			]
 */
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF((PyObject *)__pyx_9genexpr16__pyx_v_blk); __pyx_9genexpr16__pyx_v_blk = 0;
      __Pyx_XDECREF((PyObject *)__pyx_9genexpr16__pyx_v_cseq); __pyx_9genexpr16__pyx_v_cseq = 0;
      goto __pyx_L36_exit_scope;
      __pyx_L31_error:;
      __Pyx_XDECREF((PyObject *)__pyx_9genexpr16__pyx_v_blk); __pyx_9genexpr16__pyx_v_blk = 0;
      __Pyx_XDECREF((PyObject *)__pyx_9genexpr16__pyx_v_cseq); __pyx_9genexpr16__pyx_v_cseq = 0;
      goto __pyx_L1_error;
      __pyx_L36_exit_scope:;
    } /* exit inner scope */
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "gnocis/models.pyx":227
 * 					for score in blkScores
 * 				]
 * 		if isinstance(seqs, sequenceStream):             # <<<<<<<<<<<<<<
 * 			return [
 * 				self.scoreSequence(cseq)
 */
  }

  /* "gnocis/models.pyx":233
 * 				for cseq in blk
 * 			]
 * 		elif isinstance(seqs, sequences) or isinstance(seqs, list):             # <<<<<<<<<<<<<<
 * 			return [ self.scoreSequence(cseq) for cseq in seqs ]
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_seqs, __pyx_ptype_6gnocis_9sequences_sequences); 
  __pyx_t_19 = (__pyx_t_1 != 0);
  if (!__pyx_t_19) {
  } else {
    __pyx_t_4 = __pyx_t_19;
    goto __pyx_L37_bool_binop_done;
  }
  __pyx_t_19 = PyList_Check(__pyx_v_seqs); 
  __pyx_t_1 = (__pyx_t_19 != 0);
  __pyx_t_4 = __pyx_t_1;
  __pyx_L37_bool_binop_done:;
  if (__pyx_t_4) {

    /* "gnocis/models.pyx":234
 * 			]
 * 		elif isinstance(seqs, sequences) or isinstance(seqs, list):
 * 			return [ self.scoreSequence(cseq) for cseq in seqs ]             # <<<<<<<<<<<<<<
 * 
 * 	# Gets the threshold that gives optimal accuracy on a pair of lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 */
    __Pyx_XDECREF(__pyx_r);
    { /* enter inner scope */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 234, __pyx_L41_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (likely(PyList_CheckExact(__pyx_v_seqs)) || PyTuple_CheckExact(__pyx_v_seqs)) {
        __pyx_t_8 = __pyx_v_seqs; __Pyx_INCREF(__pyx_t_8); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_11 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_seqs); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 234, __pyx_L41_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_12 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 234, __pyx_L41_error)
      }
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 234, __pyx_L41_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 234, __pyx_L41_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_12(__pyx_t_8);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 234, __pyx_L41_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_9sequences_sequence))))) __PYX_ERR(0, 234, __pyx_L41_error)
        __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_cseq, ((struct __pyx_obj_6gnocis_9sequences_sequence *)__pyx_t_3));
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_scoreSequence); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L41_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_3 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_13, ((PyObject *)__pyx_9genexpr17__pyx_v_cseq)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_9genexpr17__pyx_v_cseq));
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L41_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 234, __pyx_L41_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF((PyObject *)__pyx_9genexpr17__pyx_v_cseq); __pyx_9genexpr17__pyx_v_cseq = 0;
      goto __pyx_L44_exit_scope;
      __pyx_L41_error:;
      __Pyx_XDECREF((PyObject *)__pyx_9genexpr17__pyx_v_cseq); __pyx_9genexpr17__pyx_v_cseq = 0;
      goto __pyx_L1_error;
      __pyx_L44_exit_scope:;
    } /* exit inner scope */
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "gnocis/models.pyx":233
 * 				for cseq in blk
 * 			]
 * 		elif isinstance(seqs, sequences) or isinstance(seqs, list):             # <<<<<<<<<<<<<<
 * 			return [ self.scoreSequence(cseq) for cseq in seqs ]
 * 
 */
  }

  /* "gnocis/models.pyx":186
 * 
 * 	# Gets a list of scores for a sequence list or stream. A stream is recommended when the sequence list is large, in order to avoid running out of memory.
 * 	def getSequenceScores(self, seqs):             # <<<<<<<<<<<<<<
 * 		""" Scores a set of sequences, returning the maximum window score for each. Multiprocessing is enabled by default, but can be disabled in the constructor.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getSequenceScores", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr14__pyx_v_blk);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_blkScores);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_blkScores);
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr16__pyx_v_blk);
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr16__pyx_v_cseq);
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr17__pyx_v_cseq);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":237
 * 
 * 	# Gets the threshold that gives optimal accuracy on a pair of lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getOptimalAccuracyThreshold(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value optimized for accuracy to a set of positive and a set of negative sequences.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_34__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_9getOptimalAccuracyThreshold(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_8getOptimalAccuracyThreshold[] = " Gets a threshold value optimized for accuracy to a set of positive and a set of negative sequences.\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t\n\t\t:return: List of the maximum window score per sequence\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_9getOptimalAccuracyThreshold = {"getOptimalAccuracyThreshold", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_9getOptimalAccuracyThreshold, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_8getOptimalAccuracyThreshold};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_9getOptimalAccuracyThreshold(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getOptimalAccuracyThreshold (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getOptimalAccuracyThreshold", 0, 2, 4, 1); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getOptimalAccuracyThreshold") < 0)) __PYX_ERR(0, 237, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getOptimalAccuracyThreshold", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 237, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getOptimalAccuracyThreshold", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_8getOptimalAccuracyThreshold(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":254
 * 		positiveScores = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		vPairs = sorted(positiveScores + negativeScores, key = lambda x: x.score)             # <<<<<<<<<<<<<<
 * 		TP, FP, TN, FN = len(positiveScores), len(negativeScores), 0, 0
 * 		threshold = float('-INF')
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_27getOptimalAccuracyThreshold_lambda6(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_27getOptimalAccuracyThreshold_lambda6 = {"lambda6", (PyCFunction)__pyx_pw_6gnocis_6models_13sequenceModel_27getOptimalAccuracyThreshold_lambda6, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_27getOptimalAccuracyThreshold_lambda6(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda6 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda6(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda6(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda6", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getOptimalAccuracyThreshold.lambda6", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":237
 * 
 * 	# Gets the threshold that gives optimal accuracy on a pair of lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getOptimalAccuracyThreshold(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value optimized for accuracy to a set of positive and a set of negative sequences.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_8getOptimalAccuracyThreshold(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_v_positives = NULL;
  PyObject *__pyx_v_negatives = NULL;
  PyObject *__pyx_v_positiveScores = NULL;
  PyObject *__pyx_v_negativeScores = NULL;
  PyObject *__pyx_v_vPairs = NULL;
  PyObject *__pyx_v_TP = NULL;
  PyObject *__pyx_v_FP = NULL;
  PyObject *__pyx_v_TN = NULL;
  PyObject *__pyx_v_FN = NULL;
  PyObject *__pyx_v_threshold = NULL;
  double __pyx_v_ACC;
  PyObject *__pyx_v_vp = NULL;
  double __pyx_v_cACC;
  PyObject *__pyx_9genexpr18__pyx_v_score = NULL;
  PyObject *__pyx_9genexpr19__pyx_v_score = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  double __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getOptimalAccuracyThreshold", 0);

  /* "gnocis/models.pyx":251
 * 		:rtype: list
 * 		"""
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])             # <<<<<<<<<<<<<<
 * 		positiveScores = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_seqs, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_labelPositive);
  __Pyx_GIVEREF(__pyx_v_labelPositive);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_labelPositive);
  __Pyx_INCREF(__pyx_v_labelNegative);
  __Pyx_GIVEREF(__pyx_v_labelNegative);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_v_labelNegative);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 251, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 251, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_positives = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_negatives = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":252
 * 		"""
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 		positiveScores = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]             # <<<<<<<<<<<<<<
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		vPairs = sorted(positiveScores + negativeScores, key = lambda x: x.score)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_positives) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_positives);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 252, __pyx_L7_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 252, __pyx_L7_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 252, __pyx_L7_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 252, __pyx_L7_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_score, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_score, __pyx_9genexpr18__pyx_v_score) < 0) __PYX_ERR(0, 252, __pyx_L7_error)
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_label, Py_True) < 0) __PYX_ERR(0, 252, __pyx_L7_error)
      __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 252, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 252, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_score); __pyx_9genexpr18__pyx_v_score = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_score); __pyx_9genexpr18__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_positiveScores = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":253
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 		positiveScores = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]             # <<<<<<<<<<<<<<
 * 		vPairs = sorted(positiveScores + negativeScores, key = lambda x: x.score)
 * 		TP, FP, TN, FN = len(positiveScores), len(negativeScores), 0, 0
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_negatives) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_negatives);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 253, __pyx_L13_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 253, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 253, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 253, __pyx_L13_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr19__pyx_v_score, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_score, __pyx_9genexpr19__pyx_v_score) < 0) __PYX_ERR(0, 253, __pyx_L13_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_label, Py_False) < 0) __PYX_ERR(0, 253, __pyx_L13_error)
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 253, __pyx_L13_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_score); __pyx_9genexpr19__pyx_v_score = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_score); __pyx_9genexpr19__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  __pyx_v_negativeScores = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":254
 * 		positiveScores = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		vPairs = sorted(positiveScores + negativeScores, key = lambda x: x.score)             # <<<<<<<<<<<<<<
 * 		TP, FP, TN, FN = len(positiveScores), len(negativeScores), 0, 0
 * 		threshold = float('-INF')
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_positiveScores, __pyx_v_negativeScores); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_27getOptimalAccuracyThreshold_lambda6, 0, __pyx_n_s_sequenceModel_getOptimalAccuracy, NULL, __pyx_n_s_gnocis_models, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_vPairs = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":255
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		vPairs = sorted(positiveScores + negativeScores, key = lambda x: x.score)
 * 		TP, FP, TN, FN = len(positiveScores), len(negativeScores), 0, 0             # <<<<<<<<<<<<<<
 * 		threshold = float('-INF')
 * 		ACC = float( TP + TN ) / float( TP + TN + FP + FN )
 */
  __pyx_t_6 = PyList_GET_SIZE(__pyx_v_positiveScores); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 255, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = PyList_GET_SIZE(__pyx_v_negativeScores); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 255, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_int_0;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_2 = __pyx_int_0;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_TP = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_FP = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_TN = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_FN = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "gnocis/models.pyx":256
 * 		vPairs = sorted(positiveScores + negativeScores, key = lambda x: x.score)
 * 		TP, FP, TN, FN = len(positiveScores), len(negativeScores), 0, 0
 * 		threshold = float('-INF')             # <<<<<<<<<<<<<<
 * 		ACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 		for vp in vPairs:
 */
  __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_kp_u_INF); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_threshold = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "gnocis/models.pyx":257
 * 		TP, FP, TN, FN = len(positiveScores), len(negativeScores), 0, 0
 * 		threshold = float('-INF')
 * 		ACC = float( TP + TN ) / float( TP + TN + FP + FN )             # <<<<<<<<<<<<<<
 * 		for vp in vPairs:
 * 			if vp.label:
 */
  __pyx_t_2 = PyNumber_Add(__pyx_v_TP, __pyx_v_TN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_PyObject_AsDouble(__pyx_t_2); if (unlikely(__pyx_t_8 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_v_TP, __pyx_v_TN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyNumber_Add(__pyx_t_2, __pyx_v_FP); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_4, __pyx_v_FN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_2); if (unlikely(__pyx_t_9 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_9 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 257, __pyx_L1_error)
  }
  __pyx_v_ACC = (__pyx_t_8 / __pyx_t_9);

  /* "gnocis/models.pyx":258
 * 		threshold = float('-INF')
 * 		ACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 		for vp in vPairs:             # <<<<<<<<<<<<<<
 * 			if vp.label:
 * 				TP -= 1
 */
  if (likely(PyList_CheckExact(__pyx_v_vPairs)) || PyTuple_CheckExact(__pyx_v_vPairs)) {
    __pyx_t_2 = __pyx_v_vPairs; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_vPairs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 258, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 258, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 258, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 258, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_vp, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "gnocis/models.pyx":259
 * 		ACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 		for vp in vPairs:
 * 			if vp.label:             # <<<<<<<<<<<<<<
 * 				TP -= 1
 * 				FN += 1
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_vp, __pyx_n_s_label); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_10) {

      /* "gnocis/models.pyx":260
 * 		for vp in vPairs:
 * 			if vp.label:
 * 				TP -= 1             # <<<<<<<<<<<<<<
 * 				FN += 1
 * 			else:
 */
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_TP, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_TP, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "gnocis/models.pyx":261
 * 			if vp.label:
 * 				TP -= 1
 * 				FN += 1             # <<<<<<<<<<<<<<
 * 			else:
 * 				FP -= 1
 */
      __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_FN, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_FN, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "gnocis/models.pyx":259
 * 		ACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 		for vp in vPairs:
 * 			if vp.label:             # <<<<<<<<<<<<<<
 * 				TP -= 1
 * 				FN += 1
 */
      goto __pyx_L19;
    }

    /* "gnocis/models.pyx":263
 * 				FN += 1
 * 			else:
 * 				FP -= 1             # <<<<<<<<<<<<<<
 * 				TN += 1
 * 			cACC = float( TP + TN ) / float( TP + TN + FP + FN )
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_FP, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 263, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_FP, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "gnocis/models.pyx":264
 * 			else:
 * 				FP -= 1
 * 				TN += 1             # <<<<<<<<<<<<<<
 * 			cACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 			if cACC >= ACC:
 */
      __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_TN, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_TN, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L19:;

    /* "gnocis/models.pyx":265
 * 				FP -= 1
 * 				TN += 1
 * 			cACC = float( TP + TN ) / float( TP + TN + FP + FN )             # <<<<<<<<<<<<<<
 * 			if cACC >= ACC:
 * 				ACC = cACC
 */
    __pyx_t_4 = PyNumber_Add(__pyx_v_TP, __pyx_v_TN); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_4); if (unlikely(__pyx_t_9 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_v_TP, __pyx_v_TN); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyNumber_Add(__pyx_t_4, __pyx_v_FP); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_v_FN); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyObject_AsDouble(__pyx_t_4); if (unlikely(__pyx_t_8 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__pyx_t_8 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 265, __pyx_L1_error)
    }
    __pyx_v_cACC = (__pyx_t_9 / __pyx_t_8);

    /* "gnocis/models.pyx":266
 * 				TN += 1
 * 			cACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 			if cACC >= ACC:             # <<<<<<<<<<<<<<
 * 				ACC = cACC
 * 				threshold = vp.score
 */
    __pyx_t_10 = ((__pyx_v_cACC >= __pyx_v_ACC) != 0);
    if (__pyx_t_10) {

      /* "gnocis/models.pyx":267
 * 			cACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 			if cACC >= ACC:
 * 				ACC = cACC             # <<<<<<<<<<<<<<
 * 				threshold = vp.score
 * 		return threshold
 */
      __pyx_v_ACC = __pyx_v_cACC;

      /* "gnocis/models.pyx":268
 * 			if cACC >= ACC:
 * 				ACC = cACC
 * 				threshold = vp.score             # <<<<<<<<<<<<<<
 * 		return threshold
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_vp, __pyx_n_s_score); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_threshold, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "gnocis/models.pyx":266
 * 				TN += 1
 * 			cACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 			if cACC >= ACC:             # <<<<<<<<<<<<<<
 * 				ACC = cACC
 * 				threshold = vp.score
 */
    }

    /* "gnocis/models.pyx":258
 * 		threshold = float('-INF')
 * 		ACC = float( TP + TN ) / float( TP + TN + FP + FN )
 * 		for vp in vPairs:             # <<<<<<<<<<<<<<
 * 			if vp.label:
 * 				TP -= 1
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":269
 * 				ACC = cACC
 * 				threshold = vp.score
 * 		return threshold             # <<<<<<<<<<<<<<
 * 
 * 	# Gets the threshold that gives a desired precision for two lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_threshold);
  __pyx_r = __pyx_v_threshold;
  goto __pyx_L0;

  /* "gnocis/models.pyx":237
 * 
 * 	# Gets the threshold that gives optimal accuracy on a pair of lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getOptimalAccuracyThreshold(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value optimized for accuracy to a set of positive and a set of negative sequences.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getOptimalAccuracyThreshold", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_positives);
  __Pyx_XDECREF(__pyx_v_negatives);
  __Pyx_XDECREF(__pyx_v_positiveScores);
  __Pyx_XDECREF(__pyx_v_negativeScores);
  __Pyx_XDECREF(__pyx_v_vPairs);
  __Pyx_XDECREF(__pyx_v_TP);
  __Pyx_XDECREF(__pyx_v_FP);
  __Pyx_XDECREF(__pyx_v_TN);
  __Pyx_XDECREF(__pyx_v_FN);
  __Pyx_XDECREF(__pyx_v_threshold);
  __Pyx_XDECREF(__pyx_v_vp);
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_score);
  __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_score);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":272
 * 
 * 	# Gets the threshold that gives a desired precision for two lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getPrecisionThreshold(self, positives, negatives, wantedPrecision):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value for a desired precision to a set of positive and a set of negative sequences. Linear interpolation is used in order to achieve a close approximation.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_11getPrecisionThreshold(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_10getPrecisionThreshold[] = " Gets a threshold value for a desired precision to a set of positive and a set of negative sequences. Linear interpolation is used in order to achieve a close approximation.\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t:param wantedPrecision: The precision to approximate.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t:type wantedPrecision: float\n\t\t\n\t\t:return: List of the maximum window score per sequence\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_11getPrecisionThreshold = {"getPrecisionThreshold", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_11getPrecisionThreshold, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_10getPrecisionThreshold};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_11getPrecisionThreshold(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_positives = 0;
  PyObject *__pyx_v_negatives = 0;
  PyObject *__pyx_v_wantedPrecision = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getPrecisionThreshold (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_positives,&__pyx_n_s_negatives,&__pyx_n_s_wantedPrecision,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_positives)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getPrecisionThreshold", 1, 4, 4, 1); __PYX_ERR(0, 272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_negatives)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getPrecisionThreshold", 1, 4, 4, 2); __PYX_ERR(0, 272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_wantedPrecision)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getPrecisionThreshold", 1, 4, 4, 3); __PYX_ERR(0, 272, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getPrecisionThreshold") < 0)) __PYX_ERR(0, 272, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_positives = values[1];
    __pyx_v_negatives = values[2];
    __pyx_v_wantedPrecision = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getPrecisionThreshold", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 272, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getPrecisionThreshold", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_10getPrecisionThreshold(__pyx_self, __pyx_v_self, __pyx_v_positives, __pyx_v_negatives, __pyx_v_wantedPrecision);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":291
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		#
 * 		classScoresSorted = sorted( positiveScores + negativeScores, key = lambda x: x.score )             # <<<<<<<<<<<<<<
 * 		# Search for optimal threshold
 * 		TP, FP = ( len(positiveScores), len(negativeScores) )
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_21getPrecisionThreshold_lambda7(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_21getPrecisionThreshold_lambda7 = {"lambda7", (PyCFunction)__pyx_pw_6gnocis_6models_13sequenceModel_21getPrecisionThreshold_lambda7, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_21getPrecisionThreshold_lambda7(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda7 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda7(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda7(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda7", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getPrecisionThreshold.lambda7", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":272
 * 
 * 	# Gets the threshold that gives a desired precision for two lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getPrecisionThreshold(self, positives, negatives, wantedPrecision):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value for a desired precision to a set of positive and a set of negative sequences. Linear interpolation is used in order to achieve a close approximation.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_10getPrecisionThreshold(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_positives, PyObject *__pyx_v_negatives, PyObject *__pyx_v_wantedPrecision) {
  PyObject *__pyx_v_positiveScores = NULL;
  PyObject *__pyx_v_negativeScores = NULL;
  PyObject *__pyx_v_classScoresSorted = NULL;
  PyObject *__pyx_v_TP = NULL;
  PyObject *__pyx_v_FP = NULL;
  PyObject *__pyx_v_aThreshold = NULL;
  PyObject *__pyx_v_bThreshold = NULL;
  double __pyx_v_aPrecision;
  double __pyx_v_bPrecision;
  PyObject *__pyx_v_e = NULL;
  double __pyx_v_wPrecision;
  PyObject *__pyx_v_wThreshold = NULL;
  PyObject *__pyx_v_threshold = NULL;
  PyObject *__pyx_9genexpr20__pyx_v_score = NULL;
  PyObject *__pyx_9genexpr21__pyx_v_score = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  double __pyx_t_7;
  double __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getPrecisionThreshold", 0);

  /* "gnocis/models.pyx":288
 * 		:rtype: list
 * 		"""
 * 		positiveScores = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]             # <<<<<<<<<<<<<<
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		#
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_positives) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_positives);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 288, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 288, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 288, __pyx_L5_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_6(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 288, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr20__pyx_v_score, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_score, __pyx_9genexpr20__pyx_v_score) < 0) __PYX_ERR(0, 288, __pyx_L5_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_label, Py_True) < 0) __PYX_ERR(0, 288, __pyx_L5_error)
      __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 288, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 288, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_score); __pyx_9genexpr20__pyx_v_score = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_score); __pyx_9genexpr20__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_v_positiveScores = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":289
 * 		"""
 * 		positiveScores = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]             # <<<<<<<<<<<<<<
 * 		#
 * 		classScoresSorted = sorted( positiveScores + negativeScores, key = lambda x: x.score )
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_v_negatives) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_negatives);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 289, __pyx_L11_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 289, __pyx_L11_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 289, __pyx_L11_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 289, __pyx_L11_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr21__pyx_v_score, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_score, __pyx_9genexpr21__pyx_v_score) < 0) __PYX_ERR(0, 289, __pyx_L11_error)
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_label, Py_False) < 0) __PYX_ERR(0, 289, __pyx_L11_error)
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 289, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_score); __pyx_9genexpr21__pyx_v_score = 0;
    goto __pyx_L14_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_score); __pyx_9genexpr21__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L14_exit_scope:;
  } /* exit inner scope */
  __pyx_v_negativeScores = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":291
 * 		negativeScores = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		#
 * 		classScoresSorted = sorted( positiveScores + negativeScores, key = lambda x: x.score )             # <<<<<<<<<<<<<<
 * 		# Search for optimal threshold
 * 		TP, FP = ( len(positiveScores), len(negativeScores) )
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_positiveScores, __pyx_v_negativeScores); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_21getPrecisionThreshold_lambda7, 0, __pyx_n_s_sequenceModel_getPrecisionThresh, NULL, __pyx_n_s_gnocis_models, __pyx_d, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_key, __pyx_t_2) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_classScoresSorted = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "gnocis/models.pyx":293
 * 		classScoresSorted = sorted( positiveScores + negativeScores, key = lambda x: x.score )
 * 		# Search for optimal threshold
 * 		TP, FP = ( len(positiveScores), len(negativeScores) )             # <<<<<<<<<<<<<<
 * 		aThreshold, bThreshold = (0, 0)
 * 		aPrecision, bPrecision = (-1, -1)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_positiveScores); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 293, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_negativeScores); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 293, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_TP = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_FP = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":294
 * 		# Search for optimal threshold
 * 		TP, FP = ( len(positiveScores), len(negativeScores) )
 * 		aThreshold, bThreshold = (0, 0)             # <<<<<<<<<<<<<<
 * 		aPrecision, bPrecision = (-1, -1)
 * 		for e in classScoresSorted:
 */
  __pyx_t_1 = __pyx_int_0;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_int_0;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_aThreshold = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_bThreshold = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "gnocis/models.pyx":295
 * 		TP, FP = ( len(positiveScores), len(negativeScores) )
 * 		aThreshold, bThreshold = (0, 0)
 * 		aPrecision, bPrecision = (-1, -1)             # <<<<<<<<<<<<<<
 * 		for e in classScoresSorted:
 * 			aPrecision = bPrecision
 */
  __pyx_t_7 = -1.0;
  __pyx_t_8 = -1.0;
  __pyx_v_aPrecision = __pyx_t_7;
  __pyx_v_bPrecision = __pyx_t_8;

  /* "gnocis/models.pyx":296
 * 		aThreshold, bThreshold = (0, 0)
 * 		aPrecision, bPrecision = (-1, -1)
 * 		for e in classScoresSorted:             # <<<<<<<<<<<<<<
 * 			aPrecision = bPrecision
 * 			bPrecision = float(TP)/float(TP+FP)
 */
  if (likely(PyList_CheckExact(__pyx_v_classScoresSorted)) || PyTuple_CheckExact(__pyx_v_classScoresSorted)) {
    __pyx_t_2 = __pyx_v_classScoresSorted; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_classScoresSorted); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 296, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 296, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 296, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 296, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":297
 * 		aPrecision, bPrecision = (-1, -1)
 * 		for e in classScoresSorted:
 * 			aPrecision = bPrecision             # <<<<<<<<<<<<<<
 * 			bPrecision = float(TP)/float(TP+FP)
 * 			aThreshold = bThreshold
 */
    __pyx_v_aPrecision = __pyx_v_bPrecision;

    /* "gnocis/models.pyx":298
 * 		for e in classScoresSorted:
 * 			aPrecision = bPrecision
 * 			bPrecision = float(TP)/float(TP+FP)             # <<<<<<<<<<<<<<
 * 			aThreshold = bThreshold
 * 			bThreshold = e.score
 */
    __pyx_t_8 = __Pyx_PyObject_AsDouble(__pyx_v_TP); if (unlikely(__pyx_t_8 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L1_error)
    __pyx_t_1 = PyNumber_Add(__pyx_v_TP, __pyx_v_FP); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__pyx_t_7 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_t_7 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 298, __pyx_L1_error)
    }
    __pyx_v_bPrecision = (__pyx_t_8 / __pyx_t_7);

    /* "gnocis/models.pyx":299
 * 			aPrecision = bPrecision
 * 			bPrecision = float(TP)/float(TP+FP)
 * 			aThreshold = bThreshold             # <<<<<<<<<<<<<<
 * 			bThreshold = e.score
 * 			# This observation will be below the threshold in the next iteration,
 */
    __Pyx_INCREF(__pyx_v_bThreshold);
    __Pyx_DECREF_SET(__pyx_v_aThreshold, __pyx_v_bThreshold);

    /* "gnocis/models.pyx":300
 * 			bPrecision = float(TP)/float(TP+FP)
 * 			aThreshold = bThreshold
 * 			bThreshold = e.score             # <<<<<<<<<<<<<<
 * 			# This observation will be below the threshold in the next iteration,
 * 			# so remove count depending on whether it was a true or false positive
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_score); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_bThreshold, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":303
 * 			# This observation will be below the threshold in the next iteration,
 * 			# so remove count depending on whether it was a true or false positive
 * 			if bPrecision >= wantedPrecision:             # <<<<<<<<<<<<<<
 * 				break
 * 			if e.label:
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_bPrecision); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_v_wantedPrecision, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_9) {

      /* "gnocis/models.pyx":304
 * 			# so remove count depending on whether it was a true or false positive
 * 			if bPrecision >= wantedPrecision:
 * 				break             # <<<<<<<<<<<<<<
 * 			if e.label:
 * 				TP -= 1
 */
      goto __pyx_L16_break;

      /* "gnocis/models.pyx":303
 * 			# This observation will be below the threshold in the next iteration,
 * 			# so remove count depending on whether it was a true or false positive
 * 			if bPrecision >= wantedPrecision:             # <<<<<<<<<<<<<<
 * 				break
 * 			if e.label:
 */
    }

    /* "gnocis/models.pyx":305
 * 			if bPrecision >= wantedPrecision:
 * 				break
 * 			if e.label:             # <<<<<<<<<<<<<<
 * 				TP -= 1
 * 			else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_label); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_9) {

      /* "gnocis/models.pyx":306
 * 				break
 * 			if e.label:
 * 				TP -= 1             # <<<<<<<<<<<<<<
 * 			else:
 * 				FP -= 1
 */
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_TP, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_TP, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "gnocis/models.pyx":305
 * 			if bPrecision >= wantedPrecision:
 * 				break
 * 			if e.label:             # <<<<<<<<<<<<<<
 * 				TP -= 1
 * 			else:
 */
      goto __pyx_L18;
    }

    /* "gnocis/models.pyx":308
 * 				TP -= 1
 * 			else:
 * 				FP -= 1             # <<<<<<<<<<<<<<
 * 		if aPrecision != bPrecision:
 * 			wPrecision = bPrecision-aPrecision
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_FP, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_FP, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L18:;

    /* "gnocis/models.pyx":296
 * 		aThreshold, bThreshold = (0, 0)
 * 		aPrecision, bPrecision = (-1, -1)
 * 		for e in classScoresSorted:             # <<<<<<<<<<<<<<
 * 			aPrecision = bPrecision
 * 			bPrecision = float(TP)/float(TP+FP)
 */
  }
  __pyx_L16_break:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":309
 * 			else:
 * 				FP -= 1
 * 		if aPrecision != bPrecision:             # <<<<<<<<<<<<<<
 * 			wPrecision = bPrecision-aPrecision
 * 			wThreshold = bThreshold-aThreshold
 */
  __pyx_t_9 = ((__pyx_v_aPrecision != __pyx_v_bPrecision) != 0);
  if (__pyx_t_9) {

    /* "gnocis/models.pyx":310
 * 				FP -= 1
 * 		if aPrecision != bPrecision:
 * 			wPrecision = bPrecision-aPrecision             # <<<<<<<<<<<<<<
 * 			wThreshold = bThreshold-aThreshold
 * 			threshold = (wantedPrecision-aPrecision)*(wThreshold/wPrecision) + aThreshold
 */
    __pyx_v_wPrecision = (__pyx_v_bPrecision - __pyx_v_aPrecision);

    /* "gnocis/models.pyx":311
 * 		if aPrecision != bPrecision:
 * 			wPrecision = bPrecision-aPrecision
 * 			wThreshold = bThreshold-aThreshold             # <<<<<<<<<<<<<<
 * 			threshold = (wantedPrecision-aPrecision)*(wThreshold/wPrecision) + aThreshold
 * 		else:
 */
    __pyx_t_2 = PyNumber_Subtract(__pyx_v_bThreshold, __pyx_v_aThreshold); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_wThreshold = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":312
 * 			wPrecision = bPrecision-aPrecision
 * 			wThreshold = bThreshold-aThreshold
 * 			threshold = (wantedPrecision-aPrecision)*(wThreshold/wPrecision) + aThreshold             # <<<<<<<<<<<<<<
 * 		else:
 * 			threshold = aThreshold
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_aPrecision); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyNumber_Subtract(__pyx_v_wantedPrecision, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_wPrecision); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_wThreshold, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_aThreshold); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_threshold = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":309
 * 			else:
 * 				FP -= 1
 * 		if aPrecision != bPrecision:             # <<<<<<<<<<<<<<
 * 			wPrecision = bPrecision-aPrecision
 * 			wThreshold = bThreshold-aThreshold
 */
    goto __pyx_L19;
  }

  /* "gnocis/models.pyx":314
 * 			threshold = (wantedPrecision-aPrecision)*(wThreshold/wPrecision) + aThreshold
 * 		else:
 * 			threshold = aThreshold             # <<<<<<<<<<<<<<
 * 		return threshold
 * 
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_aThreshold);
    __pyx_v_threshold = __pyx_v_aThreshold;
  }
  __pyx_L19:;

  /* "gnocis/models.pyx":315
 * 		else:
 * 			threshold = aThreshold
 * 		return threshold             # <<<<<<<<<<<<<<
 * 
 * 	# Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_threshold);
  __pyx_r = __pyx_v_threshold;
  goto __pyx_L0;

  /* "gnocis/models.pyx":272
 * 
 * 	# Gets the threshold that gives a desired precision for two lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getPrecisionThreshold(self, positives, negatives, wantedPrecision):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value for a desired precision to a set of positive and a set of negative sequences. Linear interpolation is used in order to achieve a close approximation.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getPrecisionThreshold", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_positiveScores);
  __Pyx_XDECREF(__pyx_v_negativeScores);
  __Pyx_XDECREF(__pyx_v_classScoresSorted);
  __Pyx_XDECREF(__pyx_v_TP);
  __Pyx_XDECREF(__pyx_v_FP);
  __Pyx_XDECREF(__pyx_v_aThreshold);
  __Pyx_XDECREF(__pyx_v_bThreshold);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_wThreshold);
  __Pyx_XDECREF(__pyx_v_threshold);
  __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_score);
  __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_score);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":318
 * 
 * 	# Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 	def calibrateGenomewidePrecision(self, positives, genome, factor = 1.0, precision = 0.8, bgModelOrder = 4, bgModel = None):             # <<<<<<<<<<<<<<
 * 		""" Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_13calibrateGenomewidePrecision(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_12calibrateGenomewidePrecision[] = " Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.\n\t\t\n\t\t:param positives: Positive sequences.\n\t\t:param genome: The genome.\n\t\t:param factor: Scaling factor for positive set versus genome.\n\t\t:param precision: The precision to approximate.\n\t\t:type bgModelOrder: The order of the background model.\n\t\t:type bgModel: A background model to use. If specified, the `bgModelOrder` argument will be ignored, and this model will be used to generate the background. Otherwise, a background model is trained on the genome.\n\t\t\n\t\t:type positives: sequences/sequenceStream\n\t\t:type genome: sequences/sequenceStream/str\n\t\t:type factor: float, optional\n\t\t:type precision: float, optional\n\t\t:type bgModelOrder: float, optional\n\t\t:type bgModel: object, optional\n\t\t\n\t\t:return: List of the maximum window score per sequence\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_13calibrateGenomewidePrecision = {"calibrateGenomewidePrecision", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_13calibrateGenomewidePrecision, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_12calibrateGenomewidePrecision};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_13calibrateGenomewidePrecision(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_positives = 0;
  PyObject *__pyx_v_genome = 0;
  PyObject *__pyx_v_factor = 0;
  PyObject *__pyx_v_precision = 0;
  PyObject *__pyx_v_bgModelOrder = 0;
  PyObject *__pyx_v_bgModel = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calibrateGenomewidePrecision (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_positives,&__pyx_n_s_genome,&__pyx_n_s_factor,&__pyx_n_s_precision,&__pyx_n_s_bgModelOrder,&__pyx_n_s_bgModel,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[3] = ((PyObject *)((PyObject*)__pyx_float_1_0));
    values[4] = ((PyObject *)((PyObject*)__pyx_float_0_8));
    values[5] = ((PyObject *)((PyObject *)__pyx_int_4));
    values[6] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_positives)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calibrateGenomewidePrecision", 0, 3, 7, 1); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_genome)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calibrateGenomewidePrecision", 0, 3, 7, 2); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_factor);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_precision);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bgModelOrder);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bgModel);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calibrateGenomewidePrecision") < 0)) __PYX_ERR(0, 318, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_positives = values[1];
    __pyx_v_genome = values[2];
    __pyx_v_factor = values[3];
    __pyx_v_precision = values[4];
    __pyx_v_bgModelOrder = values[5];
    __pyx_v_bgModel = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calibrateGenomewidePrecision", 0, 3, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 318, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.calibrateGenomewidePrecision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_12calibrateGenomewidePrecision(__pyx_self, __pyx_v_self, __pyx_v_positives, __pyx_v_genome, __pyx_v_factor, __pyx_v_precision, __pyx_v_bgModelOrder, __pyx_v_bgModel);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_2generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":342
 * 			genome = getSequenceStreamFromPath(genome)
 * 		seqLens = genome.sequenceLengths()
 * 		genomeLen = sum(seqLens[chrom] for chrom in seqLens.keys())             # <<<<<<<<<<<<<<
 * 		# Train background model
 * 		if bgModel == None:
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_11_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_11_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 342, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_2generator6, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_sequenceModel_calibrateGenomewid, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.calibrateGenomewidePrecision.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_2generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens)) { __Pyx_RaiseClosureNameError("seqLens"); __PYX_ERR(0, 342, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(0, 342, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens, 0, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_chrom);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_chrom, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens)) { __Pyx_RaiseClosureNameError("seqLens"); __PYX_ERR(0, 342, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seqLens, __pyx_cur_scope->__pyx_v_chrom); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 342, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_5generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":349
 * 				else MarkovChain(trainingSequences = genome, degree = bgModelOrder)
 * 		# Generate calibration sequences
 * 		meanLen = int(sum(len(s) for s in positives) / len(positives))             # <<<<<<<<<<<<<<
 * 		negatives = bgModel.generateStream(n = int(factor * genomeLen / meanLen), length = meanLen)
 * 		# Calibrate threshold
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_12_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_12_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 349, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_5generator7, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_sequenceModel_calibrateGenomewid, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.calibrateGenomewidePrecision.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_5generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 349, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_positives)) { __Pyx_RaiseClosureNameError("positives"); __PYX_ERR(0, 349, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_positives)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_positives)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_positives; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_positives); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 349, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 349, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 349, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 349, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = PyObject_Length(__pyx_cur_scope->__pyx_v_s); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 349, __pyx_L1_error)
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 349, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":318
 * 
 * 	# Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 	def calibrateGenomewidePrecision(self, positives, genome, factor = 1.0, precision = 0.8, bgModelOrder = 4, bgModel = None):             # <<<<<<<<<<<<<<
 * 		""" Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_12calibrateGenomewidePrecision(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_positives, PyObject *__pyx_v_genome, PyObject *__pyx_v_factor, PyObject *__pyx_v_precision, PyObject *__pyx_v_bgModelOrder, PyObject *__pyx_v_bgModel) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *__pyx_cur_scope;
  PyObject *__pyx_v_genomeLen = NULL;
  PyObject *__pyx_v_meanLen = NULL;
  PyObject *__pyx_v_negatives = NULL;
  PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_2generator6 = 0;
  PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_5generator7 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calibrateGenomewidePrecision", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision(__pyx_ptype_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 318, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_positives = __pyx_v_positives;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_positives);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_positives);
  __Pyx_INCREF(__pyx_v_genome);
  __Pyx_INCREF(__pyx_v_bgModel);

  /* "gnocis/models.pyx":339
 * 		"""
 * 		# Find genome size.
 * 		if isinstance(genome, str):             # <<<<<<<<<<<<<<
 * 			genome = getSequenceStreamFromPath(genome)
 * 		seqLens = genome.sequenceLengths()
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_genome); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "gnocis/models.pyx":340
 * 		# Find genome size.
 * 		if isinstance(genome, str):
 * 			genome = getSequenceStreamFromPath(genome)             # <<<<<<<<<<<<<<
 * 		seqLens = genome.sequenceLengths()
 * 		genomeLen = sum(seqLens[chrom] for chrom in seqLens.keys())
 */
    __pyx_t_3 = __pyx_f_6gnocis_9sequences_getSequenceStreamFromPath(__pyx_v_genome, 0, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_genome, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "gnocis/models.pyx":339
 * 		"""
 * 		# Find genome size.
 * 		if isinstance(genome, str):             # <<<<<<<<<<<<<<
 * 			genome = getSequenceStreamFromPath(genome)
 * 		seqLens = genome.sequenceLengths()
 */
  }

  /* "gnocis/models.pyx":341
 * 		if isinstance(genome, str):
 * 			genome = getSequenceStreamFromPath(genome)
 * 		seqLens = genome.sequenceLengths()             # <<<<<<<<<<<<<<
 * 		genomeLen = sum(seqLens[chrom] for chrom in seqLens.keys())
 * 		# Train background model
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_genome, __pyx_n_s_sequenceLengths); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_seqLens = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":342
 * 			genome = getSequenceStreamFromPath(genome)
 * 		seqLens = genome.sequenceLengths()
 * 		genomeLen = sum(seqLens[chrom] for chrom in seqLens.keys())             # <<<<<<<<<<<<<<
 * 		# Train background model
 * 		if bgModel == None:
 */
  __pyx_t_3 = __pyx_pf_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_genomeLen = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "gnocis/models.pyx":344
 * 		genomeLen = sum(seqLens[chrom] for chrom in seqLens.keys())
 * 		# Train background model
 * 		if bgModel == None:             # <<<<<<<<<<<<<<
 * 			bgModel = IID(trainingSequences = genome) \
 * 				if bgModelOrder <= 0 \
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_bgModel, Py_None, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 344, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {

    /* "gnocis/models.pyx":346
 * 		if bgModel == None:
 * 			bgModel = IID(trainingSequences = genome) \
 * 				if bgModelOrder <= 0 \             # <<<<<<<<<<<<<<
 * 				else MarkovChain(trainingSequences = genome, degree = bgModelOrder)
 * 		# Generate calibration sequences
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_bgModelOrder, __pyx_int_0, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "gnocis/models.pyx":345
 * 		# Train background model
 * 		if bgModel == None:
 * 			bgModel = IID(trainingSequences = genome) \             # <<<<<<<<<<<<<<
 * 				if bgModelOrder <= 0 \
 * 				else MarkovChain(trainingSequences = genome, degree = bgModelOrder)
 */
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_trainingSequences, __pyx_v_genome) < 0) __PYX_ERR(0, 345, __pyx_L1_error)
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_IID), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_4 = __pyx_t_5;
      __pyx_t_5 = 0;
    } else {

      /* "gnocis/models.pyx":347
 * 			bgModel = IID(trainingSequences = genome) \
 * 				if bgModelOrder <= 0 \
 * 				else MarkovChain(trainingSequences = genome, degree = bgModelOrder)             # <<<<<<<<<<<<<<
 * 		# Generate calibration sequences
 * 		meanLen = int(sum(len(s) for s in positives) / len(positives))
 */
      __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_trainingSequences, __pyx_v_genome) < 0) __PYX_ERR(0, 347, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_degree, __pyx_v_bgModelOrder) < 0) __PYX_ERR(0, 347, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_MarkovChain), __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_4 = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_bgModel, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "gnocis/models.pyx":344
 * 		genomeLen = sum(seqLens[chrom] for chrom in seqLens.keys())
 * 		# Train background model
 * 		if bgModel == None:             # <<<<<<<<<<<<<<
 * 			bgModel = IID(trainingSequences = genome) \
 * 				if bgModelOrder <= 0 \
 */
  }

  /* "gnocis/models.pyx":349
 * 				else MarkovChain(trainingSequences = genome, degree = bgModelOrder)
 * 		# Generate calibration sequences
 * 		meanLen = int(sum(len(s) for s in positives) / len(positives))             # <<<<<<<<<<<<<<
 * 		negatives = bgModel.generateStream(n = int(factor * genomeLen / meanLen), length = meanLen)
 * 		# Calibrate threshold
 */
  __pyx_t_4 = __pyx_pf_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_cur_scope->__pyx_v_positives;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_6 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_meanLen = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "gnocis/models.pyx":350
 * 		# Generate calibration sequences
 * 		meanLen = int(sum(len(s) for s in positives) / len(positives))
 * 		negatives = bgModel.generateStream(n = int(factor * genomeLen / meanLen), length = meanLen)             # <<<<<<<<<<<<<<
 * 		# Calibrate threshold
 * 		self.threshold = self.getPrecisionThreshold(positives = positives, negatives = negatives, wantedPrecision = precision)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_bgModel, __pyx_n_s_generateStream); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_factor, __pyx_v_genomeLen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_v_meanLen); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_n, __pyx_t_3) < 0) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_length, __pyx_v_meanLen) < 0) __PYX_ERR(0, 350, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_negatives = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":352
 * 		negatives = bgModel.generateStream(n = int(factor * genomeLen / meanLen), length = meanLen)
 * 		# Calibrate threshold
 * 		self.threshold = self.getPrecisionThreshold(positives = positives, negatives = negatives, wantedPrecision = precision)             # <<<<<<<<<<<<<<
 * 		return self
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getPrecisionThreshold); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_positives, __pyx_cur_scope->__pyx_v_positives) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_negatives, __pyx_v_negatives) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_wantedPrecision, __pyx_v_precision) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_threshold, __pyx_t_4) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gnocis/models.pyx":353
 * 		# Calibrate threshold
 * 		self.threshold = self.getPrecisionThreshold(positives = positives, negatives = negatives, wantedPrecision = precision)
 * 		return self             # <<<<<<<<<<<<<<
 * 
 * 	# Applies model to a sequence with a sliding window, and returns the maximum score.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "gnocis/models.pyx":318
 * 
 * 	# Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 	def calibrateGenomewidePrecision(self, positives, genome, factor = 1.0, precision = 0.8, bgModelOrder = 4, bgModel = None):             # <<<<<<<<<<<<<<
 * 		""" Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.calibrateGenomewidePrecision", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_genomeLen);
  __Pyx_XDECREF(__pyx_v_meanLen);
  __Pyx_XDECREF(__pyx_v_negatives);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_2generator6);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_13sequenceModel_28calibrateGenomewidePrecision_5generator7);
  __Pyx_XDECREF(__pyx_v_genome);
  __Pyx_XDECREF(__pyx_v_bgModel);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":356
 * 
 * 	# Applies model to a sequence with a sliding window, and returns the maximum score.
 * 	def scoreSequence(self, seq):             # <<<<<<<<<<<<<<
 * 		""" Scores a single sequence. The score is determined by applying the model with a sliding window, and taking the maximum score.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_15scoreSequence(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_14scoreSequence[] = " Scores a single sequence. The score is determined by applying the model with a sliding window, and taking the maximum score.\n\t\t\n\t\t:param seq: The sequence.\n\t\t\n\t\t:type seq: sequences/sequenceStream\n\t\t\n\t\t:return: Maximal window score\n\t\t:rtype: float\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_15scoreSequence = {"scoreSequence", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_15scoreSequence, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_14scoreSequence};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_15scoreSequence(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seq = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scoreSequence (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seq,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scoreSequence", 1, 2, 2, 1); __PYX_ERR(0, 356, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scoreSequence") < 0)) __PYX_ERR(0, 356, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_seq = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scoreSequence", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 356, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.scoreSequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_14scoreSequence(__pyx_self, __pyx_v_self, __pyx_v_seq);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_13scoreSequence_2generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":369
 * 		score = max(
 * 			(
 * 				self.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(self.windowSize, self.windowStep)
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_13scoreSequence_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_14_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_14_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 369, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_13sequenceModel_13scoreSequence_2generator8, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_sequenceModel_scoreSequence_loca, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 369, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.scoreSequence.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_13scoreSequence_2generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 369, __pyx_L1_error)

  /* "gnocis/models.pyx":371
 * 				self.scoreWindow(win)
 * 				for win in
 * 				seq.windows(self.windowSize, self.windowStep)             # <<<<<<<<<<<<<<
 * 				if len(win) == self.windowSize
 * 			),
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seq)) { __Pyx_RaiseClosureNameError("seq"); __PYX_ERR(0, 371, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_seq, __pyx_n_s_windows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 371, __pyx_L1_error) }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_windowSize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 371, __pyx_L1_error) }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_windowStep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 371, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 371, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 371, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 371, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_win);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_win, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":372
 * 				for win in
 * 				seq.windows(self.windowSize, self.windowStep)
 * 				if len(win) == self.windowSize             # <<<<<<<<<<<<<<
 * 			),
 * 			default = float('-INF')
 */
    __pyx_t_10 = PyObject_Length(__pyx_cur_scope->__pyx_v_win); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 372, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 372, __pyx_L1_error) }
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_windowSize); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_11) {

      /* "gnocis/models.pyx":369
 * 		score = max(
 * 			(
 * 				self.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(self.windowSize, self.windowStep)
 */
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 369, __pyx_L1_error) }
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_scoreWindow); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_1, __pyx_cur_scope->__pyx_v_win) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_cur_scope->__pyx_v_win);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      __Pyx_XGIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_8;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_9;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_8 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_9 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 369, __pyx_L1_error)

      /* "gnocis/models.pyx":372
 * 				for win in
 * 				seq.windows(self.windowSize, self.windowStep)
 * 				if len(win) == self.windowSize             # <<<<<<<<<<<<<<
 * 			),
 * 			default = float('-INF')
 */
    }

    /* "gnocis/models.pyx":370
 * 			(
 * 				self.scoreWindow(win)
 * 				for win in             # <<<<<<<<<<<<<<
 * 				seq.windows(self.windowSize, self.windowStep)
 * 				if len(win) == self.windowSize
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "gnocis/models.pyx":369
 * 		score = max(
 * 			(
 * 				self.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(self.windowSize, self.windowStep)
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":356
 * 
 * 	# Applies model to a sequence with a sliding window, and returns the maximum score.
 * 	def scoreSequence(self, seq):             # <<<<<<<<<<<<<<
 * 		""" Scores a single sequence. The score is determined by applying the model with a sliding window, and taking the maximum score.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_14scoreSequence(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seq) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *__pyx_cur_scope;
  float __pyx_v_score;
  PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_13scoreSequence_2generator8 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  float __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("scoreSequence", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_13_scoreSequence(__pyx_ptype_6gnocis_6models___pyx_scope_struct_13_scoreSequence, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 356, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_seq = __pyx_v_seq;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_seq);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_seq);

  /* "gnocis/models.pyx":369
 * 		score = max(
 * 			(
 * 				self.scoreWindow(win)             # <<<<<<<<<<<<<<
 * 				for win in
 * 				seq.windows(self.windowSize, self.windowStep)
 */
  __pyx_t_1 = __pyx_pf_6gnocis_6models_13sequenceModel_13scoreSequence_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gnocis/models.pyx":367
 * 		"""
 * 		cdef float score
 * 		score = max(             # <<<<<<<<<<<<<<
 * 			(
 * 				self.scoreWindow(win)
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":374
 * 				if len(win) == self.windowSize
 * 			),
 * 			default = float('-INF')             # <<<<<<<<<<<<<<
 * 		)
 * 		return score
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_kp_u_INF); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_default, __pyx_t_3) < 0) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":367
 * 		"""
 * 		cdef float score
 * 		score = max(             # <<<<<<<<<<<<<<
 * 			(
 * 				self.scoreWindow(win)
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_score = __pyx_t_4;

  /* "gnocis/models.pyx":376
 * 			default = float('-INF')
 * 		)
 * 		return score             # <<<<<<<<<<<<<<
 * 
 * 	# Gets and returns a confusion matrix.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_score); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":356
 * 
 * 	# Applies model to a sequence with a sliding window, and returns the maximum score.
 * 	def scoreSequence(self, seq):             # <<<<<<<<<<<<<<
 * 		""" Scores a single sequence. The score is determined by applying the model with a sliding window, and taking the maximum score.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.scoreSequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_13sequenceModel_13scoreSequence_2generator8);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":379
 * 
 * 	# Gets and returns a confusion matrix.
 * 	def getConfusionMatrix(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a confusion matrix for sets of positive and negative sequences.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_36__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_17getConfusionMatrix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_16getConfusionMatrix[] = " Calculates and returns a confusion matrix for sets of positive and negative sequences.\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t\n\t\t:return: Confusion matrix\n\t\t:rtype: dict\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_17getConfusionMatrix = {"getConfusionMatrix", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_17getConfusionMatrix, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_16getConfusionMatrix};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_17getConfusionMatrix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConfusionMatrix (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults1 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getConfusionMatrix", 0, 2, 4, 1); __PYX_ERR(0, 379, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getConfusionMatrix") < 0)) __PYX_ERR(0, 379, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getConfusionMatrix", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 379, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getConfusionMatrix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_16getConfusionMatrix(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_16getConfusionMatrix(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_v_positives = NULL;
  PyObject *__pyx_v_negatives = NULL;
  PyObject *__pyx_v_vPos = NULL;
  PyObject *__pyx_v_vNeg = NULL;
  PyObject *__pyx_9genexpr25__pyx_v_score = NULL;
  PyObject *__pyx_9genexpr26__pyx_v_score = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getConfusionMatrix", 0);

  /* "gnocis/models.pyx":393
 * 		:rtype: dict
 * 		"""
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])             # <<<<<<<<<<<<<<
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_seqs, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_labelPositive);
  __Pyx_GIVEREF(__pyx_v_labelPositive);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_labelPositive);
  __Pyx_INCREF(__pyx_v_labelNegative);
  __Pyx_GIVEREF(__pyx_v_labelNegative);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_v_labelNegative);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 393, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 393, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 393, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_positives = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_negatives = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":394
 * 		"""
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]             # <<<<<<<<<<<<<<
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		return getConfusionMatrix(vPos, vNeg, threshold = self.threshold)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_positives) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_positives);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 394, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 394, __pyx_L7_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 394, __pyx_L7_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 394, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 394, __pyx_L7_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 394, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 394, __pyx_L7_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr25__pyx_v_score, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 394, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_score, __pyx_9genexpr25__pyx_v_score) < 0) __PYX_ERR(0, 394, __pyx_L7_error)
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_label, Py_True) < 0) __PYX_ERR(0, 394, __pyx_L7_error)
      __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 394, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 394, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_score); __pyx_9genexpr25__pyx_v_score = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_score); __pyx_9genexpr25__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_vPos = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":395
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]             # <<<<<<<<<<<<<<
 * 		return getConfusionMatrix(vPos, vNeg, threshold = self.threshold)
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 395, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_negatives) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_negatives);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 395, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 395, __pyx_L13_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 395, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 395, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 395, __pyx_L13_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr26__pyx_v_score, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_score, __pyx_9genexpr26__pyx_v_score) < 0) __PYX_ERR(0, 395, __pyx_L13_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_label, Py_False) < 0) __PYX_ERR(0, 395, __pyx_L13_error)
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 395, __pyx_L13_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_score); __pyx_9genexpr26__pyx_v_score = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_score); __pyx_9genexpr26__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  __pyx_v_vNeg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":396
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		return getConfusionMatrix(vPos, vNeg, threshold = self.threshold)             # <<<<<<<<<<<<<<
 * 
 * 	# Generates a Receiver Operating Characteristic curve.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_getConfusionMatrix); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_vPos);
  __Pyx_GIVEREF(__pyx_v_vPos);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_vPos);
  __Pyx_INCREF(__pyx_v_vNeg);
  __Pyx_GIVEREF(__pyx_v_vNeg);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_vNeg);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_threshold); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_threshold, __pyx_t_2) < 0) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":379
 * 
 * 	# Gets and returns a confusion matrix.
 * 	def getConfusionMatrix(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a confusion matrix for sets of positive and negative sequences.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getConfusionMatrix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_positives);
  __Pyx_XDECREF(__pyx_v_negatives);
  __Pyx_XDECREF(__pyx_v_vPos);
  __Pyx_XDECREF(__pyx_v_vNeg);
  __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_score);
  __Pyx_XDECREF(__pyx_9genexpr26__pyx_v_score);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":399
 * 
 * 	# Generates a Receiver Operating Characteristic curve.
 * 	def getROC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a Receiver Operating Characteristic (ROC) curve for sets of positive and negative sequences.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_38__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_19getROC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_18getROC[] = " Calculates and returns a Receiver Operating Characteristic (ROC) curve for sets of positive and negative sequences.\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t\n\t\t:return: Receiver Operating Characteristic (ROC) curve\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_19getROC = {"getROC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_19getROC, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_18getROC};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_19getROC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getROC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults2 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getROC", 0, 2, 4, 1); __PYX_ERR(0, 399, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getROC") < 0)) __PYX_ERR(0, 399, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getROC", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 399, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getROC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_18getROC(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_18getROC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_v_positives = NULL;
  PyObject *__pyx_v_negatives = NULL;
  PyObject *__pyx_v_vPos = NULL;
  PyObject *__pyx_v_vNeg = NULL;
  PyObject *__pyx_9genexpr27__pyx_v_score = NULL;
  PyObject *__pyx_9genexpr28__pyx_v_score = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getROC", 0);

  /* "gnocis/models.pyx":413
 * 		:rtype: list
 * 		"""
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])             # <<<<<<<<<<<<<<
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_seqs, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_labelPositive);
  __Pyx_GIVEREF(__pyx_v_labelPositive);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_labelPositive);
  __Pyx_INCREF(__pyx_v_labelNegative);
  __Pyx_GIVEREF(__pyx_v_labelNegative);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_v_labelNegative);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 413, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 413, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 413, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_positives = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_negatives = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":414
 * 		"""
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]             # <<<<<<<<<<<<<<
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		return getROC(vPos, vNeg)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_positives) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_positives);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 414, __pyx_L7_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 414, __pyx_L7_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 414, __pyx_L7_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 414, __pyx_L7_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr27__pyx_v_score, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_score, __pyx_9genexpr27__pyx_v_score) < 0) __PYX_ERR(0, 414, __pyx_L7_error)
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_label, Py_True) < 0) __PYX_ERR(0, 414, __pyx_L7_error)
      __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 414, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 414, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_score); __pyx_9genexpr27__pyx_v_score = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_score); __pyx_9genexpr27__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_vPos = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":415
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]             # <<<<<<<<<<<<<<
 * 		return getROC(vPos, vNeg)
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 415, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_negatives) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_negatives);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 415, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 415, __pyx_L13_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 415, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 415, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 415, __pyx_L13_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr28__pyx_v_score, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_score, __pyx_9genexpr28__pyx_v_score) < 0) __PYX_ERR(0, 415, __pyx_L13_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_label, Py_False) < 0) __PYX_ERR(0, 415, __pyx_L13_error)
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 415, __pyx_L13_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_score); __pyx_9genexpr28__pyx_v_score = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_score); __pyx_9genexpr28__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  __pyx_v_vNeg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":416
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		return getROC(vPos, vNeg)             # <<<<<<<<<<<<<<
 * 
 * 	# Calculates the area under a Receiver Operating Characteristic curve.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_getROC); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_vPos, __pyx_v_vNeg};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_vPos, __pyx_v_vNeg};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_vPos);
    __Pyx_GIVEREF(__pyx_v_vPos);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_8, __pyx_v_vPos);
    __Pyx_INCREF(__pyx_v_vNeg);
    __Pyx_GIVEREF(__pyx_v_vNeg);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_8, __pyx_v_vNeg);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":399
 * 
 * 	# Generates a Receiver Operating Characteristic curve.
 * 	def getROC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a Receiver Operating Characteristic (ROC) curve for sets of positive and negative sequences.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getROC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_positives);
  __Pyx_XDECREF(__pyx_v_negatives);
  __Pyx_XDECREF(__pyx_v_vPos);
  __Pyx_XDECREF(__pyx_v_vNeg);
  __Pyx_XDECREF(__pyx_9genexpr27__pyx_v_score);
  __Pyx_XDECREF(__pyx_9genexpr28__pyx_v_score);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":419
 * 
 * 	# Calculates the area under a Receiver Operating Characteristic curve.
 * 	def getROCAUC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns the area under a Receiver Operating Characteristic (ROCAUC) curve for sets of positive and negative sequences.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_40__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_21getROCAUC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_20getROCAUC[] = " Calculates and returns the area under a Receiver Operating Characteristic (ROCAUC) curve for sets of positive and negative sequences.\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t\n\t\t:return: Area under the Receiver Operating Characteristic (ROCAUC) curve\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_21getROCAUC = {"getROCAUC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_21getROCAUC, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_20getROCAUC};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_21getROCAUC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getROCAUC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults3 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getROCAUC", 0, 2, 4, 1); __PYX_ERR(0, 419, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getROCAUC") < 0)) __PYX_ERR(0, 419, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getROCAUC", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 419, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getROCAUC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_20getROCAUC(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_20getROCAUC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getROCAUC", 0);

  /* "gnocis/models.pyx":433
 * 		:rtype: list
 * 		"""
 * 		return getAUC(self.getROC(seqs, labelPositive = labelPositive, labelNegative = labelNegative))             # <<<<<<<<<<<<<<
 * 
 * 	# Generates a Precision/Recall curve.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getROC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_seqs);
  __Pyx_GIVEREF(__pyx_v_seqs);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_seqs);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 433, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 433, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":419
 * 
 * 	# Calculates the area under a Receiver Operating Characteristic curve.
 * 	def getROCAUC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns the area under a Receiver Operating Characteristic (ROCAUC) curve for sets of positive and negative sequences.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getROCAUC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":436
 * 
 * 	# Generates a Precision/Recall curve.
 * 	def getPRC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a Precision/Recall curve (PRC) for sets of positive and negative sequences.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_42__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_23getPRC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_22getPRC[] = " Calculates and returns a Precision/Recall curve (PRC) for sets of positive and negative sequences.\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t\n\t\t:return: Precision/Recall curve (PRC)\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_23getPRC = {"getPRC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_23getPRC, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_22getPRC};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_23getPRC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getPRC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults4 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getPRC", 0, 2, 4, 1); __PYX_ERR(0, 436, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getPRC") < 0)) __PYX_ERR(0, 436, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getPRC", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 436, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_22getPRC(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_22getPRC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_v_positives = NULL;
  PyObject *__pyx_v_negatives = NULL;
  PyObject *__pyx_v_vPos = NULL;
  PyObject *__pyx_v_vNeg = NULL;
  PyObject *__pyx_9genexpr29__pyx_v_score = NULL;
  PyObject *__pyx_9genexpr30__pyx_v_score = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getPRC", 0);

  /* "gnocis/models.pyx":450
 * 		:rtype: list
 * 		"""
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])             # <<<<<<<<<<<<<<
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_seqs, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_labelPositive);
  __Pyx_GIVEREF(__pyx_v_labelPositive);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_labelPositive);
  __Pyx_INCREF(__pyx_v_labelNegative);
  __Pyx_GIVEREF(__pyx_v_labelNegative);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_v_labelNegative);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 450, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 450, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 450, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_positives = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_negatives = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":451
 * 		"""
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]             # <<<<<<<<<<<<<<
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		return getPRC(vPos, vNeg)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_positives) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_positives);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 451, __pyx_L7_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 451, __pyx_L7_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 451, __pyx_L7_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 451, __pyx_L7_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr29__pyx_v_score, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_score, __pyx_9genexpr29__pyx_v_score) < 0) __PYX_ERR(0, 451, __pyx_L7_error)
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_label, Py_True) < 0) __PYX_ERR(0, 451, __pyx_L7_error)
      __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 451, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr29__pyx_v_score); __pyx_9genexpr29__pyx_v_score = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_9genexpr29__pyx_v_score); __pyx_9genexpr29__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_vPos = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":452
 * 		positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]             # <<<<<<<<<<<<<<
 * 		return getPRC(vPos, vNeg)
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 452, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_negatives) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_negatives);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 452, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 452, __pyx_L13_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 452, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 452, __pyx_L13_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 452, __pyx_L13_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr30__pyx_v_score, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_score, __pyx_9genexpr30__pyx_v_score) < 0) __PYX_ERR(0, 452, __pyx_L13_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_label, Py_False) < 0) __PYX_ERR(0, 452, __pyx_L13_error)
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_validationPair), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 452, __pyx_L13_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr30__pyx_v_score); __pyx_9genexpr30__pyx_v_score = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_9genexpr30__pyx_v_score); __pyx_9genexpr30__pyx_v_score = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  __pyx_v_vNeg = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":453
 * 		vPos = [ validationPair(score = score, label = True) for score in self.getSequenceScores(positives) ]
 * 		vNeg = [ validationPair(score = score, label = False) for score in self.getSequenceScores(negatives) ]
 * 		return getPRC(vPos, vNeg)             # <<<<<<<<<<<<<<
 * 
 * 	# Calculates the area under a Precision/Recall Curve.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_getPRC); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_vPos, __pyx_v_vNeg};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_vPos, __pyx_v_vNeg};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_vPos);
    __Pyx_GIVEREF(__pyx_v_vPos);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_8, __pyx_v_vPos);
    __Pyx_INCREF(__pyx_v_vNeg);
    __Pyx_GIVEREF(__pyx_v_vNeg);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_8, __pyx_v_vNeg);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":436
 * 
 * 	# Generates a Precision/Recall curve.
 * 	def getPRC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a Precision/Recall curve (PRC) for sets of positive and negative sequences.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_positives);
  __Pyx_XDECREF(__pyx_v_negatives);
  __Pyx_XDECREF(__pyx_v_vPos);
  __Pyx_XDECREF(__pyx_v_vNeg);
  __Pyx_XDECREF(__pyx_9genexpr29__pyx_v_score);
  __Pyx_XDECREF(__pyx_9genexpr30__pyx_v_score);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":456
 * 
 * 	# Calculates the area under a Precision/Recall Curve.
 * 	def getPRCAUC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns the area under a Precision/Recall curve (PRCAUC) for sets of positive and negative sequences.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_44__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_25getPRCAUC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_24getPRCAUC[] = " Calculates and returns the area under a Precision/Recall curve (PRCAUC) for sets of positive and negative sequences.\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t\n\t\t:return: Area under the Precision/Recall curve (PRCAUC)\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_25getPRCAUC = {"getPRCAUC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_25getPRCAUC, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_24getPRCAUC};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_25getPRCAUC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getPRCAUC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults5 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getPRCAUC", 0, 2, 4, 1); __PYX_ERR(0, 456, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getPRCAUC") < 0)) __PYX_ERR(0, 456, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getPRCAUC", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 456, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getPRCAUC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_24getPRCAUC(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_24getPRCAUC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getPRCAUC", 0);

  /* "gnocis/models.pyx":470
 * 		:rtype: list
 * 		"""
 * 		return getAUC(self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative))             # <<<<<<<<<<<<<<
 * 
 * 	def plotPRC(self, seqs, labelPositive = positive, labelNegative = negative, figsize = (8, 8), outpath = None, style = 'ggplot'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getPRC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_seqs);
  __Pyx_GIVEREF(__pyx_v_seqs);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_seqs);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 470, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 470, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":456
 * 
 * 	# Calculates the area under a Precision/Recall Curve.
 * 	def getPRCAUC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns the area under a Precision/Recall curve (PRCAUC) for sets of positive and negative sequences.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getPRCAUC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":472
 * 		return getAUC(self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative))
 * 
 * 	def plotPRC(self, seqs, labelPositive = positive, labelNegative = negative, figsize = (8, 8), outpath = None, style = 'ggplot'):             # <<<<<<<<<<<<<<
 * 		""" Plots a Precision/Recall curve and either displays it in an IPython session or saves it to a file.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_46__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_INCREF(((PyObject*)__pyx_tuple__4));
  __Pyx_GIVEREF(((PyObject*)__pyx_tuple__4));
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject*)__pyx_tuple__4));
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 3, ((PyObject *)Py_None));
  __Pyx_INCREF(((PyObject*)__pyx_n_u_ggplot));
  __Pyx_GIVEREF(((PyObject*)__pyx_n_u_ggplot));
  PyTuple_SET_ITEM(__pyx_t_1, 4, ((PyObject*)__pyx_n_u_ggplot));
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_27plotPRC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_26plotPRC[] = " Plots a Precision/Recall curve and either displays it in an IPython session or saves it to a file.\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t:param figsize: Tuple of figure dimensions.\n\t\t:param outpath: Path to save generated plot to. If not set, the plot will be output to IPython.\n\t\t:param style: Matplotlib style to use.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t:type figsize: tuple, optional\n\t\t:type outpath: str, optional\n\t\t:type style: str, optional\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_27plotPRC = {"plotPRC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_27plotPRC, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_26plotPRC};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_27plotPRC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_style = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("plotPRC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_style,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    __pyx_defaults6 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    values[4] = ((PyObject *)((PyObject*)__pyx_tuple__4));
    values[5] = ((PyObject *)((PyObject *)Py_None));
    values[6] = ((PyObject *)((PyObject*)__pyx_n_u_ggplot));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("plotPRC", 0, 2, 7, 1); __PYX_ERR(0, 472, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "plotPRC") < 0)) __PYX_ERR(0, 472, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
    __pyx_v_figsize = values[4];
    __pyx_v_outpath = values[5];
    __pyx_v_style = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("plotPRC", 0, 2, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 472, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.plotPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_26plotPRC(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_style);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_26plotPRC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_style) {
  PyObject *__pyx_v_plt = NULL;
  PyObject *__pyx_v_base64 = NULL;
  PyObject *__pyx_v_BytesIO = NULL;
  PyObject *__pyx_v_display = NULL;
  PyObject *__pyx_v_HTML = NULL;
  PyObject *__pyx_v_fig = NULL;
  PyObject *__pyx_v_positives = NULL;
  PyObject *__pyx_v_negatives = NULL;
  PyObject *__pyx_v_curve = NULL;
  PyObject *__pyx_v_bio = NULL;
  PyObject *__pyx_v_encoded = NULL;
  PyObject *__pyx_v_html = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_9genexpr31__pyx_v_v = NULL;
  PyObject *__pyx_9genexpr32__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *(*__pyx_t_18)(PyObject *);
  Py_UCS4 __pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  int __pyx_t_23;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_t_25;
  int __pyx_t_26;
  char const *__pyx_t_27;
  PyObject *__pyx_t_28 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("plotPRC", 0);

  /* "gnocis/models.pyx":489
 * 		:type style: str, optional
 * 		"""
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "gnocis/models.pyx":490
 * 		"""
 * 		try:
 * 			import matplotlib.pyplot as plt             # <<<<<<<<<<<<<<
 * 			import base64
 * 			from io import BytesIO
 */
      __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_n_s__5);
      __Pyx_GIVEREF(__pyx_n_s__5);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s__5);
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_matplotlib_pyplot, __pyx_t_4, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 490, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_plt = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "gnocis/models.pyx":491
 * 		try:
 * 			import matplotlib.pyplot as plt
 * 			import base64             # <<<<<<<<<<<<<<
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 */
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_base64, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 491, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_base64 = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "gnocis/models.pyx":492
 * 			import matplotlib.pyplot as plt
 * 			import base64
 * 			from io import BytesIO             # <<<<<<<<<<<<<<
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):
 */
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 492, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_n_s_BytesIO);
      __Pyx_GIVEREF(__pyx_n_s_BytesIO);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_BytesIO);
      __pyx_t_4 = __Pyx_Import(__pyx_n_s_io, __pyx_t_5, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 492, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_BytesIO); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 492, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_BytesIO = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "gnocis/models.pyx":493
 * 			import base64
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML             # <<<<<<<<<<<<<<
 * 			with plt.style.context(style):
 * 				fig = plt.figure(figsize = figsize)
 */
      __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_n_s_display);
      __Pyx_GIVEREF(__pyx_n_s_display);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_display);
      __Pyx_INCREF(__pyx_n_s_HTML);
      __Pyx_GIVEREF(__pyx_n_s_HTML);
      PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_HTML);
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_IPython_core_display, __pyx_t_4, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_display); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_display = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_HTML); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_HTML = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "gnocis/models.pyx":494
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 				fig = plt.figure(figsize = figsize)
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 */
      /*with:*/ {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_style); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 494, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_context); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 494, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_style) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_style);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 494, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 494, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 494, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 494, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_9);
            __Pyx_XGOTREF(__pyx_t_10);
            __Pyx_XGOTREF(__pyx_t_11);
            /*try:*/ {

              /* "gnocis/models.pyx":495
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):
 * 				fig = plt.figure(figsize = figsize)             # <<<<<<<<<<<<<<
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 				curve = self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative)
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_figure); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 495, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 495, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 495, __pyx_L13_error)
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 495, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_v_fig = __pyx_t_4;
              __pyx_t_4 = 0;

              /* "gnocis/models.pyx":496
 * 			with plt.style.context(style):
 * 				fig = plt.figure(figsize = figsize)
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])             # <<<<<<<<<<<<<<
 * 				curve = self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative)
 * 				plt.plot([ 0., 1. ],
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_seqs, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 496, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 496, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_INCREF(__pyx_v_labelPositive);
              __Pyx_GIVEREF(__pyx_v_labelPositive);
              PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_labelPositive);
              __Pyx_INCREF(__pyx_v_labelNegative);
              __Pyx_GIVEREF(__pyx_v_labelNegative);
              PyList_SET_ITEM(__pyx_t_5, 1, __pyx_v_labelNegative);
              __pyx_t_8 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_8)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_8);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 496, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
                PyObject* sequence = __pyx_t_4;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 496, __pyx_L13_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
                  __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
                } else {
                  __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
                  __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
                }
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_5);
                #else
                __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 496, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 496, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                #endif
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else {
                Py_ssize_t index = -1;
                __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 496, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_12 = Py_TYPE(__pyx_t_8)->tp_iternext;
                index = 0; __pyx_t_6 = __pyx_t_12(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L19_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_6);
                index = 1; __pyx_t_5 = __pyx_t_12(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L19_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_5);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_8), 2) < 0) __PYX_ERR(0, 496, __pyx_L13_error)
                __pyx_t_12 = NULL;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                goto __pyx_L20_unpacking_done;
                __pyx_L19_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __pyx_t_12 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 496, __pyx_L13_error)
                __pyx_L20_unpacking_done:;
              }
              __pyx_v_positives = __pyx_t_6;
              __pyx_t_6 = 0;
              __pyx_v_negatives = __pyx_t_5;
              __pyx_t_5 = 0;

              /* "gnocis/models.pyx":497
 * 				fig = plt.figure(figsize = figsize)
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 				curve = self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative)             # <<<<<<<<<<<<<<
 * 				plt.plot([ 0., 1. ],
 * 					[ len(positives) / (len(positives) + len(negatives)),
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getPRC); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 497, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 497, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_INCREF(__pyx_v_seqs);
              __Pyx_GIVEREF(__pyx_v_seqs);
              PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_seqs);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 497, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 497, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 497, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 497, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_v_curve = __pyx_t_8;
              __pyx_t_8 = 0;

              /* "gnocis/models.pyx":498
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 				curve = self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative)
 * 				plt.plot([ 0., 1. ],             # <<<<<<<<<<<<<<
 * 					[ len(positives) / (len(positives) + len(negatives)),
 * 						len(positives) / (len(positives) + len(negatives))],
 */
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_plot); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 498, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_6 = PyList_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 498, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_INCREF(__pyx_float_0_);
              __Pyx_GIVEREF(__pyx_float_0_);
              PyList_SET_ITEM(__pyx_t_6, 0, __pyx_float_0_);
              __Pyx_INCREF(__pyx_float_1_);
              __Pyx_GIVEREF(__pyx_float_1_);
              PyList_SET_ITEM(__pyx_t_6, 1, __pyx_float_1_);

              /* "gnocis/models.pyx":499
 * 				curve = self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative)
 * 				plt.plot([ 0., 1. ],
 * 					[ len(positives) / (len(positives) + len(negatives)),             # <<<<<<<<<<<<<<
 * 						len(positives) / (len(positives) + len(negatives))],
 * 					linestyle = '--',
 */
              __pyx_t_13 = PyObject_Length(__pyx_v_positives); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 499, __pyx_L13_error)
              __pyx_t_14 = PyObject_Length(__pyx_v_positives); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 499, __pyx_L13_error)
              __pyx_t_15 = PyObject_Length(__pyx_v_negatives); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 499, __pyx_L13_error)
              __pyx_t_16 = (__pyx_t_14 + __pyx_t_15);
              if (unlikely(__pyx_t_16 == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 499, __pyx_L13_error)
              }
              __pyx_t_5 = PyFloat_FromDouble((((double)__pyx_t_13) / ((double)__pyx_t_16))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);

              /* "gnocis/models.pyx":500
 * 				plt.plot([ 0., 1. ],
 * 					[ len(positives) / (len(positives) + len(negatives)),
 * 						len(positives) / (len(positives) + len(negatives))],             # <<<<<<<<<<<<<<
 * 					linestyle = '--',
 * 					label = 'Expected at random')
 */
              __pyx_t_16 = PyObject_Length(__pyx_v_positives); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 500, __pyx_L13_error)
              __pyx_t_13 = PyObject_Length(__pyx_v_positives); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 500, __pyx_L13_error)
              __pyx_t_15 = PyObject_Length(__pyx_v_negatives); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 500, __pyx_L13_error)
              __pyx_t_14 = (__pyx_t_13 + __pyx_t_15);
              if (unlikely(__pyx_t_14 == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 500, __pyx_L13_error)
              }
              __pyx_t_4 = PyFloat_FromDouble((((double)__pyx_t_16) / ((double)__pyx_t_14))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 500, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);

              /* "gnocis/models.pyx":499
 * 				curve = self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative)
 * 				plt.plot([ 0., 1. ],
 * 					[ len(positives) / (len(positives) + len(negatives)),             # <<<<<<<<<<<<<<
 * 						len(positives) / (len(positives) + len(negatives))],
 * 					linestyle = '--',
 */
              __pyx_t_17 = PyList_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 499, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_GIVEREF(__pyx_t_5);
              PyList_SET_ITEM(__pyx_t_17, 0, __pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_4);
              PyList_SET_ITEM(__pyx_t_17, 1, __pyx_t_4);
              __pyx_t_5 = 0;
              __pyx_t_4 = 0;

              /* "gnocis/models.pyx":498
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 				curve = self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative)
 * 				plt.plot([ 0., 1. ],             # <<<<<<<<<<<<<<
 * 					[ len(positives) / (len(positives) + len(negatives)),
 * 						len(positives) / (len(positives) + len(negatives))],
 */
              __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 498, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_6);
              PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_17);
              PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_17);
              __pyx_t_6 = 0;
              __pyx_t_17 = 0;

              /* "gnocis/models.pyx":501
 * 					[ len(positives) / (len(positives) + len(negatives)),
 * 						len(positives) / (len(positives) + len(negatives))],
 * 					linestyle = '--',             # <<<<<<<<<<<<<<
 * 					label = 'Expected at random')
 * 				plt.plot([ v.x for v in curve ],
 */
              __pyx_t_17 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 501, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_linestyle, __pyx_kp_u__6) < 0) __PYX_ERR(0, 501, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_label, __pyx_kp_u_Expected_at_random) < 0) __PYX_ERR(0, 501, __pyx_L13_error)

              /* "gnocis/models.pyx":498
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 * 				curve = self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative)
 * 				plt.plot([ 0., 1. ],             # <<<<<<<<<<<<<<
 * 					[ len(positives) / (len(positives) + len(negatives)),
 * 						len(positives) / (len(positives) + len(negatives))],
 */
              __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_4, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 498, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

              /* "gnocis/models.pyx":503
 * 					linestyle = '--',
 * 					label = 'Expected at random')
 * 				plt.plot([ v.x for v in curve ],             # <<<<<<<<<<<<<<
 * 					[ v.y for v in curve ],
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_plot); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 503, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              { /* enter inner scope */
                __pyx_t_17 = PyList_New(0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 503, __pyx_L23_error)
                __Pyx_GOTREF(__pyx_t_17);
                if (likely(PyList_CheckExact(__pyx_v_curve)) || PyTuple_CheckExact(__pyx_v_curve)) {
                  __pyx_t_4 = __pyx_v_curve; __Pyx_INCREF(__pyx_t_4); __pyx_t_14 = 0;
                  __pyx_t_18 = NULL;
                } else {
                  __pyx_t_14 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_curve); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 503, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __pyx_t_18 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 503, __pyx_L23_error)
                }
                for (;;) {
                  if (likely(!__pyx_t_18)) {
                    if (likely(PyList_CheckExact(__pyx_t_4))) {
                      if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_4)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_8 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_8); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 503, __pyx_L23_error)
                      #else
                      __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 503, __pyx_L23_error)
                      __Pyx_GOTREF(__pyx_t_8);
                      #endif
                    } else {
                      if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_8); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 503, __pyx_L23_error)
                      #else
                      __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 503, __pyx_L23_error)
                      __Pyx_GOTREF(__pyx_t_8);
                      #endif
                    }
                  } else {
                    __pyx_t_8 = __pyx_t_18(__pyx_t_4);
                    if (unlikely(!__pyx_t_8)) {
                      PyObject* exc_type = PyErr_Occurred();
                      if (exc_type) {
                        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                        else __PYX_ERR(0, 503, __pyx_L23_error)
                      }
                      break;
                    }
                    __Pyx_GOTREF(__pyx_t_8);
                  }
                  __Pyx_XDECREF_SET(__pyx_9genexpr31__pyx_v_v, __pyx_t_8);
                  __pyx_t_8 = 0;
                  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr31__pyx_v_v, __pyx_n_s_x); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 503, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  if (unlikely(__Pyx_ListComp_Append(__pyx_t_17, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 503, __pyx_L23_error)
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                }
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_XDECREF(__pyx_9genexpr31__pyx_v_v); __pyx_9genexpr31__pyx_v_v = 0;
                goto __pyx_L26_exit_scope;
                __pyx_L23_error:;
                __Pyx_XDECREF(__pyx_9genexpr31__pyx_v_v); __pyx_9genexpr31__pyx_v_v = 0;
                goto __pyx_L13_error;
                __pyx_L26_exit_scope:;
              } /* exit inner scope */
              { /* enter inner scope */

                /* "gnocis/models.pyx":504
 * 					label = 'Expected at random')
 * 				plt.plot([ v.x for v in curve ],
 * 					[ v.y for v in curve ],             # <<<<<<<<<<<<<<
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))
 * 				plt.xlabel('Recall', fontsize = 14)
 */
                __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 504, __pyx_L29_error)
                __Pyx_GOTREF(__pyx_t_4);
                if (likely(PyList_CheckExact(__pyx_v_curve)) || PyTuple_CheckExact(__pyx_v_curve)) {
                  __pyx_t_8 = __pyx_v_curve; __Pyx_INCREF(__pyx_t_8); __pyx_t_14 = 0;
                  __pyx_t_18 = NULL;
                } else {
                  __pyx_t_14 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_curve); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 504, __pyx_L29_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __pyx_t_18 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 504, __pyx_L29_error)
                }
                for (;;) {
                  if (likely(!__pyx_t_18)) {
                    if (likely(PyList_CheckExact(__pyx_t_8))) {
                      if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_8)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_5); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 504, __pyx_L29_error)
                      #else
                      __pyx_t_5 = PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L29_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      #endif
                    } else {
                      if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_5); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 504, __pyx_L29_error)
                      #else
                      __pyx_t_5 = PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L29_error)
                      __Pyx_GOTREF(__pyx_t_5);
                      #endif
                    }
                  } else {
                    __pyx_t_5 = __pyx_t_18(__pyx_t_8);
                    if (unlikely(!__pyx_t_5)) {
                      PyObject* exc_type = PyErr_Occurred();
                      if (exc_type) {
                        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                        else __PYX_ERR(0, 504, __pyx_L29_error)
                      }
                      break;
                    }
                    __Pyx_GOTREF(__pyx_t_5);
                  }
                  __Pyx_XDECREF_SET(__pyx_9genexpr32__pyx_v_v, __pyx_t_5);
                  __pyx_t_5 = 0;
                  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr32__pyx_v_v, __pyx_n_s_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 504, __pyx_L29_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 504, __pyx_L29_error)
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                }
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_XDECREF(__pyx_9genexpr32__pyx_v_v); __pyx_9genexpr32__pyx_v_v = 0;
                goto __pyx_L32_exit_scope;
                __pyx_L29_error:;
                __Pyx_XDECREF(__pyx_9genexpr32__pyx_v_v); __pyx_9genexpr32__pyx_v_v = 0;
                goto __pyx_L13_error;
                __pyx_L32_exit_scope:;
              } /* exit inner scope */

              /* "gnocis/models.pyx":503
 * 					linestyle = '--',
 * 					label = 'Expected at random')
 * 				plt.plot([ v.x for v in curve ],             # <<<<<<<<<<<<<<
 * 					[ v.y for v in curve ],
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))
 */
              __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 503, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_17);
              PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_17);
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
              __pyx_t_17 = 0;
              __pyx_t_4 = 0;

              /* "gnocis/models.pyx":505
 * 				plt.plot([ v.x for v in curve ],
 * 					[ v.y for v in curve ],
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))             # <<<<<<<<<<<<<<
 * 				plt.xlabel('Recall', fontsize = 14)
 * 				plt.ylabel('Precision', fontsize = 14)
 */
              __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_17 = PyTuple_New(4); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_14 = 0;
              __pyx_t_19 = 127;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_20 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_5), __pyx_empty_unicode); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_19 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_20) > __pyx_t_19) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_20) : __pyx_t_19;
              __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_20);
              __Pyx_GIVEREF(__pyx_t_20);
              PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_20);
              __pyx_t_20 = 0;
              __Pyx_INCREF(__pyx_kp_u_AUC);
              __pyx_t_14 += 9;
              __Pyx_GIVEREF(__pyx_kp_u_AUC);
              PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_kp_u_AUC);
              __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_21 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_21)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_21);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              __pyx_t_20 = (__pyx_t_21) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_21, __pyx_v_curve) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_curve);
              __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
              if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = PyNumber_Multiply(__pyx_t_20, __pyx_float_100_); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              __pyx_t_20 = __Pyx_PyObject_Format(__pyx_t_5, __pyx_kp_u_2f); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_19 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_20) > __pyx_t_19) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_20) : __pyx_t_19;
              __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_20);
              __Pyx_GIVEREF(__pyx_t_20);
              PyTuple_SET_ITEM(__pyx_t_17, 2, __pyx_t_20);
              __pyx_t_20 = 0;
              __Pyx_INCREF(__pyx_kp_u__7);
              __pyx_t_14 += 2;
              __Pyx_GIVEREF(__pyx_kp_u__7);
              PyTuple_SET_ITEM(__pyx_t_17, 3, __pyx_kp_u__7);
              __pyx_t_20 = __Pyx_PyUnicode_Join(__pyx_t_17, 4, __pyx_t_14, __pyx_t_19); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_label, __pyx_t_20) < 0) __PYX_ERR(0, 505, __pyx_L13_error)
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

              /* "gnocis/models.pyx":503
 * 					linestyle = '--',
 * 					label = 'Expected at random')
 * 				plt.plot([ v.x for v in curve ],             # <<<<<<<<<<<<<<
 * 					[ v.y for v in curve ],
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))
 */
              __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, __pyx_t_4); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 503, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

              /* "gnocis/models.pyx":506
 * 					[ v.y for v in curve ],
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))
 * 				plt.xlabel('Recall', fontsize = 14)             # <<<<<<<<<<<<<<
 * 				plt.ylabel('Precision', fontsize = 14)
 * 				plt.legend(loc = 'upper right', fancybox = True)
 */
              __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_xlabel); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 506, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_20);
              __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 506, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fontsize, __pyx_int_14) < 0) __PYX_ERR(0, 506, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_tuple__8, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 506, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":507
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))
 * 				plt.xlabel('Recall', fontsize = 14)
 * 				plt.ylabel('Precision', fontsize = 14)             # <<<<<<<<<<<<<<
 * 				plt.legend(loc = 'upper right', fancybox = True)
 * 				if outpath is None:
 */
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_ylabel); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 507, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 507, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fontsize, __pyx_int_14) < 0) __PYX_ERR(0, 507, __pyx_L13_error)
              __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__9, __pyx_t_4); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 507, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

              /* "gnocis/models.pyx":508
 * 				plt.xlabel('Recall', fontsize = 14)
 * 				plt.ylabel('Precision', fontsize = 14)
 * 				plt.legend(loc = 'upper right', fancybox = True)             # <<<<<<<<<<<<<<
 * 				if outpath is None:
 * 					bio = BytesIO()
 */
              __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_legend); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 508, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_20);
              __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 508, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_loc, __pyx_kp_u_upper_right) < 0) __PYX_ERR(0, 508, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fancybox, Py_True) < 0) __PYX_ERR(0, 508, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 508, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":509
 * 				plt.ylabel('Precision', fontsize = 14)
 * 				plt.legend(loc = 'upper right', fancybox = True)
 * 				if outpath is None:             # <<<<<<<<<<<<<<
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 */
              __pyx_t_22 = (__pyx_v_outpath == Py_None);
              __pyx_t_23 = (__pyx_t_22 != 0);
              if (__pyx_t_23) {

                /* "gnocis/models.pyx":510
 * 				plt.legend(loc = 'upper right', fancybox = True)
 * 				if outpath is None:
 * 					bio = BytesIO()             # <<<<<<<<<<<<<<
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')
 */
                __Pyx_INCREF(__pyx_v_BytesIO);
                __pyx_t_4 = __pyx_v_BytesIO; __pyx_t_20 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_20)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_20);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                __pyx_t_8 = (__pyx_t_20) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_20) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
                __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 510, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_v_bio = __pyx_t_8;
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":511
 * 				if outpath is None:
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')             # <<<<<<<<<<<<<<
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 */
                __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_savefig); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 511, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 511, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_INCREF(__pyx_v_bio);
                __Pyx_GIVEREF(__pyx_v_bio);
                PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_bio);
                __pyx_t_20 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 511, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_format, __pyx_n_u_png) < 0) __PYX_ERR(0, 511, __pyx_L13_error)
                __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_4, __pyx_t_20); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 511, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "gnocis/models.pyx":512
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')             # <<<<<<<<<<<<<<
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 */
                __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_close); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 512, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __pyx_t_4 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_20))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_20);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_20, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_4, __pyx_n_u_all) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_u_all);
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 512, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "gnocis/models.pyx":513
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')             # <<<<<<<<<<<<<<
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					display(HTML(html))
 */
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_base64, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_bio, __pyx_n_s_getvalue); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 513, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __pyx_t_5 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
                  __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_17);
                  if (likely(__pyx_t_5)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                    __Pyx_INCREF(__pyx_t_5);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_17, function);
                  }
                }
                __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_17);
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 513, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_17 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_17)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_17);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                __pyx_t_20 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_17, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8);
                __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 513, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __pyx_t_20 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_20)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_20);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_20, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_utf_8);
                __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 513, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_v_encoded = __pyx_t_6;
                __pyx_t_6 = 0;

                /* "gnocis/models.pyx":514
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded             # <<<<<<<<<<<<<<
 * 					display(HTML(html))
 * 				else:
 */
                __pyx_t_6 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_img_src_data_image_png_base64_s, __pyx_v_encoded); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 514, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_v_html = ((PyObject*)__pyx_t_6);
                __pyx_t_6 = 0;

                /* "gnocis/models.pyx":515
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					display(HTML(html))             # <<<<<<<<<<<<<<
 * 				else:
 * 					fig.savefig(outpath)
 */
                __Pyx_INCREF(__pyx_v_HTML);
                __pyx_t_20 = __pyx_v_HTML; __pyx_t_8 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
                  __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_20);
                  if (likely(__pyx_t_8)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
                    __Pyx_INCREF(__pyx_t_8);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_20, function);
                  }
                }
                __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_8, __pyx_v_html) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_v_html);
                __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 515, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_INCREF(__pyx_v_display);
                __pyx_t_20 = __pyx_v_display; __pyx_t_8 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
                  __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_20);
                  if (likely(__pyx_t_8)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
                    __Pyx_INCREF(__pyx_t_8);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_20, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_t_4);
                __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 515, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "gnocis/models.pyx":509
 * 				plt.ylabel('Precision', fontsize = 14)
 * 				plt.legend(loc = 'upper right', fancybox = True)
 * 				if outpath is None:             # <<<<<<<<<<<<<<
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 */
                goto __pyx_L33;
              }

              /* "gnocis/models.pyx":517
 * 					display(HTML(html))
 * 				else:
 * 					fig.savefig(outpath)             # <<<<<<<<<<<<<<
 * 					plt.close('all')
 * 		except ImportError as err:
 */
              /*else*/ {
                __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_savefig); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 517, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __pyx_t_4 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_20))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_20);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_20, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_4, __pyx_v_outpath) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_v_outpath);
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 517, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "gnocis/models.pyx":518
 * 				else:
 * 					fig.savefig(outpath)
 * 					plt.close('all')             # <<<<<<<<<<<<<<
 * 		except ImportError as err:
 * 			raise err
 */
                __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_close); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 518, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __pyx_t_4 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_20))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_20);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_20, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_4, __pyx_n_u_all) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_n_u_all);
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 518, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              }
              __pyx_L33:;

              /* "gnocis/models.pyx":494
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 				fig = plt.figure(figsize = figsize)
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 */
            }
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            goto __pyx_L18_try_end;
            __pyx_L13_error:;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("gnocis.models.sequenceModel.plotPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_20, &__pyx_t_4) < 0) __PYX_ERR(0, 494, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_8 = PyTuple_Pack(3, __pyx_t_6, __pyx_t_20, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 494, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 494, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_24);
              __pyx_t_23 = __Pyx_PyObject_IsTrue(__pyx_t_24);
              __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
              if (__pyx_t_23 < 0) __PYX_ERR(0, 494, __pyx_L15_except_error)
              __pyx_t_22 = ((!(__pyx_t_23 != 0)) != 0);
              if (__pyx_t_22) {
                __Pyx_GIVEREF(__pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_4);
                __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_20, __pyx_t_4);
                __pyx_t_6 = 0; __pyx_t_20 = 0; __pyx_t_4 = 0; 
                __PYX_ERR(0, 494, __pyx_L15_except_error)
              }
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              goto __pyx_L14_exception_handled;
            }
            __pyx_L15_except_error:;
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
            goto __pyx_L3_error;
            __pyx_L14_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
            __pyx_L18_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_7) {
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__10, NULL);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 494, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            }
            goto __pyx_L12;
          }
          __pyx_L12:;
        }
        goto __pyx_L37;
        __pyx_L9_error:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L3_error;
        __pyx_L37:;
      }

      /* "gnocis/models.pyx":489
 * 		:type style: str, optional
 * 		"""
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "gnocis/models.pyx":519
 * 					fig.savefig(outpath)
 * 					plt.close('all')
 * 		except ImportError as err:             # <<<<<<<<<<<<<<
 * 			raise err
 * 
 */
    __pyx_t_25 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_25) {
      __Pyx_AddTraceback("gnocis.models.sequenceModel.plotPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_20, &__pyx_t_6) < 0) __PYX_ERR(0, 519, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_20);
      __pyx_v_err = __pyx_t_20;
      /*try:*/ {

        /* "gnocis/models.pyx":520
 * 					plt.close('all')
 * 		except ImportError as err:
 * 			raise err             # <<<<<<<<<<<<<<
 * 
 * 	# Returns validation statistics dictionary.
 */
        __Pyx_Raise(__pyx_v_err, 0, 0, 0);
        __PYX_ERR(0, 520, __pyx_L43_error)
      }

      /* "gnocis/models.pyx":519
 * 					fig.savefig(outpath)
 * 					plt.close('all')
 * 		except ImportError as err:             # <<<<<<<<<<<<<<
 * 			raise err
 * 
 */
      /*finally:*/ {
        __pyx_L43_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_7 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_24 = 0; __pyx_t_28 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_9, &__pyx_t_24, &__pyx_t_28);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_11, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_11, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_24);
          __Pyx_XGOTREF(__pyx_t_28);
          __pyx_t_25 = __pyx_lineno; __pyx_t_26 = __pyx_clineno; __pyx_t_27 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err);
            __pyx_v_err = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_24);
            __Pyx_XGIVEREF(__pyx_t_28);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_24, __pyx_t_28);
          }
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ErrRestore(__pyx_t_7, __pyx_t_11, __pyx_t_10);
          __pyx_t_7 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_24 = 0; __pyx_t_28 = 0;
          __pyx_lineno = __pyx_t_25; __pyx_clineno = __pyx_t_26; __pyx_filename = __pyx_t_27;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "gnocis/models.pyx":489
 * 		:type style: str, optional
 * 		"""
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "gnocis/models.pyx":472
 * 		return getAUC(self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative))
 * 
 * 	def plotPRC(self, seqs, labelPositive = positive, labelNegative = negative, figsize = (8, 8), outpath = None, style = 'ggplot'):             # <<<<<<<<<<<<<<
 * 		""" Plots a Precision/Recall curve and either displays it in an IPython session or saves it to a file.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.plotPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_plt);
  __Pyx_XDECREF(__pyx_v_base64);
  __Pyx_XDECREF(__pyx_v_BytesIO);
  __Pyx_XDECREF(__pyx_v_display);
  __Pyx_XDECREF(__pyx_v_HTML);
  __Pyx_XDECREF(__pyx_v_fig);
  __Pyx_XDECREF(__pyx_v_positives);
  __Pyx_XDECREF(__pyx_v_negatives);
  __Pyx_XDECREF(__pyx_v_curve);
  __Pyx_XDECREF(__pyx_v_bio);
  __Pyx_XDECREF(__pyx_v_encoded);
  __Pyx_XDECREF(__pyx_v_html);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_9genexpr31__pyx_v_v);
  __Pyx_XDECREF(__pyx_9genexpr32__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":523
 * 
 * 	# Returns validation statistics dictionary.
 * 	def getValidationStatistics(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Returns common model validation statistics (confusion matrix values; ROCAUC; PRCAUC).
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_48__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_29getValidationStatistics(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_28getValidationStatistics[] = " Returns common model validation statistics (confusion matrix values; ROCAUC; PRCAUC).\n\t\t\n\t\t:param seqs: Sequences.\n\t\t:param labelPositive: Label of positives.\n\t\t:param labelNegative: Label of negatives.\n\t\t\n\t\t:type seqs: sequences\n\t\t:type labelPositive: sequenceLabel\n\t\t:type labelNegative: sequenceLabel\n\t\t\n\t\t:return: Validation statistics\n\t\t:rtype: list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_29getValidationStatistics = {"getValidationStatistics", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_29getValidationStatistics, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_28getValidationStatistics};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_29getValidationStatistics(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getValidationStatistics (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults7 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getValidationStatistics", 0, 2, 4, 1); __PYX_ERR(0, 523, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getValidationStatistics") < 0)) __PYX_ERR(0, 523, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getValidationStatistics", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 523, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getValidationStatistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_28getValidationStatistics(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_28getValidationStatistics(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_v_CMStats = NULL;
  PyObject *__pyx_v_stats = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getValidationStatistics", 0);

  /* "gnocis/models.pyx":537
 * 		:rtype: list
 * 		"""
 * 		CMStats = getConfusionMatrixStatistics( self.getConfusionMatrix(seqs, labelPositive = labelPositive, labelNegative = labelNegative) )             # <<<<<<<<<<<<<<
 * 		stats = {
 * 			'title': 'Validation', 'positives': labelPositive, 'negatives': labelNegative, 'model': self,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_getConfusionMatrixStatistics); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getConfusionMatrix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_seqs);
  __Pyx_GIVEREF(__pyx_v_seqs);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_seqs);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 537, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 537, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_CMStats = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":539
 * 		CMStats = getConfusionMatrixStatistics( self.getConfusionMatrix(seqs, labelPositive = labelPositive, labelNegative = labelNegative) )
 * 		stats = {
 * 			'title': 'Validation', 'positives': labelPositive, 'negatives': labelNegative, 'model': self,             # <<<<<<<<<<<<<<
 * 			'ROC AUC': self.getROCAUC(seqs, labelPositive = labelPositive, labelNegative = labelNegative),
 * 			'PRC AUC': self.getPRCAUC(seqs, labelPositive = labelPositive, labelNegative = labelNegative),
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_title, __pyx_n_u_Validation) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_positives, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_negatives, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_model, __pyx_v_self) < 0) __PYX_ERR(0, 539, __pyx_L1_error)

  /* "gnocis/models.pyx":540
 * 		stats = {
 * 			'title': 'Validation', 'positives': labelPositive, 'negatives': labelNegative, 'model': self,
 * 			'ROC AUC': self.getROCAUC(seqs, labelPositive = labelPositive, labelNegative = labelNegative),             # <<<<<<<<<<<<<<
 * 			'PRC AUC': self.getPRCAUC(seqs, labelPositive = labelPositive, labelNegative = labelNegative),
 * 		}
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getROCAUC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_seqs);
  __Pyx_GIVEREF(__pyx_v_seqs);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_seqs);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_kp_u_ROC_AUC, __pyx_t_4) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gnocis/models.pyx":541
 * 			'title': 'Validation', 'positives': labelPositive, 'negatives': labelNegative, 'model': self,
 * 			'ROC AUC': self.getROCAUC(seqs, labelPositive = labelPositive, labelNegative = labelNegative),
 * 			'PRC AUC': self.getPRCAUC(seqs, labelPositive = labelPositive, labelNegative = labelNegative),             # <<<<<<<<<<<<<<
 * 		}
 * 		for k in CMStats.keys():
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getPRCAUC); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_seqs);
  __Pyx_GIVEREF(__pyx_v_seqs);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_seqs);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 541, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 541, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_kp_u_PRC_AUC, __pyx_t_2) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_stats = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":543
 * 			'PRC AUC': self.getPRCAUC(seqs, labelPositive = labelPositive, labelNegative = labelNegative),
 * 		}
 * 		for k in CMStats.keys():             # <<<<<<<<<<<<<<
 * 			stats[k] = CMStats[k]
 * 		return stats
 */
  __pyx_t_7 = 0;
  if (unlikely(__pyx_v_CMStats == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(0, 543, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_dict_iterator(__pyx_v_CMStats, 0, __pyx_n_s_keys, (&__pyx_t_8), (&__pyx_t_9)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_2;
  __pyx_t_2 = 0;
  while (1) {
    __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_8, &__pyx_t_7, &__pyx_t_2, NULL, NULL, __pyx_t_9);
    if (unlikely(__pyx_t_10 == 0)) break;
    if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":544
 * 		}
 * 		for k in CMStats.keys():
 * 			stats[k] = CMStats[k]             # <<<<<<<<<<<<<<
 * 		return stats
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_CMStats, __pyx_v_k); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(PyDict_SetItem(__pyx_v_stats, __pyx_v_k, __pyx_t_2) < 0)) __PYX_ERR(0, 544, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":545
 * 		for k in CMStats.keys():
 * 			stats[k] = CMStats[k]
 * 		return stats             # <<<<<<<<<<<<<<
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_stats);
  __pyx_r = __pyx_v_stats;
  goto __pyx_L0;

  /* "gnocis/models.pyx":523
 * 
 * 	# Returns validation statistics dictionary.
 * 	def getValidationStatistics(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Returns common model validation statistics (confusion matrix values; ROCAUC; PRCAUC).
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.getValidationStatistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_CMStats);
  __Pyx_XDECREF(__pyx_v_stats);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":548
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamRegions(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_31predictSequenceStreamRegions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_30predictSequenceStreamRegions[] = " Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.\n\t\t\n\t\t:param stream: Sequence material that the model is applied to for prediction.\n\t\t:type stream: sequences/sequenceStream/str\n\t\t\n\t\t:return: Set of non-overlapping (merged) predicted regions\n\t\t:rtype: regions\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_31predictSequenceStreamRegions = {"predictSequenceStreamRegions", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_31predictSequenceStreamRegions, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_30predictSequenceStreamRegions};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_31predictSequenceStreamRegions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_stream = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predictSequenceStreamRegions (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_stream,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("predictSequenceStreamRegions", 1, 2, 2, 1); __PYX_ERR(0, 548, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "predictSequenceStreamRegions") < 0)) __PYX_ERR(0, 548, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_stream = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("predictSequenceStreamRegions", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 548, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.predictSequenceStreamRegions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_30predictSequenceStreamRegions(__pyx_self, __pyx_v_self, __pyx_v_stream);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_30predictSequenceStreamRegions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_stream) {
  struct __pyx_obj_6gnocis_7regions_regions *__pyx_v_pred = 0;
  PyObject *__pyx_v_winPosSet = 0;
  PyObject *__pyx_v_winScoreSet = 0;
  int __pyx_v_i;
  int __pyx_v_nNT;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_v_winPos = 0;
  struct __pyx_obj_6gnocis_9sequences_sequence *__pyx_v_winSeq = 0;
  struct __pyx_obj_6gnocis_9sequences_sequences *__pyx_v_winSeqSet = 0;
  float __pyx_v_wScore;
  PyObject *__pyx_v_cScores = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  float __pyx_t_14;
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("predictSequenceStreamRegions", 0);

  /* "gnocis/models.pyx":558
 * 		"""
 * 		cdef regions pred
 * 		cdef list winPosSet = []             # <<<<<<<<<<<<<<
 * 		cdef list winScoreSet = []
 * 		cdef int i = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_winPosSet = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":559
 * 		cdef regions pred
 * 		cdef list winPosSet = []
 * 		cdef list winScoreSet = []             # <<<<<<<<<<<<<<
 * 		cdef int i = 0
 * 		cdef int nNT = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_winScoreSet = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":560
 * 		cdef list winPosSet = []
 * 		cdef list winScoreSet = []
 * 		cdef int i = 0             # <<<<<<<<<<<<<<
 * 		cdef int nNT = 0
 * 		cdef region winPos
 */
  __pyx_v_i = 0;

  /* "gnocis/models.pyx":561
 * 		cdef list winScoreSet = []
 * 		cdef int i = 0
 * 		cdef int nNT = 0             # <<<<<<<<<<<<<<
 * 		cdef region winPos
 * 		cdef sequence winSeq
 */
  __pyx_v_nNT = 0;

  /* "gnocis/models.pyx":564
 * 		cdef region winPos
 * 		cdef sequence winSeq
 * 		cdef sequences winSeqSet = sequences('', [])             # <<<<<<<<<<<<<<
 * 		cdef float wScore
 * 		cdef list cScores
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_u__2);
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_sequences), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_winSeqSet = ((struct __pyx_obj_6gnocis_9sequences_sequences *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":567
 * 		cdef float wScore
 * 		cdef list cScores
 * 		pred = regions('Predictions', [])             # <<<<<<<<<<<<<<
 * 
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_u_Predictions);
  __Pyx_GIVEREF(__pyx_n_u_Predictions);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_u_Predictions);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_7regions_regions), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_pred = ((struct __pyx_obj_6gnocis_7regions_regions *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":569
 * 		pred = regions('Predictions', [])
 * 
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):             # <<<<<<<<<<<<<<
 * 			winPosSet.append(winSeq.sourceRegion)
 * 			winSeqSet.sequences.append(winSeq)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_streamSequenceWindows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_windowSize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_windowStep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_stream, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_stream, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_stream);
    __Pyx_GIVEREF(__pyx_v_stream);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_stream);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 569, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 569, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 569, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 569, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6gnocis_9sequences_sequence))))) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_winSeq, ((struct __pyx_obj_6gnocis_9sequences_sequence *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":570
 * 
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):
 * 			winPosSet.append(winSeq.sourceRegion)             # <<<<<<<<<<<<<<
 * 			winSeqSet.sequences.append(winSeq)
 * 			i += 1
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_winSeq->sourceRegion);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_winPosSet, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":571
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):
 * 			winPosSet.append(winSeq.sourceRegion)
 * 			winSeqSet.sequences.append(winSeq)             # <<<<<<<<<<<<<<
 * 			i += 1
 * 			nNT += len(winSeq)
 */
    if (unlikely(__pyx_v_winSeqSet->sequences == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 571, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_winSeqSet->sequences, ((PyObject *)__pyx_v_winSeq)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 571, __pyx_L1_error)

    /* "gnocis/models.pyx":572
 * 			winPosSet.append(winSeq.sourceRegion)
 * 			winSeqSet.sequences.append(winSeq)
 * 			i += 1             # <<<<<<<<<<<<<<
 * 			nNT += len(winSeq)
 * 			if nNT >= maxThreadFetchNT:
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "gnocis/models.pyx":573
 * 			winSeqSet.sequences.append(winSeq)
 * 			i += 1
 * 			nNT += len(winSeq)             # <<<<<<<<<<<<<<
 * 			if nNT >= maxThreadFetchNT:
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 */
    __pyx_t_11 = PyObject_Length(((PyObject *)__pyx_v_winSeq)); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 573, __pyx_L1_error)
    __pyx_v_nNT = (__pyx_v_nNT + __pyx_t_11);

    /* "gnocis/models.pyx":574
 * 			i += 1
 * 			nNT += len(winSeq)
 * 			if nNT >= maxThreadFetchNT:             # <<<<<<<<<<<<<<
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 * 				winSeqSet.sequences = []
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nNT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_maxThreadFetchNT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_12) {

      /* "gnocis/models.pyx":575
 * 			nNT += len(winSeq)
 * 			if nNT >= maxThreadFetchNT:
 * 				winScoreSet += self.getSequenceScores(winSeqSet)             # <<<<<<<<<<<<<<
 * 				winSeqSet.sequences = []
 * 				i = 0
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_1, ((PyObject *)__pyx_v_winSeqSet)) : __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_v_winSeqSet));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_winScoreSet, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(PyList_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_winScoreSet, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "gnocis/models.pyx":576
 * 			if nNT >= maxThreadFetchNT:
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 * 				winSeqSet.sequences = []             # <<<<<<<<<<<<<<
 * 				i = 0
 * 				nNT = 0
 */
      __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_v_winSeqSet->sequences);
      __Pyx_DECREF(__pyx_v_winSeqSet->sequences);
      __pyx_v_winSeqSet->sequences = ((PyObject*)__pyx_t_7);
      __pyx_t_7 = 0;

      /* "gnocis/models.pyx":577
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 * 				winSeqSet.sequences = []
 * 				i = 0             # <<<<<<<<<<<<<<
 * 				nNT = 0
 * 		if len(winSeqSet.sequences) > 0:
 */
      __pyx_v_i = 0;

      /* "gnocis/models.pyx":578
 * 				winSeqSet.sequences = []
 * 				i = 0
 * 				nNT = 0             # <<<<<<<<<<<<<<
 * 		if len(winSeqSet.sequences) > 0:
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 */
      __pyx_v_nNT = 0;

      /* "gnocis/models.pyx":574
 * 			i += 1
 * 			nNT += len(winSeq)
 * 			if nNT >= maxThreadFetchNT:             # <<<<<<<<<<<<<<
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 * 				winSeqSet.sequences = []
 */
    }

    /* "gnocis/models.pyx":569
 * 		pred = regions('Predictions', [])
 * 
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):             # <<<<<<<<<<<<<<
 * 			winPosSet.append(winSeq.sourceRegion)
 * 			winSeqSet.sequences.append(winSeq)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":579
 * 				i = 0
 * 				nNT = 0
 * 		if len(winSeqSet.sequences) > 0:             # <<<<<<<<<<<<<<
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 * 
 */
  __pyx_t_2 = __pyx_v_winSeqSet->sequences;
  __Pyx_INCREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 579, __pyx_L1_error)
  }
  __pyx_t_8 = PyList_GET_SIZE(__pyx_t_2); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = ((__pyx_t_8 > 0) != 0);
  if (__pyx_t_12) {

    /* "gnocis/models.pyx":580
 * 				nNT = 0
 * 		if len(winSeqSet.sequences) > 0:
 * 			winScoreSet += self.getSequenceScores(winSeqSet)             # <<<<<<<<<<<<<<
 * 
 * 		for winPos, wScore in zip(winPosSet, winScoreSet):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, ((PyObject *)__pyx_v_winSeqSet)) : __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_v_winSeqSet));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_winScoreSet, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_winScoreSet, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "gnocis/models.pyx":579
 * 				i = 0
 * 				nNT = 0
 * 		if len(winSeqSet.sequences) > 0:             # <<<<<<<<<<<<<<
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 * 
 */
  }

  /* "gnocis/models.pyx":582
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 * 
 * 		for winPos, wScore in zip(winPosSet, winScoreSet):             # <<<<<<<<<<<<<<
 * 			if wScore <= self.threshold: continue
 * 			if len(pred) == 0 or pred[-1].seq != winPos.seq or pred[-1].end < winPos.start-1:
 */
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_winPosSet);
  __Pyx_GIVEREF(__pyx_v_winPosSet);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_winPosSet);
  __Pyx_INCREF(__pyx_v_winScoreSet);
  __Pyx_GIVEREF(__pyx_v_winScoreSet);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_winScoreSet);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_7 = __pyx_t_2; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 582, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 582, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 582, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_9(__pyx_t_7);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 582, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 582, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_13(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_1 = __pyx_t_13(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_3), 2) < 0) __PYX_ERR(0, 582, __pyx_L1_error)
      __pyx_t_13 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_13 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 582, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 582, __pyx_L1_error)
    __pyx_t_14 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_14 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_winPos, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_4));
    __pyx_t_4 = 0;
    __pyx_v_wScore = __pyx_t_14;

    /* "gnocis/models.pyx":583
 * 
 * 		for winPos, wScore in zip(winPosSet, winScoreSet):
 * 			if wScore <= self.threshold: continue             # <<<<<<<<<<<<<<
 * 			if len(pred) == 0 or pred[-1].seq != winPos.seq or pred[-1].end < winPos.start-1:
 * 				cScores = [ ]
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_wScore); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_threshold); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 583, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 583, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_12) {
      goto __pyx_L7_continue;
    }

    /* "gnocis/models.pyx":584
 * 		for winPos, wScore in zip(winPosSet, winScoreSet):
 * 			if wScore <= self.threshold: continue
 * 			if len(pred) == 0 or pred[-1].seq != winPos.seq or pred[-1].end < winPos.start-1:             # <<<<<<<<<<<<<<
 * 				cScores = [ ]
 * 				pred.regions.append(winPos)
 */
    __pyx_t_11 = PyObject_Length(((PyObject *)__pyx_v_pred)); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 584, __pyx_L1_error)
    __pyx_t_15 = ((__pyx_t_11 == 0) != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_12 = __pyx_t_15;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_GetItemInt(((PyObject *)__pyx_v_pred), -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_seq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_15 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_v_winPos->seq, Py_NE)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_15) {
    } else {
      __pyx_t_12 = __pyx_t_15;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_GetItemInt(((PyObject *)__pyx_v_pred), -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG((__pyx_v_winPos->start - 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_12 = __pyx_t_15;
    __pyx_L13_bool_binop_done:;
    if (__pyx_t_12) {

      /* "gnocis/models.pyx":585
 * 			if wScore <= self.threshold: continue
 * 			if len(pred) == 0 or pred[-1].seq != winPos.seq or pred[-1].end < winPos.start-1:
 * 				cScores = [ ]             # <<<<<<<<<<<<<<
 * 				pred.regions.append(winPos)
 * 			else:
 */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_cScores, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "gnocis/models.pyx":586
 * 			if len(pred) == 0 or pred[-1].seq != winPos.seq or pred[-1].end < winPos.start-1:
 * 				cScores = [ ]
 * 				pred.regions.append(winPos)             # <<<<<<<<<<<<<<
 * 			else:
 * 				pred[-1].end = winPos.end
 */
      if (unlikely(__pyx_v_pred->regions == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 586, __pyx_L1_error)
      }
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_pred->regions, ((PyObject *)__pyx_v_winPos)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 586, __pyx_L1_error)

      /* "gnocis/models.pyx":584
 * 		for winPos, wScore in zip(winPosSet, winScoreSet):
 * 			if wScore <= self.threshold: continue
 * 			if len(pred) == 0 or pred[-1].seq != winPos.seq or pred[-1].end < winPos.start-1:             # <<<<<<<<<<<<<<
 * 				cScores = [ ]
 * 				pred.regions.append(winPos)
 */
      goto __pyx_L12;
    }

    /* "gnocis/models.pyx":588
 * 				pred.regions.append(winPos)
 * 			else:
 * 				pred[-1].end = winPos.end             # <<<<<<<<<<<<<<
 * 			cScores.append(wScore)
 * 			pred[-1].score = sum(cScores) / len(cScores)
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_winPos->end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_GetItemInt(((PyObject *)__pyx_v_pred), -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_end, __pyx_t_2) < 0) __PYX_ERR(0, 588, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L12:;

    /* "gnocis/models.pyx":589
 * 			else:
 * 				pred[-1].end = winPos.end
 * 			cScores.append(wScore)             # <<<<<<<<<<<<<<
 * 			pred[-1].score = sum(cScores) / len(cScores)
 * 		pred.sort()
 */
    if (unlikely(!__pyx_v_cScores)) { __Pyx_RaiseUnboundLocalError("cScores"); __PYX_ERR(0, 589, __pyx_L1_error) }
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_wScore); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_cScores, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 589, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":590
 * 				pred[-1].end = winPos.end
 * 			cScores.append(wScore)
 * 			pred[-1].score = sum(cScores) / len(cScores)             # <<<<<<<<<<<<<<
 * 		pred.sort()
 * 		return pred
 */
    if (unlikely(!__pyx_v_cScores)) { __Pyx_RaiseUnboundLocalError("cScores"); __PYX_ERR(0, 590, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_v_cScores); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(!__pyx_v_cScores)) { __Pyx_RaiseUnboundLocalError("cScores"); __PYX_ERR(0, 590, __pyx_L1_error) }
    __pyx_t_11 = PyList_GET_SIZE(__pyx_v_cScores); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 590, __pyx_L1_error)
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt(((PyObject *)__pyx_v_pred), -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_score, __pyx_t_4) < 0) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":582
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 * 
 * 		for winPos, wScore in zip(winPosSet, winScoreSet):             # <<<<<<<<<<<<<<
 * 			if wScore <= self.threshold: continue
 * 			if len(pred) == 0 or pred[-1].seq != winPos.seq or pred[-1].end < winPos.start-1:
 */
    __pyx_L7_continue:;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "gnocis/models.pyx":591
 * 			cScores.append(wScore)
 * 			pred[-1].score = sum(cScores) / len(cScores)
 * 		pred.sort()             # <<<<<<<<<<<<<<
 * 		return pred
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_pred), __pyx_n_s_sort); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "gnocis/models.pyx":592
 * 			pred[-1].score = sum(cScores) / len(cScores)
 * 		pred.sort()
 * 		return pred             # <<<<<<<<<<<<<<
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_pred));
  __pyx_r = ((PyObject *)__pyx_v_pred);
  goto __pyx_L0;

  /* "gnocis/models.pyx":548
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamRegions(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.predictSequenceStreamRegions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_pred);
  __Pyx_XDECREF(__pyx_v_winPosSet);
  __Pyx_XDECREF(__pyx_v_winScoreSet);
  __Pyx_XDECREF((PyObject *)__pyx_v_winPos);
  __Pyx_XDECREF((PyObject *)__pyx_v_winSeq);
  __Pyx_XDECREF((PyObject *)__pyx_v_winSeqSet);
  __Pyx_XDECREF(__pyx_v_cScores);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":595
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamCurves(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_33predictSequenceStreamCurves(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_32predictSequenceStreamCurves[] = " Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.\n\t\t\n\t\t:param stream: Sequence material that the model is applied to for prediction.\n\t\t:type stream: sequences/sequenceStream/str\n\t\t\n\t\t:return: Set of non-overlapping (merged) predicted regions\n\t\t:rtype: regions\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_33predictSequenceStreamCurves = {"predictSequenceStreamCurves", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_33predictSequenceStreamCurves, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_32predictSequenceStreamCurves};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_33predictSequenceStreamCurves(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_stream = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predictSequenceStreamCurves (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_stream,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("predictSequenceStreamCurves", 1, 2, 2, 1); __PYX_ERR(0, 595, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "predictSequenceStreamCurves") < 0)) __PYX_ERR(0, 595, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_stream = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("predictSequenceStreamCurves", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 595, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.predictSequenceStreamCurves", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_32predictSequenceStreamCurves(__pyx_self, __pyx_v_self, __pyx_v_stream);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_27predictSequenceStreamCurves_2generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":630
 * 
 * 		posScores = {}
 * 		for seq in set(winPos.seq for winPos in winPosSet):             # <<<<<<<<<<<<<<
 * 			scores = [
 * 				(int(winPos.start/self.windowStep), wScore)
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_27predictSequenceStreamCurves_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_16_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_16_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 630, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_13sequenceModel_27predictSequenceStreamCurves_2generator9, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_sequenceModel_predictSequenceStr, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.predictSequenceStreamCurves.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_27predictSequenceStreamCurves_2generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 630, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_winPosSet)) { __Pyx_RaiseClosureNameError("winPosSet"); __PYX_ERR(0, 630, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_winPosSet == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 630, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_winPosSet; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 630, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_winPos);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_winPos, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_winPos, __pyx_n_s_seq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":595
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamCurves(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_32predictSequenceStreamCurves(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_stream) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *__pyx_cur_scope;
  PyObject *__pyx_v_winScoreSet = 0;
  int __pyx_v_i;
  int __pyx_v_nNT;
  struct __pyx_obj_6gnocis_9sequences_sequence *__pyx_v_winSeq = 0;
  struct __pyx_obj_6gnocis_9sequences_sequences *__pyx_v_winSeqSet = 0;
  PyObject *__pyx_v_sScores = 0;
  double __pyx_v_s;
  PyObject *__pyx_v_posScores = NULL;
  PyObject *__pyx_v_seq = NULL;
  PyObject *__pyx_v_scores = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_gb_6gnocis_6models_13sequenceModel_27predictSequenceStreamCurves_2generator9 = 0;
  struct __pyx_obj_6gnocis_7regions_region *__pyx_9genexpr34__pyx_v_winPos = NULL;
  float __pyx_9genexpr34__pyx_v_wScore;
  CYTHON_UNUSED Py_ssize_t __pyx_9genexpr35__pyx_v__;
  PyObject *__pyx_9genexpr36__pyx_v_seq = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  float __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  double __pyx_t_20;
  int __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("predictSequenceStreamCurves", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves(__pyx_ptype_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 595, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/models.pyx":605
 * 		"""
 * 		#cdef regions pred
 * 		cdef list winPosSet = []             # <<<<<<<<<<<<<<
 * 		cdef list winScoreSet = []
 * 		cdef int i = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_winPosSet = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":606
 * 		#cdef regions pred
 * 		cdef list winPosSet = []
 * 		cdef list winScoreSet = []             # <<<<<<<<<<<<<<
 * 		cdef int i = 0
 * 		cdef int nNT = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_winScoreSet = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":607
 * 		cdef list winPosSet = []
 * 		cdef list winScoreSet = []
 * 		cdef int i = 0             # <<<<<<<<<<<<<<
 * 		cdef int nNT = 0
 * 		cdef region winPos
 */
  __pyx_v_i = 0;

  /* "gnocis/models.pyx":608
 * 		cdef list winScoreSet = []
 * 		cdef int i = 0
 * 		cdef int nNT = 0             # <<<<<<<<<<<<<<
 * 		cdef region winPos
 * 		cdef sequence winSeq
 */
  __pyx_v_nNT = 0;

  /* "gnocis/models.pyx":611
 * 		cdef region winPos
 * 		cdef sequence winSeq
 * 		cdef sequences winSeqSet = sequences('', [])             # <<<<<<<<<<<<<<
 * 		cdef float wScore
 * 		cdef list sScores
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_u__2);
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_sequences), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_winSeqSet = ((struct __pyx_obj_6gnocis_9sequences_sequences *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":616
 * 		cdef double s
 * 
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):             # <<<<<<<<<<<<<<
 * 			winPosSet.append(winSeq.sourceRegion)
 * 			winSeqSet.sequences.append(winSeq)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_streamSequenceWindows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_windowSize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_windowStep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_stream, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_stream, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_stream);
    __Pyx_GIVEREF(__pyx_v_stream);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_stream);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 616, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 616, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 616, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 616, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6gnocis_9sequences_sequence))))) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_winSeq, ((struct __pyx_obj_6gnocis_9sequences_sequence *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":617
 * 
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):
 * 			winPosSet.append(winSeq.sourceRegion)             # <<<<<<<<<<<<<<
 * 			winSeqSet.sequences.append(winSeq)
 * 			i += 1
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_winSeq->sourceRegion);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_winPosSet, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":618
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):
 * 			winPosSet.append(winSeq.sourceRegion)
 * 			winSeqSet.sequences.append(winSeq)             # <<<<<<<<<<<<<<
 * 			i += 1
 * 			nNT += len(winSeq)
 */
    if (unlikely(__pyx_v_winSeqSet->sequences == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 618, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_winSeqSet->sequences, ((PyObject *)__pyx_v_winSeq)); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 618, __pyx_L1_error)

    /* "gnocis/models.pyx":619
 * 			winPosSet.append(winSeq.sourceRegion)
 * 			winSeqSet.sequences.append(winSeq)
 * 			i += 1             # <<<<<<<<<<<<<<
 * 			nNT += len(winSeq)
 * 			if nNT >= maxThreadFetchNT:
 */
    __pyx_v_i = (__pyx_v_i + 1);

    /* "gnocis/models.pyx":620
 * 			winSeqSet.sequences.append(winSeq)
 * 			i += 1
 * 			nNT += len(winSeq)             # <<<<<<<<<<<<<<
 * 			if nNT >= maxThreadFetchNT:
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 */
    __pyx_t_11 = PyObject_Length(((PyObject *)__pyx_v_winSeq)); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 620, __pyx_L1_error)
    __pyx_v_nNT = (__pyx_v_nNT + __pyx_t_11);

    /* "gnocis/models.pyx":621
 * 			i += 1
 * 			nNT += len(winSeq)
 * 			if nNT >= maxThreadFetchNT:             # <<<<<<<<<<<<<<
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 * 				winSeqSet.sequences = []
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nNT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_maxThreadFetchNT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_12) {

      /* "gnocis/models.pyx":622
 * 			nNT += len(winSeq)
 * 			if nNT >= maxThreadFetchNT:
 * 				winScoreSet += self.getSequenceScores(winSeqSet)             # <<<<<<<<<<<<<<
 * 				winSeqSet.sequences = []
 * 				i = 0
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_1, ((PyObject *)__pyx_v_winSeqSet)) : __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_v_winSeqSet));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_winScoreSet, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(PyList_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_winScoreSet, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "gnocis/models.pyx":623
 * 			if nNT >= maxThreadFetchNT:
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 * 				winSeqSet.sequences = []             # <<<<<<<<<<<<<<
 * 				i = 0
 * 				nNT = 0
 */
      __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 623, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_v_winSeqSet->sequences);
      __Pyx_DECREF(__pyx_v_winSeqSet->sequences);
      __pyx_v_winSeqSet->sequences = ((PyObject*)__pyx_t_7);
      __pyx_t_7 = 0;

      /* "gnocis/models.pyx":624
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 * 				winSeqSet.sequences = []
 * 				i = 0             # <<<<<<<<<<<<<<
 * 				nNT = 0
 * 		if len(winSeqSet.sequences) > 0:
 */
      __pyx_v_i = 0;

      /* "gnocis/models.pyx":625
 * 				winSeqSet.sequences = []
 * 				i = 0
 * 				nNT = 0             # <<<<<<<<<<<<<<
 * 		if len(winSeqSet.sequences) > 0:
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 */
      __pyx_v_nNT = 0;

      /* "gnocis/models.pyx":621
 * 			i += 1
 * 			nNT += len(winSeq)
 * 			if nNT >= maxThreadFetchNT:             # <<<<<<<<<<<<<<
 * 				winScoreSet += self.getSequenceScores(winSeqSet)
 * 				winSeqSet.sequences = []
 */
    }

    /* "gnocis/models.pyx":616
 * 		cdef double s
 * 
 * 		for winSeq in streamSequenceWindows(stream, self.windowSize, self.windowStep):             # <<<<<<<<<<<<<<
 * 			winPosSet.append(winSeq.sourceRegion)
 * 			winSeqSet.sequences.append(winSeq)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":626
 * 				i = 0
 * 				nNT = 0
 * 		if len(winSeqSet.sequences) > 0:             # <<<<<<<<<<<<<<
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 * 
 */
  __pyx_t_2 = __pyx_v_winSeqSet->sequences;
  __Pyx_INCREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 626, __pyx_L1_error)
  }
  __pyx_t_8 = PyList_GET_SIZE(__pyx_t_2); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = ((__pyx_t_8 > 0) != 0);
  if (__pyx_t_12) {

    /* "gnocis/models.pyx":627
 * 				nNT = 0
 * 		if len(winSeqSet.sequences) > 0:
 * 			winScoreSet += self.getSequenceScores(winSeqSet)             # <<<<<<<<<<<<<<
 * 
 * 		posScores = {}
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_4, ((PyObject *)__pyx_v_winSeqSet)) : __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_v_winSeqSet));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_winScoreSet, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_winScoreSet, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "gnocis/models.pyx":626
 * 				i = 0
 * 				nNT = 0
 * 		if len(winSeqSet.sequences) > 0:             # <<<<<<<<<<<<<<
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 * 
 */
  }

  /* "gnocis/models.pyx":629
 * 			winScoreSet += self.getSequenceScores(winSeqSet)
 * 
 * 		posScores = {}             # <<<<<<<<<<<<<<
 * 		for seq in set(winPos.seq for winPos in winPosSet):
 * 			scores = [
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_posScores = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "gnocis/models.pyx":630
 * 
 * 		posScores = {}
 * 		for seq in set(winPos.seq for winPos in winPosSet):             # <<<<<<<<<<<<<<
 * 			scores = [
 * 				(int(winPos.start/self.windowStep), wScore)
 */
  __pyx_t_8 = 0;
  __pyx_t_2 = __pyx_pf_6gnocis_6models_13sequenceModel_27predictSequenceStreamCurves_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_set_iterator(__pyx_t_4, 1, (&__pyx_t_11), (&__pyx_t_6)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_XDECREF(__pyx_t_7);
  __pyx_t_7 = __pyx_t_2;
  __pyx_t_2 = 0;
  while (1) {
    __pyx_t_13 = __Pyx_set_iter_next(__pyx_t_7, __pyx_t_11, &__pyx_t_8, &__pyx_t_2, __pyx_t_6);
    if (unlikely(__pyx_t_13 == 0)) break;
    if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_seq, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":631
 * 		posScores = {}
 * 		for seq in set(winPos.seq for winPos in winPosSet):
 * 			scores = [             # <<<<<<<<<<<<<<
 * 				(int(winPos.start/self.windowStep), wScore)
 * 				for winPos, wScore in zip(winPosSet, winScoreSet)
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "gnocis/models.pyx":633
 * 			scores = [
 * 				(int(winPos.start/self.windowStep), wScore)
 * 				for winPos, wScore in zip(winPosSet, winScoreSet)             # <<<<<<<<<<<<<<
 * 				if winPos.seq == seq
 * 			]
 */
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 633, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_winPosSet);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_winPosSet);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_cur_scope->__pyx_v_winPosSet);
      __Pyx_INCREF(__pyx_v_winScoreSet);
      __Pyx_GIVEREF(__pyx_v_winScoreSet);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_winScoreSet);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
        __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_14 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_14 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 633, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 633, __pyx_L11_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_1); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 633, __pyx_L11_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_1); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 633, __pyx_L11_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_9(__pyx_t_4);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 633, __pyx_L11_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
          PyObject* sequence = __pyx_t_1;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 633, __pyx_L11_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          #else
          __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 633, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 633, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_15 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 633, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_16 = Py_TYPE(__pyx_t_15)->tp_iternext;
          index = 0; __pyx_t_3 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_3)) goto __pyx_L14_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_3);
          index = 1; __pyx_t_5 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_15), 2) < 0) __PYX_ERR(0, 633, __pyx_L11_error)
          __pyx_t_16 = NULL;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L15_unpacking_done;
          __pyx_L14_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_16 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 633, __pyx_L11_error)
          __pyx_L15_unpacking_done:;
        }
        if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6gnocis_7regions_region))))) __PYX_ERR(0, 633, __pyx_L11_error)
        __pyx_t_17 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_17 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 633, __pyx_L11_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF_SET(__pyx_9genexpr34__pyx_v_winPos, ((struct __pyx_obj_6gnocis_7regions_region *)__pyx_t_3));
        __pyx_t_3 = 0;
        __pyx_9genexpr34__pyx_v_wScore = __pyx_t_17;

        /* "gnocis/models.pyx":634
 * 				(int(winPos.start/self.windowStep), wScore)
 * 				for winPos, wScore in zip(winPosSet, winScoreSet)
 * 				if winPos.seq == seq             # <<<<<<<<<<<<<<
 * 			]
 * 			sScores = [ 0. for _ in range(len(scores)) ]
 */
        __pyx_t_12 = (__Pyx_PyUnicode_Equals(__pyx_9genexpr34__pyx_v_winPos->seq, __pyx_v_seq, Py_EQ)); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 634, __pyx_L11_error)
        if (__pyx_t_12) {

          /* "gnocis/models.pyx":632
 * 		for seq in set(winPos.seq for winPos in winPosSet):
 * 			scores = [
 * 				(int(winPos.start/self.windowStep), wScore)             # <<<<<<<<<<<<<<
 * 				for winPos, wScore in zip(winPosSet, winScoreSet)
 * 				if winPos.seq == seq
 */
          __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_9genexpr34__pyx_v_winPos->start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_windowStep); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 632, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 632, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 632, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyFloat_FromDouble(__pyx_9genexpr34__pyx_v_wScore); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 632, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
          __pyx_t_5 = 0;
          __pyx_t_3 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 631, __pyx_L11_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "gnocis/models.pyx":634
 * 				(int(winPos.start/self.windowStep), wScore)
 * 				for winPos, wScore in zip(winPosSet, winScoreSet)
 * 				if winPos.seq == seq             # <<<<<<<<<<<<<<
 * 			]
 * 			sScores = [ 0. for _ in range(len(scores)) ]
 */
        }

        /* "gnocis/models.pyx":633
 * 			scores = [
 * 				(int(winPos.start/self.windowStep), wScore)
 * 				for winPos, wScore in zip(winPosSet, winScoreSet)             # <<<<<<<<<<<<<<
 * 				if winPos.seq == seq
 * 			]
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF((PyObject *)__pyx_9genexpr34__pyx_v_winPos); __pyx_9genexpr34__pyx_v_winPos = 0;
      goto __pyx_L17_exit_scope;
      __pyx_L11_error:;
      __Pyx_XDECREF((PyObject *)__pyx_9genexpr34__pyx_v_winPos); __pyx_9genexpr34__pyx_v_winPos = 0;
      goto __pyx_L1_error;
      __pyx_L17_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_scores, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":636
 * 				if winPos.seq == seq
 * 			]
 * 			sScores = [ 0. for _ in range(len(scores)) ]             # <<<<<<<<<<<<<<
 * 			for i, s in scores:
 * 				sScores[i] = s
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_14 = PyList_GET_SIZE(__pyx_v_scores); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 636, __pyx_L1_error)
      __pyx_t_18 = __pyx_t_14;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_9genexpr35__pyx_v__ = __pyx_t_19;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_float_0_))) __PYX_ERR(0, 636, __pyx_L1_error)
      }
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_sScores, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":637
 * 			]
 * 			sScores = [ 0. for _ in range(len(scores)) ]
 * 			for i, s in scores:             # <<<<<<<<<<<<<<
 * 				sScores[i] = s
 * 			posScores[seq] = sScores
 */
    __pyx_t_2 = __pyx_v_scores; __Pyx_INCREF(__pyx_t_2); __pyx_t_14 = 0;
    for (;;) {
      if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 637, __pyx_L1_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 637, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_16 = Py_TYPE(__pyx_t_5)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_16(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L22_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_3 = __pyx_t_16(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L22_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_5), 2) < 0) __PYX_ERR(0, 637, __pyx_L1_error)
        __pyx_t_16 = NULL;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L23_unpacking_done;
        __pyx_L22_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_16 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 637, __pyx_L1_error)
        __pyx_L23_unpacking_done:;
      }
      __pyx_t_13 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_13 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 637, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_20 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_20 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 637, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_i = __pyx_t_13;
      __pyx_v_s = __pyx_t_20;

      /* "gnocis/models.pyx":638
 * 			sScores = [ 0. for _ in range(len(scores)) ]
 * 			for i, s in scores:
 * 				sScores[i] = s             # <<<<<<<<<<<<<<
 * 			posScores[seq] = sScores
 * 
 */
      __pyx_t_4 = PyFloat_FromDouble(__pyx_v_s); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_SetItemInt(__pyx_v_sScores, __pyx_v_i, __pyx_t_4, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0)) __PYX_ERR(0, 638, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "gnocis/models.pyx":637
 * 			]
 * 			sScores = [ 0. for _ in range(len(scores)) ]
 * 			for i, s in scores:             # <<<<<<<<<<<<<<
 * 				sScores[i] = s
 * 			posScores[seq] = sScores
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":639
 * 			for i, s in scores:
 * 				sScores[i] = s
 * 			posScores[seq] = sScores             # <<<<<<<<<<<<<<
 * 
 * 		ret = curves('Predictions', [
 */
    if (unlikely(PyDict_SetItem(__pyx_v_posScores, __pyx_v_seq, __pyx_v_sScores) < 0)) __PYX_ERR(0, 639, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "gnocis/models.pyx":641
 * 			posScores[seq] = sScores
 * 
 * 		ret = curves('Predictions', [             # <<<<<<<<<<<<<<
 * 			fixedStepCurve(
 * 				seq = seq,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_curves); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 641, __pyx_L26_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "gnocis/models.pyx":649
 * 				values = posScores[seq]
 * 			)
 * 			for seq in posScores             # <<<<<<<<<<<<<<
 * 		])
 * 		if self.threshold is not None:
 */
    __pyx_t_11 = 0;
    __pyx_t_1 = __Pyx_dict_iterator(__pyx_v_posScores, 1, ((PyObject *)NULL), (&__pyx_t_8), (&__pyx_t_6)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L26_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_1;
    __pyx_t_1 = 0;
    while (1) {
      __pyx_t_13 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_8, &__pyx_t_11, &__pyx_t_1, NULL, NULL, __pyx_t_6);
      if (unlikely(__pyx_t_13 == 0)) break;
      if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 649, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_9genexpr36__pyx_v_seq, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "gnocis/models.pyx":642
 * 
 * 		ret = curves('Predictions', [
 * 			fixedStepCurve(             # <<<<<<<<<<<<<<
 * 				seq = seq,
 * 				start = 0,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_fixedStepCurve); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "gnocis/models.pyx":643
 * 		ret = curves('Predictions', [
 * 			fixedStepCurve(
 * 				seq = seq,             # <<<<<<<<<<<<<<
 * 				start = 0,
 * 				span = self.windowSize,
 */
      __pyx_t_5 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_seq, __pyx_9genexpr36__pyx_v_seq) < 0) __PYX_ERR(0, 643, __pyx_L26_error)
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_start, __pyx_int_0) < 0) __PYX_ERR(0, 643, __pyx_L26_error)

      /* "gnocis/models.pyx":645
 * 				seq = seq,
 * 				start = 0,
 * 				span = self.windowSize,             # <<<<<<<<<<<<<<
 * 				step = self.windowStep,
 * 				values = posScores[seq]
 */
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_windowSize); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 645, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_span, __pyx_t_15) < 0) __PYX_ERR(0, 643, __pyx_L26_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

      /* "gnocis/models.pyx":646
 * 				start = 0,
 * 				span = self.windowSize,
 * 				step = self.windowStep,             # <<<<<<<<<<<<<<
 * 				values = posScores[seq]
 * 			)
 */
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_windowStep); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 646, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_step, __pyx_t_15) < 0) __PYX_ERR(0, 643, __pyx_L26_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

      /* "gnocis/models.pyx":647
 * 				span = self.windowSize,
 * 				step = self.windowStep,
 * 				values = posScores[seq]             # <<<<<<<<<<<<<<
 * 			)
 * 			for seq in posScores
 */
      __pyx_t_15 = __Pyx_PyDict_GetItem(__pyx_v_posScores, __pyx_9genexpr36__pyx_v_seq); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 647, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_values, __pyx_t_15) < 0) __PYX_ERR(0, 643, __pyx_L26_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

      /* "gnocis/models.pyx":642
 * 
 * 		ret = curves('Predictions', [
 * 			fixedStepCurve(             # <<<<<<<<<<<<<<
 * 				seq = seq,
 * 				start = 0,
 */
      __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 642, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_15))) __PYX_ERR(0, 641, __pyx_L26_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr36__pyx_v_seq); __pyx_9genexpr36__pyx_v_seq = 0;
    goto __pyx_L29_exit_scope;
    __pyx_L26_error:;
    __Pyx_XDECREF(__pyx_9genexpr36__pyx_v_seq); __pyx_9genexpr36__pyx_v_seq = 0;
    goto __pyx_L1_error;
    __pyx_L29_exit_scope:;
  } /* exit inner scope */
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_n_u_Predictions, __pyx_t_4};
    __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_n_u_Predictions, __pyx_t_4};
    __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_15 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_n_u_Predictions);
    __Pyx_GIVEREF(__pyx_n_u_Predictions);
    PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_6, __pyx_n_u_Predictions);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_6, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_15, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ret = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "gnocis/models.pyx":651
 * 			for seq in posScores
 * 		])
 * 		if self.threshold is not None:             # <<<<<<<<<<<<<<
 * 			ret = ret.threshold(self.threshold)
 * 		return ret
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_threshold); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_12 = (__pyx_t_7 != Py_None);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_21 = (__pyx_t_12 != 0);
  if (__pyx_t_21) {

    /* "gnocis/models.pyx":652
 * 		])
 * 		if self.threshold is not None:
 * 			ret = ret.threshold(self.threshold)             # <<<<<<<<<<<<<<
 * 		return ret
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ret, __pyx_n_s_threshold); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_threshold); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_15);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "gnocis/models.pyx":651
 * 			for seq in posScores
 * 		])
 * 		if self.threshold is not None:             # <<<<<<<<<<<<<<
 * 			ret = ret.threshold(self.threshold)
 * 		return ret
 */
  }

  /* "gnocis/models.pyx":653
 * 		if self.threshold is not None:
 * 			ret = ret.threshold(self.threshold)
 * 		return ret             # <<<<<<<<<<<<<<
 * 
 * 	# Predicts core regions.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "gnocis/models.pyx":595
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamCurves(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.predictSequenceStreamCurves", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_winScoreSet);
  __Pyx_XDECREF((PyObject *)__pyx_v_winSeq);
  __Pyx_XDECREF((PyObject *)__pyx_v_winSeqSet);
  __Pyx_XDECREF(__pyx_v_sScores);
  __Pyx_XDECREF(__pyx_v_posScores);
  __Pyx_XDECREF(__pyx_v_seq);
  __Pyx_XDECREF(__pyx_v_scores);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_13sequenceModel_27predictSequenceStreamCurves_2generator9);
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr34__pyx_v_winPos);
  __Pyx_XDECREF(__pyx_9genexpr36__pyx_v_seq);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":656
 * 
 * 	# Predicts core regions.
 * 	def predictCore(self, genome):             # <<<<<<<<<<<<<<
 * 		""" Predicts cores. Not implemented by default. Models that use core prediction should implement this.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_35predictCore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_34predictCore[] = " Predicts cores. Not implemented by default. Models that use core prediction should implement this.\n\t\t\n\t\t:param genome: Genome.\n\t\t:type genome: genome\n\t\t\n\t\t:return: Predicted core regions if implemented, or None otherwise\n\t\t:rtype: regions\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_35predictCore = {"predictCore", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_35predictCore, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_34predictCore};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_35predictCore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_genome = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predictCore (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_genome,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_genome)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("predictCore", 1, 2, 2, 1); __PYX_ERR(0, 656, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "predictCore") < 0)) __PYX_ERR(0, 656, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_genome = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("predictCore", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 656, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.predictCore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_34predictCore(__pyx_self, __pyx_v_self, __pyx_v_genome);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_34predictCore(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_genome) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predictCore", 0);

  /* "gnocis/models.pyx":665
 * 		:rtype: regions
 * 		"""
 * 		return None             # <<<<<<<<<<<<<<
 * 
 * 	# Short-hand for predictSequenceStreamRegions.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "gnocis/models.pyx":656
 * 
 * 	# Predicts core regions.
 * 	def predictCore(self, genome):             # <<<<<<<<<<<<<<
 * 		""" Predicts cores. Not implemented by default. Models that use core prediction should implement this.
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":668
 * 
 * 	# Short-hand for predictSequenceStreamRegions.
 * 	def predict(self, stream, curves = True):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_37predict(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_36predict[] = " Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.\n\t\t\n\t\t:param stream: Sequence material that the model is applied to for prediction.\n\t\t:type stream: sequences/sequenceStream/str\n\t\t\n\t\t:return: Set of non-overlapping (merged) predicted regions\n\t\t:rtype: regions\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_37predict = {"predict", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_37predict, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_36predict};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_37predict(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_stream = 0;
  PyObject *__pyx_v_curves = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predict (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_stream,&__pyx_n_s_curves,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("predict", 0, 2, 3, 1); __PYX_ERR(0, 668, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_curves);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "predict") < 0)) __PYX_ERR(0, 668, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_stream = values[1];
    __pyx_v_curves = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("predict", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 668, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_36predict(__pyx_self, __pyx_v_self, __pyx_v_stream, __pyx_v_curves);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_36predict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_stream, PyObject *__pyx_v_curves) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("predict", 0);

  /* "gnocis/models.pyx":677
 * 		:rtype: regions
 * 		"""
 * 		if curves: return self.predictSequenceStreamCurves(stream)             # <<<<<<<<<<<<<<
 * 		return self.predictSequenceStreamRegions(stream)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_curves); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 677, __pyx_L1_error)
  if (__pyx_t_1) {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictSequenceStreamCurves); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_stream) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_stream);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "gnocis/models.pyx":678
 * 		"""
 * 		if curves: return self.predictSequenceStreamCurves(stream)
 * 		return self.predictSequenceStreamRegions(stream)             # <<<<<<<<<<<<<<
 * 
 * 	# Short-hand for getSequenceScores/scoreSequence
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictSequenceStreamRegions); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_stream) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_stream);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":668
 * 
 * 	# Short-hand for predictSequenceStreamRegions.
 * 	def predict(self, stream, curves = True):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":681
 * 
 * 	# Short-hand for getSequenceScores/scoreSequence
 * 	def score(self, target, asTable = False):             # <<<<<<<<<<<<<<
 * 		""" Scores a sequence of set of sequences.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_39score(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_38score[] = " Scores a sequence of set of sequences.\n\t\t\n\t\t:param target: Sequence(s) to score.\n\t\t:param asTable: If True, a table will be output. Otherwise, a list.\n\t\t\n\t\t:type target: sequence/sequences/sequenceStream\n\t\t:type asTable: bool\n\t\t\n\t\t:return: Score or list of scores\n\t\t:rtype: float/list\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_39score = {"score", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_39score, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_38score};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_39score(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_target = 0;
  PyObject *__pyx_v_asTable = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("score (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_target,&__pyx_n_s_asTable,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("score", 0, 2, 3, 1); __PYX_ERR(0, 681, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_asTable);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "score") < 0)) __PYX_ERR(0, 681, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_target = values[1];
    __pyx_v_asTable = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("score", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 681, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.score", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_38score(__pyx_self, __pyx_v_self, __pyx_v_target, __pyx_v_asTable);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_38score(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_target, PyObject *__pyx_v_asTable) {
  PyObject *__pyx_9genexpr37__pyx_v_seq = NULL;
  PyObject *__pyx_9genexpr37__pyx_v_score = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("score", 0);

  /* "gnocis/models.pyx":693
 * 		:rtype: float/list
 * 		"""
 * 		if isinstance(target, sequence):             # <<<<<<<<<<<<<<
 * 			return self.scoreSequence(target)
 * 		if asTable:
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_target, __pyx_ptype_6gnocis_9sequences_sequence); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "gnocis/models.pyx":694
 * 		"""
 * 		if isinstance(target, sequence):
 * 			return self.scoreSequence(target)             # <<<<<<<<<<<<<<
 * 		if asTable:
 * 			return nctable(
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_scoreSequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_target) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_target);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "gnocis/models.pyx":693
 * 		:rtype: float/list
 * 		"""
 * 		if isinstance(target, sequence):             # <<<<<<<<<<<<<<
 * 			return self.scoreSequence(target)
 * 		if asTable:
 */
  }

  /* "gnocis/models.pyx":695
 * 		if isinstance(target, sequence):
 * 			return self.scoreSequence(target)
 * 		if asTable:             # <<<<<<<<<<<<<<
 * 			return nctable(
 * 				'Scored sequences - Model: ' + self.name + '; Sequences: ' + target.name,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_asTable); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 695, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "gnocis/models.pyx":696
 * 			return self.scoreSequence(target)
 * 		if asTable:
 * 			return nctable(             # <<<<<<<<<<<<<<
 * 				'Scored sequences - Model: ' + self.name + '; Sequences: ' + target.name,
 * 				[
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_nctable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "gnocis/models.pyx":697
 * 		if asTable:
 * 			return nctable(
 * 				'Scored sequences - Model: ' + self.name + '; Sequences: ' + target.name,             # <<<<<<<<<<<<<<
 * 				[
 * 					{
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyNumber_Add(__pyx_kp_u_Scored_sequences_Model, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Add(__pyx_t_6, __pyx_kp_u_Sequences); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_target, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyNumber_Add(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    { /* enter inner scope */

      /* "gnocis/models.pyx":698
 * 			return nctable(
 * 				'Scored sequences - Model: ' + self.name + '; Sequences: ' + target.name,
 * 				[             # <<<<<<<<<<<<<<
 * 					{
 * 						'Name': seq.name,
 */
      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 698, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "gnocis/models.pyx":703
 * 						'Score': score,
 * 					}
 * 					for seq, score in zip(target, self.getSequenceScores(target))             # <<<<<<<<<<<<<<
 * 				],
 * 			)
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 703, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_v_target) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_target);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 703, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 703, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_v_target);
      __Pyx_GIVEREF(__pyx_v_target);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_target);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 703, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
        __pyx_t_8 = __pyx_t_5; __Pyx_INCREF(__pyx_t_8); __pyx_t_10 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_10 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 703, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 703, __pyx_L7_error)
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 703, __pyx_L7_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_8, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 703, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          } else {
            if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_10); __Pyx_INCREF(__pyx_t_5); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 703, __pyx_L7_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_8, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 703, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          }
        } else {
          __pyx_t_5 = __pyx_t_11(__pyx_t_8);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 703, __pyx_L7_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_5);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
          PyObject* sequence = __pyx_t_5;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 703, __pyx_L7_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_12 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_12 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_12);
          #else
          __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 703, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_12 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 703, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_12);
          #endif
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_13 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 703, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_14 = Py_TYPE(__pyx_t_13)->tp_iternext;
          index = 0; __pyx_t_9 = __pyx_t_14(__pyx_t_13); if (unlikely(!__pyx_t_9)) goto __pyx_L10_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_9);
          index = 1; __pyx_t_12 = __pyx_t_14(__pyx_t_13); if (unlikely(!__pyx_t_12)) goto __pyx_L10_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_12);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_13), 2) < 0) __PYX_ERR(0, 703, __pyx_L7_error)
          __pyx_t_14 = NULL;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L11_unpacking_done;
          __pyx_L10_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_14 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 703, __pyx_L7_error)
          __pyx_L11_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr37__pyx_v_seq, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_XDECREF_SET(__pyx_9genexpr37__pyx_v_score, __pyx_t_12);
        __pyx_t_12 = 0;

        /* "gnocis/models.pyx":700
 * 				[
 * 					{
 * 						'Name': seq.name,             # <<<<<<<<<<<<<<
 * 						'Score': score,
 * 					}
 */
        __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 700, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr37__pyx_v_seq, __pyx_n_s_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 700, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_u_Name, __pyx_t_12) < 0) __PYX_ERR(0, 700, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

        /* "gnocis/models.pyx":701
 * 					{
 * 						'Name': seq.name,
 * 						'Score': score,             # <<<<<<<<<<<<<<
 * 					}
 * 					for seq, score in zip(target, self.getSequenceScores(target))
 */
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_u_Score, __pyx_9genexpr37__pyx_v_score) < 0) __PYX_ERR(0, 700, __pyx_L7_error)
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 698, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "gnocis/models.pyx":703
 * 						'Score': score,
 * 					}
 * 					for seq, score in zip(target, self.getSequenceScores(target))             # <<<<<<<<<<<<<<
 * 				],
 * 			)
 */
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_9genexpr37__pyx_v_score); __pyx_9genexpr37__pyx_v_score = 0;
      __Pyx_XDECREF(__pyx_9genexpr37__pyx_v_seq); __pyx_9genexpr37__pyx_v_seq = 0;
      goto __pyx_L12_exit_scope;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_9genexpr37__pyx_v_score); __pyx_9genexpr37__pyx_v_score = 0;
      __Pyx_XDECREF(__pyx_9genexpr37__pyx_v_seq); __pyx_9genexpr37__pyx_v_seq = 0;
      goto __pyx_L1_error;
      __pyx_L12_exit_scope:;
    } /* exit inner scope */
    __pyx_t_8 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_7, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 696, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_7, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 696, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_15, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_15, __pyx_t_6);
      __pyx_t_7 = 0;
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 696, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "gnocis/models.pyx":695
 * 		if isinstance(target, sequence):
 * 			return self.scoreSequence(target)
 * 		if asTable:             # <<<<<<<<<<<<<<
 * 			return nctable(
 * 				'Scored sequences - Model: ' + self.name + '; Sequences: ' + target.name,
 */
  }

  /* "gnocis/models.pyx":706
 * 				],
 * 			)
 * 		return self.getSequenceScores(target)             # <<<<<<<<<<<<<<
 * 
 * 	# Short-hand for training
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getSequenceScores); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_target) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_target);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":681
 * 
 * 	# Short-hand for getSequenceScores/scoreSequence
 * 	def score(self, target, asTable = False):             # <<<<<<<<<<<<<<
 * 		""" Scores a sequence of set of sequences.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.score", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr37__pyx_v_seq);
  __Pyx_XDECREF(__pyx_9genexpr37__pyx_v_score);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":709
 * 
 * 	# Short-hand for training
 * 	def train(self, ts):             # <<<<<<<<<<<<<<
 * 		""" Scores a sequence of set of sequences.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_41train(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gnocis_6models_13sequenceModel_40train[] = " Scores a sequence of set of sequences.\n\t\t\n\t\t:param ts: Training set.\n\t\t\n\t\t:type ts: sequences\n\t\t\n\t\t:return: Trained model\n\t\t:rtype: sequenceModel\n\t\t";
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_41train = {"train", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_41train, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gnocis_6models_13sequenceModel_40train};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_41train(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ts = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("train (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ts,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ts)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("train", 1, 2, 2, 1); __PYX_ERR(0, 709, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "train") < 0)) __PYX_ERR(0, 709, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ts = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("train", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 709, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.train", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_40train(__pyx_self, __pyx_v_self, __pyx_v_ts);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_40train(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ts) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("train", 0);

  /* "gnocis/models.pyx":719
 * 		:rtype: sequenceModel
 * 		"""
 * 		return self.getTrainer()(ts)             # <<<<<<<<<<<<<<
 * 
 * 	# Prints out test statistics.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getTrainer); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_ts) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_ts);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":709
 * 
 * 	# Short-hand for training
 * 	def train(self, ts):             # <<<<<<<<<<<<<<
 * 		""" Scores a sequence of set of sequences.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.train", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":722
 * 
 * 	# Prints out test statistics.
 * 	def printTestStatistics(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		printValidationStatistics( self.getValidationStatistics(seqs, labelPositive = labelPositive, labelNegative = labelNegative) )
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_50__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_43printTestStatistics(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_13sequenceModel_43printTestStatistics = {"printTestStatistics", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_13sequenceModel_43printTestStatistics, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_13sequenceModel_43printTestStatistics(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seqs = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("printTestStatistics (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seqs,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults8 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[3] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seqs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("printTestStatistics", 0, 2, 4, 1); __PYX_ERR(0, 722, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "printTestStatistics") < 0)) __PYX_ERR(0, 722, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_seqs = values[1];
    __pyx_v_labelPositive = values[2];
    __pyx_v_labelNegative = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("printTestStatistics", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 722, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModel.printTestStatistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_13sequenceModel_42printTestStatistics(__pyx_self, __pyx_v_self, __pyx_v_seqs, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_13sequenceModel_42printTestStatistics(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seqs, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("printTestStatistics", 0);

  /* "gnocis/models.pyx":723
 * 	# Prints out test statistics.
 * 	def printTestStatistics(self, seqs, labelPositive = positive, labelNegative = negative):
 * 		printValidationStatistics( self.getValidationStatistics(seqs, labelPositive = labelPositive, labelNegative = labelNegative) )             # <<<<<<<<<<<<<<
 * 
 * # Uniformly weighted model
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_printValidationStatistics); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getValidationStatistics); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_seqs);
  __Pyx_GIVEREF(__pyx_v_seqs);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_seqs);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 723, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 723, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":722
 * 
 * 	# Prints out test statistics.
 * 	def printTestStatistics(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		printValidationStatistics( self.getValidationStatistics(seqs, labelPositive = labelPositive, labelNegative = labelNegative) )
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.models.sequenceModel.printTestStatistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":741
 * 	"""
 * 
 * 	def __init__(self, name, features, windowSize, windowStep):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18sequenceModelDummy_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_18sequenceModelDummy_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_features = 0;
  PyObject *__pyx_v_windowSize = 0;
  PyObject *__pyx_v_windowStep = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_features,&__pyx_n_s_windowSize,&__pyx_n_s_windowStep,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 1); __PYX_ERR(0, 741, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 2); __PYX_ERR(0, 741, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 3); __PYX_ERR(0, 741, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowStep)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 4); __PYX_ERR(0, 741, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 741, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_features = values[2];
    __pyx_v_windowSize = values[3];
    __pyx_v_windowStep = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 741, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModelDummy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_18sequenceModelDummy___init__(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_features, __pyx_v_windowSize, __pyx_v_windowStep);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_features, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gnocis/models.pyx":742
 * 
 * 	def __init__(self, name, features, windowSize, windowStep):
 * 		super().__init__(name)             # <<<<<<<<<<<<<<
 * 		self.threshold = 0.0
 * 		self.features = features
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 742, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_name);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":743
 * 	def __init__(self, name, features, windowSize, windowStep):
 * 		super().__init__(name)
 * 		self.threshold = 0.0             # <<<<<<<<<<<<<<
 * 		self.features = features
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_threshold, __pyx_float_0_0) < 0) __PYX_ERR(0, 743, __pyx_L1_error)

  /* "gnocis/models.pyx":744
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 * 		self.features = features             # <<<<<<<<<<<<<<
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_features, __pyx_v_features) < 0) __PYX_ERR(0, 744, __pyx_L1_error)

  /* "gnocis/models.pyx":745
 * 		self.threshold = 0.0
 * 		self.features = features
 * 		self.windowSize, self.windowStep = windowSize, windowStep             # <<<<<<<<<<<<<<
 * 
 * 	def __str__(self):
 */
  __pyx_t_1 = __pyx_v_windowSize;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_v_windowStep;
  __Pyx_INCREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_windowSize, __pyx_t_1) < 0) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_windowStep, __pyx_t_3) < 0) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":741
 * 	"""
 * 
 * 	def __init__(self, name, features, windowSize, windowStep):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.models.sequenceModelDummy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":747
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Dummy model<Features: %s>'%(str(self.features))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_3__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18sequenceModelDummy_3__str__ = {"__str__", (PyCFunction)__pyx_pw_6gnocis_6models_18sequenceModelDummy_3__str__, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_3__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_18sequenceModelDummy_2__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "gnocis/models.pyx":748
 * 
 * 	def __str__(self):
 * 		return 'Dummy model<Features: %s>'%(str(self.features))             # <<<<<<<<<<<<<<
 * 
 * 	def __repr__(self): return self.__str__()
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_features); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyUnicode_Format(__pyx_kp_u_Dummy_model_Features_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":747
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Dummy model<Features: %s>'%(str(self.features))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.sequenceModelDummy.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":750
 * 		return 'Dummy model<Features: %s>'%(str(self.features))
 * 
 * 	def __repr__(self): return self.__str__()             # <<<<<<<<<<<<<<
 * 
 * 	def getTrainer(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_5__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18sequenceModelDummy_5__repr__ = {"__repr__", (PyCFunction)__pyx_pw_6gnocis_6models_18sequenceModelDummy_5__repr__, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_5__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_18sequenceModelDummy_4__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy_4__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.models.sequenceModelDummy.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":752
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_7getTrainer(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18sequenceModelDummy_7getTrainer = {"getTrainer", (PyCFunction)__pyx_pw_6gnocis_6models_18sequenceModelDummy_7getTrainer, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_7getTrainer(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getTrainer (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_18sequenceModelDummy_6getTrainer(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":753
 * 
 * 	def getTrainer(self):
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)             # <<<<<<<<<<<<<<
 * 
 * 	def scoreWindow(self, seq):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_10getTrainer_lambda12(PyObject *__pyx_self, PyObject *__pyx_v_ts); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18sequenceModelDummy_10getTrainer_lambda12 = {"lambda12", (PyCFunction)__pyx_pw_6gnocis_6models_18sequenceModelDummy_10getTrainer_lambda12, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_10getTrainer_lambda12(PyObject *__pyx_self, PyObject *__pyx_v_ts) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda12 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda12(__pyx_self, ((PyObject *)__pyx_v_ts));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda12(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_ts) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *__pyx_cur_scope;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda12", 0);
  __pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_latin1_sequenceModelDummy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 753, __pyx_L1_error) }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 753, __pyx_L1_error) }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_features); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 753, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_windowSize); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 753, __pyx_L1_error) }
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_windowStep); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("gnocis.models.sequenceModelDummy.getTrainer.lambda12", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":752
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy_6getTrainer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getTrainer", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_17_getTrainer(__pyx_ptype_6gnocis_6models___pyx_scope_struct_17_getTrainer, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 752, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "gnocis/models.pyx":753
 * 
 * 	def getTrainer(self):
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)             # <<<<<<<<<<<<<<
 * 
 * 	def scoreWindow(self, seq):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18sequenceModelDummy_10getTrainer_lambda12, 0, __pyx_n_s_sequenceModelDummy_getTrainer_lo, ((PyObject*)__pyx_cur_scope), __pyx_n_s_gnocis_models, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":752
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.models.sequenceModelDummy.getTrainer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":755
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 * 	def scoreWindow(self, seq):             # <<<<<<<<<<<<<<
 * 		cdef list fv
 * 		cdef float ret, v
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_9scoreWindow(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18sequenceModelDummy_9scoreWindow = {"scoreWindow", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_18sequenceModelDummy_9scoreWindow, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_18sequenceModelDummy_9scoreWindow(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seq = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scoreWindow (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seq,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scoreWindow", 1, 2, 2, 1); __PYX_ERR(0, 755, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scoreWindow") < 0)) __PYX_ERR(0, 755, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_seq = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scoreWindow", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 755, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModelDummy.scoreWindow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_18sequenceModelDummy_8scoreWindow(__pyx_self, __pyx_v_self, __pyx_v_seq);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18sequenceModelDummy_8scoreWindow(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seq) {
  PyObject *__pyx_v_fv = 0;
  float __pyx_v_ret;
  float __pyx_v_v;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  float __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("scoreWindow", 0);

  /* "gnocis/models.pyx":758
 * 		cdef list fv
 * 		cdef float ret, v
 * 		fv = self.features.getAll(seq)             # <<<<<<<<<<<<<<
 * 		ret = 0.0
 * 		for v in fv:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_features); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_getAll); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_seq) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_seq);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 758, __pyx_L1_error)
  __pyx_v_fv = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":759
 * 		cdef float ret, v
 * 		fv = self.features.getAll(seq)
 * 		ret = 0.0             # <<<<<<<<<<<<<<
 * 		for v in fv:
 * 			ret += v
 */
  __pyx_v_ret = 0.0;

  /* "gnocis/models.pyx":760
 * 		fv = self.features.getAll(seq)
 * 		ret = 0.0
 * 		for v in fv:             # <<<<<<<<<<<<<<
 * 			ret += v
 * 		return ret
 */
  if (unlikely(__pyx_v_fv == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 760, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_fv; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 760, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __pyx_t_5 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_5 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_v = __pyx_t_5;

    /* "gnocis/models.pyx":761
 * 		ret = 0.0
 * 		for v in fv:
 * 			ret += v             # <<<<<<<<<<<<<<
 * 		return ret
 * 
 */
    __pyx_v_ret = (__pyx_v_ret + __pyx_v_v);

    /* "gnocis/models.pyx":760
 * 		fv = self.features.getAll(seq)
 * 		ret = 0.0
 * 		for v in fv:             # <<<<<<<<<<<<<<
 * 			ret += v
 * 		return ret
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":762
 * 		for v in fv:
 * 			ret += v
 * 		return ret             # <<<<<<<<<<<<<<
 * 
 * # Log-odds model
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_ret); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":755
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 * 	def scoreWindow(self, seq):             # <<<<<<<<<<<<<<
 * 		cdef list fv
 * 		cdef float ret, v
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.models.sequenceModelDummy.scoreWindow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fv);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":786
 * 	"""
 * 
 * 	def __init__(self, name, features, trainingSet, windowSize, windowStep, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */

static PyObject *__pyx_pf_6gnocis_6models_52__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_self)->__pyx_arg_labelNegative);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_features = 0;
  PyObject *__pyx_v_trainingSet = 0;
  PyObject *__pyx_v_windowSize = 0;
  PyObject *__pyx_v_windowStep = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_features,&__pyx_n_s_trainingSet,&__pyx_n_s_windowSize,&__pyx_n_s_windowStep,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    __pyx_defaults9 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_self);
    values[6] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[7] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 6, 8, 1); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_features)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 6, 8, 2); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trainingSet)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 6, 8, 3); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 6, 8, 4); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowStep)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 6, 8, 5); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 786, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_features = values[2];
    __pyx_v_trainingSet = values[3];
    __pyx_v_windowSize = values[4];
    __pyx_v_windowStep = values[5];
    __pyx_v_labelPositive = values[6];
    __pyx_v_labelNegative = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 6, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 786, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_20sequenceModelLogOdds___init__(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_features, __pyx_v_trainingSet, __pyx_v_windowSize, __pyx_v_windowStep, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":798
 * 		self.sortedweights = []
 * 		for f in features:
 * 			vPos = sum( f.get(s) for s in trainingPositives )             # <<<<<<<<<<<<<<
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )
 * 			if vPos == 0.0 or vNeg == 0.0:
 */

static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_19_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_19_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 798, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___2generator10, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_sequenceModelLogOdds___init___lo, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 798, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingPositives)) { __Pyx_RaiseClosureNameError("trainingPositives"); __PYX_ERR(0, 798, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingPositives)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingPositives)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingPositives; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingPositives); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 798, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f)) { __Pyx_RaiseClosureNameError("f"); __PYX_ERR(0, 798, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f, __pyx_n_s_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_cur_scope->__pyx_v_s) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 798, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___5generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":799
 * 		for f in features:
 * 			vPos = sum( f.get(s) for s in trainingPositives )
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )             # <<<<<<<<<<<<<<
 * 			if vPos == 0.0 or vNeg == 0.0:
 * 				vPos += 1.0
 */

static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8__init___3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_20_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_20_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 799, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___5generator11, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_sequenceModelLogOdds___init___lo, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___5generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 799, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingNegatives)) { __Pyx_RaiseClosureNameError("trainingNegatives"); __PYX_ERR(0, 799, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingNegatives)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingNegatives)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingNegatives; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_trainingNegatives); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 799, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 799, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 799, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 799, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 799, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 799, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f)) { __Pyx_RaiseClosureNameError("f"); __PYX_ERR(0, 799, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_f, __pyx_n_s_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_cur_scope->__pyx_v_s) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 799, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":786
 * 	"""
 * 
 * 	def __init__(self, name, features, trainingSet, windowSize, windowStep, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */

static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_features, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_vPos = NULL;
  PyObject *__pyx_v_vNeg = NULL;
  PyObject *__pyx_v_weight = NULL;
  PyObject *__pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___2generator10 = 0;
  PyObject *__pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___5generator11 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_18___init__(__pyx_ptype_6gnocis_6models___pyx_scope_struct_18___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 786, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/models.pyx":787
 * 
 * 	def __init__(self, name, features, trainingSet, windowSize, windowStep, labelPositive = positive, labelNegative = negative):
 * 		super().__init__(name)             # <<<<<<<<<<<<<<
 * 		self.threshold = 0.0
 * 		self.features = features
 */
  __pyx_t_2 = __Pyx_CyFunction_GetClassObj(__pyx_self);
  if (!__pyx_t_2) { PyErr_SetString(PyExc_SystemError, "super(): empty __class__ cell"); __PYX_ERR(0, 787, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_name);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":788
 * 	def __init__(self, name, features, trainingSet, windowSize, windowStep, labelPositive = positive, labelNegative = negative):
 * 		super().__init__(name)
 * 		self.threshold = 0.0             # <<<<<<<<<<<<<<
 * 		self.features = features
 * 		self.trainingSet = trainingSet
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_threshold, __pyx_float_0_0) < 0) __PYX_ERR(0, 788, __pyx_L1_error)

  /* "gnocis/models.pyx":789
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 * 		self.features = features             # <<<<<<<<<<<<<<
 * 		self.trainingSet = trainingSet
 * 		trainingPositives, trainingNegatives = trainingSet.withLabel([ labelPositive, labelNegative ])
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_features, __pyx_v_features) < 0) __PYX_ERR(0, 789, __pyx_L1_error)

  /* "gnocis/models.pyx":790
 * 		self.threshold = 0.0
 * 		self.features = features
 * 		self.trainingSet = trainingSet             # <<<<<<<<<<<<<<
 * 		trainingPositives, trainingNegatives = trainingSet.withLabel([ labelPositive, labelNegative ])
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_trainingSet, __pyx_v_trainingSet) < 0) __PYX_ERR(0, 790, __pyx_L1_error)

  /* "gnocis/models.pyx":791
 * 		self.features = features
 * 		self.trainingSet = trainingSet
 * 		trainingPositives, trainingNegatives = trainingSet.withLabel([ labelPositive, labelNegative ])             # <<<<<<<<<<<<<<
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 * 		self.trainingPositives, self.trainingNegatives = trainingPositives, trainingNegatives
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_trainingSet, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_labelPositive);
  __Pyx_GIVEREF(__pyx_v_labelPositive);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_labelPositive);
  __Pyx_INCREF(__pyx_v_labelNegative);
  __Pyx_GIVEREF(__pyx_v_labelNegative);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_labelNegative);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 791, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 791, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 791, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_trainingPositives = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_trainingNegatives = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "gnocis/models.pyx":792
 * 		self.trainingSet = trainingSet
 * 		trainingPositives, trainingNegatives = trainingSet.withLabel([ labelPositive, labelNegative ])
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative             # <<<<<<<<<<<<<<
 * 		self.trainingPositives, self.trainingNegatives = trainingPositives, trainingNegatives
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 */
  __pyx_t_1 = __pyx_v_labelPositive;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_labelNegative;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_labelPositive, __pyx_t_1) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_labelNegative, __pyx_t_2) < 0) __PYX_ERR(0, 792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":793
 * 		trainingPositives, trainingNegatives = trainingSet.withLabel([ labelPositive, labelNegative ])
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 * 		self.trainingPositives, self.trainingNegatives = trainingPositives, trainingNegatives             # <<<<<<<<<<<<<<
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 * 		self.weights = {}
 */
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_trainingPositives;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_trainingNegatives;
  __Pyx_INCREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_trainingPositives, __pyx_t_2) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_trainingNegatives, __pyx_t_1) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":794
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 * 		self.trainingPositives, self.trainingNegatives = trainingPositives, trainingNegatives
 * 		self.windowSize, self.windowStep = windowSize, windowStep             # <<<<<<<<<<<<<<
 * 		self.weights = {}
 * 		self.sortedweights = []
 */
  __pyx_t_1 = __pyx_v_windowSize;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_windowStep;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_windowSize, __pyx_t_1) < 0) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_windowStep, __pyx_t_2) < 0) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":795
 * 		self.trainingPositives, self.trainingNegatives = trainingPositives, trainingNegatives
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 * 		self.weights = {}             # <<<<<<<<<<<<<<
 * 		self.sortedweights = []
 * 		for f in features:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_weights, __pyx_t_2) < 0) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":796
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 * 		self.weights = {}
 * 		self.sortedweights = []             # <<<<<<<<<<<<<<
 * 		for f in features:
 * 			vPos = sum( f.get(s) for s in trainingPositives )
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_sortedweights, __pyx_t_2) < 0) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":797
 * 		self.weights = {}
 * 		self.sortedweights = []
 * 		for f in features:             # <<<<<<<<<<<<<<
 * 			vPos = sum( f.get(s) for s in trainingPositives )
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )
 */
  if (likely(PyList_CheckExact(__pyx_v_features)) || PyTuple_CheckExact(__pyx_v_features)) {
    __pyx_t_2 = __pyx_v_features; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_features); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 797, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 797, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 797, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 797, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_f);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":798
 * 		self.sortedweights = []
 * 		for f in features:
 * 			vPos = sum( f.get(s) for s in trainingPositives )             # <<<<<<<<<<<<<<
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )
 * 			if vPos == 0.0 or vNeg == 0.0:
 */
    __pyx_t_1 = __pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_vPos, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "gnocis/models.pyx":799
 * 		for f in features:
 * 			vPos = sum( f.get(s) for s in trainingPositives )
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )             # <<<<<<<<<<<<<<
 * 			if vPos == 0.0 or vNeg == 0.0:
 * 				vPos += 1.0
 */
    __pyx_t_3 = __pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8__init___3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_vNeg, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":800
 * 			vPos = sum( f.get(s) for s in trainingPositives )
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )
 * 			if vPos == 0.0 or vNeg == 0.0:             # <<<<<<<<<<<<<<
 * 				vPos += 1.0
 * 				vNeg += 1.0
 */
    __pyx_t_1 = __Pyx_PyFloat_EqObjC(__pyx_v_vPos, __pyx_float_0_0, 0.0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyFloat_EqObjC(__pyx_v_vNeg, __pyx_float_0_0, 0.0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __pyx_t_9;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_8) {

      /* "gnocis/models.pyx":801
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )
 * 			if vPos == 0.0 or vNeg == 0.0:
 * 				vPos += 1.0             # <<<<<<<<<<<<<<
 * 				vNeg += 1.0
 * 			weight = math.log(vPos) - math.log(float(len(trainingPositives))) - (math.log(vNeg) - math.log(float(len(trainingNegatives))))
 */
      __pyx_t_1 = __Pyx_PyFloat_AddObjC(__pyx_v_vPos, __pyx_float_1_0, 1.0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_vPos, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "gnocis/models.pyx":802
 * 			if vPos == 0.0 or vNeg == 0.0:
 * 				vPos += 1.0
 * 				vNeg += 1.0             # <<<<<<<<<<<<<<
 * 			weight = math.log(vPos) - math.log(float(len(trainingPositives))) - (math.log(vNeg) - math.log(float(len(trainingNegatives))))
 * 			self.weights[f] = weight
 */
      __pyx_t_1 = __Pyx_PyFloat_AddObjC(__pyx_v_vNeg, __pyx_float_1_0, 1.0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 802, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_vNeg, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "gnocis/models.pyx":800
 * 			vPos = sum( f.get(s) for s in trainingPositives )
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )
 * 			if vPos == 0.0 or vNeg == 0.0:             # <<<<<<<<<<<<<<
 * 				vPos += 1.0
 * 				vNeg += 1.0
 */
    }

    /* "gnocis/models.pyx":803
 * 				vPos += 1.0
 * 				vNeg += 1.0
 * 			weight = math.log(vPos) - math.log(float(len(trainingPositives))) - (math.log(vNeg) - math.log(float(len(trainingNegatives))))             # <<<<<<<<<<<<<<
 * 			self.weights[f] = weight
 * 			self.sortedweights.append(weight)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_math); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_vPos) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_vPos);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_math); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_log); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __pyx_cur_scope->__pyx_v_trainingPositives;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_11 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyFloat_FromDouble(((double)__pyx_t_11)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_4 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_12, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyNumber_Subtract(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_math); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_vNeg) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_vNeg);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_math); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_log); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_trainingNegatives;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_11 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyFloat_FromDouble(((double)__pyx_t_11)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_3 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = PyNumber_Subtract(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_10, __pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF_SET(__pyx_v_weight, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "gnocis/models.pyx":804
 * 				vNeg += 1.0
 * 			weight = math.log(vPos) - math.log(float(len(trainingPositives))) - (math.log(vNeg) - math.log(float(len(trainingNegatives))))
 * 			self.weights[f] = weight             # <<<<<<<<<<<<<<
 * 			self.sortedweights.append(weight)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_weights); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PyObject_SetItem(__pyx_t_3, __pyx_cur_scope->__pyx_v_f, __pyx_v_weight) < 0)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "gnocis/models.pyx":805
 * 			weight = math.log(vPos) - math.log(float(len(trainingPositives))) - (math.log(vNeg) - math.log(float(len(trainingNegatives))))
 * 			self.weights[f] = weight
 * 			self.sortedweights.append(weight)             # <<<<<<<<<<<<<<
 * 
 * 	def __str__(self):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sortedweights); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 805, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_14 = __Pyx_PyObject_Append(__pyx_t_3, __pyx_v_weight); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 805, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "gnocis/models.pyx":797
 * 		self.weights = {}
 * 		self.sortedweights = []
 * 		for f in features:             # <<<<<<<<<<<<<<
 * 			vPos = sum( f.get(s) for s in trainingPositives )
 * 			vNeg = sum( f.get(s) for s in trainingNegatives )
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":786
 * 	"""
 * 
 * 	def __init__(self, name, features, trainingSet, windowSize, windowStep, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vPos);
  __Pyx_XDECREF(__pyx_v_vNeg);
  __Pyx_XDECREF(__pyx_v_weight);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___2generator10);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_20sequenceModelLogOdds_8__init___5generator11);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":807
 * 			self.sortedweights.append(weight)
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Log-odds model<Features: %s; Training set: %s; Positive label: %s; Negative label: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_3__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_3__str__ = {"__str__", (PyCFunction)__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_3__str__, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_3__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_20sequenceModelLogOdds_2__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "gnocis/models.pyx":808
 * 
 * 	def __str__(self):
 * 		return 'Log-odds model<Features: %s; Training set: %s; Positive label: %s; Negative label: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative))             # <<<<<<<<<<<<<<
 * 
 * 	def __repr__(self): return self.__str__()
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_Log_odds_model_Features);
  __pyx_t_2 += 25;
  __Pyx_GIVEREF(__pyx_kp_u_Log_odds_model_Features);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Log_odds_model_Features);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_features); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_Training_set);
  __pyx_t_2 += 16;
  __Pyx_GIVEREF(__pyx_kp_u_Training_set);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_Training_set);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainingSet); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_Positive_label);
  __pyx_t_2 += 18;
  __Pyx_GIVEREF(__pyx_kp_u_Positive_label);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_Positive_label);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_labelPositive); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_Negative_label);
  __pyx_t_2 += 18;
  __Pyx_GIVEREF(__pyx_kp_u_Negative_label);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u_Negative_label);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_labelNegative); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__11);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__11);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_kp_u__11);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 9, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":807
 * 			self.sortedweights.append(weight)
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Log-odds model<Features: %s; Training set: %s; Positive label: %s; Negative label: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":810
 * 		return 'Log-odds model<Features: %s; Training set: %s; Positive label: %s; Negative label: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative))
 * 
 * 	def __repr__(self): return self.__str__()             # <<<<<<<<<<<<<<
 * 
 * 	def getTrainer(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_5__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_5__repr__ = {"__repr__", (PyCFunction)__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_5__repr__, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_5__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_20sequenceModelLogOdds_4__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_4__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":812
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_7getTrainer(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_7getTrainer = {"getTrainer", (PyCFunction)__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_7getTrainer, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_7getTrainer(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getTrainer (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_20sequenceModelLogOdds_6getTrainer(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":813
 * 
 * 	def getTrainer(self):
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)             # <<<<<<<<<<<<<<
 * 
 * 	def scoreWindow(self, seq):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_10getTrainer_lambda15(PyObject *__pyx_self, PyObject *__pyx_v_ts); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_10getTrainer_lambda15 = {"lambda15", (PyCFunction)__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_10getTrainer_lambda15, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_10getTrainer_lambda15(PyObject *__pyx_self, PyObject *__pyx_v_ts) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda15 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda15(__pyx_self, ((PyObject *)__pyx_v_ts));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda15(PyObject *__pyx_self, PyObject *__pyx_v_ts) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *__pyx_cur_scope;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda15", 0);
  __pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_latin1_sequenceModelLogOdds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 813, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 813, __pyx_L1_error) }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_features); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_ts);
  __Pyx_GIVEREF(__pyx_v_ts);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_ts);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 813, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_windowSize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_windowSize, __pyx_t_2) < 0) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 813, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_windowStep); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_windowStep, __pyx_t_2) < 0) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 813, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_labelPositive); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_labelPositive, __pyx_t_2) < 0) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 813, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_labelNegative); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_labelNegative, __pyx_t_2) < 0) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.getTrainer.lambda15", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":812
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 */

static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_6getTrainer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getTrainer", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_21_getTrainer(__pyx_ptype_6gnocis_6models___pyx_scope_struct_21_getTrainer, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 812, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "gnocis/models.pyx":813
 * 
 * 	def getTrainer(self):
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)             # <<<<<<<<<<<<<<
 * 
 * 	def scoreWindow(self, seq):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_10getTrainer_lambda15, 0, __pyx_n_s_sequenceModelLogOdds_getTrainer, ((PyObject*)__pyx_cur_scope), __pyx_n_s_gnocis_models, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":812
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.getTrainer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":815
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 * 	def scoreWindow(self, seq):             # <<<<<<<<<<<<<<
 * 		cdef list fv, w
 * 		cdef int i
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_9scoreWindow(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_9scoreWindow = {"scoreWindow", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_9scoreWindow, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_20sequenceModelLogOdds_9scoreWindow(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_seq = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scoreWindow (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_seq,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scoreWindow", 1, 2, 2, 1); __PYX_ERR(0, 815, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scoreWindow") < 0)) __PYX_ERR(0, 815, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_seq = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scoreWindow", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 815, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.scoreWindow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8scoreWindow(__pyx_self, __pyx_v_self, __pyx_v_seq);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_20sequenceModelLogOdds_8scoreWindow(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_seq) {
  PyObject *__pyx_v_fv = 0;
  PyObject *__pyx_v_w = 0;
  int __pyx_v_i;
  float __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  float __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("scoreWindow", 0);

  /* "gnocis/models.pyx":818
 * 		cdef list fv, w
 * 		cdef int i
 * 		cdef float ret = 0.0             # <<<<<<<<<<<<<<
 * 		w = self.sortedweights
 * 		fv = self.features.getAll(seq)
 */
  __pyx_v_ret = 0.0;

  /* "gnocis/models.pyx":819
 * 		cdef int i
 * 		cdef float ret = 0.0
 * 		w = self.sortedweights             # <<<<<<<<<<<<<<
 * 		fv = self.features.getAll(seq)
 * 		for i in range(len(fv)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sortedweights); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 819, __pyx_L1_error)
  __pyx_v_w = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":820
 * 		cdef float ret = 0.0
 * 		w = self.sortedweights
 * 		fv = self.features.getAll(seq)             # <<<<<<<<<<<<<<
 * 		for i in range(len(fv)):
 * 			ret += fv[i] * w[i]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_features); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_getAll); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_seq) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_seq);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 820, __pyx_L1_error)
  __pyx_v_fv = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":821
 * 		w = self.sortedweights
 * 		fv = self.features.getAll(seq)
 * 		for i in range(len(fv)):             # <<<<<<<<<<<<<<
 * 			ret += fv[i] * w[i]
 * 		return ret
 */
  if (unlikely(__pyx_v_fv == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 821, __pyx_L1_error)
  }
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_fv); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 821, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "gnocis/models.pyx":822
 * 		fv = self.features.getAll(seq)
 * 		for i in range(len(fv)):
 * 			ret += fv[i] * w[i]             # <<<<<<<<<<<<<<
 * 		return ret
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_ret); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_fv == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 822, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_fv, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_w == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 822, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_w, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = PyNumber_Multiply(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = __pyx_PyFloat_AsFloat(__pyx_t_2); if (unlikely((__pyx_t_8 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_ret = __pyx_t_8;
  }

  /* "gnocis/models.pyx":823
 * 		for i in range(len(fv)):
 * 			ret += fv[i] * w[i]
 * 		return ret             # <<<<<<<<<<<<<<
 * 
 * # Trains a singular-motif PREdictor model with a given set of motifs and positive and negative training sequences.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_ret); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":815
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 * 	def scoreWindow(self, seq):             # <<<<<<<<<<<<<<
 * 		cdef list fv, w
 * 		cdef int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.models.sequenceModelLogOdds.scoreWindow", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fv);
  __Pyx_XDECREF(__pyx_v_w);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":826
 * 
 * # Trains a singular-motif PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainSinglePREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 	return sequenceModelLogOdds(name, features.motifSpectrum(motifs), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_27trainSinglePREdictorModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_27trainSinglePREdictorModel = {"trainSinglePREdictorModel", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_27trainSinglePREdictorModel, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_27trainSinglePREdictorModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_motifs = 0;
  PyObject *__pyx_v_trainingSet = 0;
  PyObject *__pyx_v_windowSize = 0;
  PyObject *__pyx_v_windowStep = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trainSinglePREdictorModel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_motifs,&__pyx_n_s_trainingSet,&__pyx_n_s_windowSize,&__pyx_n_s_windowStep,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_int_500);
    values[4] = ((PyObject *)__pyx_int_10);
    values[5] = __pyx_k__12;
    values[6] = __pyx_k__13;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_motifs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trainSinglePREdictorModel", 0, 3, 7, 1); __PYX_ERR(0, 826, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trainingSet)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trainSinglePREdictorModel", 0, 3, 7, 2); __PYX_ERR(0, 826, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowSize);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowStep);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "trainSinglePREdictorModel") < 0)) __PYX_ERR(0, 826, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_motifs = values[1];
    __pyx_v_trainingSet = values[2];
    __pyx_v_windowSize = values[3];
    __pyx_v_windowStep = values[4];
    __pyx_v_labelPositive = values[5];
    __pyx_v_labelNegative = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("trainSinglePREdictorModel", 0, 3, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 826, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.trainSinglePREdictorModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_26trainSinglePREdictorModel(__pyx_self, __pyx_v_name, __pyx_v_motifs, __pyx_v_trainingSet, __pyx_v_windowSize, __pyx_v_windowStep, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_26trainSinglePREdictorModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_motifs, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("trainSinglePREdictorModel", 0);

  /* "gnocis/models.pyx":827
 * # Trains a singular-motif PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainSinglePREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):
 * 	return sequenceModelLogOdds(name, features.motifSpectrum(motifs), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)             # <<<<<<<<<<<<<<
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_latin1_sequenceModelLogOdds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6gnocis_8features_features), __pyx_n_s_motifSpectrum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_motifs) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_motifs);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_name);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_trainingSet);
  __Pyx_GIVEREF(__pyx_v_trainingSet);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_trainingSet);
  __Pyx_INCREF(__pyx_v_windowSize);
  __Pyx_GIVEREF(__pyx_v_windowSize);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_windowSize);
  __Pyx_INCREF(__pyx_v_windowStep);
  __Pyx_GIVEREF(__pyx_v_windowStep);
  PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_windowStep);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 827, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":826
 * 
 * # Trains a singular-motif PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainSinglePREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 	return sequenceModelLogOdds(name, features.motifSpectrum(motifs), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.trainSinglePREdictorModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":830
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def createDummyPREdictorModel(name, motifs, windowSize=500, windowStep=10):             # <<<<<<<<<<<<<<
 * 	return sequenceModelDummy(name, motifs.pairFreq(219), windowSize, windowStep)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_29createDummyPREdictorModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_29createDummyPREdictorModel = {"createDummyPREdictorModel", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_29createDummyPREdictorModel, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_29createDummyPREdictorModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_motifs = 0;
  PyObject *__pyx_v_windowSize = 0;
  PyObject *__pyx_v_windowStep = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createDummyPREdictorModel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_motifs,&__pyx_n_s_windowSize,&__pyx_n_s_windowStep,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)__pyx_int_500);
    values[3] = ((PyObject *)__pyx_int_10);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_motifs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("createDummyPREdictorModel", 0, 2, 4, 1); __PYX_ERR(0, 830, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowSize);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowStep);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "createDummyPREdictorModel") < 0)) __PYX_ERR(0, 830, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_motifs = values[1];
    __pyx_v_windowSize = values[2];
    __pyx_v_windowStep = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("createDummyPREdictorModel", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 830, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.createDummyPREdictorModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_28createDummyPREdictorModel(__pyx_self, __pyx_v_name, __pyx_v_motifs, __pyx_v_windowSize, __pyx_v_windowStep);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_28createDummyPREdictorModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_motifs, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("createDummyPREdictorModel", 0);

  /* "gnocis/models.pyx":831
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def createDummyPREdictorModel(name, motifs, windowSize=500, windowStep=10):
 * 	return sequenceModelDummy(name, motifs.pairFreq(219), windowSize, windowStep)             # <<<<<<<<<<<<<<
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_latin1_sequenceModelDummy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_motifs, __pyx_n_s_pairFreq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_int_219) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_int_219);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_name, __pyx_t_3, __pyx_v_windowSize, __pyx_v_windowStep};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_name, __pyx_t_3, __pyx_v_windowSize, __pyx_v_windowStep};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_name);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_windowSize);
    __Pyx_GIVEREF(__pyx_v_windowSize);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_v_windowSize);
    __Pyx_INCREF(__pyx_v_windowStep);
    __Pyx_GIVEREF(__pyx_v_windowStep);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_6, __pyx_v_windowStep);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":830
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def createDummyPREdictorModel(name, motifs, windowSize=500, windowStep=10):             # <<<<<<<<<<<<<<
 * 	return sequenceModelDummy(name, motifs.pairFreq(219), windowSize, windowStep)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.models.createDummyPREdictorModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":834
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainPREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 	return sequenceModelLogOdds(name, features.motifPairSpectrum(motifs, 219), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_31trainPREdictorModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_31trainPREdictorModel = {"trainPREdictorModel", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_31trainPREdictorModel, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_31trainPREdictorModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_motifs = 0;
  PyObject *__pyx_v_trainingSet = 0;
  PyObject *__pyx_v_windowSize = 0;
  PyObject *__pyx_v_windowStep = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trainPREdictorModel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_motifs,&__pyx_n_s_trainingSet,&__pyx_n_s_windowSize,&__pyx_n_s_windowStep,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_int_500);
    values[4] = ((PyObject *)__pyx_int_10);
    values[5] = __pyx_k__14;
    values[6] = __pyx_k__15;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_motifs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trainPREdictorModel", 0, 3, 7, 1); __PYX_ERR(0, 834, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trainingSet)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("trainPREdictorModel", 0, 3, 7, 2); __PYX_ERR(0, 834, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowSize);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_windowStep);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "trainPREdictorModel") < 0)) __PYX_ERR(0, 834, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_motifs = values[1];
    __pyx_v_trainingSet = values[2];
    __pyx_v_windowSize = values[3];
    __pyx_v_windowStep = values[4];
    __pyx_v_labelPositive = values[5];
    __pyx_v_labelNegative = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("trainPREdictorModel", 0, 3, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 834, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.trainPREdictorModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_30trainPREdictorModel(__pyx_self, __pyx_v_name, __pyx_v_motifs, __pyx_v_trainingSet, __pyx_v_windowSize, __pyx_v_windowStep, __pyx_v_labelPositive, __pyx_v_labelNegative);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_30trainPREdictorModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_motifs, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_windowSize, PyObject *__pyx_v_windowStep, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("trainPREdictorModel", 0);

  /* "gnocis/models.pyx":835
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainPREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):
 * 	return sequenceModelLogOdds(name, features.motifPairSpectrum(motifs, 219), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)             # <<<<<<<<<<<<<<
 * 
 * class CVModelPredictions:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_latin1_sequenceModelLogOdds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6gnocis_8features_features), __pyx_n_s_motifPairSpectrum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_motifs, __pyx_int_219};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_motifs, __pyx_int_219};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_motifs);
    __Pyx_GIVEREF(__pyx_v_motifs);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_motifs);
    __Pyx_INCREF(__pyx_int_219);
    __Pyx_GIVEREF(__pyx_int_219);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_219);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_name);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_trainingSet);
  __Pyx_GIVEREF(__pyx_v_trainingSet);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_trainingSet);
  __Pyx_INCREF(__pyx_v_windowSize);
  __Pyx_GIVEREF(__pyx_v_windowSize);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_windowSize);
  __Pyx_INCREF(__pyx_v_windowStep);
  __Pyx_GIVEREF(__pyx_v_windowStep);
  PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_windowStep);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":834
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainPREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 	return sequenceModelLogOdds(name, features.motifPairSpectrum(motifs, 219), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.models.trainPREdictorModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":848
 * 	"""
 * 
 * 	def __init__(self, model, trainedModels = [], labelPredictionCurves = None):             # <<<<<<<<<<<<<<
 * 		self.name = model.name
 * 		self.model = model
 */

static PyObject *__pyx_pf_6gnocis_6models_54__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults10, __pyx_self)->__pyx_arg_trainedModels);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults10, __pyx_self)->__pyx_arg_trainedModels);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults10, __pyx_self)->__pyx_arg_trainedModels);
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_None));
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18CVModelPredictions_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_18CVModelPredictions_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_model = 0;
  PyObject *__pyx_v_trainedModels = 0;
  PyObject *__pyx_v_labelPredictionCurves = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_model,&__pyx_n_s_trainedModels,&__pyx_n_s_labelPredictionCurves,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults10 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults10, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_trainedModels;
    values[3] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_model)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 848, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trainedModels);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPredictionCurves);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 848, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_model = values[1];
    __pyx_v_trainedModels = values[2];
    __pyx_v_labelPredictionCurves = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 848, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_18CVModelPredictions___init__(__pyx_self, __pyx_v_self, __pyx_v_model, __pyx_v_trainedModels, __pyx_v_labelPredictionCurves);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_model, PyObject *__pyx_v_trainedModels, PyObject *__pyx_v_labelPredictionCurves) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gnocis/models.pyx":849
 * 
 * 	def __init__(self, model, trainedModels = [], labelPredictionCurves = None):
 * 		self.name = model.name             # <<<<<<<<<<<<<<
 * 		self.model = model
 * 		self.trainedModels = trainedModels
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_model, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_t_1) < 0) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":850
 * 	def __init__(self, model, trainedModels = [], labelPredictionCurves = None):
 * 		self.name = model.name
 * 		self.model = model             # <<<<<<<<<<<<<<
 * 		self.trainedModels = trainedModels
 * 		self.predictionCurves = {} if labelPredictionCurves is None else labelPredictionCurves
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_model, __pyx_v_model) < 0) __PYX_ERR(0, 850, __pyx_L1_error)

  /* "gnocis/models.pyx":851
 * 		self.name = model.name
 * 		self.model = model
 * 		self.trainedModels = trainedModels             # <<<<<<<<<<<<<<
 * 		self.predictionCurves = {} if labelPredictionCurves is None else labelPredictionCurves
 * 		self.predictedRegions = {}
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels, __pyx_v_trainedModels) < 0) __PYX_ERR(0, 851, __pyx_L1_error)

  /* "gnocis/models.pyx":852
 * 		self.model = model
 * 		self.trainedModels = trainedModels
 * 		self.predictionCurves = {} if labelPredictionCurves is None else labelPredictionCurves             # <<<<<<<<<<<<<<
 * 		self.predictedRegions = {}
 * 		self.hasCores = False
 */
  __pyx_t_2 = (__pyx_v_labelPredictionCurves == Py_None);
  if ((__pyx_t_2 != 0)) {
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_labelPredictionCurves);
    __pyx_t_1 = __pyx_v_labelPredictionCurves;
  }
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_predictionCurves, __pyx_t_1) < 0) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":853
 * 		self.trainedModels = trainedModels
 * 		self.predictionCurves = {} if labelPredictionCurves is None else labelPredictionCurves
 * 		self.predictedRegions = {}             # <<<<<<<<<<<<<<
 * 		self.hasCores = False
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_predictedRegions, __pyx_t_1) < 0) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":854
 * 		self.predictionCurves = {} if labelPredictionCurves is None else labelPredictionCurves
 * 		self.predictedRegions = {}
 * 		self.hasCores = False             # <<<<<<<<<<<<<<
 * 
 * 	def regions(self, label):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_hasCores, Py_False) < 0) __PYX_ERR(0, 854, __pyx_L1_error)

  /* "gnocis/models.pyx":848
 * 	"""
 * 
 * 	def __init__(self, model, trainedModels = [], labelPredictionCurves = None):             # <<<<<<<<<<<<<<
 * 		self.name = model.name
 * 		self.model = model
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":856
 * 		self.hasCores = False
 * 
 * 	def regions(self, label):             # <<<<<<<<<<<<<<
 * 		if label in self.predictedRegions:
 * 			return self.predictedRegions[label]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_3regions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18CVModelPredictions_3regions = {"regions", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_18CVModelPredictions_3regions, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_3regions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_label = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("regions (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_label,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_label)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("regions", 1, 2, 2, 1); __PYX_ERR(0, 856, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "regions") < 0)) __PYX_ERR(0, 856, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_label = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("regions", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 856, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.regions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_18CVModelPredictions_2regions(__pyx_self, __pyx_v_self, __pyx_v_label);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_2regions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_label) {
  PyObject *__pyx_9genexpr40__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("regions", 0);

  /* "gnocis/models.pyx":857
 * 
 * 	def regions(self, label):
 * 		if label in self.predictedRegions:             # <<<<<<<<<<<<<<
 * 			return self.predictedRegions[label]
 * 		return [ c.regions() for c in self.predictionCurves[label] ]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictedRegions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_label, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "gnocis/models.pyx":858
 * 	def regions(self, label):
 * 		if label in self.predictedRegions:
 * 			return self.predictedRegions[label]             # <<<<<<<<<<<<<<
 * 		return [ c.regions() for c in self.predictionCurves[label] ]
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictedRegions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_label); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "gnocis/models.pyx":857
 * 
 * 	def regions(self, label):
 * 		if label in self.predictedRegions:             # <<<<<<<<<<<<<<
 * 			return self.predictedRegions[label]
 * 		return [ c.regions() for c in self.predictionCurves[label] ]
 */
  }

  /* "gnocis/models.pyx":859
 * 		if label in self.predictedRegions:
 * 			return self.predictedRegions[label]
 * 		return [ c.regions() for c in self.predictionCurves[label] ]             # <<<<<<<<<<<<<<
 * 
 * 	def setCorePredictions(self, label, cores):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 859, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictionCurves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_label); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 859, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_1 = __pyx_t_5; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 859, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 859, __pyx_L6_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 859, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 859, __pyx_L6_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 859, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_1);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 859, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr40__pyx_v_c, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr40__pyx_v_c, __pyx_n_s_regions); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 859, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 859, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 859, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr40__pyx_v_c); __pyx_9genexpr40__pyx_v_c = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_9genexpr40__pyx_v_c); __pyx_9genexpr40__pyx_v_c = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":856
 * 		self.hasCores = False
 * 
 * 	def regions(self, label):             # <<<<<<<<<<<<<<
 * 		if label in self.predictedRegions:
 * 			return self.predictedRegions[label]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.regions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr40__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":861
 * 		return [ c.regions() for c in self.predictionCurves[label] ]
 * 
 * 	def setCorePredictions(self, label, cores):             # <<<<<<<<<<<<<<
 * 		self.predictedRegions[label] = cores
 * 		self.hasCores = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_5setCorePredictions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18CVModelPredictions_5setCorePredictions = {"setCorePredictions", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_18CVModelPredictions_5setCorePredictions, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_5setCorePredictions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_label = 0;
  PyObject *__pyx_v_cores = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setCorePredictions (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_label,&__pyx_n_s_cores,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_label)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setCorePredictions", 1, 3, 3, 1); __PYX_ERR(0, 861, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cores)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setCorePredictions", 1, 3, 3, 2); __PYX_ERR(0, 861, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setCorePredictions") < 0)) __PYX_ERR(0, 861, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_label = values[1];
    __pyx_v_cores = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setCorePredictions", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 861, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.setCorePredictions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_18CVModelPredictions_4setCorePredictions(__pyx_self, __pyx_v_self, __pyx_v_label, __pyx_v_cores);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_4setCorePredictions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_label, PyObject *__pyx_v_cores) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setCorePredictions", 0);

  /* "gnocis/models.pyx":862
 * 
 * 	def setCorePredictions(self, label, cores):
 * 		self.predictedRegions[label] = cores             # <<<<<<<<<<<<<<
 * 		self.hasCores = True
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictedRegions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_label, __pyx_v_cores) < 0)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":863
 * 	def setCorePredictions(self, label, cores):
 * 		self.predictedRegions[label] = cores
 * 		self.hasCores = True             # <<<<<<<<<<<<<<
 * 
 * 	def predictCore(self, genome):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_hasCores, Py_True) < 0) __PYX_ERR(0, 863, __pyx_L1_error)

  /* "gnocis/models.pyx":861
 * 		return [ c.regions() for c in self.predictionCurves[label] ]
 * 
 * 	def setCorePredictions(self, label, cores):             # <<<<<<<<<<<<<<
 * 		self.predictedRegions[label] = cores
 * 		self.hasCores = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.setCorePredictions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":865
 * 		self.hasCores = True
 * 
 * 	def predictCore(self, genome):             # <<<<<<<<<<<<<<
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))
 * 		cores = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_7predictCore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18CVModelPredictions_7predictCore = {"predictCore", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_18CVModelPredictions_7predictCore, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_7predictCore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_genome = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predictCore (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_genome,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_genome)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("predictCore", 1, 2, 2, 1); __PYX_ERR(0, 865, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "predictCore") < 0)) __PYX_ERR(0, 865, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_genome = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("predictCore", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 865, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.predictCore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_18CVModelPredictions_6predictCore(__pyx_self, __pyx_v_self, __pyx_v_genome);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_6predictCore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_genome) {
  PyObject *__pyx_v_task = NULL;
  PyObject *__pyx_v_cores = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_mdl = NULL;
  PyObject *__pyx_v_core = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("predictCore", 0);

  /* "gnocis/models.pyx":866
 * 
 * 	def predictCore(self, genome):
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))             # <<<<<<<<<<<<<<
 * 		cores = []
 * 		for i, mdl in enumerate(self.trainedModels):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_steps, __pyx_t_3) < 0) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__16, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_task = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":867
 * 	def predictCore(self, genome):
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))
 * 		cores = []             # <<<<<<<<<<<<<<
 * 		for i, mdl in enumerate(self.trainedModels):
 * 			task.update(step = i)
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_cores = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":868
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))
 * 		cores = []
 * 		for i, mdl in enumerate(self.trainedModels):             # <<<<<<<<<<<<<<
 * 			task.update(step = i)
 * 			core = mdl.predictCore(genome)
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 868, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 868, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 868, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 868, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_mdl, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":869
 * 		cores = []
 * 		for i, mdl in enumerate(self.trainedModels):
 * 			task.update(step = i)             # <<<<<<<<<<<<<<
 * 			core = mdl.predictCore(genome)
 * 			if core is not None:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_step, __pyx_v_i) < 0) __PYX_ERR(0, 869, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "gnocis/models.pyx":870
 * 		for i, mdl in enumerate(self.trainedModels):
 * 			task.update(step = i)
 * 			core = mdl.predictCore(genome)             # <<<<<<<<<<<<<<
 * 			if core is not None:
 * 				cores.append(core)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_mdl, __pyx_n_s_predictCore); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 870, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_2, __pyx_v_genome) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_genome);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 870, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_core, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "gnocis/models.pyx":871
 * 			task.update(step = i)
 * 			core = mdl.predictCore(genome)
 * 			if core is not None:             # <<<<<<<<<<<<<<
 * 				cores.append(core)
 * 		if len(cores) > 0:
 */
    __pyx_t_8 = (__pyx_v_core != Py_None);
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {

      /* "gnocis/models.pyx":872
 * 			core = mdl.predictCore(genome)
 * 			if core is not None:
 * 				cores.append(core)             # <<<<<<<<<<<<<<
 * 		if len(cores) > 0:
 * 			self.setCorePredictions(label = positive, cores = cores)
 */
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_cores, __pyx_v_core); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 872, __pyx_L1_error)

      /* "gnocis/models.pyx":871
 * 			task.update(step = i)
 * 			core = mdl.predictCore(genome)
 * 			if core is not None:             # <<<<<<<<<<<<<<
 * 				cores.append(core)
 * 		if len(cores) > 0:
 */
    }

    /* "gnocis/models.pyx":868
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))
 * 		cores = []
 * 		for i, mdl in enumerate(self.trainedModels):             # <<<<<<<<<<<<<<
 * 			task.update(step = i)
 * 			core = mdl.predictCore(genome)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":873
 * 			if core is not None:
 * 				cores.append(core)
 * 		if len(cores) > 0:             # <<<<<<<<<<<<<<
 * 			self.setCorePredictions(label = positive, cores = cores)
 * 		task.done()
 */
  __pyx_t_4 = PyList_GET_SIZE(__pyx_v_cores); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 873, __pyx_L1_error)
  __pyx_t_9 = ((__pyx_t_4 > 0) != 0);
  if (__pyx_t_9) {

    /* "gnocis/models.pyx":874
 * 				cores.append(core)
 * 		if len(cores) > 0:
 * 			self.setCorePredictions(label = positive, cores = cores)             # <<<<<<<<<<<<<<
 * 		task.done()
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_setCorePredictions); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_positive); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_label, __pyx_t_7) < 0) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_cores, __pyx_v_cores) < 0) __PYX_ERR(0, 874, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "gnocis/models.pyx":873
 * 			if core is not None:
 * 				cores.append(core)
 * 		if len(cores) > 0:             # <<<<<<<<<<<<<<
 * 			self.setCorePredictions(label = positive, cores = cores)
 * 		task.done()
 */
  }

  /* "gnocis/models.pyx":875
 * 		if len(cores) > 0:
 * 			self.setCorePredictions(label = positive, cores = cores)
 * 		task.done()             # <<<<<<<<<<<<<<
 * 
 * 	def getStatTable(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_done); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "gnocis/models.pyx":865
 * 		self.hasCores = True
 * 
 * 	def predictCore(self, genome):             # <<<<<<<<<<<<<<
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))
 * 		cores = []
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.predictCore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_task);
  __Pyx_XDECREF(__pyx_v_cores);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_v_core);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":877
 * 		task.done()
 * 
 * 	def getStatTable(self):             # <<<<<<<<<<<<<<
 * 		rtable = []
 * 		for label in self.predictionCurves:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_9getStatTable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18CVModelPredictions_9getStatTable = {"getStatTable", (PyCFunction)__pyx_pw_6gnocis_6models_18CVModelPredictions_9getStatTable, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_9getStatTable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getStatTable (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_18CVModelPredictions_8getStatTable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_8getStatTable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_rtable = NULL;
  PyObject *__pyx_v_label = NULL;
  PyObject *__pyx_v_npred = NULL;
  PyObject *__pyx_v_meanlen = NULL;
  PyObject *__pyx_9genexpr41__pyx_v_p = NULL;
  PyObject *__pyx_9genexpr42__pyx_v_p = NULL;
  PyObject *__pyx_9genexpr43__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *(*__pyx_t_15)(PyObject *);
  Py_ssize_t __pyx_t_16;
  Py_UCS4 __pyx_t_17;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getStatTable", 0);

  /* "gnocis/models.pyx":878
 * 
 * 	def getStatTable(self):
 * 		rtable = []             # <<<<<<<<<<<<<<
 * 		for label in self.predictionCurves:
 * 			npred = [ len(p.regions()) for p in self.predictionCurves[label] ]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_rtable = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":879
 * 	def getStatTable(self):
 * 		rtable = []
 * 		for label in self.predictionCurves:             # <<<<<<<<<<<<<<
 * 			npred = [ len(p.regions()) for p in self.predictionCurves[label] ]
 * 			meanlen = [ mean([ len(r) for r in p.regions() ])
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictionCurves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 879, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 879, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 879, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 879, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_label, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":880
 * 		rtable = []
 * 		for label in self.predictionCurves:
 * 			npred = [ len(p.regions()) for p in self.predictionCurves[label] ]             # <<<<<<<<<<<<<<
 * 			meanlen = [ mean([ len(r) for r in p.regions() ])
 * 						for p in self.predictionCurves[label] ]
 */
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictionCurves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 880, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_v_label); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 880, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 880, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 880, __pyx_L7_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 880, __pyx_L7_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 880, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 880, __pyx_L7_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 880, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_8(__pyx_t_5);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 880, __pyx_L7_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr41__pyx_v_p, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr41__pyx_v_p, __pyx_n_s_regions); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 880, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 880, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_11 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 880, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 880, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 880, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_9genexpr41__pyx_v_p); __pyx_9genexpr41__pyx_v_p = 0;
      goto __pyx_L10_exit_scope;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_9genexpr41__pyx_v_p); __pyx_9genexpr41__pyx_v_p = 0;
      goto __pyx_L1_error;
      __pyx_L10_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_npred, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":881
 * 		for label in self.predictionCurves:
 * 			npred = [ len(p.regions()) for p in self.predictionCurves[label] ]
 * 			meanlen = [ mean([ len(r) for r in p.regions() ])             # <<<<<<<<<<<<<<
 * 						for p in self.predictionCurves[label] ]
 * 			rtable.append({
 */
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "gnocis/models.pyx":882
 * 			npred = [ len(p.regions()) for p in self.predictionCurves[label] ]
 * 			meanlen = [ mean([ len(r) for r in p.regions() ])
 * 						for p in self.predictionCurves[label] ]             # <<<<<<<<<<<<<<
 * 			rtable.append({
 * 				'Label': label.name,
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictionCurves); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 882, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_v_label); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 882, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 882, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 882, __pyx_L13_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 882, __pyx_L13_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 882, __pyx_L13_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 882, __pyx_L13_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 882, __pyx_L13_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_8(__pyx_t_5);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 882, __pyx_L13_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr42__pyx_v_p, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "gnocis/models.pyx":881
 * 		for label in self.predictionCurves:
 * 			npred = [ len(p.regions()) for p in self.predictionCurves[label] ]
 * 			meanlen = [ mean([ len(r) for r in p.regions() ])             # <<<<<<<<<<<<<<
 * 						for p in self.predictionCurves[label] ]
 * 			rtable.append({
 */
        __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_mean); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 881, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_9);
        { /* enter inner scope */
          __pyx_t_10 = PyList_New(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 881, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr42__pyx_v_p, __pyx_n_s_regions); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 881, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
            }
          }
          __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_14) : __Pyx_PyObject_CallNoArg(__pyx_t_13);
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 881, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (likely(PyList_CheckExact(__pyx_t_12)) || PyTuple_CheckExact(__pyx_t_12)) {
            __pyx_t_13 = __pyx_t_12; __Pyx_INCREF(__pyx_t_13); __pyx_t_11 = 0;
            __pyx_t_15 = NULL;
          } else {
            __pyx_t_11 = -1; __pyx_t_13 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 881, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_15 = Py_TYPE(__pyx_t_13)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 881, __pyx_L18_error)
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          for (;;) {
            if (likely(!__pyx_t_15)) {
              if (likely(PyList_CheckExact(__pyx_t_13))) {
                if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_13)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_12 = PyList_GET_ITEM(__pyx_t_13, __pyx_t_11); __Pyx_INCREF(__pyx_t_12); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 881, __pyx_L18_error)
                #else
                __pyx_t_12 = PySequence_ITEM(__pyx_t_13, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 881, __pyx_L18_error)
                __Pyx_GOTREF(__pyx_t_12);
                #endif
              } else {
                if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_13)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_13, __pyx_t_11); __Pyx_INCREF(__pyx_t_12); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 881, __pyx_L18_error)
                #else
                __pyx_t_12 = PySequence_ITEM(__pyx_t_13, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 881, __pyx_L18_error)
                __Pyx_GOTREF(__pyx_t_12);
                #endif
              }
            } else {
              __pyx_t_12 = __pyx_t_15(__pyx_t_13);
              if (unlikely(!__pyx_t_12)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 881, __pyx_L18_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_12);
            }
            __Pyx_XDECREF_SET(__pyx_9genexpr43__pyx_v_r, __pyx_t_12);
            __pyx_t_12 = 0;
            __pyx_t_16 = PyObject_Length(__pyx_9genexpr43__pyx_v_r); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 881, __pyx_L18_error)
            __pyx_t_12 = PyInt_FromSsize_t(__pyx_t_16); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 881, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (unlikely(__Pyx_ListComp_Append(__pyx_t_10, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 881, __pyx_L18_error)
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_9genexpr43__pyx_v_r); __pyx_9genexpr43__pyx_v_r = 0;
          goto __pyx_L21_exit_scope;
          __pyx_L18_error:;
          __Pyx_XDECREF(__pyx_9genexpr43__pyx_v_r); __pyx_9genexpr43__pyx_v_r = 0;
          goto __pyx_L13_error;
          __pyx_L21_exit_scope:;
        } /* exit inner scope */
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_6 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_13, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 881, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 881, __pyx_L13_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "gnocis/models.pyx":882
 * 			npred = [ len(p.regions()) for p in self.predictionCurves[label] ]
 * 			meanlen = [ mean([ len(r) for r in p.regions() ])
 * 						for p in self.predictionCurves[label] ]             # <<<<<<<<<<<<<<
 * 			rtable.append({
 * 				'Label': label.name,
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_9genexpr42__pyx_v_p); __pyx_9genexpr42__pyx_v_p = 0;
      goto __pyx_L22_exit_scope;
      __pyx_L13_error:;
      __Pyx_XDECREF(__pyx_9genexpr42__pyx_v_p); __pyx_9genexpr42__pyx_v_p = 0;
      goto __pyx_L1_error;
      __pyx_L22_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_meanlen, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":884
 * 						for p in self.predictionCurves[label] ]
 * 			rtable.append({
 * 				'Label': label.name,             # <<<<<<<<<<<<<<
 * 				'Predictions': '%.2f +/- %.2f'%(mean(npred), CI(npred)),
 * 				'Mean length': '%.2f +/- %.2f'%(mean(meanlen), CI(meanlen))
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_label, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_Label, __pyx_t_5) < 0) __PYX_ERR(0, 884, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "gnocis/models.pyx":885
 * 			rtable.append({
 * 				'Label': label.name,
 * 				'Predictions': '%.2f +/- %.2f'%(mean(npred), CI(npred)),             # <<<<<<<<<<<<<<
 * 				'Mean length': '%.2f +/- %.2f'%(mean(meanlen), CI(meanlen))
 * 			})
 */
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = 0;
    __pyx_t_17 = 127;
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_mean); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_6 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_v_npred) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_npred);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Format(__pyx_t_6, __pyx_kp_u_2f); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_17;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_9);
    __pyx_t_9 = 0;
    __Pyx_INCREF(__pyx_kp_u__17);
    __pyx_t_7 += 5;
    __Pyx_GIVEREF(__pyx_kp_u__17);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u__17);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_CI); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_9 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_10, __pyx_v_npred) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_npred);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Format(__pyx_t_9, __pyx_kp_u_2f); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_17;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_7, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_Predictions, __pyx_t_6) < 0) __PYX_ERR(0, 884, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "gnocis/models.pyx":886
 * 				'Label': label.name,
 * 				'Predictions': '%.2f +/- %.2f'%(mean(npred), CI(npred)),
 * 				'Mean length': '%.2f +/- %.2f'%(mean(meanlen), CI(meanlen))             # <<<<<<<<<<<<<<
 * 			})
 * 		return nctable('Predictions: ' + self.model.name, rtable)
 */
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 0;
    __pyx_t_17 = 127;
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_mean); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_5 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_v_meanlen) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_meanlen);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Format(__pyx_t_5, __pyx_kp_u_2f); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_17;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_9);
    __pyx_t_9 = 0;
    __Pyx_INCREF(__pyx_kp_u__17);
    __pyx_t_7 += 5;
    __Pyx_GIVEREF(__pyx_kp_u__17);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_kp_u__17);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CI); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_9 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_10, __pyx_v_meanlen) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_meanlen);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_9, __pyx_kp_u_2f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_17 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_17) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_17;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_6, 3, __pyx_t_7, __pyx_t_17); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_kp_u_Mean_length, __pyx_t_5) < 0) __PYX_ERR(0, 884, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "gnocis/models.pyx":883
 * 			meanlen = [ mean([ len(r) for r in p.regions() ])
 * 						for p in self.predictionCurves[label] ]
 * 			rtable.append({             # <<<<<<<<<<<<<<
 * 				'Label': label.name,
 * 				'Predictions': '%.2f +/- %.2f'%(mean(npred), CI(npred)),
 */
    __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_rtable, __pyx_t_1); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":879
 * 	def getStatTable(self):
 * 		rtable = []
 * 		for label in self.predictionCurves:             # <<<<<<<<<<<<<<
 * 			npred = [ len(p.regions()) for p in self.predictionCurves[label] ]
 * 			meanlen = [ mean([ len(r) for r in p.regions() ])
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":888
 * 				'Mean length': '%.2f +/- %.2f'%(mean(meanlen), CI(meanlen))
 * 			})
 * 		return nctable('Predictions: ' + self.model.name, rtable)             # <<<<<<<<<<<<<<
 * 
 * 	def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_nctable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_model); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyNumber_Add(__pyx_kp_u_Predictions_2, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_19 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_19 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_5, __pyx_v_rtable};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_5, __pyx_v_rtable};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_19); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_19, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_rtable);
    __Pyx_GIVEREF(__pyx_v_rtable);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_19, __pyx_v_rtable);
    __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":877
 * 		task.done()
 * 
 * 	def getStatTable(self):             # <<<<<<<<<<<<<<
 * 		rtable = []
 * 		for label in self.predictionCurves:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.getStatTable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rtable);
  __Pyx_XDECREF(__pyx_v_label);
  __Pyx_XDECREF(__pyx_v_npred);
  __Pyx_XDECREF(__pyx_v_meanlen);
  __Pyx_XDECREF(__pyx_9genexpr41__pyx_v_p);
  __Pyx_XDECREF(__pyx_9genexpr42__pyx_v_p);
  __Pyx_XDECREF(__pyx_9genexpr43__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":890
 * 		return nctable('Predictions: ' + self.model.name, rtable)
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_11__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18CVModelPredictions_11__repr__ = {"__repr__", (PyCFunction)__pyx_pw_6gnocis_6models_18CVModelPredictions_11__repr__, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_11__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_18CVModelPredictions_10__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_10__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "gnocis/models.pyx":891
 * 
 * 	def __repr__(self):
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()             # <<<<<<<<<<<<<<
 * 
 * 	def _repr_html_(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_model); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u_CVModelPredictions, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_kp_u__11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getStatTable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_repr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":890
 * 		return nctable('Predictions: ' + self.model.name, rtable)
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":893
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		return '<div>' + self.getStatTable()._repr_html_() + '</div>'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_13_repr_html_(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_18CVModelPredictions_13_repr_html_ = {"_repr_html_", (PyCFunction)__pyx_pw_6gnocis_6models_18CVModelPredictions_13_repr_html_, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_18CVModelPredictions_13_repr_html_(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_repr_html_ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_18CVModelPredictions_12_repr_html_(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_18CVModelPredictions_12_repr_html_(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_repr_html_", 0);

  /* "gnocis/models.pyx":894
 * 
 * 	def _repr_html_(self):
 * 		return '<div>' + self.getStatTable()._repr_html_() + '</div>'             # <<<<<<<<<<<<<<
 * 
 * class crossvalidation:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getStatTable); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_repr_html); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_kp_u_div, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_kp_u_div_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":893
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		return '<div>' + self.getStatTable()._repr_html_() + '</div>'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.CVModelPredictions._repr_html_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":919
 * 	"""
 * 
 * 	def __init__(self, models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = None, storeTrained = True):             # <<<<<<<<<<<<<<
 * 		self.models = []
 * 		self.PRC = {}
 */

static PyObject *__pyx_pf_6gnocis_6models_56__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_None));
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_self)->__pyx_arg_labelNegative);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_self)->__pyx_arg_labelNegative);
  __Pyx_INCREF(((PyObject *)__pyx_int_20));
  __Pyx_GIVEREF(((PyObject *)__pyx_int_20));
  PyTuple_SET_ITEM(__pyx_t_1, 3, ((PyObject *)__pyx_int_20));
  __Pyx_INCREF(((PyObject*)__pyx_float_0_8));
  __Pyx_GIVEREF(((PyObject*)__pyx_float_0_8));
  PyTuple_SET_ITEM(__pyx_t_1, 4, ((PyObject*)__pyx_float_0_8));
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 5, ((PyObject *)Py_None));
  __Pyx_INCREF(((PyObject *)Py_True));
  __Pyx_GIVEREF(((PyObject *)Py_True));
  PyTuple_SET_ITEM(__pyx_t_1, 6, ((PyObject *)Py_True));
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_15crossvalidation_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_models = 0;
  PyObject *__pyx_v_trainingSet = 0;
  PyObject *__pyx_v_validationSet = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  PyObject *__pyx_v_repeats = 0;
  PyObject *__pyx_v_ratioTrainTest = 0;
  PyObject *__pyx_v_ratioNegPos = 0;
  PyObject *__pyx_v_storeTrained = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_models,&__pyx_n_s_trainingSet,&__pyx_n_s_validationSet,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,&__pyx_n_s_repeats,&__pyx_n_s_ratioTrainTest,&__pyx_n_s_ratioNegPos,&__pyx_n_s_storeTrained,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    __pyx_defaults11 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_self);
    values[3] = ((PyObject *)((PyObject *)Py_None));
    values[4] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    values[5] = __pyx_dynamic_args->__pyx_arg_labelNegative;
    values[6] = ((PyObject *)((PyObject *)__pyx_int_20));
    values[7] = ((PyObject *)((PyObject*)__pyx_float_0_8));
    values[8] = ((PyObject *)((PyObject *)Py_None));
    values[9] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_models)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 10, 1); __PYX_ERR(0, 919, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trainingSet)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 10, 2); __PYX_ERR(0, 919, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validationSet);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_repeats);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ratioTrainTest);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ratioNegPos);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_storeTrained);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 919, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_models = values[1];
    __pyx_v_trainingSet = values[2];
    __pyx_v_validationSet = values[3];
    __pyx_v_labelPositive = values[4];
    __pyx_v_labelNegative = values[5];
    __pyx_v_repeats = values[6];
    __pyx_v_ratioTrainTest = values[7];
    __pyx_v_ratioNegPos = values[8];
    __pyx_v_storeTrained = values[9];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 919, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation___init__(__pyx_self, __pyx_v_self, __pyx_v_models, __pyx_v_trainingSet, __pyx_v_validationSet, __pyx_v_labelPositive, __pyx_v_labelNegative, __pyx_v_repeats, __pyx_v_ratioTrainTest, __pyx_v_ratioNegPos, __pyx_v_storeTrained);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_8__init___2generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":942
 * 			for lbl, seqs in zip(tslabels, trainingSet.withLabel(tslabels))
 * 		}
 * 		ntrain = int( min(len(tsByLabel[lbl]) for lbl in tsByLabel) * ratioTrainTest )             # <<<<<<<<<<<<<<
 * 		self.cvtrain = []
 * 		self.cvval = []
 */

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_23_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_23_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 942, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_15crossvalidation_8__init___2generator12, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_crossvalidation___init___locals, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_8__init___2generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 942, __pyx_L1_error)
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tsByLabel)) { __Pyx_RaiseClosureNameError("tsByLabel"); __PYX_ERR(0, 942, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tsByLabel == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 942, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tsByLabel, 1, ((PyObject *)NULL), (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_lbl);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_lbl, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tsByLabel)) { __Pyx_RaiseClosureNameError("tsByLabel"); __PYX_ERR(0, 942, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tsByLabel == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 942, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tsByLabel, __pyx_cur_scope->__pyx_v_lbl); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 942, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":919
 * 	"""
 * 
 * 	def __init__(self, models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = None, storeTrained = True):             # <<<<<<<<<<<<<<
 * 		self.models = []
 * 		self.PRC = {}
 */

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_models, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_validationSet, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative, PyObject *__pyx_v_repeats, PyObject *__pyx_v_ratioTrainTest, PyObject *__pyx_v_ratioNegPos, PyObject *__pyx_v_storeTrained) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_vpos = NULL;
  PyObject *__pyx_v_vneg = NULL;
  PyObject *__pyx_v_tslabels = NULL;
  PyObject *__pyx_v_ntrain = NULL;
  PyObject *__pyx_v_task = NULL;
  PyObject *__pyx_v_rep = NULL;
  PyObject *__pyx_v_lbl = NULL;
  PyObject *__pyx_v_rvpos = NULL;
  PyObject *__pyx_v_rvneg = NULL;
  PyObject *__pyx_v_nvpos = NULL;
  PyObject *__pyx_v_nvneg = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_mdl = NULL;
  PyObject *__pyx_9genexpr44__pyx_v_lbl = NULL;
  PyObject *__pyx_9genexpr44__pyx_v_seqs = NULL;
  PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_8__init___2generator12 = 0;
  PyObject *__pyx_9genexpr46__pyx_v_lbl = NULL;
  PyObject *__pyx_9genexpr46__pyx_v_s = NULL;
  PyObject *__pyx_9genexpr47__pyx_v_s = NULL;
  PyObject *__pyx_9genexpr48__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_22___init__(__pyx_ptype_6gnocis_6models___pyx_scope_struct_22___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 919, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_INCREF(__pyx_v_validationSet);
  __Pyx_INCREF(__pyx_v_ratioNegPos);

  /* "gnocis/models.pyx":920
 * 
 * 	def __init__(self, models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = None, storeTrained = True):
 * 		self.models = []             # <<<<<<<<<<<<<<
 * 		self.PRC = {}
 * 		self.ROC = {}
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_models, __pyx_t_1) < 0) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":921
 * 	def __init__(self, models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = None, storeTrained = True):
 * 		self.models = []
 * 		self.PRC = {}             # <<<<<<<<<<<<<<
 * 		self.ROC = {}
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_PRC, __pyx_t_1) < 0) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":922
 * 		self.models = []
 * 		self.PRC = {}
 * 		self.ROC = {}             # <<<<<<<<<<<<<<
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 * 		vpos, vneg = validationSet.withLabel([ labelPositive, labelNegative ])
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ROC, __pyx_t_1) < 0) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":923
 * 		self.PRC = {}
 * 		self.ROC = {}
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative             # <<<<<<<<<<<<<<
 * 		vpos, vneg = validationSet.withLabel([ labelPositive, labelNegative ])
 * 		if ratioNegPos == None:
 */
  __pyx_t_1 = __pyx_v_labelPositive;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_labelNegative;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_labelPositive, __pyx_t_1) < 0) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_labelNegative, __pyx_t_2) < 0) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":924
 * 		self.ROC = {}
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 * 		vpos, vneg = validationSet.withLabel([ labelPositive, labelNegative ])             # <<<<<<<<<<<<<<
 * 		if ratioNegPos == None:
 * 			ratioNegPos = len(vneg) / len(vpos)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_validationSet, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_labelPositive);
  __Pyx_GIVEREF(__pyx_v_labelPositive);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_labelPositive);
  __Pyx_INCREF(__pyx_v_labelNegative);
  __Pyx_GIVEREF(__pyx_v_labelNegative);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_v_labelNegative);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 924, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 924, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 924, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_vpos = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_vneg = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":925
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 * 		vpos, vneg = validationSet.withLabel([ labelPositive, labelNegative ])
 * 		if ratioNegPos == None:             # <<<<<<<<<<<<<<
 * 			ratioNegPos = len(vneg) / len(vpos)
 * 		self.ratioNegPos = ratioNegPos
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_ratioNegPos, Py_None, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 925, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "gnocis/models.pyx":926
 * 		vpos, vneg = validationSet.withLabel([ labelPositive, labelNegative ])
 * 		if ratioNegPos == None:
 * 			ratioNegPos = len(vneg) / len(vpos)             # <<<<<<<<<<<<<<
 * 		self.ratioNegPos = ratioNegPos
 * 		self.ratioTrainTest = ratioTrainTest
 */
    __pyx_t_7 = PyObject_Length(__pyx_v_vneg); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 926, __pyx_L1_error)
    __pyx_t_8 = PyObject_Length(__pyx_v_vpos); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 926, __pyx_L1_error)
    if (unlikely(__pyx_t_8 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 926, __pyx_L1_error)
    }
    __pyx_t_2 = PyFloat_FromDouble((((double)__pyx_t_7) / ((double)__pyx_t_8))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_ratioNegPos, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":925
 * 		self.labelPositive, self.labelNegative = labelPositive, labelNegative
 * 		vpos, vneg = validationSet.withLabel([ labelPositive, labelNegative ])
 * 		if ratioNegPos == None:             # <<<<<<<<<<<<<<
 * 			ratioNegPos = len(vneg) / len(vpos)
 * 		self.ratioNegPos = ratioNegPos
 */
  }

  /* "gnocis/models.pyx":927
 * 		if ratioNegPos == None:
 * 			ratioNegPos = len(vneg) / len(vpos)
 * 		self.ratioNegPos = ratioNegPos             # <<<<<<<<<<<<<<
 * 		self.ratioTrainTest = ratioTrainTest
 * 		self.trainingSet = trainingSet
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ratioNegPos, __pyx_v_ratioNegPos) < 0) __PYX_ERR(0, 927, __pyx_L1_error)

  /* "gnocis/models.pyx":928
 * 			ratioNegPos = len(vneg) / len(vpos)
 * 		self.ratioNegPos = ratioNegPos
 * 		self.ratioTrainTest = ratioTrainTest             # <<<<<<<<<<<<<<
 * 		self.trainingSet = trainingSet
 * 		self.validationSet = validationSet
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ratioTrainTest, __pyx_v_ratioTrainTest) < 0) __PYX_ERR(0, 928, __pyx_L1_error)

  /* "gnocis/models.pyx":929
 * 		self.ratioNegPos = ratioNegPos
 * 		self.ratioTrainTest = ratioTrainTest
 * 		self.trainingSet = trainingSet             # <<<<<<<<<<<<<<
 * 		self.validationSet = validationSet
 * 		self.storeTrained = storeTrained
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_trainingSet, __pyx_v_trainingSet) < 0) __PYX_ERR(0, 929, __pyx_L1_error)

  /* "gnocis/models.pyx":930
 * 		self.ratioTrainTest = ratioTrainTest
 * 		self.trainingSet = trainingSet
 * 		self.validationSet = validationSet             # <<<<<<<<<<<<<<
 * 		self.storeTrained = storeTrained
 * 		self.trainedModels = {}
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_validationSet, __pyx_v_validationSet) < 0) __PYX_ERR(0, 930, __pyx_L1_error)

  /* "gnocis/models.pyx":931
 * 		self.trainingSet = trainingSet
 * 		self.validationSet = validationSet
 * 		self.storeTrained = storeTrained             # <<<<<<<<<<<<<<
 * 		self.trainedModels = {}
 * 		self.predictions = {}
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_storeTrained, __pyx_v_storeTrained) < 0) __PYX_ERR(0, 931, __pyx_L1_error)

  /* "gnocis/models.pyx":932
 * 		self.validationSet = validationSet
 * 		self.storeTrained = storeTrained
 * 		self.trainedModels = {}             # <<<<<<<<<<<<<<
 * 		self.predictions = {}
 * 		# Construct cross-validation set
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels, __pyx_t_2) < 0) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":933
 * 		self.storeTrained = storeTrained
 * 		self.trainedModels = {}
 * 		self.predictions = {}             # <<<<<<<<<<<<<<
 * 		# Construct cross-validation set
 * 		if validationSet == None:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_predictions, __pyx_t_2) < 0) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":935
 * 		self.predictions = {}
 * 		# Construct cross-validation set
 * 		if validationSet == None:             # <<<<<<<<<<<<<<
 * 			validationSet = trainingSet
 * 		tslabels = list(trainingSet.labels())
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_validationSet, Py_None, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 935, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 935, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "gnocis/models.pyx":936
 * 		# Construct cross-validation set
 * 		if validationSet == None:
 * 			validationSet = trainingSet             # <<<<<<<<<<<<<<
 * 		tslabels = list(trainingSet.labels())
 * 		tsByLabel = {
 */
    __Pyx_INCREF(__pyx_v_trainingSet);
    __Pyx_DECREF_SET(__pyx_v_validationSet, __pyx_v_trainingSet);

    /* "gnocis/models.pyx":935
 * 		self.predictions = {}
 * 		# Construct cross-validation set
 * 		if validationSet == None:             # <<<<<<<<<<<<<<
 * 			validationSet = trainingSet
 * 		tslabels = list(trainingSet.labels())
 */
  }

  /* "gnocis/models.pyx":937
 * 		if validationSet == None:
 * 			validationSet = trainingSet
 * 		tslabels = list(trainingSet.labels())             # <<<<<<<<<<<<<<
 * 		tsByLabel = {
 * 			lbl: seqs.sequences[:]
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_trainingSet, __pyx_n_s_labels); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tslabels = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":938
 * 			validationSet = trainingSet
 * 		tslabels = list(trainingSet.labels())
 * 		tsByLabel = {             # <<<<<<<<<<<<<<
 * 			lbl: seqs.sequences[:]
 * 			for lbl, seqs in zip(tslabels, trainingSet.withLabel(tslabels))
 */
  { /* enter inner scope */
    __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 938, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "gnocis/models.pyx":940
 * 		tsByLabel = {
 * 			lbl: seqs.sequences[:]
 * 			for lbl, seqs in zip(tslabels, trainingSet.withLabel(tslabels))             # <<<<<<<<<<<<<<
 * 		}
 * 		ntrain = int( min(len(tsByLabel[lbl]) for lbl in tsByLabel) * ratioTrainTest )
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_trainingSet, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_v_tslabels) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_tslabels);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 940, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_tslabels);
    __Pyx_GIVEREF(__pyx_v_tslabels);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_tslabels);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 940, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 940, __pyx_L9_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 940, __pyx_L9_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 940, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 940, __pyx_L9_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 940, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_9(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 940, __pyx_L9_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
        PyObject* sequence = __pyx_t_2;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 940, __pyx_L9_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 940, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 940, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_11 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 940, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_5 = Py_TYPE(__pyx_t_11)->tp_iternext;
        index = 0; __pyx_t_4 = __pyx_t_5(__pyx_t_11); if (unlikely(!__pyx_t_4)) goto __pyx_L12_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_10 = __pyx_t_5(__pyx_t_11); if (unlikely(!__pyx_t_10)) goto __pyx_L12_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_10);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_11), 2) < 0) __PYX_ERR(0, 940, __pyx_L9_error)
        __pyx_t_5 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L13_unpacking_done;
        __pyx_L12_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_5 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 940, __pyx_L9_error)
        __pyx_L13_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr44__pyx_v_lbl, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr44__pyx_v_seqs, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "gnocis/models.pyx":939
 * 		tslabels = list(trainingSet.labels())
 * 		tsByLabel = {
 * 			lbl: seqs.sequences[:]             # <<<<<<<<<<<<<<
 * 			for lbl, seqs in zip(tslabels, trainingSet.withLabel(tslabels))
 * 		}
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr44__pyx_v_seqs, __pyx_n_s_sequences); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__18, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 939, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_3, (PyObject*)__pyx_9genexpr44__pyx_v_lbl, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 939, __pyx_L9_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "gnocis/models.pyx":940
 * 		tsByLabel = {
 * 			lbl: seqs.sequences[:]
 * 			for lbl, seqs in zip(tslabels, trainingSet.withLabel(tslabels))             # <<<<<<<<<<<<<<
 * 		}
 * 		ntrain = int( min(len(tsByLabel[lbl]) for lbl in tsByLabel) * ratioTrainTest )
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr44__pyx_v_lbl); __pyx_9genexpr44__pyx_v_lbl = 0;
    __Pyx_XDECREF(__pyx_9genexpr44__pyx_v_seqs); __pyx_9genexpr44__pyx_v_seqs = 0;
    goto __pyx_L14_exit_scope;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_9genexpr44__pyx_v_lbl); __pyx_9genexpr44__pyx_v_lbl = 0;
    __Pyx_XDECREF(__pyx_9genexpr44__pyx_v_seqs); __pyx_9genexpr44__pyx_v_seqs = 0;
    goto __pyx_L1_error;
    __pyx_L14_exit_scope:;
  } /* exit inner scope */
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_tsByLabel = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":942
 * 			for lbl, seqs in zip(tslabels, trainingSet.withLabel(tslabels))
 * 		}
 * 		ntrain = int( min(len(tsByLabel[lbl]) for lbl in tsByLabel) * ratioTrainTest )             # <<<<<<<<<<<<<<
 * 		self.cvtrain = []
 * 		self.cvval = []
 */
  __pyx_t_3 = __pyx_pf_6gnocis_6models_15crossvalidation_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_v_ratioTrainTest); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ntrain = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":943
 * 		}
 * 		ntrain = int( min(len(tsByLabel[lbl]) for lbl in tsByLabel) * ratioTrainTest )
 * 		self.cvtrain = []             # <<<<<<<<<<<<<<
 * 		self.cvval = []
 * 		self.repeats = repeats
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_cvtrain, __pyx_t_1) < 0) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":944
 * 		ntrain = int( min(len(tsByLabel[lbl]) for lbl in tsByLabel) * ratioTrainTest )
 * 		self.cvtrain = []
 * 		self.cvval = []             # <<<<<<<<<<<<<<
 * 		self.repeats = repeats
 * 		print('Cross-validation')
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_cvval, __pyx_t_1) < 0) __PYX_ERR(0, 944, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":945
 * 		self.cvtrain = []
 * 		self.cvval = []
 * 		self.repeats = repeats             # <<<<<<<<<<<<<<
 * 		print('Cross-validation')
 * 		print(' - Training set: %s'%str(trainingSet))
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_repeats, __pyx_v_repeats) < 0) __PYX_ERR(0, 945, __pyx_L1_error)

  /* "gnocis/models.pyx":946
 * 		self.cvval = []
 * 		self.repeats = repeats
 * 		print('Cross-validation')             # <<<<<<<<<<<<<<
 * 		print(' - Training set: %s'%str(trainingSet))
 * 		print(' - Validation set: %s'%str(validationSet))
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":947
 * 		self.repeats = repeats
 * 		print('Cross-validation')
 * 		print(' - Training set: %s'%str(trainingSet))             # <<<<<<<<<<<<<<
 * 		print(' - Validation set: %s'%str(validationSet))
 * 		print(' - Training sequences per class: %d'%ntrain)
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_trainingSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Training_set_s, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":948
 * 		print('Cross-validation')
 * 		print(' - Training set: %s'%str(trainingSet))
 * 		print(' - Validation set: %s'%str(validationSet))             # <<<<<<<<<<<<<<
 * 		print(' - Training sequences per class: %d'%ntrain)
 * 		print(' - Repeats: %d'%repeats)
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_validationSet); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Validation_set_s, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":949
 * 		print(' - Training set: %s'%str(trainingSet))
 * 		print(' - Validation set: %s'%str(validationSet))
 * 		print(' - Training sequences per class: %d'%ntrain)             # <<<<<<<<<<<<<<
 * 		print(' - Repeats: %d'%repeats)
 * 		task = progressTask('Generating training/test sets', steps = self.repeats)
 */
  __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Training_sequences_per_class_d, __pyx_v_ntrain); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":950
 * 		print(' - Validation set: %s'%str(validationSet))
 * 		print(' - Training sequences per class: %d'%ntrain)
 * 		print(' - Repeats: %d'%repeats)             # <<<<<<<<<<<<<<
 * 		task = progressTask('Generating training/test sets', steps = self.repeats)
 * 		for rep in range(repeats):
 */
  __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Repeats_d, __pyx_v_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":951
 * 		print(' - Training sequences per class: %d'%ntrain)
 * 		print(' - Repeats: %d'%repeats)
 * 		task = progressTask('Generating training/test sets', steps = self.repeats)             # <<<<<<<<<<<<<<
 * 		for rep in range(repeats):
 * 			task.update(step = rep)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_steps, __pyx_t_10) < 0) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__20, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_task = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "gnocis/models.pyx":952
 * 		print(' - Repeats: %d'%repeats)
 * 		task = progressTask('Generating training/test sets', steps = self.repeats)
 * 		for rep in range(repeats):             # <<<<<<<<<<<<<<
 * 			task.update(step = rep)
 * 			# Construct training set as balanced subset of shuffled sequences
 */
  __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_repeats); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
    __pyx_t_3 = __pyx_t_10; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 952, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 952, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 952, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 952, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 952, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 952, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
      }
    } else {
      __pyx_t_10 = __pyx_t_9(__pyx_t_3);
      if (unlikely(!__pyx_t_10)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 952, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_10);
    }
    __Pyx_XDECREF_SET(__pyx_v_rep, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "gnocis/models.pyx":953
 * 		task = progressTask('Generating training/test sets', steps = self.repeats)
 * 		for rep in range(repeats):
 * 			task.update(step = rep)             # <<<<<<<<<<<<<<
 * 			# Construct training set as balanced subset of shuffled sequences
 * 			for lbl in tsByLabel:
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_step, __pyx_v_rep) < 0) __PYX_ERR(0, 953, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":955
 * 			task.update(step = rep)
 * 			# Construct training set as balanced subset of shuffled sequences
 * 			for lbl in tsByLabel:             # <<<<<<<<<<<<<<
 * 				random.shuffle(tsByLabel[lbl])
 * 			self.cvtrain.append( sequences('Training set', [
 */
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_v_tsByLabel, 1, ((PyObject *)NULL), (&__pyx_t_12), (&__pyx_t_13)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_1;
    __pyx_t_1 = 0;
    while (1) {
      __pyx_t_14 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_12, &__pyx_t_7, &__pyx_t_1, NULL, NULL, __pyx_t_13);
      if (unlikely(__pyx_t_14 == 0)) break;
      if (unlikely(__pyx_t_14 == -1)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_lbl, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "gnocis/models.pyx":956
 * 			# Construct training set as balanced subset of shuffled sequences
 * 			for lbl in tsByLabel:
 * 				random.shuffle(tsByLabel[lbl])             # <<<<<<<<<<<<<<
 * 			self.cvtrain.append( sequences('Training set', [
 * 				s
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_random); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 956, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_shuffle); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 956, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_tsByLabel, __pyx_v_lbl); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 956, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_11, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 956, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":957
 * 			for lbl in tsByLabel:
 * 				random.shuffle(tsByLabel[lbl])
 * 			self.cvtrain.append( sequences('Training set', [             # <<<<<<<<<<<<<<
 * 				s
 * 				for lbl in tsByLabel
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvtrain); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "gnocis/models.pyx":959
 * 			self.cvtrain.append( sequences('Training set', [
 * 				s
 * 				for lbl in tsByLabel             # <<<<<<<<<<<<<<
 * 				for s in tsByLabel[lbl][:ntrain]
 * 			]) )
 */
      __pyx_t_12 = 0;
      __pyx_t_10 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_v_tsByLabel, 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_13)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 959, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_4);
      __pyx_t_4 = __pyx_t_10;
      __pyx_t_10 = 0;
      while (1) {
        __pyx_t_14 = __Pyx_dict_iter_next(__pyx_t_4, __pyx_t_7, &__pyx_t_12, &__pyx_t_10, NULL, NULL, __pyx_t_13);
        if (unlikely(__pyx_t_14 == 0)) break;
        if (unlikely(__pyx_t_14 == -1)) __PYX_ERR(0, 959, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_XDECREF_SET(__pyx_9genexpr46__pyx_v_lbl, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "gnocis/models.pyx":960
 * 				s
 * 				for lbl in tsByLabel
 * 				for s in tsByLabel[lbl][:ntrain]             # <<<<<<<<<<<<<<
 * 			]) )
 * 			# Construct validation set from independent sequences
 */
        __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_tsByLabel, __pyx_9genexpr46__pyx_v_lbl); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 960, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_GetSlice(__pyx_t_10, 0, 0, NULL, &__pyx_v_ntrain, NULL, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 960, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (likely(PyList_CheckExact(__pyx_t_11)) || PyTuple_CheckExact(__pyx_t_11)) {
          __pyx_t_10 = __pyx_t_11; __Pyx_INCREF(__pyx_t_10); __pyx_t_15 = 0;
          __pyx_t_16 = NULL;
        } else {
          __pyx_t_15 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 960, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_16 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 960, __pyx_L21_error)
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        for (;;) {
          if (likely(!__pyx_t_16)) {
            if (likely(PyList_CheckExact(__pyx_t_10))) {
              if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_10)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_11 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_15); __Pyx_INCREF(__pyx_t_11); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 960, __pyx_L21_error)
              #else
              __pyx_t_11 = PySequence_ITEM(__pyx_t_10, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 960, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_11);
              #endif
            } else {
              if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_15); __Pyx_INCREF(__pyx_t_11); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 960, __pyx_L21_error)
              #else
              __pyx_t_11 = PySequence_ITEM(__pyx_t_10, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 960, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_11);
              #endif
            }
          } else {
            __pyx_t_11 = __pyx_t_16(__pyx_t_10);
            if (unlikely(!__pyx_t_11)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 960, __pyx_L21_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_11);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr46__pyx_v_s, __pyx_t_11);
          __pyx_t_11 = 0;

          /* "gnocis/models.pyx":958
 * 				random.shuffle(tsByLabel[lbl])
 * 			self.cvtrain.append( sequences('Training set', [
 * 				s             # <<<<<<<<<<<<<<
 * 				for lbl in tsByLabel
 * 				for s in tsByLabel[lbl][:ntrain]
 */
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr46__pyx_v_s))) __PYX_ERR(0, 957, __pyx_L21_error)

          /* "gnocis/models.pyx":960
 * 				s
 * 				for lbl in tsByLabel
 * 				for s in tsByLabel[lbl][:ntrain]             # <<<<<<<<<<<<<<
 * 			]) )
 * 			# Construct validation set from independent sequences
 */
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_9genexpr46__pyx_v_lbl); __pyx_9genexpr46__pyx_v_lbl = 0;
      __Pyx_XDECREF(__pyx_9genexpr46__pyx_v_s); __pyx_9genexpr46__pyx_v_s = 0;
      goto __pyx_L26_exit_scope;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_9genexpr46__pyx_v_lbl); __pyx_9genexpr46__pyx_v_lbl = 0;
      __Pyx_XDECREF(__pyx_9genexpr46__pyx_v_s); __pyx_9genexpr46__pyx_v_s = 0;
      goto __pyx_L1_error;
      __pyx_L26_exit_scope:;
    } /* exit inner scope */

    /* "gnocis/models.pyx":957
 * 			for lbl in tsByLabel:
 * 				random.shuffle(tsByLabel[lbl])
 * 			self.cvtrain.append( sequences('Training set', [             # <<<<<<<<<<<<<<
 * 				s
 * 				for lbl in tsByLabel
 */
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_kp_u_Training_set_2);
    __Pyx_GIVEREF(__pyx_kp_u_Training_set_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Training_set_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_sequences), __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_17 = __Pyx_PyObject_Append(__pyx_t_2, __pyx_t_1); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":963
 * 			]) )
 * 			# Construct validation set from independent sequences
 * 			rvpos = [ s for s in vpos if s not in self.cvtrain[-1] ]             # <<<<<<<<<<<<<<
 * 			rvneg = [ s for s in vneg if s not in self.cvtrain[-1] ]
 * 			random.shuffle(rvpos)
 */
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 963, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (likely(PyList_CheckExact(__pyx_v_vpos)) || PyTuple_CheckExact(__pyx_v_vpos)) {
        __pyx_t_2 = __pyx_v_vpos; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
        __pyx_t_16 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_vpos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 963, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_16 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 963, __pyx_L29_error)
      }
      for (;;) {
        if (likely(!__pyx_t_16)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 963, __pyx_L29_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 963, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 963, __pyx_L29_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 963, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_16(__pyx_t_2);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 963, __pyx_L29_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr47__pyx_v_s, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvtrain); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 963, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_4, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 963, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_9genexpr47__pyx_v_s, __pyx_t_10, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 963, __pyx_L29_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_18 = (__pyx_t_6 != 0);
        if (__pyx_t_18) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr47__pyx_v_s))) __PYX_ERR(0, 963, __pyx_L29_error)
        }
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_9genexpr47__pyx_v_s); __pyx_9genexpr47__pyx_v_s = 0;
      goto __pyx_L33_exit_scope;
      __pyx_L29_error:;
      __Pyx_XDECREF(__pyx_9genexpr47__pyx_v_s); __pyx_9genexpr47__pyx_v_s = 0;
      goto __pyx_L1_error;
      __pyx_L33_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_rvpos, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":964
 * 			# Construct validation set from independent sequences
 * 			rvpos = [ s for s in vpos if s not in self.cvtrain[-1] ]
 * 			rvneg = [ s for s in vneg if s not in self.cvtrain[-1] ]             # <<<<<<<<<<<<<<
 * 			random.shuffle(rvpos)
 * 			random.shuffle(rvneg)
 */
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L36_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (likely(PyList_CheckExact(__pyx_v_vneg)) || PyTuple_CheckExact(__pyx_v_vneg)) {
        __pyx_t_2 = __pyx_v_vneg; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
        __pyx_t_16 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_vneg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 964, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_16 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 964, __pyx_L36_error)
      }
      for (;;) {
        if (likely(!__pyx_t_16)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_10); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 964, __pyx_L36_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 964, __pyx_L36_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_10); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 964, __pyx_L36_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 964, __pyx_L36_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          }
        } else {
          __pyx_t_10 = __pyx_t_16(__pyx_t_2);
          if (unlikely(!__pyx_t_10)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 964, __pyx_L36_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_10);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr48__pyx_v_s, __pyx_t_10);
        __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvtrain); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 964, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_10, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 964, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_18 = (__Pyx_PySequence_ContainsTF(__pyx_9genexpr48__pyx_v_s, __pyx_t_4, Py_NE)); if (unlikely(__pyx_t_18 < 0)) __PYX_ERR(0, 964, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = (__pyx_t_18 != 0);
        if (__pyx_t_6) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr48__pyx_v_s))) __PYX_ERR(0, 964, __pyx_L36_error)
        }
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_9genexpr48__pyx_v_s); __pyx_9genexpr48__pyx_v_s = 0;
      goto __pyx_L40_exit_scope;
      __pyx_L36_error:;
      __Pyx_XDECREF(__pyx_9genexpr48__pyx_v_s); __pyx_9genexpr48__pyx_v_s = 0;
      goto __pyx_L1_error;
      __pyx_L40_exit_scope:;
    } /* exit inner scope */
    __Pyx_XDECREF_SET(__pyx_v_rvneg, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":965
 * 			rvpos = [ s for s in vpos if s not in self.cvtrain[-1] ]
 * 			rvneg = [ s for s in vneg if s not in self.cvtrain[-1] ]
 * 			random.shuffle(rvpos)             # <<<<<<<<<<<<<<
 * 			random.shuffle(rvneg)
 * 			nvpos = int( min(len(rvpos), len(rvneg)/ratioNegPos) )
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_random); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_shuffle); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_2, __pyx_v_rvpos) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_rvpos);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":966
 * 			rvneg = [ s for s in vneg if s not in self.cvtrain[-1] ]
 * 			random.shuffle(rvpos)
 * 			random.shuffle(rvneg)             # <<<<<<<<<<<<<<
 * 			nvpos = int( min(len(rvpos), len(rvneg)/ratioNegPos) )
 * 			nvneg = int(nvpos * ratioNegPos)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_random); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_shuffle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v_rvneg) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_rvneg);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":967
 * 			random.shuffle(rvpos)
 * 			random.shuffle(rvneg)
 * 			nvpos = int( min(len(rvpos), len(rvneg)/ratioNegPos) )             # <<<<<<<<<<<<<<
 * 			nvneg = int(nvpos * ratioNegPos)
 * 			self.cvval.append( sequences('Validation set', rvpos[:nvpos] + rvneg[:nvneg]) )
 */
    __pyx_t_7 = PyList_GET_SIZE(__pyx_v_rvneg); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 967, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_v_ratioNegPos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = PyList_GET_SIZE(__pyx_v_rvpos); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 967, __pyx_L1_error)
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_1 = __pyx_t_2;
    } else {
      __pyx_t_10 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 967, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_1 = __pyx_t_10;
      __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_nvpos, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":968
 * 			random.shuffle(rvneg)
 * 			nvpos = int( min(len(rvpos), len(rvneg)/ratioNegPos) )
 * 			nvneg = int(nvpos * ratioNegPos)             # <<<<<<<<<<<<<<
 * 			self.cvval.append( sequences('Validation set', rvpos[:nvpos] + rvneg[:nvneg]) )
 * 			task.update(step = rep+1)
 */
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_nvpos, __pyx_v_ratioNegPos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_nvneg, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":969
 * 			nvpos = int( min(len(rvpos), len(rvneg)/ratioNegPos) )
 * 			nvneg = int(nvpos * ratioNegPos)
 * 			self.cvval.append( sequences('Validation set', rvpos[:nvpos] + rvneg[:nvneg]) )             # <<<<<<<<<<<<<<
 * 			task.update(step = rep+1)
 * 		task.done()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvval); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_nvpos);
    __pyx_t_2 = __pyx_v_nvpos;
    __pyx_t_6 = (__pyx_t_2 == Py_None);
    if (__pyx_t_6) {
      __pyx_t_7 = PY_SSIZE_T_MAX;
    } else {
      __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L1_error)
      __pyx_t_7 = __pyx_t_12;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyList_GetSlice(__pyx_v_rvpos, 0, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_nvneg);
    __pyx_t_10 = __pyx_v_nvneg;
    __pyx_t_6 = (__pyx_t_10 == Py_None);
    if (__pyx_t_6) {
      __pyx_t_7 = PY_SSIZE_T_MAX;
    } else {
      __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_10); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L1_error)
      __pyx_t_7 = __pyx_t_12;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyList_GetSlice(__pyx_v_rvneg, 0, __pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = PyNumber_Add(__pyx_t_2, __pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_kp_u_Validation_set);
    __Pyx_GIVEREF(__pyx_kp_u_Validation_set);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_kp_u_Validation_set);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_9sequences_sequences), __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_17 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_t_4); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 969, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "gnocis/models.pyx":970
 * 			nvneg = int(nvpos * ratioNegPos)
 * 			self.cvval.append( sequences('Validation set', rvpos[:nvpos] + rvneg[:nvneg]) )
 * 			task.update(step = rep+1)             # <<<<<<<<<<<<<<
 * 		task.done()
 * 		# Cross-validate models
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_v_rep, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_step, __pyx_t_10) < 0) __PYX_ERR(0, 970, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "gnocis/models.pyx":952
 * 		print(' - Repeats: %d'%repeats)
 * 		task = progressTask('Generating training/test sets', steps = self.repeats)
 * 		for rep in range(repeats):             # <<<<<<<<<<<<<<
 * 			task.update(step = rep)
 * 			# Construct training set as balanced subset of shuffled sequences
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":971
 * 			self.cvval.append( sequences('Validation set', rvpos[:nvpos] + rvneg[:nvneg]) )
 * 			task.update(step = rep+1)
 * 		task.done()             # <<<<<<<<<<<<<<
 * 		# Cross-validate models
 * 		task = progressTask('CV', steps = len(models))
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_done); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":973
 * 		task.done()
 * 		# Cross-validate models
 * 		task = progressTask('CV', steps = len(models))             # <<<<<<<<<<<<<<
 * 		for i, mdl in enumerate(models):
 * 			task.update(step = i)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_8 = PyObject_Length(__pyx_v_models); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 973, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_steps, __pyx_t_1) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__21, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF_SET(__pyx_v_task, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":974
 * 		# Cross-validate models
 * 		task = progressTask('CV', steps = len(models))
 * 		for i, mdl in enumerate(models):             # <<<<<<<<<<<<<<
 * 			task.update(step = i)
 * 			self.addModel(mdl)
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_models)) || PyTuple_CheckExact(__pyx_v_models)) {
    __pyx_t_10 = __pyx_v_models; __Pyx_INCREF(__pyx_t_10); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_models); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_9 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 974, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_10))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_10)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 974, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_10, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 974, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 974, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_10, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 974, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_9(__pyx_t_10);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 974, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_mdl, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "gnocis/models.pyx":975
 * 		task = progressTask('CV', steps = len(models))
 * 		for i, mdl in enumerate(models):
 * 			task.update(step = i)             # <<<<<<<<<<<<<<
 * 			self.addModel(mdl)
 * 		task.done()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_step, __pyx_v_i) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":976
 * 		for i, mdl in enumerate(models):
 * 			task.update(step = i)
 * 			self.addModel(mdl)             # <<<<<<<<<<<<<<
 * 		task.done()
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_addModel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_mdl) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_mdl);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":974
 * 		# Cross-validate models
 * 		task = progressTask('CV', steps = len(models))
 * 		for i, mdl in enumerate(models):             # <<<<<<<<<<<<<<
 * 			task.update(step = i)
 * 			self.addModel(mdl)
 */
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":977
 * 			task.update(step = i)
 * 			self.addModel(mdl)
 * 		task.done()             # <<<<<<<<<<<<<<
 * 
 * 	"""
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_done); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":919
 * 	"""
 * 
 * 	def __init__(self, models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = None, storeTrained = True):             # <<<<<<<<<<<<<<
 * 		self.models = []
 * 		self.PRC = {}
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vpos);
  __Pyx_XDECREF(__pyx_v_vneg);
  __Pyx_XDECREF(__pyx_v_tslabels);
  __Pyx_XDECREF(__pyx_v_ntrain);
  __Pyx_XDECREF(__pyx_v_task);
  __Pyx_XDECREF(__pyx_v_rep);
  __Pyx_XDECREF(__pyx_v_lbl);
  __Pyx_XDECREF(__pyx_v_rvpos);
  __Pyx_XDECREF(__pyx_v_rvneg);
  __Pyx_XDECREF(__pyx_v_nvpos);
  __Pyx_XDECREF(__pyx_v_nvneg);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_9genexpr44__pyx_v_lbl);
  __Pyx_XDECREF(__pyx_9genexpr44__pyx_v_seqs);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_15crossvalidation_8__init___2generator12);
  __Pyx_XDECREF(__pyx_9genexpr46__pyx_v_lbl);
  __Pyx_XDECREF(__pyx_9genexpr46__pyx_v_s);
  __Pyx_XDECREF(__pyx_9genexpr47__pyx_v_s);
  __Pyx_XDECREF(__pyx_9genexpr48__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_validationSet);
  __Pyx_XDECREF(__pyx_v_ratioNegPos);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":986
 * 	:type mdl: model
 * 	"""
 * 	def addModel(self, mdl):             # <<<<<<<<<<<<<<
 * 		self.PRC[mdl] = []
 * 		self.ROC[mdl] = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_3addModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_3addModel = {"addModel", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_15crossvalidation_3addModel, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_3addModel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_mdl = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addModel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_mdl,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mdl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addModel", 1, 2, 2, 1); __PYX_ERR(0, 986, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "addModel") < 0)) __PYX_ERR(0, 986, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_mdl = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addModel", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 986, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.addModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_2addModel(__pyx_self, __pyx_v_self, __pyx_v_mdl);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_2addModel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_mdl) {
  PyObject *__pyx_v_task = NULL;
  PyObject *__pyx_v_rep = NULL;
  PyObject *__pyx_v_imdl = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addModel", 0);

  /* "gnocis/models.pyx":987
 * 	"""
 * 	def addModel(self, mdl):
 * 		self.PRC[mdl] = []             # <<<<<<<<<<<<<<
 * 		self.ROC[mdl] = []
 * 		self.trainedModels[mdl] = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_PRC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_v_mdl, __pyx_t_1) < 0)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":988
 * 	def addModel(self, mdl):
 * 		self.PRC[mdl] = []
 * 		self.ROC[mdl] = []             # <<<<<<<<<<<<<<
 * 		self.trainedModels[mdl] = []
 * 		task = progressTask(mdl.name, steps = self.repeats*3)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ROC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_v_mdl, __pyx_t_1) < 0)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":989
 * 		self.PRC[mdl] = []
 * 		self.ROC[mdl] = []
 * 		self.trainedModels[mdl] = []             # <<<<<<<<<<<<<<
 * 		task = progressTask(mdl.name, steps = self.repeats*3)
 * 		for rep in range(self.repeats):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_v_mdl, __pyx_t_1) < 0)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":990
 * 		self.ROC[mdl] = []
 * 		self.trainedModels[mdl] = []
 * 		task = progressTask(mdl.name, steps = self.repeats*3)             # <<<<<<<<<<<<<<
 * 		for rep in range(self.repeats):
 * 			task.update(step = rep*3)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_mdl, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyNumber_Multiply(__pyx_t_4, __pyx_int_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_steps, __pyx_t_5) < 0) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_task = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "gnocis/models.pyx":991
 * 		self.trainedModels[mdl] = []
 * 		task = progressTask(mdl.name, steps = self.repeats*3)
 * 		for rep in range(self.repeats):             # <<<<<<<<<<<<<<
 * 			task.update(step = rep*3)
 * 			imdl = mdl.train(self.cvtrain[rep])
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 991, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 991, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 991, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 991, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_5);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 991, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_rep, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":992
 * 		task = progressTask(mdl.name, steps = self.repeats*3)
 * 		for rep in range(self.repeats):
 * 			task.update(step = rep*3)             # <<<<<<<<<<<<<<
 * 			imdl = mdl.train(self.cvtrain[rep])
 * 			if self.storeTrained: self.trainedModels[mdl].append(imdl)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyNumber_Multiply(__pyx_v_rep, __pyx_int_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_step, __pyx_t_1) < 0) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gnocis/models.pyx":993
 * 		for rep in range(self.repeats):
 * 			task.update(step = rep*3)
 * 			imdl = mdl.train(self.cvtrain[rep])             # <<<<<<<<<<<<<<
 * 			if self.storeTrained: self.trainedModels[mdl].append(imdl)
 * 			task.update(step = rep*3+1)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mdl, __pyx_n_s_train); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvtrain); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_rep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_imdl, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gnocis/models.pyx":994
 * 			task.update(step = rep*3)
 * 			imdl = mdl.train(self.cvtrain[rep])
 * 			if self.storeTrained: self.trainedModels[mdl].append(imdl)             # <<<<<<<<<<<<<<
 * 			task.update(step = rep*3+1)
 * 			self.PRC[mdl].append(imdl.getPRC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_storeTrained); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_8) {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_mdl); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 994, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = __Pyx_PyObject_Append(__pyx_t_3, __pyx_v_imdl); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 994, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "gnocis/models.pyx":995
 * 			imdl = mdl.train(self.cvtrain[rep])
 * 			if self.storeTrained: self.trainedModels[mdl].append(imdl)
 * 			task.update(step = rep*3+1)             # <<<<<<<<<<<<<<
 * 			self.PRC[mdl].append(imdl.getPRC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))
 * 			task.update(step = rep*3+2)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_rep, __pyx_int_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_step, __pyx_t_2) < 0) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":996
 * 			if self.storeTrained: self.trainedModels[mdl].append(imdl)
 * 			task.update(step = rep*3+1)
 * 			self.PRC[mdl].append(imdl.getPRC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))             # <<<<<<<<<<<<<<
 * 			task.update(step = rep*3+2)
 * 			self.ROC[mdl].append(imdl.getROC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_PRC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_mdl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_imdl, __pyx_n_s_getPRC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvval); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_rep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_labelPositive); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_labelPositive, __pyx_t_10) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_labelNegative); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_labelNegative, __pyx_t_10) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_t_10); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "gnocis/models.pyx":997
 * 			task.update(step = rep*3+1)
 * 			self.PRC[mdl].append(imdl.getPRC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))
 * 			task.update(step = rep*3+2)             # <<<<<<<<<<<<<<
 * 			self.ROC[mdl].append(imdl.getROC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))
 * 			task.update(step = rep*3+3)
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_rep, __pyx_int_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_step, __pyx_t_3) < 0) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "gnocis/models.pyx":998
 * 			self.PRC[mdl].append(imdl.getPRC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))
 * 			task.update(step = rep*3+2)
 * 			self.ROC[mdl].append(imdl.getROC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))             # <<<<<<<<<<<<<<
 * 			task.update(step = rep*3+3)
 * 		self.models.append(mdl)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ROC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_mdl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_imdl, __pyx_n_s_getROC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvval); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_v_rep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_labelPositive); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_labelPositive, __pyx_t_2) < 0) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_labelNegative); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_labelNegative, __pyx_t_2) < 0) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_t_2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "gnocis/models.pyx":999
 * 			task.update(step = rep*3+2)
 * 			self.ROC[mdl].append(imdl.getROC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))
 * 			task.update(step = rep*3+3)             # <<<<<<<<<<<<<<
 * 		self.models.append(mdl)
 * 		task.done()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyNumber_Multiply(__pyx_v_rep, __pyx_int_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_3, 3, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_step, __pyx_t_10) < 0) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "gnocis/models.pyx":991
 * 		self.trainedModels[mdl] = []
 * 		task = progressTask(mdl.name, steps = self.repeats*3)
 * 		for rep in range(self.repeats):             # <<<<<<<<<<<<<<
 * 			task.update(step = rep*3)
 * 			imdl = mdl.train(self.cvtrain[rep])
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":1000
 * 			self.ROC[mdl].append(imdl.getROC(self.cvval[rep], labelPositive = self.labelPositive, labelNegative = self.labelNegative))
 * 			task.update(step = rep*3+3)
 * 		self.models.append(mdl)             # <<<<<<<<<<<<<<
 * 		task.done()
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyObject_Append(__pyx_t_5, __pyx_v_mdl); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":1001
 * 			task.update(step = rep*3+3)
 * 		self.models.append(mdl)
 * 		task.done()             # <<<<<<<<<<<<<<
 * 
 * 	def calibrate(self, calibrator, labelPositive = positive):
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_done); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":986
 * 	:type mdl: model
 * 	"""
 * 	def addModel(self, mdl):             # <<<<<<<<<<<<<<
 * 		self.PRC[mdl] = []
 * 		self.ROC[mdl] = []
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.addModel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_task);
  __Pyx_XDECREF(__pyx_v_rep);
  __Pyx_XDECREF(__pyx_v_imdl);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1003
 * 		task.done()
 * 
 * 	def calibrate(self, calibrator, labelPositive = positive):             # <<<<<<<<<<<<<<
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 */

static PyObject *__pyx_pf_6gnocis_6models_58__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults12, __pyx_self)->__pyx_arg_labelPositive);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults12, __pyx_self)->__pyx_arg_labelPositive);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults12, __pyx_self)->__pyx_arg_labelPositive);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gnocis.models.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_5calibrate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_5calibrate = {"calibrate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_15crossvalidation_5calibrate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_5calibrate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_calibrator = 0;
  PyObject *__pyx_v_labelPositive = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calibrate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_calibrator,&__pyx_n_s_labelPositive,0};
    PyObject* values[3] = {0,0,0};
    __pyx_defaults12 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults12, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_labelPositive;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_calibrator)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calibrate", 0, 2, 3, 1); __PYX_ERR(0, 1003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calibrate") < 0)) __PYX_ERR(0, 1003, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_calibrator = values[1];
    __pyx_v_labelPositive = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calibrate", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1003, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.calibrate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_4calibrate(__pyx_self, __pyx_v_self, __pyx_v_calibrator, __pyx_v_labelPositive);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_4calibrate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_calibrator, PyObject *__pyx_v_labelPositive) {
  PyObject *__pyx_v_mtask = NULL;
  PyObject *__pyx_v_imdl = NULL;
  PyObject *__pyx_v_mdl = NULL;
  PyObject *__pyx_v_rtask = NULL;
  PyObject *__pyx_v_rep = NULL;
  PyObject *__pyx_v_vpos = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calibrate", 0);

  /* "gnocis/models.pyx":1004
 * 
 * 	def calibrate(self, calibrator, labelPositive = positive):
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))             # <<<<<<<<<<<<<<
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_steps, __pyx_t_3) < 0) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__22, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_mtask = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":1005
 * 	def calibrate(self, calibrator, labelPositive = positive):
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):             # <<<<<<<<<<<<<<
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1005, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1005, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1005, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1005, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_mdl, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_imdl, __pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":1006
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)             # <<<<<<<<<<<<<<
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_mtask, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_step, __pyx_v_imdl) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "gnocis/models.pyx":1007
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:             # <<<<<<<<<<<<<<
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_v_mdl, __pyx_t_7, Py_NE)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_10 = (__pyx_t_9 != 0);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_8 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_7, __pyx_v_mdl); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_11 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_6, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_8) {

      /* "gnocis/models.pyx":1008
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))             # <<<<<<<<<<<<<<
 * 				continue
 * 			rtask = progressTask('Repeat', steps = self.repeats)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_mdl, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Warning_model_s_not_trained_Skip, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "gnocis/models.pyx":1009
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue             # <<<<<<<<<<<<<<
 * 			rtask = progressTask('Repeat', steps = self.repeats)
 * 			# TODO Implement multiclass calibration
 */
      goto __pyx_L3_continue;

      /* "gnocis/models.pyx":1007
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:             # <<<<<<<<<<<<<<
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue
 */
    }

    /* "gnocis/models.pyx":1010
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue
 * 			rtask = progressTask('Repeat', steps = self.repeats)             # <<<<<<<<<<<<<<
 * 			# TODO Implement multiclass calibration
 * 			for rep in range(self.repeats):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_steps, __pyx_t_6) < 0) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__23, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_rtask, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "gnocis/models.pyx":1012
 * 			rtask = progressTask('Repeat', steps = self.repeats)
 * 			# TODO Implement multiclass calibration
 * 			for rep in range(self.repeats):             # <<<<<<<<<<<<<<
 * 				rtask.update(step = rep)
 * 				imdl = self.trainedModels[mdl][rep]
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_6 = __pyx_t_7; __Pyx_INCREF(__pyx_t_6); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_12 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1012, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_11); __Pyx_INCREF(__pyx_t_7); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 1012, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1012, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_11); __Pyx_INCREF(__pyx_t_7); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 1012, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1012, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_12(__pyx_t_6);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1012, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_rep, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "gnocis/models.pyx":1013
 * 			# TODO Implement multiclass calibration
 * 			for rep in range(self.repeats):
 * 				rtask.update(step = rep)             # <<<<<<<<<<<<<<
 * 				imdl = self.trainedModels[mdl][rep]
 * 				vpos = self.cvval[rep].withLabel(labelPositive)
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_rtask, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1013, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1013, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_step, __pyx_v_rep) < 0) __PYX_ERR(0, 1013, __pyx_L1_error)
      __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1013, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "gnocis/models.pyx":1014
 * 			for rep in range(self.repeats):
 * 				rtask.update(step = rep)
 * 				imdl = self.trainedModels[mdl][rep]             # <<<<<<<<<<<<<<
 * 				vpos = self.cvval[rep].withLabel(labelPositive)
 * 				calibrator(imdl, vpos, rep)
 */
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_13, __pyx_v_mdl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_rep); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_imdl, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "gnocis/models.pyx":1015
 * 				rtask.update(step = rep)
 * 				imdl = self.trainedModels[mdl][rep]
 * 				vpos = self.cvval[rep].withLabel(labelPositive)             # <<<<<<<<<<<<<<
 * 				calibrator(imdl, vpos, rep)
 * 			rtask.done()
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvval); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_rep); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_13 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_v_labelPositive) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_labelPositive);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_vpos, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "gnocis/models.pyx":1016
 * 				imdl = self.trainedModels[mdl][rep]
 * 				vpos = self.cvval[rep].withLabel(labelPositive)
 * 				calibrator(imdl, vpos, rep)             # <<<<<<<<<<<<<<
 * 			rtask.done()
 * 		mtask.done()
 */
      __Pyx_INCREF(__pyx_v_calibrator);
      __pyx_t_2 = __pyx_v_calibrator; __pyx_t_7 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_14 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_v_imdl, __pyx_v_vpos, __pyx_v_rep};
        __pyx_t_13 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_14, 3+__pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_13);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_v_imdl, __pyx_v_vpos, __pyx_v_rep};
        __pyx_t_13 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_14, 3+__pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_13);
      } else
      #endif
      {
        __pyx_t_15 = PyTuple_New(3+__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_INCREF(__pyx_v_imdl);
        __Pyx_GIVEREF(__pyx_v_imdl);
        PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_14, __pyx_v_imdl);
        __Pyx_INCREF(__pyx_v_vpos);
        __Pyx_GIVEREF(__pyx_v_vpos);
        PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_14, __pyx_v_vpos);
        __Pyx_INCREF(__pyx_v_rep);
        __Pyx_GIVEREF(__pyx_v_rep);
        PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_14, __pyx_v_rep);
        __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_15, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "gnocis/models.pyx":1012
 * 			rtask = progressTask('Repeat', steps = self.repeats)
 * 			# TODO Implement multiclass calibration
 * 			for rep in range(self.repeats):             # <<<<<<<<<<<<<<
 * 				rtask.update(step = rep)
 * 				imdl = self.trainedModels[mdl][rep]
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "gnocis/models.pyx":1017
 * 				vpos = self.cvval[rep].withLabel(labelPositive)
 * 				calibrator(imdl, vpos, rep)
 * 			rtask.done()             # <<<<<<<<<<<<<<
 * 		mtask.done()
 * 
 */
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_rtask, __pyx_n_s_done); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_13);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "gnocis/models.pyx":1005
 * 	def calibrate(self, calibrator, labelPositive = positive):
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):             # <<<<<<<<<<<<<<
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":1018
 * 				calibrator(imdl, vpos, rep)
 * 			rtask.done()
 * 		mtask.done()             # <<<<<<<<<<<<<<
 * 
 * 	def predict(self, genome):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mtask, __pyx_n_s_done); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":1003
 * 		task.done()
 * 
 * 	def calibrate(self, calibrator, labelPositive = positive):             # <<<<<<<<<<<<<<
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.calibrate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mtask);
  __Pyx_XDECREF(__pyx_v_imdl);
  __Pyx_XDECREF(__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_v_rtask);
  __Pyx_XDECREF(__pyx_v_rep);
  __Pyx_XDECREF(__pyx_v_vpos);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1020
 * 		mtask.done()
 * 
 * 	def predict(self, genome):             # <<<<<<<<<<<<<<
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_7predict(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_7predict = {"predict", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_15crossvalidation_7predict, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_7predict(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_genome = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("predict (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_genome,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_genome)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("predict", 1, 2, 2, 1); __PYX_ERR(0, 1020, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "predict") < 0)) __PYX_ERR(0, 1020, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_genome = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("predict", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1020, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_6predict(__pyx_self, __pyx_v_self, __pyx_v_genome);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_6predict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_genome) {
  PyObject *__pyx_v_mtask = NULL;
  PyObject *__pyx_v_imdl = NULL;
  PyObject *__pyx_v_mdl = NULL;
  PyObject *__pyx_v_cvpred = NULL;
  PyObject *__pyx_v_rtask = NULL;
  PyObject *__pyx_v_rep = NULL;
  PyObject *__pyx_v_pred = NULL;
  PyObject *__pyx_9genexpr49__pyx_v_mdl = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("predict", 0);

  /* "gnocis/models.pyx":1021
 * 
 * 	def predict(self, genome):
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))             # <<<<<<<<<<<<<<
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_steps, __pyx_t_3) < 0) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__24, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_mtask = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gnocis/models.pyx":1022
 * 	def predict(self, genome):
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):             # <<<<<<<<<<<<<<
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1022, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1022, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1022, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1022, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_mdl, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_imdl, __pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":1023
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)             # <<<<<<<<<<<<<<
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_mtask, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_step, __pyx_v_imdl) < 0) __PYX_ERR(0, 1023, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "gnocis/models.pyx":1024
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:             # <<<<<<<<<<<<<<
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_v_mdl, __pyx_t_7, Py_NE)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_10 = (__pyx_t_9 != 0);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_8 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_7, __pyx_v_mdl); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_11 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_6, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_8) {

      /* "gnocis/models.pyx":1025
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))             # <<<<<<<<<<<<<<
 * 				continue
 * 			cvpred = []
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_mdl, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Warning_model_s_not_trained_Skip, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "gnocis/models.pyx":1026
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue             # <<<<<<<<<<<<<<
 * 			cvpred = []
 * 			rtask = progressTask('Repeat', steps = self.repeats)
 */
      goto __pyx_L3_continue;

      /* "gnocis/models.pyx":1024
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:             # <<<<<<<<<<<<<<
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue
 */
    }

    /* "gnocis/models.pyx":1027
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue
 * 			cvpred = []             # <<<<<<<<<<<<<<
 * 			rtask = progressTask('Repeat', steps = self.repeats)
 * 			# TODO Implement multiclass prediction
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_cvpred, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "gnocis/models.pyx":1028
 * 				continue
 * 			cvpred = []
 * 			rtask = progressTask('Repeat', steps = self.repeats)             # <<<<<<<<<<<<<<
 * 			# TODO Implement multiclass prediction
 * 			for rep in range(self.repeats):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_steps, __pyx_t_6) < 0) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__23, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_rtask, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "gnocis/models.pyx":1030
 * 			rtask = progressTask('Repeat', steps = self.repeats)
 * 			# TODO Implement multiclass prediction
 * 			for rep in range(self.repeats):             # <<<<<<<<<<<<<<
 * 				rtask.update(step = rep)
 * 				imdl = self.trainedModels[mdl][rep]
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_6 = __pyx_t_7; __Pyx_INCREF(__pyx_t_6); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1030, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_12 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1030, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_11); __Pyx_INCREF(__pyx_t_7); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 1030, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1030, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_11); __Pyx_INCREF(__pyx_t_7); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 1030, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1030, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_12(__pyx_t_6);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1030, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_rep, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "gnocis/models.pyx":1031
 * 			# TODO Implement multiclass prediction
 * 			for rep in range(self.repeats):
 * 				rtask.update(step = rep)             # <<<<<<<<<<<<<<
 * 				imdl = self.trainedModels[mdl][rep]
 * 				pred = imdl.predict(genome.sequences)
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_rtask, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1031, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1031, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_step, __pyx_v_rep) < 0) __PYX_ERR(0, 1031, __pyx_L1_error)
      __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1031, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "gnocis/models.pyx":1032
 * 			for rep in range(self.repeats):
 * 				rtask.update(step = rep)
 * 				imdl = self.trainedModels[mdl][rep]             # <<<<<<<<<<<<<<
 * 				pred = imdl.predict(genome.sequences)
 * 				cvpred.append(pred)
 */
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_13, __pyx_v_mdl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_rep); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_imdl, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "gnocis/models.pyx":1033
 * 				rtask.update(step = rep)
 * 				imdl = self.trainedModels[mdl][rep]
 * 				pred = imdl.predict(genome.sequences)             # <<<<<<<<<<<<<<
 * 				cvpred.append(pred)
 * 			rtask.done()
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_imdl, __pyx_n_s_predict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_genome, __pyx_n_s_sequences); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_13 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_14, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_pred, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "gnocis/models.pyx":1034
 * 				imdl = self.trainedModels[mdl][rep]
 * 				pred = imdl.predict(genome.sequences)
 * 				cvpred.append(pred)             # <<<<<<<<<<<<<<
 * 			rtask.done()
 * 			# TODO Once multiclass prediction is implemented, replace with labels
 */
      __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_cvpred, __pyx_v_pred); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1034, __pyx_L1_error)

      /* "gnocis/models.pyx":1030
 * 			rtask = progressTask('Repeat', steps = self.repeats)
 * 			# TODO Implement multiclass prediction
 * 			for rep in range(self.repeats):             # <<<<<<<<<<<<<<
 * 				rtask.update(step = rep)
 * 				imdl = self.trainedModels[mdl][rep]
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "gnocis/models.pyx":1035
 * 				pred = imdl.predict(genome.sequences)
 * 				cvpred.append(pred)
 * 			rtask.done()             # <<<<<<<<<<<<<<
 * 			# TODO Once multiclass prediction is implemented, replace with labels
 * 			self.predictions[mdl] = CVModelPredictions(
 */
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_rtask, __pyx_n_s_done); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_13);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "gnocis/models.pyx":1037
 * 			rtask.done()
 * 			# TODO Once multiclass prediction is implemented, replace with labels
 * 			self.predictions[mdl] = CVModelPredictions(             # <<<<<<<<<<<<<<
 * 				model = mdl,
 * 				trainedModels = self.trainedModels[mdl],
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_latin1_CVModelPredictions_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "gnocis/models.pyx":1038
 * 			# TODO Once multiclass prediction is implemented, replace with labels
 * 			self.predictions[mdl] = CVModelPredictions(
 * 				model = mdl,             # <<<<<<<<<<<<<<
 * 				trainedModels = self.trainedModels[mdl],
 * 				labelPredictionCurves = { positive: cvpred })
 */
    __pyx_t_13 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_model, __pyx_v_mdl) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)

    /* "gnocis/models.pyx":1039
 * 			self.predictions[mdl] = CVModelPredictions(
 * 				model = mdl,
 * 				trainedModels = self.trainedModels[mdl],             # <<<<<<<<<<<<<<
 * 				labelPredictionCurves = { positive: cvpred })
 * 		mtask.done()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainedModels); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_mdl); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_trainedModels, __pyx_t_7) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "gnocis/models.pyx":1040
 * 				model = mdl,
 * 				trainedModels = self.trainedModels[mdl],
 * 				labelPredictionCurves = { positive: cvpred })             # <<<<<<<<<<<<<<
 * 		mtask.done()
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 */
    __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_positive); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_7, __pyx_t_2, __pyx_v_cvpred) < 0) __PYX_ERR(0, 1040, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_13, __pyx_n_s_labelPredictionCurves, __pyx_t_7) < 0) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "gnocis/models.pyx":1037
 * 			rtask.done()
 * 			# TODO Once multiclass prediction is implemented, replace with labels
 * 			self.predictions[mdl] = CVModelPredictions(             # <<<<<<<<<<<<<<
 * 				model = mdl,
 * 				trainedModels = self.trainedModels[mdl],
 */
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictions); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (unlikely(PyObject_SetItem(__pyx_t_13, __pyx_v_mdl, __pyx_t_7) < 0)) __PYX_ERR(0, 1037, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "gnocis/models.pyx":1022
 * 	def predict(self, genome):
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):             # <<<<<<<<<<<<<<
 * 			mtask.update(step = imdl)
 * 			if mdl not in self.trainedModels or len(self.trainedModels[mdl]) != self.repeats:
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":1041
 * 				trainedModels = self.trainedModels[mdl],
 * 				labelPredictionCurves = { positive: cvpred })
 * 		mtask.done()             # <<<<<<<<<<<<<<
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mtask, __pyx_n_s_done); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1041, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":1042
 * 				labelPredictionCurves = { positive: cvpred })
 * 		mtask.done()
 * 		return [ self.predictions[mdl] for mdl in self.models ]             # <<<<<<<<<<<<<<
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1042, __pyx_L12_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L12_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_7 = __pyx_t_1; __Pyx_INCREF(__pyx_t_7); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1042, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1042, __pyx_L12_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1042, __pyx_L12_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_7, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1042, __pyx_L12_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_7, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_5(__pyx_t_7);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1042, __pyx_L12_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr49__pyx_v_mdl, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_predictions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_9genexpr49__pyx_v_mdl); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1042, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_13))) __PYX_ERR(0, 1042, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_9genexpr49__pyx_v_mdl); __pyx_9genexpr49__pyx_v_mdl = 0;
    goto __pyx_L15_exit_scope;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_9genexpr49__pyx_v_mdl); __pyx_9genexpr49__pyx_v_mdl = 0;
    goto __pyx_L1_error;
    __pyx_L15_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":1020
 * 		mtask.done()
 * 
 * 	def predict(self, genome):             # <<<<<<<<<<<<<<
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.predict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mtask);
  __Pyx_XDECREF(__pyx_v_imdl);
  __Pyx_XDECREF(__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_v_cvpred);
  __Pyx_XDECREF(__pyx_v_rtask);
  __Pyx_XDECREF(__pyx_v_rep);
  __Pyx_XDECREF(__pyx_v_pred);
  __Pyx_XDECREF(__pyx_9genexpr49__pyx_v_mdl);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1044
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_9plotPRC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_9plotPRC = {"plotPRC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_15crossvalidation_9plotPRC, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_9plotPRC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_style = 0;
  PyObject *__pyx_v_returnHTML = 0;
  PyObject *__pyx_v_fontsizeLabels = 0;
  PyObject *__pyx_v_fontsizeLegend = 0;
  PyObject *__pyx_v_fontsizeAxis = 0;
  PyObject *__pyx_v_bboxAnchorTo = 0;
  PyObject *__pyx_v_legendLoc = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("plotPRC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_style,&__pyx_n_s_returnHTML,&__pyx_n_s_fontsizeLabels,&__pyx_n_s_fontsizeLegend,&__pyx_n_s_fontsizeAxis,&__pyx_n_s_bboxAnchorTo,&__pyx_n_s_legendLoc,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject*)__pyx_tuple__25));
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject*)__pyx_n_u_ggplot));
    values[4] = ((PyObject *)((PyObject *)Py_False));
    values[5] = ((PyObject *)((PyObject *)__pyx_int_18));
    values[6] = ((PyObject *)((PyObject *)__pyx_int_12));
    values[7] = ((PyObject *)((PyObject *)__pyx_int_10));
    values[8] = ((PyObject *)((PyObject*)__pyx_tuple__26));
    values[9] = ((PyObject *)((PyObject*)__pyx_kp_u_upper_left));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_returnHTML);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLabels);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLegend);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeAxis);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bboxAnchorTo);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_legendLoc);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "plotPRC") < 0)) __PYX_ERR(0, 1044, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_figsize = values[1];
    __pyx_v_outpath = values[2];
    __pyx_v_style = values[3];
    __pyx_v_returnHTML = values[4];
    __pyx_v_fontsizeLabels = values[5];
    __pyx_v_fontsizeLegend = values[6];
    __pyx_v_fontsizeAxis = values[7];
    __pyx_v_bboxAnchorTo = values[8];
    __pyx_v_legendLoc = values[9];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("plotPRC", 0, 1, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1044, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.plotPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_8plotPRC(__pyx_self, __pyx_v_self, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_style, __pyx_v_returnHTML, __pyx_v_fontsizeLabels, __pyx_v_fontsizeLegend, __pyx_v_fontsizeAxis, __pyx_v_bboxAnchorTo, __pyx_v_legendLoc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_7plotPRC_2generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":1064
 * 				for mdl in self.models:
 * 					curves = self.PRC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))             # <<<<<<<<<<<<<<
 * 					curvebyx = [
 * 						{
 */

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_7plotPRC_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_25_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_25_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1064, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_15crossvalidation_7plotPRC_2generator13, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_crossvalidation_plotPRC_locals_g, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.plotPRC.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_7plotPRC_2generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves)) { __Pyx_RaiseClosureNameError("curves"); __PYX_ERR(0, 1064, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1064, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1064, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1064, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1064, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_c)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_c)) {
      __pyx_t_4 = __pyx_cur_scope->__pyx_v_c; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_c); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1064, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1064, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1064, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1064, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1064, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1064, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1064, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_pt);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_pt, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_pt, __pyx_n_s_x); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1064, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 1064, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1044
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_8plotPRC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_style, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *__pyx_cur_scope;
  PyObject *__pyx_v_plt = NULL;
  PyObject *__pyx_v_base64 = NULL;
  PyObject *__pyx_v_BytesIO = NULL;
  PyObject *__pyx_v_display = NULL;
  PyObject *__pyx_v_HTML = NULL;
  PyObject *__pyx_v_fig = NULL;
  PyObject *__pyx_v_positives = NULL;
  PyObject *__pyx_v_negatives = NULL;
  double __pyx_v_ry;
  PyObject *__pyx_v_mdl = NULL;
  PyObject *__pyx_v_xs = NULL;
  PyObject *__pyx_v_curvebyx = NULL;
  PyObject *__pyx_v_curvebyxmax = NULL;
  PyObject *__pyx_v_curvebyxmin = NULL;
  PyObject *__pyx_v_meanCurve = NULL;
  PyObject *__pyx_v_CICurve = NULL;
  PyObject *__pyx_v_AUCs = NULL;
  PyObject *__pyx_v_bio = NULL;
  PyObject *__pyx_v_encoded = NULL;
  PyObject *__pyx_v_html = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_7plotPRC_2generator13 = 0;
  PyObject *__pyx_9genexpr51__pyx_v_curve = NULL;
  PyObject *__pyx_9genexpr52__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr53__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr54__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr55__pyx_v_i = NULL;
  CYTHON_UNUSED PyObject *__pyx_9genexpr55__pyx_v_curve = NULL;
  PyObject *__pyx_9genexpr56__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr57__pyx_v_i = NULL;
  CYTHON_UNUSED PyObject *__pyx_9genexpr57__pyx_v_curve = NULL;
  PyObject *__pyx_9genexpr58__pyx_v_x = NULL;
  struct __pyx_obj_6gnocis_10validation_point2D *__pyx_9genexpr58__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr59__pyx_v_x = NULL;
  struct __pyx_obj_6gnocis_10validation_point2D *__pyx_9genexpr59__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr60__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr61__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr61__pyx_v_ci = NULL;
  PyObject *__pyx_9genexpr62__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr62__pyx_v_ci = NULL;
  PyObject *__pyx_9genexpr63__pyx_v_curve = NULL;
  PyObject *__pyx_9genexpr64__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr65__pyx_v_pt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  int __pyx_t_19;
  PyObject *(*__pyx_t_20)(PyObject *);
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *(*__pyx_t_23)(PyObject *);
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  int __pyx_t_26;
  PyObject *__pyx_t_27 = NULL;
  Py_UCS4 __pyx_t_28;
  int __pyx_t_29;
  PyObject *__pyx_t_30 = NULL;
  int __pyx_t_31;
  int __pyx_t_32;
  char const *__pyx_t_33;
  PyObject *__pyx_t_34 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("plotPRC", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_24_plotPRC(__pyx_ptype_6gnocis_6models___pyx_scope_struct_24_plotPRC, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1044, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/models.pyx":1045
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "gnocis/models.pyx":1046
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 * 		try:
 * 			import matplotlib.pyplot as plt             # <<<<<<<<<<<<<<
 * 			import base64
 * 			from io import BytesIO
 */
      __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1046, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_n_s__5);
      __Pyx_GIVEREF(__pyx_n_s__5);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s__5);
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_matplotlib_pyplot, __pyx_t_4, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1046, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_plt = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "gnocis/models.pyx":1047
 * 		try:
 * 			import matplotlib.pyplot as plt
 * 			import base64             # <<<<<<<<<<<<<<
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 */
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_base64, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1047, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_base64 = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "gnocis/models.pyx":1048
 * 			import matplotlib.pyplot as plt
 * 			import base64
 * 			from io import BytesIO             # <<<<<<<<<<<<<<
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):
 */
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1048, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_n_s_BytesIO);
      __Pyx_GIVEREF(__pyx_n_s_BytesIO);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_BytesIO);
      __pyx_t_4 = __Pyx_Import(__pyx_n_s_io, __pyx_t_5, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1048, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_BytesIO); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1048, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_BytesIO = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "gnocis/models.pyx":1049
 * 			import base64
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML             # <<<<<<<<<<<<<<
 * 			with plt.style.context(style):
 * 				fig = plt.figure(figsize = figsize)
 */
      __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1049, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_n_s_display);
      __Pyx_GIVEREF(__pyx_n_s_display);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_display);
      __Pyx_INCREF(__pyx_n_s_HTML);
      __Pyx_GIVEREF(__pyx_n_s_HTML);
      PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_HTML);
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_IPython_core_display, __pyx_t_4, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1049, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_display); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1049, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_display = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_HTML); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1049, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_HTML = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "gnocis/models.pyx":1050
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 				fig = plt.figure(figsize = figsize)
 * 				# Expected random generalization
 */
      /*with:*/ {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_style); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1050, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_context); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1050, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_style) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_style);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1050, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1050, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1050, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1050, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_9);
            __Pyx_XGOTREF(__pyx_t_10);
            __Pyx_XGOTREF(__pyx_t_11);
            /*try:*/ {

              /* "gnocis/models.pyx":1051
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):
 * 				fig = plt.figure(figsize = figsize)             # <<<<<<<<<<<<<<
 * 				# Expected random generalization
 * 				positives, negatives = self.cvval[0].withLabel([ self.labelPositive, self.labelNegative ])
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_figure); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1051, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1051, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 1051, __pyx_L13_error)
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1051, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_v_fig = __pyx_t_4;
              __pyx_t_4 = 0;

              /* "gnocis/models.pyx":1053
 * 				fig = plt.figure(figsize = figsize)
 * 				# Expected random generalization
 * 				positives, negatives = self.cvval[0].withLabel([ self.labelPositive, self.labelNegative ])             # <<<<<<<<<<<<<<
 * 				ry = len(positives) / (len(positives) + len(negatives))
 * 				plt.plot(
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvval); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1053, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1053, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_withLabel); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1053, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_labelPositive); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1053, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_labelNegative); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1053, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_12 = PyList_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1053, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_GIVEREF(__pyx_t_5);
              PyList_SET_ITEM(__pyx_t_12, 0, __pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_8);
              PyList_SET_ITEM(__pyx_t_12, 1, __pyx_t_8);
              __pyx_t_5 = 0;
              __pyx_t_8 = 0;
              __pyx_t_8 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_8)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_8);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_12);
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1053, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
                PyObject* sequence = __pyx_t_4;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 1053, __pyx_L13_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
                  __pyx_t_12 = PyTuple_GET_ITEM(sequence, 1); 
                } else {
                  __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
                  __pyx_t_12 = PyList_GET_ITEM(sequence, 1); 
                }
                __Pyx_INCREF(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_12);
                #else
                __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1053, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_12 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1053, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_12);
                #endif
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              } else {
                Py_ssize_t index = -1;
                __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1053, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_13 = Py_TYPE(__pyx_t_8)->tp_iternext;
                index = 0; __pyx_t_6 = __pyx_t_13(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L19_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_6);
                index = 1; __pyx_t_12 = __pyx_t_13(__pyx_t_8); if (unlikely(!__pyx_t_12)) goto __pyx_L19_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_12);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_8), 2) < 0) __PYX_ERR(0, 1053, __pyx_L13_error)
                __pyx_t_13 = NULL;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                goto __pyx_L20_unpacking_done;
                __pyx_L19_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __pyx_t_13 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 1053, __pyx_L13_error)
                __pyx_L20_unpacking_done:;
              }
              __pyx_v_positives = __pyx_t_6;
              __pyx_t_6 = 0;
              __pyx_v_negatives = __pyx_t_12;
              __pyx_t_12 = 0;

              /* "gnocis/models.pyx":1054
 * 				# Expected random generalization
 * 				positives, negatives = self.cvval[0].withLabel([ self.labelPositive, self.labelNegative ])
 * 				ry = len(positives) / (len(positives) + len(negatives))             # <<<<<<<<<<<<<<
 * 				plt.plot(
 * 					[ 0., 1. ],
 */
              __pyx_t_14 = PyObject_Length(__pyx_v_positives); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1054, __pyx_L13_error)
              __pyx_t_15 = PyObject_Length(__pyx_v_positives); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1054, __pyx_L13_error)
              __pyx_t_16 = PyObject_Length(__pyx_v_negatives); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1054, __pyx_L13_error)
              __pyx_t_17 = (__pyx_t_15 + __pyx_t_16);
              if (unlikely(__pyx_t_17 == 0)) {
                PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                __PYX_ERR(0, 1054, __pyx_L13_error)
              }
              __pyx_v_ry = (((double)__pyx_t_14) / ((double)__pyx_t_17));

              /* "gnocis/models.pyx":1055
 * 				positives, negatives = self.cvval[0].withLabel([ self.labelPositive, self.labelNegative ])
 * 				ry = len(positives) / (len(positives) + len(negatives))
 * 				plt.plot(             # <<<<<<<<<<<<<<
 * 					[ 0., 1. ],
 * 					[ ry, ry ],
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_plot); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1055, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);

              /* "gnocis/models.pyx":1056
 * 				ry = len(positives) / (len(positives) + len(negatives))
 * 				plt.plot(
 * 					[ 0., 1. ],             # <<<<<<<<<<<<<<
 * 					[ ry, ry ],
 * 					linestyle = '--',
 */
              __pyx_t_12 = PyList_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1056, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_INCREF(__pyx_float_0_);
              __Pyx_GIVEREF(__pyx_float_0_);
              PyList_SET_ITEM(__pyx_t_12, 0, __pyx_float_0_);
              __Pyx_INCREF(__pyx_float_1_);
              __Pyx_GIVEREF(__pyx_float_1_);
              PyList_SET_ITEM(__pyx_t_12, 1, __pyx_float_1_);

              /* "gnocis/models.pyx":1057
 * 				plt.plot(
 * 					[ 0., 1. ],
 * 					[ ry, ry ],             # <<<<<<<<<<<<<<
 * 					linestyle = '--',
 * 					color = 'grey',
 */
              __pyx_t_6 = PyFloat_FromDouble(__pyx_v_ry); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1057, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_8 = PyFloat_FromDouble(__pyx_v_ry); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1057, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1057, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_6);
              PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_8);
              PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_8);
              __pyx_t_6 = 0;
              __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1055
 * 				positives, negatives = self.cvval[0].withLabel([ self.labelPositive, self.labelNegative ])
 * 				ry = len(positives) / (len(positives) + len(negatives))
 * 				plt.plot(             # <<<<<<<<<<<<<<
 * 					[ 0., 1. ],
 * 					[ ry, ry ],
 */
              __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1055, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_12);
              PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_12);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_5);
              __pyx_t_12 = 0;
              __pyx_t_5 = 0;

              /* "gnocis/models.pyx":1058
 * 					[ 0., 1. ],
 * 					[ ry, ry ],
 * 					linestyle = '--',             # <<<<<<<<<<<<<<
 * 					color = 'grey',
 * 					label = 'Expected at random')
 */
              __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1058, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_linestyle, __pyx_kp_u__6) < 0) __PYX_ERR(0, 1058, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_color, __pyx_n_u_grey) < 0) __PYX_ERR(0, 1058, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_label, __pyx_kp_u_Expected_at_random) < 0) __PYX_ERR(0, 1058, __pyx_L13_error)

              /* "gnocis/models.pyx":1055
 * 				positives, negatives = self.cvval[0].withLabel([ self.labelPositive, self.labelNegative ])
 * 				ry = len(positives) / (len(positives) + len(negatives))
 * 				plt.plot(             # <<<<<<<<<<<<<<
 * 					[ 0., 1. ],
 * 					[ ry, ry ],
 */
              __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, __pyx_t_5); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1055, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

              /* "gnocis/models.pyx":1062
 * 					label = 'Expected at random')
 * 				# Curves per model
 * 				for mdl in self.models:             # <<<<<<<<<<<<<<
 * 					curves = self.PRC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 */
              __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1062, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_12);
              if (likely(PyList_CheckExact(__pyx_t_12)) || PyTuple_CheckExact(__pyx_t_12)) {
                __pyx_t_5 = __pyx_t_12; __Pyx_INCREF(__pyx_t_5); __pyx_t_17 = 0;
                __pyx_t_18 = NULL;
              } else {
                __pyx_t_17 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1062, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_18 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1062, __pyx_L13_error)
              }
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              for (;;) {
                if (likely(!__pyx_t_18)) {
                  if (likely(PyList_CheckExact(__pyx_t_5))) {
                    if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_5)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_12 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_17); __Pyx_INCREF(__pyx_t_12); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 1062, __pyx_L13_error)
                    #else
                    __pyx_t_12 = PySequence_ITEM(__pyx_t_5, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1062, __pyx_L13_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    #endif
                  } else {
                    if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_17); __Pyx_INCREF(__pyx_t_12); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 1062, __pyx_L13_error)
                    #else
                    __pyx_t_12 = PySequence_ITEM(__pyx_t_5, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1062, __pyx_L13_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    #endif
                  }
                } else {
                  __pyx_t_12 = __pyx_t_18(__pyx_t_5);
                  if (unlikely(!__pyx_t_12)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 1062, __pyx_L13_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_12);
                }
                __Pyx_XDECREF_SET(__pyx_v_mdl, __pyx_t_12);
                __pyx_t_12 = 0;

                /* "gnocis/models.pyx":1063
 * 				# Curves per model
 * 				for mdl in self.models:
 * 					curves = self.PRC[mdl]             # <<<<<<<<<<<<<<
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 * 					curvebyx = [
 */
                __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_PRC); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1063, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_12);
                __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_t_12, __pyx_v_mdl); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1063, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_curves);
                __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_curves, __pyx_t_8);
                __Pyx_GIVEREF(__pyx_t_8);
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1064
 * 				for mdl in self.models:
 * 					curves = self.PRC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))             # <<<<<<<<<<<<<<
 * 					curvebyx = [
 * 						{
 */
                __pyx_t_12 = __pyx_pf_6gnocis_6models_15crossvalidation_7plotPRC_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1064, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_12);
                __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1064, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __pyx_t_12 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1064, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_12);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = PySequence_List(__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1064, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __pyx_t_8 = ((PyObject*)__pyx_t_4);
                __pyx_t_4 = 0;
                __pyx_t_19 = PyList_Sort(__pyx_t_8); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 1064, __pyx_L13_error)
                __Pyx_XDECREF_SET(__pyx_v_xs, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1065
 * 					curves = self.PRC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 * 					curvebyx = [             # <<<<<<<<<<<<<<
 * 						{
 * 							x: [ pt.y for pt in curve if pt.x == x ]
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1065, __pyx_L25_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1070
 * 							for x in xs
 * 						}
 * 						for curve in curves             # <<<<<<<<<<<<<<
 * 					]
 * 					curvebyxmax = {
 */
                  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_curves)) {
                    __pyx_t_4 = __pyx_cur_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_4); __pyx_t_14 = 0;
                    __pyx_t_20 = NULL;
                  } else {
                    __pyx_t_14 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_curves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1070, __pyx_L25_error)
                    __Pyx_GOTREF(__pyx_t_4);
                    __pyx_t_20 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1070, __pyx_L25_error)
                  }
                  for (;;) {
                    if (likely(!__pyx_t_20)) {
                      if (likely(PyList_CheckExact(__pyx_t_4))) {
                        if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_4)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_12 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_12); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1070, __pyx_L25_error)
                        #else
                        __pyx_t_12 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1070, __pyx_L25_error)
                        __Pyx_GOTREF(__pyx_t_12);
                        #endif
                      } else {
                        if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_12); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1070, __pyx_L25_error)
                        #else
                        __pyx_t_12 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1070, __pyx_L25_error)
                        __Pyx_GOTREF(__pyx_t_12);
                        #endif
                      }
                    } else {
                      __pyx_t_12 = __pyx_t_20(__pyx_t_4);
                      if (unlikely(!__pyx_t_12)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 1070, __pyx_L25_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_12);
                    }
                    __Pyx_XDECREF_SET(__pyx_9genexpr51__pyx_v_curve, __pyx_t_12);
                    __pyx_t_12 = 0;
                    { /* enter inner scope */

                      /* "gnocis/models.pyx":1066
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 * 					curvebyx = [
 * 						{             # <<<<<<<<<<<<<<
 * 							x: [ pt.y for pt in curve if pt.x == x ]
 * 							for x in xs
 */
                      __pyx_t_12 = PyDict_New(); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1066, __pyx_L30_error)
                      __Pyx_GOTREF(__pyx_t_12);

                      /* "gnocis/models.pyx":1068
 * 						{
 * 							x: [ pt.y for pt in curve if pt.x == x ]
 * 							for x in xs             # <<<<<<<<<<<<<<
 * 						}
 * 						for curve in curves
 */
                      if (unlikely(__pyx_v_xs == Py_None)) {
                        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                        __PYX_ERR(0, 1068, __pyx_L30_error)
                      }
                      __pyx_t_6 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_6); __pyx_t_16 = 0;
                      for (;;) {
                        if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_6)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_21 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_16); __Pyx_INCREF(__pyx_t_21); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1068, __pyx_L30_error)
                        #else
                        __pyx_t_21 = PySequence_ITEM(__pyx_t_6, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1068, __pyx_L30_error)
                        __Pyx_GOTREF(__pyx_t_21);
                        #endif
                        __Pyx_XDECREF_SET(__pyx_9genexpr52__pyx_v_x, __pyx_t_21);
                        __pyx_t_21 = 0;

                        /* "gnocis/models.pyx":1067
 * 					curvebyx = [
 * 						{
 * 							x: [ pt.y for pt in curve if pt.x == x ]             # <<<<<<<<<<<<<<
 * 							for x in xs
 * 						}
 */
                        { /* enter inner scope */
                          __pyx_t_21 = PyList_New(0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1067, __pyx_L35_error)
                          __Pyx_GOTREF(__pyx_t_21);
                          if (likely(PyList_CheckExact(__pyx_9genexpr51__pyx_v_curve)) || PyTuple_CheckExact(__pyx_9genexpr51__pyx_v_curve)) {
                            __pyx_t_22 = __pyx_9genexpr51__pyx_v_curve; __Pyx_INCREF(__pyx_t_22); __pyx_t_15 = 0;
                            __pyx_t_23 = NULL;
                          } else {
                            __pyx_t_15 = -1; __pyx_t_22 = PyObject_GetIter(__pyx_9genexpr51__pyx_v_curve); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1067, __pyx_L35_error)
                            __Pyx_GOTREF(__pyx_t_22);
                            __pyx_t_23 = Py_TYPE(__pyx_t_22)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1067, __pyx_L35_error)
                          }
                          for (;;) {
                            if (likely(!__pyx_t_23)) {
                              if (likely(PyList_CheckExact(__pyx_t_22))) {
                                if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_22)) break;
                                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                                __pyx_t_24 = PyList_GET_ITEM(__pyx_t_22, __pyx_t_15); __Pyx_INCREF(__pyx_t_24); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1067, __pyx_L35_error)
                                #else
                                __pyx_t_24 = PySequence_ITEM(__pyx_t_22, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1067, __pyx_L35_error)
                                __Pyx_GOTREF(__pyx_t_24);
                                #endif
                              } else {
                                if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_22)) break;
                                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                                __pyx_t_24 = PyTuple_GET_ITEM(__pyx_t_22, __pyx_t_15); __Pyx_INCREF(__pyx_t_24); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1067, __pyx_L35_error)
                                #else
                                __pyx_t_24 = PySequence_ITEM(__pyx_t_22, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1067, __pyx_L35_error)
                                __Pyx_GOTREF(__pyx_t_24);
                                #endif
                              }
                            } else {
                              __pyx_t_24 = __pyx_t_23(__pyx_t_22);
                              if (unlikely(!__pyx_t_24)) {
                                PyObject* exc_type = PyErr_Occurred();
                                if (exc_type) {
                                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                                  else __PYX_ERR(0, 1067, __pyx_L35_error)
                                }
                                break;
                              }
                              __Pyx_GOTREF(__pyx_t_24);
                            }
                            __Pyx_XDECREF_SET(__pyx_9genexpr53__pyx_v_pt, __pyx_t_24);
                            __pyx_t_24 = 0;
                            __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr53__pyx_v_pt, __pyx_n_s_x); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1067, __pyx_L35_error)
                            __Pyx_GOTREF(__pyx_t_24);
                            __pyx_t_25 = PyObject_RichCompare(__pyx_t_24, __pyx_9genexpr52__pyx_v_x, Py_EQ); __Pyx_XGOTREF(__pyx_t_25); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1067, __pyx_L35_error)
                            __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                            __pyx_t_26 = __Pyx_PyObject_IsTrue(__pyx_t_25); if (unlikely(__pyx_t_26 < 0)) __PYX_ERR(0, 1067, __pyx_L35_error)
                            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                            if (__pyx_t_26) {
                              __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr53__pyx_v_pt, __pyx_n_s_y); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1067, __pyx_L35_error)
                              __Pyx_GOTREF(__pyx_t_25);
                              if (unlikely(__Pyx_ListComp_Append(__pyx_t_21, (PyObject*)__pyx_t_25))) __PYX_ERR(0, 1067, __pyx_L35_error)
                              __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                            }
                          }
                          __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                          __Pyx_XDECREF(__pyx_9genexpr53__pyx_v_pt); __pyx_9genexpr53__pyx_v_pt = 0;
                          goto __pyx_L39_exit_scope;
                          __pyx_L35_error:;
                          __Pyx_XDECREF(__pyx_9genexpr53__pyx_v_pt); __pyx_9genexpr53__pyx_v_pt = 0;
                          goto __pyx_L30_error;
                          __pyx_L39_exit_scope:;
                        } /* exit inner scope */
                        if (unlikely(PyDict_SetItem(__pyx_t_12, (PyObject*)__pyx_9genexpr52__pyx_v_x, (PyObject*)__pyx_t_21))) __PYX_ERR(0, 1067, __pyx_L30_error)
                        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

                        /* "gnocis/models.pyx":1068
 * 						{
 * 							x: [ pt.y for pt in curve if pt.x == x ]
 * 							for x in xs             # <<<<<<<<<<<<<<
 * 						}
 * 						for curve in curves
 */
                      }
                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                      __Pyx_XDECREF(__pyx_9genexpr52__pyx_v_x); __pyx_9genexpr52__pyx_v_x = 0;
                      goto __pyx_L40_exit_scope;
                      __pyx_L30_error:;
                      __Pyx_XDECREF(__pyx_9genexpr52__pyx_v_x); __pyx_9genexpr52__pyx_v_x = 0;
                      goto __pyx_L25_error;
                      __pyx_L40_exit_scope:;
                    } /* exit inner scope */
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 1065, __pyx_L25_error)
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

                    /* "gnocis/models.pyx":1070
 * 							for x in xs
 * 						}
 * 						for curve in curves             # <<<<<<<<<<<<<<
 * 					]
 * 					curvebyxmax = {
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr51__pyx_v_curve); __pyx_9genexpr51__pyx_v_curve = 0;
                  goto __pyx_L41_exit_scope;
                  __pyx_L25_error:;
                  __Pyx_XDECREF(__pyx_9genexpr51__pyx_v_curve); __pyx_9genexpr51__pyx_v_curve = 0;
                  goto __pyx_L13_error;
                  __pyx_L41_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_curvebyx, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1072
 * 						for curve in curves
 * 					]
 * 					curvebyxmax = {             # <<<<<<<<<<<<<<
 * 						x: [
 * 							max(curvebyx[i][x])
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1072, __pyx_L44_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1077
 * 							for i, curve in enumerate(curves)
 * 						]
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 					}
 * 					curvebyxmin = {
 */
                  if (unlikely(__pyx_v_xs == Py_None)) {
                    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                    __PYX_ERR(0, 1077, __pyx_L44_error)
                  }
                  __pyx_t_4 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_4); __pyx_t_14 = 0;
                  for (;;) {
                    if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_4)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_12 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_12); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1077, __pyx_L44_error)
                    #else
                    __pyx_t_12 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1077, __pyx_L44_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr54__pyx_v_x, __pyx_t_12);
                    __pyx_t_12 = 0;

                    /* "gnocis/models.pyx":1073
 * 					]
 * 					curvebyxmax = {
 * 						x: [             # <<<<<<<<<<<<<<
 * 							max(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)
 */
                    { /* enter inner scope */
                      __pyx_t_12 = PyList_New(0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1073, __pyx_L49_error)
                      __Pyx_GOTREF(__pyx_t_12);
                      __Pyx_INCREF(__pyx_int_0);
                      __pyx_t_6 = __pyx_int_0;

                      /* "gnocis/models.pyx":1075
 * 						x: [
 * 							max(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)             # <<<<<<<<<<<<<<
 * 						]
 * 						for x in xs
 */
                      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_curves)) {
                        __pyx_t_21 = __pyx_cur_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_21); __pyx_t_16 = 0;
                        __pyx_t_20 = NULL;
                      } else {
                        __pyx_t_16 = -1; __pyx_t_21 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_curves); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1075, __pyx_L49_error)
                        __Pyx_GOTREF(__pyx_t_21);
                        __pyx_t_20 = Py_TYPE(__pyx_t_21)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1075, __pyx_L49_error)
                      }
                      for (;;) {
                        if (likely(!__pyx_t_20)) {
                          if (likely(PyList_CheckExact(__pyx_t_21))) {
                            if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_21)) break;
                            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                            __pyx_t_22 = PyList_GET_ITEM(__pyx_t_21, __pyx_t_16); __Pyx_INCREF(__pyx_t_22); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1075, __pyx_L49_error)
                            #else
                            __pyx_t_22 = PySequence_ITEM(__pyx_t_21, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1075, __pyx_L49_error)
                            __Pyx_GOTREF(__pyx_t_22);
                            #endif
                          } else {
                            if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_21)) break;
                            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                            __pyx_t_22 = PyTuple_GET_ITEM(__pyx_t_21, __pyx_t_16); __Pyx_INCREF(__pyx_t_22); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1075, __pyx_L49_error)
                            #else
                            __pyx_t_22 = PySequence_ITEM(__pyx_t_21, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1075, __pyx_L49_error)
                            __Pyx_GOTREF(__pyx_t_22);
                            #endif
                          }
                        } else {
                          __pyx_t_22 = __pyx_t_20(__pyx_t_21);
                          if (unlikely(!__pyx_t_22)) {
                            PyObject* exc_type = PyErr_Occurred();
                            if (exc_type) {
                              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                              else __PYX_ERR(0, 1075, __pyx_L49_error)
                            }
                            break;
                          }
                          __Pyx_GOTREF(__pyx_t_22);
                        }
                        __Pyx_XDECREF_SET(__pyx_9genexpr55__pyx_v_curve, __pyx_t_22);
                        __pyx_t_22 = 0;
                        __Pyx_INCREF(__pyx_t_6);
                        __Pyx_XDECREF_SET(__pyx_9genexpr55__pyx_v_i, __pyx_t_6);
                        __pyx_t_22 = __Pyx_PyInt_AddObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1075, __pyx_L49_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        __Pyx_DECREF(__pyx_t_6);
                        __pyx_t_6 = __pyx_t_22;
                        __pyx_t_22 = 0;

                        /* "gnocis/models.pyx":1074
 * 					curvebyxmax = {
 * 						x: [
 * 							max(curvebyx[i][x])             # <<<<<<<<<<<<<<
 * 							for i, curve in enumerate(curves)
 * 						]
 */
                        __pyx_t_22 = __Pyx_PyObject_GetItem(__pyx_v_curvebyx, __pyx_9genexpr55__pyx_v_i); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1074, __pyx_L49_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        __pyx_t_25 = __Pyx_PyObject_GetItem(__pyx_t_22, __pyx_9genexpr54__pyx_v_x); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1074, __pyx_L49_error)
                        __Pyx_GOTREF(__pyx_t_25);
                        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                        __pyx_t_22 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_25); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1074, __pyx_L49_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                        if (unlikely(__Pyx_ListComp_Append(__pyx_t_12, (PyObject*)__pyx_t_22))) __PYX_ERR(0, 1073, __pyx_L49_error)
                        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

                        /* "gnocis/models.pyx":1075
 * 						x: [
 * 							max(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)             # <<<<<<<<<<<<<<
 * 						]
 * 						for x in xs
 */
                      }
                      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                      __Pyx_XDECREF(__pyx_9genexpr55__pyx_v_curve); __pyx_9genexpr55__pyx_v_curve = 0;
                      __Pyx_XDECREF(__pyx_9genexpr55__pyx_v_i); __pyx_9genexpr55__pyx_v_i = 0;
                      goto __pyx_L52_exit_scope;
                      __pyx_L49_error:;
                      __Pyx_XDECREF(__pyx_9genexpr55__pyx_v_curve); __pyx_9genexpr55__pyx_v_curve = 0;
                      __Pyx_XDECREF(__pyx_9genexpr55__pyx_v_i); __pyx_9genexpr55__pyx_v_i = 0;
                      goto __pyx_L44_error;
                      __pyx_L52_exit_scope:;
                    } /* exit inner scope */
                    if (unlikely(PyDict_SetItem(__pyx_t_8, (PyObject*)__pyx_9genexpr54__pyx_v_x, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 1073, __pyx_L44_error)
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

                    /* "gnocis/models.pyx":1077
 * 							for i, curve in enumerate(curves)
 * 						]
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 					}
 * 					curvebyxmin = {
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr54__pyx_v_x); __pyx_9genexpr54__pyx_v_x = 0;
                  goto __pyx_L53_exit_scope;
                  __pyx_L44_error:;
                  __Pyx_XDECREF(__pyx_9genexpr54__pyx_v_x); __pyx_9genexpr54__pyx_v_x = 0;
                  goto __pyx_L13_error;
                  __pyx_L53_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_curvebyxmax, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1079
 * 						for x in xs
 * 					}
 * 					curvebyxmin = {             # <<<<<<<<<<<<<<
 * 						x: [
 * 							min(curvebyx[i][x])
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1079, __pyx_L56_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1084
 * 							for i, curve in enumerate(curves)
 * 						]
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 					}
 * 					meanCurve = [
 */
                  if (unlikely(__pyx_v_xs == Py_None)) {
                    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                    __PYX_ERR(0, 1084, __pyx_L56_error)
                  }
                  __pyx_t_4 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_4); __pyx_t_14 = 0;
                  for (;;) {
                    if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_4)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_12 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_12); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1084, __pyx_L56_error)
                    #else
                    __pyx_t_12 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1084, __pyx_L56_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr56__pyx_v_x, __pyx_t_12);
                    __pyx_t_12 = 0;

                    /* "gnocis/models.pyx":1080
 * 					}
 * 					curvebyxmin = {
 * 						x: [             # <<<<<<<<<<<<<<
 * 							min(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)
 */
                    { /* enter inner scope */
                      __pyx_t_12 = PyList_New(0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1080, __pyx_L61_error)
                      __Pyx_GOTREF(__pyx_t_12);
                      __Pyx_INCREF(__pyx_int_0);
                      __pyx_t_6 = __pyx_int_0;

                      /* "gnocis/models.pyx":1082
 * 						x: [
 * 							min(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)             # <<<<<<<<<<<<<<
 * 						]
 * 						for x in xs
 */
                      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_curves)) {
                        __pyx_t_21 = __pyx_cur_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_21); __pyx_t_16 = 0;
                        __pyx_t_20 = NULL;
                      } else {
                        __pyx_t_16 = -1; __pyx_t_21 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_curves); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1082, __pyx_L61_error)
                        __Pyx_GOTREF(__pyx_t_21);
                        __pyx_t_20 = Py_TYPE(__pyx_t_21)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1082, __pyx_L61_error)
                      }
                      for (;;) {
                        if (likely(!__pyx_t_20)) {
                          if (likely(PyList_CheckExact(__pyx_t_21))) {
                            if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_21)) break;
                            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                            __pyx_t_22 = PyList_GET_ITEM(__pyx_t_21, __pyx_t_16); __Pyx_INCREF(__pyx_t_22); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1082, __pyx_L61_error)
                            #else
                            __pyx_t_22 = PySequence_ITEM(__pyx_t_21, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1082, __pyx_L61_error)
                            __Pyx_GOTREF(__pyx_t_22);
                            #endif
                          } else {
                            if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_21)) break;
                            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                            __pyx_t_22 = PyTuple_GET_ITEM(__pyx_t_21, __pyx_t_16); __Pyx_INCREF(__pyx_t_22); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1082, __pyx_L61_error)
                            #else
                            __pyx_t_22 = PySequence_ITEM(__pyx_t_21, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1082, __pyx_L61_error)
                            __Pyx_GOTREF(__pyx_t_22);
                            #endif
                          }
                        } else {
                          __pyx_t_22 = __pyx_t_20(__pyx_t_21);
                          if (unlikely(!__pyx_t_22)) {
                            PyObject* exc_type = PyErr_Occurred();
                            if (exc_type) {
                              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                              else __PYX_ERR(0, 1082, __pyx_L61_error)
                            }
                            break;
                          }
                          __Pyx_GOTREF(__pyx_t_22);
                        }
                        __Pyx_XDECREF_SET(__pyx_9genexpr57__pyx_v_curve, __pyx_t_22);
                        __pyx_t_22 = 0;
                        __Pyx_INCREF(__pyx_t_6);
                        __Pyx_XDECREF_SET(__pyx_9genexpr57__pyx_v_i, __pyx_t_6);
                        __pyx_t_22 = __Pyx_PyInt_AddObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1082, __pyx_L61_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        __Pyx_DECREF(__pyx_t_6);
                        __pyx_t_6 = __pyx_t_22;
                        __pyx_t_22 = 0;

                        /* "gnocis/models.pyx":1081
 * 					curvebyxmin = {
 * 						x: [
 * 							min(curvebyx[i][x])             # <<<<<<<<<<<<<<
 * 							for i, curve in enumerate(curves)
 * 						]
 */
                        __pyx_t_22 = __Pyx_PyObject_GetItem(__pyx_v_curvebyx, __pyx_9genexpr57__pyx_v_i); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1081, __pyx_L61_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        __pyx_t_25 = __Pyx_PyObject_GetItem(__pyx_t_22, __pyx_9genexpr56__pyx_v_x); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1081, __pyx_L61_error)
                        __Pyx_GOTREF(__pyx_t_25);
                        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                        __pyx_t_22 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_25); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1081, __pyx_L61_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                        if (unlikely(__Pyx_ListComp_Append(__pyx_t_12, (PyObject*)__pyx_t_22))) __PYX_ERR(0, 1080, __pyx_L61_error)
                        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

                        /* "gnocis/models.pyx":1082
 * 						x: [
 * 							min(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)             # <<<<<<<<<<<<<<
 * 						]
 * 						for x in xs
 */
                      }
                      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                      __Pyx_XDECREF(__pyx_9genexpr57__pyx_v_curve); __pyx_9genexpr57__pyx_v_curve = 0;
                      __Pyx_XDECREF(__pyx_9genexpr57__pyx_v_i); __pyx_9genexpr57__pyx_v_i = 0;
                      goto __pyx_L64_exit_scope;
                      __pyx_L61_error:;
                      __Pyx_XDECREF(__pyx_9genexpr57__pyx_v_curve); __pyx_9genexpr57__pyx_v_curve = 0;
                      __Pyx_XDECREF(__pyx_9genexpr57__pyx_v_i); __pyx_9genexpr57__pyx_v_i = 0;
                      goto __pyx_L56_error;
                      __pyx_L64_exit_scope:;
                    } /* exit inner scope */
                    if (unlikely(PyDict_SetItem(__pyx_t_8, (PyObject*)__pyx_9genexpr56__pyx_v_x, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 1080, __pyx_L56_error)
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

                    /* "gnocis/models.pyx":1084
 * 							for i, curve in enumerate(curves)
 * 						]
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 					}
 * 					meanCurve = [
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr56__pyx_v_x); __pyx_9genexpr56__pyx_v_x = 0;
                  goto __pyx_L65_exit_scope;
                  __pyx_L56_error:;
                  __Pyx_XDECREF(__pyx_9genexpr56__pyx_v_x); __pyx_9genexpr56__pyx_v_x = 0;
                  goto __pyx_L13_error;
                  __pyx_L65_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_curvebyxmin, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1086
 * 						for x in xs
 * 					}
 * 					meanCurve = [             # <<<<<<<<<<<<<<
 * 						pt
 * 						for x in xs
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1086, __pyx_L68_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1088
 * 					meanCurve = [
 * 						pt
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 						for pt in [
 * 							point2D(x, mean(curvebyxmax[x])),
 */
                  if (unlikely(__pyx_v_xs == Py_None)) {
                    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                    __PYX_ERR(0, 1088, __pyx_L68_error)
                  }
                  __pyx_t_4 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_4); __pyx_t_14 = 0;
                  for (;;) {
                    if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_4)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_12 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_12); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1088, __pyx_L68_error)
                    #else
                    __pyx_t_12 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1088, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr58__pyx_v_x, __pyx_t_12);
                    __pyx_t_12 = 0;

                    /* "gnocis/models.pyx":1090
 * 						for x in xs
 * 						for pt in [
 * 							point2D(x, mean(curvebyxmax[x])),             # <<<<<<<<<<<<<<
 * 							point2D(x, mean(curvebyxmin[x]))
 * 						]
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_mean); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1090, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __pyx_t_21 = __Pyx_PyDict_GetItem(__pyx_v_curvebyxmax, __pyx_9genexpr58__pyx_v_x); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1090, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __pyx_t_22 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                      __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_6);
                      if (likely(__pyx_t_22)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                        __Pyx_INCREF(__pyx_t_22);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_6, function);
                      }
                    }
                    __pyx_t_12 = (__pyx_t_22) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_22, __pyx_t_21) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_21);
                    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1090, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1090, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_INCREF(__pyx_9genexpr58__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_9genexpr58__pyx_v_x);
                    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_9genexpr58__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_t_12);
                    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_12);
                    __pyx_t_12 = 0;
                    __pyx_t_12 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_point2D), __pyx_t_6, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1090, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1091
 * 						for pt in [
 * 							point2D(x, mean(curvebyxmax[x])),
 * 							point2D(x, mean(curvebyxmin[x]))             # <<<<<<<<<<<<<<
 * 						]
 * 					]
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_mean); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1091, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __pyx_t_22 = __Pyx_PyDict_GetItem(__pyx_v_curvebyxmin, __pyx_9genexpr58__pyx_v_x); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1091, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    __pyx_t_25 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
                      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_21);
                      if (likely(__pyx_t_25)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
                        __Pyx_INCREF(__pyx_t_25);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_21, function);
                      }
                    }
                    __pyx_t_6 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_21, __pyx_t_25, __pyx_t_22) : __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_t_22);
                    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
                    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1091, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                    __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1091, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __Pyx_INCREF(__pyx_9genexpr58__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_9genexpr58__pyx_v_x);
                    PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_9genexpr58__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_t_6);
                    PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_6);
                    __pyx_t_6 = 0;
                    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_point2D), __pyx_t_21, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1091, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

                    /* "gnocis/models.pyx":1089
 * 						pt
 * 						for x in xs
 * 						for pt in [             # <<<<<<<<<<<<<<
 * 							point2D(x, mean(curvebyxmax[x])),
 * 							point2D(x, mean(curvebyxmin[x]))
 */
                    __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1089, __pyx_L68_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __Pyx_GIVEREF(__pyx_t_12);
                    PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_12);
                    __Pyx_GIVEREF(__pyx_t_6);
                    PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_6);
                    __pyx_t_12 = 0;
                    __pyx_t_6 = 0;
                    __pyx_t_6 = __pyx_t_21; __Pyx_INCREF(__pyx_t_6); __pyx_t_16 = 0;
                    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                    for (;;) {
                      if (__pyx_t_16 >= 2) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_21 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_16); __Pyx_INCREF(__pyx_t_21); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1089, __pyx_L68_error)
                      #else
                      __pyx_t_21 = PySequence_ITEM(__pyx_t_6, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1089, __pyx_L68_error)
                      __Pyx_GOTREF(__pyx_t_21);
                      #endif
                      if (!(likely(((__pyx_t_21) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_21, __pyx_ptype_6gnocis_10validation_point2D))))) __PYX_ERR(0, 1089, __pyx_L68_error)
                      __Pyx_XDECREF_SET(__pyx_9genexpr58__pyx_v_pt, ((struct __pyx_obj_6gnocis_10validation_point2D *)__pyx_t_21));
                      __pyx_t_21 = 0;

                      /* "gnocis/models.pyx":1087
 * 					}
 * 					meanCurve = [
 * 						pt             # <<<<<<<<<<<<<<
 * 						for x in xs
 * 						for pt in [
 */
                      if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_9genexpr58__pyx_v_pt))) __PYX_ERR(0, 1086, __pyx_L68_error)

                      /* "gnocis/models.pyx":1089
 * 						pt
 * 						for x in xs
 * 						for pt in [             # <<<<<<<<<<<<<<
 * 							point2D(x, mean(curvebyxmax[x])),
 * 							point2D(x, mean(curvebyxmin[x]))
 */
                    }
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1088
 * 					meanCurve = [
 * 						pt
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 						for pt in [
 * 							point2D(x, mean(curvebyxmax[x])),
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF((PyObject *)__pyx_9genexpr58__pyx_v_pt); __pyx_9genexpr58__pyx_v_pt = 0;
                  __Pyx_XDECREF(__pyx_9genexpr58__pyx_v_x); __pyx_9genexpr58__pyx_v_x = 0;
                  goto __pyx_L73_exit_scope;
                  __pyx_L68_error:;
                  __Pyx_XDECREF((PyObject *)__pyx_9genexpr58__pyx_v_pt); __pyx_9genexpr58__pyx_v_pt = 0;
                  __Pyx_XDECREF(__pyx_9genexpr58__pyx_v_x); __pyx_9genexpr58__pyx_v_x = 0;
                  goto __pyx_L13_error;
                  __pyx_L73_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_meanCurve, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1094
 * 						]
 * 					]
 * 					CICurve = [             # <<<<<<<<<<<<<<
 * 						pt
 * 						for x in xs
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1094, __pyx_L76_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1096
 * 					CICurve = [
 * 						pt
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 						for pt in [
 * 							point2D(x, CI(curvebyxmax[x])),
 */
                  if (unlikely(__pyx_v_xs == Py_None)) {
                    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                    __PYX_ERR(0, 1096, __pyx_L76_error)
                  }
                  __pyx_t_4 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_4); __pyx_t_14 = 0;
                  for (;;) {
                    if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_4)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_14); __Pyx_INCREF(__pyx_t_6); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1096, __pyx_L76_error)
                    #else
                    __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1096, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr59__pyx_v_x, __pyx_t_6);
                    __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1098
 * 						for x in xs
 * 						for pt in [
 * 							point2D(x, CI(curvebyxmax[x])),             # <<<<<<<<<<<<<<
 * 							point2D(x, CI(curvebyxmin[x]))
 * 						]
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_CI); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1098, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __pyx_t_12 = __Pyx_PyDict_GetItem(__pyx_v_curvebyxmax, __pyx_9genexpr59__pyx_v_x); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1098, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_22 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
                      __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_21);
                      if (likely(__pyx_t_22)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
                        __Pyx_INCREF(__pyx_t_22);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_21, function);
                      }
                    }
                    __pyx_t_6 = (__pyx_t_22) ? __Pyx_PyObject_Call2Args(__pyx_t_21, __pyx_t_22, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_t_12);
                    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1098, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                    __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1098, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __Pyx_INCREF(__pyx_9genexpr59__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_9genexpr59__pyx_v_x);
                    PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_9genexpr59__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_t_6);
                    PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_6);
                    __pyx_t_6 = 0;
                    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_point2D), __pyx_t_21, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1098, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

                    /* "gnocis/models.pyx":1099
 * 						for pt in [
 * 							point2D(x, CI(curvebyxmax[x])),
 * 							point2D(x, CI(curvebyxmin[x]))             # <<<<<<<<<<<<<<
 * 						]
 * 					]
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_CI); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1099, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_22 = __Pyx_PyDict_GetItem(__pyx_v_curvebyxmin, __pyx_9genexpr59__pyx_v_x); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1099, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    __pyx_t_25 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
                      __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_12);
                      if (likely(__pyx_t_25)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                        __Pyx_INCREF(__pyx_t_25);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_12, function);
                      }
                    }
                    __pyx_t_21 = (__pyx_t_25) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_25, __pyx_t_22) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_22);
                    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
                    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                    if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1099, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1099, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __Pyx_INCREF(__pyx_9genexpr59__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_9genexpr59__pyx_v_x);
                    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_9genexpr59__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_t_21);
                    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_21);
                    __pyx_t_21 = 0;
                    __pyx_t_21 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_point2D), __pyx_t_12, NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1099, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_21);
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

                    /* "gnocis/models.pyx":1097
 * 						pt
 * 						for x in xs
 * 						for pt in [             # <<<<<<<<<<<<<<
 * 							point2D(x, CI(curvebyxmax[x])),
 * 							point2D(x, CI(curvebyxmin[x]))
 */
                    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1097, __pyx_L76_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __Pyx_GIVEREF(__pyx_t_6);
                    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_6);
                    __Pyx_GIVEREF(__pyx_t_21);
                    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_21);
                    __pyx_t_6 = 0;
                    __pyx_t_21 = 0;
                    __pyx_t_21 = __pyx_t_12; __Pyx_INCREF(__pyx_t_21); __pyx_t_16 = 0;
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    for (;;) {
                      if (__pyx_t_16 >= 2) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_21, __pyx_t_16); __Pyx_INCREF(__pyx_t_12); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1097, __pyx_L76_error)
                      #else
                      __pyx_t_12 = PySequence_ITEM(__pyx_t_21, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1097, __pyx_L76_error)
                      __Pyx_GOTREF(__pyx_t_12);
                      #endif
                      if (!(likely(((__pyx_t_12) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_12, __pyx_ptype_6gnocis_10validation_point2D))))) __PYX_ERR(0, 1097, __pyx_L76_error)
                      __Pyx_XDECREF_SET(__pyx_9genexpr59__pyx_v_pt, ((struct __pyx_obj_6gnocis_10validation_point2D *)__pyx_t_12));
                      __pyx_t_12 = 0;

                      /* "gnocis/models.pyx":1095
 * 					]
 * 					CICurve = [
 * 						pt             # <<<<<<<<<<<<<<
 * 						for x in xs
 * 						for pt in [
 */
                      if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_9genexpr59__pyx_v_pt))) __PYX_ERR(0, 1094, __pyx_L76_error)

                      /* "gnocis/models.pyx":1097
 * 						pt
 * 						for x in xs
 * 						for pt in [             # <<<<<<<<<<<<<<
 * 							point2D(x, CI(curvebyxmax[x])),
 * 							point2D(x, CI(curvebyxmin[x]))
 */
                    }
                    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

                    /* "gnocis/models.pyx":1096
 * 					CICurve = [
 * 						pt
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 						for pt in [
 * 							point2D(x, CI(curvebyxmax[x])),
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF((PyObject *)__pyx_9genexpr59__pyx_v_pt); __pyx_9genexpr59__pyx_v_pt = 0;
                  __Pyx_XDECREF(__pyx_9genexpr59__pyx_v_x); __pyx_9genexpr59__pyx_v_x = 0;
                  goto __pyx_L81_exit_scope;
                  __pyx_L76_error:;
                  __Pyx_XDECREF((PyObject *)__pyx_9genexpr59__pyx_v_pt); __pyx_9genexpr59__pyx_v_pt = 0;
                  __Pyx_XDECREF(__pyx_9genexpr59__pyx_v_x); __pyx_9genexpr59__pyx_v_x = 0;
                  goto __pyx_L13_error;
                  __pyx_L81_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_CICurve, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1102
 * 						]
 * 					]
 * 					plt.fill_between(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 */
                __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_fill_between); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1102, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1103
 * 					]
 * 					plt.fill_between(
 * 						[ pt.x for pt in meanCurve ],             # <<<<<<<<<<<<<<
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],
 */
                  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1103, __pyx_L84_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __pyx_t_21 = __pyx_v_meanCurve; __Pyx_INCREF(__pyx_t_21); __pyx_t_14 = 0;
                  for (;;) {
                    if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_21)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_12 = PyList_GET_ITEM(__pyx_t_21, __pyx_t_14); __Pyx_INCREF(__pyx_t_12); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1103, __pyx_L84_error)
                    #else
                    __pyx_t_12 = PySequence_ITEM(__pyx_t_21, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1103, __pyx_L84_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr60__pyx_v_pt, __pyx_t_12);
                    __pyx_t_12 = 0;
                    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr60__pyx_v_pt, __pyx_n_s_x); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1103, __pyx_L84_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 1103, __pyx_L84_error)
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr60__pyx_v_pt); __pyx_9genexpr60__pyx_v_pt = 0;
                  goto __pyx_L87_exit_scope;
                  __pyx_L84_error:;
                  __Pyx_XDECREF(__pyx_9genexpr60__pyx_v_pt); __pyx_9genexpr60__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L87_exit_scope:;
                } /* exit inner scope */
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1104
 * 					plt.fill_between(
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],             # <<<<<<<<<<<<<<
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						alpha=.3)
 */
                  __pyx_t_21 = PyList_New(0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1104, __pyx_L90_error)
                  __Pyx_GOTREF(__pyx_t_21);
                  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1104, __pyx_L90_error)
                  __Pyx_GOTREF(__pyx_t_12);
                  __Pyx_INCREF(__pyx_v_meanCurve);
                  __Pyx_GIVEREF(__pyx_v_meanCurve);
                  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_meanCurve);
                  __Pyx_INCREF(__pyx_v_CICurve);
                  __Pyx_GIVEREF(__pyx_v_CICurve);
                  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_CICurve);
                  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_12, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1104, __pyx_L90_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                  if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
                    __pyx_t_12 = __pyx_t_6; __Pyx_INCREF(__pyx_t_12); __pyx_t_14 = 0;
                    __pyx_t_20 = NULL;
                  } else {
                    __pyx_t_14 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1104, __pyx_L90_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_20 = Py_TYPE(__pyx_t_12)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1104, __pyx_L90_error)
                  }
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  for (;;) {
                    if (likely(!__pyx_t_20)) {
                      if (likely(PyList_CheckExact(__pyx_t_12))) {
                        if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_12)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_12, __pyx_t_14); __Pyx_INCREF(__pyx_t_6); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1104, __pyx_L90_error)
                        #else
                        __pyx_t_6 = PySequence_ITEM(__pyx_t_12, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1104, __pyx_L90_error)
                        __Pyx_GOTREF(__pyx_t_6);
                        #endif
                      } else {
                        if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_12)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_14); __Pyx_INCREF(__pyx_t_6); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1104, __pyx_L90_error)
                        #else
                        __pyx_t_6 = PySequence_ITEM(__pyx_t_12, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1104, __pyx_L90_error)
                        __Pyx_GOTREF(__pyx_t_6);
                        #endif
                      }
                    } else {
                      __pyx_t_6 = __pyx_t_20(__pyx_t_12);
                      if (unlikely(!__pyx_t_6)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 1104, __pyx_L90_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_6);
                    }
                    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
                      PyObject* sequence = __pyx_t_6;
                      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                      if (unlikely(size != 2)) {
                        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                        __PYX_ERR(0, 1104, __pyx_L90_error)
                      }
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      if (likely(PyTuple_CheckExact(sequence))) {
                        __pyx_t_22 = PyTuple_GET_ITEM(sequence, 0); 
                        __pyx_t_25 = PyTuple_GET_ITEM(sequence, 1); 
                      } else {
                        __pyx_t_22 = PyList_GET_ITEM(sequence, 0); 
                        __pyx_t_25 = PyList_GET_ITEM(sequence, 1); 
                      }
                      __Pyx_INCREF(__pyx_t_22);
                      __Pyx_INCREF(__pyx_t_25);
                      #else
                      __pyx_t_22 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1104, __pyx_L90_error)
                      __Pyx_GOTREF(__pyx_t_22);
                      __pyx_t_25 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1104, __pyx_L90_error)
                      __Pyx_GOTREF(__pyx_t_25);
                      #endif
                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                    } else {
                      Py_ssize_t index = -1;
                      __pyx_t_24 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1104, __pyx_L90_error)
                      __Pyx_GOTREF(__pyx_t_24);
                      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                      __pyx_t_13 = Py_TYPE(__pyx_t_24)->tp_iternext;
                      index = 0; __pyx_t_22 = __pyx_t_13(__pyx_t_24); if (unlikely(!__pyx_t_22)) goto __pyx_L93_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_22);
                      index = 1; __pyx_t_25 = __pyx_t_13(__pyx_t_24); if (unlikely(!__pyx_t_25)) goto __pyx_L93_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_25);
                      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_24), 2) < 0) __PYX_ERR(0, 1104, __pyx_L90_error)
                      __pyx_t_13 = NULL;
                      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                      goto __pyx_L94_unpacking_done;
                      __pyx_L93_unpacking_failed:;
                      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                      __pyx_t_13 = NULL;
                      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                      __PYX_ERR(0, 1104, __pyx_L90_error)
                      __pyx_L94_unpacking_done:;
                    }
                    __Pyx_XDECREF_SET(__pyx_9genexpr61__pyx_v_pt, __pyx_t_22);
                    __pyx_t_22 = 0;
                    __Pyx_XDECREF_SET(__pyx_9genexpr61__pyx_v_ci, __pyx_t_25);
                    __pyx_t_25 = 0;
                    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr61__pyx_v_pt, __pyx_n_s_y); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1104, __pyx_L90_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr61__pyx_v_ci, __pyx_n_s_y); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1104, __pyx_L90_error)
                    __Pyx_GOTREF(__pyx_t_25);
                    __pyx_t_22 = PyNumber_Subtract(__pyx_t_6, __pyx_t_25); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1104, __pyx_L90_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_21, (PyObject*)__pyx_t_22))) __PYX_ERR(0, 1104, __pyx_L90_error)
                    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr61__pyx_v_ci); __pyx_9genexpr61__pyx_v_ci = 0;
                  __Pyx_XDECREF(__pyx_9genexpr61__pyx_v_pt); __pyx_9genexpr61__pyx_v_pt = 0;
                  goto __pyx_L95_exit_scope;
                  __pyx_L90_error:;
                  __Pyx_XDECREF(__pyx_9genexpr61__pyx_v_ci); __pyx_9genexpr61__pyx_v_ci = 0;
                  __Pyx_XDECREF(__pyx_9genexpr61__pyx_v_pt); __pyx_9genexpr61__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L95_exit_scope:;
                } /* exit inner scope */
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1105
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],             # <<<<<<<<<<<<<<
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 */
                  __pyx_t_12 = PyList_New(0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1105, __pyx_L98_error)
                  __Pyx_GOTREF(__pyx_t_12);
                  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1105, __pyx_L98_error)
                  __Pyx_GOTREF(__pyx_t_22);
                  __Pyx_INCREF(__pyx_v_meanCurve);
                  __Pyx_GIVEREF(__pyx_v_meanCurve);
                  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_v_meanCurve);
                  __Pyx_INCREF(__pyx_v_CICurve);
                  __Pyx_GIVEREF(__pyx_v_CICurve);
                  PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_v_CICurve);
                  __pyx_t_25 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_22, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1105, __pyx_L98_error)
                  __Pyx_GOTREF(__pyx_t_25);
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                  if (likely(PyList_CheckExact(__pyx_t_25)) || PyTuple_CheckExact(__pyx_t_25)) {
                    __pyx_t_22 = __pyx_t_25; __Pyx_INCREF(__pyx_t_22); __pyx_t_14 = 0;
                    __pyx_t_20 = NULL;
                  } else {
                    __pyx_t_14 = -1; __pyx_t_22 = PyObject_GetIter(__pyx_t_25); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1105, __pyx_L98_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    __pyx_t_20 = Py_TYPE(__pyx_t_22)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1105, __pyx_L98_error)
                  }
                  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                  for (;;) {
                    if (likely(!__pyx_t_20)) {
                      if (likely(PyList_CheckExact(__pyx_t_22))) {
                        if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_22)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_25 = PyList_GET_ITEM(__pyx_t_22, __pyx_t_14); __Pyx_INCREF(__pyx_t_25); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1105, __pyx_L98_error)
                        #else
                        __pyx_t_25 = PySequence_ITEM(__pyx_t_22, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1105, __pyx_L98_error)
                        __Pyx_GOTREF(__pyx_t_25);
                        #endif
                      } else {
                        if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_22)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_25 = PyTuple_GET_ITEM(__pyx_t_22, __pyx_t_14); __Pyx_INCREF(__pyx_t_25); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1105, __pyx_L98_error)
                        #else
                        __pyx_t_25 = PySequence_ITEM(__pyx_t_22, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1105, __pyx_L98_error)
                        __Pyx_GOTREF(__pyx_t_25);
                        #endif
                      }
                    } else {
                      __pyx_t_25 = __pyx_t_20(__pyx_t_22);
                      if (unlikely(!__pyx_t_25)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 1105, __pyx_L98_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_25);
                    }
                    if ((likely(PyTuple_CheckExact(__pyx_t_25))) || (PyList_CheckExact(__pyx_t_25))) {
                      PyObject* sequence = __pyx_t_25;
                      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                      if (unlikely(size != 2)) {
                        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                        __PYX_ERR(0, 1105, __pyx_L98_error)
                      }
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      if (likely(PyTuple_CheckExact(sequence))) {
                        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
                        __pyx_t_24 = PyTuple_GET_ITEM(sequence, 1); 
                      } else {
                        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
                        __pyx_t_24 = PyList_GET_ITEM(sequence, 1); 
                      }
                      __Pyx_INCREF(__pyx_t_6);
                      __Pyx_INCREF(__pyx_t_24);
                      #else
                      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1105, __pyx_L98_error)
                      __Pyx_GOTREF(__pyx_t_6);
                      __pyx_t_24 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1105, __pyx_L98_error)
                      __Pyx_GOTREF(__pyx_t_24);
                      #endif
                      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                    } else {
                      Py_ssize_t index = -1;
                      __pyx_t_27 = PyObject_GetIter(__pyx_t_25); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1105, __pyx_L98_error)
                      __Pyx_GOTREF(__pyx_t_27);
                      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                      __pyx_t_13 = Py_TYPE(__pyx_t_27)->tp_iternext;
                      index = 0; __pyx_t_6 = __pyx_t_13(__pyx_t_27); if (unlikely(!__pyx_t_6)) goto __pyx_L101_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_6);
                      index = 1; __pyx_t_24 = __pyx_t_13(__pyx_t_27); if (unlikely(!__pyx_t_24)) goto __pyx_L101_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_24);
                      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_27), 2) < 0) __PYX_ERR(0, 1105, __pyx_L98_error)
                      __pyx_t_13 = NULL;
                      __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
                      goto __pyx_L102_unpacking_done;
                      __pyx_L101_unpacking_failed:;
                      __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
                      __pyx_t_13 = NULL;
                      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                      __PYX_ERR(0, 1105, __pyx_L98_error)
                      __pyx_L102_unpacking_done:;
                    }
                    __Pyx_XDECREF_SET(__pyx_9genexpr62__pyx_v_pt, __pyx_t_6);
                    __pyx_t_6 = 0;
                    __Pyx_XDECREF_SET(__pyx_9genexpr62__pyx_v_ci, __pyx_t_24);
                    __pyx_t_24 = 0;
                    __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr62__pyx_v_pt, __pyx_n_s_y); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 1105, __pyx_L98_error)
                    __Pyx_GOTREF(__pyx_t_25);
                    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr62__pyx_v_ci, __pyx_n_s_y); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1105, __pyx_L98_error)
                    __Pyx_GOTREF(__pyx_t_24);
                    __pyx_t_6 = PyNumber_Add(__pyx_t_25, __pyx_t_24); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1105, __pyx_L98_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
                    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_12, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1105, __pyx_L98_error)
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr62__pyx_v_ci); __pyx_9genexpr62__pyx_v_ci = 0;
                  __Pyx_XDECREF(__pyx_9genexpr62__pyx_v_pt); __pyx_9genexpr62__pyx_v_pt = 0;
                  goto __pyx_L103_exit_scope;
                  __pyx_L98_error:;
                  __Pyx_XDECREF(__pyx_9genexpr62__pyx_v_ci); __pyx_9genexpr62__pyx_v_ci = 0;
                  __Pyx_XDECREF(__pyx_9genexpr62__pyx_v_pt); __pyx_9genexpr62__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L103_exit_scope:;
                } /* exit inner scope */

                /* "gnocis/models.pyx":1102
 * 						]
 * 					]
 * 					plt.fill_between(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 */
                __pyx_t_22 = PyTuple_New(3); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1102, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_21);
                PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_21);
                __Pyx_GIVEREF(__pyx_t_12);
                PyTuple_SET_ITEM(__pyx_t_22, 2, __pyx_t_12);
                __pyx_t_4 = 0;
                __pyx_t_21 = 0;
                __pyx_t_12 = 0;

                /* "gnocis/models.pyx":1106
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						alpha=.3)             # <<<<<<<<<<<<<<
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(
 */
                __pyx_t_12 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1106, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_12);
                if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_alpha, __pyx_float__3) < 0) __PYX_ERR(0, 1106, __pyx_L13_error)

                /* "gnocis/models.pyx":1102
 * 						]
 * 					]
 * 					plt.fill_between(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 */
                __pyx_t_21 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_22, __pyx_t_12); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1102, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_21);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

                /* "gnocis/models.pyx":1107
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]             # <<<<<<<<<<<<<<
 * 					plt.plot(
 * 						[ pt.x for pt in meanCurve ],
 */
                { /* enter inner scope */
                  __pyx_t_21 = PyList_New(0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1107, __pyx_L106_error)
                  __Pyx_GOTREF(__pyx_t_21);
                  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_curves)) {
                    __pyx_t_12 = __pyx_cur_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_12); __pyx_t_14 = 0;
                    __pyx_t_20 = NULL;
                  } else {
                    __pyx_t_14 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_curves); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1107, __pyx_L106_error)
                    __Pyx_GOTREF(__pyx_t_12);
                    __pyx_t_20 = Py_TYPE(__pyx_t_12)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1107, __pyx_L106_error)
                  }
                  for (;;) {
                    if (likely(!__pyx_t_20)) {
                      if (likely(PyList_CheckExact(__pyx_t_12))) {
                        if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_12)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_22 = PyList_GET_ITEM(__pyx_t_12, __pyx_t_14); __Pyx_INCREF(__pyx_t_22); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1107, __pyx_L106_error)
                        #else
                        __pyx_t_22 = PySequence_ITEM(__pyx_t_12, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1107, __pyx_L106_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        #endif
                      } else {
                        if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_12)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_22 = PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_14); __Pyx_INCREF(__pyx_t_22); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1107, __pyx_L106_error)
                        #else
                        __pyx_t_22 = PySequence_ITEM(__pyx_t_12, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1107, __pyx_L106_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        #endif
                      }
                    } else {
                      __pyx_t_22 = __pyx_t_20(__pyx_t_12);
                      if (unlikely(!__pyx_t_22)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 1107, __pyx_L106_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_22);
                    }
                    __Pyx_XDECREF_SET(__pyx_9genexpr63__pyx_v_curve, __pyx_t_22);
                    __pyx_t_22 = 0;
                    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1107, __pyx_L106_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __pyx_t_4 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
                      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
                      if (likely(__pyx_t_4)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                        __Pyx_INCREF(__pyx_t_4);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_8, function);
                      }
                    }
                    __pyx_t_22 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_9genexpr63__pyx_v_curve) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_9genexpr63__pyx_v_curve);
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1107, __pyx_L106_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __pyx_t_8 = PyNumber_Multiply(__pyx_t_22, __pyx_float_100_); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1107, __pyx_L106_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_21, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 1107, __pyx_L106_error)
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr63__pyx_v_curve); __pyx_9genexpr63__pyx_v_curve = 0;
                  goto __pyx_L109_exit_scope;
                  __pyx_L106_error:;
                  __Pyx_XDECREF(__pyx_9genexpr63__pyx_v_curve); __pyx_9genexpr63__pyx_v_curve = 0;
                  goto __pyx_L13_error;
                  __pyx_L109_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_AUCs, ((PyObject*)__pyx_t_21));
                __pyx_t_21 = 0;

                /* "gnocis/models.pyx":1108
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],
 */
                __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_plot); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1108, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_21);
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1109
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(
 * 						[ pt.x for pt in meanCurve ],             # <<<<<<<<<<<<<<
 * 						[ pt.y for pt in meanCurve ],
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))
 */
                  __pyx_t_12 = PyList_New(0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1109, __pyx_L112_error)
                  __Pyx_GOTREF(__pyx_t_12);
                  __pyx_t_8 = __pyx_v_meanCurve; __Pyx_INCREF(__pyx_t_8); __pyx_t_14 = 0;
                  for (;;) {
                    if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_8)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_22 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_22); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1109, __pyx_L112_error)
                    #else
                    __pyx_t_22 = PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1109, __pyx_L112_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr64__pyx_v_pt, __pyx_t_22);
                    __pyx_t_22 = 0;
                    __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr64__pyx_v_pt, __pyx_n_s_x); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1109, __pyx_L112_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_12, (PyObject*)__pyx_t_22))) __PYX_ERR(0, 1109, __pyx_L112_error)
                    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr64__pyx_v_pt); __pyx_9genexpr64__pyx_v_pt = 0;
                  goto __pyx_L115_exit_scope;
                  __pyx_L112_error:;
                  __Pyx_XDECREF(__pyx_9genexpr64__pyx_v_pt); __pyx_9genexpr64__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L115_exit_scope:;
                } /* exit inner scope */
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1110
 * 					plt.plot(
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],             # <<<<<<<<<<<<<<
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 */
                  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1110, __pyx_L118_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __pyx_t_22 = __pyx_v_meanCurve; __Pyx_INCREF(__pyx_t_22); __pyx_t_14 = 0;
                  for (;;) {
                    if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_22)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_22, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1110, __pyx_L118_error)
                    #else
                    __pyx_t_4 = PySequence_ITEM(__pyx_t_22, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1110, __pyx_L118_error)
                    __Pyx_GOTREF(__pyx_t_4);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr65__pyx_v_pt, __pyx_t_4);
                    __pyx_t_4 = 0;
                    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr65__pyx_v_pt, __pyx_n_s_y); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1110, __pyx_L118_error)
                    __Pyx_GOTREF(__pyx_t_4);
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1110, __pyx_L118_error)
                    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr65__pyx_v_pt); __pyx_9genexpr65__pyx_v_pt = 0;
                  goto __pyx_L121_exit_scope;
                  __pyx_L118_error:;
                  __Pyx_XDECREF(__pyx_9genexpr65__pyx_v_pt); __pyx_9genexpr65__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L121_exit_scope:;
                } /* exit inner scope */

                /* "gnocis/models.pyx":1108
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],
 */
                __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1108, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __Pyx_GIVEREF(__pyx_t_12);
                PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_12);
                __Pyx_GIVEREF(__pyx_t_8);
                PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_8);
                __pyx_t_12 = 0;
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1111
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))             # <<<<<<<<<<<<<<
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 */
                __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_12 = PyTuple_New(6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_12);
                __pyx_t_14 = 0;
                __pyx_t_28 = 127;
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_mdl, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_28 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_28) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_28;
                __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_6);
                __pyx_t_6 = 0;
                __Pyx_INCREF(__pyx_kp_u_AUC_2);
                __pyx_t_14 += 8;
                __Pyx_GIVEREF(__pyx_kp_u_AUC_2);
                PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_kp_u_AUC_2);
                __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_mean); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_24 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_24)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_24);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_24, __pyx_v_AUCs) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_AUCs);
                __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = __Pyx_PyObject_Format(__pyx_t_6, __pyx_kp_u_2f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_28 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_28) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_28;
                __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_t_4);
                __pyx_t_4 = 0;
                __Pyx_INCREF(__pyx_kp_u__17);
                __pyx_t_14 += 5;
                __Pyx_GIVEREF(__pyx_kp_u__17);
                PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_kp_u__17);
                __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_CI); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_24 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_24)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_24);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_4 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_24, __pyx_v_AUCs) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_AUCs);
                __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_6 = __Pyx_PyObject_Format(__pyx_t_4, __pyx_kp_u_2f); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_28 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_28) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_28;
                __pyx_t_14 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_12, 4, __pyx_t_6);
                __pyx_t_6 = 0;
                __Pyx_INCREF(__pyx_kp_u__7);
                __pyx_t_14 += 2;
                __Pyx_GIVEREF(__pyx_kp_u__7);
                PyTuple_SET_ITEM(__pyx_t_12, 5, __pyx_kp_u__7);
                __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_12, 6, __pyx_t_14, __pyx_t_28); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_label, __pyx_t_6) < 0) __PYX_ERR(0, 1111, __pyx_L13_error)
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "gnocis/models.pyx":1108
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],
 */
                __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_t_22, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1108, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "gnocis/models.pyx":1062
 * 					label = 'Expected at random')
 * 				# Curves per model
 * 				for mdl in self.models:             # <<<<<<<<<<<<<<
 * 					curves = self.PRC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 */
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "gnocis/models.pyx":1112
 * 						[ pt.y for pt in meanCurve ],
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)             # <<<<<<<<<<<<<<
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.xlabel('Recall', fontsize = fontsizeLabels)
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_xticks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1112, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1112, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_fontsize, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 1112, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_rotation, __pyx_int_0) < 0) __PYX_ERR(0, 1112, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1112, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1113
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)             # <<<<<<<<<<<<<<
 * 				plt.xlabel('Recall', fontsize = fontsizeLabels)
 * 				plt.ylabel('Precision', fontsize = fontsizeLabels)
 */
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_yticks); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1113, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1113, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_fontsize, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 1113, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_rotation, __pyx_int_0) < 0) __PYX_ERR(0, 1113, __pyx_L13_error)
              __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1113, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "gnocis/models.pyx":1114
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.xlabel('Recall', fontsize = fontsizeLabels)             # <<<<<<<<<<<<<<
 * 				plt.ylabel('Precision', fontsize = fontsizeLabels)
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_xlabel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1114, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1114, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_fontsize, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 1114, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__8, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1114, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1115
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.xlabel('Recall', fontsize = fontsizeLabels)
 * 				plt.ylabel('Precision', fontsize = fontsizeLabels)             # <<<<<<<<<<<<<<
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()
 */
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_ylabel); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1115, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1115, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_fontsize, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 1115, __pyx_L13_error)
              __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__9, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1115, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "gnocis/models.pyx":1116
 * 				plt.xlabel('Recall', fontsize = fontsizeLabels)
 * 				plt.ylabel('Precision', fontsize = fontsizeLabels)
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)             # <<<<<<<<<<<<<<
 * 				fig.tight_layout()
 * 				if outpath is None:
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_legend); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1116, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1116, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_bbox_to_anchor, __pyx_v_bboxAnchorTo) < 0) __PYX_ERR(0, 1116, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_loc, __pyx_v_legendLoc) < 0) __PYX_ERR(0, 1116, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_fontsize, __pyx_v_fontsizeLegend) < 0) __PYX_ERR(0, 1116, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_fancybox, Py_True) < 0) __PYX_ERR(0, 1116, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1116, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1117
 * 				plt.ylabel('Precision', fontsize = fontsizeLabels)
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()             # <<<<<<<<<<<<<<
 * 				if outpath is None:
 * 					bio = BytesIO()
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_tight_layout); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1117, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1117, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1118
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()
 * 				if outpath is None:             # <<<<<<<<<<<<<<
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 */
              __pyx_t_26 = (__pyx_v_outpath == Py_None);
              __pyx_t_29 = (__pyx_t_26 != 0);
              if (__pyx_t_29) {

                /* "gnocis/models.pyx":1119
 * 				fig.tight_layout()
 * 				if outpath is None:
 * 					bio = BytesIO()             # <<<<<<<<<<<<<<
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')
 */
                __Pyx_INCREF(__pyx_v_BytesIO);
                __pyx_t_6 = __pyx_v_BytesIO; __pyx_t_5 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_5)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_5);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1119, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_v_bio = __pyx_t_8;
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1120
 * 				if outpath is None:
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')             # <<<<<<<<<<<<<<
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 */
                __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_savefig); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1120, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1120, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_INCREF(__pyx_v_bio);
                __Pyx_GIVEREF(__pyx_v_bio);
                PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_bio);
                __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1120, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_format, __pyx_n_u_png) < 0) __PYX_ERR(0, 1120, __pyx_L13_error)
                __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1120, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

                /* "gnocis/models.pyx":1121
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')             # <<<<<<<<<<<<<<
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 */
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_close); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1121, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_22 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_u_all) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_u_all);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1121, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

                /* "gnocis/models.pyx":1122
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')             # <<<<<<<<<<<<<<
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					if returnHTML:
 */
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_base64, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1122, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_bio, __pyx_n_s_getvalue); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1122, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_21);
                __pyx_t_12 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_21))) {
                  __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_21);
                  if (likely(__pyx_t_12)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
                    __Pyx_INCREF(__pyx_t_12);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_21, function);
                  }
                }
                __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_21);
                __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1122, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                __pyx_t_21 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_21)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_21);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_5 = (__pyx_t_21) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_21, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_8);
                __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1122, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_decode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1122, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __pyx_t_5 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_5)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_5);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_22 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_utf_8);
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1122, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_v_encoded = __pyx_t_22;
                __pyx_t_22 = 0;

                /* "gnocis/models.pyx":1123
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded             # <<<<<<<<<<<<<<
 * 					if returnHTML:
 * 						return html
 */
                __pyx_t_22 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_img_src_data_image_png_base64_s, __pyx_v_encoded); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1123, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __pyx_v_html = ((PyObject*)__pyx_t_22);
                __pyx_t_22 = 0;

                /* "gnocis/models.pyx":1124
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					if returnHTML:             # <<<<<<<<<<<<<<
 * 						return html
 * 					display(HTML(html))
 */
                __pyx_t_29 = __Pyx_PyObject_IsTrue(__pyx_v_returnHTML); if (unlikely(__pyx_t_29 < 0)) __PYX_ERR(0, 1124, __pyx_L13_error)
                if (__pyx_t_29) {

                  /* "gnocis/models.pyx":1125
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					if returnHTML:
 * 						return html             # <<<<<<<<<<<<<<
 * 					display(HTML(html))
 * 				else:
 */
                  __Pyx_XDECREF(__pyx_r);
                  __Pyx_INCREF(__pyx_v_html);
                  __pyx_r = __pyx_v_html;
                  goto __pyx_L17_try_return;

                  /* "gnocis/models.pyx":1124
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					if returnHTML:             # <<<<<<<<<<<<<<
 * 						return html
 * 					display(HTML(html))
 */
                }

                /* "gnocis/models.pyx":1126
 * 					if returnHTML:
 * 						return html
 * 					display(HTML(html))             # <<<<<<<<<<<<<<
 * 				else:
 * 					fig.savefig(outpath)
 */
                __Pyx_INCREF(__pyx_v_HTML);
                __pyx_t_5 = __pyx_v_HTML; __pyx_t_8 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_8)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_8);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_v_html) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_html);
                __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1126, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_INCREF(__pyx_v_display);
                __pyx_t_5 = __pyx_v_display; __pyx_t_8 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_8)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_8);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_22 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
                __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1126, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

                /* "gnocis/models.pyx":1118
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()
 * 				if outpath is None:             # <<<<<<<<<<<<<<
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 */
                goto __pyx_L122;
              }

              /* "gnocis/models.pyx":1128
 * 					display(HTML(html))
 * 				else:
 * 					fig.savefig(outpath)             # <<<<<<<<<<<<<<
 * 					plt.close('all')
 * 		except ImportError as err:
 */
              /*else*/ {
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_savefig); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1128, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_22 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_outpath) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_outpath);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1128, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

                /* "gnocis/models.pyx":1129
 * 				else:
 * 					fig.savefig(outpath)
 * 					plt.close('all')             # <<<<<<<<<<<<<<
 * 		except ImportError as err:
 * 			raise err
 */
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_close); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1129, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_22 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_u_all) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_u_all);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1129, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_22);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              }
              __pyx_L122:;

              /* "gnocis/models.pyx":1050
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 				fig = plt.figure(figsize = figsize)
 * 				# Expected random generalization
 */
            }
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            goto __pyx_L18_try_end;
            __pyx_L13_error:;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
            __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
            __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
            __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("gnocis.models.crossvalidation.plotPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_22, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 1050, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_22);
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GOTREF(__pyx_t_6);
              __pyx_t_8 = PyTuple_Pack(3, __pyx_t_22, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1050, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_30 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1050, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_30);
              __pyx_t_29 = __Pyx_PyObject_IsTrue(__pyx_t_30);
              __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
              if (__pyx_t_29 < 0) __PYX_ERR(0, 1050, __pyx_L15_except_error)
              __pyx_t_26 = ((!(__pyx_t_29 != 0)) != 0);
              if (__pyx_t_26) {
                __Pyx_GIVEREF(__pyx_t_22);
                __Pyx_GIVEREF(__pyx_t_5);
                __Pyx_XGIVEREF(__pyx_t_6);
                __Pyx_ErrRestoreWithState(__pyx_t_22, __pyx_t_5, __pyx_t_6);
                __pyx_t_22 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0; 
                __PYX_ERR(0, 1050, __pyx_L15_except_error)
              }
              __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              goto __pyx_L14_exception_handled;
            }
            __pyx_L15_except_error:;
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
            goto __pyx_L3_error;
            __pyx_L17_try_return:;
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
            goto __pyx_L10_return;
            __pyx_L14_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
            __pyx_L18_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_7) {
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__10, NULL);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1050, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            }
            goto __pyx_L12;
          }
          __pyx_L10_return: {
            __pyx_t_11 = __pyx_r;
            __pyx_r = 0;
            if (__pyx_t_7) {
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__10, NULL);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1050, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
            __pyx_r = __pyx_t_11;
            __pyx_t_11 = 0;
            goto __pyx_L7_try_return;
          }
          __pyx_L12:;
        }
        goto __pyx_L127;
        __pyx_L9_error:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L3_error;
        __pyx_L127:;
      }

      /* "gnocis/models.pyx":1045
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "gnocis/models.pyx":1130
 * 					fig.savefig(outpath)
 * 					plt.close('all')
 * 		except ImportError as err:             # <<<<<<<<<<<<<<
 * 			raise err
 * 
 */
    __pyx_t_31 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_31) {
      __Pyx_AddTraceback("gnocis.models.crossvalidation.plotPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_22) < 0) __PYX_ERR(0, 1130, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_22);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_err = __pyx_t_5;
      /*try:*/ {

        /* "gnocis/models.pyx":1131
 * 					plt.close('all')
 * 		except ImportError as err:
 * 			raise err             # <<<<<<<<<<<<<<
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 */
        __Pyx_Raise(__pyx_v_err, 0, 0, 0);
        __PYX_ERR(0, 1131, __pyx_L133_error)
      }

      /* "gnocis/models.pyx":1130
 * 					fig.savefig(outpath)
 * 					plt.close('all')
 * 		except ImportError as err:             # <<<<<<<<<<<<<<
 * 			raise err
 * 
 */
      /*finally:*/ {
        __pyx_L133_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_7 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_30 = 0; __pyx_t_34 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
          __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
          __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
          __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_9, &__pyx_t_30, &__pyx_t_34);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_11, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_11, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_30);
          __Pyx_XGOTREF(__pyx_t_34);
          __pyx_t_31 = __pyx_lineno; __pyx_t_32 = __pyx_clineno; __pyx_t_33 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err);
            __pyx_v_err = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_30);
            __Pyx_XGIVEREF(__pyx_t_34);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_30, __pyx_t_34);
          }
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ErrRestore(__pyx_t_7, __pyx_t_11, __pyx_t_10);
          __pyx_t_7 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_30 = 0; __pyx_t_34 = 0;
          __pyx_lineno = __pyx_t_31; __pyx_clineno = __pyx_t_32; __pyx_filename = __pyx_t_33;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "gnocis/models.pyx":1045
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "gnocis/models.pyx":1044
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.plotPRC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_plt);
  __Pyx_XDECREF(__pyx_v_base64);
  __Pyx_XDECREF(__pyx_v_BytesIO);
  __Pyx_XDECREF(__pyx_v_display);
  __Pyx_XDECREF(__pyx_v_HTML);
  __Pyx_XDECREF(__pyx_v_fig);
  __Pyx_XDECREF(__pyx_v_positives);
  __Pyx_XDECREF(__pyx_v_negatives);
  __Pyx_XDECREF(__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_v_xs);
  __Pyx_XDECREF(__pyx_v_curvebyx);
  __Pyx_XDECREF(__pyx_v_curvebyxmax);
  __Pyx_XDECREF(__pyx_v_curvebyxmin);
  __Pyx_XDECREF(__pyx_v_meanCurve);
  __Pyx_XDECREF(__pyx_v_CICurve);
  __Pyx_XDECREF(__pyx_v_AUCs);
  __Pyx_XDECREF(__pyx_v_bio);
  __Pyx_XDECREF(__pyx_v_encoded);
  __Pyx_XDECREF(__pyx_v_html);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_15crossvalidation_7plotPRC_2generator13);
  __Pyx_XDECREF(__pyx_9genexpr51__pyx_v_curve);
  __Pyx_XDECREF(__pyx_9genexpr52__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr53__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr54__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr55__pyx_v_i);
  __Pyx_XDECREF(__pyx_9genexpr55__pyx_v_curve);
  __Pyx_XDECREF(__pyx_9genexpr56__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr57__pyx_v_i);
  __Pyx_XDECREF(__pyx_9genexpr57__pyx_v_curve);
  __Pyx_XDECREF(__pyx_9genexpr58__pyx_v_x);
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr58__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr59__pyx_v_x);
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr59__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr60__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr61__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr61__pyx_v_ci);
  __Pyx_XDECREF(__pyx_9genexpr62__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr62__pyx_v_ci);
  __Pyx_XDECREF(__pyx_9genexpr63__pyx_v_curve);
  __Pyx_XDECREF(__pyx_9genexpr64__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr65__pyx_v_pt);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1133
 * 			raise err
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_11plotROC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_11plotROC = {"plotROC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_15crossvalidation_11plotROC, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_11plotROC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_figsize = 0;
  PyObject *__pyx_v_outpath = 0;
  PyObject *__pyx_v_style = 0;
  PyObject *__pyx_v_returnHTML = 0;
  PyObject *__pyx_v_fontsizeLabels = 0;
  PyObject *__pyx_v_fontsizeLegend = 0;
  PyObject *__pyx_v_fontsizeAxis = 0;
  PyObject *__pyx_v_bboxAnchorTo = 0;
  PyObject *__pyx_v_legendLoc = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("plotROC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_figsize,&__pyx_n_s_outpath,&__pyx_n_s_style,&__pyx_n_s_returnHTML,&__pyx_n_s_fontsizeLabels,&__pyx_n_s_fontsizeLegend,&__pyx_n_s_fontsizeAxis,&__pyx_n_s_bboxAnchorTo,&__pyx_n_s_legendLoc,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject*)__pyx_tuple__4));
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject*)__pyx_n_u_ggplot));
    values[4] = ((PyObject *)((PyObject *)Py_False));
    values[5] = ((PyObject *)((PyObject *)__pyx_int_18));
    values[6] = ((PyObject *)((PyObject *)__pyx_int_12));
    values[7] = ((PyObject *)((PyObject *)__pyx_int_10));
    values[8] = ((PyObject *)((PyObject*)__pyx_tuple__26));
    values[9] = ((PyObject *)((PyObject*)__pyx_kp_u_upper_left));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_figsize);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outpath);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_returnHTML);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLabels);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeLegend);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontsizeAxis);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bboxAnchorTo);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_legendLoc);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "plotROC") < 0)) __PYX_ERR(0, 1133, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_figsize = values[1];
    __pyx_v_outpath = values[2];
    __pyx_v_style = values[3];
    __pyx_v_returnHTML = values[4];
    __pyx_v_fontsizeLabels = values[5];
    __pyx_v_fontsizeLegend = values[6];
    __pyx_v_fontsizeAxis = values[7];
    __pyx_v_bboxAnchorTo = values[8];
    __pyx_v_legendLoc = values[9];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("plotROC", 0, 1, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1133, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.plotROC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_10plotROC(__pyx_self, __pyx_v_self, __pyx_v_figsize, __pyx_v_outpath, __pyx_v_style, __pyx_v_returnHTML, __pyx_v_fontsizeLabels, __pyx_v_fontsizeLegend, __pyx_v_fontsizeAxis, __pyx_v_bboxAnchorTo, __pyx_v_legendLoc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_7plotROC_2generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "gnocis/models.pyx":1151
 * 				for mdl in self.models:
 * 					curves = self.ROC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))             # <<<<<<<<<<<<<<
 * 					curvebyx = [
 * 						{
 */

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_7plotROC_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_27_genexpr(__pyx_ptype_6gnocis_6models___pyx_scope_struct_27_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1151, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6gnocis_6models_15crossvalidation_7plotROC_2generator14, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_crossvalidation_plotROC_locals_g, __pyx_n_s_gnocis_models); if (unlikely(!gen)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidation.plotROC.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_7plotROC_2generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves)) { __Pyx_RaiseClosureNameError("curves"); __PYX_ERR(0, 1151, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_curves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1151, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1151, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1151, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1151, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_c)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_c)) {
      __pyx_t_4 = __pyx_cur_scope->__pyx_v_c; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_c); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1151, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1151, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1151, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1151, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1151, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1151, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_pt);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_pt, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_pt, __pyx_n_s_x); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 1151, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1133
 * 			raise err
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_10plotROC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_figsize, PyObject *__pyx_v_outpath, PyObject *__pyx_v_style, PyObject *__pyx_v_returnHTML, PyObject *__pyx_v_fontsizeLabels, PyObject *__pyx_v_fontsizeLegend, PyObject *__pyx_v_fontsizeAxis, PyObject *__pyx_v_bboxAnchorTo, PyObject *__pyx_v_legendLoc) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *__pyx_cur_scope;
  PyObject *__pyx_v_plt = NULL;
  PyObject *__pyx_v_base64 = NULL;
  PyObject *__pyx_v_BytesIO = NULL;
  PyObject *__pyx_v_display = NULL;
  PyObject *__pyx_v_HTML = NULL;
  PyObject *__pyx_v_fig = NULL;
  PyObject *__pyx_v_mdl = NULL;
  PyObject *__pyx_v_xs = NULL;
  PyObject *__pyx_v_curvebyx = NULL;
  PyObject *__pyx_v_curvebyxmax = NULL;
  PyObject *__pyx_v_curvebyxmin = NULL;
  PyObject *__pyx_v_meanCurve = NULL;
  PyObject *__pyx_v_CICurve = NULL;
  PyObject *__pyx_v_AUCs = NULL;
  PyObject *__pyx_v_bio = NULL;
  PyObject *__pyx_v_encoded = NULL;
  PyObject *__pyx_v_html = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_gb_6gnocis_6models_15crossvalidation_7plotROC_2generator14 = 0;
  PyObject *__pyx_9genexpr67__pyx_v_curve = NULL;
  PyObject *__pyx_9genexpr68__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr69__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr70__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr71__pyx_v_i = NULL;
  CYTHON_UNUSED PyObject *__pyx_9genexpr71__pyx_v_curve = NULL;
  PyObject *__pyx_9genexpr72__pyx_v_x = NULL;
  PyObject *__pyx_9genexpr73__pyx_v_i = NULL;
  CYTHON_UNUSED PyObject *__pyx_9genexpr73__pyx_v_curve = NULL;
  PyObject *__pyx_9genexpr74__pyx_v_x = NULL;
  struct __pyx_obj_6gnocis_10validation_point2D *__pyx_9genexpr74__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr75__pyx_v_x = NULL;
  struct __pyx_obj_6gnocis_10validation_point2D *__pyx_9genexpr75__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr76__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr77__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr77__pyx_v_ci = NULL;
  PyObject *__pyx_9genexpr78__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr78__pyx_v_ci = NULL;
  PyObject *__pyx_9genexpr79__pyx_v_curve = NULL;
  PyObject *__pyx_9genexpr80__pyx_v_pt = NULL;
  PyObject *__pyx_9genexpr81__pyx_v_pt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  PyObject *__pyx_t_17 = NULL;
  Py_ssize_t __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  Py_ssize_t __pyx_t_21;
  PyObject *(*__pyx_t_22)(PyObject *);
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_t_25;
  PyObject *(*__pyx_t_26)(PyObject *);
  PyObject *__pyx_t_27 = NULL;
  Py_UCS4 __pyx_t_28;
  int __pyx_t_29;
  PyObject *__pyx_t_30 = NULL;
  int __pyx_t_31;
  int __pyx_t_32;
  char const *__pyx_t_33;
  PyObject *__pyx_t_34 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("plotROC", 0);
  __pyx_cur_scope = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *)__pyx_tp_new_6gnocis_6models___pyx_scope_struct_26_plotROC(__pyx_ptype_6gnocis_6models___pyx_scope_struct_26_plotROC, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1133, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "gnocis/models.pyx":1134
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "gnocis/models.pyx":1135
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 * 		try:
 * 			import matplotlib.pyplot as plt             # <<<<<<<<<<<<<<
 * 			import base64
 * 			from io import BytesIO
 */
      __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1135, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_n_s__5);
      __Pyx_GIVEREF(__pyx_n_s__5);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s__5);
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_matplotlib_pyplot, __pyx_t_4, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1135, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_plt = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "gnocis/models.pyx":1136
 * 		try:
 * 			import matplotlib.pyplot as plt
 * 			import base64             # <<<<<<<<<<<<<<
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 */
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_base64, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1136, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_base64 = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "gnocis/models.pyx":1137
 * 			import matplotlib.pyplot as plt
 * 			import base64
 * 			from io import BytesIO             # <<<<<<<<<<<<<<
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):
 */
      __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1137, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_n_s_BytesIO);
      __Pyx_GIVEREF(__pyx_n_s_BytesIO);
      PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_BytesIO);
      __pyx_t_4 = __Pyx_Import(__pyx_n_s_io, __pyx_t_5, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1137, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_BytesIO); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1137, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_BytesIO = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "gnocis/models.pyx":1138
 * 			import base64
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML             # <<<<<<<<<<<<<<
 * 			with plt.style.context(style):
 * 				fig = plt.figure(figsize = figsize)
 */
      __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1138, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_n_s_display);
      __Pyx_GIVEREF(__pyx_n_s_display);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_display);
      __Pyx_INCREF(__pyx_n_s_HTML);
      __Pyx_GIVEREF(__pyx_n_s_HTML);
      PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_HTML);
      __pyx_t_5 = __Pyx_Import(__pyx_n_s_IPython_core_display, __pyx_t_4, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1138, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_display); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1138, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_display = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_5, __pyx_n_s_HTML); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1138, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_v_HTML = __pyx_t_4;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "gnocis/models.pyx":1139
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 				fig = plt.figure(figsize = figsize)
 * 				# Expected random generalization
 */
      /*with:*/ {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_style); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1139, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_context); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1139, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_style) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_style);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1139, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1139, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1139, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1139, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_9);
            __Pyx_XGOTREF(__pyx_t_10);
            __Pyx_XGOTREF(__pyx_t_11);
            /*try:*/ {

              /* "gnocis/models.pyx":1140
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):
 * 				fig = plt.figure(figsize = figsize)             # <<<<<<<<<<<<<<
 * 				# Expected random generalization
 * 				plt.plot(
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_figure); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1140, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1140, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_figsize, __pyx_v_figsize) < 0) __PYX_ERR(0, 1140, __pyx_L13_error)
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1140, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_v_fig = __pyx_t_4;
              __pyx_t_4 = 0;

              /* "gnocis/models.pyx":1142
 * 				fig = plt.figure(figsize = figsize)
 * 				# Expected random generalization
 * 				plt.plot(             # <<<<<<<<<<<<<<
 * 					[ 0., 1. ],
 * 					[ 0., 1. ],
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_plot); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1142, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_4);

              /* "gnocis/models.pyx":1143
 * 				# Expected random generalization
 * 				plt.plot(
 * 					[ 0., 1. ],             # <<<<<<<<<<<<<<
 * 					[ 0., 1. ],
 * 					linestyle = '--',
 */
              __pyx_t_6 = PyList_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1143, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_INCREF(__pyx_float_0_);
              __Pyx_GIVEREF(__pyx_float_0_);
              PyList_SET_ITEM(__pyx_t_6, 0, __pyx_float_0_);
              __Pyx_INCREF(__pyx_float_1_);
              __Pyx_GIVEREF(__pyx_float_1_);
              PyList_SET_ITEM(__pyx_t_6, 1, __pyx_float_1_);

              /* "gnocis/models.pyx":1144
 * 				plt.plot(
 * 					[ 0., 1. ],
 * 					[ 0., 1. ],             # <<<<<<<<<<<<<<
 * 					linestyle = '--',
 * 					color = 'grey',
 */
              __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1144, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_INCREF(__pyx_float_0_);
              __Pyx_GIVEREF(__pyx_float_0_);
              PyList_SET_ITEM(__pyx_t_5, 0, __pyx_float_0_);
              __Pyx_INCREF(__pyx_float_1_);
              __Pyx_GIVEREF(__pyx_float_1_);
              PyList_SET_ITEM(__pyx_t_5, 1, __pyx_float_1_);

              /* "gnocis/models.pyx":1142
 * 				fig = plt.figure(figsize = figsize)
 * 				# Expected random generalization
 * 				plt.plot(             # <<<<<<<<<<<<<<
 * 					[ 0., 1. ],
 * 					[ 0., 1. ],
 */
              __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1142, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_6);
              PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_5);
              __pyx_t_6 = 0;
              __pyx_t_5 = 0;

              /* "gnocis/models.pyx":1145
 * 					[ 0., 1. ],
 * 					[ 0., 1. ],
 * 					linestyle = '--',             # <<<<<<<<<<<<<<
 * 					color = 'grey',
 * 					label = 'Expected at random')
 */
              __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1145, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_linestyle, __pyx_kp_u__6) < 0) __PYX_ERR(0, 1145, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_color, __pyx_n_u_grey) < 0) __PYX_ERR(0, 1145, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_label, __pyx_kp_u_Expected_at_random) < 0) __PYX_ERR(0, 1145, __pyx_L13_error)

              /* "gnocis/models.pyx":1142
 * 				fig = plt.figure(figsize = figsize)
 * 				# Expected random generalization
 * 				plt.plot(             # <<<<<<<<<<<<<<
 * 					[ 0., 1. ],
 * 					[ 0., 1. ],
 */
              __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1142, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

              /* "gnocis/models.pyx":1149
 * 					label = 'Expected at random')
 * 				# Curves per model
 * 				for mdl in self.models:             # <<<<<<<<<<<<<<
 * 					curves = self.ROC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 */
              __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1149, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_6);
              if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
                __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5); __pyx_t_12 = 0;
                __pyx_t_13 = NULL;
              } else {
                __pyx_t_12 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1149, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_13 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1149, __pyx_L13_error)
              }
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              for (;;) {
                if (likely(!__pyx_t_13)) {
                  if (likely(PyList_CheckExact(__pyx_t_5))) {
                    if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_5)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_12); __Pyx_INCREF(__pyx_t_6); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1149, __pyx_L13_error)
                    #else
                    __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1149, __pyx_L13_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    #endif
                  } else {
                    if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_12); __Pyx_INCREF(__pyx_t_6); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1149, __pyx_L13_error)
                    #else
                    __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1149, __pyx_L13_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    #endif
                  }
                } else {
                  __pyx_t_6 = __pyx_t_13(__pyx_t_5);
                  if (unlikely(!__pyx_t_6)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 1149, __pyx_L13_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_6);
                }
                __Pyx_XDECREF_SET(__pyx_v_mdl, __pyx_t_6);
                __pyx_t_6 = 0;

                /* "gnocis/models.pyx":1150
 * 				# Curves per model
 * 				for mdl in self.models:
 * 					curves = self.ROC[mdl]             # <<<<<<<<<<<<<<
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 * 					curvebyx = [
 */
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ROC); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1150, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_v_mdl); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1150, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_curves);
                __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_curves, __pyx_t_8);
                __Pyx_GIVEREF(__pyx_t_8);
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1151
 * 				for mdl in self.models:
 * 					curves = self.ROC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))             # <<<<<<<<<<<<<<
 * 					curvebyx = [
 * 						{
 */
                __pyx_t_6 = __pyx_pf_6gnocis_6models_15crossvalidation_7plotROC_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1151, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1151, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_6 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1151, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = PySequence_List(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1151, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_8 = ((PyObject*)__pyx_t_4);
                __pyx_t_4 = 0;
                __pyx_t_14 = PyList_Sort(__pyx_t_8); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 1151, __pyx_L13_error)
                __Pyx_XDECREF_SET(__pyx_v_xs, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1152
 * 					curves = self.ROC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 * 					curvebyx = [             # <<<<<<<<<<<<<<
 * 						{
 * 							x: [ pt.y for pt in curve if pt.x == x ]
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1152, __pyx_L23_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1157
 * 							for x in xs
 * 						}
 * 						for curve in curves             # <<<<<<<<<<<<<<
 * 					]
 * 					curvebyxmax = {
 */
                  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_curves)) {
                    __pyx_t_4 = __pyx_cur_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_4); __pyx_t_15 = 0;
                    __pyx_t_16 = NULL;
                  } else {
                    __pyx_t_15 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_curves); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1157, __pyx_L23_error)
                    __Pyx_GOTREF(__pyx_t_4);
                    __pyx_t_16 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1157, __pyx_L23_error)
                  }
                  for (;;) {
                    if (likely(!__pyx_t_16)) {
                      if (likely(PyList_CheckExact(__pyx_t_4))) {
                        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_4)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1157, __pyx_L23_error)
                        #else
                        __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1157, __pyx_L23_error)
                        __Pyx_GOTREF(__pyx_t_6);
                        #endif
                      } else {
                        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1157, __pyx_L23_error)
                        #else
                        __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1157, __pyx_L23_error)
                        __Pyx_GOTREF(__pyx_t_6);
                        #endif
                      }
                    } else {
                      __pyx_t_6 = __pyx_t_16(__pyx_t_4);
                      if (unlikely(!__pyx_t_6)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 1157, __pyx_L23_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_6);
                    }
                    __Pyx_XDECREF_SET(__pyx_9genexpr67__pyx_v_curve, __pyx_t_6);
                    __pyx_t_6 = 0;
                    { /* enter inner scope */

                      /* "gnocis/models.pyx":1153
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 * 					curvebyx = [
 * 						{             # <<<<<<<<<<<<<<
 * 							x: [ pt.y for pt in curve if pt.x == x ]
 * 							for x in xs
 */
                      __pyx_t_6 = PyDict_New(); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1153, __pyx_L28_error)
                      __Pyx_GOTREF(__pyx_t_6);

                      /* "gnocis/models.pyx":1155
 * 						{
 * 							x: [ pt.y for pt in curve if pt.x == x ]
 * 							for x in xs             # <<<<<<<<<<<<<<
 * 						}
 * 						for curve in curves
 */
                      if (unlikely(__pyx_v_xs == Py_None)) {
                        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                        __PYX_ERR(0, 1155, __pyx_L28_error)
                      }
                      __pyx_t_17 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_17); __pyx_t_18 = 0;
                      for (;;) {
                        if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_17)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_19 = PyList_GET_ITEM(__pyx_t_17, __pyx_t_18); __Pyx_INCREF(__pyx_t_19); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1155, __pyx_L28_error)
                        #else
                        __pyx_t_19 = PySequence_ITEM(__pyx_t_17, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1155, __pyx_L28_error)
                        __Pyx_GOTREF(__pyx_t_19);
                        #endif
                        __Pyx_XDECREF_SET(__pyx_9genexpr68__pyx_v_x, __pyx_t_19);
                        __pyx_t_19 = 0;

                        /* "gnocis/models.pyx":1154
 * 					curvebyx = [
 * 						{
 * 							x: [ pt.y for pt in curve if pt.x == x ]             # <<<<<<<<<<<<<<
 * 							for x in xs
 * 						}
 */
                        { /* enter inner scope */
                          __pyx_t_19 = PyList_New(0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1154, __pyx_L33_error)
                          __Pyx_GOTREF(__pyx_t_19);
                          if (likely(PyList_CheckExact(__pyx_9genexpr67__pyx_v_curve)) || PyTuple_CheckExact(__pyx_9genexpr67__pyx_v_curve)) {
                            __pyx_t_20 = __pyx_9genexpr67__pyx_v_curve; __Pyx_INCREF(__pyx_t_20); __pyx_t_21 = 0;
                            __pyx_t_22 = NULL;
                          } else {
                            __pyx_t_21 = -1; __pyx_t_20 = PyObject_GetIter(__pyx_9genexpr67__pyx_v_curve); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1154, __pyx_L33_error)
                            __Pyx_GOTREF(__pyx_t_20);
                            __pyx_t_22 = Py_TYPE(__pyx_t_20)->tp_iternext; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1154, __pyx_L33_error)
                          }
                          for (;;) {
                            if (likely(!__pyx_t_22)) {
                              if (likely(PyList_CheckExact(__pyx_t_20))) {
                                if (__pyx_t_21 >= PyList_GET_SIZE(__pyx_t_20)) break;
                                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                                __pyx_t_23 = PyList_GET_ITEM(__pyx_t_20, __pyx_t_21); __Pyx_INCREF(__pyx_t_23); __pyx_t_21++; if (unlikely(0 < 0)) __PYX_ERR(0, 1154, __pyx_L33_error)
                                #else
                                __pyx_t_23 = PySequence_ITEM(__pyx_t_20, __pyx_t_21); __pyx_t_21++; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1154, __pyx_L33_error)
                                __Pyx_GOTREF(__pyx_t_23);
                                #endif
                              } else {
                                if (__pyx_t_21 >= PyTuple_GET_SIZE(__pyx_t_20)) break;
                                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                                __pyx_t_23 = PyTuple_GET_ITEM(__pyx_t_20, __pyx_t_21); __Pyx_INCREF(__pyx_t_23); __pyx_t_21++; if (unlikely(0 < 0)) __PYX_ERR(0, 1154, __pyx_L33_error)
                                #else
                                __pyx_t_23 = PySequence_ITEM(__pyx_t_20, __pyx_t_21); __pyx_t_21++; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1154, __pyx_L33_error)
                                __Pyx_GOTREF(__pyx_t_23);
                                #endif
                              }
                            } else {
                              __pyx_t_23 = __pyx_t_22(__pyx_t_20);
                              if (unlikely(!__pyx_t_23)) {
                                PyObject* exc_type = PyErr_Occurred();
                                if (exc_type) {
                                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                                  else __PYX_ERR(0, 1154, __pyx_L33_error)
                                }
                                break;
                              }
                              __Pyx_GOTREF(__pyx_t_23);
                            }
                            __Pyx_XDECREF_SET(__pyx_9genexpr69__pyx_v_pt, __pyx_t_23);
                            __pyx_t_23 = 0;
                            __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr69__pyx_v_pt, __pyx_n_s_x); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1154, __pyx_L33_error)
                            __Pyx_GOTREF(__pyx_t_23);
                            __pyx_t_24 = PyObject_RichCompare(__pyx_t_23, __pyx_9genexpr68__pyx_v_x, Py_EQ); __Pyx_XGOTREF(__pyx_t_24); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1154, __pyx_L33_error)
                            __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                            __pyx_t_25 = __Pyx_PyObject_IsTrue(__pyx_t_24); if (unlikely(__pyx_t_25 < 0)) __PYX_ERR(0, 1154, __pyx_L33_error)
                            __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                            if (__pyx_t_25) {
                              __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr69__pyx_v_pt, __pyx_n_s_y); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1154, __pyx_L33_error)
                              __Pyx_GOTREF(__pyx_t_24);
                              if (unlikely(__Pyx_ListComp_Append(__pyx_t_19, (PyObject*)__pyx_t_24))) __PYX_ERR(0, 1154, __pyx_L33_error)
                              __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                            }
                          }
                          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                          __Pyx_XDECREF(__pyx_9genexpr69__pyx_v_pt); __pyx_9genexpr69__pyx_v_pt = 0;
                          goto __pyx_L37_exit_scope;
                          __pyx_L33_error:;
                          __Pyx_XDECREF(__pyx_9genexpr69__pyx_v_pt); __pyx_9genexpr69__pyx_v_pt = 0;
                          goto __pyx_L28_error;
                          __pyx_L37_exit_scope:;
                        } /* exit inner scope */
                        if (unlikely(PyDict_SetItem(__pyx_t_6, (PyObject*)__pyx_9genexpr68__pyx_v_x, (PyObject*)__pyx_t_19))) __PYX_ERR(0, 1154, __pyx_L28_error)
                        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                        /* "gnocis/models.pyx":1155
 * 						{
 * 							x: [ pt.y for pt in curve if pt.x == x ]
 * 							for x in xs             # <<<<<<<<<<<<<<
 * 						}
 * 						for curve in curves
 */
                      }
                      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                      __Pyx_XDECREF(__pyx_9genexpr68__pyx_v_x); __pyx_9genexpr68__pyx_v_x = 0;
                      goto __pyx_L38_exit_scope;
                      __pyx_L28_error:;
                      __Pyx_XDECREF(__pyx_9genexpr68__pyx_v_x); __pyx_9genexpr68__pyx_v_x = 0;
                      goto __pyx_L23_error;
                      __pyx_L38_exit_scope:;
                    } /* exit inner scope */
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1152, __pyx_L23_error)
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1157
 * 							for x in xs
 * 						}
 * 						for curve in curves             # <<<<<<<<<<<<<<
 * 					]
 * 					curvebyxmax = {
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr67__pyx_v_curve); __pyx_9genexpr67__pyx_v_curve = 0;
                  goto __pyx_L39_exit_scope;
                  __pyx_L23_error:;
                  __Pyx_XDECREF(__pyx_9genexpr67__pyx_v_curve); __pyx_9genexpr67__pyx_v_curve = 0;
                  goto __pyx_L13_error;
                  __pyx_L39_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_curvebyx, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1159
 * 						for curve in curves
 * 					]
 * 					curvebyxmax = {             # <<<<<<<<<<<<<<
 * 						x: [
 * 							max(curvebyx[i][x])
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1159, __pyx_L42_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1164
 * 							for i, curve in enumerate(curves)
 * 						]
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 					}
 * 					curvebyxmin = {
 */
                  if (unlikely(__pyx_v_xs == Py_None)) {
                    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                    __PYX_ERR(0, 1164, __pyx_L42_error)
                  }
                  __pyx_t_4 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_4); __pyx_t_15 = 0;
                  for (;;) {
                    if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_4)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1164, __pyx_L42_error)
                    #else
                    __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1164, __pyx_L42_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr70__pyx_v_x, __pyx_t_6);
                    __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1160
 * 					]
 * 					curvebyxmax = {
 * 						x: [             # <<<<<<<<<<<<<<
 * 							max(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)
 */
                    { /* enter inner scope */
                      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1160, __pyx_L47_error)
                      __Pyx_GOTREF(__pyx_t_6);
                      __Pyx_INCREF(__pyx_int_0);
                      __pyx_t_17 = __pyx_int_0;

                      /* "gnocis/models.pyx":1162
 * 						x: [
 * 							max(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)             # <<<<<<<<<<<<<<
 * 						]
 * 						for x in xs
 */
                      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_curves)) {
                        __pyx_t_19 = __pyx_cur_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_19); __pyx_t_18 = 0;
                        __pyx_t_16 = NULL;
                      } else {
                        __pyx_t_18 = -1; __pyx_t_19 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_curves); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1162, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_19);
                        __pyx_t_16 = Py_TYPE(__pyx_t_19)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1162, __pyx_L47_error)
                      }
                      for (;;) {
                        if (likely(!__pyx_t_16)) {
                          if (likely(PyList_CheckExact(__pyx_t_19))) {
                            if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_19)) break;
                            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                            __pyx_t_20 = PyList_GET_ITEM(__pyx_t_19, __pyx_t_18); __Pyx_INCREF(__pyx_t_20); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1162, __pyx_L47_error)
                            #else
                            __pyx_t_20 = PySequence_ITEM(__pyx_t_19, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1162, __pyx_L47_error)
                            __Pyx_GOTREF(__pyx_t_20);
                            #endif
                          } else {
                            if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_19)) break;
                            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                            __pyx_t_20 = PyTuple_GET_ITEM(__pyx_t_19, __pyx_t_18); __Pyx_INCREF(__pyx_t_20); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1162, __pyx_L47_error)
                            #else
                            __pyx_t_20 = PySequence_ITEM(__pyx_t_19, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1162, __pyx_L47_error)
                            __Pyx_GOTREF(__pyx_t_20);
                            #endif
                          }
                        } else {
                          __pyx_t_20 = __pyx_t_16(__pyx_t_19);
                          if (unlikely(!__pyx_t_20)) {
                            PyObject* exc_type = PyErr_Occurred();
                            if (exc_type) {
                              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                              else __PYX_ERR(0, 1162, __pyx_L47_error)
                            }
                            break;
                          }
                          __Pyx_GOTREF(__pyx_t_20);
                        }
                        __Pyx_XDECREF_SET(__pyx_9genexpr71__pyx_v_curve, __pyx_t_20);
                        __pyx_t_20 = 0;
                        __Pyx_INCREF(__pyx_t_17);
                        __Pyx_XDECREF_SET(__pyx_9genexpr71__pyx_v_i, __pyx_t_17);
                        __pyx_t_20 = __Pyx_PyInt_AddObjC(__pyx_t_17, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1162, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_20);
                        __Pyx_DECREF(__pyx_t_17);
                        __pyx_t_17 = __pyx_t_20;
                        __pyx_t_20 = 0;

                        /* "gnocis/models.pyx":1161
 * 					curvebyxmax = {
 * 						x: [
 * 							max(curvebyx[i][x])             # <<<<<<<<<<<<<<
 * 							for i, curve in enumerate(curves)
 * 						]
 */
                        __pyx_t_20 = __Pyx_PyObject_GetItem(__pyx_v_curvebyx, __pyx_9genexpr71__pyx_v_i); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1161, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_20);
                        __pyx_t_24 = __Pyx_PyObject_GetItem(__pyx_t_20, __pyx_9genexpr70__pyx_v_x); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1161, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_24);
                        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                        __pyx_t_20 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_24); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1161, __pyx_L47_error)
                        __Pyx_GOTREF(__pyx_t_20);
                        __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                        if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_20))) __PYX_ERR(0, 1160, __pyx_L47_error)
                        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

                        /* "gnocis/models.pyx":1162
 * 						x: [
 * 							max(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)             # <<<<<<<<<<<<<<
 * 						]
 * 						for x in xs
 */
                      }
                      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                      __Pyx_XDECREF(__pyx_9genexpr71__pyx_v_curve); __pyx_9genexpr71__pyx_v_curve = 0;
                      __Pyx_XDECREF(__pyx_9genexpr71__pyx_v_i); __pyx_9genexpr71__pyx_v_i = 0;
                      goto __pyx_L50_exit_scope;
                      __pyx_L47_error:;
                      __Pyx_XDECREF(__pyx_9genexpr71__pyx_v_curve); __pyx_9genexpr71__pyx_v_curve = 0;
                      __Pyx_XDECREF(__pyx_9genexpr71__pyx_v_i); __pyx_9genexpr71__pyx_v_i = 0;
                      goto __pyx_L42_error;
                      __pyx_L50_exit_scope:;
                    } /* exit inner scope */
                    if (unlikely(PyDict_SetItem(__pyx_t_8, (PyObject*)__pyx_9genexpr70__pyx_v_x, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1160, __pyx_L42_error)
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1164
 * 							for i, curve in enumerate(curves)
 * 						]
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 					}
 * 					curvebyxmin = {
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr70__pyx_v_x); __pyx_9genexpr70__pyx_v_x = 0;
                  goto __pyx_L51_exit_scope;
                  __pyx_L42_error:;
                  __Pyx_XDECREF(__pyx_9genexpr70__pyx_v_x); __pyx_9genexpr70__pyx_v_x = 0;
                  goto __pyx_L13_error;
                  __pyx_L51_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_curvebyxmax, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1166
 * 						for x in xs
 * 					}
 * 					curvebyxmin = {             # <<<<<<<<<<<<<<
 * 						x: [
 * 							min(curvebyx[i][x])
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1166, __pyx_L54_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1171
 * 							for i, curve in enumerate(curves)
 * 						]
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 					}
 * 					meanCurve = [
 */
                  if (unlikely(__pyx_v_xs == Py_None)) {
                    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                    __PYX_ERR(0, 1171, __pyx_L54_error)
                  }
                  __pyx_t_4 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_4); __pyx_t_15 = 0;
                  for (;;) {
                    if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_4)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1171, __pyx_L54_error)
                    #else
                    __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1171, __pyx_L54_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr72__pyx_v_x, __pyx_t_6);
                    __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1167
 * 					}
 * 					curvebyxmin = {
 * 						x: [             # <<<<<<<<<<<<<<
 * 							min(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)
 */
                    { /* enter inner scope */
                      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1167, __pyx_L59_error)
                      __Pyx_GOTREF(__pyx_t_6);
                      __Pyx_INCREF(__pyx_int_0);
                      __pyx_t_17 = __pyx_int_0;

                      /* "gnocis/models.pyx":1169
 * 						x: [
 * 							min(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)             # <<<<<<<<<<<<<<
 * 						]
 * 						for x in xs
 */
                      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_curves)) {
                        __pyx_t_19 = __pyx_cur_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_19); __pyx_t_18 = 0;
                        __pyx_t_16 = NULL;
                      } else {
                        __pyx_t_18 = -1; __pyx_t_19 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_curves); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1169, __pyx_L59_error)
                        __Pyx_GOTREF(__pyx_t_19);
                        __pyx_t_16 = Py_TYPE(__pyx_t_19)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1169, __pyx_L59_error)
                      }
                      for (;;) {
                        if (likely(!__pyx_t_16)) {
                          if (likely(PyList_CheckExact(__pyx_t_19))) {
                            if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_19)) break;
                            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                            __pyx_t_20 = PyList_GET_ITEM(__pyx_t_19, __pyx_t_18); __Pyx_INCREF(__pyx_t_20); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1169, __pyx_L59_error)
                            #else
                            __pyx_t_20 = PySequence_ITEM(__pyx_t_19, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1169, __pyx_L59_error)
                            __Pyx_GOTREF(__pyx_t_20);
                            #endif
                          } else {
                            if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_19)) break;
                            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                            __pyx_t_20 = PyTuple_GET_ITEM(__pyx_t_19, __pyx_t_18); __Pyx_INCREF(__pyx_t_20); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1169, __pyx_L59_error)
                            #else
                            __pyx_t_20 = PySequence_ITEM(__pyx_t_19, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1169, __pyx_L59_error)
                            __Pyx_GOTREF(__pyx_t_20);
                            #endif
                          }
                        } else {
                          __pyx_t_20 = __pyx_t_16(__pyx_t_19);
                          if (unlikely(!__pyx_t_20)) {
                            PyObject* exc_type = PyErr_Occurred();
                            if (exc_type) {
                              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                              else __PYX_ERR(0, 1169, __pyx_L59_error)
                            }
                            break;
                          }
                          __Pyx_GOTREF(__pyx_t_20);
                        }
                        __Pyx_XDECREF_SET(__pyx_9genexpr73__pyx_v_curve, __pyx_t_20);
                        __pyx_t_20 = 0;
                        __Pyx_INCREF(__pyx_t_17);
                        __Pyx_XDECREF_SET(__pyx_9genexpr73__pyx_v_i, __pyx_t_17);
                        __pyx_t_20 = __Pyx_PyInt_AddObjC(__pyx_t_17, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1169, __pyx_L59_error)
                        __Pyx_GOTREF(__pyx_t_20);
                        __Pyx_DECREF(__pyx_t_17);
                        __pyx_t_17 = __pyx_t_20;
                        __pyx_t_20 = 0;

                        /* "gnocis/models.pyx":1168
 * 					curvebyxmin = {
 * 						x: [
 * 							min(curvebyx[i][x])             # <<<<<<<<<<<<<<
 * 							for i, curve in enumerate(curves)
 * 						]
 */
                        __pyx_t_20 = __Pyx_PyObject_GetItem(__pyx_v_curvebyx, __pyx_9genexpr73__pyx_v_i); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1168, __pyx_L59_error)
                        __Pyx_GOTREF(__pyx_t_20);
                        __pyx_t_24 = __Pyx_PyObject_GetItem(__pyx_t_20, __pyx_9genexpr72__pyx_v_x); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1168, __pyx_L59_error)
                        __Pyx_GOTREF(__pyx_t_24);
                        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                        __pyx_t_20 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_24); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1168, __pyx_L59_error)
                        __Pyx_GOTREF(__pyx_t_20);
                        __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                        if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_20))) __PYX_ERR(0, 1167, __pyx_L59_error)
                        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

                        /* "gnocis/models.pyx":1169
 * 						x: [
 * 							min(curvebyx[i][x])
 * 							for i, curve in enumerate(curves)             # <<<<<<<<<<<<<<
 * 						]
 * 						for x in xs
 */
                      }
                      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                      __Pyx_XDECREF(__pyx_9genexpr73__pyx_v_curve); __pyx_9genexpr73__pyx_v_curve = 0;
                      __Pyx_XDECREF(__pyx_9genexpr73__pyx_v_i); __pyx_9genexpr73__pyx_v_i = 0;
                      goto __pyx_L62_exit_scope;
                      __pyx_L59_error:;
                      __Pyx_XDECREF(__pyx_9genexpr73__pyx_v_curve); __pyx_9genexpr73__pyx_v_curve = 0;
                      __Pyx_XDECREF(__pyx_9genexpr73__pyx_v_i); __pyx_9genexpr73__pyx_v_i = 0;
                      goto __pyx_L54_error;
                      __pyx_L62_exit_scope:;
                    } /* exit inner scope */
                    if (unlikely(PyDict_SetItem(__pyx_t_8, (PyObject*)__pyx_9genexpr72__pyx_v_x, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1167, __pyx_L54_error)
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1171
 * 							for i, curve in enumerate(curves)
 * 						]
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 					}
 * 					meanCurve = [
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr72__pyx_v_x); __pyx_9genexpr72__pyx_v_x = 0;
                  goto __pyx_L63_exit_scope;
                  __pyx_L54_error:;
                  __Pyx_XDECREF(__pyx_9genexpr72__pyx_v_x); __pyx_9genexpr72__pyx_v_x = 0;
                  goto __pyx_L13_error;
                  __pyx_L63_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_curvebyxmin, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1173
 * 						for x in xs
 * 					}
 * 					meanCurve = [             # <<<<<<<<<<<<<<
 * 						pt
 * 						for x in xs
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1173, __pyx_L66_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1175
 * 					meanCurve = [
 * 						pt
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 						for pt in [
 * 							point2D(x, mean(curvebyxmax[x])),
 */
                  if (unlikely(__pyx_v_xs == Py_None)) {
                    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                    __PYX_ERR(0, 1175, __pyx_L66_error)
                  }
                  __pyx_t_4 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_4); __pyx_t_15 = 0;
                  for (;;) {
                    if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_4)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1175, __pyx_L66_error)
                    #else
                    __pyx_t_6 = PySequence_ITEM(__pyx_t_4, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1175, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr74__pyx_v_x, __pyx_t_6);
                    __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1177
 * 						for x in xs
 * 						for pt in [
 * 							point2D(x, mean(curvebyxmax[x])),             # <<<<<<<<<<<<<<
 * 							point2D(x, mean(curvebyxmin[x]))
 * 						]
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_mean); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1177, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_curvebyxmax, __pyx_9genexpr74__pyx_v_x); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1177, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __pyx_t_20 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
                      __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_17);
                      if (likely(__pyx_t_20)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                        __Pyx_INCREF(__pyx_t_20);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_17, function);
                      }
                    }
                    __pyx_t_6 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_20, __pyx_t_19) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_19);
                    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1177, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                    __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1177, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __Pyx_INCREF(__pyx_9genexpr74__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_9genexpr74__pyx_v_x);
                    PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_9genexpr74__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_t_6);
                    PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_6);
                    __pyx_t_6 = 0;
                    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_point2D), __pyx_t_17, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1177, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

                    /* "gnocis/models.pyx":1178
 * 						for pt in [
 * 							point2D(x, mean(curvebyxmax[x])),
 * 							point2D(x, mean(curvebyxmin[x]))             # <<<<<<<<<<<<<<
 * 						]
 * 					]
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_mean); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1178, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __pyx_t_20 = __Pyx_PyDict_GetItem(__pyx_v_curvebyxmin, __pyx_9genexpr74__pyx_v_x); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1178, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_20);
                    __pyx_t_24 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
                      __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_19);
                      if (likely(__pyx_t_24)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
                        __Pyx_INCREF(__pyx_t_24);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_19, function);
                      }
                    }
                    __pyx_t_17 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_19, __pyx_t_24, __pyx_t_20) : __Pyx_PyObject_CallOneArg(__pyx_t_19, __pyx_t_20);
                    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                    if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1178, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                    __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1178, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __Pyx_INCREF(__pyx_9genexpr74__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_9genexpr74__pyx_v_x);
                    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_9genexpr74__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_t_17);
                    PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_17);
                    __pyx_t_17 = 0;
                    __pyx_t_17 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_point2D), __pyx_t_19, NULL); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1178, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                    /* "gnocis/models.pyx":1176
 * 						pt
 * 						for x in xs
 * 						for pt in [             # <<<<<<<<<<<<<<
 * 							point2D(x, mean(curvebyxmax[x])),
 * 							point2D(x, mean(curvebyxmin[x]))
 */
                    __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1176, __pyx_L66_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __Pyx_GIVEREF(__pyx_t_6);
                    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_6);
                    __Pyx_GIVEREF(__pyx_t_17);
                    PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_17);
                    __pyx_t_6 = 0;
                    __pyx_t_17 = 0;
                    __pyx_t_17 = __pyx_t_19; __Pyx_INCREF(__pyx_t_17); __pyx_t_18 = 0;
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                    for (;;) {
                      if (__pyx_t_18 >= 2) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_19 = PyTuple_GET_ITEM(__pyx_t_17, __pyx_t_18); __Pyx_INCREF(__pyx_t_19); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1176, __pyx_L66_error)
                      #else
                      __pyx_t_19 = PySequence_ITEM(__pyx_t_17, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1176, __pyx_L66_error)
                      __Pyx_GOTREF(__pyx_t_19);
                      #endif
                      if (!(likely(((__pyx_t_19) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_19, __pyx_ptype_6gnocis_10validation_point2D))))) __PYX_ERR(0, 1176, __pyx_L66_error)
                      __Pyx_XDECREF_SET(__pyx_9genexpr74__pyx_v_pt, ((struct __pyx_obj_6gnocis_10validation_point2D *)__pyx_t_19));
                      __pyx_t_19 = 0;

                      /* "gnocis/models.pyx":1174
 * 					}
 * 					meanCurve = [
 * 						pt             # <<<<<<<<<<<<<<
 * 						for x in xs
 * 						for pt in [
 */
                      if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_9genexpr74__pyx_v_pt))) __PYX_ERR(0, 1173, __pyx_L66_error)

                      /* "gnocis/models.pyx":1176
 * 						pt
 * 						for x in xs
 * 						for pt in [             # <<<<<<<<<<<<<<
 * 							point2D(x, mean(curvebyxmax[x])),
 * 							point2D(x, mean(curvebyxmin[x]))
 */
                    }
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

                    /* "gnocis/models.pyx":1175
 * 					meanCurve = [
 * 						pt
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 						for pt in [
 * 							point2D(x, mean(curvebyxmax[x])),
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF((PyObject *)__pyx_9genexpr74__pyx_v_pt); __pyx_9genexpr74__pyx_v_pt = 0;
                  __Pyx_XDECREF(__pyx_9genexpr74__pyx_v_x); __pyx_9genexpr74__pyx_v_x = 0;
                  goto __pyx_L71_exit_scope;
                  __pyx_L66_error:;
                  __Pyx_XDECREF((PyObject *)__pyx_9genexpr74__pyx_v_pt); __pyx_9genexpr74__pyx_v_pt = 0;
                  __Pyx_XDECREF(__pyx_9genexpr74__pyx_v_x); __pyx_9genexpr74__pyx_v_x = 0;
                  goto __pyx_L13_error;
                  __pyx_L71_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_meanCurve, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1181
 * 						]
 * 					]
 * 					CICurve = [             # <<<<<<<<<<<<<<
 * 						pt
 * 						for x in xs
 */
                { /* enter inner scope */
                  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1181, __pyx_L74_error)
                  __Pyx_GOTREF(__pyx_t_8);

                  /* "gnocis/models.pyx":1183
 * 					CICurve = [
 * 						pt
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 						for pt in [
 * 							point2D(x, CI(curvebyxmax[x])),
 */
                  if (unlikely(__pyx_v_xs == Py_None)) {
                    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
                    __PYX_ERR(0, 1183, __pyx_L74_error)
                  }
                  __pyx_t_4 = __pyx_v_xs; __Pyx_INCREF(__pyx_t_4); __pyx_t_15 = 0;
                  for (;;) {
                    if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_4)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_17 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_15); __Pyx_INCREF(__pyx_t_17); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1183, __pyx_L74_error)
                    #else
                    __pyx_t_17 = PySequence_ITEM(__pyx_t_4, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1183, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr75__pyx_v_x, __pyx_t_17);
                    __pyx_t_17 = 0;

                    /* "gnocis/models.pyx":1185
 * 						for x in xs
 * 						for pt in [
 * 							point2D(x, CI(curvebyxmax[x])),             # <<<<<<<<<<<<<<
 * 							point2D(x, CI(curvebyxmin[x]))
 * 						]
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_CI); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1185, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_curvebyxmax, __pyx_9genexpr75__pyx_v_x); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1185, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __pyx_t_20 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
                      __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_19);
                      if (likely(__pyx_t_20)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
                        __Pyx_INCREF(__pyx_t_20);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_19, function);
                      }
                    }
                    __pyx_t_17 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_19, __pyx_t_20, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_19, __pyx_t_6);
                    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                    if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1185, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                    __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1185, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __Pyx_INCREF(__pyx_9genexpr75__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_9genexpr75__pyx_v_x);
                    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_9genexpr75__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_t_17);
                    PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_17);
                    __pyx_t_17 = 0;
                    __pyx_t_17 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_point2D), __pyx_t_19, NULL); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1185, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                    /* "gnocis/models.pyx":1186
 * 						for pt in [
 * 							point2D(x, CI(curvebyxmax[x])),
 * 							point2D(x, CI(curvebyxmin[x]))             # <<<<<<<<<<<<<<
 * 						]
 * 					]
 */
                    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_CI); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1186, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __pyx_t_20 = __Pyx_PyDict_GetItem(__pyx_v_curvebyxmin, __pyx_9genexpr75__pyx_v_x); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1186, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_20);
                    __pyx_t_24 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                      __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_6);
                      if (likely(__pyx_t_24)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                        __Pyx_INCREF(__pyx_t_24);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_6, function);
                      }
                    }
                    __pyx_t_19 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_24, __pyx_t_20) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_20);
                    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                    if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1186, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1186, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_INCREF(__pyx_9genexpr75__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_9genexpr75__pyx_v_x);
                    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_9genexpr75__pyx_v_x);
                    __Pyx_GIVEREF(__pyx_t_19);
                    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_19);
                    __pyx_t_19 = 0;
                    __pyx_t_19 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gnocis_10validation_point2D), __pyx_t_6, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1186, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_19);
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                    /* "gnocis/models.pyx":1184
 * 						pt
 * 						for x in xs
 * 						for pt in [             # <<<<<<<<<<<<<<
 * 							point2D(x, CI(curvebyxmax[x])),
 * 							point2D(x, CI(curvebyxmin[x]))
 */
                    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1184, __pyx_L74_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __Pyx_GIVEREF(__pyx_t_17);
                    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_17);
                    __Pyx_GIVEREF(__pyx_t_19);
                    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_19);
                    __pyx_t_17 = 0;
                    __pyx_t_19 = 0;
                    __pyx_t_19 = __pyx_t_6; __Pyx_INCREF(__pyx_t_19); __pyx_t_18 = 0;
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                    for (;;) {
                      if (__pyx_t_18 >= 2) break;
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_19, __pyx_t_18); __Pyx_INCREF(__pyx_t_6); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 1184, __pyx_L74_error)
                      #else
                      __pyx_t_6 = PySequence_ITEM(__pyx_t_19, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1184, __pyx_L74_error)
                      __Pyx_GOTREF(__pyx_t_6);
                      #endif
                      if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_6gnocis_10validation_point2D))))) __PYX_ERR(0, 1184, __pyx_L74_error)
                      __Pyx_XDECREF_SET(__pyx_9genexpr75__pyx_v_pt, ((struct __pyx_obj_6gnocis_10validation_point2D *)__pyx_t_6));
                      __pyx_t_6 = 0;

                      /* "gnocis/models.pyx":1182
 * 					]
 * 					CICurve = [
 * 						pt             # <<<<<<<<<<<<<<
 * 						for x in xs
 * 						for pt in [
 */
                      if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_9genexpr75__pyx_v_pt))) __PYX_ERR(0, 1181, __pyx_L74_error)

                      /* "gnocis/models.pyx":1184
 * 						pt
 * 						for x in xs
 * 						for pt in [             # <<<<<<<<<<<<<<
 * 							point2D(x, CI(curvebyxmax[x])),
 * 							point2D(x, CI(curvebyxmin[x]))
 */
                    }
                    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                    /* "gnocis/models.pyx":1183
 * 					CICurve = [
 * 						pt
 * 						for x in xs             # <<<<<<<<<<<<<<
 * 						for pt in [
 * 							point2D(x, CI(curvebyxmax[x])),
 */
                  }
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF((PyObject *)__pyx_9genexpr75__pyx_v_pt); __pyx_9genexpr75__pyx_v_pt = 0;
                  __Pyx_XDECREF(__pyx_9genexpr75__pyx_v_x); __pyx_9genexpr75__pyx_v_x = 0;
                  goto __pyx_L79_exit_scope;
                  __pyx_L74_error:;
                  __Pyx_XDECREF((PyObject *)__pyx_9genexpr75__pyx_v_pt); __pyx_9genexpr75__pyx_v_pt = 0;
                  __Pyx_XDECREF(__pyx_9genexpr75__pyx_v_x); __pyx_9genexpr75__pyx_v_x = 0;
                  goto __pyx_L13_error;
                  __pyx_L79_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_CICurve, ((PyObject*)__pyx_t_8));
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1189
 * 						]
 * 					]
 * 					plt.fill_between(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 */
                __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_fill_between); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1189, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1190
 * 					]
 * 					plt.fill_between(
 * 						[ pt.x for pt in meanCurve ],             # <<<<<<<<<<<<<<
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],
 */
                  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1190, __pyx_L82_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __pyx_t_19 = __pyx_v_meanCurve; __Pyx_INCREF(__pyx_t_19); __pyx_t_15 = 0;
                  for (;;) {
                    if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_19)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_19, __pyx_t_15); __Pyx_INCREF(__pyx_t_6); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1190, __pyx_L82_error)
                    #else
                    __pyx_t_6 = PySequence_ITEM(__pyx_t_19, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1190, __pyx_L82_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr76__pyx_v_pt, __pyx_t_6);
                    __pyx_t_6 = 0;
                    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr76__pyx_v_pt, __pyx_n_s_x); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1190, __pyx_L82_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1190, __pyx_L82_error)
                    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr76__pyx_v_pt); __pyx_9genexpr76__pyx_v_pt = 0;
                  goto __pyx_L85_exit_scope;
                  __pyx_L82_error:;
                  __Pyx_XDECREF(__pyx_9genexpr76__pyx_v_pt); __pyx_9genexpr76__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L85_exit_scope:;
                } /* exit inner scope */
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1191
 * 					plt.fill_between(
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],             # <<<<<<<<<<<<<<
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						alpha=.3)
 */
                  __pyx_t_19 = PyList_New(0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1191, __pyx_L88_error)
                  __Pyx_GOTREF(__pyx_t_19);
                  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1191, __pyx_L88_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_INCREF(__pyx_v_meanCurve);
                  __Pyx_GIVEREF(__pyx_v_meanCurve);
                  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_meanCurve);
                  __Pyx_INCREF(__pyx_v_CICurve);
                  __Pyx_GIVEREF(__pyx_v_CICurve);
                  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_CICurve);
                  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_6, NULL); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1191, __pyx_L88_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  if (likely(PyList_CheckExact(__pyx_t_17)) || PyTuple_CheckExact(__pyx_t_17)) {
                    __pyx_t_6 = __pyx_t_17; __Pyx_INCREF(__pyx_t_6); __pyx_t_15 = 0;
                    __pyx_t_16 = NULL;
                  } else {
                    __pyx_t_15 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1191, __pyx_L88_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __pyx_t_16 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1191, __pyx_L88_error)
                  }
                  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  for (;;) {
                    if (likely(!__pyx_t_16)) {
                      if (likely(PyList_CheckExact(__pyx_t_6))) {
                        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_6)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_17 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_15); __Pyx_INCREF(__pyx_t_17); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1191, __pyx_L88_error)
                        #else
                        __pyx_t_17 = PySequence_ITEM(__pyx_t_6, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1191, __pyx_L88_error)
                        __Pyx_GOTREF(__pyx_t_17);
                        #endif
                      } else {
                        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_17 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_15); __Pyx_INCREF(__pyx_t_17); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1191, __pyx_L88_error)
                        #else
                        __pyx_t_17 = PySequence_ITEM(__pyx_t_6, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1191, __pyx_L88_error)
                        __Pyx_GOTREF(__pyx_t_17);
                        #endif
                      }
                    } else {
                      __pyx_t_17 = __pyx_t_16(__pyx_t_6);
                      if (unlikely(!__pyx_t_17)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 1191, __pyx_L88_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_17);
                    }
                    if ((likely(PyTuple_CheckExact(__pyx_t_17))) || (PyList_CheckExact(__pyx_t_17))) {
                      PyObject* sequence = __pyx_t_17;
                      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                      if (unlikely(size != 2)) {
                        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                        __PYX_ERR(0, 1191, __pyx_L88_error)
                      }
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      if (likely(PyTuple_CheckExact(sequence))) {
                        __pyx_t_20 = PyTuple_GET_ITEM(sequence, 0); 
                        __pyx_t_24 = PyTuple_GET_ITEM(sequence, 1); 
                      } else {
                        __pyx_t_20 = PyList_GET_ITEM(sequence, 0); 
                        __pyx_t_24 = PyList_GET_ITEM(sequence, 1); 
                      }
                      __Pyx_INCREF(__pyx_t_20);
                      __Pyx_INCREF(__pyx_t_24);
                      #else
                      __pyx_t_20 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1191, __pyx_L88_error)
                      __Pyx_GOTREF(__pyx_t_20);
                      __pyx_t_24 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1191, __pyx_L88_error)
                      __Pyx_GOTREF(__pyx_t_24);
                      #endif
                      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                    } else {
                      Py_ssize_t index = -1;
                      __pyx_t_23 = PyObject_GetIter(__pyx_t_17); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1191, __pyx_L88_error)
                      __Pyx_GOTREF(__pyx_t_23);
                      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                      __pyx_t_26 = Py_TYPE(__pyx_t_23)->tp_iternext;
                      index = 0; __pyx_t_20 = __pyx_t_26(__pyx_t_23); if (unlikely(!__pyx_t_20)) goto __pyx_L91_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_20);
                      index = 1; __pyx_t_24 = __pyx_t_26(__pyx_t_23); if (unlikely(!__pyx_t_24)) goto __pyx_L91_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_24);
                      if (__Pyx_IternextUnpackEndCheck(__pyx_t_26(__pyx_t_23), 2) < 0) __PYX_ERR(0, 1191, __pyx_L88_error)
                      __pyx_t_26 = NULL;
                      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                      goto __pyx_L92_unpacking_done;
                      __pyx_L91_unpacking_failed:;
                      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                      __pyx_t_26 = NULL;
                      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                      __PYX_ERR(0, 1191, __pyx_L88_error)
                      __pyx_L92_unpacking_done:;
                    }
                    __Pyx_XDECREF_SET(__pyx_9genexpr77__pyx_v_pt, __pyx_t_20);
                    __pyx_t_20 = 0;
                    __Pyx_XDECREF_SET(__pyx_9genexpr77__pyx_v_ci, __pyx_t_24);
                    __pyx_t_24 = 0;
                    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr77__pyx_v_pt, __pyx_n_s_y); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1191, __pyx_L88_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr77__pyx_v_ci, __pyx_n_s_y); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1191, __pyx_L88_error)
                    __Pyx_GOTREF(__pyx_t_24);
                    __pyx_t_20 = PyNumber_Subtract(__pyx_t_17, __pyx_t_24); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1191, __pyx_L88_error)
                    __Pyx_GOTREF(__pyx_t_20);
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_19, (PyObject*)__pyx_t_20))) __PYX_ERR(0, 1191, __pyx_L88_error)
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr77__pyx_v_ci); __pyx_9genexpr77__pyx_v_ci = 0;
                  __Pyx_XDECREF(__pyx_9genexpr77__pyx_v_pt); __pyx_9genexpr77__pyx_v_pt = 0;
                  goto __pyx_L93_exit_scope;
                  __pyx_L88_error:;
                  __Pyx_XDECREF(__pyx_9genexpr77__pyx_v_ci); __pyx_9genexpr77__pyx_v_ci = 0;
                  __Pyx_XDECREF(__pyx_9genexpr77__pyx_v_pt); __pyx_9genexpr77__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L93_exit_scope:;
                } /* exit inner scope */
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1192
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],             # <<<<<<<<<<<<<<
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 */
                  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1192, __pyx_L96_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1192, __pyx_L96_error)
                  __Pyx_GOTREF(__pyx_t_20);
                  __Pyx_INCREF(__pyx_v_meanCurve);
                  __Pyx_GIVEREF(__pyx_v_meanCurve);
                  PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_v_meanCurve);
                  __Pyx_INCREF(__pyx_v_CICurve);
                  __Pyx_GIVEREF(__pyx_v_CICurve);
                  PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_v_CICurve);
                  __pyx_t_24 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_20, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1192, __pyx_L96_error)
                  __Pyx_GOTREF(__pyx_t_24);
                  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  if (likely(PyList_CheckExact(__pyx_t_24)) || PyTuple_CheckExact(__pyx_t_24)) {
                    __pyx_t_20 = __pyx_t_24; __Pyx_INCREF(__pyx_t_20); __pyx_t_15 = 0;
                    __pyx_t_16 = NULL;
                  } else {
                    __pyx_t_15 = -1; __pyx_t_20 = PyObject_GetIter(__pyx_t_24); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1192, __pyx_L96_error)
                    __Pyx_GOTREF(__pyx_t_20);
                    __pyx_t_16 = Py_TYPE(__pyx_t_20)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1192, __pyx_L96_error)
                  }
                  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                  for (;;) {
                    if (likely(!__pyx_t_16)) {
                      if (likely(PyList_CheckExact(__pyx_t_20))) {
                        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_20)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_24 = PyList_GET_ITEM(__pyx_t_20, __pyx_t_15); __Pyx_INCREF(__pyx_t_24); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1192, __pyx_L96_error)
                        #else
                        __pyx_t_24 = PySequence_ITEM(__pyx_t_20, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1192, __pyx_L96_error)
                        __Pyx_GOTREF(__pyx_t_24);
                        #endif
                      } else {
                        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_20)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_24 = PyTuple_GET_ITEM(__pyx_t_20, __pyx_t_15); __Pyx_INCREF(__pyx_t_24); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1192, __pyx_L96_error)
                        #else
                        __pyx_t_24 = PySequence_ITEM(__pyx_t_20, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1192, __pyx_L96_error)
                        __Pyx_GOTREF(__pyx_t_24);
                        #endif
                      }
                    } else {
                      __pyx_t_24 = __pyx_t_16(__pyx_t_20);
                      if (unlikely(!__pyx_t_24)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 1192, __pyx_L96_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_24);
                    }
                    if ((likely(PyTuple_CheckExact(__pyx_t_24))) || (PyList_CheckExact(__pyx_t_24))) {
                      PyObject* sequence = __pyx_t_24;
                      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                      if (unlikely(size != 2)) {
                        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                        __PYX_ERR(0, 1192, __pyx_L96_error)
                      }
                      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                      if (likely(PyTuple_CheckExact(sequence))) {
                        __pyx_t_17 = PyTuple_GET_ITEM(sequence, 0); 
                        __pyx_t_23 = PyTuple_GET_ITEM(sequence, 1); 
                      } else {
                        __pyx_t_17 = PyList_GET_ITEM(sequence, 0); 
                        __pyx_t_23 = PyList_GET_ITEM(sequence, 1); 
                      }
                      __Pyx_INCREF(__pyx_t_17);
                      __Pyx_INCREF(__pyx_t_23);
                      #else
                      __pyx_t_17 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1192, __pyx_L96_error)
                      __Pyx_GOTREF(__pyx_t_17);
                      __pyx_t_23 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1192, __pyx_L96_error)
                      __Pyx_GOTREF(__pyx_t_23);
                      #endif
                      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                    } else {
                      Py_ssize_t index = -1;
                      __pyx_t_27 = PyObject_GetIter(__pyx_t_24); if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 1192, __pyx_L96_error)
                      __Pyx_GOTREF(__pyx_t_27);
                      __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                      __pyx_t_26 = Py_TYPE(__pyx_t_27)->tp_iternext;
                      index = 0; __pyx_t_17 = __pyx_t_26(__pyx_t_27); if (unlikely(!__pyx_t_17)) goto __pyx_L99_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_17);
                      index = 1; __pyx_t_23 = __pyx_t_26(__pyx_t_27); if (unlikely(!__pyx_t_23)) goto __pyx_L99_unpacking_failed;
                      __Pyx_GOTREF(__pyx_t_23);
                      if (__Pyx_IternextUnpackEndCheck(__pyx_t_26(__pyx_t_27), 2) < 0) __PYX_ERR(0, 1192, __pyx_L96_error)
                      __pyx_t_26 = NULL;
                      __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
                      goto __pyx_L100_unpacking_done;
                      __pyx_L99_unpacking_failed:;
                      __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
                      __pyx_t_26 = NULL;
                      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                      __PYX_ERR(0, 1192, __pyx_L96_error)
                      __pyx_L100_unpacking_done:;
                    }
                    __Pyx_XDECREF_SET(__pyx_9genexpr78__pyx_v_pt, __pyx_t_17);
                    __pyx_t_17 = 0;
                    __Pyx_XDECREF_SET(__pyx_9genexpr78__pyx_v_ci, __pyx_t_23);
                    __pyx_t_23 = 0;
                    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr78__pyx_v_pt, __pyx_n_s_y); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1192, __pyx_L96_error)
                    __Pyx_GOTREF(__pyx_t_24);
                    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr78__pyx_v_ci, __pyx_n_s_y); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1192, __pyx_L96_error)
                    __Pyx_GOTREF(__pyx_t_23);
                    __pyx_t_17 = PyNumber_Add(__pyx_t_24, __pyx_t_23); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1192, __pyx_L96_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
                    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_17))) __PYX_ERR(0, 1192, __pyx_L96_error)
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr78__pyx_v_ci); __pyx_9genexpr78__pyx_v_ci = 0;
                  __Pyx_XDECREF(__pyx_9genexpr78__pyx_v_pt); __pyx_9genexpr78__pyx_v_pt = 0;
                  goto __pyx_L101_exit_scope;
                  __pyx_L96_error:;
                  __Pyx_XDECREF(__pyx_9genexpr78__pyx_v_ci); __pyx_9genexpr78__pyx_v_ci = 0;
                  __Pyx_XDECREF(__pyx_9genexpr78__pyx_v_pt); __pyx_9genexpr78__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L101_exit_scope:;
                } /* exit inner scope */

                /* "gnocis/models.pyx":1189
 * 						]
 * 					]
 * 					plt.fill_between(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 */
                __pyx_t_20 = PyTuple_New(3); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1189, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_19);
                PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_19);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_20, 2, __pyx_t_6);
                __pyx_t_4 = 0;
                __pyx_t_19 = 0;
                __pyx_t_6 = 0;

                /* "gnocis/models.pyx":1193
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						alpha=.3)             # <<<<<<<<<<<<<<
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(
 */
                __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1193, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_alpha, __pyx_float__3) < 0) __PYX_ERR(0, 1193, __pyx_L13_error)

                /* "gnocis/models.pyx":1189
 * 						]
 * 					]
 * 					plt.fill_between(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y - ci.y for pt, ci in zip(meanCurve, CICurve) ],
 */
                __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_20, __pyx_t_6); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1189, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                /* "gnocis/models.pyx":1194
 * 						[ pt.y + ci.y for pt, ci in zip(meanCurve, CICurve) ],
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]             # <<<<<<<<<<<<<<
 * 					plt.plot(
 * 						[ pt.x for pt in meanCurve ],
 */
                { /* enter inner scope */
                  __pyx_t_19 = PyList_New(0); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1194, __pyx_L104_error)
                  __Pyx_GOTREF(__pyx_t_19);
                  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_curves)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_curves)) {
                    __pyx_t_6 = __pyx_cur_scope->__pyx_v_curves; __Pyx_INCREF(__pyx_t_6); __pyx_t_15 = 0;
                    __pyx_t_16 = NULL;
                  } else {
                    __pyx_t_15 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_curves); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1194, __pyx_L104_error)
                    __Pyx_GOTREF(__pyx_t_6);
                    __pyx_t_16 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1194, __pyx_L104_error)
                  }
                  for (;;) {
                    if (likely(!__pyx_t_16)) {
                      if (likely(PyList_CheckExact(__pyx_t_6))) {
                        if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_6)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_20 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_15); __Pyx_INCREF(__pyx_t_20); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1194, __pyx_L104_error)
                        #else
                        __pyx_t_20 = PySequence_ITEM(__pyx_t_6, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1194, __pyx_L104_error)
                        __Pyx_GOTREF(__pyx_t_20);
                        #endif
                      } else {
                        if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
                        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                        __pyx_t_20 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_15); __Pyx_INCREF(__pyx_t_20); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1194, __pyx_L104_error)
                        #else
                        __pyx_t_20 = PySequence_ITEM(__pyx_t_6, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1194, __pyx_L104_error)
                        __Pyx_GOTREF(__pyx_t_20);
                        #endif
                      }
                    } else {
                      __pyx_t_20 = __pyx_t_16(__pyx_t_6);
                      if (unlikely(!__pyx_t_20)) {
                        PyObject* exc_type = PyErr_Occurred();
                        if (exc_type) {
                          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                          else __PYX_ERR(0, 1194, __pyx_L104_error)
                        }
                        break;
                      }
                      __Pyx_GOTREF(__pyx_t_20);
                    }
                    __Pyx_XDECREF_SET(__pyx_9genexpr79__pyx_v_curve, __pyx_t_20);
                    __pyx_t_20 = 0;
                    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1194, __pyx_L104_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __pyx_t_4 = NULL;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
                      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
                      if (likely(__pyx_t_4)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                        __Pyx_INCREF(__pyx_t_4);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_8, function);
                      }
                    }
                    __pyx_t_20 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_4, __pyx_9genexpr79__pyx_v_curve) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_9genexpr79__pyx_v_curve);
                    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                    if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1194, __pyx_L104_error)
                    __Pyx_GOTREF(__pyx_t_20);
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __pyx_t_8 = PyNumber_Multiply(__pyx_t_20, __pyx_float_100_); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1194, __pyx_L104_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_19, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 1194, __pyx_L104_error)
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr79__pyx_v_curve); __pyx_9genexpr79__pyx_v_curve = 0;
                  goto __pyx_L107_exit_scope;
                  __pyx_L104_error:;
                  __Pyx_XDECREF(__pyx_9genexpr79__pyx_v_curve); __pyx_9genexpr79__pyx_v_curve = 0;
                  goto __pyx_L13_error;
                  __pyx_L107_exit_scope:;
                } /* exit inner scope */
                __Pyx_XDECREF_SET(__pyx_v_AUCs, ((PyObject*)__pyx_t_19));
                __pyx_t_19 = 0;

                /* "gnocis/models.pyx":1195
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],
 */
                __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_plot); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1195, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_19);
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1196
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(
 * 						[ pt.x for pt in meanCurve ],             # <<<<<<<<<<<<<<
 * 						[ pt.y for pt in meanCurve ],
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))
 */
                  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1196, __pyx_L110_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __pyx_t_8 = __pyx_v_meanCurve; __Pyx_INCREF(__pyx_t_8); __pyx_t_15 = 0;
                  for (;;) {
                    if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_8)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_20 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_15); __Pyx_INCREF(__pyx_t_20); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1196, __pyx_L110_error)
                    #else
                    __pyx_t_20 = PySequence_ITEM(__pyx_t_8, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1196, __pyx_L110_error)
                    __Pyx_GOTREF(__pyx_t_20);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr80__pyx_v_pt, __pyx_t_20);
                    __pyx_t_20 = 0;
                    __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr80__pyx_v_pt, __pyx_n_s_x); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1196, __pyx_L110_error)
                    __Pyx_GOTREF(__pyx_t_20);
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_20))) __PYX_ERR(0, 1196, __pyx_L110_error)
                    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr80__pyx_v_pt); __pyx_9genexpr80__pyx_v_pt = 0;
                  goto __pyx_L113_exit_scope;
                  __pyx_L110_error:;
                  __Pyx_XDECREF(__pyx_9genexpr80__pyx_v_pt); __pyx_9genexpr80__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L113_exit_scope:;
                } /* exit inner scope */
                { /* enter inner scope */

                  /* "gnocis/models.pyx":1197
 * 					plt.plot(
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],             # <<<<<<<<<<<<<<
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 */
                  __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1197, __pyx_L116_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __pyx_t_20 = __pyx_v_meanCurve; __Pyx_INCREF(__pyx_t_20); __pyx_t_15 = 0;
                  for (;;) {
                    if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_20)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_20, __pyx_t_15); __Pyx_INCREF(__pyx_t_4); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(0, 1197, __pyx_L116_error)
                    #else
                    __pyx_t_4 = PySequence_ITEM(__pyx_t_20, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1197, __pyx_L116_error)
                    __Pyx_GOTREF(__pyx_t_4);
                    #endif
                    __Pyx_XDECREF_SET(__pyx_9genexpr81__pyx_v_pt, __pyx_t_4);
                    __pyx_t_4 = 0;
                    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr81__pyx_v_pt, __pyx_n_s_y); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1197, __pyx_L116_error)
                    __Pyx_GOTREF(__pyx_t_4);
                    if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1197, __pyx_L116_error)
                    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_XDECREF(__pyx_9genexpr81__pyx_v_pt); __pyx_9genexpr81__pyx_v_pt = 0;
                  goto __pyx_L119_exit_scope;
                  __pyx_L116_error:;
                  __Pyx_XDECREF(__pyx_9genexpr81__pyx_v_pt); __pyx_9genexpr81__pyx_v_pt = 0;
                  goto __pyx_L13_error;
                  __pyx_L119_exit_scope:;
                } /* exit inner scope */

                /* "gnocis/models.pyx":1195
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],
 */
                __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1195, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_GIVEREF(__pyx_t_6);
                PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_6);
                __Pyx_GIVEREF(__pyx_t_8);
                PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_8);
                __pyx_t_6 = 0;
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1198
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))             # <<<<<<<<<<<<<<
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 */
                __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_6 = PyTuple_New(6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_15 = 0;
                __pyx_t_28 = 127;
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_mdl, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_17 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_28 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) > __pyx_t_28) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) : __pyx_t_28;
                __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_17);
                __Pyx_GIVEREF(__pyx_t_17);
                PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_17);
                __pyx_t_17 = 0;
                __Pyx_INCREF(__pyx_kp_u_AUC_2);
                __pyx_t_15 += 8;
                __Pyx_GIVEREF(__pyx_kp_u_AUC_2);
                PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_kp_u_AUC_2);
                __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_mean); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __pyx_t_23 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_23)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_23);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                  }
                }
                __pyx_t_17 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_23, __pyx_v_AUCs) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_AUCs);
                __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
                if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_4 = __Pyx_PyObject_Format(__pyx_t_17, __pyx_kp_u_2f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_28 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_28) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_28;
                __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_4);
                __pyx_t_4 = 0;
                __Pyx_INCREF(__pyx_kp_u__17);
                __pyx_t_15 += 5;
                __Pyx_GIVEREF(__pyx_kp_u__17);
                PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_kp_u__17);
                __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_CI); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __pyx_t_23 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
                  __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_17);
                  if (likely(__pyx_t_23)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                    __Pyx_INCREF(__pyx_t_23);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_17, function);
                  }
                }
                __pyx_t_4 = (__pyx_t_23) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_23, __pyx_v_AUCs) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_v_AUCs);
                __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
                if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_17 = __Pyx_PyObject_Format(__pyx_t_4, __pyx_kp_u_2f); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_28 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) > __pyx_t_28) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) : __pyx_t_28;
                __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_17);
                __Pyx_GIVEREF(__pyx_t_17);
                PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_17);
                __pyx_t_17 = 0;
                __Pyx_INCREF(__pyx_kp_u__7);
                __pyx_t_15 += 2;
                __Pyx_GIVEREF(__pyx_kp_u__7);
                PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_kp_u__7);
                __pyx_t_17 = __Pyx_PyUnicode_Join(__pyx_t_6, 6, __pyx_t_15, __pyx_t_28); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_label, __pyx_t_17) < 0) __PYX_ERR(0, 1198, __pyx_L13_error)
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

                /* "gnocis/models.pyx":1195
 * 						alpha=.3)
 * 					AUCs = [ getAUC(curve) * 100. for curve in curves ]
 * 					plt.plot(             # <<<<<<<<<<<<<<
 * 						[ pt.x for pt in meanCurve ],
 * 						[ pt.y for pt in meanCurve ],
 */
                __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_20, __pyx_t_8); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1195, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

                /* "gnocis/models.pyx":1149
 * 					label = 'Expected at random')
 * 				# Curves per model
 * 				for mdl in self.models:             # <<<<<<<<<<<<<<
 * 					curves = self.ROC[mdl]
 * 					xs = sorted(list(set(pt.x for c in curves for pt in c)))
 */
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "gnocis/models.pyx":1199
 * 						[ pt.y for pt in meanCurve ],
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)             # <<<<<<<<<<<<<<
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.xlabel('False Positive Rate', fontsize = fontsizeLabels)
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_xticks); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1199, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_17 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1199, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_fontsize, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 1199, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_rotation, __pyx_int_0) < 0) __PYX_ERR(0, 1199, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1199, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1200
 * 						label = '%s; AUC = %.2f +/- %.2f %%'%(mdl.name, mean(AUCs), CI(AUCs)))
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)             # <<<<<<<<<<<<<<
 * 				plt.xlabel('False Positive Rate', fontsize = fontsizeLabels)
 * 				plt.ylabel('True Positive Rate', fontsize = fontsizeLabels)
 */
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_yticks); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1200, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_17 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1200, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_fontsize, __pyx_v_fontsizeAxis) < 0) __PYX_ERR(0, 1200, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_rotation, __pyx_int_0) < 0) __PYX_ERR(0, 1200, __pyx_L13_error)
              __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_empty_tuple, __pyx_t_17); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1200, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "gnocis/models.pyx":1201
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.xlabel('False Positive Rate', fontsize = fontsizeLabels)             # <<<<<<<<<<<<<<
 * 				plt.ylabel('True Positive Rate', fontsize = fontsizeLabels)
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_xlabel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1201, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_17 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1201, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_fontsize, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 1201, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__27, __pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1201, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1202
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.xlabel('False Positive Rate', fontsize = fontsizeLabels)
 * 				plt.ylabel('True Positive Rate', fontsize = fontsizeLabels)             # <<<<<<<<<<<<<<
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()
 */
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_ylabel); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1202, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_17 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1202, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_fontsize, __pyx_v_fontsizeLabels) < 0) __PYX_ERR(0, 1202, __pyx_L13_error)
              __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__28, __pyx_t_17); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1202, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "gnocis/models.pyx":1203
 * 				plt.xlabel('False Positive Rate', fontsize = fontsizeLabels)
 * 				plt.ylabel('True Positive Rate', fontsize = fontsizeLabels)
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)             # <<<<<<<<<<<<<<
 * 				fig.tight_layout()
 * 				if outpath is None:
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_legend); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1203, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_17 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1203, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_bbox_to_anchor, __pyx_v_bboxAnchorTo) < 0) __PYX_ERR(0, 1203, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_loc, __pyx_v_legendLoc) < 0) __PYX_ERR(0, 1203, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_fontsize, __pyx_v_fontsizeLegend) < 0) __PYX_ERR(0, 1203, __pyx_L13_error)
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_fancybox, Py_True) < 0) __PYX_ERR(0, 1203, __pyx_L13_error)
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1203, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1204
 * 				plt.ylabel('True Positive Rate', fontsize = fontsizeLabels)
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()             # <<<<<<<<<<<<<<
 * 				if outpath is None:
 * 					bio = BytesIO()
 */
              __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_tight_layout); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1204, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_17);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_17, function);
                }
              }
              __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_17);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1204, __pyx_L13_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "gnocis/models.pyx":1205
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()
 * 				if outpath is None:             # <<<<<<<<<<<<<<
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 */
              __pyx_t_25 = (__pyx_v_outpath == Py_None);
              __pyx_t_29 = (__pyx_t_25 != 0);
              if (__pyx_t_29) {

                /* "gnocis/models.pyx":1206
 * 				fig.tight_layout()
 * 				if outpath is None:
 * 					bio = BytesIO()             # <<<<<<<<<<<<<<
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')
 */
                __Pyx_INCREF(__pyx_v_BytesIO);
                __pyx_t_17 = __pyx_v_BytesIO; __pyx_t_5 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
                  __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_17);
                  if (likely(__pyx_t_5)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                    __Pyx_INCREF(__pyx_t_5);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_17, function);
                  }
                }
                __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_17);
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1206, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_v_bio = __pyx_t_8;
                __pyx_t_8 = 0;

                /* "gnocis/models.pyx":1207
 * 				if outpath is None:
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')             # <<<<<<<<<<<<<<
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 */
                __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_savefig); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1207, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_17 = PyTuple_New(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1207, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_INCREF(__pyx_v_bio);
                __Pyx_GIVEREF(__pyx_v_bio);
                PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_v_bio);
                __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1207, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_format, __pyx_n_u_png) < 0) __PYX_ERR(0, 1207, __pyx_L13_error)
                __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_17, __pyx_t_5); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1207, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

                /* "gnocis/models.pyx":1208
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')             # <<<<<<<<<<<<<<
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 */
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_close); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1208, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_17 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_17)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_17);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_20 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_17, __pyx_n_u_all) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_u_all);
                __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1208, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

                /* "gnocis/models.pyx":1209
 * 					fig.savefig(bio, format='png')
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')             # <<<<<<<<<<<<<<
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					if returnHTML:
 */
                __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_base64, __pyx_n_s_b64encode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1209, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_bio, __pyx_n_s_getvalue); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1209, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_19);
                __pyx_t_6 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_19))) {
                  __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_19);
                  if (likely(__pyx_t_6)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
                    __Pyx_INCREF(__pyx_t_6);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_19, function);
                  }
                }
                __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_19, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_19);
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1209, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __pyx_t_19 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
                  __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_17);
                  if (likely(__pyx_t_19)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                    __Pyx_INCREF(__pyx_t_19);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_17, function);
                  }
                }
                __pyx_t_5 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_19, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_8);
                __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1209, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_decode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1209, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __pyx_t_5 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
                  __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_17);
                  if (likely(__pyx_t_5)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                    __Pyx_INCREF(__pyx_t_5);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_17, function);
                  }
                }
                __pyx_t_20 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_5, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_kp_u_utf_8);
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1209, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_v_encoded = __pyx_t_20;
                __pyx_t_20 = 0;

                /* "gnocis/models.pyx":1210
 * 					plt.close('all')
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded             # <<<<<<<<<<<<<<
 * 					if returnHTML:
 * 						return html
 */
                __pyx_t_20 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_img_src_data_image_png_base64_s, __pyx_v_encoded); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1210, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __pyx_v_html = ((PyObject*)__pyx_t_20);
                __pyx_t_20 = 0;

                /* "gnocis/models.pyx":1211
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					if returnHTML:             # <<<<<<<<<<<<<<
 * 						return html
 * 					display(HTML(html))
 */
                __pyx_t_29 = __Pyx_PyObject_IsTrue(__pyx_v_returnHTML); if (unlikely(__pyx_t_29 < 0)) __PYX_ERR(0, 1211, __pyx_L13_error)
                if (__pyx_t_29) {

                  /* "gnocis/models.pyx":1212
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					if returnHTML:
 * 						return html             # <<<<<<<<<<<<<<
 * 					display(HTML(html))
 * 				else:
 */
                  __Pyx_XDECREF(__pyx_r);
                  __Pyx_INCREF(__pyx_v_html);
                  __pyx_r = __pyx_v_html;
                  goto __pyx_L17_try_return;

                  /* "gnocis/models.pyx":1211
 * 					encoded = base64.b64encode(bio.getvalue()).decode('utf-8')
 * 					html = '<img src=\'data:image/png;base64,%s\'>'%encoded
 * 					if returnHTML:             # <<<<<<<<<<<<<<
 * 						return html
 * 					display(HTML(html))
 */
                }

                /* "gnocis/models.pyx":1213
 * 					if returnHTML:
 * 						return html
 * 					display(HTML(html))             # <<<<<<<<<<<<<<
 * 				else:
 * 					fig.savefig(outpath)
 */
                __Pyx_INCREF(__pyx_v_HTML);
                __pyx_t_5 = __pyx_v_HTML; __pyx_t_8 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_8)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_8);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_17 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_v_html) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_html);
                __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1213, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_17);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_INCREF(__pyx_v_display);
                __pyx_t_5 = __pyx_v_display; __pyx_t_8 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_8)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_8);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_20 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_17);
                __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1213, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

                /* "gnocis/models.pyx":1205
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()
 * 				if outpath is None:             # <<<<<<<<<<<<<<
 * 					bio = BytesIO()
 * 					fig.savefig(bio, format='png')
 */
                goto __pyx_L120;
              }

              /* "gnocis/models.pyx":1215
 * 					display(HTML(html))
 * 				else:
 * 					fig.savefig(outpath)             # <<<<<<<<<<<<<<
 * 					plt.close('all')
 * 		except ImportError as err:
 */
              /*else*/ {
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_fig, __pyx_n_s_savefig); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1215, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_17 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_17)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_17);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_20 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_17, __pyx_v_outpath) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_outpath);
                __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1215, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

                /* "gnocis/models.pyx":1216
 * 				else:
 * 					fig.savefig(outpath)
 * 					plt.close('all')             # <<<<<<<<<<<<<<
 * 		except ImportError as err:
 * 			raise err
 */
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_plt, __pyx_n_s_close); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1216, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_17 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                  __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_5);
                  if (likely(__pyx_t_17)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                    __Pyx_INCREF(__pyx_t_17);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_5, function);
                  }
                }
                __pyx_t_20 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_17, __pyx_n_u_all) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_u_all);
                __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1216, __pyx_L13_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              }
              __pyx_L120:;

              /* "gnocis/models.pyx":1139
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 				fig = plt.figure(figsize = figsize)
 * 				# Expected random generalization
 */
            }
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            goto __pyx_L18_try_end;
            __pyx_L13_error:;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
            __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
            __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("gnocis.models.crossvalidation.plotROC", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_20, &__pyx_t_5, &__pyx_t_17) < 0) __PYX_ERR(0, 1139, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_8 = PyTuple_Pack(3, __pyx_t_20, __pyx_t_5, __pyx_t_17); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1139, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_30 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1139, __pyx_L15_except_error)
              __Pyx_GOTREF(__pyx_t_30);
              __pyx_t_29 = __Pyx_PyObject_IsTrue(__pyx_t_30);
              __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
              if (__pyx_t_29 < 0) __PYX_ERR(0, 1139, __pyx_L15_except_error)
              __pyx_t_25 = ((!(__pyx_t_29 != 0)) != 0);
              if (__pyx_t_25) {
                __Pyx_GIVEREF(__pyx_t_20);
                __Pyx_GIVEREF(__pyx_t_5);
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_ErrRestoreWithState(__pyx_t_20, __pyx_t_5, __pyx_t_17);
                __pyx_t_20 = 0; __pyx_t_5 = 0; __pyx_t_17 = 0; 
                __PYX_ERR(0, 1139, __pyx_L15_except_error)
              }
              __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              goto __pyx_L14_exception_handled;
            }
            __pyx_L15_except_error:;
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
            goto __pyx_L3_error;
            __pyx_L17_try_return:;
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
            goto __pyx_L10_return;
            __pyx_L14_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
            __pyx_L18_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_7) {
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__10, NULL);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1139, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            }
            goto __pyx_L12;
          }
          __pyx_L10_return: {
            __pyx_t_11 = __pyx_r;
            __pyx_r = 0;
            if (__pyx_t_7) {
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__10, NULL);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1139, __pyx_L3_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
            __pyx_r = __pyx_t_11;
            __pyx_t_11 = 0;
            goto __pyx_L7_try_return;
          }
          __pyx_L12:;
        }
        goto __pyx_L125;
        __pyx_L9_error:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L3_error;
        __pyx_L125:;
      }

      /* "gnocis/models.pyx":1134
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "gnocis/models.pyx":1217
 * 					fig.savefig(outpath)
 * 					plt.close('all')
 * 		except ImportError as err:             # <<<<<<<<<<<<<<
 * 			raise err
 * 
 */
    __pyx_t_31 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_31) {
      __Pyx_AddTraceback("gnocis.models.crossvalidation.plotROC", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_17, &__pyx_t_5, &__pyx_t_20) < 0) __PYX_ERR(0, 1217, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_err = __pyx_t_5;
      /*try:*/ {

        /* "gnocis/models.pyx":1218
 * 					plt.close('all')
 * 		except ImportError as err:
 * 			raise err             # <<<<<<<<<<<<<<
 * 
 * 	def getConfigurationTable(self):
 */
        __Pyx_Raise(__pyx_v_err, 0, 0, 0);
        __PYX_ERR(0, 1218, __pyx_L131_error)
      }

      /* "gnocis/models.pyx":1217
 * 					fig.savefig(outpath)
 * 					plt.close('all')
 * 		except ImportError as err:             # <<<<<<<<<<<<<<
 * 			raise err
 * 
 */
      /*finally:*/ {
        __pyx_L131_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_7 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_30 = 0; __pyx_t_34 = 0;
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
          __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
          __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_9, &__pyx_t_30, &__pyx_t_34);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_11, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_11, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_30);
          __Pyx_XGOTREF(__pyx_t_34);
          __pyx_t_31 = __pyx_lineno; __pyx_t_32 = __pyx_clineno; __pyx_t_33 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err);
            __pyx_v_err = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_XGIVEREF(__pyx_t_30);
            __Pyx_XGIVEREF(__pyx_t_34);
            __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_30, __pyx_t_34);
          }
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ErrRestore(__pyx_t_7, __pyx_t_11, __pyx_t_10);
          __pyx_t_7 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_30 = 0; __pyx_t_34 = 0;
          __pyx_lineno = __pyx_t_31; __pyx_clineno = __pyx_t_32; __pyx_filename = __pyx_t_33;
          goto __pyx_L5_except_error;
        }
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "gnocis/models.pyx":1134
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):
 * 		try:             # <<<<<<<<<<<<<<
 * 			import matplotlib.pyplot as plt
 * 			import base64
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "gnocis/models.pyx":1133
 * 			raise err
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.plotROC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_plt);
  __Pyx_XDECREF(__pyx_v_base64);
  __Pyx_XDECREF(__pyx_v_BytesIO);
  __Pyx_XDECREF(__pyx_v_display);
  __Pyx_XDECREF(__pyx_v_HTML);
  __Pyx_XDECREF(__pyx_v_fig);
  __Pyx_XDECREF(__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_v_xs);
  __Pyx_XDECREF(__pyx_v_curvebyx);
  __Pyx_XDECREF(__pyx_v_curvebyxmax);
  __Pyx_XDECREF(__pyx_v_curvebyxmin);
  __Pyx_XDECREF(__pyx_v_meanCurve);
  __Pyx_XDECREF(__pyx_v_CICurve);
  __Pyx_XDECREF(__pyx_v_AUCs);
  __Pyx_XDECREF(__pyx_v_bio);
  __Pyx_XDECREF(__pyx_v_encoded);
  __Pyx_XDECREF(__pyx_v_html);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_gb_6gnocis_6models_15crossvalidation_7plotROC_2generator14);
  __Pyx_XDECREF(__pyx_9genexpr67__pyx_v_curve);
  __Pyx_XDECREF(__pyx_9genexpr68__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr69__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr70__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr71__pyx_v_i);
  __Pyx_XDECREF(__pyx_9genexpr71__pyx_v_curve);
  __Pyx_XDECREF(__pyx_9genexpr72__pyx_v_x);
  __Pyx_XDECREF(__pyx_9genexpr73__pyx_v_i);
  __Pyx_XDECREF(__pyx_9genexpr73__pyx_v_curve);
  __Pyx_XDECREF(__pyx_9genexpr74__pyx_v_x);
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr74__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr75__pyx_v_x);
  __Pyx_XDECREF((PyObject *)__pyx_9genexpr75__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr76__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr77__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr77__pyx_v_ci);
  __Pyx_XDECREF(__pyx_9genexpr78__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr78__pyx_v_ci);
  __Pyx_XDECREF(__pyx_9genexpr79__pyx_v_curve);
  __Pyx_XDECREF(__pyx_9genexpr80__pyx_v_pt);
  __Pyx_XDECREF(__pyx_9genexpr81__pyx_v_pt);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1220
 * 			raise err
 * 
 * 	def getConfigurationTable(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Configuration',
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_13getConfigurationTable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_13getConfigurationTable = {"getConfigurationTable", (PyCFunction)__pyx_pw_6gnocis_6models_15crossvalidation_13getConfigurationTable, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_13getConfigurationTable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getConfigurationTable (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_12getConfigurationTable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_12getConfigurationTable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getConfigurationTable", 0);

  /* "gnocis/models.pyx":1221
 * 
 * 	def getConfigurationTable(self):
 * 		return nctable(             # <<<<<<<<<<<<<<
 * 			'Configuration',
 * 			{
 */
  __Pyx_XDECREF(__pyx_r);

  /* "gnocis/models.pyx":1232
 * 				'Train/test ratio:': [ str(self.ratioTrainTest) ],
 * 			}
 * 		).noCropNames()             # <<<<<<<<<<<<<<
 * 
 * 	def getAUCTable(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_nctable); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gnocis/models.pyx":1224
 * 			'Configuration',
 * 			{
 * 				'Training set:': [ str(self.trainingSet) ],             # <<<<<<<<<<<<<<
 * 				'Validation set:': [ str(self.validationSet) ],
 * 				'Training sequences per repeat': [ len(self.cvtrain[0]) ],
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trainingSet); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Training_set_3, __pyx_t_5) < 0) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":1225
 * 			{
 * 				'Training set:': [ str(self.trainingSet) ],
 * 				'Validation set:': [ str(self.validationSet) ],             # <<<<<<<<<<<<<<
 * 				'Training sequences per repeat': [ len(self.cvtrain[0]) ],
 * 				'Validation sequences per repeat': [ len(self.cvval[0]) ],
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_validationSet); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Validation_set_2, __pyx_t_5) < 0) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":1226
 * 				'Training set:': [ str(self.trainingSet) ],
 * 				'Validation set:': [ str(self.validationSet) ],
 * 				'Training sequences per repeat': [ len(self.cvtrain[0]) ],             # <<<<<<<<<<<<<<
 * 				'Validation sequences per repeat': [ len(self.cvval[0]) ],
 * 				'Repeats:': [ str(self.repeats) ],
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvtrain); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_7 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Training_sequences_per_repeat, __pyx_t_5) < 0) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":1227
 * 				'Validation set:': [ str(self.validationSet) ],
 * 				'Training sequences per repeat': [ len(self.cvtrain[0]) ],
 * 				'Validation sequences per repeat': [ len(self.cvval[0]) ],             # <<<<<<<<<<<<<<
 * 				'Repeats:': [ str(self.repeats) ],
 * 				'Negatives per positive:': [ str(self.ratioNegPos) ],
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cvval); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_7 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Validation_sequences_per_repeat, __pyx_t_5) < 0) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":1228
 * 				'Training sequences per repeat': [ len(self.cvtrain[0]) ],
 * 				'Validation sequences per repeat': [ len(self.cvval[0]) ],
 * 				'Repeats:': [ str(self.repeats) ],             # <<<<<<<<<<<<<<
 * 				'Negatives per positive:': [ str(self.ratioNegPos) ],
 * 				'Train/test ratio:': [ str(self.ratioTrainTest) ],
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_repeats); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Repeats, __pyx_t_5) < 0) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":1229
 * 				'Validation sequences per repeat': [ len(self.cvval[0]) ],
 * 				'Repeats:': [ str(self.repeats) ],
 * 				'Negatives per positive:': [ str(self.ratioNegPos) ],             # <<<<<<<<<<<<<<
 * 				'Train/test ratio:': [ str(self.ratioTrainTest) ],
 * 			}
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ratioNegPos); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Negatives_per_positive, __pyx_t_5) < 0) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":1230
 * 				'Repeats:': [ str(self.repeats) ],
 * 				'Negatives per positive:': [ str(self.ratioNegPos) ],
 * 				'Train/test ratio:': [ str(self.ratioTrainTest) ],             # <<<<<<<<<<<<<<
 * 			}
 * 		).noCropNames()
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ratioTrainTest); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_kp_u_Train_test_ratio, __pyx_t_5) < 0) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_n_u_Configuration, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_n_u_Configuration, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_n_u_Configuration);
    __Pyx_GIVEREF(__pyx_n_u_Configuration);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_8, __pyx_n_u_Configuration);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_8, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gnocis/models.pyx":1232
 * 				'Train/test ratio:': [ str(self.ratioTrainTest) ],
 * 			}
 * 		).noCropNames()             # <<<<<<<<<<<<<<
 * 
 * 	def getAUCTable(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_noCropNames); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":1220
 * 			raise err
 * 
 * 	def getConfigurationTable(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Configuration',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.getConfigurationTable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1234
 * 		).noCropNames()
 * 
 * 	def getAUCTable(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Evaluation statistics',
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_15getAUCTable(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_15getAUCTable = {"getAUCTable", (PyCFunction)__pyx_pw_6gnocis_6models_15crossvalidation_15getAUCTable, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_15getAUCTable(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getAUCTable (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_14getAUCTable(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_14getAUCTable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_9genexpr82__pyx_v_mdl = NULL;
  PyObject *__pyx_9genexpr83__pyx_v_mdl = NULL;
  PyObject *__pyx_9genexpr84__pyx_v_c = NULL;
  PyObject *__pyx_9genexpr85__pyx_v_c = NULL;
  PyObject *__pyx_9genexpr86__pyx_v_mdl = NULL;
  PyObject *__pyx_9genexpr87__pyx_v_c = NULL;
  PyObject *__pyx_9genexpr88__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getAUCTable", 0);

  /* "gnocis/models.pyx":1235
 * 
 * 	def getAUCTable(self):
 * 		return nctable(             # <<<<<<<<<<<<<<
 * 			'Evaluation statistics',
 * 			{
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nctable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gnocis/models.pyx":1238
 * 			'Evaluation statistics',
 * 			{
 * 				'Model': [ mdl.name for mdl in self.models ],             # <<<<<<<<<<<<<<
 * 				'PRC AUC': [
 * 					'%.2f +/- %.2f %%'%(
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1238, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_6 = __pyx_t_5; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1238, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1238, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1238, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1238, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_8(__pyx_t_6);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1238, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr82__pyx_v_mdl, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr82__pyx_v_mdl, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 1238, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_9genexpr82__pyx_v_mdl); __pyx_9genexpr82__pyx_v_mdl = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr82__pyx_v_mdl); __pyx_9genexpr82__pyx_v_mdl = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_Model, __pyx_t_4) < 0) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  { /* enter inner scope */

    /* "gnocis/models.pyx":1239
 * 			{
 * 				'Model': [ mdl.name for mdl in self.models ],
 * 				'PRC AUC': [             # <<<<<<<<<<<<<<
 * 					'%.2f +/- %.2f %%'%(
 * 						mean([getAUC(c) for c in self.PRC[mdl]]) * 100.,
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1239, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "gnocis/models.pyx":1244
 * 						CI([getAUC(c) for c in self.PRC[mdl]]) * 100.
 * 					)
 * 					for mdl in self.models             # <<<<<<<<<<<<<<
 * 				],
 * 				'ROC AUC': [
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1244, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1244, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1244, __pyx_L11_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1244, __pyx_L11_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1244, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1244, __pyx_L11_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1244, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_8(__pyx_t_5);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1244, __pyx_L11_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr83__pyx_v_mdl, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "gnocis/models.pyx":1240
 * 				'Model': [ mdl.name for mdl in self.models ],
 * 				'PRC AUC': [
 * 					'%.2f +/- %.2f %%'%(             # <<<<<<<<<<<<<<
 * 						mean([getAUC(c) for c in self.PRC[mdl]]) * 100.,
 * 						CI([getAUC(c) for c in self.PRC[mdl]]) * 100.
 */
      __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1240, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;

      /* "gnocis/models.pyx":1241
 * 				'PRC AUC': [
 * 					'%.2f +/- %.2f %%'%(
 * 						mean([getAUC(c) for c in self.PRC[mdl]]) * 100.,             # <<<<<<<<<<<<<<
 * 						CI([getAUC(c) for c in self.PRC[mdl]]) * 100.
 * 					)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_mean); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1241, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_12);
      { /* enter inner scope */
        __pyx_t_13 = PyList_New(0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1241, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_PRC); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1241, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = __Pyx_PyObject_GetItem(__pyx_t_14, __pyx_9genexpr83__pyx_v_mdl); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1241, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
          __pyx_t_14 = __pyx_t_15; __Pyx_INCREF(__pyx_t_14); __pyx_t_16 = 0;
          __pyx_t_17 = NULL;
        } else {
          __pyx_t_16 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1241, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_17 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1241, __pyx_L16_error)
        }
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        for (;;) {
          if (likely(!__pyx_t_17)) {
            if (likely(PyList_CheckExact(__pyx_t_14))) {
              if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_14)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_16); __Pyx_INCREF(__pyx_t_15); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1241, __pyx_L16_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_14, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1241, __pyx_L16_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            } else {
              if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_16); __Pyx_INCREF(__pyx_t_15); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1241, __pyx_L16_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_14, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1241, __pyx_L16_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            }
          } else {
            __pyx_t_15 = __pyx_t_17(__pyx_t_14);
            if (unlikely(!__pyx_t_15)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1241, __pyx_L16_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_15);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr84__pyx_v_c, __pyx_t_15);
          __pyx_t_15 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1241, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_19 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_18))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_18);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_18, function);
            }
          }
          __pyx_t_15 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_19, __pyx_9genexpr84__pyx_v_c) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_9genexpr84__pyx_v_c);
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1241, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_13, (PyObject*)__pyx_t_15))) __PYX_ERR(0, 1241, __pyx_L16_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_9genexpr84__pyx_v_c); __pyx_9genexpr84__pyx_v_c = 0;
        goto __pyx_L19_exit_scope;
        __pyx_L16_error:;
        __Pyx_XDECREF(__pyx_9genexpr84__pyx_v_c); __pyx_9genexpr84__pyx_v_c = 0;
        goto __pyx_L11_error;
        __pyx_L19_exit_scope:;
      } /* exit inner scope */
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_11 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1241, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = PyNumber_Multiply(__pyx_t_11, __pyx_float_100_); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1241, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_kp_u_2f); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1241, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_INCREF(__pyx_kp_u__17);
      __pyx_t_9 += 5;
      __Pyx_GIVEREF(__pyx_kp_u__17);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_kp_u__17);

      /* "gnocis/models.pyx":1242
 * 					'%.2f +/- %.2f %%'%(
 * 						mean([getAUC(c) for c in self.PRC[mdl]]) * 100.,
 * 						CI([getAUC(c) for c in self.PRC[mdl]]) * 100.             # <<<<<<<<<<<<<<
 * 					)
 * 					for mdl in self.models
 */
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_CI); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1242, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_12);
      { /* enter inner scope */
        __pyx_t_13 = PyList_New(0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1242, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_PRC); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1242, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = __Pyx_PyObject_GetItem(__pyx_t_14, __pyx_9genexpr83__pyx_v_mdl); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1242, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
          __pyx_t_14 = __pyx_t_15; __Pyx_INCREF(__pyx_t_14); __pyx_t_16 = 0;
          __pyx_t_17 = NULL;
        } else {
          __pyx_t_16 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1242, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_17 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1242, __pyx_L22_error)
        }
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        for (;;) {
          if (likely(!__pyx_t_17)) {
            if (likely(PyList_CheckExact(__pyx_t_14))) {
              if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_14)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_16); __Pyx_INCREF(__pyx_t_15); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1242, __pyx_L22_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_14, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1242, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            } else {
              if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_16); __Pyx_INCREF(__pyx_t_15); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1242, __pyx_L22_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_14, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1242, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            }
          } else {
            __pyx_t_15 = __pyx_t_17(__pyx_t_14);
            if (unlikely(!__pyx_t_15)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1242, __pyx_L22_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_15);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr85__pyx_v_c, __pyx_t_15);
          __pyx_t_15 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1242, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_19 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_18))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_18);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_18, function);
            }
          }
          __pyx_t_15 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_19, __pyx_9genexpr85__pyx_v_c) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_9genexpr85__pyx_v_c);
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1242, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_13, (PyObject*)__pyx_t_15))) __PYX_ERR(0, 1242, __pyx_L22_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_9genexpr85__pyx_v_c); __pyx_9genexpr85__pyx_v_c = 0;
        goto __pyx_L25_exit_scope;
        __pyx_L22_error:;
        __Pyx_XDECREF(__pyx_9genexpr85__pyx_v_c); __pyx_9genexpr85__pyx_v_c = 0;
        goto __pyx_L11_error;
        __pyx_L25_exit_scope:;
      } /* exit inner scope */
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_11 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1242, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = PyNumber_Multiply(__pyx_t_11, __pyx_float_100_); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1242, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_kp_u_2f); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1242, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_INCREF(__pyx_kp_u__7);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__7);
      PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_kp_u__7);

      /* "gnocis/models.pyx":1240
 * 				'Model': [ mdl.name for mdl in self.models ],
 * 				'PRC AUC': [
 * 					'%.2f +/- %.2f %%'%(             # <<<<<<<<<<<<<<
 * 						mean([getAUC(c) for c in self.PRC[mdl]]) * 100.,
 * 						CI([getAUC(c) for c in self.PRC[mdl]]) * 100.
 */
      __pyx_t_11 = __Pyx_PyUnicode_Join(__pyx_t_6, 4, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1240, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_11))) __PYX_ERR(0, 1239, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "gnocis/models.pyx":1244
 * 						CI([getAUC(c) for c in self.PRC[mdl]]) * 100.
 * 					)
 * 					for mdl in self.models             # <<<<<<<<<<<<<<
 * 				],
 * 				'ROC AUC': [
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_9genexpr83__pyx_v_mdl); __pyx_9genexpr83__pyx_v_mdl = 0;
    goto __pyx_L26_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_9genexpr83__pyx_v_mdl); __pyx_9genexpr83__pyx_v_mdl = 0;
    goto __pyx_L1_error;
    __pyx_L26_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_3, __pyx_kp_u_PRC_AUC, __pyx_t_4) < 0) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  { /* enter inner scope */

    /* "gnocis/models.pyx":1246
 * 					for mdl in self.models
 * 				],
 * 				'ROC AUC': [             # <<<<<<<<<<<<<<
 * 					'%.2f +/- %.2f %%'%(
 * 						mean([getAUC(c) for c in self.ROC[mdl]]) * 100.,
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1246, __pyx_L29_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "gnocis/models.pyx":1251
 * 						CI([getAUC(c) for c in self.ROC[mdl]]) * 100.
 * 					)
 * 					for mdl in self.models             # <<<<<<<<<<<<<<
 * 				],
 * 			})
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_models); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1251, __pyx_L29_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_11 = __pyx_t_5; __Pyx_INCREF(__pyx_t_11); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1251, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_8 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1251, __pyx_L29_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_11))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_11)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1251, __pyx_L29_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_11, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1251, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1251, __pyx_L29_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_11, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1251, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_8(__pyx_t_11);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1251, __pyx_L29_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr86__pyx_v_mdl, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "gnocis/models.pyx":1247
 * 				],
 * 				'ROC AUC': [
 * 					'%.2f +/- %.2f %%'%(             # <<<<<<<<<<<<<<
 * 						mean([getAUC(c) for c in self.ROC[mdl]]) * 100.,
 * 						CI([getAUC(c) for c in self.ROC[mdl]]) * 100.
 */
      __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1247, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;

      /* "gnocis/models.pyx":1248
 * 				'ROC AUC': [
 * 					'%.2f +/- %.2f %%'%(
 * 						mean([getAUC(c) for c in self.ROC[mdl]]) * 100.,             # <<<<<<<<<<<<<<
 * 						CI([getAUC(c) for c in self.ROC[mdl]]) * 100.
 * 					)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_mean); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1248, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_12);
      { /* enter inner scope */
        __pyx_t_13 = PyList_New(0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1248, __pyx_L34_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ROC); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1248, __pyx_L34_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = __Pyx_PyObject_GetItem(__pyx_t_14, __pyx_9genexpr86__pyx_v_mdl); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1248, __pyx_L34_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
          __pyx_t_14 = __pyx_t_15; __Pyx_INCREF(__pyx_t_14); __pyx_t_16 = 0;
          __pyx_t_17 = NULL;
        } else {
          __pyx_t_16 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1248, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_17 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1248, __pyx_L34_error)
        }
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        for (;;) {
          if (likely(!__pyx_t_17)) {
            if (likely(PyList_CheckExact(__pyx_t_14))) {
              if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_14)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_16); __Pyx_INCREF(__pyx_t_15); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1248, __pyx_L34_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_14, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1248, __pyx_L34_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            } else {
              if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_16); __Pyx_INCREF(__pyx_t_15); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1248, __pyx_L34_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_14, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1248, __pyx_L34_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            }
          } else {
            __pyx_t_15 = __pyx_t_17(__pyx_t_14);
            if (unlikely(!__pyx_t_15)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1248, __pyx_L34_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_15);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr87__pyx_v_c, __pyx_t_15);
          __pyx_t_15 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1248, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_19 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_18))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_18);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_18, function);
            }
          }
          __pyx_t_15 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_19, __pyx_9genexpr87__pyx_v_c) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_9genexpr87__pyx_v_c);
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1248, __pyx_L34_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_13, (PyObject*)__pyx_t_15))) __PYX_ERR(0, 1248, __pyx_L34_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_9genexpr87__pyx_v_c); __pyx_9genexpr87__pyx_v_c = 0;
        goto __pyx_L37_exit_scope;
        __pyx_L34_error:;
        __Pyx_XDECREF(__pyx_9genexpr87__pyx_v_c); __pyx_9genexpr87__pyx_v_c = 0;
        goto __pyx_L29_error;
        __pyx_L37_exit_scope:;
      } /* exit inner scope */
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_6 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1248, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = PyNumber_Multiply(__pyx_t_6, __pyx_float_100_); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1248, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_kp_u_2f); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1248, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u__17);
      __pyx_t_9 += 5;
      __Pyx_GIVEREF(__pyx_kp_u__17);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u__17);

      /* "gnocis/models.pyx":1249
 * 					'%.2f +/- %.2f %%'%(
 * 						mean([getAUC(c) for c in self.ROC[mdl]]) * 100.,
 * 						CI([getAUC(c) for c in self.ROC[mdl]]) * 100.             # <<<<<<<<<<<<<<
 * 					)
 * 					for mdl in self.models
 */
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_CI); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1249, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_12);
      { /* enter inner scope */
        __pyx_t_13 = PyList_New(0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1249, __pyx_L40_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ROC); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1249, __pyx_L40_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = __Pyx_PyObject_GetItem(__pyx_t_14, __pyx_9genexpr86__pyx_v_mdl); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1249, __pyx_L40_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
          __pyx_t_14 = __pyx_t_15; __Pyx_INCREF(__pyx_t_14); __pyx_t_16 = 0;
          __pyx_t_17 = NULL;
        } else {
          __pyx_t_16 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1249, __pyx_L40_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_17 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1249, __pyx_L40_error)
        }
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        for (;;) {
          if (likely(!__pyx_t_17)) {
            if (likely(PyList_CheckExact(__pyx_t_14))) {
              if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_14)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_16); __Pyx_INCREF(__pyx_t_15); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1249, __pyx_L40_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_14, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1249, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            } else {
              if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_16); __Pyx_INCREF(__pyx_t_15); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1249, __pyx_L40_error)
              #else
              __pyx_t_15 = PySequence_ITEM(__pyx_t_14, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1249, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_15);
              #endif
            }
          } else {
            __pyx_t_15 = __pyx_t_17(__pyx_t_14);
            if (unlikely(!__pyx_t_15)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1249, __pyx_L40_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_15);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr88__pyx_v_c, __pyx_t_15);
          __pyx_t_15 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1249, __pyx_L40_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_19 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_18))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_18);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_18, function);
            }
          }
          __pyx_t_15 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_19, __pyx_9genexpr88__pyx_v_c) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_9genexpr88__pyx_v_c);
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1249, __pyx_L40_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_13, (PyObject*)__pyx_t_15))) __PYX_ERR(0, 1249, __pyx_L40_error)
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_9genexpr88__pyx_v_c); __pyx_9genexpr88__pyx_v_c = 0;
        goto __pyx_L43_exit_scope;
        __pyx_L40_error:;
        __Pyx_XDECREF(__pyx_9genexpr88__pyx_v_c); __pyx_9genexpr88__pyx_v_c = 0;
        goto __pyx_L29_error;
        __pyx_L43_exit_scope:;
      } /* exit inner scope */
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_6 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_14, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1249, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = PyNumber_Multiply(__pyx_t_6, __pyx_float_100_); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1249, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Format(__pyx_t_12, __pyx_kp_u_2f); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1249, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u__7);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__7);
      PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u__7);

      /* "gnocis/models.pyx":1247
 * 				],
 * 				'ROC AUC': [
 * 					'%.2f +/- %.2f %%'%(             # <<<<<<<<<<<<<<
 * 						mean([getAUC(c) for c in self.ROC[mdl]]) * 100.,
 * 						CI([getAUC(c) for c in self.ROC[mdl]]) * 100.
 */
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 4, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1247, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1246, __pyx_L29_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "gnocis/models.pyx":1251
 * 						CI([getAUC(c) for c in self.ROC[mdl]]) * 100.
 * 					)
 * 					for mdl in self.models             # <<<<<<<<<<<<<<
 * 				],
 * 			})
 */
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_9genexpr86__pyx_v_mdl); __pyx_9genexpr86__pyx_v_mdl = 0;
    goto __pyx_L44_exit_scope;
    __pyx_L29_error:;
    __Pyx_XDECREF(__pyx_9genexpr86__pyx_v_mdl); __pyx_9genexpr86__pyx_v_mdl = 0;
    goto __pyx_L1_error;
    __pyx_L44_exit_scope:;
  } /* exit inner scope */
  if (PyDict_SetItem(__pyx_t_3, __pyx_kp_u_ROC_AUC, __pyx_t_4) < 0) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_20 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_20 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_Evaluation_statistics, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_Evaluation_statistics, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_11 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_u_Evaluation_statistics);
    __Pyx_GIVEREF(__pyx_kp_u_Evaluation_statistics);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_20, __pyx_kp_u_Evaluation_statistics);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_20, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":1234
 * 		).noCropNames()
 * 
 * 	def getAUCTable(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Evaluation statistics',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.getAUCTable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_9genexpr82__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_9genexpr83__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_9genexpr84__pyx_v_c);
  __Pyx_XDECREF(__pyx_9genexpr85__pyx_v_c);
  __Pyx_XDECREF(__pyx_9genexpr86__pyx_v_mdl);
  __Pyx_XDECREF(__pyx_9genexpr87__pyx_v_c);
  __Pyx_XDECREF(__pyx_9genexpr88__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1255
 * 			})
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		hdr = 'Cross-validation\n'
 * 		config = self.getConfigurationTable()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_17__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_17__repr__ = {"__repr__", (PyCFunction)__pyx_pw_6gnocis_6models_15crossvalidation_17__repr__, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_17__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_16__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_16__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_hdr = NULL;
  PyObject *__pyx_v_config = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "gnocis/models.pyx":1256
 * 
 * 	def __repr__(self):
 * 		hdr = 'Cross-validation\n'             # <<<<<<<<<<<<<<
 * 		config = self.getConfigurationTable()
 * 		t = self.getAUCTable()
 */
  __Pyx_INCREF(__pyx_kp_u_Cross_validation_2);
  __pyx_v_hdr = __pyx_kp_u_Cross_validation_2;

  /* "gnocis/models.pyx":1257
 * 	def __repr__(self):
 * 		hdr = 'Cross-validation\n'
 * 		config = self.getConfigurationTable()             # <<<<<<<<<<<<<<
 * 		t = self.getAUCTable()
 * 		return hdr + '\n' + config.__repr__() + '\n' + t.__repr__()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getConfigurationTable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_config = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":1258
 * 		hdr = 'Cross-validation\n'
 * 		config = self.getConfigurationTable()
 * 		t = self.getAUCTable()             # <<<<<<<<<<<<<<
 * 		return hdr + '\n' + config.__repr__() + '\n' + t.__repr__()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getAUCTable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_t = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":1259
 * 		config = self.getConfigurationTable()
 * 		t = self.getAUCTable()
 * 		return hdr + '\n' + config.__repr__() + '\n' + t.__repr__()             # <<<<<<<<<<<<<<
 * 
 * 	def _repr_html_(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_v_hdr, __pyx_kp_u__29); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_config, __pyx_n_s_repr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_3, __pyx_kp_u__29); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_repr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":1255
 * 			})
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		hdr = 'Cross-validation\n'
 * 		config = self.getConfigurationTable()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gnocis.models.crossvalidation.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hdr);
  __Pyx_XDECREF(__pyx_v_config);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1261
 * 		return hdr + '\n' + config.__repr__() + '\n' + t.__repr__()
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		hdr = '<div><b>Cross-validation</b></div>'
 * 		config = self.getConfigurationTable()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_19_repr_html_(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_15crossvalidation_19_repr_html_ = {"_repr_html_", (PyCFunction)__pyx_pw_6gnocis_6models_15crossvalidation_19_repr_html_, METH_O, 0};
static PyObject *__pyx_pw_6gnocis_6models_15crossvalidation_19_repr_html_(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_repr_html_ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gnocis_6models_15crossvalidation_18_repr_html_(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_15crossvalidation_18_repr_html_(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_hdr = NULL;
  PyObject *__pyx_v_config = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_repr_html_", 0);

  /* "gnocis/models.pyx":1262
 * 
 * 	def _repr_html_(self):
 * 		hdr = '<div><b>Cross-validation</b></div>'             # <<<<<<<<<<<<<<
 * 		config = self.getConfigurationTable()
 * 		t = self.getAUCTable()
 */
  __Pyx_INCREF(__pyx_kp_u_div_b_Cross_validation_b_div);
  __pyx_v_hdr = __pyx_kp_u_div_b_Cross_validation_b_div;

  /* "gnocis/models.pyx":1263
 * 	def _repr_html_(self):
 * 		hdr = '<div><b>Cross-validation</b></div>'
 * 		config = self.getConfigurationTable()             # <<<<<<<<<<<<<<
 * 		t = self.getAUCTable()
 * 		return '<div>' + hdr + config._repr_html_() + t._repr_html_() +\
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getConfigurationTable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_config = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":1264
 * 		hdr = '<div><b>Cross-validation</b></div>'
 * 		config = self.getConfigurationTable()
 * 		t = self.getAUCTable()             # <<<<<<<<<<<<<<
 * 		return '<div>' + hdr + config._repr_html_() + t._repr_html_() +\
 * 			'<div style="float: left;">%s</div><div style="float: right;">%s</div></div>'%(
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_getAUCTable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_t = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gnocis/models.pyx":1265
 * 		config = self.getConfigurationTable()
 * 		t = self.getAUCTable()
 * 		return '<div>' + hdr + config._repr_html_() + t._repr_html_() +\             # <<<<<<<<<<<<<<
 * 			'<div style="float: left;">%s</div><div style="float: right;">%s</div></div>'%(
 * 				self.plotPRC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7),
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u_div, __pyx_v_hdr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_config, __pyx_n_s_repr_html); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_repr_html); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1266
 * 		t = self.getAUCTable()
 * 		return '<div>' + hdr + config._repr_html_() + t._repr_html_() +\
 * 			'<div style="float: left;">%s</div><div style="float: right;">%s</div></div>'%(             # <<<<<<<<<<<<<<
 * 				self.plotPRC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7),
 * 				self.plotROC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7)
 */
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 0;
  __pyx_t_6 = 127;
  __Pyx_INCREF(__pyx_kp_u_div_style_float_left);
  __pyx_t_5 += 26;
  __Pyx_GIVEREF(__pyx_kp_u_div_style_float_left);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_div_style_float_left);

  /* "gnocis/models.pyx":1267
 * 		return '<div>' + hdr + config._repr_html_() + t._repr_html_() +\
 * 			'<div style="float: left;">%s</div><div style="float: right;">%s</div></div>'%(
 * 				self.plotPRC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7),             # <<<<<<<<<<<<<<
 * 				self.plotROC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7)
 * 			) + '</div>'
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_plotPRC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_returnHTML, Py_True) < 0) __PYX_ERR(0, 1267, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_figsize, __pyx_tuple__30) < 0) __PYX_ERR(0, 1267, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fontsizeLabels, __pyx_int_12) < 0) __PYX_ERR(0, 1267, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fontsizeLegend, __pyx_int_10) < 0) __PYX_ERR(0, 1267, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_fontsizeAxis, __pyx_int_7) < 0) __PYX_ERR(0, 1267, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_7), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_6;
  __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_div_div_style_float_right);
  __pyx_t_5 += 33;
  __Pyx_GIVEREF(__pyx_kp_u_div_div_style_float_right);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_div_div_style_float_right);

  /* "gnocis/models.pyx":1268
 * 			'<div style="float: left;">%s</div><div style="float: right;">%s</div></div>'%(
 * 				self.plotPRC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7),
 * 				self.plotROC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7)             # <<<<<<<<<<<<<<
 * 			) + '</div>'
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_plotROC); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_returnHTML, Py_True) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_figsize, __pyx_tuple__30) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_fontsizeLabels, __pyx_int_12) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_fontsizeLegend, __pyx_int_10) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_fontsizeAxis, __pyx_int_7) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_3), __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
  __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_7);
  __pyx_t_7 = 0;
  __Pyx_INCREF(__pyx_kp_u_div_div);
  __pyx_t_5 += 12;
  __Pyx_GIVEREF(__pyx_kp_u_div_div);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_kp_u_div_div);

  /* "gnocis/models.pyx":1266
 * 		t = self.getAUCTable()
 * 		return '<div>' + hdr + config._repr_html_() + t._repr_html_() +\
 * 			'<div style="float: left;">%s</div><div style="float: right;">%s</div></div>'%(             # <<<<<<<<<<<<<<
 * 				self.plotPRC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7),
 * 				self.plotROC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7)
 */
  __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_2, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1265
 * 		config = self.getConfigurationTable()
 * 		t = self.getAUCTable()
 * 		return '<div>' + hdr + config._repr_html_() + t._repr_html_() +\             # <<<<<<<<<<<<<<
 * 			'<div style="float: left;">%s</div><div style="float: right;">%s</div></div>'%(
 * 				self.plotPRC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7),
 */
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "gnocis/models.pyx":1269
 * 				self.plotPRC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7),
 * 				self.plotROC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7)
 * 			) + '</div>'             # <<<<<<<<<<<<<<
 * 
 * def crossvalidate(models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = 100., storeTrained = True):
 */
  __pyx_t_7 = PyNumber_Add(__pyx_t_2, __pyx_kp_u_div_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":1261
 * 		return hdr + '\n' + config.__repr__() + '\n' + t.__repr__()
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		hdr = '<div><b>Cross-validation</b></div>'
 * 		config = self.getConfigurationTable()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gnocis.models.crossvalidation._repr_html_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hdr);
  __Pyx_XDECREF(__pyx_v_config);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gnocis/models.pyx":1271
 * 			) + '</div>'
 * 
 * def crossvalidate(models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = 100., storeTrained = True):             # <<<<<<<<<<<<<<
 * 	return crossvalidation(models = models, trainingSet = trainingSet, validationSet = validationSet, labelPositive = labelPositive, labelNegative = labelNegative, repeats = repeats, ratioTrainTest = ratioTrainTest, ratioNegPos = ratioNegPos, storeTrained = storeTrained)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gnocis_6models_33crossvalidate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gnocis_6models_33crossvalidate = {"crossvalidate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6gnocis_6models_33crossvalidate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gnocis_6models_33crossvalidate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_models = 0;
  PyObject *__pyx_v_trainingSet = 0;
  PyObject *__pyx_v_validationSet = 0;
  PyObject *__pyx_v_labelPositive = 0;
  PyObject *__pyx_v_labelNegative = 0;
  PyObject *__pyx_v_repeats = 0;
  PyObject *__pyx_v_ratioTrainTest = 0;
  PyObject *__pyx_v_ratioNegPos = 0;
  PyObject *__pyx_v_storeTrained = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("crossvalidate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_models,&__pyx_n_s_trainingSet,&__pyx_n_s_validationSet,&__pyx_n_s_labelPositive,&__pyx_n_s_labelNegative,&__pyx_n_s_repeats,&__pyx_n_s_ratioTrainTest,&__pyx_n_s_ratioNegPos,&__pyx_n_s_storeTrained,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[3] = __pyx_k__31;
    values[4] = __pyx_k__32;
    values[5] = ((PyObject *)__pyx_int_20);
    values[6] = ((PyObject *)__pyx_float_0_8);
    values[7] = ((PyObject *)__pyx_float_100_);
    values[8] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_models)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trainingSet)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("crossvalidate", 0, 2, 9, 1); __PYX_ERR(0, 1271, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validationSet);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelPositive);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_labelNegative);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_repeats);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ratioTrainTest);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ratioNegPos);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_storeTrained);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "crossvalidate") < 0)) __PYX_ERR(0, 1271, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_models = values[0];
    __pyx_v_trainingSet = values[1];
    __pyx_v_validationSet = values[2];
    __pyx_v_labelPositive = values[3];
    __pyx_v_labelNegative = values[4];
    __pyx_v_repeats = values[5];
    __pyx_v_ratioTrainTest = values[6];
    __pyx_v_ratioNegPos = values[7];
    __pyx_v_storeTrained = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("crossvalidate", 0, 2, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1271, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gnocis.models.crossvalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gnocis_6models_32crossvalidate(__pyx_self, __pyx_v_models, __pyx_v_trainingSet, __pyx_v_validationSet, __pyx_v_labelPositive, __pyx_v_labelNegative, __pyx_v_repeats, __pyx_v_ratioTrainTest, __pyx_v_ratioNegPos, __pyx_v_storeTrained);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gnocis_6models_32crossvalidate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_models, PyObject *__pyx_v_trainingSet, PyObject *__pyx_v_validationSet, PyObject *__pyx_v_labelPositive, PyObject *__pyx_v_labelNegative, PyObject *__pyx_v_repeats, PyObject *__pyx_v_ratioTrainTest, PyObject *__pyx_v_ratioNegPos, PyObject *__pyx_v_storeTrained) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("crossvalidate", 0);

  /* "gnocis/models.pyx":1272
 * 
 * def crossvalidate(models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = 100., storeTrained = True):
 * 	return crossvalidation(models = models, trainingSet = trainingSet, validationSet = validationSet, labelPositive = labelPositive, labelNegative = labelNegative, repeats = repeats, ratioTrainTest = ratioTrainTest, ratioNegPos = ratioNegPos, storeTrained = storeTrained)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_latin1_crossvalidation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_models, __pyx_v_models) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_trainingSet, __pyx_v_trainingSet) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_validationSet, __pyx_v_validationSet) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_labelPositive, __pyx_v_labelPositive) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_labelNegative, __pyx_v_labelNegative) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_repeats, __pyx_v_repeats) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ratioTrainTest, __pyx_v_ratioTrainTest) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ratioNegPos, __pyx_v_ratioNegPos) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_storeTrained, __pyx_v_storeTrained) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gnocis/models.pyx":1271
 * 			) + '</div>'
 * 
 * def crossvalidate(models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = 100., storeTrained = True):             # <<<<<<<<<<<<<<
 * 	return crossvalidation(models = models, trainingSet = trainingSet, validationSet = validationSet, labelPositive = labelPositive, labelNegative = labelNegative, repeats = repeats, ratioTrainTest = ratioTrainTest, ratioNegPos = ratioNegPos, storeTrained = storeTrained)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gnocis.models.crossvalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *__pyx_freelist_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer[--__pyx_freecount_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_7genexpr__pyx_v_seq);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer[__pyx_freecount_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *)o;
  if (p->__pyx_7genexpr__pyx_v_seq) {
    e = (*v)(((PyObject *)p->__pyx_7genexpr__pyx_v_seq), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer *)o;
  tmp = ((PyObject*)p->__pyx_7genexpr__pyx_v_seq);
  p->__pyx_7genexpr__pyx_v_seq = ((struct __pyx_obj_6gnocis_9sequences_sequence *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct__multiprocessModelDeployer", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_1_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_win);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_1_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_win) {
    e = (*v)(p->__pyx_v_win, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *__pyx_freelist_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_pool);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs[__pyx_freecount_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *)o;
  if (p->__pyx_v_pool) {
    e = (*v)(p->__pyx_v_pool, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs *)o;
  tmp = ((PyObject*)p->__pyx_v_pool);
  p->__pyx_v_pool = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_2_getPoolProcessIDs", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_3_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_3_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_3_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_3_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_3_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *__pyx_freelist_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_r);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs[__pyx_freecount_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *)o;
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs *)o;
  tmp = ((PyObject*)p->__pyx_v_r);
  p->__pyx_v_r = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_4_getPoolModelIDs", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_5_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_5_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_5_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_5_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_5_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_5_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_5_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_mid);
  Py_CLEAR(p->__pyx_v_pid);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_5_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_5_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_5_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_5_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_mid) {
    e = (*v)(p->__pyx_v_mid, a); if (e) return e;
  }
  if (p->__pyx_v_pid) {
    e = (*v)(p->__pyx_v_pid, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_5_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_5_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_5_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_5_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_5_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_5_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_6_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_6_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_6_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_6_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_6_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_6_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_6_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_mid);
  Py_CLEAR(p->__pyx_v_pid);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_6_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_6_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_6_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_6_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_mid) {
    e = (*v)(p->__pyx_v_mid, a); if (e) return e;
  }
  if (p->__pyx_v_pid) {
    e = (*v)(p->__pyx_v_pid, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_6_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_6_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_6_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_6_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_6_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_6_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *__pyx_freelist_6gnocis_6models___pyx_scope_struct_7_setPoolModel[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_7_setPoolModel = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_7_setPoolModel(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_7_setPoolModel > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_7_setPoolModel[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_7_setPoolModel];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_7_setPoolModel(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_r);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_7_setPoolModel < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_7_setPoolModel[__pyx_freecount_6gnocis_6models___pyx_scope_struct_7_setPoolModel++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_7_setPoolModel(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *)o;
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_7_setPoolModel(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel *)o;
  tmp = ((PyObject*)p->__pyx_v_r);
  p->__pyx_v_r = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_7_setPoolModel = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_7_setPoolModel", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_7_setPoolModel), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_7_setPoolModel, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_7_setPoolModel, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_7_setPoolModel, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_7_setPoolModel, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_8_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_8_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_8_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_8_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_8_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_8_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_8_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_hasModel);
  Py_CLEAR(p->__pyx_v_pid);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_8_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_8_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_8_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_8_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_hasModel) {
    e = (*v)(p->__pyx_v_hasModel, a); if (e) return e;
  }
  if (p->__pyx_v_pid) {
    e = (*v)(p->__pyx_v_pid, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_8_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_8_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_8_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_8_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_8_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_8_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_9_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_9_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_9_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_9_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_9_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_9_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_9_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_hasModel);
  Py_CLEAR(p->__pyx_v_pid);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_9_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_9_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_9_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_9_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_hasModel) {
    e = (*v)(p->__pyx_v_hasModel, a); if (e) return e;
  }
  if (p->__pyx_v_pid) {
    e = (*v)(p->__pyx_v_pid, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_9_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_9_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_9_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_9_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_9_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_9_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *__pyx_freelist_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_positives);
  Py_CLEAR(p->__pyx_v_seqLens);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision[__pyx_freecount_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *)o;
  if (p->__pyx_v_positives) {
    e = (*v)(p->__pyx_v_positives, a); if (e) return e;
  }
  if (p->__pyx_v_seqLens) {
    e = (*v)(p->__pyx_v_seqLens, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision *)o;
  tmp = ((PyObject*)p->__pyx_v_positives);
  p->__pyx_v_positives = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_seqLens);
  p->__pyx_v_seqLens = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_10_calibrateGenomewidePrecision", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_11_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_11_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_11_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_11_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_11_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_11_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_11_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_chrom);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_11_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_11_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_11_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_11_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_chrom) {
    e = (*v)(p->__pyx_v_chrom, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_11_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_11_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_11_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_11_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_11_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_11_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_12_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_12_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_12_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_12_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_12_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_12_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_12_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_12_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_12_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_12_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_12_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_12_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_12_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_12_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_12_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_12_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_12_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *__pyx_freelist_6gnocis_6models___pyx_scope_struct_13_scoreSequence[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_13_scoreSequence = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_13_scoreSequence(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_13_scoreSequence > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_13_scoreSequence[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_13_scoreSequence];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_13_scoreSequence(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_seq);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_13_scoreSequence < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_13_scoreSequence[__pyx_freecount_6gnocis_6models___pyx_scope_struct_13_scoreSequence++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_13_scoreSequence(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  if (p->__pyx_v_seq) {
    e = (*v)(p->__pyx_v_seq, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_13_scoreSequence(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_seq);
  p->__pyx_v_seq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_13_scoreSequence = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_13_scoreSequence", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_13_scoreSequence), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_13_scoreSequence, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_13_scoreSequence, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_13_scoreSequence, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_13_scoreSequence, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_14_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_14_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_14_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_14_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_14_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_14_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_14_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_win);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_14_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_14_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_14_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_14_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_win) {
    e = (*v)(p->__pyx_v_win, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_14_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_14_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_14_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_14_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_14_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_14_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *__pyx_freelist_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_winPosSet);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves[__pyx_freecount_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *)o;
  if (p->__pyx_v_winPosSet) {
    e = (*v)(p->__pyx_v_winPosSet, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves *)o;
  tmp = ((PyObject*)p->__pyx_v_winPosSet);
  p->__pyx_v_winPosSet = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_15_predictSequenceStreamCurves", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_16_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_16_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_16_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_16_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_16_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_16_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_16_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_winPos);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_16_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_16_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_16_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_16_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_winPos) {
    e = (*v)(p->__pyx_v_winPos, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_16_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_16_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_16_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_16_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_16_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_16_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *__pyx_freelist_6gnocis_6models___pyx_scope_struct_17_getTrainer[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_17_getTrainer = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_17_getTrainer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_17_getTrainer > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_17_getTrainer[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_17_getTrainer];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_17_getTrainer(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_17_getTrainer < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_17_getTrainer[__pyx_freecount_6gnocis_6models___pyx_scope_struct_17_getTrainer++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_17_getTrainer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_17_getTrainer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_17_getTrainer = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_17_getTrainer", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_17_getTrainer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_17_getTrainer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_17_getTrainer, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_17_getTrainer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_17_getTrainer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *__pyx_freelist_6gnocis_6models___pyx_scope_struct_18___init__[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_18___init__ = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_18___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_18___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_18___init__[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_18___init__];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_18___init__(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_f);
  Py_CLEAR(p->__pyx_v_trainingNegatives);
  Py_CLEAR(p->__pyx_v_trainingPositives);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_18___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_18___init__[__pyx_freecount_6gnocis_6models___pyx_scope_struct_18___init__++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_18___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *)o;
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  if (p->__pyx_v_trainingNegatives) {
    e = (*v)(p->__pyx_v_trainingNegatives, a); if (e) return e;
  }
  if (p->__pyx_v_trainingPositives) {
    e = (*v)(p->__pyx_v_trainingPositives, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_18___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_f);
  p->__pyx_v_f = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_trainingNegatives);
  p->__pyx_v_trainingNegatives = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_trainingPositives);
  p->__pyx_v_trainingPositives = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_18___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_18___init__", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_18___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_18___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_18___init__, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_18___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_18___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_19_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_19_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_19_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_19_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_19_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_19_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_19_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_19_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_19_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_19_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_19_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_19_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_19_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_19_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_19_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_19_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_19_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_20_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_20_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_20_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_20_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_20_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_20_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_20_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_20_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_20_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_20_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_20_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_20_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_20_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_20_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_20_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_20_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_20_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *__pyx_freelist_6gnocis_6models___pyx_scope_struct_21_getTrainer[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_21_getTrainer = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_21_getTrainer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_21_getTrainer > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_21_getTrainer[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_21_getTrainer];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_21_getTrainer(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_21_getTrainer < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_21_getTrainer[__pyx_freecount_6gnocis_6models___pyx_scope_struct_21_getTrainer++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_21_getTrainer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_21_getTrainer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_21_getTrainer = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_21_getTrainer", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_21_getTrainer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_21_getTrainer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_21_getTrainer, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_21_getTrainer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_21_getTrainer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *__pyx_freelist_6gnocis_6models___pyx_scope_struct_22___init__[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_22___init__ = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_22___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_22___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_22___init__[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_22___init__];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_22___init__(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_tsByLabel);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_22___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_22___init__[__pyx_freecount_6gnocis_6models___pyx_scope_struct_22___init__++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_22___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *)o;
  if (p->__pyx_v_tsByLabel) {
    e = (*v)(p->__pyx_v_tsByLabel, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_22___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_tsByLabel);
  p->__pyx_v_tsByLabel = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_22___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_22___init__", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_22___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_22___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_22___init__, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_22___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_22___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_23_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_23_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_23_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_23_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_23_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_23_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_23_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_lbl);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_23_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_23_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_23_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_23_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_lbl) {
    e = (*v)(p->__pyx_v_lbl, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_23_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_23_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_23_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_23_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_23_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_23_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *__pyx_freelist_6gnocis_6models___pyx_scope_struct_24_plotPRC[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_24_plotPRC = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_24_plotPRC(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_24_plotPRC > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_24_plotPRC[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_24_plotPRC];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_24_plotPRC(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_curves);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_24_plotPRC < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_24_plotPRC[__pyx_freecount_6gnocis_6models___pyx_scope_struct_24_plotPRC++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_24_plotPRC(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *)o;
  if (p->__pyx_v_curves) {
    e = (*v)(p->__pyx_v_curves, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_24_plotPRC(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC *)o;
  tmp = ((PyObject*)p->__pyx_v_curves);
  p->__pyx_v_curves = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_24_plotPRC = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_24_plotPRC", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_24_plotPRC), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_24_plotPRC, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_24_plotPRC, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_24_plotPRC, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_24_plotPRC, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_25_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_25_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_25_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_25_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_25_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_25_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_25_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  Py_CLEAR(p->__pyx_v_pt);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_25_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_25_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_25_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_25_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  if (p->__pyx_v_pt) {
    e = (*v)(p->__pyx_v_pt, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_25_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_25_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_25_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_25_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_25_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_25_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *__pyx_freelist_6gnocis_6models___pyx_scope_struct_26_plotROC[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_26_plotROC = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_26_plotROC(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_26_plotROC > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_26_plotROC[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_26_plotROC];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_26_plotROC(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_curves);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_26_plotROC < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_26_plotROC[__pyx_freecount_6gnocis_6models___pyx_scope_struct_26_plotROC++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_26_plotROC(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *)o;
  if (p->__pyx_v_curves) {
    e = (*v)(p->__pyx_v_curves, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_26_plotROC(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC *)o;
  tmp = ((PyObject*)p->__pyx_v_curves);
  p->__pyx_v_curves = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_26_plotROC = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_26_plotROC", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_26_plotROC), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_26_plotROC, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_26_plotROC, /*tp_traverse*/
  __pyx_tp_clear_6gnocis_6models___pyx_scope_struct_26_plotROC, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_26_plotROC, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *__pyx_freelist_6gnocis_6models___pyx_scope_struct_27_genexpr[8];
static int __pyx_freecount_6gnocis_6models___pyx_scope_struct_27_genexpr = 0;

static PyObject *__pyx_tp_new_6gnocis_6models___pyx_scope_struct_27_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6gnocis_6models___pyx_scope_struct_27_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6gnocis_6models___pyx_scope_struct_27_genexpr[--__pyx_freecount_6gnocis_6models___pyx_scope_struct_27_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_27_genexpr(PyObject *o) {
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  Py_CLEAR(p->__pyx_v_pt);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6gnocis_6models___pyx_scope_struct_27_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr)))) {
    __pyx_freelist_6gnocis_6models___pyx_scope_struct_27_genexpr[__pyx_freecount_6gnocis_6models___pyx_scope_struct_27_genexpr++] = ((struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_27_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *p = (struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  if (p->__pyx_v_pt) {
    e = (*v)(p->__pyx_v_pt, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6gnocis_6models___pyx_scope_struct_27_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "gnocis.models.__pyx_scope_struct_27_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6gnocis_6models___pyx_scope_struct_27_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gnocis_6models___pyx_scope_struct_27_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gnocis_6models___pyx_scope_struct_27_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gnocis_6models___pyx_scope_struct_27_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_models(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_models},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "models",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_kp_u_2f, __pyx_k_2f, sizeof(__pyx_k_2f), 0, 1, 0, 0},
  {&__pyx_n_s_ACC, __pyx_k_ACC, sizeof(__pyx_k_ACC), 0, 0, 1, 1},
  {&__pyx_kp_u_AUC, __pyx_k_AUC, sizeof(__pyx_k_AUC), 0, 1, 0, 0},
  {&__pyx_kp_u_AUC_2, __pyx_k_AUC_2, sizeof(__pyx_k_AUC_2), 0, 1, 0, 0},
  {&__pyx_n_s_AUCs, __pyx_k_AUCs, sizeof(__pyx_k_AUCs), 0, 0, 1, 1},
  {&__pyx_n_s_BytesIO, __pyx_k_BytesIO, sizeof(__pyx_k_BytesIO), 0, 0, 1, 1},
  {&__pyx_n_s_CI, __pyx_k_CI, sizeof(__pyx_k_CI), 0, 0, 1, 1},
  {&__pyx_n_s_CICurve, __pyx_k_CICurve, sizeof(__pyx_k_CICurve), 0, 0, 1, 1},
  {&__pyx_n_s_CMStats, __pyx_k_CMStats, sizeof(__pyx_k_CMStats), 0, 0, 1, 1},
  {&__pyx_n_u_CV, __pyx_k_CV, sizeof(__pyx_k_CV), 0, 1, 0, 1},
  {&__pyx_kp_u_CVModelPredictions, __pyx_k_CVModelPredictions, sizeof(__pyx_k_CVModelPredictions), 0, 1, 0, 0},
  {&__pyx_n_s_CVModelPredictions_2, __pyx_k_CVModelPredictions_2, sizeof(__pyx_k_CVModelPredictions_2), 0, 0, 1, 1},
  {&__pyx_n_s_latin1_CVModelPredictions_2, __pyx_k_CVModelPredictions_2, sizeof(__pyx_k_CVModelPredictions_2), "latin-1", 0, 1, 1},
  {&__pyx_n_s_CVModelPredictions___init, __pyx_k_CVModelPredictions___init, sizeof(__pyx_k_CVModelPredictions___init), 0, 0, 1, 1},
  {&__pyx_n_s_CVModelPredictions___repr, __pyx_k_CVModelPredictions___repr, sizeof(__pyx_k_CVModelPredictions___repr), 0, 0, 1, 1},
  {&__pyx_n_s_CVModelPredictions__repr_html, __pyx_k_CVModelPredictions__repr_html, sizeof(__pyx_k_CVModelPredictions__repr_html), 0, 0, 1, 1},
  {&__pyx_n_s_CVModelPredictions_getStatTable, __pyx_k_CVModelPredictions_getStatTable, sizeof(__pyx_k_CVModelPredictions_getStatTable), 0, 0, 1, 1},
  {&__pyx_n_s_CVModelPredictions_predictCore, __pyx_k_CVModelPredictions_predictCore, sizeof(__pyx_k_CVModelPredictions_predictCore), 0, 0, 1, 1},
  {&__pyx_n_s_CVModelPredictions_regions, __pyx_k_CVModelPredictions_regions, sizeof(__pyx_k_CVModelPredictions_regions), 0, 0, 1, 1},
  {&__pyx_n_s_CVModelPredictions_setCorePredic, __pyx_k_CVModelPredictions_setCorePredic, sizeof(__pyx_k_CVModelPredictions_setCorePredic), 0, 0, 1, 1},
  {&__pyx_kp_u_Calibrating_models, __pyx_k_Calibrating_models, sizeof(__pyx_k_Calibrating_models), 0, 1, 0, 0},
  {&__pyx_n_u_Configuration, __pyx_k_Configuration, sizeof(__pyx_k_Configuration), 0, 1, 0, 1},
  {&__pyx_kp_s_Constructs_a_log_odds_model_bas, __pyx_k_Constructs_a_log_odds_model_bas, sizeof(__pyx_k_Constructs_a_log_odds_model_bas), 0, 0, 1, 0},
  {&__pyx_kp_u_Cross_validation, __pyx_k_Cross_validation, sizeof(__pyx_k_Cross_validation), 0, 1, 0, 0},
  {&__pyx_kp_u_Cross_validation_2, __pyx_k_Cross_validation_2, sizeof(__pyx_k_Cross_validation_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Dummy_model_Features_s, __pyx_k_Dummy_model_Features_s, sizeof(__pyx_k_Dummy_model_Features_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Evaluation_statistics, __pyx_k_Evaluation_statistics, sizeof(__pyx_k_Evaluation_statistics), 0, 1, 0, 0},
  {&__pyx_kp_u_Expected_at_random, __pyx_k_Expected_at_random, sizeof(__pyx_k_Expected_at_random), 0, 1, 0, 0},
  {&__pyx_n_s_FN, __pyx_k_FN, sizeof(__pyx_k_FN), 0, 0, 1, 1},
  {&__pyx_n_s_FP, __pyx_k_FP, sizeof(__pyx_k_FP), 0, 0, 1, 1},
  {&__pyx_kp_u_False_Positive_Rate, __pyx_k_False_Positive_Rate, sizeof(__pyx_k_False_Positive_Rate), 0, 1, 0, 0},
  {&__pyx_kp_u_Generating_training_test_sets, __pyx_k_Generating_training_test_sets, sizeof(__pyx_k_Generating_training_test_sets), 0, 1, 0, 0},
  {&__pyx_n_s_HTML, __pyx_k_HTML, sizeof(__pyx_k_HTML), 0, 0, 1, 1},
  {&__pyx_kp_s_Helper_class_for_cross_validati, __pyx_k_Helper_class_for_cross_validati, sizeof(__pyx_k_Helper_class_for_cross_validati), 0, 0, 1, 0},
  {&__pyx_kp_s_Helper_class_that_represents_cr, __pyx_k_Helper_class_that_represents_cr, sizeof(__pyx_k_Helper_class_that_represents_cr), 0, 0, 1, 0},
  {&__pyx_kp_u_INF, __pyx_k_INF, sizeof(__pyx_k_INF), 0, 1, 0, 0},
  {&__pyx_n_s_IPython_core_display, __pyx_k_IPython_core_display, sizeof(__pyx_k_IPython_core_display), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_u_Label, __pyx_k_Label, sizeof(__pyx_k_Label), 0, 1, 0, 1},
  {&__pyx_kp_u_Log_odds_model_Features, __pyx_k_Log_odds_model_Features, sizeof(__pyx_k_Log_odds_model_Features), 0, 1, 0, 0},
  {&__pyx_kp_u_Mean_length, __pyx_k_Mean_length, sizeof(__pyx_k_Mean_length), 0, 1, 0, 0},
  {&__pyx_n_u_Model, __pyx_k_Model, sizeof(__pyx_k_Model), 0, 1, 0, 1},
  {&__pyx_n_u_Name, __pyx_k_Name, sizeof(__pyx_k_Name), 0, 1, 0, 1},
  {&__pyx_kp_u_Negative_label, __pyx_k_Negative_label, sizeof(__pyx_k_Negative_label), 0, 1, 0, 0},
  {&__pyx_kp_u_Negatives_per_positive, __pyx_k_Negatives_per_positive, sizeof(__pyx_k_Negatives_per_positive), 0, 1, 0, 0},
  {&__pyx_n_s_PRC, __pyx_k_PRC, sizeof(__pyx_k_PRC), 0, 0, 1, 1},
  {&__pyx_kp_u_PRC_AUC, __pyx_k_PRC_AUC, sizeof(__pyx_k_PRC_AUC), 0, 1, 0, 0},
  {&__pyx_n_s_Pool, __pyx_k_Pool, sizeof(__pyx_k_Pool), 0, 0, 1, 1},
  {&__pyx_kp_u_Positive_label, __pyx_k_Positive_label, sizeof(__pyx_k_Positive_label), 0, 1, 0, 0},
  {&__pyx_n_u_Precision, __pyx_k_Precision, sizeof(__pyx_k_Precision), 0, 1, 0, 1},
  {&__pyx_kp_u_Predicting_cores, __pyx_k_Predicting_cores, sizeof(__pyx_k_Predicting_cores), 0, 1, 0, 0},
  {&__pyx_kp_u_Predicting_genome_wide, __pyx_k_Predicting_genome_wide, sizeof(__pyx_k_Predicting_genome_wide), 0, 1, 0, 0},
  {&__pyx_n_u_Predictions, __pyx_k_Predictions, sizeof(__pyx_k_Predictions), 0, 1, 0, 1},
  {&__pyx_kp_u_Predictions_2, __pyx_k_Predictions_2, sizeof(__pyx_k_Predictions_2), 0, 1, 0, 0},
  {&__pyx_n_s_ROC, __pyx_k_ROC, sizeof(__pyx_k_ROC), 0, 0, 1, 1},
  {&__pyx_kp_u_ROC_AUC, __pyx_k_ROC_AUC, sizeof(__pyx_k_ROC_AUC), 0, 1, 0, 0},
  {&__pyx_n_u_Recall, __pyx_k_Recall, sizeof(__pyx_k_Recall), 0, 1, 0, 1},
  {&__pyx_n_u_Repeat, __pyx_k_Repeat, sizeof(__pyx_k_Repeat), 0, 1, 0, 1},
  {&__pyx_kp_u_Repeats, __pyx_k_Repeats, sizeof(__pyx_k_Repeats), 0, 1, 0, 0},
  {&__pyx_kp_u_Repeats_d, __pyx_k_Repeats_d, sizeof(__pyx_k_Repeats_d), 0, 1, 0, 0},
  {&__pyx_n_u_Score, __pyx_k_Score, sizeof(__pyx_k_Score), 0, 1, 0, 1},
  {&__pyx_kp_u_Scored_sequences_Model, __pyx_k_Scored_sequences_Model, sizeof(__pyx_k_Scored_sequences_Model), 0, 1, 0, 0},
  {&__pyx_kp_u_Sequences, __pyx_k_Sequences, sizeof(__pyx_k_Sequences), 0, 1, 0, 0},
  {&__pyx_n_s_TN, __pyx_k_TN, sizeof(__pyx_k_TN), 0, 0, 1, 1},
  {&__pyx_n_s_TP, __pyx_k_TP, sizeof(__pyx_k_TP), 0, 0, 1, 1},
  {&__pyx_kp_s_The_sequenceModel_class_is_an_a, __pyx_k_The_sequenceModel_class_is_an_a, sizeof(__pyx_k_The_sequenceModel_class_is_an_a), 0, 0, 1, 0},
  {&__pyx_kp_s_This_model_takes_a_feature_set, __pyx_k_This_model_takes_a_feature_set, sizeof(__pyx_k_This_model_takes_a_feature_set), 0, 0, 1, 0},
  {&__pyx_kp_u_Train_test_ratio, __pyx_k_Train_test_ratio, sizeof(__pyx_k_Train_test_ratio), 0, 1, 0, 0},
  {&__pyx_kp_u_Training_sequences_per_class_d, __pyx_k_Training_sequences_per_class_d, sizeof(__pyx_k_Training_sequences_per_class_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Training_sequences_per_repeat, __pyx_k_Training_sequences_per_repeat, sizeof(__pyx_k_Training_sequences_per_repeat), 0, 1, 0, 0},
  {&__pyx_kp_u_Training_set, __pyx_k_Training_set, sizeof(__pyx_k_Training_set), 0, 1, 0, 0},
  {&__pyx_kp_u_Training_set_2, __pyx_k_Training_set_2, sizeof(__pyx_k_Training_set_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Training_set_3, __pyx_k_Training_set_3, sizeof(__pyx_k_Training_set_3), 0, 1, 0, 0},
  {&__pyx_kp_u_Training_set_s, __pyx_k_Training_set_s, sizeof(__pyx_k_Training_set_s), 0, 1, 0, 0},
  {&__pyx_kp_u_True_Positive_Rate, __pyx_k_True_Positive_Rate, sizeof(__pyx_k_True_Positive_Rate), 0, 1, 0, 0},
  {&__pyx_n_u_Validation, __pyx_k_Validation, sizeof(__pyx_k_Validation), 0, 1, 0, 1},
  {&__pyx_kp_u_Validation_sequences_per_repeat, __pyx_k_Validation_sequences_per_repeat, sizeof(__pyx_k_Validation_sequences_per_repeat), 0, 1, 0, 0},
  {&__pyx_kp_u_Validation_set, __pyx_k_Validation_set, sizeof(__pyx_k_Validation_set), 0, 1, 0, 0},
  {&__pyx_kp_u_Validation_set_2, __pyx_k_Validation_set_2, sizeof(__pyx_k_Validation_set_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Validation_set_s, __pyx_k_Validation_set_s, sizeof(__pyx_k_Validation_set_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Warning_model_s_not_trained_Skip, __pyx_k_Warning_model_s_not_trained_Skip, sizeof(__pyx_k_Warning_model_s_not_trained_Skip), 0, 1, 0, 0},
  {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
  {&__pyx_kp_u__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 1, 0, 0},
  {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
  {&__pyx_kp_u__29, __pyx_k__29, sizeof(__pyx_k__29), 0, 1, 0, 0},
  {&__pyx_n_s__45, __pyx_k__45, sizeof(__pyx_k__45), 0, 0, 1, 1},
  {&__pyx_n_s__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 1, 1},
  {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
  {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
  {&__pyx_n_s_aPrecision, __pyx_k_aPrecision, sizeof(__pyx_k_aPrecision), 0, 0, 1, 1},
  {&__pyx_n_s_aThreshold, __pyx_k_aThreshold, sizeof(__pyx_k_aThreshold), 0, 0, 1, 1},
  {&__pyx_n_s_addModel, __pyx_k_addModel, sizeof(__pyx_k_addModel), 0, 0, 1, 1},
  {&__pyx_n_u_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 1, 0, 1},
  {&__pyx_n_s_alpha, __pyx_k_alpha, sizeof(__pyx_k_alpha), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_arg, __pyx_k_arg, sizeof(__pyx_k_arg), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_asTable, __pyx_k_asTable, sizeof(__pyx_k_asTable), 0, 0, 1, 1},
  {&__pyx_n_s_b64encode, __pyx_k_b64encode, sizeof(__pyx_k_b64encode), 0, 0, 1, 1},
  {&__pyx_n_s_bPrecision, __pyx_k_bPrecision, sizeof(__pyx_k_bPrecision), 0, 0, 1, 1},
  {&__pyx_n_s_bThreshold, __pyx_k_bThreshold, sizeof(__pyx_k_bThreshold), 0, 0, 1, 1},
  {&__pyx_n_s_base64, __pyx_k_base64, sizeof(__pyx_k_base64), 0, 0, 1, 1},
  {&__pyx_n_s_bboxAnchorTo, __pyx_k_bboxAnchorTo, sizeof(__pyx_k_bboxAnchorTo), 0, 0, 1, 1},
  {&__pyx_n_s_bbox_to_anchor, __pyx_k_bbox_to_anchor, sizeof(__pyx_k_bbox_to_anchor), 0, 0, 1, 1},
  {&__pyx_n_s_bgModel, __pyx_k_bgModel, sizeof(__pyx_k_bgModel), 0, 0, 1, 1},
  {&__pyx_n_s_bgModelOrder, __pyx_k_bgModelOrder, sizeof(__pyx_k_bgModelOrder), 0, 0, 1, 1},
  {&__pyx_n_s_bio, __pyx_k_bio, sizeof(__pyx_k_bio), 0, 0, 1, 1},
  {&__pyx_n_s_blk, __pyx_k_blk, sizeof(__pyx_k_blk), 0, 0, 1, 1},
  {&__pyx_n_s_blkScores, __pyx_k_blkScores, sizeof(__pyx_k_blkScores), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_cACC, __pyx_k_cACC, sizeof(__pyx_k_cACC), 0, 0, 1, 1},
  {&__pyx_n_s_cScores, __pyx_k_cScores, sizeof(__pyx_k_cScores), 0, 0, 1, 1},
  {&__pyx_n_s_calibrate, __pyx_k_calibrate, sizeof(__pyx_k_calibrate), 0, 0, 1, 1},
  {&__pyx_n_s_calibrateGenomewidePrecision, __pyx_k_calibrateGenomewidePrecision, sizeof(__pyx_k_calibrateGenomewidePrecision), 0, 0, 1, 1},
  {&__pyx_n_s_calibrator, __pyx_k_calibrator, sizeof(__pyx_k_calibrator), 0, 0, 1, 1},
  {&__pyx_n_s_ci, __pyx_k_ci, sizeof(__pyx_k_ci), 0, 0, 1, 1},
  {&__pyx_n_s_classScoresSorted, __pyx_k_classScoresSorted, sizeof(__pyx_k_classScoresSorted), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_color, __pyx_k_color, sizeof(__pyx_k_color), 0, 0, 1, 1},
  {&__pyx_n_s_common, __pyx_k_common, sizeof(__pyx_k_common), 0, 0, 1, 1},
  {&__pyx_n_s_config, __pyx_k_config, sizeof(__pyx_k_config), 0, 0, 1, 1},
  {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
  {&__pyx_n_s_core, __pyx_k_core, sizeof(__pyx_k_core), 0, 0, 1, 1},
  {&__pyx_n_s_cores, __pyx_k_cores, sizeof(__pyx_k_cores), 0, 0, 1, 1},
  {&__pyx_n_s_createDummyPREdictorModel, __pyx_k_createDummyPREdictorModel, sizeof(__pyx_k_createDummyPREdictorModel), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidate, __pyx_k_crossvalidate, sizeof(__pyx_k_crossvalidate), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation, __pyx_k_crossvalidation, sizeof(__pyx_k_crossvalidation), 0, 0, 1, 1},
  {&__pyx_n_s_latin1_crossvalidation, __pyx_k_crossvalidation, sizeof(__pyx_k_crossvalidation), "latin-1", 0, 1, 1},
  {&__pyx_n_s_crossvalidation___init, __pyx_k_crossvalidation___init, sizeof(__pyx_k_crossvalidation___init), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation___init___locals, __pyx_k_crossvalidation___init___locals, sizeof(__pyx_k_crossvalidation___init___locals), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation___repr, __pyx_k_crossvalidation___repr, sizeof(__pyx_k_crossvalidation___repr), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation__repr_html, __pyx_k_crossvalidation__repr_html, sizeof(__pyx_k_crossvalidation__repr_html), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_addModel, __pyx_k_crossvalidation_addModel, sizeof(__pyx_k_crossvalidation_addModel), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_calibrate, __pyx_k_crossvalidation_calibrate, sizeof(__pyx_k_crossvalidation_calibrate), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_getAUCTable, __pyx_k_crossvalidation_getAUCTable, sizeof(__pyx_k_crossvalidation_getAUCTable), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_getConfiguration, __pyx_k_crossvalidation_getConfiguration, sizeof(__pyx_k_crossvalidation_getConfiguration), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_plotPRC, __pyx_k_crossvalidation_plotPRC, sizeof(__pyx_k_crossvalidation_plotPRC), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_plotPRC_locals_g, __pyx_k_crossvalidation_plotPRC_locals_g, sizeof(__pyx_k_crossvalidation_plotPRC_locals_g), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_plotROC, __pyx_k_crossvalidation_plotROC, sizeof(__pyx_k_crossvalidation_plotROC), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_plotROC_locals_g, __pyx_k_crossvalidation_plotROC_locals_g, sizeof(__pyx_k_crossvalidation_plotROC_locals_g), 0, 0, 1, 1},
  {&__pyx_n_s_crossvalidation_predict, __pyx_k_crossvalidation_predict, sizeof(__pyx_k_crossvalidation_predict), 0, 0, 1, 1},
  {&__pyx_n_s_cseq, __pyx_k_cseq, sizeof(__pyx_k_cseq), 0, 0, 1, 1},
  {&__pyx_n_s_curve, __pyx_k_curve, sizeof(__pyx_k_curve), 0, 0, 1, 1},
  {&__pyx_n_s_curvebyx, __pyx_k_curvebyx, sizeof(__pyx_k_curvebyx), 0, 0, 1, 1},
  {&__pyx_n_s_curvebyxmax, __pyx_k_curvebyxmax, sizeof(__pyx_k_curvebyxmax), 0, 0, 1, 1},
  {&__pyx_n_s_curvebyxmin, __pyx_k_curvebyxmin, sizeof(__pyx_k_curvebyxmin), 0, 0, 1, 1},
  {&__pyx_n_s_curves, __pyx_k_curves, sizeof(__pyx_k_curves), 0, 0, 1, 1},
  {&__pyx_n_s_cvpred, __pyx_k_cvpred, sizeof(__pyx_k_cvpred), 0, 0, 1, 1},
  {&__pyx_n_s_cvtrain, __pyx_k_cvtrain, sizeof(__pyx_k_cvtrain), 0, 0, 1, 1},
  {&__pyx_n_s_cvval, __pyx_k_cvval, sizeof(__pyx_k_cvval), 0, 0, 1, 1},
  {&__pyx_n_u_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 1, 0, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_s_defaultValue, __pyx_k_defaultValue, sizeof(__pyx_k_defaultValue), 0, 0, 1, 1},
  {&__pyx_n_s_degree, __pyx_k_degree, sizeof(__pyx_k_degree), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_display, __pyx_k_display, sizeof(__pyx_k_display), 0, 0, 1, 1},
  {&__pyx_kp_u_div, __pyx_k_div, sizeof(__pyx_k_div), 0, 1, 0, 0},
  {&__pyx_kp_u_div_2, __pyx_k_div_2, sizeof(__pyx_k_div_2), 0, 1, 0, 0},
  {&__pyx_kp_u_div_b_Cross_validation_b_div, __pyx_k_div_b_Cross_validation_b_div, sizeof(__pyx_k_div_b_Cross_validation_b_div), 0, 1, 0, 0},
  {&__pyx_kp_u_div_div, __pyx_k_div_div, sizeof(__pyx_k_div_div), 0, 1, 0, 0},
  {&__pyx_kp_u_div_div_style_float_right, __pyx_k_div_div_style_float_right, sizeof(__pyx_k_div_div_style_float_right), 0, 1, 0, 0},
  {&__pyx_kp_u_div_style_float_left, __pyx_k_div_style_float_left, sizeof(__pyx_k_div_style_float_left), 0, 1, 0, 0},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_done, __pyx_k_done, sizeof(__pyx_k_done), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_enableMultiprocessing, __pyx_k_enableMultiprocessing, sizeof(__pyx_k_enableMultiprocessing), 0, 0, 1, 1},
  {&__pyx_n_s_encoded, __pyx_k_encoded, sizeof(__pyx_k_encoded), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_factor, __pyx_k_factor, sizeof(__pyx_k_factor), 0, 0, 1, 1},
  {&__pyx_n_s_fancybox, __pyx_k_fancybox, sizeof(__pyx_k_fancybox), 0, 0, 1, 1},
  {&__pyx_n_s_features, __pyx_k_features, sizeof(__pyx_k_features), 0, 0, 1, 1},
  {&__pyx_n_s_fetch, __pyx_k_fetch, sizeof(__pyx_k_fetch), 0, 0, 1, 1},
  {&__pyx_n_s_fig, __pyx_k_fig, sizeof(__pyx_k_fig), 0, 0, 1, 1},
  {&__pyx_n_s_figsize, __pyx_k_figsize, sizeof(__pyx_k_figsize), 0, 0, 1, 1},
  {&__pyx_n_s_figure, __pyx_k_figure, sizeof(__pyx_k_figure), 0, 0, 1, 1},
  {&__pyx_n_s_fill_between, __pyx_k_fill_between, sizeof(__pyx_k_fill_between), 0, 0, 1, 1},
  {&__pyx_n_s_fixedStepCurve, __pyx_k_fixedStepCurve, sizeof(__pyx_k_fixedStepCurve), 0, 0, 1, 1},
  {&__pyx_n_s_fontsize, __pyx_k_fontsize, sizeof(__pyx_k_fontsize), 0, 0, 1, 1},
  {&__pyx_n_s_fontsizeAxis, __pyx_k_fontsizeAxis, sizeof(__pyx_k_fontsizeAxis), 0, 0, 1, 1},
  {&__pyx_n_s_fontsizeLabels, __pyx_k_fontsizeLabels, sizeof(__pyx_k_fontsizeLabels), 0, 0, 1, 1},
  {&__pyx_n_s_fontsizeLegend, __pyx_k_fontsizeLegend, sizeof(__pyx_k_fontsizeLegend), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fv, __pyx_k_fv, sizeof(__pyx_k_fv), 0, 0, 1, 1},
  {&__pyx_n_s_generateStream, __pyx_k_generateStream, sizeof(__pyx_k_generateStream), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_genome, __pyx_k_genome, sizeof(__pyx_k_genome), 0, 0, 1, 1},
  {&__pyx_n_s_genomeLen, __pyx_k_genomeLen, sizeof(__pyx_k_genomeLen), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getAUC, __pyx_k_getAUC, sizeof(__pyx_k_getAUC), 0, 0, 1, 1},
  {&__pyx_n_s_getAUCTable, __pyx_k_getAUCTable, sizeof(__pyx_k_getAUCTable), 0, 0, 1, 1},
  {&__pyx_n_s_getAll, __pyx_k_getAll, sizeof(__pyx_k_getAll), 0, 0, 1, 1},
  {&__pyx_n_s_getConfigurationTable, __pyx_k_getConfigurationTable, sizeof(__pyx_k_getConfigurationTable), 0, 0, 1, 1},
  {&__pyx_n_s_getConfusionMatrix, __pyx_k_getConfusionMatrix, sizeof(__pyx_k_getConfusionMatrix), 0, 0, 1, 1},
  {&__pyx_n_s_getConfusionMatrixStatistics, __pyx_k_getConfusionMatrixStatistics, sizeof(__pyx_k_getConfusionMatrixStatistics), 0, 0, 1, 1},
  {&__pyx_n_s_getOptimalAccuracyThreshold, __pyx_k_getOptimalAccuracyThreshold, sizeof(__pyx_k_getOptimalAccuracyThreshold), 0, 0, 1, 1},
  {&__pyx_n_s_getPRC, __pyx_k_getPRC, sizeof(__pyx_k_getPRC), 0, 0, 1, 1},
  {&__pyx_n_s_getPRCAUC, __pyx_k_getPRCAUC, sizeof(__pyx_k_getPRCAUC), 0, 0, 1, 1},
  {&__pyx_n_s_getPoolModelIDs, __pyx_k_getPoolModelIDs, sizeof(__pyx_k_getPoolModelIDs), 0, 0, 1, 1},
  {&__pyx_n_s_getPoolModelIDs_locals_genexpr, __pyx_k_getPoolModelIDs_locals_genexpr, sizeof(__pyx_k_getPoolModelIDs_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_getPoolProcessHasModel, __pyx_k_getPoolProcessHasModel, sizeof(__pyx_k_getPoolProcessHasModel), 0, 0, 1, 1},
  {&__pyx_n_s_getPoolProcessID, __pyx_k_getPoolProcessID, sizeof(__pyx_k_getPoolProcessID), 0, 0, 1, 1},
  {&__pyx_n_s_getPoolProcessIDs, __pyx_k_getPoolProcessIDs, sizeof(__pyx_k_getPoolProcessIDs), 0, 0, 1, 1},
  {&__pyx_n_s_getPoolProcessIDs_locals_genexpr, __pyx_k_getPoolProcessIDs_locals_genexpr, sizeof(__pyx_k_getPoolProcessIDs_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_getPoolProcessModelID, __pyx_k_getPoolProcessModelID, sizeof(__pyx_k_getPoolProcessModelID), 0, 0, 1, 1},
  {&__pyx_n_s_getPrecisionThreshold, __pyx_k_getPrecisionThreshold, sizeof(__pyx_k_getPrecisionThreshold), 0, 0, 1, 1},
  {&__pyx_n_s_getROC, __pyx_k_getROC, sizeof(__pyx_k_getROC), 0, 0, 1, 1},
  {&__pyx_n_s_getROCAUC, __pyx_k_getROCAUC, sizeof(__pyx_k_getROCAUC), 0, 0, 1, 1},
  {&__pyx_n_s_getSequenceScores, __pyx_k_getSequenceScores, sizeof(__pyx_k_getSequenceScores), 0, 0, 1, 1},
  {&__pyx_n_s_getStatTable, __pyx_k_getStatTable, sizeof(__pyx_k_getStatTable), 0, 0, 1, 1},
  {&__pyx_n_s_getTrainer, __pyx_k_getTrainer, sizeof(__pyx_k_getTrainer), 0, 0, 1, 1},
  {&__pyx_n_s_getValidationStatistics, __pyx_k_getValidationStatistics, sizeof(__pyx_k_getValidationStatistics), 0, 0, 1, 1},
  {&__pyx_n_s_get_context, __pyx_k_get_context, sizeof(__pyx_k_get_context), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_getvalue, __pyx_k_getvalue, sizeof(__pyx_k_getvalue), 0, 0, 1, 1},
  {&__pyx_n_u_ggplot, __pyx_k_ggplot, sizeof(__pyx_k_ggplot), 0, 1, 0, 1},
  {&__pyx_n_s_gnocis_models, __pyx_k_gnocis_models, sizeof(__pyx_k_gnocis_models), 0, 0, 1, 1},
  {&__pyx_kp_s_gnocis_models_pyx, __pyx_k_gnocis_models_pyx, sizeof(__pyx_k_gnocis_models_pyx), 0, 0, 1, 0},
  {&__pyx_n_u_grey, __pyx_k_grey, sizeof(__pyx_k_grey), 0, 1, 0, 1},
  {&__pyx_n_s_hasCores, __pyx_k_hasCores, sizeof(__pyx_k_hasCores), 0, 0, 1, 1},
  {&__pyx_n_s_hdr, __pyx_k_hdr, sizeof(__pyx_k_hdr), 0, 0, 1, 1},
  {&__pyx_n_s_html, __pyx_k_html, sizeof(__pyx_k_html), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_imdl, __pyx_k_imdl, sizeof(__pyx_k_imdl), 0, 0, 1, 1},
  {&__pyx_kp_u_img_src_data_image_png_base64_s, __pyx_k_img_src_data_image_png_base64_s, sizeof(__pyx_k_img_src_data_image_png_base64_s), 0, 1, 0, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_initargs, __pyx_k_initargs, sizeof(__pyx_k_initargs), 0, 0, 1, 1},
  {&__pyx_n_s_initializePoolProcess, __pyx_k_initializePoolProcess, sizeof(__pyx_k_initializePoolProcess), 0, 0, 1, 1},
  {&__pyx_n_s_initializePoolProcessIDs, __pyx_k_initializePoolProcessIDs, sizeof(__pyx_k_initializePoolProcessIDs), 0, 0, 1, 1},
  {&__pyx_n_s_initializer, __pyx_k_initializer, sizeof(__pyx_k_initializer), 0, 0, 1, 1},
  {&__pyx_n_s_io, __pyx_k_io, sizeof(__pyx_k_io), 0, 0, 1, 1},
  {&__pyx_n_s_ioputil, __pyx_k_ioputil, sizeof(__pyx_k_ioputil), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_label, __pyx_k_label, sizeof(__pyx_k_label), 0, 0, 1, 1},
  {&__pyx_n_s_labelNegative, __pyx_k_labelNegative, sizeof(__pyx_k_labelNegative), 0, 0, 1, 1},
  {&__pyx_n_s_labelPositive, __pyx_k_labelPositive, sizeof(__pyx_k_labelPositive), 0, 0, 1, 1},
  {&__pyx_n_s_labelPredictionCurves, __pyx_k_labelPredictionCurves, sizeof(__pyx_k_labelPredictionCurves), 0, 0, 1, 1},
  {&__pyx_n_s_labels, __pyx_k_labels, sizeof(__pyx_k_labels), 0, 0, 1, 1},
  {&__pyx_n_s_lbl, __pyx_k_lbl, sizeof(__pyx_k_lbl), 0, 0, 1, 1},
  {&__pyx_n_s_legend, __pyx_k_legend, sizeof(__pyx_k_legend), 0, 0, 1, 1},
  {&__pyx_n_s_legendLoc, __pyx_k_legendLoc, sizeof(__pyx_k_legendLoc), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_s_linestyle, __pyx_k_linestyle, sizeof(__pyx_k_linestyle), 0, 0, 1, 1},
  {&__pyx_n_s_loc, __pyx_k_loc, sizeof(__pyx_k_loc), 0, 0, 1, 1},
  {&__pyx_n_s_log, __pyx_k_log, sizeof(__pyx_k_log), 0, 0, 1, 1},
  {&__pyx_n_s_mIDs, __pyx_k_mIDs, sizeof(__pyx_k_mIDs), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
  {&__pyx_n_s_matplotlib_pyplot, __pyx_k_matplotlib_pyplot, sizeof(__pyx_k_matplotlib_pyplot), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_maxModelID, __pyx_k_maxModelID, sizeof(__pyx_k_maxModelID), 0, 0, 1, 1},
  {&__pyx_n_s_maxThreadFetchNT, __pyx_k_maxThreadFetchNT, sizeof(__pyx_k_maxThreadFetchNT), 0, 0, 1, 1},
  {&__pyx_n_s_mdl, __pyx_k_mdl, sizeof(__pyx_k_mdl), 0, 0, 1, 1},
  {&__pyx_n_s_mean, __pyx_k_mean, sizeof(__pyx_k_mean), 0, 0, 1, 1},
  {&__pyx_n_s_meanCurve, __pyx_k_meanCurve, sizeof(__pyx_k_meanCurve), 0, 0, 1, 1},
  {&__pyx_n_s_meanLen, __pyx_k_meanLen, sizeof(__pyx_k_meanLen), 0, 0, 1, 1},
  {&__pyx_n_s_meanlen, __pyx_k_meanlen, sizeof(__pyx_k_meanlen), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_model, __pyx_k_model, sizeof(__pyx_k_model), 0, 0, 1, 1},
  {&__pyx_n_u_model, __pyx_k_model, sizeof(__pyx_k_model), 0, 1, 0, 1},
  {&__pyx_n_s_modelID, __pyx_k_modelID, sizeof(__pyx_k_modelID), 0, 0, 1, 1},
  {&__pyx_n_u_modelID, __pyx_k_modelID, sizeof(__pyx_k_modelID), 0, 1, 0, 1},
  {&__pyx_n_s_models, __pyx_k_models, sizeof(__pyx_k_models), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_motifPairSpectrum, __pyx_k_motifPairSpectrum, sizeof(__pyx_k_motifPairSpectrum), 0, 0, 1, 1},
  {&__pyx_n_s_motifSpectrum, __pyx_k_motifSpectrum, sizeof(__pyx_k_motifSpectrum), 0, 0, 1, 1},
  {&__pyx_n_s_motifs, __pyx_k_motifs, sizeof(__pyx_k_motifs), 0, 0, 1, 1},
  {&__pyx_n_s_mp, __pyx_k_mp, sizeof(__pyx_k_mp), 0, 0, 1, 1},
  {&__pyx_n_s_mtask, __pyx_k_mtask, sizeof(__pyx_k_mtask), 0, 0, 1, 1},
  {&__pyx_n_s_multiprocessModelDeployer, __pyx_k_multiprocessModelDeployer, sizeof(__pyx_k_multiprocessModelDeployer), 0, 0, 1, 1},
  {&__pyx_n_s_multiprocessModelDeployer_locals, __pyx_k_multiprocessModelDeployer_locals, sizeof(__pyx_k_multiprocessModelDeployer_locals), 0, 0, 1, 1},
  {&__pyx_n_s_multiprocessing, __pyx_k_multiprocessing, sizeof(__pyx_k_multiprocessing), 0, 0, 1, 1},
  {&__pyx_n_s_multiprocessingPool, __pyx_k_multiprocessingPool, sizeof(__pyx_k_multiprocessingPool), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_nCoresUse, __pyx_k_nCoresUse, sizeof(__pyx_k_nCoresUse), 0, 0, 1, 1},
  {&__pyx_n_s_nNT, __pyx_k_nNT, sizeof(__pyx_k_nNT), 0, 0, 1, 1},
  {&__pyx_n_s_nTreadFetch, __pyx_k_nTreadFetch, sizeof(__pyx_k_nTreadFetch), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_nctable, __pyx_k_nctable, sizeof(__pyx_k_nctable), 0, 0, 1, 1},
  {&__pyx_n_s_negative, __pyx_k_negative, sizeof(__pyx_k_negative), 0, 0, 1, 1},
  {&__pyx_n_s_negativeScores, __pyx_k_negativeScores, sizeof(__pyx_k_negativeScores), 0, 0, 1, 1},
  {&__pyx_n_s_negatives, __pyx_k_negatives, sizeof(__pyx_k_negatives), 0, 0, 1, 1},
  {&__pyx_n_u_negatives, __pyx_k_negatives, sizeof(__pyx_k_negatives), 0, 1, 0, 1},
  {&__pyx_n_s_newID, __pyx_k_newID, sizeof(__pyx_k_newID), 0, 0, 1, 1},
  {&__pyx_n_s_noCropNames, __pyx_k_noCropNames, sizeof(__pyx_k_noCropNames), 0, 0, 1, 1},
  {&__pyx_n_s_npred, __pyx_k_npred, sizeof(__pyx_k_npred), 0, 0, 1, 1},
  {&__pyx_n_s_ntrain, __pyx_k_ntrain, sizeof(__pyx_k_ntrain), 0, 0, 1, 1},
  {&__pyx_n_s_nvneg, __pyx_k_nvneg, sizeof(__pyx_k_nvneg), 0, 0, 1, 1},
  {&__pyx_n_s_nvpos, __pyx_k_nvpos, sizeof(__pyx_k_nvpos), 0, 0, 1, 1},
  {&__pyx_n_s_outpath, __pyx_k_outpath, sizeof(__pyx_k_outpath), 0, 0, 1, 1},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_n_s_pIDs, __pyx_k_pIDs, sizeof(__pyx_k_pIDs), 0, 0, 1, 1},
  {&__pyx_n_s_pairFreq, __pyx_k_pairFreq, sizeof(__pyx_k_pairFreq), 0, 0, 1, 1},
  {&__pyx_n_s_plot, __pyx_k_plot, sizeof(__pyx_k_plot), 0, 0, 1, 1},
  {&__pyx_n_s_plotPRC, __pyx_k_plotPRC, sizeof(__pyx_k_plotPRC), 0, 0, 1, 1},
  {&__pyx_n_s_plotROC, __pyx_k_plotROC, sizeof(__pyx_k_plotROC), 0, 0, 1, 1},
  {&__pyx_n_s_plt, __pyx_k_plt, sizeof(__pyx_k_plt), 0, 0, 1, 1},
  {&__pyx_n_u_png, __pyx_k_png, sizeof(__pyx_k_png), 0, 1, 0, 1},
  {&__pyx_n_s_pool, __pyx_k_pool, sizeof(__pyx_k_pool), 0, 0, 1, 1},
  {&__pyx_n_s_pool_2, __pyx_k_pool_2, sizeof(__pyx_k_pool_2), 0, 0, 1, 1},
  {&__pyx_n_s_posScores, __pyx_k_posScores, sizeof(__pyx_k_posScores), 0, 0, 1, 1},
  {&__pyx_n_s_positive, __pyx_k_positive, sizeof(__pyx_k_positive), 0, 0, 1, 1},
  {&__pyx_n_s_positiveScores, __pyx_k_positiveScores, sizeof(__pyx_k_positiveScores), 0, 0, 1, 1},
  {&__pyx_n_s_positives, __pyx_k_positives, sizeof(__pyx_k_positives), 0, 0, 1, 1},
  {&__pyx_n_u_positives, __pyx_k_positives, sizeof(__pyx_k_positives), 0, 1, 0, 1},
  {&__pyx_n_s_precision, __pyx_k_precision, sizeof(__pyx_k_precision), 0, 0, 1, 1},
  {&__pyx_n_s_pred, __pyx_k_pred, sizeof(__pyx_k_pred), 0, 0, 1, 1},
  {&__pyx_n_s_predict, __pyx_k_predict, sizeof(__pyx_k_predict), 0, 0, 1, 1},
  {&__pyx_n_s_predictCore, __pyx_k_predictCore, sizeof(__pyx_k_predictCore), 0, 0, 1, 1},
  {&__pyx_n_s_predictSequenceStreamCurves, __pyx_k_predictSequenceStreamCurves, sizeof(__pyx_k_predictSequenceStreamCurves), 0, 0, 1, 1},
  {&__pyx_n_s_predictSequenceStreamRegions, __pyx_k_predictSequenceStreamRegions, sizeof(__pyx_k_predictSequenceStreamRegions), 0, 0, 1, 1},
  {&__pyx_n_s_predictedRegions, __pyx_k_predictedRegions, sizeof(__pyx_k_predictedRegions), 0, 0, 1, 1},
  {&__pyx_n_s_predictionCurves, __pyx_k_predictionCurves, sizeof(__pyx_k_predictionCurves), 0, 0, 1, 1},
  {&__pyx_n_s_predictions, __pyx_k_predictions, sizeof(__pyx_k_predictions), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_printTestStatistics, __pyx_k_printTestStatistics, sizeof(__pyx_k_printTestStatistics), 0, 0, 1, 1},
  {&__pyx_n_s_printValidationStatistics, __pyx_k_printValidationStatistics, sizeof(__pyx_k_printValidationStatistics), 0, 0, 1, 1},
  {&__pyx_n_s_procIDs, __pyx_k_procIDs, sizeof(__pyx_k_procIDs), 0, 0, 1, 1},
  {&__pyx_n_s_processID, __pyx_k_processID, sizeof(__pyx_k_processID), 0, 0, 1, 1},
  {&__pyx_n_s_progressTask, __pyx_k_progressTask, sizeof(__pyx_k_progressTask), 0, 0, 1, 1},
  {&__pyx_n_s_pt, __pyx_k_pt, sizeof(__pyx_k_pt), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_ratioNegPos, __pyx_k_ratioNegPos, sizeof(__pyx_k_ratioNegPos), 0, 0, 1, 1},
  {&__pyx_n_s_ratioTrainTest, __pyx_k_ratioTrainTest, sizeof(__pyx_k_ratioTrainTest), 0, 0, 1, 1},
  {&__pyx_n_s_regions, __pyx_k_regions, sizeof(__pyx_k_regions), 0, 0, 1, 1},
  {&__pyx_n_s_rep, __pyx_k_rep, sizeof(__pyx_k_rep), 0, 0, 1, 1},
  {&__pyx_n_s_repeats, __pyx_k_repeats, sizeof(__pyx_k_repeats), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_repr_html, __pyx_k_repr_html, sizeof(__pyx_k_repr_html), 0, 0, 1, 1},
  {&__pyx_n_s_ret, __pyx_k_ret, sizeof(__pyx_k_ret), 0, 0, 1, 1},
  {&__pyx_n_s_returnHTML, __pyx_k_returnHTML, sizeof(__pyx_k_returnHTML), 0, 0, 1, 1},
  {&__pyx_n_s_rotation, __pyx_k_rotation, sizeof(__pyx_k_rotation), 0, 0, 1, 1},
  {&__pyx_n_s_rtable, __pyx_k_rtable, sizeof(__pyx_k_rtable), 0, 0, 1, 1},
  {&__pyx_n_s_rtask, __pyx_k_rtask, sizeof(__pyx_k_rtask), 0, 0, 1, 1},
  {&__pyx_n_s_rvneg, __pyx_k_rvneg, sizeof(__pyx_k_rvneg), 0, 0, 1, 1},
  {&__pyx_n_s_rvpos, __pyx_k_rvpos, sizeof(__pyx_k_rvpos), 0, 0, 1, 1},
  {&__pyx_n_s_ry, __pyx_k_ry, sizeof(__pyx_k_ry), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_sScores, __pyx_k_sScores, sizeof(__pyx_k_sScores), 0, 0, 1, 1},
  {&__pyx_n_s_savefig, __pyx_k_savefig, sizeof(__pyx_k_savefig), 0, 0, 1, 1},
  {&__pyx_n_s_score, __pyx_k_score, sizeof(__pyx_k_score), 0, 0, 1, 1},
  {&__pyx_n_s_scoreSequence, __pyx_k_scoreSequence, sizeof(__pyx_k_scoreSequence), 0, 0, 1, 1},
  {&__pyx_n_s_scoreWindow, __pyx_k_scoreWindow, sizeof(__pyx_k_scoreWindow), 0, 0, 1, 1},
  {&__pyx_n_s_scores, __pyx_k_scores, sizeof(__pyx_k_scores), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_seq, __pyx_k_seq, sizeof(__pyx_k_seq), 0, 0, 1, 1},
  {&__pyx_n_s_seqLens, __pyx_k_seqLens, sizeof(__pyx_k_seqLens), 0, 0, 1, 1},
  {&__pyx_n_s_seqs, __pyx_k_seqs, sizeof(__pyx_k_seqs), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceLengths, __pyx_k_sequenceLengths, sizeof(__pyx_k_sequenceLengths), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel, __pyx_k_sequenceModel, sizeof(__pyx_k_sequenceModel), 0, 0, 1, 1},
  {&__pyx_n_s_latin1_sequenceModel, __pyx_k_sequenceModel, sizeof(__pyx_k_sequenceModel), "latin-1", 0, 1, 1},
  {&__pyx_n_s_sequenceModelDummy, __pyx_k_sequenceModelDummy, sizeof(__pyx_k_sequenceModelDummy), 0, 0, 1, 1},
  {&__pyx_n_s_latin1_sequenceModelDummy, __pyx_k_sequenceModelDummy, sizeof(__pyx_k_sequenceModelDummy), "latin-1", 0, 1, 1},
  {&__pyx_n_s_sequenceModelDummy___init, __pyx_k_sequenceModelDummy___init, sizeof(__pyx_k_sequenceModelDummy___init), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelDummy___repr, __pyx_k_sequenceModelDummy___repr, sizeof(__pyx_k_sequenceModelDummy___repr), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelDummy___str, __pyx_k_sequenceModelDummy___str, sizeof(__pyx_k_sequenceModelDummy___str), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelDummy_getTrainer, __pyx_k_sequenceModelDummy_getTrainer, sizeof(__pyx_k_sequenceModelDummy_getTrainer), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelDummy_getTrainer_lo, __pyx_k_sequenceModelDummy_getTrainer_lo, sizeof(__pyx_k_sequenceModelDummy_getTrainer_lo), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelDummy_scoreWindow, __pyx_k_sequenceModelDummy_scoreWindow, sizeof(__pyx_k_sequenceModelDummy_scoreWindow), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelLogOdds, __pyx_k_sequenceModelLogOdds, sizeof(__pyx_k_sequenceModelLogOdds), 0, 0, 1, 1},
  {&__pyx_n_s_latin1_sequenceModelLogOdds, __pyx_k_sequenceModelLogOdds, sizeof(__pyx_k_sequenceModelLogOdds), "latin-1", 0, 1, 1},
  {&__pyx_n_s_sequenceModelLogOdds___init, __pyx_k_sequenceModelLogOdds___init, sizeof(__pyx_k_sequenceModelLogOdds___init), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelLogOdds___init___lo, __pyx_k_sequenceModelLogOdds___init___lo, sizeof(__pyx_k_sequenceModelLogOdds___init___lo), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelLogOdds___repr, __pyx_k_sequenceModelLogOdds___repr, sizeof(__pyx_k_sequenceModelLogOdds___repr), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelLogOdds___str, __pyx_k_sequenceModelLogOdds___str, sizeof(__pyx_k_sequenceModelLogOdds___str), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelLogOdds_getTrainer, __pyx_k_sequenceModelLogOdds_getTrainer, sizeof(__pyx_k_sequenceModelLogOdds_getTrainer), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelLogOdds_getTrainer_2, __pyx_k_sequenceModelLogOdds_getTrainer_2, sizeof(__pyx_k_sequenceModelLogOdds_getTrainer_2), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModelLogOdds_scoreWindow, __pyx_k_sequenceModelLogOdds_scoreWindow, sizeof(__pyx_k_sequenceModelLogOdds_scoreWindow), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel___getstate, __pyx_k_sequenceModel___getstate, sizeof(__pyx_k_sequenceModel___getstate), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel___init, __pyx_k_sequenceModel___init, sizeof(__pyx_k_sequenceModel___init), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel___setstate, __pyx_k_sequenceModel___setstate, sizeof(__pyx_k_sequenceModel___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_calibrateGenomewid, __pyx_k_sequenceModel_calibrateGenomewid, sizeof(__pyx_k_sequenceModel_calibrateGenomewid), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_calibrateGenomewid_2, __pyx_k_sequenceModel_calibrateGenomewid_2, sizeof(__pyx_k_sequenceModel_calibrateGenomewid_2), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getConfusionMatrix, __pyx_k_sequenceModel_getConfusionMatrix, sizeof(__pyx_k_sequenceModel_getConfusionMatrix), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getOptimalAccuracy, __pyx_k_sequenceModel_getOptimalAccuracy, sizeof(__pyx_k_sequenceModel_getOptimalAccuracy), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getOptimalAccuracy_2, __pyx_k_sequenceModel_getOptimalAccuracy_2, sizeof(__pyx_k_sequenceModel_getOptimalAccuracy_2), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getPRC, __pyx_k_sequenceModel_getPRC, sizeof(__pyx_k_sequenceModel_getPRC), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getPRCAUC, __pyx_k_sequenceModel_getPRCAUC, sizeof(__pyx_k_sequenceModel_getPRCAUC), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getPrecisionThresh, __pyx_k_sequenceModel_getPrecisionThresh, sizeof(__pyx_k_sequenceModel_getPrecisionThresh), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getPrecisionThresh_2, __pyx_k_sequenceModel_getPrecisionThresh_2, sizeof(__pyx_k_sequenceModel_getPrecisionThresh_2), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getROC, __pyx_k_sequenceModel_getROC, sizeof(__pyx_k_sequenceModel_getROC), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getROCAUC, __pyx_k_sequenceModel_getROCAUC, sizeof(__pyx_k_sequenceModel_getROCAUC), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getSequenceScores, __pyx_k_sequenceModel_getSequenceScores, sizeof(__pyx_k_sequenceModel_getSequenceScores), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_getValidationStati, __pyx_k_sequenceModel_getValidationStati, sizeof(__pyx_k_sequenceModel_getValidationStati), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_plotPRC, __pyx_k_sequenceModel_plotPRC, sizeof(__pyx_k_sequenceModel_plotPRC), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_predict, __pyx_k_sequenceModel_predict, sizeof(__pyx_k_sequenceModel_predict), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_predictCore, __pyx_k_sequenceModel_predictCore, sizeof(__pyx_k_sequenceModel_predictCore), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_predictSequenceStr, __pyx_k_sequenceModel_predictSequenceStr, sizeof(__pyx_k_sequenceModel_predictSequenceStr), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_predictSequenceStr_2, __pyx_k_sequenceModel_predictSequenceStr_2, sizeof(__pyx_k_sequenceModel_predictSequenceStr_2), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_predictSequenceStr_3, __pyx_k_sequenceModel_predictSequenceStr_3, sizeof(__pyx_k_sequenceModel_predictSequenceStr_3), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_printTestStatistic, __pyx_k_sequenceModel_printTestStatistic, sizeof(__pyx_k_sequenceModel_printTestStatistic), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_score, __pyx_k_sequenceModel_score, sizeof(__pyx_k_sequenceModel_score), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_scoreSequence, __pyx_k_sequenceModel_scoreSequence, sizeof(__pyx_k_sequenceModel_scoreSequence), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_scoreSequence_loca, __pyx_k_sequenceModel_scoreSequence_loca, sizeof(__pyx_k_sequenceModel_scoreSequence_loca), 0, 0, 1, 1},
  {&__pyx_n_s_sequenceModel_train, __pyx_k_sequenceModel_train, sizeof(__pyx_k_sequenceModel_train), 0, 0, 1, 1},
  {&__pyx_n_s_sequences, __pyx_k_sequences, sizeof(__pyx_k_sequences), 0, 0, 1, 1},
  {&__pyx_n_s_setCorePredictions, __pyx_k_setCorePredictions, sizeof(__pyx_k_setCorePredictions), 0, 0, 1, 1},
  {&__pyx_n_s_setNCores, __pyx_k_setNCores, sizeof(__pyx_k_setNCores), 0, 0, 1, 1},
  {&__pyx_n_s_setNThreadFetch, __pyx_k_setNThreadFetch, sizeof(__pyx_k_setNThreadFetch), 0, 0, 1, 1},
  {&__pyx_n_s_setPoolModel, __pyx_k_setPoolModel, sizeof(__pyx_k_setPoolModel), 0, 0, 1, 1},
  {&__pyx_n_s_setPoolModel_locals_genexpr, __pyx_k_setPoolModel_locals_genexpr, sizeof(__pyx_k_setPoolModel_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_setPoolProcessID, __pyx_k_setPoolProcessID, sizeof(__pyx_k_setPoolProcessID), 0, 0, 1, 1},
  {&__pyx_n_s_setPoolProcessModel, __pyx_k_setPoolProcessModel, sizeof(__pyx_k_setPoolProcessModel), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_shuffle, __pyx_k_shuffle, sizeof(__pyx_k_shuffle), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_sortedweights, __pyx_k_sortedweights, sizeof(__pyx_k_sortedweights), 0, 0, 1, 1},
  {&__pyx_n_s_span, __pyx_k_span, sizeof(__pyx_k_span), 0, 0, 1, 1},
  {&__pyx_n_u_spawn, __pyx_k_spawn, sizeof(__pyx_k_spawn), 0, 1, 0, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_stats, __pyx_k_stats, sizeof(__pyx_k_stats), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_steps, __pyx_k_steps, sizeof(__pyx_k_steps), 0, 0, 1, 1},
  {&__pyx_n_s_storeTrained, __pyx_k_storeTrained, sizeof(__pyx_k_storeTrained), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 0, 1, 1},
  {&__pyx_n_s_streamSequenceWindows, __pyx_k_streamSequenceWindows, sizeof(__pyx_k_streamSequenceWindows), 0, 0, 1, 1},
  {&__pyx_n_s_style, __pyx_k_style, sizeof(__pyx_k_style), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_s_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 0, 1, 1},
  {&__pyx_n_s_task, __pyx_k_task, sizeof(__pyx_k_task), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_threshold, __pyx_k_threshold, sizeof(__pyx_k_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tight_layout, __pyx_k_tight_layout, sizeof(__pyx_k_tight_layout), 0, 0, 1, 1},
  {&__pyx_n_u_title, __pyx_k_title, sizeof(__pyx_k_title), 0, 1, 0, 1},
  {&__pyx_n_s_train, __pyx_k_train, sizeof(__pyx_k_train), 0, 0, 1, 1},
  {&__pyx_n_s_trainPREdictorModel, __pyx_k_trainPREdictorModel, sizeof(__pyx_k_trainPREdictorModel), 0, 0, 1, 1},
  {&__pyx_n_s_trainSinglePREdictorModel, __pyx_k_trainSinglePREdictorModel, sizeof(__pyx_k_trainSinglePREdictorModel), 0, 0, 1, 1},
  {&__pyx_n_s_trainedModels, __pyx_k_trainedModels, sizeof(__pyx_k_trainedModels), 0, 0, 1, 1},
  {&__pyx_n_s_trainingNegatives, __pyx_k_trainingNegatives, sizeof(__pyx_k_trainingNegatives), 0, 0, 1, 1},
  {&__pyx_n_s_trainingPositives, __pyx_k_trainingPositives, sizeof(__pyx_k_trainingPositives), 0, 0, 1, 1},
  {&__pyx_n_s_trainingSequences, __pyx_k_trainingSequences, sizeof(__pyx_k_trainingSequences), 0, 0, 1, 1},
  {&__pyx_n_s_trainingSet, __pyx_k_trainingSet, sizeof(__pyx_k_trainingSet), 0, 0, 1, 1},
  {&__pyx_n_s_ts, __pyx_k_ts, sizeof(__pyx_k_ts), 0, 0, 1, 1},
  {&__pyx_n_s_tsByLabel, __pyx_k_tsByLabel, sizeof(__pyx_k_tsByLabel), 0, 0, 1, 1},
  {&__pyx_n_s_tslabels, __pyx_k_tslabels, sizeof(__pyx_k_tslabels), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_kp_u_upper_left, __pyx_k_upper_left, sizeof(__pyx_k_upper_left), 0, 1, 0, 0},
  {&__pyx_kp_u_upper_right, __pyx_k_upper_right, sizeof(__pyx_k_upper_right), 0, 1, 0, 0},
  {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_vNeg, __pyx_k_vNeg, sizeof(__pyx_k_vNeg), 0, 0, 1, 1},
  {&__pyx_n_s_vPairs, __pyx_k_vPairs, sizeof(__pyx_k_vPairs), 0, 0, 1, 1},
  {&__pyx_n_s_vPos, __pyx_k_vPos, sizeof(__pyx_k_vPos), 0, 0, 1, 1},
  {&__pyx_n_s_validation, __pyx_k_validation, sizeof(__pyx_k_validation), 0, 0, 1, 1},
  {&__pyx_n_s_validationSet, __pyx_k_validationSet, sizeof(__pyx_k_validationSet), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_vneg, __pyx_k_vneg, sizeof(__pyx_k_vneg), 0, 0, 1, 1},
  {&__pyx_n_s_vp, __pyx_k_vp, sizeof(__pyx_k_vp), 0, 0, 1, 1},
  {&__pyx_n_s_vpos, __pyx_k_vpos, sizeof(__pyx_k_vpos), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_s_wPrecision, __pyx_k_wPrecision, sizeof(__pyx_k_wPrecision), 0, 0, 1, 1},
  {&__pyx_n_s_wScore, __pyx_k_wScore, sizeof(__pyx_k_wScore), 0, 0, 1, 1},
  {&__pyx_n_s_wThreshold, __pyx_k_wThreshold, sizeof(__pyx_k_wThreshold), 0, 0, 1, 1},
  {&__pyx_n_s_wantedPrecision, __pyx_k_wantedPrecision, sizeof(__pyx_k_wantedPrecision), 0, 0, 1, 1},
  {&__pyx_n_s_weight, __pyx_k_weight, sizeof(__pyx_k_weight), 0, 0, 1, 1},
  {&__pyx_n_s_weights, __pyx_k_weights, sizeof(__pyx_k_weights), 0, 0, 1, 1},
  {&__pyx_n_s_win, __pyx_k_win, sizeof(__pyx_k_win), 0, 0, 1, 1},
  {&__pyx_n_s_winPos, __pyx_k_winPos, sizeof(__pyx_k_winPos), 0, 0, 1, 1},
  {&__pyx_n_s_winPosSet, __pyx_k_winPosSet, sizeof(__pyx_k_winPosSet), 0, 0, 1, 1},
  {&__pyx_n_s_winScoreSet, __pyx_k_winScoreSet, sizeof(__pyx_k_winScoreSet), 0, 0, 1, 1},
  {&__pyx_n_s_winSeq, __pyx_k_winSeq, sizeof(__pyx_k_winSeq), 0, 0, 1, 1},
  {&__pyx_n_s_winSeqSet, __pyx_k_winSeqSet, sizeof(__pyx_k_winSeqSet), 0, 0, 1, 1},
  {&__pyx_n_s_windowSize, __pyx_k_windowSize, sizeof(__pyx_k_windowSize), 0, 0, 1, 1},
  {&__pyx_n_s_windowStep, __pyx_k_windowStep, sizeof(__pyx_k_windowStep), 0, 0, 1, 1},
  {&__pyx_n_s_windows, __pyx_k_windows, sizeof(__pyx_k_windows), 0, 0, 1, 1},
  {&__pyx_n_s_withLabel, __pyx_k_withLabel, sizeof(__pyx_k_withLabel), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_xlabel, __pyx_k_xlabel, sizeof(__pyx_k_xlabel), 0, 0, 1, 1},
  {&__pyx_n_s_xs, __pyx_k_xs, sizeof(__pyx_k_xs), 0, 0, 1, 1},
  {&__pyx_n_s_xticks, __pyx_k_xticks, sizeof(__pyx_k_xticks), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_ylabel, __pyx_k_ylabel, sizeof(__pyx_k_ylabel), 0, 0, 1, 1},
  {&__pyx_n_s_yticks, __pyx_k_yticks, sizeof(__pyx_k_yticks), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 49, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 82, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 254, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(0, 519, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 582, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 742, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 868, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 942, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 946, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "gnocis/models.pyx":153
 * 		multiprocessingPool = None
 * 	if n > 1:
 * 		multiprocessingPool = mp.get_context('spawn').Pool(nCoresUse, initializer = initializePoolProcess, initargs = ('',))             # <<<<<<<<<<<<<<
 * 		initializePoolProcessIDs(multiprocessingPool)
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u__2); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "gnocis/models.pyx":472
 * 		return getAUC(self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative))
 * 
 * 	def plotPRC(self, seqs, labelPositive = positive, labelNegative = negative, figsize = (8, 8), outpath = None, style = 'ggplot'):             # <<<<<<<<<<<<<<
 * 		""" Plots a Precision/Recall curve and either displays it in an IPython session or saves it to a file.
 * 
 */
  __pyx_tuple__4 = PyTuple_Pack(2, __pyx_int_8, __pyx_int_8); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "gnocis/models.pyx":506
 * 					[ v.y for v in curve ],
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))
 * 				plt.xlabel('Recall', fontsize = 14)             # <<<<<<<<<<<<<<
 * 				plt.ylabel('Precision', fontsize = 14)
 * 				plt.legend(loc = 'upper right', fancybox = True)
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_n_u_Recall); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "gnocis/models.pyx":507
 * 					label = '%s - AUC = %.2f %%'%(self.name, getAUC(curve)*100.))
 * 				plt.xlabel('Recall', fontsize = 14)
 * 				plt.ylabel('Precision', fontsize = 14)             # <<<<<<<<<<<<<<
 * 				plt.legend(loc = 'upper right', fancybox = True)
 * 				if outpath is None:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_n_u_Precision); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "gnocis/models.pyx":494
 * 			from io import BytesIO
 * 			from IPython.core.display import display, HTML
 * 			with plt.style.context(style):             # <<<<<<<<<<<<<<
 * 				fig = plt.figure(figsize = figsize)
 * 				positives, negatives = seqs.withLabel([ labelPositive, labelNegative ])
 */
  __pyx_tuple__10 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "gnocis/models.pyx":866
 * 
 * 	def predictCore(self, genome):
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))             # <<<<<<<<<<<<<<
 * 		cores = []
 * 		for i, mdl in enumerate(self.trainedModels):
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_u_Predicting_cores); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "gnocis/models.pyx":939
 * 		tslabels = list(trainingSet.labels())
 * 		tsByLabel = {
 * 			lbl: seqs.sequences[:]             # <<<<<<<<<<<<<<
 * 			for lbl, seqs in zip(tslabels, trainingSet.withLabel(tslabels))
 * 		}
 */
  __pyx_slice__18 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__18)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__18);
  __Pyx_GIVEREF(__pyx_slice__18);

  /* "gnocis/models.pyx":946
 * 		self.cvval = []
 * 		self.repeats = repeats
 * 		print('Cross-validation')             # <<<<<<<<<<<<<<
 * 		print(' - Training set: %s'%str(trainingSet))
 * 		print(' - Validation set: %s'%str(validationSet))
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_Cross_validation); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "gnocis/models.pyx":951
 * 		print(' - Training sequences per class: %d'%ntrain)
 * 		print(' - Repeats: %d'%repeats)
 * 		task = progressTask('Generating training/test sets', steps = self.repeats)             # <<<<<<<<<<<<<<
 * 		for rep in range(repeats):
 * 			task.update(step = rep)
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_Generating_training_test_sets); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "gnocis/models.pyx":973
 * 		task.done()
 * 		# Cross-validate models
 * 		task = progressTask('CV', steps = len(models))             # <<<<<<<<<<<<<<
 * 		for i, mdl in enumerate(models):
 * 			task.update(step = i)
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_n_u_CV); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "gnocis/models.pyx":1004
 * 
 * 	def calibrate(self, calibrator, labelPositive = positive):
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))             # <<<<<<<<<<<<<<
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_u_Calibrating_models); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "gnocis/models.pyx":1010
 * 				print('Warning: model %s not trained. Skipping.'%(mdl.name))
 * 				continue
 * 			rtask = progressTask('Repeat', steps = self.repeats)             # <<<<<<<<<<<<<<
 * 			# TODO Implement multiclass calibration
 * 			for rep in range(self.repeats):
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_n_u_Repeat); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "gnocis/models.pyx":1021
 * 
 * 	def predict(self, genome):
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))             # <<<<<<<<<<<<<<
 * 		for imdl, mdl in enumerate(self.models):
 * 			mtask.update(step = imdl)
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_u_Predicting_genome_wide); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "gnocis/models.pyx":1044
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_int_9, __pyx_int_9); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_float_0_, __pyx_float_neg_0_15); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "gnocis/models.pyx":1201
 * 				plt.xticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.xlabel('False Positive Rate', fontsize = fontsizeLabels)             # <<<<<<<<<<<<<<
 * 				plt.ylabel('True Positive Rate', fontsize = fontsizeLabels)
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_u_False_Positive_Rate); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "gnocis/models.pyx":1202
 * 				plt.yticks(fontsize = fontsizeAxis, rotation = 0)
 * 				plt.xlabel('False Positive Rate', fontsize = fontsizeLabels)
 * 				plt.ylabel('True Positive Rate', fontsize = fontsizeLabels)             # <<<<<<<<<<<<<<
 * 				plt.legend(bbox_to_anchor = bboxAnchorTo, loc = legendLoc, fontsize = fontsizeLegend, fancybox = True)
 * 				fig.tight_layout()
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u_True_Positive_Rate); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "gnocis/models.pyx":1267
 * 		return '<div>' + hdr + config._repr_html_() + t._repr_html_() +\
 * 			'<div style="float: left;">%s</div><div style="float: right;">%s</div></div>'%(
 * 				self.plotPRC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7),             # <<<<<<<<<<<<<<
 * 				self.plotROC(returnHTML = True, figsize = (6., 6.), fontsizeLabels = 12, fontsizeLegend = 10, fontsizeAxis = 7)
 * 			) + '</div>'
 */
  __pyx_tuple__30 = PyTuple_Pack(2, __pyx_float_6_, __pyx_float_6_); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "gnocis/models.pyx":33
 * maxThreadFetchNT = nTreadFetch * 1000
 * 
 * def setNThreadFetch(v):             # <<<<<<<<<<<<<<
 * 	global nTreadFetch
 * 	global maxThreadFetchNT
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_n_s_v); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_setNThreadFetch, 33, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 33, __pyx_L1_error)

  /* "gnocis/models.pyx":40
 * 
 * # Used for applying models in separate processes
 * def multiprocessModelDeployer(args):             # <<<<<<<<<<<<<<
 * 	cdef sequences seqs
 * 	cdef sequence seq, win
 */
  __pyx_tuple__35 = PyTuple_Pack(10, __pyx_n_s_args, __pyx_n_s_seqs, __pyx_n_s_seq, __pyx_n_s_win, __pyx_n_s_defaultValue, __pyx_n_s_windowSize, __pyx_n_s_windowStep, __pyx_n_s_seq, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(1, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_multiprocessModelDeployer, 40, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 40, __pyx_L1_error)

  /* "gnocis/models.pyx":61
 * 	]
 * 
 * def initializePoolProcess(arg):             # <<<<<<<<<<<<<<
 * 	global processID, model
 * 	processID, model = None, None
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_initializePoolProcess, 61, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 61, __pyx_L1_error)

  /* "gnocis/models.pyx":65
 * 	processID, model = None, None
 * 
 * def setPoolProcessID(newID):             # <<<<<<<<<<<<<<
 * 	global processID
 * 	processID = newID
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_n_s_newID); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_setPoolProcessID, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 65, __pyx_L1_error)

  /* "gnocis/models.pyx":70
 * 	return True
 * 
 * def getPoolProcessID(arg):             # <<<<<<<<<<<<<<
 * 	global processID
 * 	return processID
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getPoolProcessID, 70, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 70, __pyx_L1_error)

  /* "gnocis/models.pyx":74
 * 	return processID
 * 
 * def getPoolProcessModelID(arg):             # <<<<<<<<<<<<<<
 * 	global processID, model
 * 	if model == None:
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getPoolProcessModelID, 74, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 74, __pyx_L1_error)

  /* "gnocis/models.pyx":80
 * 	return processID, '%d %s'%(model.modelID, str(model))
 * 
 * def initializePoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	while True:
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )
 */
  __pyx_tuple__46 = PyTuple_Pack(3, __pyx_n_s_pool_2, __pyx_n_s_procIDs, __pyx_n_s__45); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_initializePoolProcessIDs, 80, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 80, __pyx_L1_error)

  /* "gnocis/models.pyx":87
 * 			break
 * 
 * def getPoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	while True:
 */
  __pyx_tuple__48 = PyTuple_Pack(4, __pyx_n_s_pool_2, __pyx_n_s_pIDs, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getPoolProcessIDs, 87, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 87, __pyx_L1_error)

  /* "gnocis/models.pyx":94
 * 			return pIDs
 * 
 * def getPoolModelIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	mIDs = set()
 */
  __pyx_tuple__50 = PyTuple_Pack(8, __pyx_n_s_pool_2, __pyx_n_s_pIDs, __pyx_n_s_mIDs, __pyx_n_s_r, __pyx_n_s__45, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getPoolModelIDs, 94, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 94, __pyx_L1_error)

  /* "gnocis/models.pyx":104
 * 			return mIDs
 * 
 * def setPoolProcessModel(mdl):             # <<<<<<<<<<<<<<
 * 	global model
 * 	model = mdl
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_n_s_mdl); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_setPoolProcessModel, 104, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 104, __pyx_L1_error)

  /* "gnocis/models.pyx":109
 * 	return True
 * 
 * def getPoolProcessHasModel(arg):             # <<<<<<<<<<<<<<
 * 	global model, processID
 * 	return (processID, model != None)
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_n_s_arg); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getPoolProcessHasModel, 109, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 109, __pyx_L1_error)

  /* "gnocis/models.pyx":113
 * 	return (processID, model != None)
 * 
 * def setPoolModel(pool, mdl):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	if mdl == None:
 */
  __pyx_tuple__56 = PyTuple_Pack(11, __pyx_n_s_pool_2, __pyx_n_s_mdl, __pyx_n_s_pIDs, __pyx_n_s_r, __pyx_n_s__45, __pyx_n_s__45, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s__45, __pyx_n_s__45, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(2, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_setPoolModel, 113, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 113, __pyx_L1_error)

  /* "gnocis/models.pyx":137
 * 				break
 * 
 * def setNCores(n):             # <<<<<<<<<<<<<<
 * 	"""
 * 	Sets the number of cores to use for multiprocessing.
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_n_s_n); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_setNCores, 137, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 137, __pyx_L1_error)

  /* "gnocis/models.pyx":174
 * 	"""
 * 
 * 	def __init__(self, name, enableMultiprocessing = True):             # <<<<<<<<<<<<<<
 * 		self.name = name
 * 		self.enableMultiprocessing = enableMultiprocessing
 */
  __pyx_tuple__60 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_enableMultiprocessing); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_init, 174, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 174, __pyx_L1_error)
  __pyx_tuple__62 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "gnocis/models.pyx":179
 * 
 * 	# For pickling (used with multiprocessing)
 * 	def __getstate__(self):             # <<<<<<<<<<<<<<
 * 		return self.__dict__
 * 
 */
  __pyx_tuple__63 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getstate, 179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 179, __pyx_L1_error)

  /* "gnocis/models.pyx":182
 * 		return self.__dict__
 * 
 * 	def __setstate__(self, state):             # <<<<<<<<<<<<<<
 * 		self.__dict__ = state
 * 
 */
  __pyx_tuple__65 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_state); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_setstate, 182, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 182, __pyx_L1_error)

  /* "gnocis/models.pyx":186
 * 
 * 	# Gets a list of scores for a sequence list or stream. A stream is recommended when the sequence list is large, in order to avoid running out of memory.
 * 	def getSequenceScores(self, seqs):             # <<<<<<<<<<<<<<
 * 		""" Scores a set of sequences, returning the maximum window score for each. Multiprocessing is enabled by default, but can be disabled in the constructor.
 * 
 */
  __pyx_tuple__67 = PyTuple_Pack(14, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_score, __pyx_n_s_blkScores, __pyx_n_s_blk, __pyx_n_s_cseq, __pyx_n_s_blk, __pyx_n_s_blkScores, __pyx_n_s_score, __pyx_n_s_blkScores, __pyx_n_s_score, __pyx_n_s_blk, __pyx_n_s_cseq, __pyx_n_s_cseq); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(2, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getSequenceScores, 186, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 186, __pyx_L1_error)

  /* "gnocis/models.pyx":237
 * 
 * 	# Gets the threshold that gives optimal accuracy on a pair of lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getOptimalAccuracyThreshold(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value optimized for accuracy to a set of positive and a set of negative sequences.
 * 
 */
  __pyx_tuple__69 = PyTuple_Pack(19, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_positives, __pyx_n_s_negatives, __pyx_n_s_positiveScores, __pyx_n_s_negativeScores, __pyx_n_s_vPairs, __pyx_n_s_TP, __pyx_n_s_FP, __pyx_n_s_TN, __pyx_n_s_FN, __pyx_n_s_threshold, __pyx_n_s_ACC, __pyx_n_s_vp, __pyx_n_s_cACC, __pyx_n_s_score, __pyx_n_s_score); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(4, 0, 19, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getOptimalAccuracyThreshold, 237, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 237, __pyx_L1_error)

  /* "gnocis/models.pyx":272
 * 
 * 	# Gets the threshold that gives a desired precision for two lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getPrecisionThreshold(self, positives, negatives, wantedPrecision):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value for a desired precision to a set of positive and a set of negative sequences. Linear interpolation is used in order to achieve a close approximation.
 * 
 */
  __pyx_tuple__71 = PyTuple_Pack(19, __pyx_n_s_self, __pyx_n_s_positives, __pyx_n_s_negatives, __pyx_n_s_wantedPrecision, __pyx_n_s_positiveScores, __pyx_n_s_negativeScores, __pyx_n_s_classScoresSorted, __pyx_n_s_TP, __pyx_n_s_FP, __pyx_n_s_aThreshold, __pyx_n_s_bThreshold, __pyx_n_s_aPrecision, __pyx_n_s_bPrecision, __pyx_n_s_e, __pyx_n_s_wPrecision, __pyx_n_s_wThreshold, __pyx_n_s_threshold, __pyx_n_s_score, __pyx_n_s_score); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(4, 0, 19, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getPrecisionThreshold, 272, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 272, __pyx_L1_error)

  /* "gnocis/models.pyx":318
 * 
 * 	# Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 	def calibrateGenomewidePrecision(self, positives, genome, factor = 1.0, precision = 0.8, bgModelOrder = 4, bgModel = None):             # <<<<<<<<<<<<<<
 * 		""" Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 
 */
  __pyx_tuple__73 = PyTuple_Pack(14, __pyx_n_s_self, __pyx_n_s_positives, __pyx_n_s_genome, __pyx_n_s_factor, __pyx_n_s_precision, __pyx_n_s_bgModelOrder, __pyx_n_s_bgModel, __pyx_n_s_seqLens, __pyx_n_s_genomeLen, __pyx_n_s_meanLen, __pyx_n_s_negatives, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(7, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_calibrateGenomewidePrecision, 318, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 318, __pyx_L1_error)
  __pyx_tuple__75 = PyTuple_Pack(4, ((PyObject*)__pyx_float_1_0), ((PyObject*)__pyx_float_0_8), ((PyObject *)__pyx_int_4), ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);

  /* "gnocis/models.pyx":356
 * 
 * 	# Applies model to a sequence with a sliding window, and returns the maximum score.
 * 	def scoreSequence(self, seq):             # <<<<<<<<<<<<<<
 * 		""" Scores a single sequence. The score is determined by applying the model with a sliding window, and taking the maximum score.
 * 
 */
  __pyx_tuple__76 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_seq, __pyx_n_s_score, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__76, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_scoreSequence, 356, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 356, __pyx_L1_error)

  /* "gnocis/models.pyx":379
 * 
 * 	# Gets and returns a confusion matrix.
 * 	def getConfusionMatrix(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a confusion matrix for sets of positive and negative sequences.
 * 
 */
  __pyx_tuple__78 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_positives, __pyx_n_s_negatives, __pyx_n_s_vPos, __pyx_n_s_vNeg, __pyx_n_s_score, __pyx_n_s_score); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(4, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__78, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getConfusionMatrix, 379, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(0, 379, __pyx_L1_error)

  /* "gnocis/models.pyx":399
 * 
 * 	# Generates a Receiver Operating Characteristic curve.
 * 	def getROC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a Receiver Operating Characteristic (ROC) curve for sets of positive and negative sequences.
 * 
 */
  __pyx_tuple__80 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_positives, __pyx_n_s_negatives, __pyx_n_s_vPos, __pyx_n_s_vNeg, __pyx_n_s_score, __pyx_n_s_score); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(4, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getROC, 399, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(0, 399, __pyx_L1_error)

  /* "gnocis/models.pyx":419
 * 
 * 	# Calculates the area under a Receiver Operating Characteristic curve.
 * 	def getROCAUC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns the area under a Receiver Operating Characteristic (ROCAUC) curve for sets of positive and negative sequences.
 * 
 */
  __pyx_tuple__82 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__82, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getROCAUC, 419, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(0, 419, __pyx_L1_error)

  /* "gnocis/models.pyx":436
 * 
 * 	# Generates a Precision/Recall curve.
 * 	def getPRC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a Precision/Recall curve (PRC) for sets of positive and negative sequences.
 * 
 */
  __pyx_tuple__84 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_positives, __pyx_n_s_negatives, __pyx_n_s_vPos, __pyx_n_s_vNeg, __pyx_n_s_score, __pyx_n_s_score); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(4, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getPRC, 436, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(0, 436, __pyx_L1_error)

  /* "gnocis/models.pyx":456
 * 
 * 	# Calculates the area under a Precision/Recall Curve.
 * 	def getPRCAUC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns the area under a Precision/Recall curve (PRCAUC) for sets of positive and negative sequences.
 * 
 */
  __pyx_tuple__86 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__86, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getPRCAUC, 456, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 456, __pyx_L1_error)

  /* "gnocis/models.pyx":472
 * 		return getAUC(self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative))
 * 
 * 	def plotPRC(self, seqs, labelPositive = positive, labelNegative = negative, figsize = (8, 8), outpath = None, style = 'ggplot'):             # <<<<<<<<<<<<<<
 * 		""" Plots a Precision/Recall curve and either displays it in an IPython session or saves it to a file.
 * 
 */
  __pyx_tuple__88 = PyTuple_Pack(22, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_style, __pyx_n_s_plt, __pyx_n_s_base64, __pyx_n_s_BytesIO, __pyx_n_s_display, __pyx_n_s_HTML, __pyx_n_s_fig, __pyx_n_s_positives, __pyx_n_s_negatives, __pyx_n_s_curve, __pyx_n_s_bio, __pyx_n_s_encoded, __pyx_n_s_html, __pyx_n_s_err, __pyx_n_s_v, __pyx_n_s_v); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(7, 0, 22, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__88, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_plotPRC, 472, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(0, 472, __pyx_L1_error)

  /* "gnocis/models.pyx":523
 * 
 * 	# Returns validation statistics dictionary.
 * 	def getValidationStatistics(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Returns common model validation statistics (confusion matrix values; ROCAUC; PRCAUC).
 * 
 */
  __pyx_tuple__90 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_CMStats, __pyx_n_s_stats, __pyx_n_s_k); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(4, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__90, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getValidationStatistics, 523, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(0, 523, __pyx_L1_error)

  /* "gnocis/models.pyx":548
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamRegions(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */
  __pyx_tuple__92 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_stream, __pyx_n_s_pred, __pyx_n_s_winPosSet, __pyx_n_s_winScoreSet, __pyx_n_s_i, __pyx_n_s_nNT, __pyx_n_s_winPos, __pyx_n_s_winSeq, __pyx_n_s_winSeqSet, __pyx_n_s_wScore, __pyx_n_s_cScores); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__92, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_predictSequenceStreamRegions, 548, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(0, 548, __pyx_L1_error)

  /* "gnocis/models.pyx":595
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamCurves(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */
  __pyx_tuple__94 = PyTuple_Pack(22, __pyx_n_s_self, __pyx_n_s_stream, __pyx_n_s_winPosSet, __pyx_n_s_winScoreSet, __pyx_n_s_i, __pyx_n_s_nNT, __pyx_n_s_winPos, __pyx_n_s_winSeq, __pyx_n_s_winSeqSet, __pyx_n_s_wScore, __pyx_n_s_sScores, __pyx_n_s_s, __pyx_n_s_posScores, __pyx_n_s_seq, __pyx_n_s_scores, __pyx_n_s_ret, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_winPos, __pyx_n_s_wScore, __pyx_n_s__45, __pyx_n_s_seq); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(2, 0, 22, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__94, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_predictSequenceStreamCurves, 595, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 595, __pyx_L1_error)

  /* "gnocis/models.pyx":656
 * 
 * 	# Predicts core regions.
 * 	def predictCore(self, genome):             # <<<<<<<<<<<<<<
 * 		""" Predicts cores. Not implemented by default. Models that use core prediction should implement this.
 * 
 */
  __pyx_tuple__96 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_genome); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_predictCore, 656, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "gnocis/models.pyx":668
 * 
 * 	# Short-hand for predictSequenceStreamRegions.
 * 	def predict(self, stream, curves = True):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */
  __pyx_tuple__98 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_stream, __pyx_n_s_curves); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_predict, 668, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(0, 668, __pyx_L1_error)
  __pyx_tuple__100 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);

  /* "gnocis/models.pyx":681
 * 
 * 	# Short-hand for getSequenceScores/scoreSequence
 * 	def score(self, target, asTable = False):             # <<<<<<<<<<<<<<
 * 		""" Scores a sequence of set of sequences.
 * 
 */
  __pyx_tuple__101 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_target, __pyx_n_s_asTable, __pyx_n_s_seq, __pyx_n_s_score); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__101, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_score, 681, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 681, __pyx_L1_error)
  __pyx_tuple__103 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);

  /* "gnocis/models.pyx":709
 * 
 * 	# Short-hand for training
 * 	def train(self, ts):             # <<<<<<<<<<<<<<
 * 		""" Scores a sequence of set of sequences.
 * 
 */
  __pyx_tuple__104 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ts); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__104, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_train, 709, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(0, 709, __pyx_L1_error)

  /* "gnocis/models.pyx":722
 * 
 * 	# Prints out test statistics.
 * 	def printTestStatistics(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		printValidationStatistics( self.getValidationStatistics(seqs, labelPositive = labelPositive, labelNegative = labelNegative) )
 * 
 */
  __pyx_tuple__106 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_seqs, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__106, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_printTestStatistics, 722, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(0, 722, __pyx_L1_error)

  /* "gnocis/models.pyx":741
 * 	"""
 * 
 * 	def __init__(self, name, features, windowSize, windowStep):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */
  __pyx_tuple__108 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_features, __pyx_n_s_windowSize, __pyx_n_s_windowStep); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_init, 741, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(0, 741, __pyx_L1_error)

  /* "gnocis/models.pyx":747
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Dummy model<Features: %s>'%(str(self.features))
 * 
 */
  __pyx_tuple__110 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__110, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_str, 747, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(0, 747, __pyx_L1_error)

  /* "gnocis/models.pyx":750
 * 		return 'Dummy model<Features: %s>'%(str(self.features))
 * 
 * 	def __repr__(self): return self.__str__()             # <<<<<<<<<<<<<<
 * 
 * 	def getTrainer(self):
 */
  __pyx_tuple__112 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__112, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_repr, 750, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(0, 750, __pyx_L1_error)

  /* "gnocis/models.pyx":752
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 */
  __pyx_tuple__114 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__114, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getTrainer, 752, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(0, 752, __pyx_L1_error)

  /* "gnocis/models.pyx":755
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 * 	def scoreWindow(self, seq):             # <<<<<<<<<<<<<<
 * 		cdef list fv
 * 		cdef float ret, v
 */
  __pyx_tuple__116 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_seq, __pyx_n_s_fv, __pyx_n_s_ret, __pyx_n_s_v); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__116, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_scoreWindow, 755, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(0, 755, __pyx_L1_error)

  /* "gnocis/models.pyx":786
 * 	"""
 * 
 * 	def __init__(self, name, features, trainingSet, windowSize, windowStep, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */
  __pyx_tuple__118 = PyTuple_Pack(17, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_features, __pyx_n_s_trainingSet, __pyx_n_s_windowSize, __pyx_n_s_windowStep, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_trainingPositives, __pyx_n_s_trainingNegatives, __pyx_n_s_f, __pyx_n_s_vPos, __pyx_n_s_vNeg, __pyx_n_s_weight, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(8, 0, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__118, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_init, 786, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(0, 786, __pyx_L1_error)

  /* "gnocis/models.pyx":807
 * 			self.sortedweights.append(weight)
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Log-odds model<Features: %s; Training set: %s; Positive label: %s; Negative label: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative))
 * 
 */
  __pyx_tuple__120 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__120, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_str, 807, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(0, 807, __pyx_L1_error)

  /* "gnocis/models.pyx":810
 * 		return 'Log-odds model<Features: %s; Training set: %s; Positive label: %s; Negative label: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative))
 * 
 * 	def __repr__(self): return self.__str__()             # <<<<<<<<<<<<<<
 * 
 * 	def getTrainer(self):
 */
  __pyx_tuple__122 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__122, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_repr, 810, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(0, 810, __pyx_L1_error)

  /* "gnocis/models.pyx":812
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 */
  __pyx_tuple__124 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__124, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getTrainer, 812, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(0, 812, __pyx_L1_error)

  /* "gnocis/models.pyx":815
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 * 	def scoreWindow(self, seq):             # <<<<<<<<<<<<<<
 * 		cdef list fv, w
 * 		cdef int i
 */
  __pyx_tuple__126 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_seq, __pyx_n_s_fv, __pyx_n_s_w, __pyx_n_s_i, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_scoreWindow, 815, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(0, 815, __pyx_L1_error)

  /* "gnocis/models.pyx":826
 * 
 * # Trains a singular-motif PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainSinglePREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 	return sequenceModelLogOdds(name, features.motifSpectrum(motifs), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 */
  __pyx_tuple__128 = PyTuple_Pack(7, __pyx_n_s_name, __pyx_n_s_motifs, __pyx_n_s_trainingSet, __pyx_n_s_windowSize, __pyx_n_s_windowStep, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  __pyx_codeobj__129 = (PyObject*)__Pyx_PyCode_New(7, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__128, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_trainSinglePREdictorModel, 826, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__129)) __PYX_ERR(0, 826, __pyx_L1_error)

  /* "gnocis/models.pyx":830
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def createDummyPREdictorModel(name, motifs, windowSize=500, windowStep=10):             # <<<<<<<<<<<<<<
 * 	return sequenceModelDummy(name, motifs.pairFreq(219), windowSize, windowStep)
 * 
 */
  __pyx_tuple__130 = PyTuple_Pack(4, __pyx_n_s_name, __pyx_n_s_motifs, __pyx_n_s_windowSize, __pyx_n_s_windowStep); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);
  __pyx_codeobj__131 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__130, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_createDummyPREdictorModel, 830, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__131)) __PYX_ERR(0, 830, __pyx_L1_error)

  /* "gnocis/models.pyx":834
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainPREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 	return sequenceModelLogOdds(name, features.motifPairSpectrum(motifs, 219), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 */
  __pyx_tuple__132 = PyTuple_Pack(7, __pyx_n_s_name, __pyx_n_s_motifs, __pyx_n_s_trainingSet, __pyx_n_s_windowSize, __pyx_n_s_windowStep, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(7, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_trainPREdictorModel, 834, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(0, 834, __pyx_L1_error)

  /* "gnocis/models.pyx":848
 * 	"""
 * 
 * 	def __init__(self, model, trainedModels = [], labelPredictionCurves = None):             # <<<<<<<<<<<<<<
 * 		self.name = model.name
 * 		self.model = model
 */
  __pyx_tuple__134 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_model, __pyx_n_s_trainedModels, __pyx_n_s_labelPredictionCurves); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);
  __pyx_codeobj__135 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_init, 848, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__135)) __PYX_ERR(0, 848, __pyx_L1_error)

  /* "gnocis/models.pyx":856
 * 		self.hasCores = False
 * 
 * 	def regions(self, label):             # <<<<<<<<<<<<<<
 * 		if label in self.predictedRegions:
 * 			return self.predictedRegions[label]
 */
  __pyx_tuple__136 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_label, __pyx_n_s_c); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);
  __pyx_codeobj__137 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__136, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_regions, 856, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__137)) __PYX_ERR(0, 856, __pyx_L1_error)

  /* "gnocis/models.pyx":861
 * 		return [ c.regions() for c in self.predictionCurves[label] ]
 * 
 * 	def setCorePredictions(self, label, cores):             # <<<<<<<<<<<<<<
 * 		self.predictedRegions[label] = cores
 * 		self.hasCores = True
 */
  __pyx_tuple__138 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_label, __pyx_n_s_cores); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_setCorePredictions, 861, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(0, 861, __pyx_L1_error)

  /* "gnocis/models.pyx":865
 * 		self.hasCores = True
 * 
 * 	def predictCore(self, genome):             # <<<<<<<<<<<<<<
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))
 * 		cores = []
 */
  __pyx_tuple__140 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_genome, __pyx_n_s_task, __pyx_n_s_cores, __pyx_n_s_i, __pyx_n_s_mdl, __pyx_n_s_core); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(2, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_predictCore, 865, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(0, 865, __pyx_L1_error)

  /* "gnocis/models.pyx":877
 * 		task.done()
 * 
 * 	def getStatTable(self):             # <<<<<<<<<<<<<<
 * 		rtable = []
 * 		for label in self.predictionCurves:
 */
  __pyx_tuple__142 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_rtable, __pyx_n_s_label, __pyx_n_s_npred, __pyx_n_s_meanlen, __pyx_n_s_p, __pyx_n_s_p, __pyx_n_s_r); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);
  __pyx_codeobj__143 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getStatTable, 877, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__143)) __PYX_ERR(0, 877, __pyx_L1_error)

  /* "gnocis/models.pyx":890
 * 		return nctable('Predictions: ' + self.model.name, rtable)
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()
 * 
 */
  __pyx_tuple__144 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);
  __pyx_codeobj__145 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__144, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_repr, 890, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__145)) __PYX_ERR(0, 890, __pyx_L1_error)

  /* "gnocis/models.pyx":893
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		return '<div>' + self.getStatTable()._repr_html_() + '</div>'
 * 
 */
  __pyx_tuple__146 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);
  __pyx_codeobj__147 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_repr_html, 893, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__147)) __PYX_ERR(0, 893, __pyx_L1_error)

  /* "gnocis/models.pyx":919
 * 	"""
 * 
 * 	def __init__(self, models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = None, storeTrained = True):             # <<<<<<<<<<<<<<
 * 		self.models = []
 * 		self.PRC = {}
 */
  __pyx_tuple__148 = PyTuple_Pack(32, __pyx_n_s_self, __pyx_n_s_models, __pyx_n_s_trainingSet, __pyx_n_s_validationSet, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_repeats, __pyx_n_s_ratioTrainTest, __pyx_n_s_ratioNegPos, __pyx_n_s_storeTrained, __pyx_n_s_vpos, __pyx_n_s_vneg, __pyx_n_s_tslabels, __pyx_n_s_tsByLabel, __pyx_n_s_ntrain, __pyx_n_s_task, __pyx_n_s_rep, __pyx_n_s_lbl, __pyx_n_s_rvpos, __pyx_n_s_rvneg, __pyx_n_s_nvpos, __pyx_n_s_nvneg, __pyx_n_s_i, __pyx_n_s_mdl, __pyx_n_s_lbl, __pyx_n_s_seqs, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_lbl, __pyx_n_s_s, __pyx_n_s_s, __pyx_n_s_s); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);
  __pyx_codeobj__149 = (PyObject*)__Pyx_PyCode_New(10, 0, 32, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__148, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_init, 919, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__149)) __PYX_ERR(0, 919, __pyx_L1_error)

  /* "gnocis/models.pyx":986
 * 	:type mdl: model
 * 	"""
 * 	def addModel(self, mdl):             # <<<<<<<<<<<<<<
 * 		self.PRC[mdl] = []
 * 		self.ROC[mdl] = []
 */
  __pyx_tuple__150 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_mdl, __pyx_n_s_task, __pyx_n_s_rep, __pyx_n_s_imdl); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_addModel, 986, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(0, 986, __pyx_L1_error)

  /* "gnocis/models.pyx":1003
 * 		task.done()
 * 
 * 	def calibrate(self, calibrator, labelPositive = positive):             # <<<<<<<<<<<<<<
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 */
  __pyx_tuple__152 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_calibrator, __pyx_n_s_labelPositive, __pyx_n_s_mtask, __pyx_n_s_imdl, __pyx_n_s_mdl, __pyx_n_s_rtask, __pyx_n_s_rep, __pyx_n_s_vpos); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);
  __pyx_codeobj__153 = (PyObject*)__Pyx_PyCode_New(3, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__152, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_calibrate, 1003, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__153)) __PYX_ERR(0, 1003, __pyx_L1_error)

  /* "gnocis/models.pyx":1020
 * 		mtask.done()
 * 
 * 	def predict(self, genome):             # <<<<<<<<<<<<<<
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 */
  __pyx_tuple__154 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_genome, __pyx_n_s_mtask, __pyx_n_s_imdl, __pyx_n_s_mdl, __pyx_n_s_cvpred, __pyx_n_s_rtask, __pyx_n_s_rep, __pyx_n_s_pred, __pyx_n_s_mdl); if (unlikely(!__pyx_tuple__154)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__154);
  __Pyx_GIVEREF(__pyx_tuple__154);
  __pyx_codeobj__155 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__154, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_predict, 1020, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__155)) __PYX_ERR(0, 1020, __pyx_L1_error)

  /* "gnocis/models.pyx":1044
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */
  __pyx_tuple__156 = PyTuple_Pack(55, __pyx_n_s_self, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_style, __pyx_n_s_returnHTML, __pyx_n_s_fontsizeLabels, __pyx_n_s_fontsizeLegend, __pyx_n_s_fontsizeAxis, __pyx_n_s_bboxAnchorTo, __pyx_n_s_legendLoc, __pyx_n_s_plt, __pyx_n_s_base64, __pyx_n_s_BytesIO, __pyx_n_s_display, __pyx_n_s_HTML, __pyx_n_s_fig, __pyx_n_s_positives, __pyx_n_s_negatives, __pyx_n_s_ry, __pyx_n_s_mdl, __pyx_n_s_curves, __pyx_n_s_xs, __pyx_n_s_curvebyx, __pyx_n_s_curvebyxmax, __pyx_n_s_curvebyxmin, __pyx_n_s_meanCurve, __pyx_n_s_CICurve, __pyx_n_s_AUCs, __pyx_n_s_bio, __pyx_n_s_encoded, __pyx_n_s_html, __pyx_n_s_err, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_curve, __pyx_n_s_x, __pyx_n_s_pt, __pyx_n_s_x, __pyx_n_s_i, __pyx_n_s_curve, __pyx_n_s_x, __pyx_n_s_i, __pyx_n_s_curve, __pyx_n_s_x, __pyx_n_s_pt, __pyx_n_s_x, __pyx_n_s_pt, __pyx_n_s_pt, __pyx_n_s_pt, __pyx_n_s_ci, __pyx_n_s_pt, __pyx_n_s_ci, __pyx_n_s_curve, __pyx_n_s_pt, __pyx_n_s_pt); if (unlikely(!__pyx_tuple__156)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__156);
  __Pyx_GIVEREF(__pyx_tuple__156);
  __pyx_codeobj__157 = (PyObject*)__Pyx_PyCode_New(10, 0, 55, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__156, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_plotPRC, 1044, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__157)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_tuple__158 = PyTuple_Pack(9, ((PyObject*)__pyx_tuple__25), ((PyObject *)Py_None), ((PyObject*)__pyx_n_u_ggplot), ((PyObject *)Py_False), ((PyObject *)__pyx_int_18), ((PyObject *)__pyx_int_12), ((PyObject *)__pyx_int_10), ((PyObject*)__pyx_tuple__26), ((PyObject*)__pyx_kp_u_upper_left)); if (unlikely(!__pyx_tuple__158)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__158);
  __Pyx_GIVEREF(__pyx_tuple__158);

  /* "gnocis/models.pyx":1133
 * 			raise err
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */
  __pyx_tuple__159 = PyTuple_Pack(52, __pyx_n_s_self, __pyx_n_s_figsize, __pyx_n_s_outpath, __pyx_n_s_style, __pyx_n_s_returnHTML, __pyx_n_s_fontsizeLabels, __pyx_n_s_fontsizeLegend, __pyx_n_s_fontsizeAxis, __pyx_n_s_bboxAnchorTo, __pyx_n_s_legendLoc, __pyx_n_s_plt, __pyx_n_s_base64, __pyx_n_s_BytesIO, __pyx_n_s_display, __pyx_n_s_HTML, __pyx_n_s_fig, __pyx_n_s_mdl, __pyx_n_s_curves, __pyx_n_s_xs, __pyx_n_s_curvebyx, __pyx_n_s_curvebyxmax, __pyx_n_s_curvebyxmin, __pyx_n_s_meanCurve, __pyx_n_s_CICurve, __pyx_n_s_AUCs, __pyx_n_s_bio, __pyx_n_s_encoded, __pyx_n_s_html, __pyx_n_s_err, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_curve, __pyx_n_s_x, __pyx_n_s_pt, __pyx_n_s_x, __pyx_n_s_i, __pyx_n_s_curve, __pyx_n_s_x, __pyx_n_s_i, __pyx_n_s_curve, __pyx_n_s_x, __pyx_n_s_pt, __pyx_n_s_x, __pyx_n_s_pt, __pyx_n_s_pt, __pyx_n_s_pt, __pyx_n_s_ci, __pyx_n_s_pt, __pyx_n_s_ci, __pyx_n_s_curve, __pyx_n_s_pt, __pyx_n_s_pt); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(0, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(10, 0, 52, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__159, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_plotROC, 1133, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(0, 1133, __pyx_L1_error)
  __pyx_tuple__161 = PyTuple_Pack(9, ((PyObject*)__pyx_tuple__4), ((PyObject *)Py_None), ((PyObject*)__pyx_n_u_ggplot), ((PyObject *)Py_False), ((PyObject *)__pyx_int_18), ((PyObject *)__pyx_int_12), ((PyObject *)__pyx_int_10), ((PyObject*)__pyx_tuple__26), ((PyObject*)__pyx_kp_u_upper_left)); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(0, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);

  /* "gnocis/models.pyx":1220
 * 			raise err
 * 
 * 	def getConfigurationTable(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Configuration',
 */
  __pyx_tuple__162 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__162)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);
  __pyx_codeobj__163 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__162, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getConfigurationTable, 1220, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__163)) __PYX_ERR(0, 1220, __pyx_L1_error)

  /* "gnocis/models.pyx":1234
 * 		).noCropNames()
 * 
 * 	def getAUCTable(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Evaluation statistics',
 */
  __pyx_tuple__164 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_mdl, __pyx_n_s_mdl, __pyx_n_s_c, __pyx_n_s_c, __pyx_n_s_mdl, __pyx_n_s_c, __pyx_n_s_c); if (unlikely(!__pyx_tuple__164)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);
  __pyx_codeobj__165 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__164, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_getAUCTable, 1234, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__165)) __PYX_ERR(0, 1234, __pyx_L1_error)

  /* "gnocis/models.pyx":1255
 * 			})
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		hdr = 'Cross-validation\n'
 * 		config = self.getConfigurationTable()
 */
  __pyx_tuple__166 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_hdr, __pyx_n_s_config, __pyx_n_s_t); if (unlikely(!__pyx_tuple__166)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);
  __pyx_codeobj__167 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__166, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_repr, 1255, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__167)) __PYX_ERR(0, 1255, __pyx_L1_error)

  /* "gnocis/models.pyx":1261
 * 		return hdr + '\n' + config.__repr__() + '\n' + t.__repr__()
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		hdr = '<div><b>Cross-validation</b></div>'
 * 		config = self.getConfigurationTable()
 */
  __pyx_tuple__168 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_hdr, __pyx_n_s_config, __pyx_n_s_t); if (unlikely(!__pyx_tuple__168)) __PYX_ERR(0, 1261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__168);
  __Pyx_GIVEREF(__pyx_tuple__168);
  __pyx_codeobj__169 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__168, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_repr_html, 1261, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__169)) __PYX_ERR(0, 1261, __pyx_L1_error)

  /* "gnocis/models.pyx":1271
 * 			) + '</div>'
 * 
 * def crossvalidate(models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = 100., storeTrained = True):             # <<<<<<<<<<<<<<
 * 	return crossvalidation(models = models, trainingSet = trainingSet, validationSet = validationSet, labelPositive = labelPositive, labelNegative = labelNegative, repeats = repeats, ratioTrainTest = ratioTrainTest, ratioNegPos = ratioNegPos, storeTrained = storeTrained)
 * 
 */
  __pyx_tuple__170 = PyTuple_Pack(9, __pyx_n_s_models, __pyx_n_s_trainingSet, __pyx_n_s_validationSet, __pyx_n_s_labelPositive, __pyx_n_s_labelNegative, __pyx_n_s_repeats, __pyx_n_s_ratioTrainTest, __pyx_n_s_ratioNegPos, __pyx_n_s_storeTrained); if (unlikely(!__pyx_tuple__170)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__170);
  __Pyx_GIVEREF(__pyx_tuple__170);
  __pyx_codeobj__171 = (PyObject*)__Pyx_PyCode_New(9, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__170, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_gnocis_models_pyx, __pyx_n_s_crossvalidate, 1271, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__171)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float__3 = PyFloat_FromDouble(.3); if (unlikely(!__pyx_float__3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_ = PyFloat_FromDouble(0.); if (unlikely(!__pyx_float_0_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_ = PyFloat_FromDouble(1.); if (unlikely(!__pyx_float_1_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_6_ = PyFloat_FromDouble(6.); if (unlikely(!__pyx_float_6_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_8 = PyFloat_FromDouble(0.8); if (unlikely(!__pyx_float_0_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_100_ = PyFloat_FromDouble(100.); if (unlikely(!__pyx_float_100_)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_neg_0_15 = PyFloat_FromDouble(-0.15); if (unlikely(!__pyx_float_neg_0_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_14 = PyInt_FromLong(14); if (unlikely(!__pyx_int_14)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_18 = PyInt_FromLong(18); if (unlikely(!__pyx_int_18)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_219 = PyInt_FromLong(219); if (unlikely(!__pyx_int_219)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_500 = PyInt_FromLong(500); if (unlikely(!__pyx_int_500)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1000 = PyInt_FromLong(1000); if (unlikely(!__pyx_int_1000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100000 = PyInt_FromLong(100000L); if (unlikely(!__pyx_int_100000)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer = &__pyx_type_6gnocis_6models___pyx_scope_struct__multiprocessModelDeployer;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_1_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs) < 0) __PYX_ERR(0, 87, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs = &__pyx_type_6gnocis_6models___pyx_scope_struct_2_getPoolProcessIDs;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_3_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_3_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_3_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_3_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_3_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_3_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs) < 0) __PYX_ERR(0, 94, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs = &__pyx_type_6gnocis_6models___pyx_scope_struct_4_getPoolModelIDs;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_5_genexpr) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_5_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_5_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_5_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_5_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_5_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_5_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_6_genexpr) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_6_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_6_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_6_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_6_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_6_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_6_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_7_setPoolModel) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_7_setPoolModel.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_7_setPoolModel.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_7_setPoolModel.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_7_setPoolModel.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_7_setPoolModel = &__pyx_type_6gnocis_6models___pyx_scope_struct_7_setPoolModel;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_8_genexpr) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_8_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_8_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_8_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_8_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_8_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_8_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_9_genexpr) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_9_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_9_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_9_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_9_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_9_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_9_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision) < 0) __PYX_ERR(0, 318, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision = &__pyx_type_6gnocis_6models___pyx_scope_struct_10_calibrateGenomewidePrecision;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_11_genexpr) < 0) __PYX_ERR(0, 342, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_11_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_11_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_11_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_11_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_11_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_11_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_12_genexpr) < 0) __PYX_ERR(0, 349, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_12_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_12_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_12_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_12_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_12_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_12_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_13_scoreSequence) < 0) __PYX_ERR(0, 356, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_13_scoreSequence.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_13_scoreSequence.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_13_scoreSequence.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_13_scoreSequence.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_13_scoreSequence = &__pyx_type_6gnocis_6models___pyx_scope_struct_13_scoreSequence;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_14_genexpr) < 0) __PYX_ERR(0, 369, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_14_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_14_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_14_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_14_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_14_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_14_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves) < 0) __PYX_ERR(0, 595, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves = &__pyx_type_6gnocis_6models___pyx_scope_struct_15_predictSequenceStreamCurves;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_16_genexpr) < 0) __PYX_ERR(0, 630, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_16_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_16_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_16_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_16_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_16_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_16_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_17_getTrainer) < 0) __PYX_ERR(0, 752, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_17_getTrainer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_17_getTrainer.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_17_getTrainer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_17_getTrainer.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_17_getTrainer = &__pyx_type_6gnocis_6models___pyx_scope_struct_17_getTrainer;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_18___init__) < 0) __PYX_ERR(0, 786, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_18___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_18___init__.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_18___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_18___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_18___init__ = &__pyx_type_6gnocis_6models___pyx_scope_struct_18___init__;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_19_genexpr) < 0) __PYX_ERR(0, 798, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_19_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_19_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_19_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_19_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_19_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_19_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_20_genexpr) < 0) __PYX_ERR(0, 799, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_20_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_20_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_20_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_20_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_20_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_20_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_21_getTrainer) < 0) __PYX_ERR(0, 812, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_21_getTrainer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_21_getTrainer.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_21_getTrainer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_21_getTrainer.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_21_getTrainer = &__pyx_type_6gnocis_6models___pyx_scope_struct_21_getTrainer;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_22___init__) < 0) __PYX_ERR(0, 919, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_22___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_22___init__.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_22___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_22___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_22___init__ = &__pyx_type_6gnocis_6models___pyx_scope_struct_22___init__;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_23_genexpr) < 0) __PYX_ERR(0, 942, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_23_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_23_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_23_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_23_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_23_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_23_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_24_plotPRC) < 0) __PYX_ERR(0, 1044, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_24_plotPRC.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_24_plotPRC.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_24_plotPRC.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_24_plotPRC.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_24_plotPRC = &__pyx_type_6gnocis_6models___pyx_scope_struct_24_plotPRC;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_25_genexpr) < 0) __PYX_ERR(0, 1064, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_25_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_25_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_25_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_25_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_25_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_25_genexpr;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_26_plotROC) < 0) __PYX_ERR(0, 1133, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_26_plotROC.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_26_plotROC.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_26_plotROC.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_26_plotROC.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_26_plotROC = &__pyx_type_6gnocis_6models___pyx_scope_struct_26_plotROC;
  if (PyType_Ready(&__pyx_type_6gnocis_6models___pyx_scope_struct_27_genexpr) < 0) __PYX_ERR(0, 1151, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6gnocis_6models___pyx_scope_struct_27_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6gnocis_6models___pyx_scope_struct_27_genexpr.tp_dictoffset && __pyx_type_6gnocis_6models___pyx_scope_struct_27_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6gnocis_6models___pyx_scope_struct_27_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6gnocis_6models___pyx_scope_struct_27_genexpr = &__pyx_type_6gnocis_6models___pyx_scope_struct_27_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("gnocis.regions"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6gnocis_7regions_region = __Pyx_ImportType(__pyx_t_1, "gnocis.regions", "region", sizeof(struct __pyx_obj_6gnocis_7regions_region), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_7regions_region) __PYX_ERR(1, 16, __pyx_L1_error)
  __pyx_ptype_6gnocis_7regions_regions = __Pyx_ImportType(__pyx_t_1, "gnocis.regions", "regions", sizeof(struct __pyx_obj_6gnocis_7regions_regions), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_7regions_regions) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("gnocis.sequences"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6gnocis_9sequences_sequenceLabel = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequenceLabel", sizeof(struct __pyx_obj_6gnocis_9sequences_sequenceLabel), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequenceLabel) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_sequence = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequence", sizeof(struct __pyx_obj_6gnocis_9sequences_sequence), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequence) __PYX_ERR(2, 27, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_9sequences_sequence = (struct __pyx_vtabstruct_6gnocis_9sequences_sequence*)__Pyx_GetVtable(__pyx_ptype_6gnocis_9sequences_sequence->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_9sequences_sequence)) __PYX_ERR(2, 27, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_sequences = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequences", sizeof(struct __pyx_obj_6gnocis_9sequences_sequences), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequences) __PYX_ERR(2, 37, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_sequenceStream = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequenceStream", sizeof(struct __pyx_obj_6gnocis_9sequences_sequenceStream), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequenceStream) __PYX_ERR(2, 49, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_sequenceGenerator = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "sequenceGenerator", sizeof(struct __pyx_obj_6gnocis_9sequences_sequenceGenerator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_sequenceGenerator) __PYX_ERR(2, 65, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_MarkovChain = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "MarkovChain", sizeof(struct __pyx_obj_6gnocis_9sequences_MarkovChain), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_MarkovChain) __PYX_ERR(2, 68, __pyx_L1_error)
  __pyx_ptype_6gnocis_9sequences_IID = __Pyx_ImportType(__pyx_t_1, "gnocis.sequences", "IID", sizeof(struct __pyx_obj_6gnocis_9sequences_IID), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_9sequences_IID) __PYX_ERR(2, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("gnocis.features"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6gnocis_8features_feature = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "feature", sizeof(struct __pyx_obj_6gnocis_8features_feature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_feature) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_feature = (struct __pyx_vtabstruct_6gnocis_8features_feature*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_feature->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_feature)) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_features = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "features", sizeof(struct __pyx_obj_6gnocis_8features_features), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_features) __PYX_ERR(3, 24, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_features = (struct __pyx_vtabstruct_6gnocis_8features_features*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_features->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_features)) __PYX_ERR(3, 24, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_scaledFeature = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "scaledFeature", sizeof(struct __pyx_obj_6gnocis_8features_scaledFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_scaledFeature) __PYX_ERR(3, 33, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_scaledFeature = (struct __pyx_vtabstruct_6gnocis_8features_scaledFeature*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_scaledFeature->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_scaledFeature)) __PYX_ERR(3, 33, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_featureScaler = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "featureScaler", sizeof(struct __pyx_obj_6gnocis_8features_featureScaler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_featureScaler) __PYX_ERR(3, 40, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_featureScaler = (struct __pyx_vtabstruct_6gnocis_8features_featureScaler*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_featureScaler->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_featureScaler)) __PYX_ERR(3, 40, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_featureMotifOccurrenceFrequency = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "featureMotifOccurrenceFrequency", sizeof(struct __pyx_obj_6gnocis_8features_featureMotifOccurrenceFrequency), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_featureMotifOccurrenceFrequency) __PYX_ERR(3, 49, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_featureMotifOccurrenceFrequency = (struct __pyx_vtabstruct_6gnocis_8features_featureMotifOccurrenceFrequency*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_featureMotifOccurrenceFrequency->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_featureMotifOccurrenceFrequency)) __PYX_ERR(3, 49, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "featurePREdictorMotifPairOccurrenceFrequency", sizeof(struct __pyx_obj_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency) __PYX_ERR(3, 56, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency = (struct __pyx_vtabstruct_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_featurePREdictorMotifPairOccurrenceFrequency)) __PYX_ERR(3, 56, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_kSpectrumFeature = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "kSpectrumFeature", sizeof(struct __pyx_obj_6gnocis_8features_kSpectrumFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_kSpectrumFeature) __PYX_ERR(3, 68, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_kSpectrumFeature = (struct __pyx_vtabstruct_6gnocis_8features_kSpectrumFeature*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_kSpectrumFeature->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_kSpectrumFeature)) __PYX_ERR(3, 68, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_kSpectrum = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "kSpectrum", sizeof(struct __pyx_obj_6gnocis_8features_kSpectrum), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_kSpectrum) __PYX_ERR(3, 77, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_kSpectrum = (struct __pyx_vtabstruct_6gnocis_8features_kSpectrum*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_kSpectrum->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_kSpectrum)) __PYX_ERR(3, 77, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_kSpectrumMMFeature = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "kSpectrumMMFeature", sizeof(struct __pyx_obj_6gnocis_8features_kSpectrumMMFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_kSpectrumMMFeature) __PYX_ERR(3, 91, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_kSpectrumMMFeature = (struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMFeature*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_kSpectrumMMFeature->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_kSpectrumMMFeature)) __PYX_ERR(3, 91, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_kSpectrumMM = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "kSpectrumMM", sizeof(struct __pyx_obj_6gnocis_8features_kSpectrumMM), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_kSpectrumMM) __PYX_ERR(3, 100, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_kSpectrumMM = (struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMM*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_kSpectrumMM->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_kSpectrumMM)) __PYX_ERR(3, 100, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_kSpectrumMMDFeature = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "kSpectrumMMDFeature", sizeof(struct __pyx_obj_6gnocis_8features_kSpectrumMMDFeature), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_kSpectrumMMDFeature) __PYX_ERR(3, 116, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_kSpectrumMMDFeature = (struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMDFeature*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_kSpectrumMMDFeature->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_kSpectrumMMDFeature)) __PYX_ERR(3, 116, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_kSpectrumMMD = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "kSpectrumMMD", sizeof(struct __pyx_obj_6gnocis_8features_kSpectrumMMD), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_kSpectrumMMD) __PYX_ERR(3, 125, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_kSpectrumMMD = (struct __pyx_vtabstruct_6gnocis_8features_kSpectrumMMD*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_kSpectrumMMD->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_kSpectrumMMD)) __PYX_ERR(3, 125, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_kSpectrumFeatureGPS = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "kSpectrumFeatureGPS", sizeof(struct __pyx_obj_6gnocis_8features_kSpectrumFeatureGPS), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_kSpectrumFeatureGPS) __PYX_ERR(3, 140, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_kSpectrumFeatureGPS = (struct __pyx_vtabstruct_6gnocis_8features_kSpectrumFeatureGPS*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_kSpectrumFeatureGPS->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_kSpectrumFeatureGPS)) __PYX_ERR(3, 140, __pyx_L1_error)
  __pyx_ptype_6gnocis_8features_kSpectrumGPS = __Pyx_ImportType(__pyx_t_1, "gnocis.features", "kSpectrumGPS", sizeof(struct __pyx_obj_6gnocis_8features_kSpectrumGPS), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_8features_kSpectrumGPS) __PYX_ERR(3, 150, __pyx_L1_error)
  __pyx_vtabptr_6gnocis_8features_kSpectrumGPS = (struct __pyx_vtabstruct_6gnocis_8features_kSpectrumGPS*)__Pyx_GetVtable(__pyx_ptype_6gnocis_8features_kSpectrumGPS->tp_dict); if (unlikely(!__pyx_vtabptr_6gnocis_8features_kSpectrumGPS)) __PYX_ERR(3, 150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("gnocis.validation"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6gnocis_10validation_point2D = __Pyx_ImportType(__pyx_t_1, "gnocis.validation", "point2D", sizeof(struct __pyx_obj_6gnocis_10validation_point2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_10validation_point2D) __PYX_ERR(4, 17, __pyx_L1_error)
  __pyx_ptype_6gnocis_10validation_validationPair = __Pyx_ImportType(__pyx_t_1, "gnocis.validation", "validationPair", sizeof(struct __pyx_obj_6gnocis_10validation_validationPair), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6gnocis_10validation_validationPair) __PYX_ERR(4, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("gnocis.sequences"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "getSequenceStreamFromPath", (void (**)(void))&__pyx_f_6gnocis_9sequences_getSequenceStreamFromPath, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_6gnocis_9sequences_getSequenceStreamFromPath *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initmodels(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initmodels(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_models(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_models(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_models(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'models' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_models(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("models", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_gnocis__models) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "gnocis.models")) {
      if (unlikely(PyDict_SetItemString(modules, "gnocis.models", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "gnocis/models.pyx":10
 * 
 * from __future__ import division
 * import random             # <<<<<<<<<<<<<<
 * from libcpp cimport bool
 * from libc.stdlib cimport malloc, free
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_random, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_random, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":14
 * from libc.stdlib cimport malloc, free
 * from libc.string cimport memcpy
 * import multiprocessing as mp             # <<<<<<<<<<<<<<
 * import math
 * from .features cimport *
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_multiprocessing, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mp, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":15
 * from libc.string cimport memcpy
 * import multiprocessing as mp
 * import math             # <<<<<<<<<<<<<<
 * from .features cimport *
 * from .validation cimport *
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_math, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_math, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":18
 * from .features cimport *
 * from .validation cimport *
 * from .validation import getConfusionMatrix, getConfusionMatrixStatistics, getROC, getPRC, getAUC, printValidationStatistics             # <<<<<<<<<<<<<<
 * from .sequences import streamSequenceWindows, positive, negative
 * from .ioputil import nctable, progressTask
 */
  __pyx_t_1 = PyList_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_getConfusionMatrix);
  __Pyx_GIVEREF(__pyx_n_s_getConfusionMatrix);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_getConfusionMatrix);
  __Pyx_INCREF(__pyx_n_s_getConfusionMatrixStatistics);
  __Pyx_GIVEREF(__pyx_n_s_getConfusionMatrixStatistics);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_getConfusionMatrixStatistics);
  __Pyx_INCREF(__pyx_n_s_getROC);
  __Pyx_GIVEREF(__pyx_n_s_getROC);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_getROC);
  __Pyx_INCREF(__pyx_n_s_getPRC);
  __Pyx_GIVEREF(__pyx_n_s_getPRC);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_getPRC);
  __Pyx_INCREF(__pyx_n_s_getAUC);
  __Pyx_GIVEREF(__pyx_n_s_getAUC);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_s_getAUC);
  __Pyx_INCREF(__pyx_n_s_printValidationStatistics);
  __Pyx_GIVEREF(__pyx_n_s_printValidationStatistics);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_s_printValidationStatistics);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_validation, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_getConfusionMatrix); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getConfusionMatrix, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_getConfusionMatrixStatistics); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getConfusionMatrixStatistics, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_getROC); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getROC, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_getPRC); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getPRC, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_getAUC); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getAUC, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_printValidationStatistics); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_printValidationStatistics, __pyx_t_1) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":19
 * from .validation cimport *
 * from .validation import getConfusionMatrix, getConfusionMatrixStatistics, getROC, getPRC, getAUC, printValidationStatistics
 * from .sequences import streamSequenceWindows, positive, negative             # <<<<<<<<<<<<<<
 * from .ioputil import nctable, progressTask
 * from .common import mean, CI
 */
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_streamSequenceWindows);
  __Pyx_GIVEREF(__pyx_n_s_streamSequenceWindows);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_streamSequenceWindows);
  __Pyx_INCREF(__pyx_n_s_positive);
  __Pyx_GIVEREF(__pyx_n_s_positive);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_positive);
  __Pyx_INCREF(__pyx_n_s_negative);
  __Pyx_GIVEREF(__pyx_n_s_negative);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_negative);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sequences, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_streamSequenceWindows); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_streamSequenceWindows, __pyx_t_2) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_positive); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_positive, __pyx_t_2) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_negative); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_negative, __pyx_t_2) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":20
 * from .validation import getConfusionMatrix, getConfusionMatrixStatistics, getROC, getPRC, getAUC, printValidationStatistics
 * from .sequences import streamSequenceWindows, positive, negative
 * from .ioputil import nctable, progressTask             # <<<<<<<<<<<<<<
 * from .common import mean, CI
 * from .curves import curves, fixedStepCurve
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_nctable);
  __Pyx_GIVEREF(__pyx_n_s_nctable);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_nctable);
  __Pyx_INCREF(__pyx_n_s_progressTask);
  __Pyx_GIVEREF(__pyx_n_s_progressTask);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_progressTask);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_ioputil, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_nctable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nctable, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_progressTask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_progressTask, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":21
 * from .sequences import streamSequenceWindows, positive, negative
 * from .ioputil import nctable, progressTask
 * from .common import mean, CI             # <<<<<<<<<<<<<<
 * from .curves import curves, fixedStepCurve
 * 
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_mean);
  __Pyx_GIVEREF(__pyx_n_s_mean);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_mean);
  __Pyx_INCREF(__pyx_n_s_CI);
  __Pyx_GIVEREF(__pyx_n_s_CI);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_CI);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_common, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_mean); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_mean, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CI); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CI, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":22
 * from .ioputil import nctable, progressTask
 * from .common import mean, CI
 * from .curves import curves, fixedStepCurve             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_curves);
  __Pyx_GIVEREF(__pyx_n_s_curves);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_curves);
  __Pyx_INCREF(__pyx_n_s_fixedStepCurve);
  __Pyx_GIVEREF(__pyx_n_s_fixedStepCurve);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_fixedStepCurve);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_curves, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_curves); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_curves, __pyx_t_1) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_fixedStepCurve); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fixedStepCurve, __pyx_t_1) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":28
 * # Multiprocessing
 * 
 * nCoresUse = 1             # <<<<<<<<<<<<<<
 * multiprocessingPool = None
 * nTreadFetch = 100000
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nCoresUse, __pyx_int_1) < 0) __PYX_ERR(0, 28, __pyx_L1_error)

  /* "gnocis/models.pyx":29
 * 
 * nCoresUse = 1
 * multiprocessingPool = None             # <<<<<<<<<<<<<<
 * nTreadFetch = 100000
 * maxThreadFetchNT = nTreadFetch * 1000
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_multiprocessingPool, Py_None) < 0) __PYX_ERR(0, 29, __pyx_L1_error)

  /* "gnocis/models.pyx":30
 * nCoresUse = 1
 * multiprocessingPool = None
 * nTreadFetch = 100000             # <<<<<<<<<<<<<<
 * maxThreadFetchNT = nTreadFetch * 1000
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nTreadFetch, __pyx_int_100000) < 0) __PYX_ERR(0, 30, __pyx_L1_error)

  /* "gnocis/models.pyx":31
 * multiprocessingPool = None
 * nTreadFetch = 100000
 * maxThreadFetchNT = nTreadFetch * 1000             # <<<<<<<<<<<<<<
 * 
 * def setNThreadFetch(v):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nTreadFetch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_int_1000); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_maxThreadFetchNT, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":33
 * maxThreadFetchNT = nTreadFetch * 1000
 * 
 * def setNThreadFetch(v):             # <<<<<<<<<<<<<<
 * 	global nTreadFetch
 * 	global maxThreadFetchNT
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_1setNThreadFetch, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setNThreadFetch, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":40
 * 
 * # Used for applying models in separate processes
 * def multiprocessModelDeployer(args):             # <<<<<<<<<<<<<<
 * 	cdef sequences seqs
 * 	cdef sequence seq, win
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_3multiprocessModelDeployer, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_multiprocessModelDeployer, __pyx_t_1) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":61
 * 	]
 * 
 * def initializePoolProcess(arg):             # <<<<<<<<<<<<<<
 * 	global processID, model
 * 	processID, model = None, None
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_5initializePoolProcess, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_initializePoolProcess, __pyx_t_1) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":65
 * 	processID, model = None, None
 * 
 * def setPoolProcessID(newID):             # <<<<<<<<<<<<<<
 * 	global processID
 * 	processID = newID
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_7setPoolProcessID, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setPoolProcessID, __pyx_t_1) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":70
 * 	return True
 * 
 * def getPoolProcessID(arg):             # <<<<<<<<<<<<<<
 * 	global processID
 * 	return processID
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_9getPoolProcessID, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getPoolProcessID, __pyx_t_1) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":74
 * 	return processID
 * 
 * def getPoolProcessModelID(arg):             # <<<<<<<<<<<<<<
 * 	global processID, model
 * 	if model == None:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_11getPoolProcessModelID, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getPoolProcessModelID, __pyx_t_1) < 0) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":80
 * 	return processID, '%d %s'%(model.modelID, str(model))
 * 
 * def initializePoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	while True:
 * 		pool.map( setPoolProcessID, list(range(len(pool._pool))) )
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_13initializePoolProcessIDs, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_initializePoolProcessIDs, __pyx_t_1) < 0) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":87
 * 			break
 * 
 * def getPoolProcessIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	while True:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_15getPoolProcessIDs, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getPoolProcessIDs, __pyx_t_1) < 0) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":94
 * 			return pIDs
 * 
 * def getPoolModelIDs(pool):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	mIDs = set()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_17getPoolModelIDs, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getPoolModelIDs, __pyx_t_1) < 0) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":104
 * 			return mIDs
 * 
 * def setPoolProcessModel(mdl):             # <<<<<<<<<<<<<<
 * 	global model
 * 	model = mdl
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_19setPoolProcessModel, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setPoolProcessModel, __pyx_t_1) < 0) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":109
 * 	return True
 * 
 * def getPoolProcessHasModel(arg):             # <<<<<<<<<<<<<<
 * 	global model, processID
 * 	return (processID, model != None)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_21getPoolProcessHasModel, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getPoolProcessHasModel, __pyx_t_1) < 0) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":113
 * 	return (processID, model != None)
 * 
 * def setPoolModel(pool, mdl):             # <<<<<<<<<<<<<<
 * 	pIDs = set()
 * 	if mdl == None:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_23setPoolModel, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setPoolModel, __pyx_t_1) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":137
 * 				break
 * 
 * def setNCores(n):             # <<<<<<<<<<<<<<
 * 	"""
 * 	Sets the number of cores to use for multiprocessing.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_25setNCores, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setNCores, __pyx_t_1) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":160
 * # Models
 * 
 * maxModelID = 0             # <<<<<<<<<<<<<<
 * 
 * # Represents sequence models.
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_maxModelID, __pyx_int_0) < 0) __PYX_ERR(0, 160, __pyx_L1_error)

  /* "gnocis/models.pyx":163
 * 
 * # Represents sequence models.
 * class sequenceModel:             # <<<<<<<<<<<<<<
 * 	"""
 * 	The `sequenceModel` class is an abstract class for sequence models. A number of methods are implemented for machine learning and prediction with DNA sequences.
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_latin1_sequenceModel, __pyx_n_s_sequenceModel, (PyObject *) NULL, __pyx_n_s_gnocis_models, __pyx_kp_s_The_sequenceModel_class_is_an_a); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gnocis/models.pyx":174
 * 	"""
 * 
 * 	def __init__(self, name, enableMultiprocessing = True):             # <<<<<<<<<<<<<<
 * 		self.name = name
 * 		self.enableMultiprocessing = enableMultiprocessing
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_1__init__, 0, __pyx_n_s_sequenceModel___init, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__62);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":179
 * 
 * 	# For pickling (used with multiprocessing)
 * 	def __getstate__(self):             # <<<<<<<<<<<<<<
 * 		return self.__dict__
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_3__getstate__, 0, __pyx_n_s_sequenceModel___getstate, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getstate, __pyx_t_2) < 0) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":182
 * 		return self.__dict__
 * 
 * 	def __setstate__(self, state):             # <<<<<<<<<<<<<<
 * 		self.__dict__ = state
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_5__setstate__, 0, __pyx_n_s_sequenceModel___setstate, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setstate, __pyx_t_2) < 0) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":186
 * 
 * 	# Gets a list of scores for a sequence list or stream. A stream is recommended when the sequence list is large, in order to avoid running out of memory.
 * 	def getSequenceScores(self, seqs):             # <<<<<<<<<<<<<<
 * 		""" Scores a set of sequences, returning the maximum window score for each. Multiprocessing is enabled by default, but can be disabled in the constructor.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_7getSequenceScores, 0, __pyx_n_s_sequenceModel_getSequenceScores, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getSequenceScores, __pyx_t_2) < 0) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":237
 * 
 * 	# Gets the threshold that gives optimal accuracy on a pair of lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getOptimalAccuracyThreshold(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value optimized for accuracy to a set of positive and a set of negative sequences.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_9getOptimalAccuracyThreshold, 0, __pyx_n_s_sequenceModel_getOptimalAccuracy_2, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults), 2)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_34__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getOptimalAccuracyThreshold, __pyx_t_2) < 0) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":272
 * 
 * 	# Gets the threshold that gives a desired precision for two lists or streams of sequences. Streams are recommended when the sequence lists are large, in order to avoid running out of memory.
 * 	def getPrecisionThreshold(self, positives, negatives, wantedPrecision):             # <<<<<<<<<<<<<<
 * 		""" Gets a threshold value for a desired precision to a set of positive and a set of negative sequences. Linear interpolation is used in order to achieve a close approximation.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_11getPrecisionThreshold, 0, __pyx_n_s_sequenceModel_getPrecisionThresh_2, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getPrecisionThreshold, __pyx_t_2) < 0) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":318
 * 
 * 	# Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 	def calibrateGenomewidePrecision(self, positives, genome, factor = 1.0, precision = 0.8, bgModelOrder = 4, bgModel = None):             # <<<<<<<<<<<<<<
 * 		""" Calibrates the model threshold for an expected precision genome-wide. Returns self to facilitate chaining of operations. However, this operation does mutate the model object. A scaling factor can be applied to the genome with the 'factor' argument. If, for instance, the positive set has been divided in half for independent training and calibration, a factor of 0.5 can be used.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_13calibrateGenomewidePrecision, 0, __pyx_n_s_sequenceModel_calibrateGenomewid_2, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__75);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_calibrateGenomewidePrecision, __pyx_t_2) < 0) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":356
 * 
 * 	# Applies model to a sequence with a sliding window, and returns the maximum score.
 * 	def scoreSequence(self, seq):             # <<<<<<<<<<<<<<
 * 		""" Scores a single sequence. The score is determined by applying the model with a sliding window, and taking the maximum score.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_15scoreSequence, 0, __pyx_n_s_sequenceModel_scoreSequence, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_scoreSequence, __pyx_t_2) < 0) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":379
 * 
 * 	# Gets and returns a confusion matrix.
 * 	def getConfusionMatrix(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a confusion matrix for sets of positive and negative sequences.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_17getConfusionMatrix, 0, __pyx_n_s_sequenceModel_getConfusionMatrix, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults1), 2)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_36__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getConfusionMatrix, __pyx_t_2) < 0) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":399
 * 
 * 	# Generates a Receiver Operating Characteristic curve.
 * 	def getROC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a Receiver Operating Characteristic (ROC) curve for sets of positive and negative sequences.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_19getROC, 0, __pyx_n_s_sequenceModel_getROC, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__81)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults2), 2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_38__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getROC, __pyx_t_2) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":419
 * 
 * 	# Calculates the area under a Receiver Operating Characteristic curve.
 * 	def getROCAUC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns the area under a Receiver Operating Characteristic (ROCAUC) curve for sets of positive and negative sequences.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_21getROCAUC, 0, __pyx_n_s_sequenceModel_getROCAUC, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__83)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults3), 2)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_40__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getROCAUC, __pyx_t_2) < 0) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":436
 * 
 * 	# Generates a Precision/Recall curve.
 * 	def getPRC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns a Precision/Recall curve (PRC) for sets of positive and negative sequences.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_23getPRC, 0, __pyx_n_s_sequenceModel_getPRC, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults4), 2)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_42__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getPRC, __pyx_t_2) < 0) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":456
 * 
 * 	# Calculates the area under a Precision/Recall Curve.
 * 	def getPRCAUC(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Calculates and returns the area under a Precision/Recall curve (PRCAUC) for sets of positive and negative sequences.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_25getPRCAUC, 0, __pyx_n_s_sequenceModel_getPRCAUC, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults5), 2)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_44__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getPRCAUC, __pyx_t_2) < 0) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":472
 * 		return getAUC(self.getPRC(seqs, labelPositive = labelPositive, labelNegative = labelNegative))
 * 
 * 	def plotPRC(self, seqs, labelPositive = positive, labelNegative = negative, figsize = (8, 8), outpath = None, style = 'ggplot'):             # <<<<<<<<<<<<<<
 * 		""" Plots a Precision/Recall curve and either displays it in an IPython session or saves it to a file.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_27plotPRC, 0, __pyx_n_s_sequenceModel_plotPRC, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__89)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults6), 2)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_46__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_plotPRC, __pyx_t_2) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":523
 * 
 * 	# Returns validation statistics dictionary.
 * 	def getValidationStatistics(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		""" Returns common model validation statistics (confusion matrix values; ROCAUC; PRCAUC).
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_29getValidationStatistics, 0, __pyx_n_s_sequenceModel_getValidationStati, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults7), 2)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_48__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getValidationStatistics, __pyx_t_2) < 0) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":548
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamRegions(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_31predictSequenceStreamRegions, 0, __pyx_n_s_sequenceModel_predictSequenceStr_2, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_predictSequenceStreamRegions, __pyx_t_2) < 0) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":595
 * 
 * 	# Predicts regions inside a sequence stream using a sliding window.
 * 	def predictSequenceStreamCurves(self, stream):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_33predictSequenceStreamCurves, 0, __pyx_n_s_sequenceModel_predictSequenceStr_3, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__95)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_predictSequenceStreamCurves, __pyx_t_2) < 0) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":656
 * 
 * 	# Predicts core regions.
 * 	def predictCore(self, genome):             # <<<<<<<<<<<<<<
 * 		""" Predicts cores. Not implemented by default. Models that use core prediction should implement this.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_35predictCore, 0, __pyx_n_s_sequenceModel_predictCore, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_predictCore, __pyx_t_2) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":668
 * 
 * 	# Short-hand for predictSequenceStreamRegions.
 * 	def predict(self, stream, curves = True):             # <<<<<<<<<<<<<<
 * 		""" Applies the model using a sliding window across an input sequence stream or sequence set. Windows with a score >= self.threshold are predicted, and predicted windows are merged into non-overlapping predictions.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_37predict, 0, __pyx_n_s_sequenceModel_predict, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__100);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_predict, __pyx_t_2) < 0) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":681
 * 
 * 	# Short-hand for getSequenceScores/scoreSequence
 * 	def score(self, target, asTable = False):             # <<<<<<<<<<<<<<
 * 		""" Scores a sequence of set of sequences.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_39score, 0, __pyx_n_s_sequenceModel_score, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__103);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_score, __pyx_t_2) < 0) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":709
 * 
 * 	# Short-hand for training
 * 	def train(self, ts):             # <<<<<<<<<<<<<<
 * 		""" Scores a sequence of set of sequences.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_41train, 0, __pyx_n_s_sequenceModel_train, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_train, __pyx_t_2) < 0) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":722
 * 
 * 	# Prints out test statistics.
 * 	def printTestStatistics(self, seqs, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		printValidationStatistics( self.getValidationStatistics(seqs, labelPositive = labelPositive, labelNegative = labelNegative) )
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_13sequenceModel_43printTestStatistics, 0, __pyx_n_s_sequenceModel_printTestStatistic, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__107)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults8), 2)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults8, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_50__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_printTestStatistics, __pyx_t_2) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":163
 * 
 * # Represents sequence models.
 * class sequenceModel:             # <<<<<<<<<<<<<<
 * 	"""
 * 	The `sequenceModel` class is an abstract class for sequence models. A number of methods are implemented for machine learning and prediction with DNA sequences.
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_latin1_sequenceModel, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_latin1_sequenceModel, __pyx_t_2) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":726
 * 
 * # Uniformly weighted model
 * class sequenceModelDummy(sequenceModel):             # <<<<<<<<<<<<<<
 * 	"""
 * 	This model takes a feature set, and scores input sequences by summing feature values, without weighting.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_latin1_sequenceModel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_latin1_sequenceModelDummy, __pyx_n_s_sequenceModelDummy, (PyObject *) NULL, __pyx_n_s_gnocis_models, __pyx_kp_s_This_model_takes_a_feature_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "gnocis/models.pyx":741
 * 	"""
 * 
 * 	def __init__(self, name, features, windowSize, windowStep):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18sequenceModelDummy_1__init__, 0, __pyx_n_s_sequenceModelDummy___init, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  PyList_Append(__pyx_t_4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_init, __pyx_t_5) < 0) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":747
 * 		self.windowSize, self.windowStep = windowSize, windowStep
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Dummy model<Features: %s>'%(str(self.features))
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18sequenceModelDummy_3__str__, 0, __pyx_n_s_sequenceModelDummy___str, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__111)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_str, __pyx_t_5) < 0) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":750
 * 		return 'Dummy model<Features: %s>'%(str(self.features))
 * 
 * 	def __repr__(self): return self.__str__()             # <<<<<<<<<<<<<<
 * 
 * 	def getTrainer(self):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18sequenceModelDummy_5__repr__, 0, __pyx_n_s_sequenceModelDummy___repr, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__113)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_repr, __pyx_t_5) < 0) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":752
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18sequenceModelDummy_7getTrainer, 0, __pyx_n_s_sequenceModelDummy_getTrainer, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__115)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_getTrainer, __pyx_t_5) < 0) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":755
 * 		return lambda ts: sequenceModelDummy(self.name, self.features, self.windowSize, self.windowStep)
 * 
 * 	def scoreWindow(self, seq):             # <<<<<<<<<<<<<<
 * 		cdef list fv
 * 		cdef float ret, v
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18sequenceModelDummy_9scoreWindow, 0, __pyx_n_s_sequenceModelDummy_scoreWindow, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__117)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_scoreWindow, __pyx_t_5) < 0) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gnocis/models.pyx":726
 * 
 * # Uniformly weighted model
 * class sequenceModelDummy(sequenceModel):             # <<<<<<<<<<<<<<
 * 	"""
 * 	This model takes a feature set, and scores input sequences by summing feature values, without weighting.
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_latin1_sequenceModelDummy, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_4, __pyx_t_5) < 0) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_latin1_sequenceModelDummy, __pyx_t_5) < 0) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":765
 * 
 * # Log-odds model
 * class sequenceModelLogOdds(sequenceModel):             # <<<<<<<<<<<<<<
 * 	"""
 * 	Constructs a log-odds model based on an input feature set and binary training set, and scores input sequences by summing log-odds-weighted feature values.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_latin1_sequenceModel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_latin1_sequenceModelLogOdds, __pyx_n_s_sequenceModelLogOdds, (PyObject *) NULL, __pyx_n_s_gnocis_models, __pyx_kp_s_Constructs_a_log_odds_model_bas); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "gnocis/models.pyx":786
 * 	"""
 * 
 * 	def __init__(self, name, features, trainingSet, windowSize, windowStep, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 		super().__init__(name)
 * 		self.threshold = 0.0
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_1__init__, 0, __pyx_n_s_sequenceModelLogOdds___init, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__119)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_t_4);
  PyList_Append(__pyx_t_5, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_4, sizeof(__pyx_defaults9), 2)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_positive); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_t_4)->__pyx_arg_labelPositive = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_negative); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_Defaults(__pyx_defaults9, __pyx_t_4)->__pyx_arg_labelNegative = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_4, __pyx_pf_6gnocis_6models_52__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_init, __pyx_t_4) < 0) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gnocis/models.pyx":807
 * 			self.sortedweights.append(weight)
 * 
 * 	def __str__(self):             # <<<<<<<<<<<<<<
 * 		return 'Log-odds model<Features: %s; Training set: %s; Positive label: %s; Negative label: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative))
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_3__str__, 0, __pyx_n_s_sequenceModelLogOdds___str, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__121)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_str, __pyx_t_4) < 0) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gnocis/models.pyx":810
 * 		return 'Log-odds model<Features: %s; Training set: %s; Positive label: %s; Negative label: %s>'%(str(self.features), str(self.trainingSet), str(self.labelPositive), str(self.labelNegative))
 * 
 * 	def __repr__(self): return self.__str__()             # <<<<<<<<<<<<<<
 * 
 * 	def getTrainer(self):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_5__repr__, 0, __pyx_n_s_sequenceModelLogOdds___repr, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__123)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_repr, __pyx_t_4) < 0) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gnocis/models.pyx":812
 * 	def __repr__(self): return self.__str__()
 * 
 * 	def getTrainer(self):             # <<<<<<<<<<<<<<
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_7getTrainer, 0, __pyx_n_s_sequenceModelLogOdds_getTrainer_2, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__125)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_getTrainer, __pyx_t_4) < 0) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gnocis/models.pyx":815
 * 		return lambda ts: sequenceModelLogOdds(self.name, self.features, ts, windowSize = self.windowSize, windowStep = self.windowStep, labelPositive = self.labelPositive, labelNegative = self.labelNegative)
 * 
 * 	def scoreWindow(self, seq):             # <<<<<<<<<<<<<<
 * 		cdef list fv, w
 * 		cdef int i
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_20sequenceModelLogOdds_9scoreWindow, 0, __pyx_n_s_sequenceModelLogOdds_scoreWindow, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__127)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_scoreWindow, __pyx_t_4) < 0) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gnocis/models.pyx":765
 * 
 * # Log-odds model
 * class sequenceModelLogOdds(sequenceModel):             # <<<<<<<<<<<<<<
 * 	"""
 * 	Constructs a log-odds model based on an input feature set and binary training set, and scores input sequences by summing log-odds-weighted feature values.
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_latin1_sequenceModelLogOdds, __pyx_t_1, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_CyFunction_InitClassCell(__pyx_t_5, __pyx_t_4) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_latin1_sequenceModelLogOdds, __pyx_t_4) < 0) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":826
 * 
 * # Trains a singular-motif PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainSinglePREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 	return sequenceModelLogOdds(name, features.motifSpectrum(motifs), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_positive); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__12 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_negative); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__13 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_27trainSinglePREdictorModel, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_trainSinglePREdictorModel, __pyx_t_1) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":830
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def createDummyPREdictorModel(name, motifs, windowSize=500, windowStep=10):             # <<<<<<<<<<<<<<
 * 	return sequenceModelDummy(name, motifs.pairFreq(219), windowSize, windowStep)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_29createDummyPREdictorModel, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_createDummyPREdictorModel, __pyx_t_1) < 0) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":834
 * 
 * # Trains a PREdictor model with a given set of motifs and positive and negative training sequences.
 * def trainPREdictorModel(name, motifs, trainingSet, windowSize=500, windowStep=10, labelPositive = positive, labelNegative = negative):             # <<<<<<<<<<<<<<
 * 	return sequenceModelLogOdds(name, features.motifPairSpectrum(motifs, 219), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_positive); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__14 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_negative); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__15 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_31trainPREdictorModel, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_trainPREdictorModel, __pyx_t_1) < 0) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":837
 * 	return sequenceModelLogOdds(name, features.motifPairSpectrum(motifs, 219), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 * class CVModelPredictions:             # <<<<<<<<<<<<<<
 * 	"""
 * 	Helper class that represents cross-validated model predictions..
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_latin1_CVModelPredictions_2, __pyx_n_s_CVModelPredictions_2, (PyObject *) NULL, __pyx_n_s_gnocis_models, __pyx_kp_s_Helper_class_that_represents_cr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gnocis/models.pyx":848
 * 	"""
 * 
 * 	def __init__(self, model, trainedModels = [], labelPredictionCurves = None):             # <<<<<<<<<<<<<<
 * 		self.name = model.name
 * 		self.model = model
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18CVModelPredictions_1__init__, 0, __pyx_n_s_CVModelPredictions___init, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__135)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults10), 1)) __PYX_ERR(0, 848, __pyx_L1_error)
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults10, __pyx_t_2)->__pyx_arg_trainedModels = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_54__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":856
 * 		self.hasCores = False
 * 
 * 	def regions(self, label):             # <<<<<<<<<<<<<<
 * 		if label in self.predictedRegions:
 * 			return self.predictedRegions[label]
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18CVModelPredictions_3regions, 0, __pyx_n_s_CVModelPredictions_regions, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__137)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_regions, __pyx_t_2) < 0) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":861
 * 		return [ c.regions() for c in self.predictionCurves[label] ]
 * 
 * 	def setCorePredictions(self, label, cores):             # <<<<<<<<<<<<<<
 * 		self.predictedRegions[label] = cores
 * 		self.hasCores = True
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18CVModelPredictions_5setCorePredictions, 0, __pyx_n_s_CVModelPredictions_setCorePredic, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__139)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setCorePredictions, __pyx_t_2) < 0) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":865
 * 		self.hasCores = True
 * 
 * 	def predictCore(self, genome):             # <<<<<<<<<<<<<<
 * 		task = progressTask('Predicting cores', steps = len(self.trainedModels))
 * 		cores = []
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18CVModelPredictions_7predictCore, 0, __pyx_n_s_CVModelPredictions_predictCore, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__141)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_predictCore, __pyx_t_2) < 0) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":877
 * 		task.done()
 * 
 * 	def getStatTable(self):             # <<<<<<<<<<<<<<
 * 		rtable = []
 * 		for label in self.predictionCurves:
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18CVModelPredictions_9getStatTable, 0, __pyx_n_s_CVModelPredictions_getStatTable, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__143)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getStatTable, __pyx_t_2) < 0) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":890
 * 		return nctable('Predictions: ' + self.model.name, rtable)
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18CVModelPredictions_11__repr__, 0, __pyx_n_s_CVModelPredictions___repr, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__145)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_repr, __pyx_t_2) < 0) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":893
 * 		return 'CVModelPredictions<' + str(self.model) + '>' + self.getStatTable().__repr__()
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		return '<div>' + self.getStatTable()._repr_html_() + '</div>'
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_18CVModelPredictions_13_repr_html_, 0, __pyx_n_s_CVModelPredictions__repr_html, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__147)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_repr_html, __pyx_t_2) < 0) __PYX_ERR(0, 893, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":837
 * 	return sequenceModelLogOdds(name, features.motifPairSpectrum(motifs, 219), trainingSet, windowSize, windowStep, labelPositive = labelPositive, labelNegative = labelNegative)
 * 
 * class CVModelPredictions:             # <<<<<<<<<<<<<<
 * 	"""
 * 	Helper class that represents cross-validated model predictions..
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_latin1_CVModelPredictions_2, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_latin1_CVModelPredictions_2, __pyx_t_2) < 0) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":896
 * 		return '<div>' + self.getStatTable()._repr_html_() + '</div>'
 * 
 * class crossvalidation:             # <<<<<<<<<<<<<<
 * 	"""
 * 	Helper class for cross-validations. Accepts binary training and validation sets and constructs cross-validation sets for a desired number of repeats. If a separate validation set is not given, the training set is used. The cross-validation set for each repeat contains numbers of training and test sequences determined by a training-to-testing sequence ratio, as well as a negative-per-positive test sequence ratio. When constructing the validation set, identities are checked for against the training set, to avoid contamination (will not work if sequences are cloned). Holds models and validation statistics. Integrates with terminal and IPython for visualization of results. Stores training and testing data, so new models can be added.
 */
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_latin1_crossvalidation, __pyx_n_s_crossvalidation, (PyObject *) NULL, __pyx_n_s_gnocis_models, __pyx_kp_s_Helper_class_for_cross_validati); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gnocis/models.pyx":919
 * 	"""
 * 
 * 	def __init__(self, models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = None, storeTrained = True):             # <<<<<<<<<<<<<<
 * 		self.models = []
 * 		self.PRC = {}
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_1__init__, 0, __pyx_n_s_crossvalidation___init, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__149)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults11), 2)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_negative); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults11, __pyx_t_2)->__pyx_arg_labelNegative = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_56__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":986
 * 	:type mdl: model
 * 	"""
 * 	def addModel(self, mdl):             # <<<<<<<<<<<<<<
 * 		self.PRC[mdl] = []
 * 		self.ROC[mdl] = []
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_3addModel, 0, __pyx_n_s_crossvalidation_addModel, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__151)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_addModel, __pyx_t_2) < 0) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1003
 * 		task.done()
 * 
 * 	def calibrate(self, calibrator, labelPositive = positive):             # <<<<<<<<<<<<<<
 * 		mtask = progressTask('Calibrating models', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_5calibrate, 0, __pyx_n_s_crossvalidation_calibrate, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__153)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_2, sizeof(__pyx_defaults12), 1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_positive); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_Defaults(__pyx_defaults12, __pyx_t_2)->__pyx_arg_labelPositive = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_2, __pyx_pf_6gnocis_6models_58__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_calibrate, __pyx_t_2) < 0) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1020
 * 		mtask.done()
 * 
 * 	def predict(self, genome):             # <<<<<<<<<<<<<<
 * 		mtask = progressTask('Predicting genome-wide', steps = len(self.models))
 * 		for imdl, mdl in enumerate(self.models):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_7predict, 0, __pyx_n_s_crossvalidation_predict, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__155)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_predict, __pyx_t_2) < 0) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1044
 * 		return [ self.predictions[mdl] for mdl in self.models ]
 * 
 * 	def plotPRC(self, figsize = (9, 9), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_9plotPRC, 0, __pyx_n_s_crossvalidation_plotPRC, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__157)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__158);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_plotPRC, __pyx_t_2) < 0) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1133
 * 			raise err
 * 
 * 	def plotROC(self, figsize = (8, 8), outpath = None, style = 'ggplot', returnHTML = False, fontsizeLabels = 18, fontsizeLegend = 12, fontsizeAxis = 10, bboxAnchorTo = (0., -0.15), legendLoc = 'upper left'):             # <<<<<<<<<<<<<<
 * 		try:
 * 			import matplotlib.pyplot as plt
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_11plotROC, 0, __pyx_n_s_crossvalidation_plotROC, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__161);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_plotROC, __pyx_t_2) < 0) __PYX_ERR(0, 1133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1220
 * 			raise err
 * 
 * 	def getConfigurationTable(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Configuration',
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_13getConfigurationTable, 0, __pyx_n_s_crossvalidation_getConfiguration, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__163)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getConfigurationTable, __pyx_t_2) < 0) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1234
 * 		).noCropNames()
 * 
 * 	def getAUCTable(self):             # <<<<<<<<<<<<<<
 * 		return nctable(
 * 			'Evaluation statistics',
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_15getAUCTable, 0, __pyx_n_s_crossvalidation_getAUCTable, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__165)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getAUCTable, __pyx_t_2) < 0) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1255
 * 			})
 * 
 * 	def __repr__(self):             # <<<<<<<<<<<<<<
 * 		hdr = 'Cross-validation\n'
 * 		config = self.getConfigurationTable()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_17__repr__, 0, __pyx_n_s_crossvalidation___repr, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__167)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_repr, __pyx_t_2) < 0) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":1261
 * 		return hdr + '\n' + config.__repr__() + '\n' + t.__repr__()
 * 
 * 	def _repr_html_(self):             # <<<<<<<<<<<<<<
 * 		hdr = '<div><b>Cross-validation</b></div>'
 * 		config = self.getConfigurationTable()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6gnocis_6models_15crossvalidation_19_repr_html_, 0, __pyx_n_s_crossvalidation__repr_html, NULL, __pyx_n_s_gnocis_models, __pyx_d, ((PyObject *)__pyx_codeobj__169)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_repr_html, __pyx_t_2) < 0) __PYX_ERR(0, 1261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gnocis/models.pyx":896
 * 		return '<div>' + self.getStatTable()._repr_html_() + '</div>'
 * 
 * class crossvalidation:             # <<<<<<<<<<<<<<
 * 	"""
 * 	Helper class for cross-validations. Accepts binary training and validation sets and constructs cross-validation sets for a desired number of repeats. If a separate validation set is not given, the training set is used. The cross-validation set for each repeat contains numbers of training and test sequences determined by a training-to-testing sequence ratio, as well as a negative-per-positive test sequence ratio. When constructing the validation set, identities are checked for against the training set, to avoid contamination (will not work if sequences are cloned). Holds models and validation statistics. Integrates with terminal and IPython for visualization of results. Stores training and testing data, so new models can be added.
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_latin1_crossvalidation, __pyx_empty_tuple, __pyx_t_1, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_latin1_crossvalidation, __pyx_t_2) < 0) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":1271
 * 			) + '</div>'
 * 
 * def crossvalidate(models, trainingSet, validationSet = None, labelPositive = positive, labelNegative = negative, repeats = 20, ratioTrainTest = 0.8, ratioNegPos = 100., storeTrained = True):             # <<<<<<<<<<<<<<
 * 	return crossvalidation(models = models, trainingSet = trainingSet, validationSet = validationSet, labelPositive = labelPositive, labelNegative = labelNegative, repeats = repeats, ratioTrainTest = ratioTrainTest, ratioNegPos = ratioNegPos, storeTrained = storeTrained)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_positive); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__31 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_negative); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__32 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6gnocis_6models_33crossvalidate, NULL, __pyx_n_s_gnocis_models); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_crossvalidate, __pyx_t_1) < 0) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gnocis/models.pyx":1
 * # distutils: language=c++             # <<<<<<<<<<<<<<
 * # -*- coding: latin-1 -*-
 * ############################################################################
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init gnocis.models", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init gnocis.models");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObjectFormat */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* obj, PyObject* format_spec) {
    int ret;
    _PyUnicodeWriter writer;
    if (likely(PyFloat_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyFloat_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else if (likely(PyLong_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyLong_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else {
        return PyObject_Format(obj, format_spec);
    }
    if (unlikely(ret == -1)) {
        _PyUnicodeWriter_Dealloc(&writer);
        return NULL;
    }
    return _PyUnicodeWriter_Finish(&writer);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    (void)is_set;
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        (void)orig_length;
        (void)ppos;
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyFloatBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_EqObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a;
    (void)inplace;
    (void)zerodivision_check;
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    if (likely(PyFloat_CheckExact(op1))) {
        a = PyFloat_AS_DOUBLE(op1);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        switch (size) {
            case  0: a = 0.0; break;
            case -1: a = -(double) digits[0]; break;
            case  1: a = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            return (
                PyFloat_Type.tp_richcompare(op2, op1, Py_EQ));
        }
    } else {
        return (
            PyObject_RichCompare(op1, op2, Py_EQ));
    }
        if (a == b) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
}
#endif

/* PyFloatBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyFloat_AddObjC(PyObject *op1, PyObject *op2, double floatval, int inplace, int zerodivision_check) {
    const double b = floatval;
    double a, result;
    (void)inplace;
    (void)zerodivision_check;
    if (likely(PyFloat_CheckExact(op1))) {
        a = PyFloat_AS_DOUBLE(op1);
        
    } else
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        a = (double) PyInt_AS_LONG(op1);
        
    } else
    #endif
    if (likely(PyLong_CheckExact(op1))) {
        #if CYTHON_USE_PYLONG_INTERNALS
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        switch (size) {
            case  0: a = 0.0; break;
            case -1: a = -(double) digits[0]; break;
            case  1: a = (double) digits[0]; break;
            case -2:
            case 2:
                if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (1 * PyLong_SHIFT < 53))) {
                    a = (double) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -2)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -3:
            case 3:
                if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (2 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -3)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            case -4:
            case 4:
                if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT && ((8 * sizeof(unsigned long) < 53) || (3 * PyLong_SHIFT < 53))) {
                    a = (double) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if ((8 * sizeof(unsigned long) < 53) || (4 * PyLong_SHIFT < 53) || (a < (double) ((PY_LONG_LONG)1 << 53))) {
                        if (size == -4)
                            a = -a;
                        break;
                    }
                }
                CYTHON_FALLTHROUGH;
            default:
        #else
        {
        #endif
            a = PyLong_AsDouble(op1);
            if (unlikely(a == -1.0 && PyErr_Occurred())) return NULL;
            
        }
    } else {
        return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
    }
        
        PyFPE_START_PROTECT("add", return NULL)
        result = a + b;
        PyFPE_END_PROTECT(result)
        return PyFloat_FromDouble(result);
}
#endif

/* PyObjectCallMethod1 */
    static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
    static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* SliceObject */
    static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* SliceTupleAndList */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject** CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyList_New(0);
    dest = PyList_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyListObject*)src)->ob_item + start,
        ((PyListObject*)dest)->ob_item,
        length);
    return dest;
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyTuple_New(0);
    dest = PyTuple_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyTupleObject*)src)->ob_item + start,
        ((PyTupleObject*)dest)->ob_item,
        length);
    return dest;
}
#endif

/* PyObject_GenericGetAttrNoDict */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* TypeImport */
    #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
    static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CalculateMetaclass */
    static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
    static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CyFunctionClassCell */
    static int __Pyx_CyFunction_InitClassCell(PyObject *cyfunctions, PyObject *classobj) {
    Py_ssize_t i, count = PyList_GET_SIZE(cyfunctions);
    for (i = 0; i < count; i++) {
        __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *)
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyList_GET_ITEM(cyfunctions, i);
#else
            PySequence_ITEM(cyfunctions, i);
        if (unlikely(!m))
            return -1;
#endif
        Py_INCREF(classobj);
        m->func_classobj = classobj;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF((PyObject*)m);
#endif
    }
    return 0;
}

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
    #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
    static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
    static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
    #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CoroutineBase */
    #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
    static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
    #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
    static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
    static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
    #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
    static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
